// DO NOT EDIT -- this file is generated automatically with langen


#define NUMEXTRA 137
const char* natchars[NUMEXTRA] = {"°","´","Á","Ä","Ç","È","É","Í","Î","Ö","Ú","Ü","ß","à","á","â","ã","ä","ç","è","é","ê","ì","í","î","ï","ò","ó","ô","õ","ö","ø","ù","ú","ü","ý","ą","ć","Č","č","Ď","ď","Ę","ę","Ě","ě","ğ","İ","ı","Ł","ł","ń","ň","Ř","ř","Ś","ś","Ş","ş","Š","š","ť","ů","Ź","ź","Ż","ż","Ž","ž","δ","ϕ","Ё","А","Б","В","Г","Д","Е","Ж","З","И","Й","К","Л","М","Н","О","П","Р","С","Т","У","Ф","Х","Ц","Ч","Ш","Щ","Ы","Ь","Э","Ю","Я","а","б","в","г","д","е","ж","з","и","й","к","л","м","н","о","п","р","с","т","у","ф","х","ц","ч","ш","щ","ъ","ы","ь","э","ю","я","ё","ᵈ",};
//javastring = "°´ÁÄÇÈÉÍÎÖÚÜßàáâãäçèéêìíîïòóôõöøùúüýąćČčĎďĘęĚěğİıŁłńňŘřŚśŞşŠšťůŹźŻżŽžδϕЁАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюяёᵈ";

int transcompleteness[NUMLAN] = {2052, 2052, 1228, 2052, 2052, 1554, 384, };

//statistics
// total:   386 nouns,  1680 sentences
// EN: 16256B nouns, 103182B sentences
// PL: 19965B nouns, 106623B sentences
// TR: 15818B nouns, 58599B sentences
// CZ: 20447B nouns, 115277B sentences
// RU: 35423B nouns, 174311B sentences
// DE: 17019B nouns, 81809B sentences
// PT-BR:  3881B nouns, 12159B sentences
// check hash: 6b8b4567
hashcode hashval = 0x6b8b4567;

sentence all_sentences[1680] = {
/*MISSING*/   {0x0, { // ""
/*MISSING*/    "",
/*MISSING*/    "",
/*MISSING*/    "",
/*MISSING*/    "",
/*MISSING*/    "",
/*MISSING*/    "",
    }},
  {0x314db7, { // "You cannot go against the wind!"
   "Nie możesz iść pod wiatr!",
/*MISSING*/    "You cannot go against the wind!",
   "Nemůžeš jít proti větru!",
   "Nie możesz iść pod wiatr!",
   "Du kannst nicht gegen den Wind laufen!",
/*MISSING*/    "You cannot go against the wind!",
    }},
  {0xbe8a00, { // "No summoning possible here!"
   "Przywołanie niemożliwe!",
   "Çağırmak burada mümkün değil!",
   "Sem se nic vyvolat nedá!",
   "Призыв невозможен!",
   "Beschwören hier nicht möglich!",
/*MISSING*/    "No summoning possible here!",
    }},
  {0xdae76d, { // "A creepy monster who follows you everywhere in the Graveyard."
   "Ten odrażający potwór chodzi za Tobą po cmentarzu!",
   "Seni Mezarlık boyunca takip eden ürpertici bir canavar.",
   "Děsivý netvor, který tě sleduje po celém Hřbitově.",
   "Пугающий монстр, вечно преследующий Вас на Кладбище.",
   "Ein gruseliges Monster das dir über den ganzen Friedhof folgt!",
/*MISSING*/    "A creepy monster who follows you everywhere in the Graveyard.",
    }},
  {0xe9a6b0, { // "local highscores"
   "lokalne wyniki",
   "yerel yüksek skorlar",
   "lokální hi-score",
   "локальные рекорды",
   "Lokale Highscores",
/*MISSING*/    "local highscores",
    }},
  {0x134fcc9, { // "Are there any magical orbs in %the1?..."
   "Czy %abl1 są jakieś magiczne sfery?",
   "%abl1 başka sihirli küre var mı?",
   "Jsou %abl1 nějaké magické sféry?",
   "Есть ли какие-нибудь сферы %abl1?",
   "Gibt es irgendwelche Orbs in %dem1 %d1?",
   "Tem algum orbe mágico %l1 %1?",
    }},
  {0x1740be6, { // "autochoose"
   "autowybór",
/*MISSING*/    "autochoose",
   "autovýbìr",
   "автовыбор",
/*MISSING*/    "autochoose",
/*MISSING*/    "autochoose",
    }},
  {0x18e3abf, { // "\n\nHint: use 'm' to toggle cells quickly"
   "\n\nWsk: użyj 'm' by szybko przestawiać pola",
/*MISSING*/    "\n\nHint: use 'm' to toggle cells quickly",
   "\n\nTip: políèka lze rychle pøepínat klávesou 'm'",
   "\n\nПодсказка: используй 'm', чтобы быстро переключать клетки",
/*MISSING*/    "\n\nHint: use 'm' to toggle cells quickly",
/*MISSING*/    "\n\nHint: use 'm' to toggle cells quickly",
    }},
  {0x1988dec, { // "ESC = return to the game"
   "ESC = powrót do gry",
   "ESC = oyuna geri dön",
   "ESC = návrat do hry",
   "ESC = вернуться в игру",
   "ESC = Zurück zum Spiel",
/*MISSING*/    "ESC = return to the game",
    }},
  {0x231516d, { // "%The1 tricks %the2."
   "%1 pokona%ł1 %a2.",
   "%1 %a2 kandırdı.",
   "%1 přelsti%l1 %a2.",
   "%1 победил%E1 %a2.",
   "%Der1 %1 überlistet %den2 %a2.",
/*MISSING*/    "%The1 tricks %the2.",
    }},
  {0x28f16c6, { // "SORT"
   "SORT",
   "SIRALA",
   "ŘAZ",
   "СОРТИРОВКА",
   "SORT",
   "ORGANIZAR",
    }},
#if REPEATED
  {0x28f16c7, { // "SORT [repeat]"
   "SORT",
/*MISSING*/    "SORT [repeat]",
   "TŘÍDIT",
   "СОРТИРУЙ",
/*MISSING*/    "SORT [repeat]",
/*MISSING*/    "SORT [repeat]",
    }},
#endif
  {0x2a95fec, { // "save the game"
   "zapisz i zakończ",
   "oyunu kaydet",
   "uložit a ukončit hru",
   "сохранить игру",
   "Speichern",
/*MISSING*/    "save the game",
    }},
  {0x2fc1d77, { // "This forest was planted by one of the wizards from the Ivory Tower to conduct experiments with gravity."
   "Ten las został zasadzony przez jednego z czarodziejów z Wieży z Kości Słoniowej, by przeprowadzać eksperymenty z grawitacją.",
/*MISSING*/    "This forest was planted by one of the wizards from the Ivory Tower to conduct experiments with gravity.",
   "Tento les vysadil jeden z čarodějů z Věže ze slonové kosti, aby tu prováděl pokusy s gravitací.",
   "Этот лес был создан одним из магов Башни из Слоновой Кости для продолжения экспериментов с гравитацией.",
   "Dieser Wald wurde von einem der Zauberer des Elfenbeinturmes gepflanzt, um Experimente mit der Graviation durchzuführen.",
/*MISSING*/    "This forest was planted by one of the wizards from the Ivory Tower to conduct experiments with gravity.",
    }},
  {0x35292aa, { // "\nSpawn rate (in Hubs): %1%/%2\n"
   "\nCzęstość występowania (w centrach): %1%/%2\n",
/*MISSING*/    "\nSpawn rate (in Hubs): %1%/%2\n",
   "\nFrekvence (v centrálních krajích): %1%/%2\n",
   "\nЧастота появления (в Центрах): %1%/%2\n",
/*MISSING*/    "\nSpawn rate (in Hubs): %1%/%2\n",
/*MISSING*/    "\nSpawn rate (in Hubs): %1%/%2\n",
    }},
  {0x3762af5, { // "This Orb allows you to grow like an Ivy. The Ivy is always rooted in your current location; moving among the Ivy cells will move the root. Moving to a new location will cause the Ivy to grow , if an Ivy could make that movement (otherwise it breaks). You can also target one of the cells adjacent to your ivy (not to you) to grow or attack there."
   "Ta Sfera pozwala Ci rosnąć jak bluszcz. Bluszcz jest zawsze ukorzeniony na Twojej pozycji; ruch w obrębie komórek bluszcza powoduje przemieszczenie korzenia. Bluszcz rośnie gdy ruszasz się na nową komórkę (o ile bluszcz mógłby tam urosnąć, w przeciwnym przypadku bluszcz się łamie). Możesz również wycelować w jedną z komórek sąsiadujących z Twoim bluszczem (nie z Tobą), by tam urosnąć lub zaatakować.",
/*MISSING*/    "This Orb allows you to grow like an Ivy. The Ivy is always rooted in your current location; moving among the Ivy cells will move the root. Moving to a new location will cause the Ivy to grow , if an Ivy could make that movement (otherwise it breaks). You can also target one of the cells adjacent to your ivy (not to you) to grow or attack there.",
   "Tato Sféra ti umožňuje růst, jako bys byl Břečťan. Kořen Břečťanu je vždy na políčku, kde se právě nacházíš; pohybem mezi políčky Břečťanu se tento kořen přesune. Pokud se pohneš na nové místo, Břečťan tam vyroste, ale jen pokud by normální Břečťan mohl provést takový tah (jinak se přetrhne). Můžeš také zacílovat jedno z políček, které sousedí se tvým Břečťanem (ale ne s tebou), a rozrůst se tam nebo tam zaútočit.",
   "Эта сфера позволяет тебе расти как плющ. Корень плюща находится в твоей позиции; движение по клеткам плюща меняет корень. При движении в новую клетку плющ растёт, если может, иначе рвётся. Также вы можете выбрать клетку рядом со своим плющом, чтобы расти или атаковать туда.",
/*MISSING*/    "This Orb allows you to grow like an Ivy. The Ivy is always rooted in your current location; moving among the Ivy cells will move the root. Moving to a new location will cause the Ivy to grow , if an Ivy could make that movement (otherwise it breaks). You can also target one of the cells adjacent to your ivy (not to you) to grow or attack there.",
/*MISSING*/    "This Orb allows you to grow like an Ivy. The Ivy is always rooted in your current location; moving among the Ivy cells will move the root. Moving to a new location will cause the Ivy to grow , if an Ivy could make that movement (otherwise it breaks). You can also target one of the cells adjacent to your ivy (not to you) to grow or attack there.",
    }},
  {0x38a34b8, { // "connected to Google Games"
   "połączenie z Google Games",
/*MISSING*/    "connected to Google Games",
   "připojeno ke Google Games",
   "подключено к Google Games",
   "verbunden mit Google Games",
/*MISSING*/    "connected to Google Games",
    }},
  {0x3c8943c, { // "%The1 sinks!"
   "%1 uton%ął1!",
   "%1 battı!",
   "%1 zmize%l1 pod vodou!",
   "%1 утонул%E1!",
   "%Der1 %1 sinkt!",
/*MISSING*/    "%The1 sinks!",
    }},
  {0x3f4b7d5, { // "A predator native to the Zebra."
   "Drapieżnik żyjący w krainie Zebry.",
   "Zebra'da yaşayan bir vahşi hayvan.",
   "Predátor pocházející ze Zebry.",
   "Типичный хищник этой земли.",
   "Ein Raubtier das im Zebra heimisch ist.",
/*MISSING*/    "A predator native to the Zebra.",
    }},
  {0x42913a9, { // "This orb can be used to invoke the lightning spell, which causes lightning bolts to shoot from you in all directions."
   "Ta sfera pozwala rzucić czar Błyskawica, który powoduje wystrzelenie błyskawicy w każdym kierunku.",
   "Bu küre şimşek büyüsünü kullanmaya yarar, senden her yöne doğru kıvılcımlar saçılır.",
   "Tuto sféru můžeš použít k seslání kouzla Blesk, které způsobí, že z tebe do všech směrů vyletí blesky.",
   "Эта сфера может использоваться для заклинания Молнии, которое бьёт молниями во всех направлениях.",
   "Du kannst diesen Orb verwenden um einen Blitzzauber zu wirken, der Blitze von deiner Position aus in sämtliche Richtungen schießt.",
/*MISSING*/    "This orb can be used to invoke the lightning spell, which causes lightning bolts to shoot from you in all directions.",
    }},
  {0x43fda2b, { // "distance: %1\n"
   "odległość: %1\n",
   "uzaklık: %1\n",
   "vzdálenost: %1\n",
   "расстояние: %1\n",
   "Distanz: %1\n",
/*MISSING*/    "distance: %1\n",
    }},
  {0x45936cc, { // "\n\nIn the Orb Strategy Mode, Orbs of Safety can be gained by collecting Phoenix Feathers in the Land of Eternal Motion. You can also find unlimited Orbs of Safety in the Crossroads and the Ocean (after collecting 25 Phoenix Feathers) and in the Prairie."
   "\n\nW trybie strategii sfer Sfery Bezpieczeństwa mogą być zdobyte przez zbieranie Piór Feniksa w Krainie Wiecznego Ruchu. Można też znaleźć nieograniczone Sfery Bezpieczeństwa na Skrzyżowaniach (po zdobyciu 25 Piór Feniksa) i na Prerii.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, Orbs of Safety can be gained by collecting Phoenix Feathers in the Land of Eternal Motion. You can also find unlimited Orbs of Safety in the Crossroads and the Ocean (after collecting 25 Phoenix Feathers) and in the Prairie.",
   "\n\nVe sférostrategickém módu mùžeš získat Sféry Bezpeènosti sbíráním Fénixových pírek v Kraji vìèného pohybu. Také jich mùžeš najít neomezenì mnoho na Køižovatce a v Oceánu (ale musíš nejprve získat 25 Fénixových pírek) a na Prérii.",
   "\n\nВ режиме стратегии сфер Сферу безопасности можно получить, Перья феникса в Земле вечного движения. Также можно найти неограниченные Сферы вечного движения на Перекрёстке и в Океане (когда соберёшь 25 Перьев феникса) и в Прерии.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, Orbs of Safety can be gained by collecting Phoenix Feathers in the Land of Eternal Motion. You can also find unlimited Orbs of Safety in the Crossroads and the Ocean (after collecting 25 Phoenix Feathers) and in the Prairie.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, Orbs of Safety can be gained by collecting Phoenix Feathers in the Land of Eternal Motion. You can also find unlimited Orbs of Safety in the Crossroads and the Ocean (after collecting 25 Phoenix Feathers) and in the Prairie.",
    }},
  {0x4b05b05, { // "polygonal"
   "wielokąt",
/*MISSING*/    "polygonal",
   "polygonální",
   "многоугольник",
   "polygonal",
/*MISSING*/    "polygonal",
    }},
  {0x50900d1, { // "Level of water bottom"
   "Poziom dna wody",
/*MISSING*/    "Level of water bottom",
   "Úroveň dna",
   "Уровень дна водоёма",
/*MISSING*/    "Level of water bottom",
/*MISSING*/    "Level of water bottom",
    }},
  {0x50fa0ab, { // "The floor has collapsed! RUN!"
   "Ziemia się zapadła! UCIEKAJ!",
   "Zemin çöktü! KAÇ!",
   "Podlaha se propadla! UTÍKEJ!",
   "Пол рухнул! БЕГИТЕ!",
   "Der Boden bricht zusammen - RENN!",
   "O chão desabou! CORRA!",
    }},
  {0x51bd008, { // "d = delete v"
   "d = skasuj v",
   "d = sil v",
   "d = vymaž v",
   "d = удалить v",
   "d = entfernen v",
/*MISSING*/    "d = delete v",
    }},
  {0x58fe401, { // "Help texts disabled."
   "Teksty pomocy wyłączone.",
/*MISSING*/    "Help texts disabled.",
   "Texty nápovědy vypnuty.",
   "Тексты помощи выключены.",
/*MISSING*/    "Help texts disabled.",
/*MISSING*/    "Help texts disabled.",
    }},
  {0x5a29c04, { // "reverse pointer control"
   "odwrotne sterowanie",
/*MISSING*/    "reverse pointer control",
   "obrácené ovládání ukazatele",
   "обратить управление",
/*MISSING*/    "reverse pointer control",
/*MISSING*/    "reverse pointer control",
    }},
  {0x5e2554f, { // "Orb power gained!"
   "Zdobyta moc!",
   "Küre gücü kazandın!",
   "Získa%l0 jsi moc sféry!",
   "Сила сферы получена!",
   "Orb Kraft erhalten!",
   "Poder do Orbe obtido!",
    }},
  {0x5ee6b22, { // "dress color"
   "kolor sukienki",
   "kıyafet rengi",
   "barva šatů",
   "цвет одежды",
   "Kleidungsfarbe",
/*MISSING*/    "dress color",
    }},
  {0x62ece6d, { // "select this color"
   "wybierz ten kolor",
/*MISSING*/    "select this color",
   "vyber tuto barvu",
   "wybierz ten kolor",
/*MISSING*/    "select this color",
/*MISSING*/    "select this color",
    }},
  {0x633ab9b, { // "which may not some day be applied to phenomena of the real world.\""
   "zastosować do zjawisk w świecie rzeczywistym.\"",
/*MISSING*/    "which may not some day be applied to phenomena of the real world.\"",
   "které by jednoho dne nemohlo být možné aplikovat na jevy ve skutečném světě.\"",
   "которая когда-нибудь не окажется применимой к явлениям действительного мира.\"",
/*MISSING*/    "which may not some day be applied to phenomena of the real world.\"",
/*MISSING*/    "which may not some day be applied to phenomena of the real world.\"",
    }},
  {0x64978a2, { // "The forests of Caribbean are too dense to be traversed by humans, and they are hard to burn. Many colorful parrots can be found there."
   "Lasy na Karaibach są zbyt gęste, by mógł przez nie przejść człowiek i trudno jest je spalić. Mieszka w nich dużo kolorowych papug.",
   "Karayip Ormanları insanlar tarafından gezilmek için fazla yoğun, ve yanmaya da dirençli. Buralarda bir sürü, renkli papağana rastlayabilirsin.",
   "Karibské lesy jsou příliš husté na to, aby jimi mohl projít člověk, a špatně hoří. Lze v nich najít mnoho barevných papoušků.",
   "Карибские леса слишком плотны, чтобы сквозь них мог пройти человек, и их трудно поджечь. Много разноуветных попугаев живёт здесь.",
   "Die karibischen Wälder sind zu dicht um von Menschen durchquert zu werden, und brennen nur schwerlich. Es gibt hier aber viele, farbenfrohe Papageien.",
/*MISSING*/    "The forests of Caribbean are too dense to be traversed by humans, and they are hard to burn. Many colorful parrots can be found there.",
    }},
  {0x6a23894, { // "You chop down the tree."
   "Ści%ąłeś0 drzewo.",
   "Ağacı kestin.",
   "Pokáce%l0 jsi strom.",
   "Вы срубили дерево.",
   "Du fällst den Baum.",
   "Você corta a árvore.",
    }},
  {0x6bdb7af, { // "This has been recorded in the Google Leaderboards."
   "To zostało zapisane w Rankingach Google.",
   "Bu Google Listelerine kaydedildi.",
   "Bylo zaznamenáno na žebříčky Googleu.",
   "Это было записано в рейтинг Google.",
   "Dies wurde in die Google Leaderboards aufgenommen.",
/*MISSING*/    "This has been recorded in the Google Leaderboards.",
    }},
  {0x6fdb6be, { // "texture size"
   "rozmiar tekstury",
/*MISSING*/    "texture size",
   "rozměr textury",
   "размер текстур",
   "Texturgröße",
/*MISSING*/    "texture size",
    }},
  {0x701c698, { // " [%1 turns]"
   " [kolejek: %1]",
   " [Tur: %1]",
   " [počet kol: %1]",
   " [ходов: %1]",
   " [%1 Züge]",
   " [%1 turnos]",
    }},
  {0x7066813, { // "band width"
   "szerokość wstęgi",
/*MISSING*/    "band width",
   "šířka pásu",
   "ширина ленты",
   "Bandbreite",
/*MISSING*/    "band width",
    }},
  {0x73c4cb8, { // "Cannot throw fire there!"
   "Nie możesz tego podpalić!",
   "Oraya ateş atamazsın!",
   "Sem nemůžeš vrhnout oheň!",
   "Nie możesz tego podpalić!",
   "Du kannst dort kein Feuer hinwerfen!",
/*MISSING*/    "Cannot throw fire there!",
    }},
  {0x758dc55, { // "\nThis is a ranged Orb. "
   "\nDziałanie tej Sfery celujesz w konkretne miejsce.",
/*MISSING*/    "\nThis is a ranged Orb. ",
   "\nToto je sféra, kterou je třeba zacílit na určité políčko.",
   "\nДействие этой сферы направляется на определённую клетку.",
   "\nDies ist ein Distanzorb.",
/*MISSING*/    "\nThis is a ranged Orb. ",
    }},
  {0x768252e, { // "aura brightness"
   "jasność aury",
/*MISSING*/    "aura brightness",
   "jas aury",
   "яркость ауры",
/*MISSING*/    "aura brightness",
/*MISSING*/    "aura brightness",
    }},
  {0x7bd70e6, { // "%The1 kisses you, and thanks you for saving %him1."
   "%1 Cię pocałowa%ł1 i podziękowa%ł1 za uratowanie %go1.",
   "%1 seni öptü ve %go1 kurtardığın için sana teşekkür etti.",
   "%1 tě políbi%l1 a děkuje ti, že jsi %ho1 zachránil.",
   "%1 поцеловал%E1 тебя и поблагодарил%E1 за спасение.",
   "%Der1 %1 küsst dich und dankt dir für die Rettung.",
/*MISSING*/    "%The1 kisses you, and thanks you for saving %him1.",
    }},
  {0x7c046df, { // "exit this menu"
   "wyjdź",
/*MISSING*/    "exit this menu",
   "ven z menu",
   "выход",
   "Menü verlassen",
/*MISSING*/    "exit this menu",
    }},
  {0x7d42ba2, { // "The Vineyard is filled with vines. A very dense pattern of straight lines here...\n\nVine Beasts and Vine Spirits change vine cells to grass, and vice versa."
   "Winnica jest wypełniona winoroślami. Bardzo gęsta siatka linii prostych...\n\nWinne Bestie i Duszki zmieniają winorośle w trawę i z powrotem.",
   "Şaraplık şaraplarla dolu. Burada çok sık düz çizgiler var ...\n\nşŞarap Canavarları ve Şarap Özleri şarap hücrelerini çimene ve tam tersine çevirir.",
   "Vinice je plná révy. Je tu velmi hustý vzor rovných čar...\n\nRévoví netvoři a Révoví duchové mění políčka s révou na trávu a obráceně.",
   "Виноградник полон виноградными лозами. Они образуют очень плотный узор из прямых...\n\nВинные звери и духи превращают траву в вино и наоборот.",
   "Der Weingarten ist voller Reben. Ein sehr dichtes Netz aus Geraden...\n\nRebenbestien und Rebenseelen verwandeln Zellen von Gras in Reben und umgekehrt.",
/*MISSING*/    "The Vineyard is filled with vines. A very dense pattern of straight lines here...\n\nVine Beasts and Vine Spirits change vine cells to grass, and vice versa.",
    }},
  {0x7e94a1b, { // "automatic"
   "automatyczny",
   "otomatik",
   "automatický",
   "авто",
   "automatisch",
   "automático",
    }},
  {0x7f6534d, { // "A spirit living in the vines!\n\nVine Spirits destroy the vines when they die."
   "Duch żyjący w winoroślach!\n\nWinne Duszki niszczą winorośle, gdy giną.",
   "Şarabın içinde yaşayan bir ruh!\n\nŞarap ruhları öldüklerinde etraflarındaki şarabı yok eder.",
   "Duch žijící v révě!\n\nRévoví duchové ničí révu, v níž se nacházejí, když zemřou.",
   "Дух, живущий в лозе\n\nВинный дух уничтожает лозу при гибели.",
   "Eine Seele die in den Reben lebt!\n\nRebenseelen zerstören Reben, wenn sie sterben.",
/*MISSING*/    "A spirit living in the vines!\n\nVine Spirits destroy the vines when they die.",
    }},
  {0x854657d, { // "vector graphics editor"
   "edytor grafiki wektorowej",
   "vektör grafiği editörü",
   "editor vektorové grafiky",
   "векторный графический редактор",
   "Vektoren-Grafiken-Editor",
   "editor de vetores gráficos",
    }},
  {0x88bb835, { // "o = zoom out"
   "o = oddalenie",
   "o = uzaklaş",
   "o = oddálení",
   "o = отдалить",
   "o = rauszoomen",
/*MISSING*/    "o = zoom out",
    }},
  {0x8cf4e7b, { // "%The1 attacks %the2!"
   "%1 atakuje %a2!",
   "%1 %2 tarafına saldırdı!",
   "%1 útočí na %a2!",
   "%1 атаковал%E1 %a2!",
   "%Der1 %1 greift %den2 %a2 an!",
/*MISSING*/    "%The1 attacks %the2!",
    }},
  {0x8dad708, { // "Hyperboloid model"
   "Model hiperboloidy",
/*MISSING*/    "Hyperboloid model",
   "Hyperboloidový model",
   "Модель гиперболоида",
/*MISSING*/    "Hyperboloid model",
/*MISSING*/    "Hyperboloid model",
    }},
  {0x90a2e13, { // "A monster who is able to live inside the living cave wall."
   "Sączaki żyją w ściane Żywej Jaskini.",
   "Yaşayan mağara duvarlarında yaşayabilen bir canavar.",
   "Netvor, který dokáže žít ve zdech Živoucí jeskyně.",
   "Этот монстр может жить внутри стен пещеры.",
   "Ein Monster, das innerhalb der lebenden Wände leben kann.",
/*MISSING*/    "A monster who is able to live inside the living cave wall.",
    }},
  {0x9104223, { // " (won at level %1!)"
   " (wygrane na poziomie %1!)",
   " (%1. seviyede kazandın!)",
   " (vítězství -- úroveň %1!)",
   " (победа на уровне %1!)",
   " (Mit Level %1 gewonnen!)",
/*MISSING*/    " (won at level %1!)",
    }},
  {0x912bb24, { // "Tall, strong, and holding a flower in his hand. You could not bring yourself to attack such a handsome man."
   "Wysoki, silny, z kwiatem w dłoni. Nie jesteś w stanie zaatakować tak przystojnego mężczyzny.",
/*MISSING*/    "Tall, strong, and holding a flower in his hand. You could not bring yourself to attack such a handsome man.",
   "Vysoký, silný a s květinou v ruce. Na tak pohledného muže se nemůžeš přimět zaútočit.",
   "Высокий, сильный, с прекрасным цветком в руках. Вы не можете ударить этого красавца.",
   "Groß, kräftig und mit einer Blume in der Hand. Du kannst dich nicht dazu überwinden so einen hübschen Mann anzugreifen. ",
/*MISSING*/    "Tall, strong, and holding a flower in his hand. You could not bring yourself to attack such a handsome man.",
    }},
  {0x9384dee, { // "Collect treasure to access more different lands..."
   "Zbieraj skarby, by znaleźć nowe krainy...",
   "Başka diyarlara ulaşmak için hazine topla...",
   "Sbírej poklady, abys získal přístup do dalších, odlišných krajů...",
   "Собирайте сокровища, чтобы найти новые земли...",
   "Sammle Schätze um andere Länder zu erreichen...",
   "Colete tesouro para acessar novas terras...",
    }},
  {0x943ad99, { // "And the Orbs of Yendor await!"
   "I sfery Yendoru czekają!",
   "Yendorun Küresi seni bekliyor!",
   "A Yendorské sféry čekají!",
   "Шары Йендора ждут!",
   "Und die Orbs von Yendor erwarten dich!",
   "E os Orbes de Yendor aguardam!",
    }},
  {0x9582b5f, { // "Ye wonder where did th' Pirates find all these riches..."
   "Zastanawiasz się, gdzie Piraci znaleźli te bogactwa...",
   "Korsanların bu kadar serveti nereden bulduğunu merak ediyorsun...",
   "Říkáš si, kde asi ti piráti našli všechno tohle bohatství...",
   "Интересно, где Пираты нашли эти богатства...",
   "Du fragst dich, woher die Piraten all diese Schätze haben...",
/*MISSING*/    "Ye wonder where did th' Pirates find all these riches...",
    }},
  {0x9a8ca13, { // "You feel even more attuned to the magic of this land!"
   "Jesteś lepiej dostrojony do magii tego świata!",
/*MISSING*/    "You feel even more attuned to the magic of this land!",
   "Cítíš se ještě více přizpůsobený magii tohoto kraje!",
   "Вы стали лучше чувствовать магию этой земли!",
/*MISSING*/    "You feel even more attuned to the magic of this land!",
/*MISSING*/    "You feel even more attuned to the magic of this land!",
    }},
  {0x9abc85a, { // "player 7 spin"
   "gracz 7 obrót",
/*MISSING*/    "player 7 spin",
   "hráč 7 otáčení",
   "игрок 7 вертит",
   "Spieler 7 drehen",
/*MISSING*/    "player 7 spin",
    }},
  {0x9ad48c4, { // "polynomial"
   "wielomian",
/*MISSING*/    "polynomial",
   "polynomiální",
   "многочлен",
   "polynomisch",
/*MISSING*/    "polynomial",
    }},
  {0x9deb9ae, { // "%The1 is immune to wind!"
   "%1 jest odporn%ya1 na wiatr!",
   "%1 rüzgârdan etkilenmiyor!",
   "%1 je imunní proti větru!",
   "%1 защищён от ветра!",
   "%Der1 %1 ist immun gegen Wind!",
/*MISSING*/    "%The1 is immune to wind!",
    }},
  {0xa089fbe, { // "Not so hardcore?"
   "Hardkor to nie to?",
   "O kadar da aşmış değil?",
   "Že bys nebyl zas až tak hardcore?",
   "не так хардкорно?",
   "Doch nicht so hardcore?",
/*MISSING*/    "Not so hardcore?",
    }},
  {0xa3a7b67, { // "black"
   "czarny",
   "siyah",
   "černý",
   "чёрный",
   "schwarz",
   "preto",
    }},
  {0xa6a987f, { // "background music volume"
   "głośność muzyki tła",
   "müzik seviyesi",
   "hlasitost hudby",
   "громкость фоновой музыки",
   "Hintergrundmusik",
   "volume da música de fundo",
    }},
  {0xa7b9017, { // "3D configuration"
   "konfiguracja 3D",
/*MISSING*/    "3D configuration",
   "3D konfigurace",
   "конфигурация 3D",
/*MISSING*/    "3D configuration",
/*MISSING*/    "3D configuration",
    }},
  {0xaadf0e6, { // "Viziers are neither pushed away nor stunned. However, you attack them automatically when escaping from them."
   "Wezyrowie ani nie są ogłuszani, ani odpychani, ale automatycznie ich atakujesz, gdy od nich uciekasz.",
   "Vezirler sersemlemez ve geriye itilmez, onlardan kaçarken otomatik olarak onlara saldırmış olursun.",
   "Vezíry nemůžeš odstrčit ani omráčit, ale automaticky na ně útočíš, když od nich utíkáš.",
   "Визиря нельзя ни оглушить, ни оттолкнуть. Вы атакуете его каждый раз, когда убегаете от него.",
   "Wesire kannst du weder zurückstoßen noch betäuben. Jedoch greifst du sie automatisch an wenn du von ihnen fliehst.",
/*MISSING*/    "Viziers are neither pushed away nor stunned. However, you attack them automatically when escaping from them.",
    }},
  {0xab0d18e, { // "%The1 shows you two fingers."
   "%1 pokaza%ł1 Ci dwa palce.",
   "%1 sana iki parmağını gösterdi.",
   "%1 ti ukáza%l1 dva prsty.",
   "%1 показал%E1 Вам два пальца.",
   "%Der1 %1 zeigt dir zwei Finger.",
   "%na1 lhe mostra dois dedos.",
    }},
  {0xae95569, { // "A typical monster from the Graveyard, who moves through walls.\n\nThere are also wandering Ghosts. They will appear if you do not explore any new places for a long time (about 100 turns). They can appear anywhere in the game."
   "Typowy cmentarny stwór. Może przechodzić przez ściany.\n\nSą też błądzące Duchy, które pojawiają się, gdy nie zwiedzasz nowych miejsc przez długi czas (około 100 kolejek). Duchy te mogą pojawić się w dowolnym miejscu w grze.",
   "Mezarlıkta sık rastlanan bir canavar. Duvarlardan geçebilir.\n\nAyrıca amaçsızca gezinen hayaletler var. Uzunca bir süre (100 tur kadar) yeni diyarlar keşfetmediğinde karşına çıkarlar. Oyundaki herhangi bir yerde karşına çıkabilirler.",
   "Typický hřbitovní netvor, který dokáže procházet zdmi.\n\nExistují také bludní Duchové, kteří se objeví, pokud dlouho (přibližně 100 kol) neprozkoumáš žádné nové místo. Ti se mohou objevit kdekoli ve hře.",
   "Типичный для Кладбища монстр, умеющий проходить сквозь стены.\n\nТакже существуют блуждающие призраки. Они появляются, если Вы не исследуете новые места в течение большого времени (примерно 100 ходов). Они могут появиться в любом месте в игре",
   "Ein Monster das vom Friedhof stammt und sich durch Wände bewegen kann.\n\nEs gibt auch umherirrende Geister. Diese erscheinen, wenn du dich lange (um die 100 Züge) am selben Ort aufhältst. Sie können überall im Spiel auftauchen.",
/*MISSING*/    "A typical monster from the Graveyard, who moves through walls.\n\nThere are also wandering Ghosts. They will appear if you do not explore any new places for a long time (about 100 turns). They can appear anywhere in the game.",
    }},
  {0xaeacc22, { // "This Orb allows you to blow your enemies away.\n\nClick a monster to blow it one cell away. It cannot be used against mimics, ghosts, sharks and other monsters restricted to a specific terrain, and multi-tile monsters."
   "Ta Sfera pozwala Ci zdmuchnąć Twoich przeciwników.\n\nKliknij potwora, by go zdmuchnąć o jedno pole. Nie można używać przeciwko mimikom, duchom i potworom zajmującym wiele pól.",
   "Bu Küre düşmanları rüzgârla uzağa ittirmene yarar.\n\nBir canavara tıklayarak onu rüzgârla bir hücre ittirebilirsin. Bu özellik, taklitçilere, hayaletlere, köpekbalıklarına, çok hücreli canavarlara ve belli bir araziye özgü canavarlara karşı kullanılamaz.",
   "Tato sféra ti umožňuje odfouknout nepřátele daleko od tebe.\n\nKliknutím na netvora ho můžeš odfouknout o jedno pole od sebe. Tuto schopnost nelze použít na mimiky, duchy, žraloky a jiné netvory, kteří jsou omezení pouze na specifický terén nebo zabírají více než jedno pole.",
   "Эта сфера позволяет Вам сдуват врагов вдаль.\n\nНажми на монстра, чтобы сдуть его на 1 клетку. Невозможно сдуть мимиков, духов, акул и других монстров, живущих на определённой территории, и монстров из нескольких частей.",
   "Dieser Orb erlaubt es dir, deine Feinde wegzuwehen.\n\nKlicke auf ein Monster, um es eine Zelle wegzuwehen. Der Orb ist wirkungslos gegenüber Mimiken, Geistern, Haien, und anderen Monstern, die nur auf bestimmtem Terrain leben können. Monster, die größer als eine Zelle sind, sind auch immun.",
/*MISSING*/    "This Orb allows you to blow your enemies away.\n\nClick a monster to blow it one cell away. It cannot be used against mimics, ghosts, sharks and other monsters restricted to a specific terrain, and multi-tile monsters.",
    }},
  {0xaf55625, { // "m = move v"
   "m = przenieś v",
   "m = taşı v",
   "m = přesuň v",
   "m = передвинуть v",
   "m = bewegen v",
/*MISSING*/    "m = move v",
    }},
  {0xb885cbb, { // "Make him hit himself by walking away from him."
   "Odejdź od niego, by sam się zaatakował.",
   "Ondan uzaklaşarak kendisine vurmasını sağla.",
   "Když se pohneš směrem od něj, zasáhne sám sebe.",
   "Заставьте его ударить самого себя, просто подойдя к нему.",
   "Bringe ihn dazu sich selbst zu treffen, indem du dich von ihm entfernst.",
/*MISSING*/    "Make him hit himself by walking away from him.",
    }},
  {0xb8ac8d0, { // "You get the powers of Shield, Horns, and Thorns after you move two moves in a straight line with this Orb."
   "Gdy ruszysz się 2 pola w linii prostej z tą Sferą, dostajesz moce Tarczy, Rogów i Cierni.",
/*MISSING*/    "You get the powers of Shield, Horns, and Thorns after you move two moves in a straight line with this Orb.",
   "Kdykoli se pohneš o dvě políčka rovně, získáš schopnosti Štítu, Rohů a Trnů.",
   "Вы получаете силы Щита, Рогов и Шипов, если совершаете два хода в одном направлении.",
/*MISSING*/    "You get the powers of Shield, Horns, and Thorns after you move two moves in a straight line with this Orb.",
/*MISSING*/    "You get the powers of Shield, Horns, and Thorns after you move two moves in a straight line with this Orb.",
    }},
  {0xbcc46d4, { // "1 turn"
   "1 kolejka",
/*MISSING*/    "1 turn",
   "1 kolo",
   "1 ход",
/*MISSING*/    "1 turn",
/*MISSING*/    "1 turn",
    }},
  {0xbe22143, { // "This orb allows you to instantly kill a non-adjacent enemy by clicking it. Each use drains 30 charges."
   "Ta sfera pozwala natychmiastowo zabić niesąsiadującego przeciwnika (klikając go). Każde użycie zmniejsza poziom naładowania o 30 jednostek.",
   "Bu küre yakında olmayan herhangi bir düşmana tıklayarak onu öldürebilmene izin verir. Her kullanım 30 şarj götürür.",
   "Tato sféra ti umožňuje zabít kliknutím libovolného nepřítele, který není na sousedním políčku. Každé použití sníží náboj sféry o 30.",
   "Эта сфера позволит Вам убить любого врага, не стоящего рядом с Вами, просто кликнув по нему. Это действие тратит 30 зарядов.",
   "Dieser Orb erlaubt es dir, einen nicht angrenzenden Feind augenblicklich zu töten, indem du auf ihn klickst. Jede Verwendung verbraucht 30 Ladungen.",
/*MISSING*/    "This orb allows you to instantly kill a non-adjacent enemy by clicking it. Each use drains 30 charges.",
    }},
  {0xbf7cd91, { // "Cheat-changed the display."
   "Zmieniam tryb wyświetlania.",
   "Cheat-changed the display.",
   "Cheat-changed the display.",
   "Отображение использованных читов.",
   "Anzeige durch Cheats geändert.",
/*MISSING*/    "Cheat-changed the display.",
    }},
  {0xc09f69c, { // "no axes"
   "bez osi",
   "eksen yok",
   "bez os",
   "без осей",
   "keine Achsen",
   "sem eixos",
    }},
  {0xc3296b1, { // "map editor"
   "edytor map",
   "harita editörü",
   "mapový editor",
   "редактор карт",
   "Map Editor",
/*MISSING*/    "map editor",
    }},
  {0xc51a284, { // "\n\nA Ghost never moves to a cell which is adjacent to another Ghost of the same kind."
   "\n\nDuch nigdy nie wchodzi na pole sąsiadujące z innym Duchem tego samego typu.",
/*MISSING*/    "\n\nA Ghost never moves to a cell which is adjacent to another Ghost of the same kind.",
   "\n\nDuch nikdy nevstoupí na políčko sousedící s jiným Duchem stejného typu.",
   "\n\nПризрак не может ходить на клетку рядом с другим таким же призраком.",
   "\n\nEin Geist bewegt sich niemals zu einer Zelle direkt neben einem anderen Geist derselben Art.",
/*MISSING*/    "\n\nA Ghost never moves to a cell which is adjacent to another Ghost of the same kind.",
    }},
  {0xc9b8ae4, { // "%The1 breathes fire!"
   "%1 zionął ogniem!",
/*MISSING*/    "%The1 breathes fire!",
   "%1 chrlí oheň!",
   "%1 дышит огнём!",
   "%Der1 %1 speit Feuer!",
/*MISSING*/    "%The1 breathes fire!",
    }},
  {0xcd8ab74, { // "Each eight turns, each rosebush at distance at most 5 from you will release a wave of alluring scent. Creatures on the frontwave will move towards where the scent came from. Even if it causes them to attack their friends or beautiful creatures, or move into water, fire, chasm, or thorns of the rosebush. Ivies, Ghosts, Rock Snakes, Rose Ladies and Lords, and monsters restricted to a specific terrain are immune to scents."
   "Co 8 kolejek każdy krzak róży w promieniu 5 od Ciebie wysyła falę uwodzicielskiego zapachu. Stworzenia na czole fali wykonają ruch w kierunku źródła zapachu, nawet jeżeli by to spowodowało, że zaatakują swoich przyjaciół lub piękne istoty, wejdą w wodę, ogień, przepaść, czy wpadną na kolce róży. Bluszcz, duchy, skalne węże, Różane Damy i Panowie, a także potwory żyjące w konkretnym terenie są odporne na zapachy.",
/*MISSING*/    "Each eight turns, each rosebush at distance at most 5 from you will release a wave of alluring scent. Creatures on the frontwave will move towards where the scent came from. Even if it causes them to attack their friends or beautiful creatures, or move into water, fire, chasm, or thorns of the rosebush. Ivies, Ghosts, Rock Snakes, Rose Ladies and Lords, and monsters restricted to a specific terrain are immune to scents.",
   "Každý růžový keřík ve vzdálenosti 5 nebo méně od tebe vypustí každých 8 kol vlnu vábivé vůně. Každý tvor zasažený touto vůní se musí pohybovat směrem k jejímu zdroji. A to i tehdy, pokud by díky tomu musel zaútočit na své kamarády nebo krásné tvory nebo musel vstoupit do vody, do ohně, do propasti nebo na trny růžového keříku. Břečťan, Kamenní hadi, Růžové dámy a pánové a netvoři omezení na určitý terén jsou vůči vůni imunní.",
   "Каждые 8 ходов каждый розовый куст на расстоянии не более 5 от Вас выпускает волну манящего аромата. Все монстры на пути двигаются к кусту, даже если им приходится атаковать союзника или прекрасное существо, пойти в воду, огонь, пропасть или в шипы куста. Это не действует на призраков, плющи, каменных змей и и монстров, ограниченных своей землёй.",
   "Alle acht Züge versprüht jeder Rosenbusch in einer Entfernung von maximal 5 Zellen von dir eine Wolke verführerischen Dufts. Wesen an der Wellenfront werden sich dorthin bewegen, wo der Duft herkommt. Auch wenn das sie dazu bringt, ihre Freunde oder schöne Wesen anzugreifen, oder sich in Wasser, Feuer, Abgründe oder Dornen der Rosenbüsche zu begeben. Efeue, Geister, Felsenschlangen, Rosendamen und -herren und Monster, die auf ein bestimmtes Gelände beschränkt sind, sind immun gegenüber dem Duft.",
/*MISSING*/    "Each eight turns, each rosebush at distance at most 5 from you will release a wave of alluring scent. Creatures on the frontwave will move towards where the scent came from. Even if it causes them to attack their friends or beautiful creatures, or move into water, fire, chasm, or thorns of the rosebush. Ivies, Ghosts, Rock Snakes, Rose Ladies and Lords, and monsters restricted to a specific terrain are immune to scents.",
    }},
  {0xcef9d5c, { // "You hear a distant roar!"
   "Słyszysz odległy ryk!",
/*MISSING*/    "You hear a distant roar!",
   "Zaslechl jsi vzdálené zařvání!",
   "Вы слышите далёкий рёв!",
   "Du hörst ein entferntes Brüllen!",
/*MISSING*/    "You hear a distant roar!",
    }},
  {0xd0aa269, { // "Press F1 or right-click things for help."
   "F1 lub klikaj prawym na elementach gry, by uzyskać pomoc.",
/*MISSING*/    "Press F1 or right-click things for help.",
   "Nápovědu můžeš zobrazit stiskem F1 nebo pravým kliknutím na různé věci.",
   "Нажмите F1 или используйте правую кнопку мыши, чтобы открыть помощь.",
/*MISSING*/    "Press F1 or right-click things for help.",
/*MISSING*/    "Press F1 or right-click things for help.",
    }},
  {0xd10afc8, { // "\"The Knights of the Horocyclic Table salute you!\""
   "\"Rycerze Horocyklicznego Stołu Cię pozdrawiają!\"",
   "\"Horosikıl Masa Şövalyeleri seni selamlıyor!\"",
   "\"Rytíři Horocyklického stolu tě zdraví!\"",
   "\"Рыцари Орициклического Стола поздравляют Вас!\"",
   "\"Die Ritter der Horozyklischen Tafel salutieren!\"",
/*MISSING*/    "\"The Knights of the Horocyclic Table salute you!\"",
    }},
  {0xd16bdfe, { // "the second page [Space]"
   "druga strona [Space]",
   "ikinci sayfa [Space]",
   "druhá stránka [mezerník]",
   "вторая страница [Пробел]",
   "Zweite Seite [LEER]",
   "a segunda página [Space]",
    }},
  {0xd1a5607, { // "Butterflies don't pursue you -- unless you get next to them, they just spin around the obstacles. They cannot be killed conventionally, but you get treasure when a Raging Bull crashes into a Butterfly. "
   "Motylki nie gonią Cię -- kręcą się wokół przeszkód, chyba że jesteś na sąsiednim polu. Nie można ich zabić zwykłą bronią, ale zdobywasz skarb, gdy Wściekły Byk zderzy się z Motylkiem.",
/*MISSING*/    "Butterflies don't pursue you -- unless you get next to them, they just spin around the obstacles. They cannot be killed conventionally, but you get treasure when a Raging Bull crashes into a Butterfly. ",
   "Motýlci tě nepronásledují -- pokud se nedostaneš těsně vedle nich, budou prostě jen poletovat kolem překážek. Motýlky nelze zabít běžným způsobem, ale když do nich vrazí Vzteklý býk, dostaneš poklad.",
   "Бабочки вас не трогают, пока вы не подойдёте прямо к ним, они просто вертятся вокруг препятствий. Их нельзя убить обычным образом, но вы получите сокровище, когда Бешеный бык врежется в Бабочку.",
/*MISSING*/    "Butterflies don't pursue you -- unless you get next to them, they just spin around the obstacles. They cannot be killed conventionally, but you get treasure when a Raging Bull crashes into a Butterfly. ",
/*MISSING*/    "Butterflies don't pursue you -- unless you get next to them, they just spin around the obstacles. They cannot be killed conventionally, but you get treasure when a Raging Bull crashes into a Butterfly. ",
    }},
  {0xd32734e, { // "Orbs of Yendor found: %1"
   "Znalezione Sfery Yendoru: %1",
   "Bulunan Yendor Küresi sayısı: %1",
   "Nalezených Yendorských sfér: %1",
   "Собрано сфер Йендора: %1",
   "Orbs von Yendor gefunden: %1",
   "Orbes de Yendor encontrados: %1",
    }},
  {0xd5882d0, { // "Cold blue gems, found in the Cocytus."
   "Zimne niebieskie kamienie, znajdowane na zamarzniętym Kocycie.",
   "Mavi soğuk mücevherler sadece Kokitos'ta bulunabilir.",
   "Chladné modré drahokamy, které se nacházejí v Cocytu.",
   "Холодные голубые камни, найденные в Коците.",
   "Kalte, blaue Edelsteine aus dem Kokytos.",
/*MISSING*/    "Cold blue gems, found in the Cocytus.",
    }},
  {0xd70069e, { // "An impassable lake in Cocytus."
   "Nie możesz przejść przez Jezioro Kocyt.",
   "Kokitos'ta, üzerinden geçilemeyen bir göl.",
   "Neprostupné jezero v Cocytu.",
   "Непроходимое озеро в Коците.",
   "Ein undurchquerbarer See im Kokytos.",
/*MISSING*/    "An impassable lake in Cocytus.",
    }},
  {0xdf22139, { // "You overheard Hedgehog Warriors talking about emeralds..."
   "Podsłucha%łeś0 Wojowników-Jeży rozmawiających o szmaragdach...",
   "Kirpi Savaşçıların zümrütler hakkında konuştuğunu duydun...",
   "Zaslechl jsi Ježčí válečníky, jak mluví o smaragdech...",
   "Вы слышите ежей-воинов, говорящих об изумрудах...",
   "Du hörst Igelkrieger über Smaragde sprechen...",
/*MISSING*/    "You overheard Hedgehog Warriors talking about emeralds...",
    }},
  {0xe38d258, { // "friends destroyed: %1"
   "przyjaciele zniszczeni: %1",
/*MISSING*/    "friends destroyed: %1",
   "zničení kamarádi: %1",
   "союзников уничтожено: %1",
   "zerstörte Freunde: %1",
/*MISSING*/    "friends destroyed: %1",
    }},
  {0xe3f1e5c, { // "You activate the Flash spell!"
   "Aktywowa%łeś0 czar Błysk!",
   "Işık büyüsünü etkinleştirdin!",
   "Aktivova%l0 jsi kouzlo Záblesk!",
   "Вы активировали заклинание вспышки!",
   "Du aktivierst den Blitz Zauber!",
   "Você ativa o feitiço de Clarão!",
    }},
  {0xea35811, { // "This Orb allows you to target monsters to stun them. 10 charges are used to stun for 5 turns. Does not work against multi-tile monsters."
   "Ta sfera pozwala ogłuszać potwory na odległość. 10 ładunków = ogłuszenie na 5 tur. Nie działa przeciwko potworom zajmującym wiele pól.",
   "Bu Küre canavarları hedefleyerek sersemletmeni sağlar. 5 tur sersemletmek için 10 güç kullanır. Çokluhücre canavarlara karşı çalışmaz.",
   "Tato Sféra ti umožňuje zacílit a omráčit netvory. Každé omráčení trvá 5 kol a stojí 10 nábojů. Tato Sféra nepůsobí na netvory, kteří zabírají více než jedno políčko.",
   "Эта сфера позволяет оглушать монстров. 10 зарядов = оглушения на 5 ходов. Не действует на \"многоклеточных\" монстров.",
   "Dieser Orb erlaubt es dir Monster zu betäuben. 10 Ladungen betäuben für 5 Züge. Dies funktioniert nicht gegen große Monster die mehrere Felder bedecken.",
/*MISSING*/    "This Orb allows you to target monsters to stun them. 10 charges are used to stun for 5 turns. Does not work against multi-tile monsters.",
    }},
  {0xee5743a, { // "Gravity does not allow this!"
   "Grawitacja nie pozwala!",
   "Yerçekimi buna izin vermiyor!",
   "Tohle ti gravitace nedovolí!",
   "Гравитация не позволяет!",
   "Die Gravitation erlaubt das nicht!",
/*MISSING*/    "Gravity does not allow this!",
    }},
  {0xefb46c9, { // "Raging Bulls charge in a straight line: on heptagons, when they can choose one of two possible directions, they choose one closer to your current location. In the case of a tie, the cell where more neighbors is closer to your current location is chosen; if still a tie, past locations are considered. They can attack you in any direction, and monsters on their way are attacked even if friendly. When they crash into something, the obstacle is usually destroyed, and they are stunned for three turns, after which they charge at you again (in any direction). Raging Bulls cannot be killed or stunned conventionally."
   "Wściekłe Byki szarżują w linii prostej: na siedmiokątach, gdy mogą wybrać jeden z dwóch kierunków, wybierają pole bliższe Twojej obecnej pozycji. W przypadku remisu pole z większą liczbą sąsiadów bliżej Ciebie jest wybierane. W przypadku dalszego remisu, rozważane są Twoje wcześniejsze pozycje. Byki mogą atakować w dowolnym kierunku; potwory na ich drodze są atakowane nawet jak są przyjazne. Gdy Byk się z czymś zderzy, przeszkoda jest niszczona, i Byk jest ogłuszony na 3 kolejki, po czym szarżuje ponownie (w dowolnym kierunku). Wściekłego Byka nie można zabić ani ogłuszyć zwykłą bronią.",
/*MISSING*/    "Raging Bulls charge in a straight line: on heptagons, when they can choose one of two possible directions, they choose one closer to your current location. In the case of a tie, the cell where more neighbors is closer to your current location is chosen; if still a tie, past locations are considered. They can attack you in any direction, and monsters on their way are attacked even if friendly. When they crash into something, the obstacle is usually destroyed, and they are stunned for three turns, after which they charge at you again (in any direction). Raging Bulls cannot be killed or stunned conventionally.",
   "Vzteklí Býci útočí po přímce: na sedmiúhelníku, kde si mohou vybrat jeden ze dvou směrů, si vždy zvolí ten, který je blíž k tobě. Pokud je to nastejno, vyberou si políčko s větším počtem sousedů bližších ke tvé pozici, a pokud je to stále nerozhodně, vezmou v úvahu minulé lokace. Býci mohou útočit ve všech směrech a na netvory, kteří jim stojí v cestě, útočí i tehdy, když jsou přátelští. Když Býk do něčeho narazí, obvykle tím překážku zničí a bude na tři kola omráčený; potom opět vyrazí proti tobě (v libovolném směru). Vzteklého Býka nelze zabít ani omráčit běžnými prostředky.",
   "Бешеные быки бегут по прямой: на семиугольниках, где есть выбор из двух направлений, они выбирают то, которое ближе к вам. В случае равенства выбирается клетка, у которой больше соседей ближе к вам. Если и здесь равенство, рассматриваются ваши предыдущие положения. Быки могут атаковать в любую сторону, и они атакуют всех на своём пути, даже дружественных существ. Когда бык врезается в препятствие, оно обычно разрушается, и бык оглушается на три хода, после чего снова направляется на вас (в любую сторону). Бешеного быка нельзя убить или оглушить обычным способом.",
/*MISSING*/    "Raging Bulls charge in a straight line: on heptagons, when they can choose one of two possible directions, they choose one closer to your current location. In the case of a tie, the cell where more neighbors is closer to your current location is chosen; if still a tie, past locations are considered. They can attack you in any direction, and monsters on their way are attacked even if friendly. When they crash into something, the obstacle is usually destroyed, and they are stunned for three turns, after which they charge at you again (in any direction). Raging Bulls cannot be killed or stunned conventionally.",
/*MISSING*/    "Raging Bulls charge in a straight line: on heptagons, when they can choose one of two possible directions, they choose one closer to your current location. In the case of a tie, the cell where more neighbors is closer to your current location is chosen; if still a tie, past locations are considered. They can attack you in any direction, and monsters on their way are attacked even if friendly. When they crash into something, the obstacle is usually destroyed, and they are stunned for three turns, after which they charge at you again (in any direction). Raging Bulls cannot be killed or stunned conventionally.",
    }},
  {0xf0f9157, { // "Quite tough, for your first fight."
   "Jak na pierwszą walkę -- całkiem trudny przeciwnik.",
/*MISSING*/    "Quite tough, for your first fight.",
   "Na první bitku docela obtížný soupeř.",
   "Довольно сложно для первого боя.",
/*MISSING*/    "Quite tough, for your first fight.",
/*MISSING*/    "Quite tough, for your first fight.",
    }},
  {0xf19397a, { // "%The1 says, \"not this place, it looks even worse...\""
   "%The1 mówi, \"nie tu, tu jest jeszcze gorzej...\"",
/*MISSING*/    "%The1 says, \"not this place, it looks even worse...\"",
   "%The1 øíká, \"tady ne, tohle vypadá ještì hùø...\"",
   "%1 сказал, \"не это место, оно выглядит ещё хуже...\"",
/*MISSING*/    "%The1 says, \"not this place, it looks even worse...\"",
/*MISSING*/    "%The1 says, \"not this place, it looks even worse...\"",
    }},
  {0xf3f7810, { // "Score: "
   "Wynik: ",
   "Puan: ",
   "Skóre: ",
   "Очки: ",
   "Punkte: ",
/*MISSING*/    "Score: ",
    }},
  {0xf582c14, { // " You need to go deep to collect lots of them."
   "By zebrać ich dużo, musisz głęboko wejść w krainę.",
/*MISSING*/    " You need to go deep to collect lots of them.",
   "Abys jich získa%l0 hodně, musíš jít hluboko do tohoto kraje.",
   "Вам нужно забраться глубже, чтобы собрать побольше их.",
/*MISSING*/    " You need to go deep to collect lots of them.",
/*MISSING*/    " You need to go deep to collect lots of them.",
    }},
  {0xf764393, { // "Halloween is a special land, that is available only in the spherical or elliptic geometry (press 'o' to switch). You play on the surface of a jack-o'-lantern, and have to collect as many Treats as possible. Each Treat you collect brings new monsters to fight, and new magical powers for you. You have to fight the monsters while effectively managing your limited resources."
   "Halloween to specjalna kraina, dostępna jedynie w geometrii sferycznej lub eliptycznej ('o' by zmienić). Grasz na powierzchni wydrążonej dyni, i musisz zdobyć jak najwięcej Cukierków. Każdy Cukierek przynosi nowe potwory do walki, i nowe magiczne moce. Walcz z potworami, efektywnie zarządzając swoimi ograniczonymi zasobami.",
/*MISSING*/    "Halloween is a special land, that is available only in the spherical or elliptic geometry (press 'o' to switch). You play on the surface of a jack-o'-lantern, and have to collect as many Treats as possible. Each Treat you collect brings new monsters to fight, and new magical powers for you. You have to fight the monsters while effectively managing your limited resources.",
   "Halloween je zvláštní kraj, který je k dispozici pouze ve sférické nebo eliptické geometrii (pro přepnutí použijte klávesu 'o'). Hraje se v něm na povrchu dýňové lucerny, kde musíš posbírat co nejvíce Pochoutek. Každá Pochoutka, kterou sebereš, vyvolá nové netvory, se kterými musíš bojovat, a dá ti nové magické schopnosti. Musíš bojovat s netvory a přitom efektivně spravovat své omezené zdroje.",
   "Хэллоуин -- специальная земля, доступная только в сферической и эллиптической геометрии (нажмите 'o', чтобы переключить). Вы играете на поверхности хэллоуинской тыквы и собираете конфеты. Каждая конфета призывает новых монстров и даёт Вам новые силы. Сражайтесь с монстрами, грамотно распоряжаясь ограниченными ресурсами.",
/*MISSING*/    "Halloween is a special land, that is available only in the spherical or elliptic geometry (press 'o' to switch). You play on the surface of a jack-o'-lantern, and have to collect as many Treats as possible. Each Treat you collect brings new monsters to fight, and new magical powers for you. You have to fight the monsters while effectively managing your limited resources.",
/*MISSING*/    "Halloween is a special land, that is available only in the spherical or elliptic geometry (press 'o' to switch). You play on the surface of a jack-o'-lantern, and have to collect as many Treats as possible. Each Treat you collect brings new monsters to fight, and new magical powers for you. You have to fight the monsters while effectively managing your limited resources.",
    }},
  {0xf87773a, { // "player 5 go"
   "gracz 5 idź",
/*MISSING*/    "player 5 go",
   "hráč 5 pohyb",
   "игрок 5 идёт",
   "Spieler 5 los",
/*MISSING*/    "player 5 go",
    }},
  {0xf90bfdd, { // "You have improved your total high score on Steam. Congratulations!"
   "Poprawi%łeś0 swój ogólny najlepszy wynik na Steam. Gratulacje!",
   "Steam'deki toplam yüksek puanını artırdın. Tebrikler!",
   "Zlepši%l0 jsi své celkové nejvyšší skóre na Steamu. Gratulujeme!",
   "Вы увеличили свой общий счёт в Steam. Поздравляем!",
   "Du hast deinen Gesamthighscore auf Steam verbessert! Glückwunsch!",
   "Você superou sua melhor pontuação da Steam. Parabéns!",
    }},
  {0xfb6fe2c, { // "hepta floor"
   "podłoga hepta",
   "heptazemin",
   "hepta-podlaha",
   "семиугольный пол",
   "Heptagonaler Boden",
   "andar hepta",
    }},
  {0xfc1035b, { // "%The1 melts away!"
   "%1 się stopi%ł1!",
   "%1 eridi!",
   "%1 roztá%l1!",
   "%1 тает!",
   "%Der1 %1 schmilzt!",
   "%na1 derrete!",
    }},
  {0x10344404, { // "towards Gans model"
   "w stronę modelu Gansa",
/*MISSING*/    "towards Gans model",
   "směr ke Gansovu modelu",
   "в сторону модели Ганса",
/*MISSING*/    "towards Gans model",
/*MISSING*/    "towards Gans model",
    }},
  {0x106bda07, { // "Human to wall ratio"
   "Stosunek człowieka do ścian",
/*MISSING*/    "Human to wall ratio",
   "Poměr člověka a zdi",
   "Отношение высот человека и стены",
/*MISSING*/    "Human to wall ratio",
/*MISSING*/    "Human to wall ratio",
    }},
  {0x10960669, { // "Fjord Trolls leave a wall when they die, causing the living fjord to rise around it. Contrary to Rock Trolls, items around are not destroyed."
   "Fjordtrolle zostawiają ścianę po zabiciu, co powoduje, że żywy fiord w okolicy się podnosi. W przeciwieństwie do Skalnych Trolli Fjordtrolle nie niszczą przedmiotów wokół.",
   "Fiyort Trolleri öldüklerinde geriye bir duvar bırakıp cesetlerinin etrafına yaşayan fiyortun yükselmesine sebep olurlar. Kaya Trollerinin aksine etraflarındaki eşyalar yok olmaz.",
   "Fjordtrollové se po smrti promění v zeď, kolem které se zvedají vody Živoucího fjordu. Fjordtrollové na rozdíl od Šutrollů neničí předměty kolem sebe.",
   "Тролли фьорда превращаются в стену при смерти, вызывая рост живых фьордов вокруг. В отличие от горных троллей они не уничтожают предметы.",
   "Fjordtrolle hinterlassen eine Wand wenn sie sterben, was den lebenden Fjord dazu bringt um sie herum anzusteigen. Im Gegensatz zu Stein Trollen werden Items nicht zerstört.",
/*MISSING*/    "Fjord Trolls leave a wall when they die, causing the living fjord to rise around it. Contrary to Rock Trolls, items around are not destroyed.",
    }},
  {0x10c001b8, { // "(most achievements are not available)"
   "(większość osiągnięć niedostępna)",
/*MISSING*/    "(most achievements are not available)",
   "(většina achievementů není k dispozici)",
   "(большинство достижений недоступно)",
/*MISSING*/    "(most achievements are not available)",
/*MISSING*/    "(most achievements are not available)",
    }},
  {0x10c5d1d4, { // "openGL & antialiasing mode"
   "openGL i antialiasing",
   "openGL & antialiasing",
   "openGL & antialiasing",
   "openGL и режим сглаживания",
   "OpenGL & Anti-Aliasing",
   "Modo OpenGL e antiserrilhamento",
    }},
  {0x10fa192b, { // "A big, beautiful, magical flower."
   "Duży, piękny, magiczny kwiat.",
/*MISSING*/    "A big, beautiful, magical flower.",
   "Velká, krásná, magická květina.",
   "Большой прекрасный магический цветок.",
   "Eine große, schöne, magische Blume.",
/*MISSING*/    "A big, beautiful, magical flower.",
    }},
  {0x110781c5, { // "This structure will disappear after some time."
   "Ta konstrukcja zniknie za jakiś czas.",
   "Bu yapı bir süre sonra yok olacak.",
   "Tato struktura za nějaký čas zmizí.",
   "Эта конструкция пропадёт через некоторое время.",
   "Diese Struktur wird nach einiger Zeit verschwinden.",
/*MISSING*/    "This structure will disappear after some time.",
    }},
  {0x1109d82c, { // "This Orb protects you from physical attacks. It lasts for more turns than the Orb of Shielding, but 10 charges are lost whenever you are attacked. It also does not protect you from fires, scents, and being eaten."
   "Ta Sfera chroni Cię przed fizycznymi atakami. Działa dłużej niż Sfera Tarczy, ale każdy atak zabiera 10 ładunków. Nie chroni przed ogniem, zapachami i byciem zjedzonym.",
/*MISSING*/    "This Orb protects you from physical attacks. It lasts for more turns than the Orb of Shielding, but 10 charges are lost whenever you are attacked. It also does not protect you from fires, scents, and being eaten.",
   "Tato Sféra tě chrání před fyzickými útoky. Vydrží déle než Sféra Štítu, ale každý útok ti ubere 10 nábojů. Nechrání proti ohni, pachu a sežrání.",
   "Эта сфера защищает вас от физических атак. Она действует дольше, чем сфера Щита, но теряет10 зарядов при каждой атаке. Не защищает от огня, запаха и съедения.",
   "Dieser Orb schützt dich vor physischen Angriffen. Es hält mehr Züge als der Orb der Abschirmung , aber immer wenn du angegriffen wirst, verlierst er 10 Ladungen. Er schützt dich nicht vor Feuer und Düften oder davor gefressen zu werden.",
/*MISSING*/    "This Orb protects you from physical attacks. It lasts for more turns than the Orb of Shielding, but 10 charges are lost whenever you are attacked. It also does not protect you from fires, scents, and being eaten.",
    }},
  {0x11cfbf44, { // "People worshipping Cthulhu. This one is especially dangerous, as he is armed with a weapon which launches fire from afar."
   "Wyznawcy Cthulhu. Ten jest szczególnie niebezpieczny, bo może z odległości rozpalić ogień w Twojej okolicy.",
   "Cthulhu'ya tapan insanlar. Bu özellikle tehlikeli ve uzaktan ateş atabilmesini sağlayan bir silahı var.",
   "Lidé uctívající Cthulhua. Tenhle je zvlášť nebezpečný, protože je vybavený zbraní, která dokáže metat oheň na dálku.",
   "Люди, поклоняющиеся Ктулху. Этот особо опасен, так как он может кинуть в Вас огонь.",
   "Verehrer Cthulhus. Dieser ist besonders gefährlich, er besitzt eine Waffe, die aus der Ferne Feuer schleudern kann.",
/*MISSING*/    "People worshipping Cthulhu. This one is especially dangerous, as he is armed with a weapon which launches fire from afar.",
    }},
  {0x122fca68, { // "Not enough power for telekinesis!"
   "Za mało mocy do telekinezy!",
   "Telekinezi için yeterince güç yok!",
   "Nemáš dost síly na telekinezi!",
   "Мало силы для телекинеза!",
   "Nicht genügend Kraft für Telekinese!",
/*MISSING*/    "Not enough power for telekinesis!",
    }},
  {0x1276d6f4, { // "use default keys"
   "klawisze domyślne",
/*MISSING*/    "use default keys",
   "standardní klávesy",
   "использовать клавиши по умолчанию",
   "Standard-Tastenbelegung nutzen",
/*MISSING*/    "use default keys",
    }},
  {0x12d4c0f9, { // "Game statistics saved to %1"
   "Statystyki gry zapisane do %1",
   "İstatistikler %1 konumuna kaydedildi.",
   "Herní statistika uložena do %1",
   "Статистика игры записана в %1",
   "Statistiken nach %1 gespeichert",
   "Estatísticas de jogo salvas em %1",
    }},
  {0x12dc77a0, { // "Powerful wizards claimed this part of the world, to perform their magical experiments in peace and solitude. They have changed the direction of gravity, to make it even harder for intruders to reach them.\n\nGravity works as follows: cells are unstable if they are empty, and there is no cell immediately below them which contains a wall. It is impossible to move from one unstable cell to another, except if moving down."
   "Część świata zajęta przez potężnych czarodziejów, w celu prowadzenia magicznych eksperymentów w spokoju i samotności. Zmienili oni kierunek grawitacji, by intruzom było jeszcze trudniej się do nich dostać.\n\nGrawitacja działa w sposób następujący: pole jest niestabilne, jeśli jest puste i żadne pole bezpośrednio pod nim nie zawiera ściany. Nie można przejść z jednego niestabilnego pola na inne, chyba że w dół.",
   "Güçlü büyücüler, dünyanın bu kısmını, büyüsel deneylerini sessiz sakin uygulayabilecekleri yer olarak seçtiler. Davetsiz misafirlerin, onlara ulaşabilmesini daha zor kılmak için yerçekiminin yönünü değiştirdiler.\n\nYerçekimi şu şekilde çalışır: Hücreler eğer boşlar ise dengesiz haldedirler ve hücrelerin hemen altında bulunan bir duvar yoktur. Bir dengesiz hücreden diğerine ilerlemek imkansızdır, aşağı doğru ilerlemek hariç.\n\n",
   "Tuto část světa si pro sebe zabrali mocní mágové, aby zde mohli v klidu a osamění provádět své magické experimenty. Změnili směr gravitace, aby bylo pro vetřelce ještě obtížnější se k nim dostat.\n\nGravitace funguje následovně: prázdné políčko, pod kterým není žádné políčko se zdí, je 'nestabilní'. Není dovoleno pohybovat se z jednoho nestabilního políčka do druhého, leda směrem dolů.",
   "Мощные колдуны создавали эту землю, чтобы проводить магические эксперименты в покое и уединении. Они изменили направление силы тяжести, чтобы сделать попадание сюда труднее.\n\nГравитация работает так: клетка нестабильна, если она пуста, и поле под ним не содердит стены. Нельзя перейти с одного нестабильного поля на другое, иначе как вниз.",
   "Mächtige Zauberer haben diesen Teil der Welt für sich beansprucht um in Ruhe und Frieden ihre magischen Experimente durchführen zu können. Sie haben die Gravitation in diesem Land manipuliert um es Eindringlingen zu erschweren sie zu erreichen.\n\nDie Gravitation funktioniert wie folgt: Zellen sind instabil wenn sie leer sind und sich direkt unter ihnen keine Zelle mit Wand befindet. Es ist unmöglich sich von einer instabilen Zelle zu einer anderen zu bewegen außer wenn man sich nach unten bewegt.",
/*MISSING*/    "Powerful wizards claimed this part of the world, to perform their magical experiments in peace and solitude. They have changed the direction of gravity, to make it even harder for intruders to reach them.\n\nGravity works as follows: cells are unstable if they are empty, and there is no cell immediately below them which contains a wall. It is impossible to move from one unstable cell to another, except if moving down.",
    }},
  {0x12edf159, { // "memory game"
   "gra pamięciowa",
/*MISSING*/    "memory game",
   "hra na paměť",
   "игра 'память'",
/*MISSING*/    "memory game",
/*MISSING*/    "memory game",
    }},
  {0x13116152, { // "A very interesting species of slime mold."
   "Bardzo ciekawy gatunek śluzowca.",
/*MISSING*/    "A very interesting species of slime mold.",
   "Velice zajímavý druh hlenky.",
   "Очень интересная разновидность слизня.",
/*MISSING*/    "A very interesting species of slime mold.",
/*MISSING*/    "A very interesting species of slime mold.",
    }},
  {0x132b8912, { // "Objects at distance less than %1 absolute units from the center will be displayed with high detail, and at distance at least %2 with low detail."
   "Obiekty w odległości bliższej niż %1 od środka będą rysowane z wieloma szczegółami, a dalszej niż %2 z niewieloma.",
/*MISSING*/    "Objects at distance less than %1 absolute units from the center will be displayed with high detail, and at distance at least %2 with low detail.",
   "Objekty ve vzdálenosti méně než %1 od středu budou zobrazovány s vysokými detaily, a objekty ve vzdálenosti méně než %2 se středními detaily.",
   "Объекты на расстоянии меньше %1 единиц от центра изображены с большой детализацией, а дальше %2 -- с низкой.",
/*MISSING*/    "Objects at distance less than %1 absolute units from the center will be displayed with high detail, and at distance at least %2 with low detail.",
/*MISSING*/    "Objects at distance less than %1 absolute units from the center will be displayed with high detail, and at distance at least %2 with low detail.",
    }},
  {0x13502f33, { // "more options"
   "więcej opcji",
/*MISSING*/    "more options",
   "více možností",
   "больше вариантов",
/*MISSING*/    "more options",
/*MISSING*/    "more options",
    }},
  {0x13552e54, { // "%The1 breathes fire at you!"
   "%1 zionął ogniem na Ciebie!",
/*MISSING*/    "%The1 breathes fire at you!",
   "%1 na tebe chrlí oheň!",
   "%1 дышит огнём на вас!",
   "%Der1 %1 speit Feuer auf dich!",
/*MISSING*/    "%The1 breathes fire at you!",
    }},
  {0x139ddc72, { // "continuous spacetime"
   "ciągła czasoprzestrzeń",
/*MISSING*/    "continuous spacetime",
   "spojitý časoprostor",
   "непрерывное пространство и время",
/*MISSING*/    "continuous spacetime",
/*MISSING*/    "continuous spacetime",
    }},
  {0x13b5fb7d, { // "distance between eyes"
   "odległość między oczami",
   "gözler arası uzaklık",
   "vzdálenost očí od sebe",
   "расстояние между глазами",
   "Augenabstand",
   "distância entre olhos",
    }},
  {0x13c016ae, { // "special features [Space]"
   "opcje specjalne [Space]",
   "özel seçenekler [Space]",
   "zvláštní možnosti [mezerník]",
   "специальные опции [Пробел]",
   "Spezielle Features [LEER]",
   "características especiais [Space]",
    }},
  {0x13d3efe5, { // "render now (length: %1)"
   "rysuj teraz (długość: %1)",
/*MISSING*/    "render now (length: %1)",
   "renderovat hned (délka: %1)",
   "рисуется сейчас (длина: %1)",
   "rendere jetzt (Länge: %1)",
/*MISSING*/    "render now (length: %1)",
    }},
  {0x13ef5280, { // "cheats"
   "oszustwa",
/*MISSING*/    "cheats",
   "cheaty",
   "Читы",
/*MISSING*/    "cheats",
/*MISSING*/    "cheats",
    }},
  {0x1414f4da, { // "see how it ended"
   "jak się skończyło",
/*MISSING*/    "see how it ended",
   "podívej se, jak to skonèilo",
   "смотри, как всё закончилось",
/*MISSING*/    "see how it ended",
/*MISSING*/    "see how it ended",
    }},
  {0x14452571, { // "You destroy %the1 with a mental blast!"
   "Zniszczy%łeś0 %a1 mocą psychiczną!",
   "%a1 bir zihin patlamasıyla yok ettin!",
   "Zniči%l0 jsi %a1 mentálním úderem!",
   "Вы уничтожили %a1 силой мысли!",
   "Du zerstörst %den1 %a1 mit einem mentalen Stoß!",
/*MISSING*/    "You destroy %the1 with a mental blast!",
    }},
  {0x148dd589, { // "cells"
   "pola",
/*MISSING*/    "cells",
   "políèek",
   "клетки",
/*MISSING*/    "cells",
/*MISSING*/    "cells",
    }},
  {0x15528312, { // "A tourist from another world. They mutter something about the 'tutorial', and claim that they are here just to learn, and to leave without any treasures. Do not kill them!"
   "Turyst%aka0 z innego świata. Coś mamrocze o 'podręczniku', i że się tylko uczy, że nie jest tu po skarby. Nie zabijać!",
/*MISSING*/    "A tourist from another world. They mutter something about the 'tutorial', and claim that they are here just to learn, and to leave without any treasures. Do not kill them!",
   "Turista z jiného světa. Mumlá něco o 'tutoriálu' a tvrdí, že se jenom učí a odejde bez pokladů. Nezabíjet!",
   "Турист из другого мира. Они бормочут что-то про 'руководство', и говорят, что хотят научить, а не забрать сокровища. Не убивай их!",
/*MISSING*/    "A tourist from another world. They mutter something about the 'tutorial', and claim that they are here just to learn, and to leave without any treasures. Do not kill them!",
/*MISSING*/    "A tourist from another world. They mutter something about the 'tutorial', and claim that they are here just to learn, and to leave without any treasures. Do not kill them!",
    }},
  {0x15671397, { // "You feel great, like a true treasure hunter."
   "Czujesz się świetnie, jak prawdziwy łowca skarbów.",
   "Şahane hissediyorsun. Sanki gerçek bir hazine avcısı gibi.",
   "Cítíš se skvěle, jako pravý lovec pokladů.",
   "Вы чувствуете себя как прекрасный охотник за сокровищами.",
   "Du fühlst dich großartig, wie ein echter Schatzjäger.",
   "Você se sente realizado, como um verdadeiro caçador de tesouros.",
    }},
  {0x158937e3, { // "Fat guards are too heavy to be pushed away by your hits."
   "Grubi Strażnicy są zbyt ciężcy, by dało się ich odepchnąć atakiem.",
   "Şişman Muhafızlar vuruşlarınla geriye itilmek için fazla ağır.",
   "Tlustí strážci jsou příliš těžcí na to, aby je tvé zásahy dokázaly odstrčit.",
   "Тучный стражник слишком тяжёл, чтобы отлетать от Ваших ударов.",
   "Dicke Wachen sind zu schwer um zurück gestoßen zu werden.",
/*MISSING*/    "Fat guards are too heavy to be pushed away by your hits.",
    }},
  {0x1589a9c4, { // "turn count"
   "liczba kolejek",
   "tur sayacı",
   "počet kol",
   "ходов потрачено",
   "Anzahl der Züge",
   "contador de turnos",
    }},
  {0x15a845e9, { // "F1 = help"
   "F1 = pomoc",
   "F1 = yardım",
   "F1 = nápověda",
   "F1 = помощь",
   "F1 = Hilfe",
/*MISSING*/    "F1 = help",
    }},
  {0x15ec3b67, { // "monster display mode"
   "tryb pokazywania potworów",
   "canavar görünümü",
   "mód zobrazování netvorů",
   "режим показа монстров",
   "Anzeigemodus (Monster)",
   "modo de exibição de monstros",
    }},
  {0x165d8a21, { // "what's this?"
   "co to jest?",
/*MISSING*/    "what's this?",
   "co to je?",
   "что это?",
   "Was ist das?",
/*MISSING*/    "what's this?",
    }},
  {0x1697e08a, { // "Rotate the camera in ball/hyperboloid model."
   "Obróć kamerę w modelu kuli/hiperboloidy.",
/*MISSING*/    "Rotate the camera in ball/hyperboloid model.",
   "Rotace kamery v modelu koule/hyperboloidu.",
   "Поворачивает камеру в модели сферы/гиперболоида.",
/*MISSING*/    "Rotate the camera in ball/hyperboloid model.",
/*MISSING*/    "Rotate the camera in ball/hyperboloid model.",
    }},
  {0x16e1d6d8, { // "A magical energy sword. Don't waste its power!"
   "Miecz energii. Nie zmarnuj jego mocy!",
/*MISSING*/    "A magical energy sword. Don't waste its power!",
   "Kouzelný energetický meč. Nepromarni jeho sílu!",
   "Магический меч энергии. Не тратьте его мощь зря!",
/*MISSING*/    "A magical energy sword. Don't waste its power!",
/*MISSING*/    "A magical energy sword. Don't waste its power!",
    }},
  {0x17160297, { // "Vikings believe that garnets improve their strength."
   "Wikingowie wierzą, że granaty dają im siłę.",
   "Vikingler nartaşlarının güçlerini artırdığını düşünür.",
   "Vikingové věří, že granáty zvyšují jejich sílu.",
   "Викинги верятЮ что гранаты дают им силу.",
   "Wikinger glauben das Granat sie stärker macht.",
/*MISSING*/    "Vikings believe that garnets improve their strength.",
    }},
  {0x17272102, { // "%The1 claws %the2!"
   "%1 zaatakowa%ł1 %a2!",
   "%1 %a2 pençeledi!",
   "%1 sek%l1 %a2!",
   "%1 атаковал%E1 %a2!",
   "%Der1 %1 zerkratzt %den2 %a2!",
   "%na1 arranha %na2!",
    }},
  {0x179cd5f0, { // "Bomberbird eggs are big and tasty, and thus valuable. They can hatch when left alone for some time (but this will never happen if you are watching)."
   "Jaja Bombardierów są wartościowe, bo są duże i smaczne. Mogą się wykluć, gdy dasz im czas (ale nie stanie się to, jeżeli patrzysz).",
   "Bombacıkuş yumurtaları büyük ve lezzetlidir, bu sebeple değerlidir. Eğer yalnız bırakılırlarsa çatlayabilirler. (Ama bu sen izlerken olmaz.)",
   "Vejce Bombarďáků jsou velká a chutná -- a tudíž i cenná. Mohou se vylíhnout, když je necháš nějaký čas na pokoji (ale to se nikdy nestane, když se na ně díváš).",
   "Яйца бомбардиров большие и вкусные, и потому ценные. Если оставть их в покое, может вылупиться бомбардир (но этого не случится, пока Вы на них смотрите).",
   "Bombenvogeleier sind groß und schmackhaft, daher wertvoll. Sie können nach einiger Zeit schlüpfen (aber machen dies nie, wenn du zusiehst).",
/*MISSING*/    "Bomberbird eggs are big and tasty, and thus valuable. They can hatch when left alone for some time (but this will never happen if you are watching).",
    }},
  {0x18dffecc, { // "Collect 5 Emeralds to access Camelot"
   "Zbierz 5 Szmaragdów, by dostać się do Camelot",
   "Kamelot'a ulaşmak için 5 Zümrüt topla.",
   "Získej 5 Smaragdů a dostaneš se do Camelotu",
   "Собери 5 изумрудов, чтобы открыть Камелот.",
   "Sammle 5 Smaragde um Zugang zu Camelot zu erhalten",
/*MISSING*/    "Collect 5 Emeralds to access Camelot",
    }},
  {0x18e20dc9, { // "You stun %the1!"
   "Ogłuszy%łeś0 %a1!",
   "%a1! sersemlettin!",
   "Omráči%l0 jsi %a1!",
   "Вы оглушили %a1!",
   "Du betäubst %den1 %a1!",
/*MISSING*/    "You stun %the1!",
    }},
  {0x18e20dd6, { // "You stun %the1."
   "Ogłuszy%łeś0 %a1.",
   "%a1 sersemlettin.",
   "Omráčil jsi %a1.",
   "Вы оглушили %a1.",
   "Du betäubst %den1 %a1.",
/*MISSING*/    "You stun %the1.",
    }},
  {0x18f803da, { // "items and monsters"
   "przedmioty i potwory",
   "eşyalar ve canavarlar",
   "předměty i netvoři",
   "предметы и монстры",
   "Items und Monster",
   "itens e monstros",
    }},
  {0x1923e5ca, { // "back to HyperRogue"
   "powrót do HyperRogue",
   "HyperRogue'a geri dön",
   "zpět do HyperRogue",
   "вернуться в игру",
   "Zurück zu HyperRogue",
/*MISSING*/    "back to HyperRogue",
    }},
  {0x193dd995, { // "Cannot summon on a monster!"
   "Nie można przywołać na potworze!",
   "Bir canavarın üstüne çağıramazsın!",
   "Na políčko s netvorem nelze nic vyvolat!",
   "Нельзя призвать монстра на занятую клетку!",
   "Kann nicht auf einem Monster beschworen werden!",
/*MISSING*/    "Cannot summon on a monster!",
    }},
  {0x1957aac2, { // "This Orb lets your allies to share your Orb powers.\n\nThe following Orbs are affected:"
   "Ta sfera pozwala Ci dzielić się mocami Sfer z przyjaciółmi.\n\nNastępujące Sfery są dzielone:",
/*MISSING*/    "This Orb lets your allies to share your Orb powers.\n\nThe following Orbs are affected:",
   "Tato Sféra ti umožňuje dělit se o moc tvých Sfér s kamarády.\n\nOvlivňuje následující Sféry:",
   "Эта сфера позволяет Вашим союзникам использовать силы Ваших сфер.\n\nДействует на следующие сферы:",
   "Dieser Orb erlaubt es deinen Verbündeten deine Orbkräfte zu nutzen.\n\nDie folgenden Orbs sind betroffen:",
/*MISSING*/    "This Orb lets your allies to share your Orb powers.\n\nThe following Orbs are affected:",
    }},
  {0x19a232a3, { // "A very cold land, full of ice walls. Your mere presence will cause these ice walls to melt, even if you don't want it."
   "Bardzo zimna kraina, pełna lodowych ścian. Wystarczy Twoja obecność, by te ściany się roztopiły. Nawet, jak tego nie chcesz.",
   "Buz duvarlarıyla dolu çok soğuk bir diyar. Sen istemesen de sadece varlığın bile duvarların erimesine sebep oluyor.",
   "Velmi chladný kraj plný ledových zdí. Tyto ledové zdi tají už tvou pouhou přítomností, ať už si to přeješ nebo ne.",
   "Очень холодная земля, полная ледяных стен. Они будут таять от Вашего присутствия, даже если Вы этого не хотите.",
   "Ein ziemlich kaltes Land - voller Eiswände. Deine Gegenwart allein wird sie allerdings zum Schmelzen bringen, ob du willst oder nicht.",
/*MISSING*/    "A very cold land, full of ice walls. Your mere presence will cause these ice walls to melt, even if you don't want it.",
    }},
  {0x19c78e5c, { // "gain kills"
   "zdobądź zabicia",
   "leşler kazan.",
   "získej zabité netvory",
   "получить убийства",
   "erhalte Tötungen",
/*MISSING*/    "gain kills",
    }},
  {0x19d3eb38, { // "No room to push %the1."
   "Brak miejsca, by popchnąć %a1.",
   "%a1 itmek için yeterince yer yok.",
   "%a1 není kam zatlačit.",
   "Нет места, чтобы толкать %a1.",
   "Kein Platz um %den1 %a1 zu schieben.",
   "Sem espaço para empurrar %na1.",
    }},
  {0x1a0f587d, { // "The Air Elemental blows you away!"
   "Powietrzny Żywiołak Cię zdmuchnie!",
   "Hava Unsuru seni uzağa uçuruyor!",
   "Elementál Vzduchu tě odfoukl!",
   "Воздушный элементаль сдул Вас!",
   "Der Luftelementar weht dich weg!",
/*MISSING*/    "The Air Elemental blows you away!",
    }},
  {0x1a161360, { // "This dangerous predator has killed many people, and has been sent to Cocytus."
   "Ten niebezpieczny drapieżnik zabił wielu ludzi i został zesłany do Kocytu.",
   "Bu tehlikeli yırtıcı çok kişiyi öldürdü ve Kokitos'a gönderildi.",
   "Tento nebezpečný predátor zabil spoustu lidí, a tak byl poslán do Cocytu.",
   "Этот опасный хищник убил немало людей и был отправлен в Коцит.",
   "Dieser gefürchtete Jäger hat schon viele das Leben gekostet und wurde nach Kokytos geschickt.",
/*MISSING*/    "This dangerous predator has killed many people, and has been sent to Cocytus.",
    }},
  {0x1a16ff81, { // "drop Dead Orb (up + down)"
   "połóż Martwą Sferę (góra+dół)",
/*MISSING*/    "drop Dead Orb (up + down)",
   "polož Mrtvou sféru (nahoru + dolů)",
   "положить Мёртвую сферу (вверх + вниз)",
/*MISSING*/    "drop Dead Orb (up + down)",
/*MISSING*/    "drop Dead Orb (up + down)",
    }},
  {0x1a8dced2, { // "0 = Klein model, 1 = Poincaré model"
   "0 = model Kleina, 1 = model Poincaré",
   "0 = Klein modeli, 1 = Poincaré modeli",
   "0 = Kleinův model, 1 = Poincarého model",
   "0 = модель Клейна, 1 = модель Пуанкаре",
   "0 = Klein Modell, 1 = Poincaré Modell",
   "0 = modelo Klein, 1 = modelo Poincaré",
    }},
  {0x1a8ded1b, { // "Horocycles are similar to circles, but you cannot reach their center at all -- they can be understood as limit circles of infinite radius centered in some point in infinity (also called an ideal point).\n\nGo to R'Lyeh, and you should quickly find a Temple of Cthulhu there. Each circle of columns is actually a horocycle. Horocycles in a given temple are concentric, and there is an infinite number of them."
   "Horocykle są podobne do okręgów, ale nie możesz dotrzeć do ich środka -- są to graniczne okręgi o nieskończonym promieniu, których środek jest w nieskończoności (takie punkty nazywamy punktami idealnymi).\n\nIdź do R'Lyeh, i szybko znajdziesz Świątynię Cthulhu.  Każdy krąg kolumn jest horocyklem. Horocykle w danej świątyni są współśrodkowe, i jest ich nieskończenie wiele.",
/*MISSING*/    "Horocycles are similar to circles, but you cannot reach their center at all -- they can be understood as limit circles of infinite radius centered in some point in infinity (also called an ideal point).\n\nGo to R'Lyeh, and you should quickly find a Temple of Cthulhu there. Each circle of columns is actually a horocycle. Horocycles in a given temple are concentric, and there is an infinite number of them.",
   "Horocykly se podobají kružnicím, ale jejich středu vůbec nelze dosáhnout -- můžeš si je představit jako limitní kružnice s nekonečným poloměrem, jejichž střed je v nějakém bodě v nekonečnu (takzvaném ideálním bodě).\n\nBěž do R'Lyeh. Měl bys tam rychle najít Cthulhuův chrám. Každý kruh ze sloupů je ve skutečnosti horocyklus. Horocykly v každém chrámu jsou soustředné a je jich nekonečně mnoho.",
   "Орициклы похожи на круги, но достичь их центра невозможно -- их можно понимать как предельные круги бесконечного радиуса с центром в точке на бесконечности (такие точки называются идеальными).\n\nИди в Р'Льех и найди Храм Ктулху. Каждый круг колонн на самом деле орицикл. Орициклы в храме концентричны, и их число бесконечно.",
/*MISSING*/    "Horocycles are similar to circles, but you cannot reach their center at all -- they can be understood as limit circles of infinite radius centered in some point in infinity (also called an ideal point).\n\nGo to R'Lyeh, and you should quickly find a Temple of Cthulhu there. Each circle of columns is actually a horocycle. Horocycles in a given temple are concentric, and there is an infinite number of them.",
/*MISSING*/    "Horocycles are similar to circles, but you cannot reach their center at all -- they can be understood as limit circles of infinite radius centered in some point in infinity (also called an ideal point).\n\nGo to R'Lyeh, and you should quickly find a Temple of Cthulhu there. Each circle of columns is actually a horocycle. Horocycles in a given temple are concentric, and there is an infinite number of them.",
    }},
  {0x1a9e2e11, { // "You need special glasses to view the game in 3D"
   "Potrzebujesz specjalnych okularów, by oglądać grę w 3D",
   "Oyunu 3D'de görüntülemek için özel gözlüklere ihtiyacın var.",
   "Abys moh%l0 hru sledovat ve 3D, potřebuješ speciální brýle",
   "Вам нужны специальные очки, чтобы видеть игру в 3D",
   "Du benötigst eine 3D Brille für den 3D Modus",
   "Você precisará de óculos especiais para ver o jogo em 3D",
    }},
  {0x1ac03d40, { // "Good to know that your fighting skills serve you well in this strange world."
   "Dobrze wiedzieć, że Twoje zdolności walki przydają się w tym świecie.",
   "Dövüş yeteneklerinin bu garip dünyada hâlâ işe yaradığını bilmek güzel.",
   "Je dobré vědět, že tvé bojové dovednosti ti v tomto podivném světě slouží dobře.",
   "Полезно знать, что Ваши боевые навыки неплохо работают в этом странном мире.",
   "Gut zu wissen, dass deine Kampfkünste dir in dieser ungewöhnlichen Welt helfen.",
   "É bom saber que suas habilidades de luta lhe servem bem nesse mundo estranho.",
    }},
  {0x1ae417df, { // "choose"
   "wybór",
/*MISSING*/    "choose",
   "výbìr",
   "выбор",
/*MISSING*/    "choose",
/*MISSING*/    "choose",
    }},
  {0x1b09e182, { // "Hell has these lakes everywhere... They are shaped like evil stars, and filled with burning sulphur."
   "Piekło jest pełne tych jezior pełnych płonącej siarki, w kształcie złych gwiazd...",
   "Cehennemde her yerde bunun gibi göller var. Şeytani star şeklinde ve yanan sülfürle dolu...",
   "Tato jezera se nacházejí po celém Pekle... Mají tvar zlých hvězd a jsou naplněna hořící sírou.",
   "В Аду эти озёра повсюду. Они похожи на звёзды и наполнны кипящей серой.",
   "Diese Seen sind überall in der Hölle... Sie sehen aus wie böse Sterne, und sind voll mit brennendem Schwefel.",
/*MISSING*/    "Hell has these lakes everywhere... They are shaped like evil stars, and filled with burning sulphur.",
    }},
  {0x1b607333, { // "animation speed"
   "prędkość animacji",
   "animasyon hızı",
   "rychlost animace",
   "скорость анимации",
   "Animationsgeschwindigkeit",
   "velocidade de animação",
    }},
#if REPEATED
  {0x1b607334, { // "animation speed [repeat]"
   "prędkość animacji",
/*MISSING*/    "animation speed [repeat]",
   "rychlost animace",
   "скорость анимации",
   "Animationsgeschwindigkeit",
/*MISSING*/    "animation speed [repeat]",
    }},
#endif
  {0x1b9d6e68, { // "ball model"
   "model kuli",
/*MISSING*/    "ball model",
   "model koule",
   "модель сферы",
/*MISSING*/    "ball model",
/*MISSING*/    "ball model",
    }},
  {0x1bf59600, { // "WARNING: you are entering a minefield!"
   "UWAGA: wchodzisz na pole minowe!",
   "UYARI: bir mayıntarlasına giriyorsun!",
   "VAROVÁNÍ: vstupuješ do minového pole!",
   "ВНИМАНИЕ: вы вошли на минное поле!",
   "ACHTUNG: Du betrittst ein Minenfeld!",
/*MISSING*/    "WARNING: you are entering a minefield!",
    }},
  {0x1c7e60fd, { // "That's all you need to unlock the Orb of Yendor! Well... as long as you are able to return to the Orb that this key unlocks...\n\nEach key unlocks only the Orb of Yendor which led you to it."
   "To jest to, czego potrzebujesz, by otworzyć Sferę Yendoru! O ile umiesz do niej trafić...\nKażdy klucz otwiera tylko jedną Sferę, która Cię do niego doprowadziła.",
   "Yendorun Küresinin kilidini açmak için ihtiyacınız olan her şeyi tamamladın. Tabi... Önce Küreyr dönmen gerekiyor...\nHer anahtar seni ona yönlendiren yendorun küresini açacaktır.",
   "To je všechno, co potřebuješ k odemčení Yendorské sféry! No... pokud se tedy dokážeš vrátit ke sféře, kterou tenhle klíč odemyká...\nKaždý klíč odemyká pouze tu Yendorskou sféru, která tě k němu zavedla.",
   "Это всё, что Вам нужно, чтобы открыть Сферу Йендора! Ну, если Вы сможете к ней вернуться...\nКаждый ключ открывает лишь ту сферу, которая на него указывает.",
   "Das ist alles, was du brauchst, um den Orb von Yendor zu öffnen! ...solange du es schaffst, zu dem Orb zurückzufinden, zu dem dieser Schlüssel passt...\nJeder Schlüssel entsperrt nur den Orb von Yendor, der dich zu ihm geführt hat.",
/*MISSING*/    "That's all you need to unlock the Orb of Yendor! Well... as long as you are able to return to the Orb that this key unlocks...\n\nEach key unlocks only the Orb of Yendor which led you to it.",
    }},
  {0x1c8904a7, { // "weapon color"
   "kolor broni",
   "silah rengi",
   "barva zbraně",
   "цвет оружия",
   "Waffenfarbe",
/*MISSING*/    "weapon color",
    }},
  {0x1c893324, { // "Your energy swords get stronger!"
   "Twoje ostrza energii stały się silniejsze!",
/*MISSING*/    "Your energy swords get stronger!",
   "Tvé energetické meče sílí!",
   "Твои энергетические мечи стали сильнее!",
/*MISSING*/    "Your energy swords get stronger!",
/*MISSING*/    "Your energy swords get stronger!",
    }},
  {0x1c8ff97a, { // "Hint: magical powers from Treats expire after a number of uses."
   "Wskazówka: magiczne moce Cukierków wyczerpują się po kilku użyciach.",
/*MISSING*/    "Hint: magical powers from Treats expire after a number of uses.",
   "Nápověda: magické schopnosti získané z Pochoutek zmizí po určitém počtu použití.",
   "Подсказка: магические силы конфет пропадают после некоторого числа использований.",
/*MISSING*/    "Hint: magical powers from Treats expire after a number of uses.",
/*MISSING*/    "Hint: magical powers from Treats expire after a number of uses.",
    }},
  {0x1c93c602, { // "Hyperbolic geometry has been discovered by the 19th century mathematicians who wondered about the nature of paralellness. Take a line L and a point A. Can a world exist where there is more than one line passing through A which does not cross L?\n\nThe Icy Land will be very dangerous if you have lots of Ice Diamonds -- lots of Yetis and Ice Wolves hunting you! But the other lands, where you have no treasures yet, will still be (relatively) safe.\n\nWander further, and you should find Crossroads quickly -- the Great Walls are straight lines, and indeed, they work differently than in Euclidean. On the other side of Great Walls, you see other lands -- there are about 50 lands in HyperRogue, based on different mechanics and aspects of hyperbolic geometry."
   "Geometria hiperboliczna została odkryta przez dziewiętnastowiecznych matematyków zainteresowanych naturą równoległości. Weź linię prostą L i punkt A. Czy może istnieć świat, w którym jest więcej niż jedna linia prosta przechodząca przez A, ale nie przecinająca L?\n\nKraina Lodu jest bardzo niebezpieczna, gdy masz mnóstwo Diamentów. Ale inne krainy, w których jeszcze nie masz skarbów, są w miarę bezpieczne.\n\nZwiedzaj dalej, a szybko znajdziesz Skrzyżowanie -- jego ściany są liniami prostymi, i działają inaczej niż w geometrii euklidesowej. Po drugiej stronie Wielkich Ścian widzisz inne krainy -- jest ich około 50, opartych na róznych aspektach geometrii hiperbolicznej.",
/*MISSING*/    "Hyperbolic geometry has been discovered by the 19th century mathematicians who wondered about the nature of paralellness. Take a line L and a point A. Can a world exist where there is more than one line passing through A which does not cross L?\n\nThe Icy Land will be very dangerous if you have lots of Ice Diamonds -- lots of Yetis and Ice Wolves hunting you! But the other lands, where you have no treasures yet, will still be (relatively) safe.\n\nWander further, and you should find Crossroads quickly -- the Great Walls are straight lines, and indeed, they work differently than in Euclidean. On the other side of Great Walls, you see other lands -- there are about 50 lands in HyperRogue, based on different mechanics and aspects of hyperbolic geometry.",
   "Hyperbolickou geometrii objevili matematikové z 19. století, kteří přemýšleli o povaze rovnoběžnosti. Máš-li přímku p a bod A, může existovat svět, ve kterém bodem A prochází více než jedna přímka, která neprotíná p?\n\nLedový kraj bude velmi nebezpečný, pokud máte hodně Ledových diamantů -- potom tě bude honit spousta Yetiů a Ledových vlků! Ale ostatní kraje, ve kterých ještě žádné poklady nemáš, budou stále bezpečné (relativně).\n\nPůjdeš-li dál, měl bys rychle narazit na Křižovatku -- Velké zdi jsou přímky a skutečně fungují jinak než v eukleidovské geometrii. Na druhé straně Velkých zdí uvidíš jiné kraje -- HyperRogue obsahuje zhruba 50 krajů založených na různých mechanikách a aspektech hyperbolické geometrie.",
   "Гиперболическая геометрия была открыта в 19 веке математиками, которые интересовались природой параллельности. Возьмём прямую L и точку A. Может ли существовать мир, в котором есть более чем одна прямая, проходящая через A и не пересекающая L?\n\nЛедяная земля становится очень опасной, когда у тебя много Ледяных алмазов -- много Йети и Ледяных волков охотятся на тебя! Но другие земли, где у тебя нет сокровищ, всё ещё (относительно) безопасны.\n\nИсследуй дальше, и ты скоро найдёшь Перекрёсток -- Великие стены являются прямыми, и они ведут себя не так, как в евклидовой геометрии. По другую сторону от Великих стен ты видишь другие земли -- их около 50, основанных на разных особенностях гиперболической геометрии.",
/*MISSING*/    "Hyperbolic geometry has been discovered by the 19th century mathematicians who wondered about the nature of paralellness. Take a line L and a point A. Can a world exist where there is more than one line passing through A which does not cross L?\n\nThe Icy Land will be very dangerous if you have lots of Ice Diamonds -- lots of Yetis and Ice Wolves hunting you! But the other lands, where you have no treasures yet, will still be (relatively) safe.\n\nWander further, and you should find Crossroads quickly -- the Great Walls are straight lines, and indeed, they work differently than in Euclidean. On the other side of Great Walls, you see other lands -- there are about 50 lands in HyperRogue, based on different mechanics and aspects of hyperbolic geometry.",
/*MISSING*/    "Hyperbolic geometry has been discovered by the 19th century mathematicians who wondered about the nature of paralellness. Take a line L and a point A. Can a world exist where there is more than one line passing through A which does not cross L?\n\nThe Icy Land will be very dangerous if you have lots of Ice Diamonds -- lots of Yetis and Ice Wolves hunting you! But the other lands, where you have no treasures yet, will still be (relatively) safe.\n\nWander further, and you should find Crossroads quickly -- the Great Walls are straight lines, and indeed, they work differently than in Euclidean. On the other side of Great Walls, you see other lands -- there are about 50 lands in HyperRogue, based on different mechanics and aspects of hyperbolic geometry.",
    }},
  {0x1cba92db, { // "summon orbs"
   "przwołaj sfery",
   "küre çıkar",
   "vyvolej sféry",
   "призвать сферы",
   "rufe Orbs herbei",
/*MISSING*/    "summon orbs",
    }},
  {0x1cc14d8a, { // "A long time ago, this was a trade route. But then, Krakens have risen out of the depths. Many trading ships sank here. Legend says that you can uncover the secret of a magical weapon spell somewhere in the depths...\n\nYou can find Sunken Treasures here, but they won't appear until you have killed a Kraken. You will also need Orb of the Fish to get the treasures, luckily you can steal one from the Viking treasure hunters."
   "Dawno, dawno temu, to był szlak handlowy, ale Krakeny wypłynęły z głębin. Wiele statków handlowych zatonęła. Według legendy, gdzieś w głębinach ukryty jest sekret magicznej broni...\n\nMożesz tu znaleźć Zatopione Skarby, ale nie będą się one pojawiać, aż pokonasz Krakena. By wyłowić skarby, potrzebna jest Sfera Ryby. Na szczęście można ukraść ją od wikińskich łowców skarbów.",
/*MISSING*/    "A long time ago, this was a trade route. But then, Krakens have risen out of the depths. Many trading ships sank here. Legend says that you can uncover the secret of a magical weapon spell somewhere in the depths...\n\nYou can find Sunken Treasures here, but they won't appear until you have killed a Kraken. You will also need Orb of the Fish to get the treasures, luckily you can steal one from the Viking treasure hunters.",
   "Tohle byla kdysi dávno obchodní trasa. Jenomže potom se z hlubin vynořili Krakeni. Mnoho obchodních lodí se tu potopilo. Legenda praví, že kdesi v hlubinách můžete najít tajemství kouzla, které vytváří magickou zbraň...\n\nV Hlubinách lze najít Potopené poklady, ale ty se objeví teprve poté, co zabiješ alespoň jednoho Krakena. K získání pokladů budeš také potřebovat Sféru Ryby, ale ty se naštěstí dají ukrást vikingským lovcům pokladů.",
   "Когда-то давно здесь были торговые пути. Но потом Кракены появились из глубин. Много торговых кораблей затонуло здесь. Легенды гласят, что где-то здесь можно найти тайны магического оружия...\n\nЗдесь можно найти затонувшие сокровища, но их нельзя найти, пока не убьёшь одного Кракена. Кроме того, чтобы до них добраться нужна Сфера Рыбы. К счастью, её можно украсть у Викингов, охотников за сокровищами.",
   "Vor langer Zeit war dies eine Handelsroute. Dann aber stiegen Kraken aus den Tiefen empor. Viele Handelsschiffe versanken hier. Legenden behaupten, dass man das Geheimnis eines Zaubers für eine magische Waffe irgendwo in den Tiefen enthüllen kann...\n\nHier kannst du versunkene Schätze finden, aber erst nachdem du einen Kraken getötet hast. Du wirst auch den Orb des Fischs brauchen, um zu den Schätzen zu gelangen.  Glücklicherweise kannst du einen von den Wikinger-Schatzjägern stehlen.",
/*MISSING*/    "A long time ago, this was a trade route. But then, Krakens have risen out of the depths. Many trading ships sank here. Legend says that you can uncover the secret of a magical weapon spell somewhere in the depths...\n\nYou can find Sunken Treasures here, but they won't appear until you have killed a Kraken. You will also need Orb of the Fish to get the treasures, luckily you can steal one from the Viking treasure hunters.",
    }},
  {0x1cf38c60, { // "Running Dogs"
   "Biegnące Psy",
/*MISSING*/    "Running Dogs",
   "Běžící psi",
   "Бегущие собаки",
/*MISSING*/    "Running Dogs",
/*MISSING*/    "Running Dogs",
    }},
  {0x1cfc7dd3, { // "hexagonal #1"
   "sześciokąt #1",
/*MISSING*/    "hexagonal #1",
   "šestiúhelníkový #1",
   "шестиугольник #1",
   "hexagonal #1",
/*MISSING*/    "hexagonal #1",
    }},
  {0x1cfcaab2, { // "These orbs can be found in the Graveyard. You think that they were once powerful magical orbs, but by now, their power is long gone. No way to use them, you could as well simply drop them...\n\n"
   "Na cmentarzu jest mnóstwo tych sfer. Chyba były to kiedyś potężne magiczne kule, ale ich moc dawno przeminęła. Nie ma jak ich użyć, także może lepiej je po prostu zostawić...\n\n",
   "Bu küreler Mezarlıkta bulunabilir. Eskiden kudretli sihrili kürelerdi, ama şimdi güçleri bitti. Onları kullanmanın bir yolu yok, bırakın gitsin...\n\n",
   "Tyto sféry můžeš najít na Hřbitově. Máš dojem, že to kdysi byly mocné magické sféry, ale jejich moc už je dávno pryč. Nedají se nijak použít, můžeš je rovnou zase položit...\n\n",
   "Эту сферу можно найти на Кладбище. Когда-то она обладала могущественной магией, но теперь её сила иссякла. Нет способа её использовать. Разве что просто бросить её...\n\n",
   "Diese Orbs kannst du auf dem Friedhof finden. Du glaubst, dass sie einst mächtige magische Orbs waren, aber inzwischen ist ihre Kraft längst erloschen. Sie sind für dich nutzlos; du könntest sie auch einfach ablegen...\n\n",
/*MISSING*/    "These orbs can be found in the Graveyard. You think that they were once powerful magical orbs, but by now, their power is long gone. No way to use them, you could as well simply drop them...\n\n",
    }},
  {0x1d1d9ed1, { // "field quotient"
   "przestrzeń ilorazowa ciała",
/*MISSING*/    "field quotient",
   "tělesový kvocient",
   "фактор поля",
/*MISSING*/    "field quotient",
/*MISSING*/    "field quotient",
    }},
  {0x1d35548f, { // "Level of water surface"
   "Poziom powierzchni wody",
/*MISSING*/    "Level of water surface",
   "Úroveň vodní hladiny",
   "Уровень поверхности воды",
/*MISSING*/    "Level of water surface",
/*MISSING*/    "Level of water surface",
    }},
  {0x1d412462, { // " (touch to activate)"
   " (dotknij, by aktywować)",
   " (etkinleştirmek için dokun)",
   " (aktivuj dotekem)",
   " (коснитесь, чтобы активировать)",
   " (Zum Aktivieren tippen)",
   " (toque para ativar)",
    }},
  {0x1dd141d2, { // "A mine is next to you!"
   "Jedna mina koło Ciebie!",
   "Yakınında bir mayın var!",
   "Ve tvé blízkosti se nachází jedna mina!",
   "Одна мина рядом с Вами!",
   "1 Mine um dich!",
/*MISSING*/    "A mine is next to you!",
    }},
  {0x1dd6b361, { // "PARTIAL"
   "CZĘŚCIOWO",
   "KISMÎ",
   "CZĘŚCIOWO",
   "ЧАСТИЧНО",
   "TEILWEISE",
/*MISSING*/    "PARTIAL",
    }},
  {0x1debe9ec, { // "(Menu button) and select the ASCII mode, which runs much faster. Depending on your device, turning the OpenGL mode on or off might make it faster, slower, or cause glitches."
   "(Menu), tryb ASCII działa dużo szybciej. Tryb OpenGL również wpływa na jakość grafiki.",
   "(Menü), ve ASCII modunu seçebilirsin. Bu mod daha hızlı çalışır. Aygıtınıza bağlı olarak OpenGL'yi açmak oyununuzu hızlandırabilir, yavaşlatabilir veya hatalara sebep olabilir.",
   "(tlačítko 'Menu') a zvolit ASCII mód, který je mnohem rychlejší. Zapnutí či vypnutí módu OpenGL může, v závislosti na tvém zařízení, hru zrychlit, zpomalit, nebo vyvolat grafické chyby.",
   "(Меню), выберите ASCII режим, который работает быстрее. В зависимости от Вашего устройства включите или выключите OpenGL. Он может быть как быстрее, так и медленнее. ",
   "(Im Menü) Versuche den ASCII-Modus zu aktivieren - der läuft viel schneller! Je nach Gerät kann der OpenGL-Modus das Spiel beschleunigen, verlangsamen, oder Bildfehler hervorrufen.",
   "(Botão Menu), e selecionar o modo ASCII, que roda bem mais rápido. Dependendo do seu dispositivo, ligar ou desligar o modo OpenGL pode fazê-lo rodar mais rápido, mais lento ou causar glitches.",
    }},
  {0x1df00010, { // "openGL mode disabled"
   "wyłączono tryb OpenGL",
   "OpenGL modu devredışı",
   "mód openGL vypnut",
   "режим OpenGL выключен",
   "OpenGL deaktiviert",
   "Modo OpenGL desabilitado",
    }},
  {0x1df9745d, { // "This guard of the Emerald Mine is wielding a huge flail. You cannot attack him directly, as the flail would still hit you then. Luckily, you have learned a trick: if you step away from him, he will hit himself with the flail!"
   "Ten strażnik Kopalni Szmaragdów jest uzbrojony w wielki cep bojowy. Nie możesz go zaatakować bezpośrednio, bo cep by Cię uderzył. Na szczęście, znasz sztuczkę: jak się od niego odsuniesz, strażnik uderzy cepem samego siebie!",
   "Zümrüt Madeninin bu muhafızı kocaman bır topuz tutuyor. Ona doğrudan saldıramazsın, çünkü topuz sana yine de vurur. Çok şükür bir hile öğrendin: Eğer geriye doğru bir adım atarsan, muhafız topuzuyla kendisine vuracak!",
   "Tento strážce Smaragdového dolu je vyzbrojen obrovským řemdihem. Nemůžeš na něj útočit přímo, protože by tě řemdih i tak zasáhl. Naštěstí ses naučil jeden trik: pokud se pohneš směrem od něj, zasáhne řemdihem sám sebe!",
   "Этот страж изумрудных шахт вооружён огромным цепом. Вы не можете атаковать его обычным путём, потому что тогда цеп убьёт Вас. К счастью, Вы знаете трюк: достаточно отойти от него, когда он замахивается, и он убьёт сам себя!",
   "Dieser Wächter der Smaragdmine hat einen riesigen Morgenstern. Ihn direkt anzugreifen würde nichts bringen da der Morgenstern dich treffen würde. Glücklicherweise kennst du einen Trick: Wenn du dich von ihm entfernst, trifft er sich selbst!",
/*MISSING*/    "This guard of the Emerald Mine is wielding a huge flail. You cannot attack him directly, as the flail would still hit you then. Luckily, you have learned a trick: if you step away from him, he will hit himself with the flail!",
    }},
  {0x1e3e87b7, { // "The Knights wish you luck!"
   "Rycerze życzą powodzenia!",
   "Şövalyeler sana şans diliyor!",
   "Rytíři ti přejí hodně štěstí!",
   "Рыцари желают Вам удачи!",
   "Die Ritter wünschen dir Glück!",
/*MISSING*/    "The Knights wish you luck!",
    }},
  {0x1e46fe41, { // "Wherever this powerful being goes, the living fjord sinks below water, non-magical boats are destroyed, and fires are extinguished.\n\nAs a special case, you can attack the Water Elemental from the water, without drowning immediately."
   "Gdziekolwiek przejdzie ta potężna istota, żywy fiord tonie, niemagiczne łódki są niszczone, a ogień gaśnie.\n\nJako szczególny przypadek, możesz atakować Wodnego Żywiołaka z wody, nie topiąc się.",
   "Bu güçlü yaratığın geçtiği her yerde Yaşayan Fiyort su altına gömülür, sihirli olmayan kayıklar yok olur ve ateşler söner. \n\n İstisna bir durum olarak Su Özütüne suyun içerisinden hemen boğulmadan vurursan onu öldürebilirsin. ",
   "Kamkoli se vydá tato mocná bytost, tam se Živoucí fjord potápí pod vodu, všechny čluny kromě magických jsou zničeny a veškerý oheň je uhašen.\n\nNa Elementála Vody můžeš zaútočit z vody, aniž by ses okamžitě utopil; to je speciální případ.",
   "Где бы ни шло это могущественное существо, фьорды расступаются перед ним, обычные лодки уничтожаются, огонь гаснет.\n\nОднако, если Вы атакуете его, находяст в воде, Вас не затянет сразу в воду.",
   "Wohin auch immer sich dieses Wesen bewegt sinkt der lebende Fjord. Dadurch werden nicht magische Boote zerstört und Feuer gelöscht.\n\nDie Besonderheit ist: du kannst den Wasserelementar vom Wasser aus angreifen ohne sofort zu ertrinken.",
/*MISSING*/    "Wherever this powerful being goes, the living fjord sinks below water, non-magical boats are destroyed, and fires are extinguished.\n\nAs a special case, you can attack the Water Elemental from the water, without drowning immediately.",
    }},
  {0x1e8e4977, { // "restart"
   "nowa gra",
/*MISSING*/    "restart",
   "nová hra",
   "новая игра",
/*MISSING*/    "restart",
/*MISSING*/    "restart",
    }},
  {0x1e9407e4, { // "Failed to load the file 'papermodeldata.txt'"
   "Nie udało się wczytać papermodeldata.txt",
   "papermodeldata.txt dosyası yüklenemedi",
   "Soubor 'papermodeldata.txt' se nepodařilo nahrát",
   "Невозможно загрузить файл papermodeldata.txt",
   "papermodeldata.txt konnte nicht geladen werden.",
/*MISSING*/    "Failed to load the file 'papermodeldata.txt'",
    }},
  {0x1ebefc13, { // "This Orb lets you jump to a place which is two cell away from you, in a single turn. You can jump over water, chasms and fire, but not over walls."
   "Ta Sfera pozwala na skok do miejsca odległego od Ciebie o dwa pola, w jednej kolejce. Możesz przeskoczyć wodę, przepaść i ogień, ale nie ścianę.",
   "Bu Küre tek tur içinde senden iki hücre uzağa zıplamanı sağlar.",
   "Tato Sféra ti umožňuje skočit v jednom kole dvě pole daleko. Můžeš skákat přes vodu, propasti a oheň, ale ne přes zdi.",
   "Эта сфера позволяет прыгать на 2 клетки за один ход. Вы можете прыгать через воду, огонь или пропасть, но не через стены.",
   "Dieser Orb lässt dich in einem Zug an einen Ort springen der 2 Zellen entfernt ist. Du kannst über Wasser, Feuer und Spalten springen aber nicht über Wände.",
/*MISSING*/    "This Orb lets you jump to a place which is two cell away from you, in a single turn. You can jump over water, chasms and fire, but not over walls.",
    }},
  {0x1ec2ea5e, { // "This orb lets you go through living walls. It also has powers in some of the other lands."
   "Ta sfera pozwala Ci przechodzić przez żywą ścianę. Ma też moce w niektórych z pozostałych krain.",
   "Bu küre Yaşayan Duvarların içinden geçebilmenizi sağlar.",
   "Tato sféra ti umožňuje procházet živoucími zdmi. Má také určité schopnosti v některých jiných krajích.",
   "Эта сфера позволяет проходить сквозь живые стены. Она имеет силу и в некоторых других землях.",
   "Dieser Orb erlaubt es dir, dich durch lebende Wände zu bewegen und er besitzt auch einige Kräfte in anderen Ländern.",
/*MISSING*/    "This orb lets you go through living walls. It also has powers in some of the other lands.",
    }},
  {0x1edbcc1a, { // "After you move while having this Orb, you immediately attack the next cell in the straight line (or two cells, when moving on a heptagon). This attack is slightly stronger than your normal attack: it can stun some of the monsters which cannot be killed or stunned normally."
   "Po każdym ruchu z tą Sferą, natychmiast atakujesz kolejne pole w linii prostej (albo dwa pola, gdy jesteś na siedmiokącie). Ten atak jest trochę silniejszy od Twojego zwykłego ataku: niektóre potwory niewrażliwe na zwykły atak można w ten sposób ogłuszyć.",
/*MISSING*/    "After you move while having this Orb, you immediately attack the next cell in the straight line (or two cells, when moving on a heptagon). This attack is slightly stronger than your normal attack: it can stun some of the monsters which cannot be killed or stunned normally.",
   "Kdykoli se pohneš s touto Sférou, okamžitě zaútočíš na další políčko v řadě (nebo na dvě, pokud se pohneš na sedmiúhelník). Tento útok je o něco silnější než tvůj běžný útok: dokáže omráčit některé netvory, které nelze zabít ani omráčit běžným způsobem.",
   "После хода с этой сферой вы сразу же атакуете следующую клетку прямой (или две клетки, если ходите на семиугольник). Эта атака немного сильнее обычной: она оглушает некоторых существ, которых нельзя убить или оглушить обычным ударом.",
/*MISSING*/    "After you move while having this Orb, you immediately attack the next cell in the straight line (or two cells, when moving on a heptagon). This attack is slightly stronger than your normal attack: it can stun some of the monsters which cannot be killed or stunned normally.",
/*MISSING*/    "After you move while having this Orb, you immediately attack the next cell in the straight line (or two cells, when moving on a heptagon). This attack is slightly stronger than your normal attack: it can stun some of the monsters which cannot be killed or stunned normally.",
    }},
  {0x1eedd5be, { // "Show me!"
   "Pokaż!",
/*MISSING*/    "Show me!",
   "Ukaž!",
   "Покажи!",
/*MISSING*/    "Show me!",
/*MISSING*/    "Show me!",
    }},
  {0x1f3db1c6, { // "%The1 squeaks gratefully!"
   "%1 pisn%ął1 w podziękowaniu!",
   "%1 minnetle ciyaklıyor.",
   "%1 vděčně piští!",
   "%1 пищит с благодарностью!",
   "%Die1 %1 piepst dankbar!",
/*MISSING*/    "%The1 squeaks gratefully!",
    }},
  {0x1f4c620a, { // "Introduction"
   "Wstęp",
/*MISSING*/    "Introduction",
   "Úvod",
   "Введение",
/*MISSING*/    "Introduction",
/*MISSING*/    "Introduction",
    }},
  {0x1f6d96e3, { // "This orb lets you instantly move to another location on the map. Just click a location which is not next to you to teleport there. "
   "Ta sfera pozwala Ci natychmiast przenieść się do innego miejsca na mapie. Wystarczy kliknąć pole, które nie jest obok Ciebie.",
   "Bu küre haritadaki balşka bir noktaya anında ışınlanmanızı sağlayacak. Işınlanmak için sadece yanında olmayan bir yere tıkla.",
   "Tato sféra ti umožňuje se okamžitě přemístit na jiné místo na mapě. Stačí, abys klikl na libovolné nesousední políčko.",
   "Эта сфера позволяет перемещаться в произвольное место. Просто кликните по не соседней с Вами клетке, чтобы телепортироваться туда.",
   "Dieser Orb versetzt dich an einen anderen Ort auf der Karte. Klicke einfach einen Ort an, der sich nicht neben dir befindet, um dorthin zu teleportieren.",
/*MISSING*/    "This orb lets you instantly move to another location on the map. Just click a location which is not next to you to teleport there. ",
    }},
  {0x1f8ebbbf, { // "An ancient sunken city which can be reached only when the stars are right.\n\nYou can find Temples of Cthulhu in R'Lyeh once you collect five Statues of Cthulhu."
   "Prastare zatopione miasto, do którego można dostać się tylko, gdy gwiazdy są na swoich miejscach.\n\nPo znalezieniu 5 Statuetek Cthulhu możesz w R'Lyeh trafić na Świątynie Cthulhu.",
   "Sadece yıldızlar burayı işaret ettiğinde ulaşılabilecek batık bir antik şehir.\n\nCthulhu'nun tapınaklarını 5 Cthulhu heykeli topladıktan sonra bulabilirsin.",
   "Prastaré potopené město, kam se lze dostat pouze při správném postavení hvězd.\n\nPo znalezieniu 5 Statuetek Cthulhu możesz w R'Lyeh trafić na Świątynie Cthulhu.",
   "Этот древний затопленный город можно найти лишь тогда, когда на него укажут звёзды.\n\nВы сможете найти храмы Ктулху, как только соберёте 5 статуй Ктулху.",
   "Eine antike, versunkene Stadt, die du nur erreichen kannst, wenn die Sterne günstig stehen.\n\nDer Tempel des Cthulhu befindet sich in R'Lyeh. Sobald du fünf Statuen von Cthulhu gesammelt hast, kannst du ihn finden.",
/*MISSING*/    "An ancient sunken city which can be reached only when the stars are right.\n\nYou can find Temples of Cthulhu in R'Lyeh once you collect five Statues of Cthulhu.",
    }},
  {0x1f9a904e, { // "native to %the1 (collect 1 %2)"
   "rodzima %abl1 (1x %2)",
   "%abl1 bulunuyor (1 tane %2)",
   "standardní %abl1 (získej 1x %2)",
   "родная %abl1 (1x %2)",
   "heimisch in %der1 (1x %2)",
/*MISSING*/    "native to %the1 (collect 1 %2)",
    }},
  {0x1fe99f69, { // "video resolution"
   "rozdzielczość obrazu",
   "çözünürlük",
   "rozlišení",
   "видео разрешение",
   "Videoauflösung",
   "resolução de vídeo",
    }},
  {0x1ff2a5f5, { // "Corals have a somewhat hyperbolic structure even in your home world, but natural corals from the Warped Sea have truly beautiful shapes. Ratlings know the value of corals, and thus keep them in boats for safety."
   "Korale mają hiperboliczną strukturę nawet w świecie, z którego pochodzisz, ale naturalne korale z Zakrzywionego Morza mają prawdziwie piękne kształty. Szczuraki znają wartość korali i dla bezpieczeństwa trzymają je na łódkach.",
/*MISSING*/    "Corals have a somewhat hyperbolic structure even in your home world, but natural corals from the Warped Sea have truly beautiful shapes. Ratlings know the value of corals, and thus keep them in boats for safety.",
   "Korály mají do jisté míry hyperbolickou strukturu i ve tvém domovském světě, ale tvary přírodních korálů z Pokřiveného moře jsou skutečně nádherné. Krysáci znají hodnotu korálů, a kvůli bezpečnosti je přechovávají ve člunech.",
   "Кораллы имеют структуру, похожую на гиперболическую, даже в Вашем обычном мире, но настоящие кораллы из Искривлённого моря имеют действительно красивые формы. Крысолюды знают цену кораллам и на всякий случай хранят из в лодках.",
   "Korallen haben auch in deiner Welt eine etwas hyperbolische Struktur. Aber natürliche Korallen der Verzerrten See haben besonders schöne Formen. Rattenmenschen kennen den Wert der Korallen und bewahren sie daher zur Sicherheit in Booten auf.",
/*MISSING*/    "Corals have a somewhat hyperbolic structure even in your home world, but natural corals from the Warped Sea have truly beautiful shapes. Ratlings know the value of corals, and thus keep them in boats for safety.",
    }},
  {0x200536cf, { // "heptagonal game board"
   "plansza do gry z siedmiokątów",
   "yedigensel tahta",
   "sedmiúhelníková herní deska",
   "семиугольная игровая доска",
   "Heptagonales Spielbrett",
   "tabuleiro de jogo heptagonal",
    }},
  {0x202534e2, { // "Collect as many Dodecahedra as you can, then return here!"
   "Zbierz jak najwięcej Dwunastościanów i tu wróć!",
/*MISSING*/    "Collect as many Dodecahedra as you can, then return here!",
   "Získej co možná nejvíc Dvanáctistěnů, a potom se vrať sem!",
   "Собери как можно больше Додекаэдров и вернись сюда!",
/*MISSING*/    "Collect as many Dodecahedra as you can, then return here!",
/*MISSING*/    "Collect as many Dodecahedra as you can, then return here!",
    }},
  {0x204ae3f4, { // "\n\nThis is a friendly being. It does not count for your total kills."
   "\n\nTo jest przyjazne stworzenie. Nie liczy się do łącznych zabić.",
/*MISSING*/    "\n\nThis is a friendly being. It does not count for your total kills.",
   "\n\nToto je přátelská bytost. Nepočítá se do celkového počtu zabitých netvorů.",
   "\n\nЭто Ваш союзник. Не учитывается в количестве убийств.",
   "\n\nDas ist ein freundliches Wesen. Es wird der Gesamtzahl deiner Tötungen nicht eingerechnet.",
/*MISSING*/    "\n\nThis is a friendly being. It does not count for your total kills.",
    }},
  {0x207af4ad, { // "The inscriptions on the Statue of Cthulhu point you toward your destiny..."
   "Inskrypcje na Statui Cthulhu wskazują Ci Twoje Przeznaczenie...",
   "Cthulhu'nun heykeli üzerindeki yazıt sana kaderinin nasıl çizileceğini gösteriyor...",
   "Nápis na Cthulhuově sošce tě směruje ke tvému osudu...",
   "Надписи на статуе Ктулху показывают Вам Вашу судьбу...",
   "Die Inschrift auf Cthulhus Statue zeigt dir den Weg zu deinem Schicksal...",
   "As inscrições na Estátua do Cthulhu lhe apontam para o seu destino...",
    }},
  {0x207f072a, { // "projection in ball model"
   "projekcja w modelu kuli",
/*MISSING*/    "projection in ball model",
   "projekce v modelu koule",
   "проекция в модели сферы",
/*MISSING*/    "projection in ball model",
/*MISSING*/    "projection in ball model",
    }},
  {0x20861ec3, { // "Save %the1 first to unlock this challenge!"
   "Uratuj %a1, by mieć dostęp do tej misji!",
   "Bu görevi açmak için önce %a1 kurtar!",
   "Pro aktivaci této mise musíš nejprve zachránit %a1!",
   "Спаси %a1, прежде чем разблокировать миссию!",
   "Rette zuerst %den1 %a1 um diese Herausforderung freizuschalten!",
/*MISSING*/    "Save %the1 first to unlock this challenge!",
    }},
  {0x20c33eab, { // "You cannot attack Sandworms directly!"
   "Nie możesz atakować Czerwi wprost!",
   "Kumkurduna doğrudan saldıramazsınız!",
   "Na Písečné červy nemůžeš útočit přímo!",
   "Вы не можете атаковать песчаных червей!",
   "Du kannst Sandwürmer nicht direkt angreifen!",
   "Você não pode atacar Minhocas de Areia diretamente!",
    }},
  {0x20cd7f3a, { // "%The1 would get you!"
   "Dopad%ł1by Cię %1!",
   "%1 seni kapar!",
   "To by tě dosta%l1 %1!",
   "%1 уничтожит вас здесь!",
   "%Der1 %1 würde dich fangen!",
   "%na1 lhe mataria!",
    }},
  {0x20d4bab0, { // "HyperRogue classic"
   "tryb klasyczny HyperRogue",
/*MISSING*/    "HyperRogue classic",
   "klasický HyperRogue",
   "классический режим HyperRogue",
/*MISSING*/    "HyperRogue classic",
/*MISSING*/    "HyperRogue classic",
    }},
  {0x20eeb3db, { // "You blow %the1 away!"
   "Zdmuchn%ąłeś0 %a1!",
   "%a1! rüzgârla ittirdin!",
   "Odfoukl jsi %a1!",
   "Вы сдули %a1!",
   "Du wehst %den1 %a1 weg!",
/*MISSING*/    "You blow %the1 away!",
    }},
  {0x20efa355, { // "male"
   "mężczyzna",
   "erkek",
   "muž",
   "мужской",
   "männlich",
   "masculino",
    }},
  {0x20f3fe95, { // "A device that attracts sandworms and other enemies. You need to activate it."
   "To urządzenie przyciąga czerwie i innych przeciwników. Musisz je aktywować.",
   "Kumkurtlarını ve diğer düşmanları çağıracak bir aygıt. Sadece etkinleştirmen gerekiyor.",
   "Zařízení, které přitahuje písečné červy a jiné nepřátele. Aby začalo pracovat, musíš ho aktivovat.",
   "Этот аппарат привлекает песчаных червей и других врагов. Вам нужно активировать его.",
   "Ein Gerät, das Sandwürmer und andere Feinde anlockt. Es muss aktiviert werden.",
/*MISSING*/    "A device that attracts sandworms and other enemies. You need to activate it.",
    }},
  {0x211b3e1f, { // "Another energy sword!"
   "Jeszcze jeden miecz energii!",
/*MISSING*/    "Another energy sword!",
   "Další energetický meč!",
   "Ещё один магический меч!",
/*MISSING*/    "Another energy sword!",
/*MISSING*/    "Another energy sword!",
    }},
  {0x21278f73, { // " Additionally, all items around the killed Troll will be destroyed."
   " Dodatkowo, wszystkie przedmioty wokół zabitego Trolla ulegają zniszczeniu.",
   " Ayrıca, Trolün etrafındaki bütün eşyalar yok olur.",
   " Zabitím Šutrolla se navíc zničí všechny předměty kolem něj.",
   " Все предметы рядом с умершим троллем также уничтожаются.",
   " Des weiteren werden alle Items um den getöteten Troll zerstört.",
/*MISSING*/    " Additionally, all items around the killed Troll will be destroyed.",
    }},
  {0x21391810, { // "You glance at your great treasures."
   "Patrzysz na swoje wspaniałe skarby.",
   "Müthiş hazinelerine göz gezdirdin.",
   "Díváš se na své velké poklady.",
   "Вы смотрите на ваши громадные сокровища.",
   "Du betrachtest deine kostbaren Schätze.",
   "Você contempla seus grandes tesouros.",
    }},
  {0x213e2e9a, { // "Hmm, he has been training in the Emerald Mine. Interesting..."
   "Ciekawe... trenował w Kopalni Szmaragdów.",
   "Çok garip... Zümrüt Madeninde alıştırma yapıyormuş.",
   "Hmm, tak on trénoval ve Smaragdovém dole. Zajímavé...",
   "Он тренировался в Изумрудной шахте. Хм, интересно...",
   "Er  hat in der Smaragdmine trainiert. Interessant...",
/*MISSING*/    "Hmm, he has been training in the Emerald Mine. Interesting...",
    }},
  {0x2175c728, { // "hyperboloid model"
   "model hiperboloidy",
/*MISSING*/    "hyperboloid model",
   "model hyperboloidu",
   "модель гиперболоида",
/*MISSING*/    "hyperboloid model",
/*MISSING*/    "hyperboloid model",
    }},
  {0x21c93b56, { // "After killing %the1, you feel able to reach the Elemental Planes!"
   "Po pokonaniu %a1 czujesz, że masz dostęp do Strefy Żywiołów!",
/*MISSING*/    "After killing %the1, you feel able to reach the Elemental Planes!",
   "Po zabití %a1 cítíš, že jsi schopen dosáhnout Elementálních rovin!",
   "Победив %a1, ты чувствуешь, что можешь попасть в Плоскость стихий!",
/*MISSING*/    "After killing %the1, you feel able to reach the Elemental Planes!",
/*MISSING*/    "After killing %the1, you feel able to reach the Elemental Planes!",
    }},
  {0x21d7b5ef, { // "(v) config"
   "(v) ust.",
   "(v) ayarlar",
   "(v) nast.",
   "(v) настр.",
   "(v) Konfig.",
   "(v) configurações",
    }},
  {0x2231398b, { // "Hyperstone Quest"
   "Misja Hiperkamień",
   "Aşkıntaş Görevi",
   "hyperkamový úkol",
   "Миссия Гиперкамней",
   "Hyperstone-Herausforderung",
/*MISSING*/    "Hyperstone Quest",
    }},
  {0x226c89da, { // "\nSpecial conduct (still valid)\n"
   "\nDodatkowe wyzwanie (wciąż spełnione):",
/*MISSING*/    "\nSpecial conduct (still valid)\n",
   "\nZvláštní výzva (stále platná):",
   "\nДополнительное условие (выполняется):",
/*MISSING*/    "\nSpecial conduct (still valid)\n",
/*MISSING*/    "\nSpecial conduct (still valid)\n",
    }},
  {0x22843718, { // "A nasty predator from the Icy Land. Contrary to other monsters, it tracks its prey by their heat."
   "Niebezpieczny drapieżnik z Krainy Lodu. W przeciwieństwie do innych stworów, śledzi ofiarę po jej cieple.",
   "Buz Diyarlarında yaşayan hınzır bir yırtıcı. Diğer canavarların aksine, avını ısısıyla takip ediyor.",
   "Ošklivý predátor z Ledového kraje. Na rozdíl od jiných netvorů sleduje kořist pomocí tepla, které vydává.",
   "Мерзкий хищник из Ледяной земли. В отличие от остальных монстров он ищет свою жертву по её теплу.",
   "Ein scheußlicher Jäger des Eisigen Landes. Anders als andere Monster findet er seine Beute durch deren Wärme.",
/*MISSING*/    "A nasty predator from the Icy Land. Contrary to other monsters, it tracks its prey by their heat.",
    }},
  {0x22923e9b, { // "Cannot slash another player!"
   "Nie możesz ciąć innego gracza!",
/*MISSING*/    "Cannot slash another player!",
   "Nemůžeš seknout jiného hráče!",
   "Нельзя разрубить другого игрока!",
   "Kann keinen anderen Spieler schlagen!",
/*MISSING*/    "Cannot slash another player!",
    }},
  {0x229f917c, { // "You are recalled!"
   "Wracasz!",
/*MISSING*/    "You are recalled!",
   "Vracíš se!",
   "Вы вернулись!",
/*MISSING*/    "You are recalled!",
/*MISSING*/    "You are recalled!",
    }},
  {0x22ccbba4, { // "throw a knife"
   "rzut nożem",
   "bıçak at",
   "házení nože",
   "кинуть нож",
   "Messer werfen",
/*MISSING*/    "throw a knife",
    }},
  {0x22ea7326, { // "Ice below you is melting! RUN!"
   "Lód pod Tobą się topi! UCIEKAJ!",
   "Altındaki buz eriyor! KAÇ!",
   "Led pod tebou taje! UTÍKEJ!",
   "Лёд под Вами тает! БЕГИТЕ!",
   "Das Eis unter dir schmilz! RENN!",
   "O gelo abaixo de você está derretendo! CORRA!",
    }},
  {0x22fd147c, { // "A natural terrain feature of the Desert."
   "Naturalny teren z Pustyni.",
   "Çölün içinde doğal bir kum oluşumu.",
   "Přírodní terénní prvek v Poušti.",
   "Натуральный ландшафт пустыни.",
   "Ein natürliches Landschaftsmerkmal der Wüste.",
/*MISSING*/    "A natural terrain feature of the Desert.",
    }},
  {0x23234993, { // "Use arrow keys to choose the teleport location."
   "Użyj strzałek, by wybrać cel teleportacji.",
   "Işınlanacağın yeri seçmek için okları kullan.",
   "Vyber cíl teleportace pomocí šipek.",
   "С помощью стрелок выберите место для телепортации.",
   "Benutze die Pfeiltasten um dein Teleportations Ziel zu wählen.",
   "Use as setas para escolher o local do teleporte.",
    }},
  {0x23a89ce4, { // "Rotate the camera. Can be used to obtain a first person perspective, or third person perspective when combined with Y shift."
   "Obróć kamerę. Można użyć do uzyskania perspektywy z pierwszej osoby, lub z trzeciej osoby po połączeniu z przesunięciem Y.",
/*MISSING*/    "Rotate the camera. Can be used to obtain a first person perspective, or third person perspective when combined with Y shift.",
   "Rotace kamery. Lze ji použít k získání perspektivy první osoby, nebo (v kombinaci s Y posuvem) třetí osoby.",
   "Поворачивает камеру. Можно использовать, чтобы получить вид от первого лица или от третьего лица в сочетании со сдвигом по Y.",
/*MISSING*/    "Rotate the camera. Can be used to obtain a first person perspective, or third person perspective when combined with Y shift.",
/*MISSING*/    "Rotate the camera. Can be used to obtain a first person perspective, or third person perspective when combined with Y shift.",
    }},
  {0x23c92250, { // "Cannot push into the same location!"
   "Nie można pchać w to samo miejsce!",
/*MISSING*/    "Cannot push into the same location!",
   "Nemůžeš tlačit na stejné místo!",
   "Нельзя толкать в оно и то же место!",
   "Kann nicht auf dasselbe Feld ziehen!",
/*MISSING*/    "Cannot push into the same location!",
    }},
  {0x23e7d6ba, { // "You feel that %the2 become%s2 more dangerous."
   "Masz wrażenie, że %P2 staje się bardziej niebezpieczn%ya2.",
   "Topladığın her %abl1, ...",
   "Cítíš, že %2 začíná být nebezpečnější.",
   "Вы чувствуете, что находиться %abl2 всё опаснее.",
   "Du spürst, dass %der2 %2 gefährlicher wird,",
   "Você sente que %na1 está mais perigoso-a.",
    }},
  {0x2406632a, { // "player 1 X"
   "gracz 1 X",
   "oyuncu 1, yatay",
   "hráč 1 X",
   "игрок 1 X",
   "Spieler 1 X",
/*MISSING*/    "player 1 X",
    }},
  {0x2406632b, { // "player 1 Y"
   "gracz 1 Y",
   "oyuncu 1, dikey",
   "hráč 1 Y",
   "игрок 1 Y",
   "Spieler 1 Y",
/*MISSING*/    "player 1 Y",
    }},
  {0x24240b83, { // "Number of uses left: %1"
   "Pozostało użyć: %1",
/*MISSING*/    "Number of uses left: %1",
   "Poèet zbývajících použití: %1",
   "Осталось применений: %1",
/*MISSING*/    "Number of uses left: %1",
/*MISSING*/    "Number of uses left: %1",
    }},
  {0x246929eb, { // "Many lost boats are spinning in this whirlpool. Some of them contain treasures and Pirates.\n\nUnmanned boats will go one cell clockwise in each turn. Sharks and manned boats are only allowed to go with the current, that is, towards the center, or clockwise."
   "Dużo statków zagubiło się w tym wirze. Na niektórych z nich są skarby i Piraci.\n\nStatki bez załogi poruszają się o 1 pole w każdej kolejce, zgodnie z ruchem wskazówek zegara. Rekiny i statki z załogą mogą poruszać się tylko z prądem, czyli w kierunku środka wiru albo zgodnie z ruchem wskazówek zegara.",
   "Bu tayfunda pek çok kayıp kayık geziniyor. Bır kısmında hazineler yahut korsanlar var.\n\nİnsansız kayıklar her tur saat yönünde bir hücre hareket eder. Köpekbalıkları ve içinde insan olan kayıklar sadece akıntıyla -yani merkeze doğru- ya da saat yönünde hareket edebilir.",
   "V tomto víru se točí mnoho ztracených člunů. Na některých z nich se nachází poklad nebo piráti.\n\nČluny bez posádky se v každém kole posunou o jedno pole po směru hodinových ručiček. Žraloci a čluny s posádkou se smějí pohybovat pouze po proudu, tedy do středu nebo po směru hodinových ručiček ",
   "Много лодок погибло в этом водовороте. В некоторых из них были ценности и Пираты.\n\nПустые лодки движутся на одну клетку по часовой стрелке каждые ход. Акулы и лодки с людьми движутся по течению, то есть к центру или по часовой стрелке.",
   "Viele verlorene Boote wirbeln in diesem Strudel... manche enthalten Schätze oder Piraten.\n\nUnbemannte Boote bewegen sich jeden Zug eine Zelle im Uhrzeigersinn. Haie und bemannte Boote dürfen nur mit der Strömung schwimmen (das heißt, zur Mitte oder im Uhrzeigersinn).",
/*MISSING*/    "Many lost boats are spinning in this whirlpool. Some of them contain treasures and Pirates.\n\nUnmanned boats will go one cell clockwise in each turn. Sharks and manned boats are only allowed to go with the current, that is, towards the center, or clockwise.",
    }},
  {0x2500ca82, { // "HyperRogue game board"
   "plansza do gry HyperRogue",
   "HyperRogue tahta",
   "herní deska HyperRogue",
   "игровая доска HyperRogue",
   "HyperRogue-Spielbrett",
   "tabuleiro de jogo HyperRogue",
    }},
  {0x25270e03, { // "Want another type of game? Want more challenge?\nHyperRogue has many special modes and challenges that significantly change the gameplay. Try them!"
   "Chcesz spróbować innego typu gry? Dodatkowych wyzwań?\nHyperRogue ma dużo specjalnych trybów, istotnie zmieniających styl gry. Wypróbuj je!",
/*MISSING*/    "Want another type of game? Want more challenge?\nHyperRogue has many special modes and challenges that significantly change the gameplay. Try them!",
   "Chceš jiný typ hry? Chceš vìtší výzvu?\nHyperRogue obsahuje øadu zvláštních módù a výzev, které výraznì mìní styl hry. Zkus je!",
   "Хочешь другой тип игры? Хочешь посложнее?\nHyperRogue содержит много специальных режимов, заметно меняющих игру. Попробуй их!",
/*MISSING*/    "Want another type of game? Want more challenge?\nHyperRogue has many special modes and challenges that significantly change the gameplay. Try them!",
/*MISSING*/    "Want another type of game? Want more challenge?\nHyperRogue has many special modes and challenges that significantly change the gameplay. Try them!",
    }},
  {0x254cf8b2, { // "You activate your terraforming powers!"
   "Uaktywni%łeś0 swoje moce edycji!",
   "Yerdüzenleyici güçlerini harekete geçirdin!",
   "Aktivova%l0 jsi své terramorfické schopnosti!",
   "Вы активировали силу редактирования!",
   "Du aktivierst deine terraforming Kräfte!",
/*MISSING*/    "You activate your terraforming powers!",
    }},
  {0x254f162a, { // "see http://www.roguetemple.com/z/hyper/conformal.php"
   "patrz http://www.roguetemple.com/z/hyper/conformal.php (po angielsku)",
/*MISSING*/    "see http://www.roguetemple.com/z/hyper/conformal.php",
   "viz http://www.roguetemple.com/z/hyper/conformal.php (anglicky)",
   "смотри http://www.roguetemple.com/z/hyper/conformal.php (по-английски))",
   "siehe http://www.roguetemple.com/z/hyper/conformal.php (in Englisch)",
/*MISSING*/    "see http://www.roguetemple.com/z/hyper/conformal.php",
    }},
  {0x25615d28, { // "high contrast"
   "wysoki kontrast",
   "yüksek kontrast",
   "vysoký kontrast",
   "высокий контраст",
   "Hoher Kontrast",
/*MISSING*/    "high contrast",
    }},
  {0x25839844, { // "Congratulations!"
   "Gratulacje!",
   "Tebrikler!",
   "Gratulujeme!",
   "Поздравляем!",
   "Glückwunsch!",
   "Parabéns!",
    }},
  {0x2586ff41, { // "A mouse squeaks at you. It seems that it wants you to go somewhere. That's interesting, what could you find here?\n\nNote: in the part of the Palace that you are exploring during this quest, the distribution of monsters, pressure plates, and items has been changed somewhat, to make it a bit more suitable for less experienced players. The number of monsters does not depend on the number of Hypersian Rugs you have collected, and there are more opening plates.\n\nHowever, even with the normal distribution, it appears that it should be always possible to reach your goal. If you want this challenge, select it from the special game modes menu."
   "Myszka piszczy na Ciebie. Wydaje się, że chce Cię gdzieś zaprowadzić. Ciekawe, co tam może być?\n\nW części Pałacu, którą zwiedzasz podczas tej misji, rozkład potworów, płyt i przedmiotów został trochę zmieniony, by misja była bardziej odpowiednia dla mniej doświadczonych graczy. W szczególności, liczba potworów nie zależy od liczby zebranych Hiperskich Dywanów i jest więcej płyt otwierających.\n\nJednak nawet przy normalnym rozkładzie wydaje się, że zawsze powinno być możliwe dotarcie do celu. Jeśli chcesz takiego wyzwania, wybierz je z menu specjalnych trybów gry.",
   "Bir fare sana ciyakladı. Galiba senin bir yere gitmeni istiyor.Gerçekten ilginç, burada ne bulabilirsin ki?\n\nW części Pałacu, którą zwiedzasz podczas tej misji, Not: Sarayın bu görev için hazırlanmış kısmında, canavarların dağılımı, baskı levhaları, ve eşyalar daha az deneyimli oyuncular düşünülerek bir nebze değiştirildi.Canavar sayısı topladığın Aşkınlı Kilim sayısına bağlı değil ve burada daha çok açma levhası var. \n\nYine de, normal dağılımla bile, bu görev her zaman tamamlanabilir. Eğer o görevi istiyorsan, özel oyun modları menüsünden onu seçebilirsin.",
   "Vykvíkla na tebe myška. Zdá se, že chce, aby ses někam vyda%l0. To je zajímavé, copak tam asi najdeš?\n\nPoznámka: V části Paláce, kterou budeš procházet při plnění této mise, je poněkud upravené rozdělení netvorů, desek a předmětů, aby byla mise o něco vhodnější i pro méně zkušené hráče. Počet netvorů nebude záviset na počtu získaných Hyperských koberců a bude tu více otvíracích desek.\n\nNicméně, dosažení cíle by mělo být vždy možné i při normálním rozdělení. Pokud si chceš tuto obtížnější misi vyzkoušet, vyber ji z menu speciálních herních módů.\n\n",
   "Мышка пищит на Вас. Как будто она хочет, чтобы Вы пошли куда-то. Интересно, что же там может быть?\n\nВ той части Дворца, которая будет доступна в Квесте, распределение монстров, плит и предметов немного изменено, чтобы подходить для не очень опытных игроков. Количество монстров не зависит от количества собранных сокровищ, и открывающих плит здесь больше.\n\nНо и при нормальном распределении Квест можно пройти. Если Вы хотите это сделать, выберите специальный режим в меню.",
   "Eine Maus piepst dich an. Es scheint als wolle sie das du dich irgendwo hin begibst. Interessant, was du dort wohl findest?\n\nIn dem Teil des Palastes den du während dieser Quest erkundest wurde die Verteilung von Monstern, Druckplatten und Items etwas geändert um es nicht so erfahrenen Spielern zu vereinfachen. Die Anzahl der Monster basiert nicht auf der Anzahl der Hypersischen Teppiche in deinem Besitzt und es gibt mehr Öffnungsmechanismen.\n\nSelbst mit der normalen Monsterverteilung sollte es allerdings möglich sein immer das Ziel zu erreichen. Dieser Modus kann im Menü ausgewählt werden.",
/*MISSING*/    "A mouse squeaks at you. It seems that it wants you to go somewhere. That's interesting, what could you find here?\n\nNote: in the part of the Palace that you are exploring during this quest, the distribution of monsters, pressure plates, and items has been changed somewhat, to make it a bit more suitable for less experienced players. The number of monsters does not depend on the number of Hypersian Rugs you have collected, and there are more opening plates.\n\nHowever, even with the normal distribution, it appears that it should be always possible to reach your goal. If you want this challenge, select it from the special game modes menu.",
    }},
  {0x25871fa3, { // "Known mines may be marked by touching while in drag mode. Your allies won't step on marked mines."
   "Znane miny można oznaczać dotykając ich w trybie przeciągania. Twoi przyjaciele nie będą stawać na oznaczonych minach.",
/*MISSING*/    "Known mines may be marked by touching while in drag mode. Your allies won't step on marked mines.",
   "Políčka, která obsahují miny, je možné označit dotekem v módu tažení. Vaši spojenci na označené miny nevstoupí.",
   "Известные мины можно помечать, кликая на них в режиме просмотра. Твои союзники не будут на них наступать.",
/*MISSING*/    "Known mines may be marked by touching while in drag mode. Your allies won't step on marked mines.",
/*MISSING*/    "Known mines may be marked by touching while in drag mode. Your allies won't step on marked mines.",
    }},
  {0x25903a7a, { // "or 't' to see the top scores"
   "lub 't' by zobaczyć najlepsze wyniki",
   "'t'  ile yüksek puanları gör",
   "'t': tabulka nejvyšších skóre",
   "либо 't', чтобы посмотреть результаты",
   "oder 't' für die Highscore Liste",
   "ou 't' para ver as pontuações mais altas",
    }},
  {0x25976d37, { // " (click to use)"
   " (klik by użyć)",
/*MISSING*/    " (click to use)",
   " (použij kliknutím)",
   " (нажми, чтобы использовать)",
/*MISSING*/    " (click to use)",
/*MISSING*/    " (click to use)",
    }},
  {0x25bb71f1, { // "This orb allows you to summon monsters. Usually, they are either Elementals or native creatures. While the monsters do not like being summoned, and will attack you once they recover from summoning sickness, such summoning often has its uses."
   "Ta sfera pozwala Ci przywoływać potwory. Zwykle są to Żywiołaki albo stworzenia pochodzące z danej krainy. Stwory nie lubią być przywoływane i będą atakować po wyjściu ze zmieszania wywołanego przywołaniem. Ale jednak takie przywoływanie ma pewne zastosowania.",
   "Bu küre canavar çağırabilmeni sağlar. Genelde Özütler veya yerel canavarlar çağrılır.Canavarlar çağrılmaktan hoşlanmasalar ve çağrılma hastalıkları biter bitmez sana saldıracak olsalar da, yine de çağırmanın işe yaradığı yerler olabilir.",
   "Tato Sféra ti umožňuje vyvolávat netvory. Obvykle se jedná buď o Elementály nebo o domorodé obyvatele kraje, kde ji použiješ. Netvoři nejsou vyvoláním nijak nadšeni a poté, co se zorientují, na tebe zaútočí, ale i tak má vyvolávání často své využití.",
   "Эта сфера позволяет призывать монстров. Обычно это Элементали или родные для земли существа. Существа не нравится быть вызванным, и как только оно оправится от вызова, будет Вас атаковать.",
   "Dieser Orb erlaubt es dir Monster zu beschwören. Diese sind meist Elementare, oder einheimische Monster. Auch wenn Monster nicht gerne beschworen werden und sie dich angreifen sobald die Beschwörungsübelkeit vorüber ist haben Beschwörungen oft einen Nutzen.",
/*MISSING*/    "This orb allows you to summon monsters. Usually, they are either Elementals or native creatures. While the monsters do not like being summoned, and will attack you once they recover from summoning sickness, such summoning often has its uses.",
    }},
  {0x2657f9d0, { // "three players"
   "trzech graczy",
/*MISSING*/    "three players",
   "tři hráči",
   "три игрока",
   "drei Spieler",
/*MISSING*/    "three players",
    }},
  {0x2660b640, { // "Ground level is actually an equidistant surface, %1 absolute units below the plane P. Theoretically, this value affects the world -- for example, eagles could fly %2 times faster by flying above the ground level, on the plane P -- but the actual game mechanics are not affected. (Distances reported by the vector graphics editor are not about points on the ground level, but about the matching points on the plane P -- divide them by the factor above to get actual distances."
   "Poziom podłoża jest w rzeczywistości ekwidystantną powierzchnią %1 jednostek pod płaszczyzną P. Teoretycznie, ta wartość wpływa na świat -- przykładowo, orły mogłyby latać %2 razy szybciej latając powyżej poziomu podłoża, na wysokości płaszczyzny P -- ale nie wpływa to na mechanikę gry w żaden sposób. (Odległości wyświetlane przez edytor grafiki dotyczą odległości między odpowiednimi punktami na płaszczyźnie P.",
/*MISSING*/    "Ground level is actually an equidistant surface, %1 absolute units below the plane P. Theoretically, this value affects the world -- for example, eagles could fly %2 times faster by flying above the ground level, on the plane P -- but the actual game mechanics are not affected. (Distances reported by the vector graphics editor are not about points on the ground level, but about the matching points on the plane P -- divide them by the factor above to get actual distances.",
   "Úroveň země je ve skutečnosti ekvidistantní plocha, která se nachází %1 absolutních jednotek pod rovinou P. Tato hodnota teoreticky ovlivňuje svět -- orli by například mohli létat %2-krát rychleji, kdyby létali nad úrovní země v rovině P -- ale herní mechaniky jako takové tím ovlivněny nejsou. (Vzdálenosti v editoru vektorové grafiky nejsou vzdálenosti bodů na úrovni země, ale odpovídajících bodů v rovině P -- skutečné vzdálenosti získáte jejich vydělením výše zmíněným faktorem.",
   "Поверхность земли является эквидистантой на глубине  %1 единицу ниже плоскости P. Теоретически, это значение могло бы влиять на мир. Например, орлы могли бы летать в %2 раз быстрее, летая над уровнем земли, на плоскости P -- но механика игры не меняется. (Расстояния в редакторе векторной графики относятсяне к точкам на поверхности, а к их проекциям на плоскость P.)",
/*MISSING*/    "Ground level is actually an equidistant surface, %1 absolute units below the plane P. Theoretically, this value affects the world -- for example, eagles could fly %2 times faster by flying above the ground level, on the plane P -- but the actual game mechanics are not affected. (Distances reported by the vector graphics editor are not about points on the ground level, but about the matching points on the plane P -- divide them by the factor above to get actual distances.",
/*MISSING*/    "Ground level is actually an equidistant surface, %1 absolute units below the plane P. Theoretically, this value affects the world -- for example, eagles could fly %2 times faster by flying above the ground level, on the plane P -- but the actual game mechanics are not affected. (Distances reported by the vector graphics editor are not about points on the ground level, but about the matching points on the plane P -- divide them by the factor above to get actual distances.",
    }},
  {0x2679cb12, { // "move up-left"
   "ruch w górę-lewo",
/*MISSING*/    "move up-left",
   "pohyb doleva nahoru",
   "ход вверх-влево",
/*MISSING*/    "move up-left",
/*MISSING*/    "move up-left",
    }},
  {0x26a6b883, { // "up"
   "góra",
/*MISSING*/    "up",
   "nahoru",
   "вверх",
   "hoch",
/*MISSING*/    "up",
    }},
  {0x26dfda10, { // " (increases treasure spawn)"
   " (więcej skarbów)",
/*MISSING*/    " (increases treasure spawn)",
   " (zvyšuje výskyt pokladů)",
   " (увеличивает генерацию сокровищ)",
/*MISSING*/    " (increases treasure spawn)",
/*MISSING*/    " (increases treasure spawn)",
    }},
  {0x272520b2, { // "You feel attuned to gravity, ready to face mountains and dungeons."
   "Jesteś dostrojon%ya1 do grawitacji, gotow%ya1 na góry i lochy.",
/*MISSING*/    "You feel attuned to gravity, ready to face mountains and dungeons.",
   "Cítíš se přizpůsoben%ý1 gravitaci, připraven%ý1 čelit horám a žalářům.",
   "Вы привыкли к гравитации и готовы покорять горы и подземелья.",
/*MISSING*/    "You feel attuned to gravity, ready to face mountains and dungeons.",
/*MISSING*/    "You feel attuned to gravity, ready to face mountains and dungeons.",
    }},
  {0x27440e71, { // "Cannot cast illusion on an item!"
   "Nie możesz rzucić iluzji na przedmiot!",
   "Bir eşyanın üzerinde ilüzyon oluşturamazsın!",
   "Nemůžeš seslat iluzi na předmět!",
   "Nie możesz rzucić iluzji na przedmiot!",
   "Du kannst Illusion nicht auf Gegenständen erschaffen!",
/*MISSING*/    "Cannot cast illusion on an item!",
    }},
  {0x276480b2, { // "You cannot move through %the1!"
   "Nie możesz przejść przez %a1!",
   "%1 hücresinden geçemezsin!",
   "Nemůžeš překonat %a1!",
   "Вы не можете пройти через %a1!",
   "Du kannst nicht durch %den1 %a1 gehen!",
   "Você não pode se mover através %g1!",
    }},
  {0x27bfdd2a, { // "zoom in"
   "zbliżenie",
/*MISSING*/    "zoom in",
   "přiblížení",
   "приблизить",
   "hereinzoomen",
/*MISSING*/    "zoom in",
    }},
  {0x27d2b4d6, { // "A long time ago a mighty warrior was guarding the mirrors from being broken. While this warrior is no longer alive, his reflections have gained life of their own, and will punish the intruders.\n\nIf you attack a Mirror Spirit physically, it is delayed, but not destroyed -- more reflections will come out of the mirror. Use Mimics to destroy them."
   "Dawno, dawno temu potężny wojownik bronił lustr przed tymi, co chcieli jej rozbić. Wojownik ten już dawno nie żyje, ale jego odbicia zaczęły żyć własnym życiem i karać najeźdźców.\n\nJeśli zaatakujesz Ducha Lustra fizycznie, jest spowolniony, ale nie zniszczony -- więcej odbić wyjdzie z lustra. Użyj Mimików, by go zniszczyć.",
/*MISSING*/    "A long time ago a mighty warrior was guarding the mirrors from being broken. While this warrior is no longer alive, his reflections have gained life of their own, and will punish the intruders.\n\nIf you attack a Mirror Spirit physically, it is delayed, but not destroyed -- more reflections will come out of the mirror. Use Mimics to destroy them.",
   "Tento mocný váleèník kdysi dávno støežil zrcadla, aby se nerozbila. Teï už nežije, ale jeho odrazy získaly svùj vlastní život a trestají vetøelce.\n\nPokud na Zrcadlového ducha fyzicky zaútoèíš, zpomalíš ho tím, ale neznièíš -- ze zrcadla vyjdou další odrazy. Pokud ho chceš znièit nadobro, použij Mimiky.",
   "Давным-давно могучий воин охранял эти зеркала, чтобы никто их не разбил. Хотя воин давно мёртв, его отражения живы, и они накажут незваных гостей.\n\nЕсли ты атакуешь Зеркального духа физически, он замедляется, но не погибает -- больше отражений выходит из зеркала. Используй Мимиков, чтобы уничтожать их.",
/*MISSING*/    "A long time ago a mighty warrior was guarding the mirrors from being broken. While this warrior is no longer alive, his reflections have gained life of their own, and will punish the intruders.\n\nIf you attack a Mirror Spirit physically, it is delayed, but not destroyed -- more reflections will come out of the mirror. Use Mimics to destroy them.",
/*MISSING*/    "A long time ago a mighty warrior was guarding the mirrors from being broken. While this warrior is no longer alive, his reflections have gained life of their own, and will punish the intruders.\n\nIf you attack a Mirror Spirit physically, it is delayed, but not destroyed -- more reflections will come out of the mirror. Use Mimics to destroy them.",
    }},
  {0x2826d66d, { // "Tiling and Tactics"
   "Posadzka i taktyka",
/*MISSING*/    "Tiling and Tactics",
   "Hyperbolické dělení a taktika",
   "Геометрия и тактика",
/*MISSING*/    "Tiling and Tactics",
/*MISSING*/    "Tiling and Tactics",
    }},
  {0x2839a534, { // "include history"
   "uwzględnij historię",
/*MISSING*/    "include history",
   "včetně historie",
   "включить историю",
   "inklusive Vergangenheit",
/*MISSING*/    "include history",
    }},
  {0x28698d2f, { // "For some reason... cannot attack!"
   "Z jakiegoś powodu... nie możesz atakować!",
/*MISSING*/    "For some reason... cannot attack!",
   "Z nějakého důvodu... nemůžeš útočit!",
   "По какой-то причине... невозможно атаковать!",
/*MISSING*/    "For some reason... cannot attack!",
/*MISSING*/    "For some reason... cannot attack!",
    }},
  {0x28af7500, { // "Hyperbolic geometry yields much more interesting periodic patterns than Euclidean."
   "W geometrii hiperbolicznej są dużo bardziej ciekawe regularne wzory niż w euklidesowej.",
/*MISSING*/    "Hyperbolic geometry yields much more interesting periodic patterns than Euclidean.",
   "V hyperbolické geometrii lze najít mnohem zajímavější periodické vzory než v eukleidovské.",
   "Гиперболическая геометрия даёт больше интересных периодических шаблонов, чем евклидова.",
/*MISSING*/    "Hyperbolic geometry yields much more interesting periodic patterns than Euclidean.",
/*MISSING*/    "Hyperbolic geometry yields much more interesting periodic patterns than Euclidean.",
    }},
  {0x28bfdd84, { // "You are now a tortoise heroine!"
   "Jesteś żółwią bohaterką!",
/*MISSING*/    "You are now a tortoise heroine!",
   "Nyní jsi želví hrdinka!",
   "Теперь вы -- черепашья героиня!",
   "Du bist jetzt ein Held der Schildkröten!",
/*MISSING*/    "You are now a tortoise heroine!",
    }},
  {0x28d62052, { // "%1 turns (%2)"
   "kolejek: %1 (%2)",
   "tur sayısı: %1 (%2)",
   "počet kol: %1 (%2)",
   "ходов: %1 (%2)",
   "%1 Züge (%2)",
   "Turnos: %1 (%2)",
    }},
  {0x2908ed86, { // "Cannot cast illusion here!"
   "Nie możesz rzucić iluzji tutaj!",
   "Burada ilüzyon oluşturamazsın!",
   "Sem nemůžeš seslat iluzi!",
   "Nie możesz rzucić ilzuji tutaj!",
   "Du kannst hier keine Illusion erschaffen!",
/*MISSING*/    "Cannot cast illusion here!",
    }},
  {0x291ece52, { // "Wine grown under hyperbolic sun would be extremely prized in your home location."
   "Wino uprawiane pod hiperbolicznym słońcem, byłoby bardzo cenne w Twoim świecie.",
   "Hiperbolik güneşin altında oluşan şarap geldiğin yerde epey değerli olsa gerek.",
   "Víno vypěstované pod hyperbolickým sluncem by mělo ve tvé domovině velkou cenu.",
   "Вино, созревшее под гиперболическим солнцем, очень ценится на Вашей родине.",
   "Wein, der unter hyperbolischer Sonne wächst, würde in deiner Heimat ziemlich viel wert sein.",
/*MISSING*/    "Wine grown under hyperbolic sun would be extremely prized in your home location.",
    }},
  {0x291fabc9, { // " Hell: %1/9"
   " Piekło: %1/9",
   " Cehennem: %1/9",
   " Peklo: %1/9",
   " Ад: %1/9",
   " Hölle: %1/9",
/*MISSING*/    " Hell: %1/9",
    }},
  {0x292a6966, { // "HyperRogue for Android"
   "HyperRogue dla Android",
   "Android için HyperRogue",
   "HyperRogue pro Android",
   "HyperRogue для Android",
   "HyperRogue für Android",
/*MISSING*/    "HyperRogue for Android",
    }},
  {0x293fb452, { // "start playing in the shmup mode"
   "zacznij grę w trybie strzelanki",
   "SEMUP modunda oyun başlat",
   "začni hrát v módu střílečky",
   "начать игру в режиме стрельбы",
   "shmup Modus starten",
/*MISSING*/    "start playing in the shmup mode",
    }},
  {0x29475e6e, { // "Some monsters have long tongues, which allow them to attack enemies in nearby cells."
   "Niektóre potwory mają długie języki, dzięki którym mogą atakować sąsiednie pola.",
   "Bazı canavarlar uzun dillere sahip, böylece yakındaki hücrelere saldırabiliyorlar.",
   "Někteří netvoři mají dlouhé jazyky, které jim umožňují útočit na sousední pole.",
   "Некоторые существа имеют длинные языки, которые позволяют им атаковать врагов на соседних клетках.",
   "Manche Monster haben lange Zungen, mit denen sie Feinde in nahen Zellen angreifen können.",
/*MISSING*/    "Some monsters have long tongues, which allow them to attack enemies in nearby cells.",
    }},
  {0x297b0403, { // "Shift-click a location to teleport there."
   "Shift-kliknij pole, by tam się teleportować.",
/*MISSING*/    "Shift-click a location to teleport there.",
   "Shift-klikni na políčko, kam se chceš teleportovat.",
   "Shift-клик на поле, чтобы телепортироваться туда.",
/*MISSING*/    "Shift-click a location to teleport there.",
/*MISSING*/    "Shift-click a location to teleport there.",
    }},
  {0x298b8919, { // "aura smoothening factor"
   "wygładzanie aury",
/*MISSING*/    "aura smoothening factor",
   "vyhlazovací faktor aury",
   "сглаживание ауры",
/*MISSING*/    "aura smoothening factor",
/*MISSING*/    "aura smoothening factor",
    }},
  {0x29ef900a, { // "Commanded %the1!"
   "%1 dosta%ł1 rozkaz!",
/*MISSING*/    "Commanded %the1!",
   "%1 dosta%l1 rozkaz!",
   "Прикажите %1!",
/*MISSING*/    "Commanded %the1!",
/*MISSING*/    "Commanded %the1!",
    }},
  {0x2a0c0556, { // "plain/3D"
   "prosty/3D",
/*MISSING*/    "plain/3D",
   "obyèejné/3D",
   "prosty/3D",
/*MISSING*/    "plain/3D",
/*MISSING*/    "plain/3D",
    }},
  {0x2a0d1a68, { // "You need to find the right Key to unlock this Orb of Yendor!"
   "Potrzebujesz prawidłowego Klucza, by otworzyć Sferę Yendoru!",
   "Bu Yendorun Küresinin kilidini açmak için bir anahtara ihtiyacın var!",
   "Abys moh%l0 odemknout tuto Yendorskou sféru, musíš najít správný klíč!",
   "Вам нужен ключ, чтобы открыть сферу Йендора!",
   "Du musst den richtigen Schlüssel finden, um den Orb von Yendor zu öffnen!",
   "Você precisa achar a Chave certa para desbloquear esse Orbe de Yendor!",
    }},
  {0x2a16fdee, { // "Reenter HyperRogue to apply this setting"
   "Wyjdź i wróć do HyperRogue, by zastosować",
/*MISSING*/    "Reenter HyperRogue to apply this setting",
   "Toto nastavení se aplikuje po návratu do HyperRogue",
   "Перезайдите в HyperRogue, чтобы применить эти настройки",
/*MISSING*/    "Reenter HyperRogue to apply this setting",
/*MISSING*/    "Reenter HyperRogue to apply this setting",
    }},
  {0x2a3a4971, { // "gender"
   "płeć",
   "cinsiyet",
   "pohlaví",
   "пол",
   "Geschlecht",
/*MISSING*/    "gender",
    }},
  {0x2a4bea57, { // "wall display mode"
   "tryb pokazywania ścian",
   "duvar görünümü",
   "mód zobrazování zdí",
   "режим показа стен",
   "Anzeigemodus (Wände)",
   "modo de exibição de paredes",
    }},
  {0x2a4e307e, { // "Unlocked by: %1 in %2"
   "Odblokowane przez: %1 %abl2",
/*MISSING*/    "Unlocked by: %1 in %2",
   "Odemèeno pøi: %1 %abl2",
   "Разблокирована: %1 %abl2",
/*MISSING*/    "Unlocked by: %1 in %2",
/*MISSING*/    "Unlocked by: %1 in %2",
    }},
  {0x2a587bc7, { // "you have cheated %1 times"
   "liczba oszustw: %1",
   "şimdiye kadar %1 kez hile yaptın.",
   "počet cheatů: %1",
   "количество читов: %1",
   "Du hast %1 mal Cheats benutzt",
   "Você trapaceou %1 vezes",
    }},
  {0x2a8ecaf0, { // "This white dog is able to run all the time. It is the only creature able to survive and breed in the Land of Eternal Motion."
   "Ten biały piesek jest w stanie biegać przez całe swoje życie. Jest to jedyne stworzenie, które jest w stanie przeżyć w Krainie Wiecznego Ruchu.",
   "Bu beyaz köpek hiç durmadan koşabilir. Sonsuz hareket diyarında yaşayabilen ve çoğalabilen tek canlıdır.",
   "Tento bílý pes dokáže neustále běžet. Je to jediný tvor, který dokáže přežít a rozmnožovat se v Kraji věčného pohybu.",
   "Эта белая собака способна всё время бежать. Это единственное существо, способное жить и размножаться в Земле вечного движения.",
   "Dieser weiße Hund kann unablässig rennen. Er ist die einzige Kreatur, die im Land der ewigen Bewegung überleben und sich vermehren kann.",
/*MISSING*/    "This white dog is able to run all the time. It is the only creature able to survive and breed in the Land of Eternal Motion.",
    }},
  {0x2ae7ad6a, { // "mirror what?"
   "co odbić?",
/*MISSING*/    "mirror what?",
   "co chceš zrcadlit?",
   "что отразить?",
/*MISSING*/    "mirror what?",
/*MISSING*/    "mirror what?",
    }},
  {0x2b02a473, { // "Kills: "
   "Zabicia: ",
   "Leşler: ",
   "Zabitých netvorů: ",
   "Убийства: ",
   "Kills: ",
/*MISSING*/    "Kills: ",
    }},
  {0x2b0733af, { // "first joystick: execute movement threshold"
   "dżojstik 1: próg wykonania ruchu",
   "birinci oyun kolu: hareket algılama eşiği",
   "joystick 1: práh provedení pohybu",
   "джойстик 1: порог выполнения движения",
   "Joystick 1: Ausgeführte Bewegungsgrenze",
/*MISSING*/    "first joystick: execute movement threshold",
    }},
  {0x2b4a42ed, { // "scale factor"
   "współczynnik skali",
   "ölçek faktörü",
   "měřítko",
   "масштаб",
   "Skalierungsfaktor",
   "fator de escala",
    }},
  {0x2b7df8b3, { // "Saved the SVG shot to %1 (sightrange %2)"
   "Zapisano zrzut SVG do %1 (zasięg %2)",
/*MISSING*/    "Saved the SVG shot to %1 (sightrange %2)",
   "Snímek SVG uložen do %1 (viditelnost %2)",
   "SVG снимок сохранён в %1 (радиус %2)",
/*MISSING*/    "Saved the SVG shot to %1 (sightrange %2)",
/*MISSING*/    "Saved the SVG shot to %1 (sightrange %2)",
    }},
  {0x2b9a4732, { // "%The1 stabs %the2."
   "%1 dźgn%ął1 %a2.",
   "%1 %2 bıçakladı.",
   "%1 probod%l1 %a2.",
   "%1 заколол%E1 %a2.",
   "%Der1 %1 ersticht %den2 %a2.",
   "%na1 apunhala %na2.",
    }},
  {0x2baa0061, { // "Unavailable in the Chaos mode.\n"
   "Niedostępne w trybie Chaosu.\n",
/*MISSING*/    "Unavailable in the Chaos mode.\n",
   "Nedostupné v chaotickém módu.\n",
   "Недоступно в режиме Хаоса.\n",
   "Nicht im Chaos-Modus verfügbar.\n",
/*MISSING*/    "Unavailable in the Chaos mode.\n",
    }},
  {0x2bce410a, { // "press Space or Home to center on the PC"
   "naciśnij Space lub Home, by wycentrować na postaci",
   "Space veya Home'a basarak PC'ye odakla.",
   "na PC můžeš vycentrovat obraz stisknutím mezerníku nebo klávesy Home",
   "Нажмите Пробел или Home, чтобы отцентрировать экран на персонаже.",
   "Drücke LEER oder Pos1 um den Spieler zu fokussieren",
   "Pressione Espaço ou Home para centralizar no PC",
    }},
  {0x2bd13413, { // "Hyperstone Quest completed!"
   "Misja alternatywna zakończona!",
   "Aşkıntaş görevi tamamlandı.",
   "Hyperkamový úkol splněn!",
   "Миссия Гиперкамня закончена!",
   "Hyperstein-Herausforderung abgeschlossen!",
   "Missão da Pedra do Ímpeto completa!",
    }},
  {0x2bf52a07, { // "You feel that %the1 slowly become%s1 dangerous..."
   "Masz wrażenie, że %P1 powoli staje się niebezpieczn%ya1...",
   "%1 yavaşça daha tehlikeli bir hâl alıyor...",
   "Máš pocit, že %1 pomalu začíná být nebezpečn%ý...",
   "Вы чувствуете, что %1 становится всё опаснее...",
   "Du spürst wie %der1 %1 langsam gefährlich wird...",
   "Você sente que %na1 está lentamente se tornando mais perigos%na1...",
    }},
  {0x2c6dc4f3, { // "You leave %the1."
   "Zostawiasz %a1.",
/*MISSING*/    "You leave %the1.",
   "Položi%l0 jsi %a1.",
   "Вы покинули %a1.",
/*MISSING*/    "You leave %the1.",
/*MISSING*/    "You leave %the1.",
    }},
  {0x2c6ef3ab, { // "These fairies would rather burn the forest, than let you get some Fern Flowers. The forest is infinite, after all...\n\nFire Fairies transform into fires when they die."
   "Wiły wolą spalić las, niż pozwolić Ci zdobyć Kwiaty Paproci. W końcu las jest nieskończony...\n\nWiły zamieniają się w ogień, gdy giną.",
   "Bu periler ormanı yakma pahasına, kimsenin Eğreltiotu Çiçeklerine dokunmasına izin vermezler.Sonuçta orman sonsuza kadar uzanıyor...\n\nAteş perileri öldüklerinde ateşe dönüşürler.",
   "Tito mužíčkové by raději spálili celý hvozd, než aby tě nechali sebrat nějaké Kapradinové květy. Konečně, hvozd je přece nekonečný...\n\nOhniví mužíčkové se promění v oheň, když zemřou.",
   "Эти феи скорее сожгут лес, чем позволят Вам собрать цветок папоротника. В конце концов, лес бесконечный...\n\nОни превращаются в огонь, когда умирают.",
   "Diese Feen würden eher den ganzen Wald niederbrennen als dich Farnblüten sammeln zu lassen. Der Wald ist schließlich grenzenlos...\n\nWenn Feuerfeen sterben, werden sie selbst zu Feuer.",
/*MISSING*/    "These fairies would rather burn the forest, than let you get some Fern Flowers. The forest is infinite, after all...\n\nFire Fairies transform into fires when they die.",
    }},
  {0x2c99ab74, { // "Press F1 or right-shift-click things for help."
   "F1 lub klikaj z prawym shiftem na elementach gry, by uzyskać pomoc.",
/*MISSING*/    "Press F1 or right-shift-click things for help.",
   "Nápovědu můžeš zobrazit stiskem F1 nebo kliknutím na věci s pravým shiftem.",
   "FНажмите F1 или используйте правую кнопку мыши с Shift'ом, чтобы открыть помощь..",
/*MISSING*/    "Press F1 or right-shift-click things for help.",
/*MISSING*/    "Press F1 or right-shift-click things for help.",
    }},
  {0x2cdaf58c, { // "Fire is extinguished!"
   "Ogień zgasł!",
/*MISSING*/    "Fire is extinguished!",
   "Oheň je uhašen!",
   "Огонь погас!",
/*MISSING*/    "Fire is extinguished!",
/*MISSING*/    "Fire is extinguished!",
    }},
  {0x2ce88365, { // "torus"
   "torus",
/*MISSING*/    "torus",
   "torus",
   "тор",
/*MISSING*/    "torus",
/*MISSING*/    "torus",
    }},
  {0x2d4b9753, { // "\"I like collecting ambers at the beach.\""
   "\"Lubię zbierać bursztyny na plaży.\"",
   "\"Sahilde Kehribarlar toplamayı seviyorum.\"",
   "\"Rád sbírám Jantar na pláži.\"",
   "\"Люблю собирать янтарь на берегу.\"",
   "\"Ich sammle gerne Bernstein am Strand.\"",
/*MISSING*/    "\"I like collecting ambers at the beach.\"",
    }},
  {0x2d55026b, { // "saved in the Euclidean mode!"
   "zapamiętywane w tym trybie!",
   "modunda kaydedilmeyecek!",
   "neukládají skóre a achievementy!",
   "сохраняются в евклидовом режиме!",
   "Punkte und Errungenschaften nicht gespeichert!",
/*MISSING*/    "saved in the Euclidean mode!",
    }},
  {0x2d5e09d9, { // "Your total treasure has been recorded in the Google Leaderboards."
   "Twój wynik został zapisany w rankingach Google.",
   "Toplam hazinen Google Listelerine kaydedildi.",
   "Tvůj celkový poklad byl zaznamenán na žebříčky na Googleu.",
   "Ваш результат записан в рейтинг Google.",
   "Deine Gesamtschätze wurden in die Google Leaderboards aufgenommen.",
/*MISSING*/    "Your total treasure has been recorded in the Google Leaderboards.",
    }},
  {0x2d6d45bb, { // "%The1 scares %the2 a bit!"
   "%1 troszkę przestraszy%ł1 %a2!",
/*MISSING*/    "%The1 scares %the2 a bit!",
   "%1 trochu vyleka%l1 %a2!",
   "%1 немного пугает %a2!",
   "%Der1 %1 ängstigt %den2 %a2 ein wenig!",
/*MISSING*/    "%The1 scares %the2 a bit!",
    }},
  {0x2d72701d, { // "Rogues will never make moves which result in their immediate death. Even when cornered, they are able to instantly teleport back to their home world at any moment, taking the treasures forever... but at least they will not steal anything further!\n\n"
   "Cwaniacy nigdy nie robią ruchów, po których by zostali od razu zabici. Nawet gdy się ich otoczy, są zdolni natychmiast teleportować się i wrócić do świata, skąd pochodzą, zabierając skarby na zawsze... ale przynajmniej nie ukradną już nic więcej!\n\n",
/*MISSING*/    "Rogues will never make moves which result in their immediate death. Even when cornered, they are able to instantly teleport back to their home world at any moment, taking the treasures forever... but at least they will not steal anything further!\n\n",
   "Chytráci nikdy neudělají tah, který by vedl k jejich okamžité smrti. I když jsou zahnaní do kouta, dokážou se kdykoli okamžitě teleportovat do svého domovského světa a navěky s sebou odnést poklady... ale aspoň už nebudou krást nic dalšího!\n\n",
   "Разбойник никогда не сделает ход, ведущий к его смерти. Когда они окружены, они могут мгновенно телепортироваться обратно в их родной мир, забирая все сокровища... но хотя бы не украдут ничего в будущем!\n\n",
/*MISSING*/    "Rogues will never make moves which result in their immediate death. Even when cornered, they are able to instantly teleport back to their home world at any moment, taking the treasures forever... but at least they will not steal anything further!\n\n",
/*MISSING*/    "Rogues will never make moves which result in their immediate death. Even when cornered, they are able to instantly teleport back to their home world at any moment, taking the treasures forever... but at least they will not steal anything further!\n\n",
    }},
  {0x2d751c51, { // "underlying tree"
   "drzewo bazowe",
/*MISSING*/    "underlying tree",
   "stromová základní struktura",
   "основное дерево",
/*MISSING*/    "underlying tree",
/*MISSING*/    "underlying tree",
    }},
  {0x2d7c2196, { // "create the model"
   "twórz model",
   "modeli üret",
   "vytvoř model",
   "создать модель",
   "Erzeuge das Modell",
/*MISSING*/    "create the model",
    }},
  {0x2db034b4, { // "first joystick: movement threshold"
   "dżojstik 1: próg ruchu",
   "birinci oyun kolu: hareket eşiği",
   "joystick 1: práh pohybu",
   "джойстик 1: порог движения",
   "Joystick 1: Bewegungsgrenze",
/*MISSING*/    "first joystick: movement threshold",
    }},
  {0x2dbddb1a, { // "game over screen"
   "ekran końca gry",
   "oyun bitiş ekranı",
   "obrazovka konce hry",
   "экран конца игры",
   "Game Over Bildschirm",
/*MISSING*/    "game over screen",
    }},
  {0x2e48d491, { // "Orbs of Safety always appear here, and may be used to escape.\n"
   "Sfery Bezpieczeństwa zawsze można tu znaleźć, by stąd uciec.",
/*MISSING*/    "Orbs of Safety always appear here, and may be used to escape.\n",
   "Sféry Bezpečí se tu dají vždy najít a můžeš je použít k úniku.",
   "Сфера безопасности всегда появляется здесь, и её можно использовать, чтобы уйти.",
/*MISSING*/    "Orbs of Safety always appear here, and may be used to escape.\n",
/*MISSING*/    "Orbs of Safety always appear here, and may be used to escape.\n",
    }},
  {0x2e755a80, { // "Collect four different Elemental Shards!"
   "Znajdź Okruch każdego żywiołu!",
/*MISSING*/    "Collect four different Elemental Shards!",
   "Získej čtyři různé Elementální úlomky!",
   "Собери Осколки всех стихий!",
   "Sammle vier verschiedene Elementscherben!",
/*MISSING*/    "Collect four different Elemental Shards!",
    }},
  {0x2e7b202e, { // "two players"
   "dwóch graczy",
   "iki kişilik",
   "dva hráči",
   "два игрока",
   "Zwei Spieler",
/*MISSING*/    "two players",
    }},
  {0x2ea0ba09, { // "No description yet."
   "Brak opisu.",
   "Henüz bir açıklama yok.",
   "Zatím bez popisu.",
   "Нет описания.",
   "Noch keine Beschreibung...",
/*MISSING*/    "No description yet.",
    }},
  {0x2ea8434f, { // "F8 = SVG shot"
   "F8 = zrzut SVG",
/*MISSING*/    "F8 = SVG shot",
   "F8 = snímek SVG",
   "F8 = снимок SVG",
/*MISSING*/    "F8 = SVG shot",
/*MISSING*/    "F8 = SVG shot",
    }},
  {0x2ee79c64, { // "Enable cheat mode or GAME OVER to use this"
   "Dostępne tylko po końcu gry albo w trybie oszusta",
/*MISSING*/    "Enable cheat mode or GAME OVER to use this",
   "Dostupné pouze v cheat módu nebo po skončení hry",
   "Доступно только в конце игры и в режиме обманщика",
   "Aktivere den Cheat-Modus oder GAME OVER um diese Funktion zu nutzen",
/*MISSING*/    "Enable cheat mode or GAME OVER to use this",
    }},
  {0x2f160195, { // "So better just to concentrate on collecting treasure. Beware the traps, and the guards, who are hard to kill!"
   "Także lepiej się skoncentrować na szukaniu skarbów. Uwaga na pułapki i trudnych do pokonania strażników!",
   "Yani sadece hazineleri toplamaya odaklansan iyi olur. Tuzaklara dikkat! Öldürmesi pek zor olan muhafızlara da dikkat et!",
   "Takže se raději soustřeď jen na sbírání pokladů. Pozor na pasti a na stráže -- je těžké je zabít!",
   "Лучше сконцентрироваться на сборе сокровищ. Остерегайтесь ловушек и стражников, которых сложно убить!",
   "Konzentriere dich besser darauf Schätze zu sammeln. Und achte auf Fallen, sowie Wachen die hart zu töten sind!",
/*MISSING*/    "So better just to concentrate on collecting treasure. Beware the traps, and the guards, who are hard to kill!",
    }},
  {0x2f44452a, { // "You kill %the1."
   "Zabi%łeś0 %a1.",
   "%a1 öldürdün.",
   "Zabi%l0 jsi %a1.",
   "Вы убили %a1.",
   "Du tötest %den1 %a1.",
   "Você mata %na1.",
    }},
  {0x2f86a320, { // "You have to run away from the water!"
   "Musisz uciec z wody!",
   "Sudan kaçmak zorundasın!",
   "Musíš utéct před vodou!",
   "Убегайте из воды!",
   "Du musst vor dem Wasser fliehen!",
/*MISSING*/    "You have to run away from the water!",
    }},
  {0x2fdfb445, { // "Find a %1 in %the2."
   "Znajdź %a1 %abl2.",
/*MISSING*/    "Find a %1 in %the2.",
   "Najdi %a1 %abl2.",
   "Найди %a1 %abl2.",
   "Finde eine %a1 %abl2.",
/*MISSING*/    "Find a %1 in %the2.",
    }},
  {0x2feee764, { // "design the net"
   "projektuj siatkę",
   "ağı tasarla",
   "vytvoř síť",
   "разработать сеть",
   "entwirf das Netz",
/*MISSING*/    "design the net",
    }},
  {0x2ff6c863, { // "One wrong move and it is game over!"
   "Jeden fałszywy ruch i koniec gry!",
   "Tek bir yanlış hamle ve oyun biter!",
   "Jediný špatný tah a hra končí!",
   "Один неверный шаг, и игра закончится!",
   "Eine falsche Bewegung kostet dich den Kopf!",
/*MISSING*/    "One wrong move and it is game over!",
    }},
  {0x2ffa6b17, { // "ON"
   "WŁ",
   "AÇIK",
   "ZAP",
   "ВКЛ",
   "EIN",
   "LIGADO",
    }},
  {0x3038e1b0, { // "\"That should teach you to take me seriously!\""
   "\"Traktuj mnie poważnie! Niech to będzie dla Ciebie nauczką!\"",
   "\"Bu sana beni ciddiye alman gerektiğini öğretir!\"",
   "\"To by tě mělo naučit brát mě vážně!\"",
   "\"Относись ко мне серьёзно! Будет тебе наука!\"",
   "\"Das sollte dir beweisen mich Ernst zu nehmen!\"",
/*MISSING*/    "\"That should teach you to take me seriously!\"",
    }},
  {0x308baacb, { // "turn count = %1 last exploration = %2 ghost timer = %3"
   "kolejek = %1 eksploracja = %2 czaso-duchy = %3",
/*MISSING*/    "turn count = %1 last exploration = %2 ghost timer = %3",
   "počet kol = %1 poslední zkoumání = %2 časovač duchů = %3",
   "количество ходов = %1, последнее исследование = %2, отсчёт времени призраков = %3",
   "Zähler = %1 letzte Erkundung = %2 Geister-Zähler = %3",
/*MISSING*/    "turn count = %1 last exploration = %2 ghost timer = %3",
    }},
  {0x309889bc, { // "camera rotation in ball model"
   "obróć kamerę w modelu kuli",
/*MISSING*/    "camera rotation in ball model",
   "rotace kamery v modelu koule",
   "поворот камеры в модели сферы",
/*MISSING*/    "camera rotation in ball model",
/*MISSING*/    "camera rotation in ball model",
    }},
  {0x30ade9f9, { // "Prince"
   "Książę",
/*MISSING*/    "Prince",
   "Princ",
   "Принц",
   "Prinz",
/*MISSING*/    "Prince",
    }},
  {0x30ae5da6, { // "Some rubble."
   "Gruzy.",
   "Döküntüler.",
   "Sutiny.",
   "Щебень.",
   "Trümmer.",
/*MISSING*/    "Some rubble.",
    }},
  {0x31276816, { // "Tutorial"
   "Podręcznik",
/*MISSING*/    "Tutorial",
   "Tutoriál",
   "Руководство",
/*MISSING*/    "Tutorial",
/*MISSING*/    "Tutorial",
    }},
  {0x3139bed2, { // "triangle grid: not rings"
   "trójkątna siatka: nie pierścienie",
/*MISSING*/    "triangle grid: not rings",
   "trojúhelníková mřížka: bez prstenců",
   "треугольная сетка: нет колец",
/*MISSING*/    "triangle grid: not rings",
/*MISSING*/    "triangle grid: not rings",
    }},
  {0x3187663a, { // "by number"
   "liczba",
/*MISSING*/    "by number",
   "podle počtu",
   "по количеству",
/*MISSING*/    "by number",
/*MISSING*/    "by number",
    }},
  {0x31ba02b2, { // "You can press 'g' or click them in the list to drop a Dead Orb."
   "Wciśnij 'g' lub kliknij Martwe Sfery na liście, by je porzucić.",
   "'g' ye basarak veya listede tıklayarak bir Ölü Küreyi bırakabilirsin.",
   "Mrtvou svéru můžeš položit stiskem klávesy 'g' nebo tím, že na ni klikneš v seznamu.",
   "Вы можете нажать 'g' либо кликнуть на сферу в списке, чтобы бросить её.",
   "Du kannst 'g' drücken oder sie in der Liste anklicken um tote Orbs abzulegen.",
   "Você pode pressionar 'g' ou clicar num Orbe Morto na lista para descartá-los.",
    }},
  {0x31d0e1f7, { // "MENU"
   "MENU",
   "MENÜ",
   "MENU",
   "МЕНЮ",
   "MENÜ",
/*MISSING*/    "MENU",
    }},
  {0x31eb9ca2, { // "Look for the Orbs of Yendor in Hell or in the Crossroads!"
   "Szukaj Sfer Yendoru w Piekle albo na Skrzyżowaniu!",
   "Yendorun Küreleri için Cehennemi yahut Arayolları ara!",
   "Hledej Yendorské sféry v Pekle nebo na Křižovatce!",
   "Ищите сферы Йендора в Аду и на Перекрёстке!",
   "Such die Orbs von Yendor in der Hölle oder auf den Kreuzungen!",
   "Procure pelos Orbes de Yendor no Inferno e nas Encruzilhadas!",
    }},
  {0x322e00a2, { // "x: %1"
   "x: %1",
   "x: %1",
   "x: %1",
   "x: %1",
   "x: %1",
/*MISSING*/    "x: %1",
    }},
  {0x32437fa4, { // "six players"
   "sześciu graczy",
/*MISSING*/    "six players",
   "šest hráčů",
   "шесть игроков",
   "sechs Spieler",
/*MISSING*/    "six players",
    }},
  {0x328fdd3c, { // "DRAG"
   "DRAG",
   "SRKLE",
   "TAŽENÍ",
   "DRAG",
   "ZIEH",
/*MISSING*/    "DRAG",
    }},
  {0x32949f8b, { // "Select the land ---"
   "Wybierz krainę ---",
   "Diyar seç ---",
   "Vyber si kraj ---",
   "Выбрать землю ---",
   "Wähle das Land ---",
/*MISSING*/    "Select the land ---",
    }},
  {0x32ae3e1f, { // "Just a nasty shark."
   "Zwykły groźny rekin.",
   "Sadece vahşi bir köpekbalığı.",
   "Obyčejný, nebezpečný žralok.",
   "Просто противная акула.",
   "Ist bloß ein fieser Hai.",
/*MISSING*/    "Just a nasty shark.",
    }},
  {0x32b6d522, { // "Aquatic region -- accessible only from coastal regions and other aquatic regions.\n"
   "Kraina morska -- dostępna jedynie zs krain przybrzeżnych i morskic.\n",
/*MISSING*/    "Aquatic region -- accessible only from coastal regions and other aquatic regions.\n",
   "Kraina morska -- dostępna jedynie zs krain przybrzeżnych i morskic.\n",
   "Морской регион -- доступен только из прибрежных и других морских регионов.\n",
/*MISSING*/    "Aquatic region -- accessible only from coastal regions and other aquatic regions.\n",
/*MISSING*/    "Aquatic region -- accessible only from coastal regions and other aquatic regions.\n",
    }},
  {0x32bb448d, { // "The Poincaré disk model is a model of a hyperbolic *plane* -- you might wonder why are the walls rendered in 3D then.\n\nHyperRogue actually assumes that the floor level is an equidistant surface in a three-dimensional hyperbolic world, and the camera is placed above the plane that the surface is equidistant to (which boils down to showing the floor level in Poincaré disk model).\n\nThis is shown on the next slide, in the Poincaré ball model, which is the 3D analog of the Poincaré disk model."
   "Model dysku Poincaré to model *płaszczyny* hiperbolicznej -- być może się zastanawiasz, jak w takim razie rozumieć ściany w 3D.\n\nHyperRogue zakłada, że powierzchnia podłogi to powierzchnia ekswidystantna w trójwymiarowym hiperbolicznym świecie, a kamera znajduje się nad powierzchnią, do której podłoga jest ekwidystantna -- co powoduje, że podłoga wygląda dokładnie jak w modelu Poincaré.\n\nPokazuje to kolejny slajd, w modelu kuli, trójwymiarowym odpowiedniku modelu dysku.",
/*MISSING*/    "The Poincaré disk model is a model of a hyperbolic *plane* -- you might wonder why are the walls rendered in 3D then.\n\nHyperRogue actually assumes that the floor level is an equidistant surface in a three-dimensional hyperbolic world, and the camera is placed above the plane that the surface is equidistant to (which boils down to showing the floor level in Poincaré disk model).\n\nThis is shown on the next slide, in the Poincaré ball model, which is the 3D analog of the Poincaré disk model.",
   "Poincarého disk je model hyperbolické *roviny* -- možná si tedy říkáš, proč se zdi zobrazují ve 3D.\n\nHyperRogue ve skutečnosti předpokládá, že podlaha je ekvidistantní plocha ve trojrozměrném hyperbolickém světě, a že kamera se nachází nad rovinou, ke které je tato plocha ekvidistantní (což ve výsledku znamená, že se podlaha zobrazuje jako Poincarého disk).\n\nTo můžeš vidět v následujícím snímku, který ukazuje Poincarého kouli -- 3D analog Poincarého disku.",
   "Модель диска Пуанкаре -- это модель гиперболической *плоскости*, и ты мог задуматься, почему же стены отображаются в 3D.\n\nHyperRogue на самом деле предполагает, что уровень пола -- это эквидистанта в трёхмерном гиперболическом мире, а камера находится выше плоскости, эквидистантой к которой является поверхность -- таким образом пол выглядит в точности как в модели Пуанкаре.\n\nЭто показано на следующем слайде, в модели сферы, трёхмерном аналоге модели диска.",
/*MISSING*/    "The Poincaré disk model is a model of a hyperbolic *plane* -- you might wonder why are the walls rendered in 3D then.\n\nHyperRogue actually assumes that the floor level is an equidistant surface in a three-dimensional hyperbolic world, and the camera is placed above the plane that the surface is equidistant to (which boils down to showing the floor level in Poincaré disk model).\n\nThis is shown on the next slide, in the Poincaré ball model, which is the 3D analog of the Poincaré disk model.",
/*MISSING*/    "The Poincaré disk model is a model of a hyperbolic *plane* -- you might wonder why are the walls rendered in 3D then.\n\nHyperRogue actually assumes that the floor level is an equidistant surface in a three-dimensional hyperbolic world, and the camera is placed above the plane that the surface is equidistant to (which boils down to showing the floor level in Poincaré disk model).\n\nThis is shown on the next slide, in the Poincaré ball model, which is the 3D analog of the Poincaré disk model.",
    }},
  {0x32c11e4f, { // "OFF"
   "WYŁ",
   "KAPALI",
   "VYP",
   "ВЫКЛ",
   "AUS",
   "DESLIGADO",
    }},
  {0x33150354, { // "You wonder what does it mean?"
   "Co to może znaczyć?",
   "Bunun ne demek olduğunu merak ettin",
   "Co to může znamenat?",
   "Вы спросите, а что это значит?",
   "Du wunderst dich, was es bedeutet.",
   "Você se pergunta: O que isso significa?",
    }},
  {0x33329502, { // "You have improved your '%1' high score on Steam!"
   "Poprawi%łeś0 swój wynik w kategorii '%P1' na Steam!",
   "Steam'deki '%P1'puanını artırdın!",
   "Zlepši%l0 jsi své nejvyšší skóre na Steamu v kategorii '%P1'!",
   "Вы подняли свой счёт в категории '%1' в Steam!",
   "Du hast deinen Highscore bei '%1' auf Steam verbessert!",
   "Você superou sua '%1' melhor pontuação da Steam!",
    }},
  {0x33491116, { // "Cheats: "
   "Oszustwa: ",
   "Hileler: ",
   "Cheaty: ",
   "Читы: ",
   "Cheats: ",
/*MISSING*/    "Cheats: ",
    }},
  {0x338f9dd8, { // "This Orb allows you to create illusions of yourself. Illusions are targeted by most monsters, just like yourself, Thumpers, and your friends.\n\nEach illusion takes 5 charges to create, and one extra charge per turn. You can also click your illusion to take it away, restoring 4 charges.\n\nIf you have both Orb of Teleport and Orb of Trickery, Illusion is cast first -- you can then teleport on your Illusion to switch places with it."
   "Ta sfera pozwala Ci tworzyć iluzje samego siebie. Większość potworów celuje w iluzje, podobnie jak w Ciebie, Dudniki i Twoich przyjaciół.\n\nKażda iluzja wymaga 5 ładunków do stworzenia i jednego dodatkowego ładunku na kolejkę. Możesz też kliknąć swoją iluzję, by ją zabrać, co przywraca 4 ładunki.\n\nJeśli masz zarówno Sferę Teleportacji jak i Sztuczek, Iluzja rzucana jest jako pierwsza -- możesz następnie teleportować się na swoją iluzję, by się z nią zamienić miejscami.",
   "Bu Küre kendinin ilüzyonlarını çıkarmanı sağlar. İlüzyonlar tıpkı sen, Gümleyenler ve dostların gibi pek çok canavar tarafından saldırıya uğrayabilir.\n\nHer ilüzyonu çıkarması 5 şarj ve her ek tur bir şarj tüketir. İlüzyona tıklayarak onu yok edip, 4 şarjını geri alabilirsin. \n\nAynı anda hem Kandırma hem Işınlanma Küren varsa, önce İlüzyon uygulanır, sonra istersen ilüzyonuna tıklayarak onla yer değiştirebilirsin. ",
   "Tato sféra ti umožňuje vytvářet své iluzorní dvojníky. Většina netvorů půjde po tvých iluzích, stejně jako jdou po tobě, po Tloucích a po tvých přátelích.\n\nStvoření každé iluze sníží náboj sféry o 5 a vyčerpává 1 náboj navíc za kolo. Můžeš také kliknout na svou iluzi a zrušit ji, čímž se náboj sféry zvýší o 4.\n\nPokud máš Sféru Teleportace a Sféru Klamu zároveň, nejprve použiješ Sféru Klamu -- poté se můžeš teleportovat na svou iluzi a vyměnit si s ní místo.",
   "Эта сфера позволяет Вам создавать иллюзии. Их будут атаковать почти все монстры, так же как Вас, Тамперы и ваших союзников.\n\nКаждая иллюзия требует 5 зарядов для создания и по одному в каждый ход. Также Вы можете кликнуть на иллюзию, чтобы забрать её, потратив 3 заряда.\n\nЕсли у Вас есть сферы Хитрости и Телепорта, Иллюзия работает первой -- Вы можетепотом телепортироваться на неё, чтобы поменяться с ней местами.",
   "Dieser Orb erlaubt es dir, Illusionen deiner selbst zu erschaffen. Diese werden von den meisten Monstern verfolgt.\n\nJede Illusion benötigt 5 Ladungen zum Erschaffen, plus eine Ladung pro Zug. Du kannst deine Illusion anklicken um sie zu entfernen, wofür du 4 Ladungen bekommst.\n\nBesitzt du einen Orb der Teleportation und einen des Truges, wird Illusion bevorzugt du kannst dich dann auf deine Illusion teleportieren, um die Plätze zu tauschen.",
/*MISSING*/    "This Orb allows you to create illusions of yourself. Illusions are targeted by most monsters, just like yourself, Thumpers, and your friends.\n\nEach illusion takes 5 charges to create, and one extra charge per turn. You can also click your illusion to take it away, restoring 4 charges.\n\nIf you have both Orb of Teleport and Orb of Trickery, Illusion is cast first -- you can then teleport on your Illusion to switch places with it.",
    }},
  {0x339454d5, { // "player 2 go"
   "gracz 2 idź",
   "oyuncu 2, git",
   "hráč 2 pohyb",
   "игрок 2 идёт",
   "Spieler 2 los",
/*MISSING*/    "player 2 go",
    }},
  {0x339e2c87, { // "puzzles and exploration"
   "zagadki i zwiedzanie",
/*MISSING*/    "puzzles and exploration",
   "hádanky a zkoumání",
   "головоломки и изучение",
/*MISSING*/    "puzzles and exploration",
/*MISSING*/    "puzzles and exploration",
    }},
  {0x33aea165, { // "Just a typical hyperbolic pirate."
   "Typowy hiperboliczny pirat.",
   "Sadece bildiğiniz bir hiperbolik korsan.",
   "Typický hyperbolický pirát.",
   "Обычный гиперболический пират.",
   "Ist bloß ein typischer, hyperbolischer Pirat.",
/*MISSING*/    "Just a typical hyperbolic pirate.",
    }},
  {0x33d27403, { // "\n\nSee sounds/credits.txt for credits for sound effects"
   "\n\nPodziękowania za efekty dźwiękowe w sounds/credits.txt",
/*MISSING*/    "\n\nSee sounds/credits.txt for credits for sound effects",
   "\n\nPodìkování za zvukové efekty najde v souboru sounds/credits.txt",
   "\n\nPodziękowania za efekty dźwiękowe w sounds/credits.txt",
/*MISSING*/    "\n\nSee sounds/credits.txt for credits for sound effects",
/*MISSING*/    "\n\nSee sounds/credits.txt for credits for sound effects",
    }},
  {0x33fb8e0b, { // "%The1 squeaks sadly."
   "%1 pisn%ął1 smutno.",
   "%1 üzüntüyle ciyaklıyor.",
   "%1 smutně piští.",
   "%1 пищит грустно.",
   "%Die1 %1 piepst traurig.",
/*MISSING*/    "%The1 squeaks sadly.",
    }},
  {0x3403ba38, { // "model used (not conformal!)"
   "użyty model (nie konforemny!)",
/*MISSING*/    "model used (not conformal!)",
   "použitý model (nekonformální!)",
   "модель используется (не конформная!)",
   "verwendetes Modell (nicht konform!)",
/*MISSING*/    "model used (not conformal!)",
    }},
  {0x3414b24b, { // "degree of the approximation"
   "stopień przybliżenia",
/*MISSING*/    "degree of the approximation",
   "stupeň opień przybliżenia",
   "степень приближения",
   "Grad der Näherung",
/*MISSING*/    "degree of the approximation",
    }},
  {0x343f8bdc, { // "Dark red birds who have created the minefield.\n\nThey create a mine on the spot where they are killed, provided that the terrain is suitable. Also note that mines are triggered by dead birds falling on them."
   "Ciemnoczerwone ptaki, które stworzyły pole minowe.\n\nNa polu, na które spadły, powstaje mina, o ile teren jest odpowiedni. Pamiętaj, że mina wybucha, gdy spadnie na nią martwy ptak.",
   "Mayıntarlasını düzenleyen kızıl, siyah kuşlar.\n\nÖldüklerinde öldükleri hücrede, eğer hücre tipi uygunsa bir mayın çıkarırlar. Ayrıca dikkat, mayınlar, üzerine düşen ölü kuşlarla da tetiklenebilir.",
   "Tmavočervení ptáci, kteří vytvořili minové pole.\n\nKdyž zemřou, vytvoří na poli, kam spadnou, minu, pokud to terén dovoluje. Měj na paměti, že i pád mrtvého ptáka na minu stačí k tomu, aby ji odpálil!",
   "Тёмно-чёрные птицы, создавшие Минное поле.\n\nКогда птица умирает, на клетке, где она была, остаётся мина. Помни, что мёртвые птицы могут активировать мины.",
   "Dunkelrote Vögel, die das Minenfeld erschufen.\n\nWenn sie sterben erschaffen sie eine Mine, vorausgesetzt das Terrain passt). Tote, fallende Vögel lösen Minen auch aus!",
/*MISSING*/    "Dark red birds who have created the minefield.\n\nThey create a mine on the spot where they are killed, provided that the terrain is suitable. Also note that mines are triggered by dead birds falling on them.",
    }},
  {0x34ba6ad5, { // "Are you sure?"
   "Czy na pewno?",
/*MISSING*/    "Are you sure?",
   "Určitě?",
   "Вы уверены?",
/*MISSING*/    "Are you sure?",
/*MISSING*/    "Are you sure?",
    }},
  {0x3517dd24, { // "vineyard pattern"
   "wzór winnicy",
/*MISSING*/    "vineyard pattern",
   "vinicový vzor",
   "узор виноградника",
/*MISSING*/    "vineyard pattern",
/*MISSING*/    "vineyard pattern",
    }},
  {0x352f28c1, { // "These bright yellow gems can be found only by those who have mastered the Crossroads."
   "Te jasne, żółte klejnoty mogą znaleźć tylko mistrze Skrzyżowań.",
   "Bu sarı parlak mücevherler Arayollarda yeterince deneyim kazananlar tarafından bulunabilir.",
   "Tyto jasně žluté drahokamy mohou najít pouze ti, kdo ovládli celou Křižovatku.",
   "Эти ярко-жёлтые камни может найти лишь тот, кто стал мастером Перекрёстка.",
   "Diese hellen, gelben Edelsteine können nur von denen gefunden werden, die die Kreuzungen gemeistert haben.",
/*MISSING*/    "These bright yellow gems can be found only by those who have mastered the Crossroads.",
    }},
  {0x35794dd5, { // "Click this to see your orbs."
   "Kliknij by zobaczyć Twoje sfery.",
/*MISSING*/    "Click this to see your orbs.",
   "Kliknutím zde zobrazíš své Sféry.",
   "Нажми, чтобы увидеть твои Сферы.",
/*MISSING*/    "Click this to see your orbs.",
/*MISSING*/    "Click this to see your orbs.",
    }},
  {0x35b1bb91, { // "rainbow landscape"
   "tęczowy krajobraz",
/*MISSING*/    "rainbow landscape",
   "duhová krajina",
   "tęczowy krajobraz",
/*MISSING*/    "rainbow landscape",
/*MISSING*/    "rainbow landscape",
    }},
  {0x35b618da, { // "second joystick: pan threshold"
   "dżojstik 2: próg przewijania",
   "ikinci oyun kolu: gezdirme hızı",
   "joystick 2: práh posuvu",
   "джойстик 2: порог панорамы",
   "Joystick 2: Scrollgrenze",
/*MISSING*/    "second joystick: pan threshold",
    }},
  {0x35be8905, { // "floor/pattern"
   "podłoga/wzorek",
   "zemin/doku",
   "podlaha/vzor",
   "пол/узор",
   "Boden/Muster",
/*MISSING*/    "floor/pattern",
    }},
  {0x35d1f8a2, { // "When Pikemen move, they attack all cells which are now adjacent to them. Luckily, they can be killed in the same way.\n\nThey never move if this would attack their friends."
   "Kiedy Pikinier się rusza, atakuje wszystkie komórki, z którymi sąsiaduje po wykonaniu ruchu. Na szczęście, jego samego można pokonać taką samą metodą.\n\nPikinier się nie rusza, jeśli by w ten sposób zaatakował przyjaciela.",
   "Mızraklılar hareket ettiğinde değdikleri tüm hücrelere saldırırlar. Çok şükür aynı şekilde öldürülebiliyorlar da. Arkadaşlarına hasar vereceklerse hareket etmiyorlar.",
   "Kopiníci po každém svém pohybu zaútočí na všechna okolní políčka Naštěstí se dají zabít stejným způsobem.\n\nKopiníci se nepohnou, pokud by to znamenalo, že zaútočí na své přátele.",
   "Когда Копейщик ходит, он атакует все клетки рядом с ним. К счастью, и его можно убить подобным образом.\n\nКопейщик никогда не будет атаковать дружественной ему существо.",
   "Bewegen sich Bergmänner, greifen sie alle Zellen um sie herum an. Glücklicherweise können sie genauso getötet werden.\n\nSie bewegen sich nicht wenn sie so ihre Freunde angreifen würden.",
/*MISSING*/    "When Pikemen move, they attack all cells which are now adjacent to them. Luckily, they can be killed in the same way.\n\nThey never move if this would attack their friends.",
    }},
  {0x35df8189, { // "1-9 = rotations: %1"
   "1-9 = obroty: %1",
   "1-9 = döndürmeler %1",
   "1-9 = rotace: %1",
   "1-9 = поворот: %1",
   "1-9 = Drehungen: %1",
/*MISSING*/    "1-9 = rotations: %1",
    }},
  {0x362bf409, { // "%The1 kills %the2!"
   "%1 zabi%ł1 %a2!",
   "%1 %a2 öldürdü!",
   "%1 zabi%l1 %a2!",
   "%1 убил%E1 %a2!",
   "%Der1 %1 tötet %den2 %a2!",
/*MISSING*/    "%The1 kills %the2!",
    }},
  {0x3671b5b6, { // "This tour shows just a small part of what you can see in the world of HyperRogue. For example, hyperbolic mazes are much nicer than their Euclidean counterparts. Have fun exploring!\n\nPress '5' to leave the tutorial mode."
   "Ten podręcznik pokazuje tylko małą część świata HyperRogue. Na przykład hiperboliczne labirynty są ciekawsze niż euklidesowe. Miłej zabawy!Naciśnij '5' by opuścić podręcznik.",
/*MISSING*/    "This tour shows just a small part of what you can see in the world of HyperRogue. For example, hyperbolic mazes are much nicer than their Euclidean counterparts. Have fun exploring!\n\nPress '5' to leave the tutorial mode.",
   "Tato prohlídka ukazuje jen malou část toho, co můžeš vidět ve světě HyperRogue. Tak například: hyperbolická bludiště jsou mnohem hezčí než jejich eukleidovské protějšky. Bav se při dalším zkoumání!\n\n.Stisknutím klávesy '5' opustíš tutoriál.",
   "Это путешествие показало лишь малую часть мира HyperRogue. Например, гиперболические лабиринты намного приятнее евклидовых. Приятной игры!Нажмите '5', чтобы выйти из режима Руководства.",
/*MISSING*/    "This tour shows just a small part of what you can see in the world of HyperRogue. For example, hyperbolic mazes are much nicer than their Euclidean counterparts. Have fun exploring!\n\nPress '5' to leave the tutorial mode.",
/*MISSING*/    "This tour shows just a small part of what you can see in the world of HyperRogue. For example, hyperbolic mazes are much nicer than their Euclidean counterparts. Have fun exploring!\n\nPress '5' to leave the tutorial mode.",
    }},
  {0x36849629, { // "gnomonic projection"
   "rzut gnomoniczny",
/*MISSING*/    "gnomonic projection",
   "gnómonická projekce",
   "гномоническая проекция",
/*MISSING*/    "gnomonic projection",
/*MISSING*/    "gnomonic projection",
    }},
  {0x36cf01a9, { // "Ground level below the plane"
   "Poziom podłoża pod płaszczyzną",
/*MISSING*/    "Ground level below the plane",
   "Úroveň země pod rovinou",
   "Глубина уровня земли",
/*MISSING*/    "Ground level below the plane",
/*MISSING*/    "Ground level below the plane",
    }},
  {0x36f58dfe, { // "A tribe of men native to the Desert. They have even tamed the huge Sandworms, who won't attack them."
   "Plemię ludzi żyjących na Pustyni. Oswoili oni Pustynne Czerwie, dzięki czemu żyją razem pokojowo.",
   "Çölde yaşayan bir insan jabilesi. Büyük Kumkurtlarını evcilleştirip kendilerine saldırmamasını sağladılar.",
   "Kmen domorodců z Pouště. Dokázali dokonce zkrotit i obrovské Písečné červy, kteří na ně neútočí.",
   "Племя людей, живущих в пустыне. Они приручили песчаных червей, и те их не трогают.",
   "Ein Volk, das in der Wüste lebt. Sie haben es sogar geschafft, die riesigen Sandwürmer so zu zähmen, dass diese sie nicht angreifen.",
/*MISSING*/    "A tribe of men native to the Desert. They have even tamed the huge Sandworms, who won't attack them.",
    }},
  {0x36fb778b, { // "The cloud turns into a bunch of images!"
   "Chmura zamienia się w obrazy!",
   "Bulut pek çok görüntüye dönüştü!",
   "Oblak se proměnil v několik obrazů!",
   "Облако превратилось в группу изображений!",
   "Die Wolke verwandelt sich in mehrere Bilder!",
   "A nuvem se transforma em várias imagens!",
    }},
  {0x376564fa, { // "b = switch auto"
   "b = ustaw auto",
/*MISSING*/    "b = switch auto",
   "b = pøepni auto",
   "b = автопереключение",
/*MISSING*/    "b = switch auto",
/*MISSING*/    "b = switch auto",
    }},
  {0x37e4291e, { // "orb power: %1"
   "moc: %1",
   "küre gücü: %1",
   "moc sfér: %1",
   "сила сферы: %1",
   "Orb Kraft: %1",
   "poder de orbes: %1",
    }},
  {0x3804a2f2, { // "movement animation speed"
   "prędkość animacji ruchu",
/*MISSING*/    "movement animation speed",
   "rychlost pohybových animací",
   "prędkość animacji ruchu",
/*MISSING*/    "movement animation speed",
/*MISSING*/    "movement animation speed",
    }},
  {0x384e4b84, { // "This tutorial is designed for computers, and keys are given for all actions. It will work without a keyboard though, although less comfortably -- just ignore the keys given and select options from MENU.\n\nSelect 'next slide' from MENU."
   "Ten tutorial jest przeznaczony głównie dla komputerów, i klawisze są podane dla wszystkich akcji. Działa jednak bez klawiatury, choć mniej wygodnie -- ignoruj klawisze i wybieraj opcje z MENU.\n\ngiven and select options from MENU.\n\nWybierz z MENU 'następny slajd'.",
/*MISSING*/    "This tutorial is designed for computers, and keys are given for all actions. It will work without a keyboard though, although less comfortably -- just ignore the keys given and select options from MENU.\n\nSelect 'next slide' from MENU.",
   "Tento tutorial je navržený pro počítače a u všech akcí jsou uvedeny klávesy. Funguje i bez klávesnice, i když ne tak pohodlně -- prostě ignorujte udávané klávesy a vybírejte možnosti z MENU.\n\nVyber z MENU 'další snímek'.",
   "Это руководство сделано для компьютеров, и клавиши требуются для всех действий. Им можно пользоваться и без клавиатуры, хотя и не так удобно -- просто игнорируй их и выбирай действия в МЕНЮ.\n\nВыбери в МЕНЮ 'следующий слайд'.",
/*MISSING*/    "This tutorial is designed for computers, and keys are given for all actions. It will work without a keyboard though, although less comfortably -- just ignore the keys given and select options from MENU.\n\nSelect 'next slide' from MENU.",
/*MISSING*/    "This tutorial is designed for computers, and keys are given for all actions. It will work without a keyboard though, although less comfortably -- just ignore the keys given and select options from MENU.\n\nSelect 'next slide' from MENU.",
    }},
  {0x387a2936, { // "leave the Tutorial"
   "opuść Podręcznik",
/*MISSING*/    "leave the Tutorial",
   "opusť Tutoriál",
   "выйти из Руководства",
/*MISSING*/    "leave the Tutorial",
/*MISSING*/    "leave the Tutorial",
    }},
  {0x38b21089, { // "Height of walls"
   "Wysokość ścian",
/*MISSING*/    "Height of walls",
   "Výška zdí",
   "Высота стен",
/*MISSING*/    "Height of walls",
/*MISSING*/    "Height of walls",
    }},
  {0x3908e6a3, { // "save the current config"
   "zapisz obecną konfigurację",
   "mevcut ayarları kaydet",
   "uložit současné nastavení",
   "сохранить текущие настройки",
   "momentane Einstellungen speichern",
   "salvar as configurações",
    }},
  {0x39419318, { // " to surface"
   " do wynurzenia",
/*MISSING*/    " to surface",
   " do vynoøení",
   " на поверхность",
/*MISSING*/    " to surface",
/*MISSING*/    " to surface",
    }},
  {0x395bc20c, { // "player 3 X"
   "gracz 3 X",
/*MISSING*/    "player 3 X",
   "hráč 3 X",
   "игрок 3 X",
   "Spieler 3 X",
/*MISSING*/    "player 3 X",
    }},
  {0x395bc20d, { // "player 3 Y"
   "gracz 3 Y",
/*MISSING*/    "player 3 Y",
   "hráč 3 Y",
   "игрок 3 Y",
   "Spieler 3 Y",
/*MISSING*/    "player 3 Y",
    }},
  {0x39741514, { // "Choose from the lands visited this game."
   "Wybierz spośród odwiedzonych krain.",
   "Oyunda eriştiğin diyarlardan birini seç.",
   "Vyber si z krajů, které jsi v této hře navštívil.",
   "Выберите одну из уже посещённых земель.",
   "Wähle aus den Ländern, die diese Sitzung besucht wurden.",
/*MISSING*/    "Choose from the lands visited this game.",
    }},
  {0x3989d384, { // "Config file: %1"
   "Plik konfiguracyjny: %1",
   "Ayar dosyası: %1",
   "Soubor s nastavením: %1",
   "Файл конфигурации: %1",
   "Confiq Datei: %1",
   "Arquivo de configuração: %1",
    }},
  {0x39cf3230, { // "Gans model"
   "Model Gansa",
/*MISSING*/    "Gans model",
   "Gansův model",
   "Модель Ганса",
/*MISSING*/    "Gans model",
/*MISSING*/    "Gans model",
    }},
  {0x3b247efe, { // "configure keys"
   "klawisze skonfigurowane",
/*MISSING*/    "configure keys",
   "konfigurace kláves",
   "настроить клавиши",
   "Tasten konfigurieren",
/*MISSING*/    "configure keys",
    }},
  {0x3b430e32, { // "Animated corpses of ancient Viking warriors. They are immune to mundane weapons, but they can be destroyed by your Orb of the Sword."
   "Ożywione zwłoki starożytnych wikińskich wojowników. Draugry są odporne na zwyczajne bronie, ale mogą zostać zniszczone przy użyciu Sfery Ostrza.",
/*MISSING*/    "Animated corpses of ancient Viking warriors. They are immune to mundane weapons, but they can be destroyed by your Orb of the Sword.",
   "Oživlé mrtvoly dávných vikingských válečníků. Jsou imunní proti obyčejným zbraním, ale tvoje Sféra Meče je může zničit.",
   "Оживлённые трупы древних викингских воинов. Драугров нельзя победить обычным оружием, но они уязвимы к Сфере Меча.",
   "Untote Helden der alten Wikinger. Sie sind immun gegenüber irdischen Waffen, aber sie können durch deinen Orb des Schwertes zerstört werden.",
/*MISSING*/    "Animated corpses of ancient Viking warriors. They are immune to mundane weapons, but they can be destroyed by your Orb of the Sword.",
    }},
  {0x3b6161d4, { // "Humans are %1 absolute units high. Your head travels %2 times the distance travelled by your feet."
   "Ludzie mają wysokość %1. Twoja głowa porusza się %2 razy szybciej niż Twoje nogi.",
/*MISSING*/    "Humans are %1 absolute units high. Your head travels %2 times the distance travelled by your feet.",
   "Lidé jsou vysocí %1 absolutních jednotek. Tvá hlava urazí %2-násobek vzdálenosti, kterou urazí tvé nohy.",
   "Человек имеет рост %1. Ваша голова проходит расстояние в %2 раз больше, чем ноги.",
/*MISSING*/    "Humans are %1 absolute units high. Your head travels %2 times the distance travelled by your feet.",
/*MISSING*/    "Humans are %1 absolute units high. Your head travels %2 times the distance travelled by your feet.",
    }},
  {0x3b63146e, { // "You become a bit nervous..."
   "Czujesz się zagrożon%y0...",
/*MISSING*/    "You become a bit nervous...",
   "Začínáš být trošku nervózní...",
   "Вы начали нервничать...",
   "Du wirst ein wenig nervös...",
/*MISSING*/    "You become a bit nervous...",
    }},
  {0x3b698746, { // "Finished lands required: %1 (collect %2 treasure)\n"
   "Wymagane ukończone krainy: %1 (zdobądź %2 skarbów)\n",
   "Gereken bitirilmiş diyarlar: %1 (%2 hazine toplananlar)\n",
   "Potřebuješ dokončit %1 krajů (získat v nich %2 pokladů)\n",
   "Земель окончено: %1 (собрано %2 сокровищ)\n",
   "Abgeschlossene Länder benötigt: %1 (sammle %2 Schätze)\n",
/*MISSING*/    "Finished lands required: %1 (collect %2 treasure)\n",
    }},
  {0x3b7bee06, { // "return to the game"
   "wróć do gry",
   "oyuna dön",
   "vrať se do hry",
   "вернуться в игру",
   "zurück zum Spiel",
/*MISSING*/    "return to the game",
    }},
#if REPEATED
  {0x3b7bee07, { // "return to the game [repeat]"
   "powrót do gry",
/*MISSING*/    "return to the game [repeat]",
   "návrat do hry",
   "вернуться в игру",
   "zum Spiel zurückkehren",
/*MISSING*/    "return to the game [repeat]",
    }},
#endif
  {0x3b960c12, { // "This cell is on fire. Most beings and items cannot survive."
   "To pole się pali. Większość stworzeń i przedmiotów tu nie przetrwa.",
   "Bu hücre yanıyor. Üzerindeki çoğu eşya ve canlı ölecek.",
   "Toto políčko hoří. Většina tvorů a předmětů zde nepřežije.",
   "Эта клетка в огне. Большинство существ и предметов не могут выжить.",
   "Diese Zelle brennt. Die meisten Items und Monster können dies nicht überleben.",
/*MISSING*/    "This cell is on fire. Most beings and items cannot survive.",
    }},
  {0x3b9db20b, { // "b = boundary"
   "i = przedmioty",
   "i = sınır",
   "c = kopírování",
   "i = предметы",
   "i = Grenze",
/*MISSING*/    "b = boundary",
    }},
  {0x3bae8fbf, { // "A Witch without any special powers. But watch out! She will pick up any basic Orbs on her path, and use their powers."
   "Czarownica bez specjalnych mocy. Uważaj! Może ona podnieść dowolną podstawową Sferę na swojej drodze i użyć jej mocy.",
   "Herhangi bir gücü olamyan bir Cadı. Ama dikkat! Yolunun üzerindeki herhangi bir Küreyi alıp güçlerini kullanabilir.",
   "Čarodějnice bez jakýchkoli zvláštních schopností. Ale pozor! Učednice sbírají jakékoli základní Sféry, na které narazí a používají jejich schopnosti.",
   "Ведьма без особых сил. Но будьте осторожны! Она может поднять сферу на своём пути и забрать её силу.",
   "Eine Hexe ohne besondere Kräfte. Aber pass auf! Sie wird alle Orbs in ihrer Nähe aufheben und verwenden.",
/*MISSING*/    "A Witch without any special powers. But watch out! She will pick up any basic Orbs on her path, and use their powers.",
    }},
  {0x3c1727e4, { // "Four mines next to you!"
   "Cztery miny koło Ciebie!",
   "Yakınında dört mayın var!",
   "Ve tvé blízkosti se nacházejí čtyři miny!",
   "Четыре мины рядом с Вами!",
   "4 Minen um dich!",
/*MISSING*/    "Four mines next to you!",
    }},
  {0x3c191f07, { // "The more Hypersian Rugs you collect, the harder it is."
   "Im więcej zbierzesz Hiperskich Dywanów, tym misja trudniejsza.",
   "Ne kadar Aşkınlı Kilim toplarsan o kadar zor.",
   "Čím víc Hyperských koberců sebereš, tím obtížnější bude.",
   "Чем больше Вы нашли гиперсидских ковров, тем это сложнее.",
   "Je mehr Hypersische Teppiche du sammelst desto schwieriger wird es.",
/*MISSING*/    "The more Hypersian Rugs you collect, the harder it is.",
    }},
  {0x3c656179, { // "The tasty treat increases your protection."
   "Pyszny Cukierek poprawia Twoją ochronę.",
/*MISSING*/    "The tasty treat increases your protection.",
   "Chutná Pochoutka zvyšuje tvou ochranu.",
   "Вкусные конфетки улучшат Вашу защиту.",
/*MISSING*/    "The tasty treat increases your protection.",
/*MISSING*/    "The tasty treat increases your protection.",
    }},
  {0x3c66765d, { // "spherical geometry"
   "geometria sferyczna",
/*MISSING*/    "spherical geometry",
   "sférická geometrie",
   "сферическая геометрия",
/*MISSING*/    "spherical geometry",
/*MISSING*/    "spherical geometry",
    }},
  {0x3c8b2988, { // "NEW"
   "NOWA",
   "YENİ",
   "NOVÁ HRA",
   "NEW",
   "NEU",
/*MISSING*/    "NEW",
    }},
  {0x3cdbf908, { // "hypersian rug mode"
   "tryb hiperskiego dywanu",
/*MISSING*/    "hypersian rug mode",
   "mód hyperského koberce",
   "режим гиперсидского ковра",
   "Hypersischer-Teppich-Modus",
/*MISSING*/    "hypersian rug mode",
    }},
  {0x3ced91b0, { // "When the tide is away, beautiful ambers can be found on the hyperbolic beaches. Perhaps there used to be a forest here?"
   "W czasie odpływu na hiperbolicznej plaży można znaleźć piękne bursztyny.",
   "Dalga uzaktayken şahane kehribarlar hiperbolik kumsallarda bulunabilir. Belki önceleri burada bir orman vardı?",
   "Za odlivu můžeš na hyperbolické pláži najít krásné kusy jantaru.Možná tu kdysi býval les?",
   "Во время отлива на гиперболических берегах можно найти прекрасный Янтарь. Може, здесь когда-то был лес?",
   "Bei Ebbe können wunderschöne Bernsteine auf den hyperbolischen Stränden gefunden werden. Vielleicht war hier früher ein Wald...?",
/*MISSING*/    "When the tide is away, beautiful ambers can be found on the hyperbolic beaches. Perhaps there used to be a forest here?",
    }},
  {0x3cfe1720, { // "You feel that you have enough treasure to access new lands!"
   "Masz wystarczająco wiele skarbów, by znaleźć nowe krainy!",
   "Yeni diyarlara ulaşmak için yeterince hazinen olduğunu düşünüyorsun!",
   "Cítíš, že už máš dost pokladů na to aby ses dostal do nových krajů!",
   "Вы чувствуете, что у Вас достаточно сокровищ, чтобы открыть новые земли!",
   "Du spürst, dass du genug Schätze gesammelt hast um neue Länder zu erreichen!",
   "Você sente que tem tesouros o suficiente para acessar novas terras!",
    }},
  {0x3d0c36c8, { // "The ivy destroys %the1!"
   "Bluszcz zniszczył %a1!",
   "Samaşık %a1 yok etti!",
   "Břečťan zničil %a1!",
   "Плющ уничтожил %a1!",
   "Der Efeu zerstört %den1 %a1!",
   "A hera destrói %na1!",
    }},
  {0x3d1cd783, { // "Half-plane model"
   "Model półpłaszczyzny",
/*MISSING*/    "Half-plane model",
   "Polorovinový model",
   "Модель полуплоскости",
/*MISSING*/    "Half-plane model",
/*MISSING*/    "Half-plane model",
    }},
  {0x3d448fe5, { // "This land is warped. Ironically, the coast is completely straight..."
   "Ta kraina jest dziwnie zakrzywiona. Co ciekawe, brzeg jest prosty jak drut...",
/*MISSING*/    "This land is warped. Ironically, the coast is completely straight...",
   "Tento kraj je pokřivený. Ironií je, že pobřeží je absolutně rovné...",
   "Эта земля искривлена. Зато берег абсолютно прямой...",
   "Dieses Land ist verzerrt. Ironischerweise ist die Küste völlig gerade...",
/*MISSING*/    "This land is warped. Ironically, the coast is completely straight...",
    }},
  {0x3d4bdfb2, { // "\nSpawn rate (as prize Orb): %1%/%2\n"
   "\nCzęstość występowania (jako nagroda): %1%/%2\n",
/*MISSING*/    "\nSpawn rate (as prize Orb): %1%/%2\n",
   "\nFrekvence (jako extra Sféra): %1%/%2\n",
   "\nЧастота появления (как награда): %1%/%2\n",
/*MISSING*/    "\nSpawn rate (as prize Orb): %1%/%2\n",
/*MISSING*/    "\nSpawn rate (as prize Orb): %1%/%2\n",
    }},
  {0x3db53675, { // "Some lands unlock at specific treasures or kills"
   "Niektóre krainy otwierają się dzięki konkretnym skarbom lub pokonanym potworom",
   "Bazı diyarlara ulaşabilmek için belirli öldürmeler ve hazineler gerekebilir.",
   "Některé kraje se otevřou získáním určitých pokladů nebo zabitím určitých netvorů.",
   "Для открытия некоторых земель нужно определённые сокровища или убитые враги.",
   "Manche Länder schaltest du durch bestimmte Anzahlen von Schätzen und Kills frei",
/*MISSING*/    "Some lands unlock at specific treasures or kills",
    }},
  {0x3ddd8b37, { // "Shmup (shoot'em up) mode: You can play a hyperbolic shoot'em up game. The game is based on the usual turn-based grid-based HyperRogue, but there are some changes. You fight by throwing knives, and you have three extra lives. There are no allies, so all Orbs related to allies give you extra lives instead (up to 5). Some other rules have been adapted too.\n\n"
   "Strzelanka: grasz w hiperboliczną strzelankę, bazowaną na turowych zasadach HyperRogue, ale z pewnymi zmianami. Twoja broń to rzucane noże, masz 3 dodatkowe życia. Nie ma za to przyjaciół, zatem wszystkie Sfery ich dotyczące dają Ci dodatkowe życia (max 5). Niektóre z pozostałych reguł również zostały dostosowane.\n\n",
/*MISSING*/    "Shmup (shoot'em up) mode: You can play a hyperbolic shoot'em up game. The game is based on the usual turn-based grid-based HyperRogue, but there are some changes. You fight by throwing knives, and you have three extra lives. There are no allies, so all Orbs related to allies give you extra lives instead (up to 5). Some other rules have been adapted too.\n\n",
   "Mód střílečky: zde můžeš hrát hyperbolickou střílečku. Hra je založena na běžném HyperRogue (s koly a políčky), ale jsou tu určité změny. Bojuješ tak, že házíš nože, a máš tři životy navíc. Nemůžeš mít žádné kamarády, takže všechny Sféry, které s nimi souvisejí, ti místo toho přidávají životy navíc (maximum je 5). I některá další pravidla jsou adaptována.\n\n",
   "Режим стрельбы: Вы можете играть в гиперболическую стрелялку, основанную на обычной HyperRogue. Ваше оружие -- метательные ножи, и вы начинаете с 3 дополнительными жизнями. Здесь нет союзников, и все сферы, с ними связанные, просто дают дополнительную жизнь (вплоть до 5). Некоторые другие правила также изменены.\n\n",
   "Shmup-Modus (Ballermodus): Du kannst ein hyperbolisches Ballerspiel spielen. Das Spiel basiert auf dem gewohnten zug- und zellenbasierten HyperRogue, aber mit einigen Änderungen. Du kämpfst indem du Messer wirfst, und hast drei Extraleben. Es gibt keine Freunde, daher geben dir alle Orbs, die Freunde betreffen, stattdessen Extraleben (bis zu fünf). Einige andere Regeln wurden ebenfalls angepasst.\n\n",
/*MISSING*/    "Shmup (shoot'em up) mode: You can play a hyperbolic shoot'em up game. The game is based on the usual turn-based grid-based HyperRogue, but there are some changes. You fight by throwing knives, and you have three extra lives. There are no allies, so all Orbs related to allies give you extra lives instead (up to 5). Some other rules have been adapted too.\n\n",
    }},
  {0x3deb988c, { // "A magical being which copies your movements. You feel that it would be much more useful in an Euclidean space."
   "Magiczna istota kopiująca Twoje ruchy. W przestrzeni euklidesowej byłaby dużo bardziej przydatna.",
   "Hareketlerini taklit eden bir büyülü varlık. Bunların Öklid uzayında daha işe yarayacağını düşünmeden edemiyorsun .",
   "Magická bytost, která kopíruje tvé pohyby. Máš pocit, že v eukleidovském prostoru by byla mnohem užitečnější.",
   "Это магическое создание повторяет Ваши движения. Вы понимаете, что в евклидовом мире это куда полезнее.",
   "Ein magisches Wesen, das deine Bewegungen kopiert.\nDu hast das Gefühl, es wäre in einer euklidischen Umgebung viel nützlicher.",
/*MISSING*/    "A magical being which copies your movements. You feel that it would be much more useful in an Euclidean space.",
    }},
  {0x3e374c5f, { // "Treasure required: %1 x %2.\n"
   "Wymagane skarby: %1 x %2.\n",
   "Gereken hazine: %1 tane %2.\n",
   "Potřebuješ %1 x %2.\n",
   "Сокровищ собрано: %1 x %2.\n",
   "Benötigte Schätze: %1 x %2.\n",
/*MISSING*/    "Treasure required: %1 x %2.\n",
    }},
  {0x3ee05d1b, { // "The rug depicts a man in a deep dungeon, unable to leave."
   "Dywan przedstawia człowieka w głębokich lochach, niezdolnego do ucieczki.",
/*MISSING*/    "The rug depicts a man in a deep dungeon, unable to leave.",
   "Na tomto koberci je vyobrazen muž v hlubokém žaláři, který nemůže ven.",
   "Ковёр изображает человека из глубокого подземелья, который не может сбежать.",
/*MISSING*/    "The rug depicts a man in a deep dungeon, unable to leave.",
/*MISSING*/    "The rug depicts a man in a deep dungeon, unable to leave.",
    }},
  {0x3eeaab06, { // "\"I have been trained to fight with a Hypersian scimitar, you know?\""
   "\"Trenowa%łem1 walkę hiperskim szamszirem, wiesz?\"",
   "Hiç Aşkınlı Palasıyla savaşmak üzere eğitilmemiştim biliyor musun?\"",
   "Jsem vycvičen%ý1 v boji s hyperskou šavlí, víš?\"",
   "Я тренировал%c1 с гиперсидским ятаганом, ты в курсе?\"",
   "Ich habe trainiert um mit einem Hypersischem Scimitar zu kämpfen, wusstest du das?\"",
/*MISSING*/    "\"I have been trained to fight with a Hypersian scimitar, you know?\"",
    }},
  {0x3f053b67, { // "%The1 picks up %the2!"
   "%1 podniosła %a2!",
   "%1 %a2 kaldırdı!",
   "%1 sebrala %a2!",
   "%1 поднял%E1 %a2!",
   "Die %1 hebt %den2 %a2 auf!",
/*MISSING*/    "%The1 picks up %the2!",
    }},
  {0x3f17bc9e, { // "Orb unlocked: %1"
   "Odblokowana Sfera: %1",
/*MISSING*/    "Orb unlocked: %1",
   "Byla odemèena Sféra: %1",
   "Открыта сфера: %1",
/*MISSING*/    "Orb unlocked: %1",
/*MISSING*/    "Orb unlocked: %1",
    }},
  {0x3f3ffe3e, { // "%The1 raises some undead!"
   "%1 ożywia nieumarłych!",
   "%1 ölü çağırdı!",
   "%1 probudil nemrtvé!",
   "%1 восстал из мёртвых!",
   "%Der1 %1 beschwört die Toten!",
   "%na1 levanta os mortos!",
    }},
  {0x3f4727cc, { // "You cannot move between the cells without dots here!"
   "Nie możesz poruszać się między polami bez kropek!",
/*MISSING*/    "You cannot move between the cells without dots here!",
   "Nie możesz tu przechodzić między polami bez kropek!",
   "Вы не можете ходить между полями без точек!",
   "Du kannst dich hier nicht zwischen Zellen ohne Punkte bewegen!",
/*MISSING*/    "You cannot move between the cells without dots here!",
    }},
  {0x3f72ebb2, { // "hyperbolic puzzles"
   "hiperboliczne zagadki",
/*MISSING*/    "hyperbolic puzzles",
   "hyperbolické hádanky",
   "гиперболические паззлы",
/*MISSING*/    "hyperbolic puzzles",
/*MISSING*/    "hyperbolic puzzles",
    }},
  {0x4002fc3c, { // "\"Bring me out of here please!\""
   "\"Proszę, zabierz mnie stąd!\"",
   "Beni buradan çıkar lütfen!\"",
   "Prosím tě, odveď mě odsud!\"",
   "Прошу, уведи меня отсюда!\"",
   "Bitte bring mich hier raus!\"",
/*MISSING*/    "\"Bring me out of here please!\"",
    }},
  {0x403512ff, { // "Orb Strategy mode"
   "tryb strategii sfer",
/*MISSING*/    "Orb Strategy mode",
   "sférostrategický mód",
   "режим стратегии сфер",
/*MISSING*/    "Orb Strategy mode",
/*MISSING*/    "Orb Strategy mode",
    }},
  {0x40656952, { // "\"Be careful in the Rose Garden! It is beautiful, but very dangerous!\""
   "\"Uważaj w Różanym Ogrodzie! Jest tam pięknie, ale bardzo niebezpiecznie!\"",
/*MISSING*/    "\"Be careful in the Rose Garden! It is beautiful, but very dangerous!\"",
   "\"V Růžové zahradě si dávej pozor! Je pěkná, ale také pěkně nebezpečná!\"",
   "\"Будьте осторожны в Розарии! Там красиво, но очень опасно!\"",
   "\"Sei vorsichtig im Rosengarten! Er ist schön, aber sehr gefährlich!\"",
/*MISSING*/    "\"Be careful in the Rose Garden! It is beautiful, but very dangerous!\"",
    }},
  {0x408f8ba3, { // "New Achievement:"
   "Nowe osiągnięcie:",
   "Yeni Başarım:",
   "Nový achievement:",
   "Новое достижение:",
   "Neues Achievement:",
   "Nova conquista:",
    }},
  {0x40a4d8a5, { // "a prized Orb, it appears only in cabinets"
   "cenna Sfera, tylko w gablotach",
   "çok kıymetli bir Küre, sadece vitrinde bulunur.",
   "cenná Sféra, která se objevuje pouze ve vitrínách",
   "ценная сфера, только в кабинетах",
   "Ein wertvoller Orb der nur in Vitrinen erscheint",
/*MISSING*/    "a prized Orb, it appears only in cabinets",
    }},
  {0x416cad19, { // "summon Thumpers"
   "przywołaj dudniki",
   "Gümleyen çıkar",
   "vyvolej Tlouky",
   "призвать Тамперы",
   "rufe Stampfer herbei",
/*MISSING*/    "summon Thumpers",
    }},
  {0x416f36a1, { // "special game modes"
   "specjalne tryby gry",
   "özel oyun modları",
   "speciální herní módy",
   "специальные режимы",
   "Spielmodi",
/*MISSING*/    "special game modes",
    }},
  {0x41877005, { // "review the scene"
   "przegląd sceny",
   "durumu gör",
   "přehled scény",
   "смотреть сцену",
   "Die Szene anzeigen",
/*MISSING*/    "review the scene",
    }},
  {0x419dd9a4, { // "The tentacles of Cthulhu are like sandworms, but longer. They also withdraw one cell at a time, instead of exploding instantly."
   "Macki Cthulhu są podobne do Pustynnych Czerwi, ale dłuższe. W przeciwieństwie do nich nie eksplodują po zablokowaniu, tylko cofają się pod ziemię, po 1 polu na kolejkę.",
   "Cthulhu'nun dokunaçları kumkurtları gibidir, ama daha uzundur. Ve birden patlamak yerine her seferinde bir hücregeri çekilirler.",
   "Cthulhuova chapadla jsou něco jako Píseční červi, ale delší. Kromě toho neexplodují okamžitě, ale ustupují políčko za políčkem.",
   "Щупальца Ктулху походи на песчаных червей, но длиннее. Они втягиваются по одной клетке, а не взрываются мгновенно.",
   "Die Tentakel von Cthulhu sind wie Sandwürmer, aber länger. Anstatt zu explodieren ziehen sie sich zurück, eine Zelle nach der anderen.",
   "Os tentáculos de Ctulhu são como minhocas de areia, porém maiores. Eles também recolhem-se uma célula por vez, ao invés de explodir instantaneamente.",
    }},
  {0x41a0800e, { // " $$$: %1"
   " $$$: %1",
   " $$$: %1",
   " $$$: %1",
   " $$$: %1",
   " $$$: %1",
/*MISSING*/    " $$$: %1",
    }},
  {0x41aeaa9f, { // "BACK"
   "OK",
   "GERİ",
   "ZPĚT",
   "BACK",
   "RÜCK",
/*MISSING*/    "BACK",
    }},
  {0x41b07440, { // "A fake Land with colored floors."
   "Fałszywa Kraina z kolorowymi podłogami.",
   "Renkli zeminli bir sahte diyar.",
   "Falešný kraj s barevnou podlahou.",
   "Фиктивная земля с разноцветным полом.",
   "Ein falsches Land mit farbigen Böden.",
/*MISSING*/    "A fake Land with colored floors.",
    }},
  {0x41ef5d8c, { // "You can hold this bush to climb the Lost Mountain, but it is not very strong -- it will get destroyed if you climb from it into an unstable location. Bushes block the movement of birds."
   "Możesz złapać ten krzak, by się wspiąć na Zagubioną Górę, ale nie jest on zbyt wytrzymały -- zniszczy się, gdy wespniesz się z niego do niestabilnej pozycji. Krzaki blokują drogę ptakom.",
/*MISSING*/    "You can hold this bush to climb the Lost Mountain, but it is not very strong -- it will get destroyed if you climb from it into an unstable location. Bushes block the movement of birds.",
   "Tohoto keře se můžeš chytit při výstupu na Ztracenou horu, ale není příliš silný -- pokud z něj vylezeš na nestabilní lokaci, zničí se. Keře blokují pohyb ptáků.",
   "Вы можете хвататься за этот куст, чтобы лезть на гору, но он не очень прочный: он развалится, если переместиться с него на нестабильную клетку. Блокирует перемещение птиц.",
/*MISSING*/    "You can hold this bush to climb the Lost Mountain, but it is not very strong -- it will get destroyed if you climb from it into an unstable location. Bushes block the movement of birds.",
/*MISSING*/    "You can hold this bush to climb the Lost Mountain, but it is not very strong -- it will get destroyed if you climb from it into an unstable location. Bushes block the movement of birds.",
    }},
  {0x4212ec3c, { // "c = readd v"
   "c = ponownie v",
   "c = yeniden ekle v",
   "c = znovu přidej v",
   "c = обновить v",
   "c = wiederhinzufügen v",
/*MISSING*/    "c = readd v",
    }},
  {0x426c48b7, { // "The Alchemists produce magical potions from pools of blue and red slime. You can go through these pools, but you cannot move from a blue pool to a red pool, or vice versa. Pools containing items count as colorless, and they change color to the PC's previous color when the item is picked up. Slime beasts also have to keep to their own color, but when they are killed, they explode, destroying items and changing the color of the slime and slime beasts around them."
   "Alchemicy produkują magiczne napoje z niebieskiej i czerwonej mazi. Możesz poruszać się poprzez maź, ale nie możesz przesunąć się z pola niebieskiego na czerwone, ani z powrotem. Pola zawierające przedmioty są bezbarwne, po zebraniu przedmiotu zmieniają kolor na kolor pola, na którym gracz był wcześniej. Maziste Stwory również ograniczone są do swojego koloru. Zabijane eksplodują, niszcząc przedmioty i zmieniając kolor mazi wokół nich.\n",
   "Simyacılar mavi ve kırmızı balçık havuzlarından sihirli iksirler üretirler. Bu havuzlardan geçebilirsin, ama kırmızı bir havuzdan mavi bir havuza veya tam tersine geçemezsin. Balçık canavarları kendi renklerinde durmak zorundadır, Ama öldüklerinde patlarlar ve etrafındaki eşyaları yok eder, etraflarındaki balçığın ve canavarların rengini değiştirirler. \n",
   "Alchymisté vytvářejí magické lektvary z kaluží modrého a červeného slizu. Ty můžeš těmito kalužemi procházet, ale nemůžeš se pohnout z modré kaluže do červené nebo obráceně. Kaluže s předměty se berou jako bezbarvé a po sebrání předmětu změní barvu na tu, na které jste byl předtím. I Slizoví netvoři se mohou pohybovat jen v kalužích své vlastní barvy, ale když zemřou, vybuchnou, což zničí předměty v jejich blízkosti a změní barvu okolního slizu a Slizových netvorů.",
   "Алхимики создают магические зелья из красной и синей слизи. Вы можете перемещаться по этой слизи, но не можете перейти с красной на синюю и наоборот. Слизь, содержащая предметы, бесцветна и меняет свой цвет на Ваш, когда Вы собираете предмет. Живая слизь тоже двигается лишь по своему цвету, но при гибели меняет цвет всей слизи в радиусе 2 клеток на свой.\n",
   "Die Alchemisten produzieren magische Tränke aus Lachen von blauem und rotem Schleim. Du kannst durch sie hindurchgehen, aber du kannst nicht von Blau zu Rot wechseln oder umgekehrt. Lachen mit Gegenständen darin sind farblos und bekommen die Farbe, auf der du zuletzt warst, wenn du den Gegenstand aufhebst. Schleimmonster müssen auch auf ihrer Farbe bleiben. Wenn sie sterben, explodieren sie, zerstören Items und ändern die Farbe von Schleim und Schleimmonstern um sie herum.",
/*MISSING*/    "The Alchemists produce magical potions from pools of blue and red slime. You can go through these pools, but you cannot move from a blue pool to a red pool, or vice versa. Pools containing items count as colorless, and they change color to the PC's previous color when the item is picked up. Slime beasts also have to keep to their own color, but when they are killed, they explode, destroying items and changing the color of the slime and slime beasts around them.",
    }},
  {0x42a2a741, { // "first page [Space]"
   "pierwsza strona [Space]",
   "ilk sayfa [Space]",
   "první stránka [mezerník]",
   "первая страница [пробел]",
   "Erste Seite [LEER]",
   "primeira página [Space]",
    }},
  {0x42f318bd, { // "anti-aliasing"
   "anti-aliasing",
/*MISSING*/    "anti-aliasing",
   "anti-aliasing",
   "anti-aliasing",
/*MISSING*/    "anti-aliasing",
/*MISSING*/    "anti-aliasing",
    }},
  {0x4321d329, { // "move down"
   "idź do dołu",
   "aşağı git",
   "dolů",
   "идти вниз",
   "Runter bewegen",
/*MISSING*/    "move down",
    }},
  {0x4380f6c5, { // " kills: %1/%2"
   " zabicia: %1/%2",
/*MISSING*/    " kills: %1/%2",
   " zabití: %1/%2",
   " убийств: %1/%2",
   " Tötungen: %1/%2",
/*MISSING*/    " kills: %1/%2",
    }},
  {0x4392dd43, { // "(ESC) tour menu"
   "(ESC) menu ucznia",
/*MISSING*/    "(ESC) tour menu",
   "(ESC) menu prohlídky",
   "(ESC) меню обучения",
/*MISSING*/    "(ESC) tour menu",
/*MISSING*/    "(ESC) tour menu",
    }},
  {0x43bec718, { // "four players"
   "czterech graczy",
/*MISSING*/    "four players",
   "čtyři hráči",
   "четыре игрока",
   "vier Spieler",
/*MISSING*/    "four players",
    }},
  {0x43e793b1, { // "There are many possible solutions to the Yendor Quest. In the Yendor Challenge, you will try many of them!\n\nEach challenge takes part in a specific land, and you have to use what you have available.\n\nYou need to obtain an Orb of Yendor in the normal game to activate this challenge, and (ever) collect 10 treasures in one or two lands to activate a specific level.\n\nAfter you complete each challenge, you can try it again, on a harder difficulty level.\n\nAll the solutions showcased in the Yendor Challenge work in the normal play too. However, passages to other lands, and (sometimes) some land features are disabled in the Yendor Challenge, so that you have to use the expected method. Also, the generation rules are changed slightly for the Palace and Minefield while you are looking for the Orb of Yendor, to make the challenge more balanced (but these changes are also active during the normal Yendor Quest).\n\nYou get 1000 points for each challenge won, and 1 extra point for each extra difficulty level."
   "Jest dużo możliwych sposobów, by zdobyć Sferę Yendoru. W tej misji spróbujesz wielu z nich!\n\nKażda misja dzieje się w konkretnej krainie i musisz użyć tego, co jest dostepne.\n\nMusisz zdobyć Sferę Yendoru w normalnej grze, by odblokować Misję Yendor i (kiedykolwiek) zdobyć 10 skarbów w jednym czy dwu światach, by uaktywnić konkretną misję.\n\nPo ukończeniu każdej misji, możesz spróbować jescze raz, na trudniejszym poziomie.\n\nWszystkie rozwiązania użyte w Misji Yendor działają w normalnej grze. Przejścia do innych krain i (czasami) niektóre inne elementy są wyłączone podczas Misji Yendor, byś musia%ł0 użyć oczekiwanej metody. Reguły generowania krain są lekko zmienione, gdy szukasz Sfer Yendoru w Pałacu i na Polu Minowym, by poprawić równowagę (te zmiany są też aktywne przy poszukiwaniu Sfery Yendoru podczas normalnej gry).\n\n",
   "Yendor Görevinin bir sürü çözümü var. Yendor Ek Görevi'nde bunların pek çoğunu deneyeceksin!\n\nHer görev belirli bir diyarda geçiyor ve elindekileri bütün verimliliğiyle kullanman gerekecek.\n\nBu görevi aktifleştirmek için normal oyunda bir kez Yendor'un Küresi'ni ele geçirmen gerekiyor. Ayrıca her hangi bir seviyeyi aktifleştirmek için ilgili diyar(lar)da 10 hazine toplamış olman gerekiyor.\n\nEk görevi tamamladıktan sonra aynı görevi bir üst zorluk seviyesinde yeniden deneyebilirsin.\n\nYendor Ek Görevi'ndeki bütün çözümler normal oyunda da işe yarar. Ancak, Yendor Ek görevinde, diğer diyarlara geçiş ve bazen diyarın bazı özellikleri devredışıdır, bu yüzden beklenen çözümü bulman beklenir. Ayrıca,  Yendor'un Küresi'ni ararken görevin daha dengeli olması için Saray ve Mayın Tarlasındaki harita üretimi biraz değiştirilmiştir. (bu değişiklikler normal Yendor Görevi'nde de mevcuttur).\n\nHer kazanılan ek görev için 1000 puan ve her fazladan zorluk seviyesi için 1 fazladan puan kazanırsın.\n\n",
   "Je mnoho různých způsobů, jak získat Yendorskou Sféru. V Misi: Yendor si jich vyzkoušíš celou řadu!\n\nKaždá mise se odehrává v určitém kraji a ty musíš využít to, co tam máš k dispozici.\n\nAbys mohl hrát Misi: Yendor, musíš získat Yendorskou Sféru v normální hře; pro aktivaci konkrétních misí musíš také (alespoň jednou) získat 10 pokladů v jednom či dvou krajích.\n\nPo splnění každé mise se o ni můžeš pokusit znovu, na vyšší obtížnosti.\n\nVšechna řešení v Misi: Yendor fungují i v normální hře. Průchody do jiných krajin a (někdy) některé jejich prvky jsou však v Misi: Yendor vypnuty, takže musíte použít metodu, jaká se od vás očekává. Kromě toho se při hledání Yendorské sféry poněkud mění pravidla pro generování Paláce a Minového pole, aby byla mise více vyvážená, (tyto změny však platí i během normálního hledání Yendorské Sféry).\n\n",
   "Есть много способов достать сферу Йендора. В этой миссии Вы можете попробовать все варианты!\n\nКаждая миссия проходит в конкретных землях, и можно использовать лишь то, что там есть.\n\nВы должны достать сферу Йендора в обычной игре, чтобы открыть этот режим, и собрать 10 сокровищ в одной или двух землях, чтобы открыть конкретные уровни.\n\nЗакончив определённую миссию, Вы можете попробовать её заново на большей сложности.\n\nВсе решения, продемонстрированные в миссии Йендора, работают и в нормальной игре. Выходы в другие земли и некоторые внутренние особенности здесь отключены, что заставляет использовать предусмотренный способ. Правила генерации во Дворце и на Минном поле немного изменены, пока Вы ищете Сферу, но в нормальной игре эти изменения во время квеста тоже действуют.\n\n",
   "Es gibt viele mögliche Lösungen für die Quest von Yendor. Während der Yendor Herausforderung wirst du viele davon ausprobieren!\n\nJede Herausforderung ist auf ein bestimmtes Land begrenzt und du musst nutzen was gegeben ist.\n\nZum Freischalten dieser Herausforderung muss ein Orb von Yendor im normalem Modus gesammelt werden sowie 10 Schätze in ein oder zwei Ländern um bestimmte Level zu aktivieren.\n\nNach dem Beenden einer Herausforderung kannst du eine schwierigere Version dieser spielen.\n\nAlle Lösungen der Yendor-Herausforderung funktionieren auch im normalem Spiel. Allerdings sind Wege zu anderen Ländern in der Yendor-Herausforderung deaktiviert, somit bleiben weniger Lösungsmöglichkeiten. Außerdem kann verändert werden wie der Palast und das Minenfeld generiert werden um es einfacher zu machen einen Orb von Yendor zu finden. (Diese Änderungen sind auch während des normalen Yendorquest aktiviert).\n\n",
/*MISSING*/    "There are many possible solutions to the Yendor Quest. In the Yendor Challenge, you will try many of them!\n\nEach challenge takes part in a specific land, and you have to use what you have available.\n\nYou need to obtain an Orb of Yendor in the normal game to activate this challenge, and (ever) collect 10 treasures in one or two lands to activate a specific level.\n\nAfter you complete each challenge, you can try it again, on a harder difficulty level.\n\nAll the solutions showcased in the Yendor Challenge work in the normal play too. However, passages to other lands, and (sometimes) some land features are disabled in the Yendor Challenge, so that you have to use the expected method. Also, the generation rules are changed slightly for the Palace and Minefield while you are looking for the Orb of Yendor, to make the challenge more balanced (but these changes are also active during the normal Yendor Quest).\n\nYou get 1000 points for each challenge won, and 1 extra point for each extra difficulty level.",
    }},
  {0x44057f41, { // "In most lands, more treasures are generated with each enemy native to this land you kill. Moreover, 100 kills is a requirement to enter the Graveyard and the Hive.\n\nFriendly creatures and parts of monsters (such as the Ivy) do appear in the list, but are not counted in the total kill count."
   "W większości krain, zabijanie przeciwników pochodzących z danej krainy powoduje, że generowane jest więcej skarbów. Ponadto, 100 zabić jest konieczne, by móc znaleźć Cmentarz i Mrowisko.\n\nPrzyjazne stworzenia i części potworów (jak np. Bluszcz) występują na liście, ale nie są liczone do łącznej liczby zabić.",
   "Çoğu diyarda, orada yaşayan canavarları öldürmen daha çok hazineyle karşılaşmanı sağlar. Ayrıca 100 leş Mezarlık ve Kovana girmen için bir önkoşuldur. \n\nDost yaratıklar ve canavarların parçaları (Sarmaşık gibi) litede gözükürt, ama toplam leşlere dahil edilmez.",
   "Ve většině krajů se objevuje tím víc pokladů, čím víc zabijete nepřátel, kteří v nich žijí. Kromě toho musíš zabít 100 netvorů, aby ses mohl dostat na Hřbitov a do Úlu.\n\nPřátelští tvorové a části netvorů (jako je Břečťan) se v seznamu objevují, ale do celkového počtu zabitých netvorů se nepočítají.",
   "В большинстве земель чем больше Вы убиваете местных жителей, тем больше сокровищ создаётся. Кроме того, Вам нужно убить 100 врагов, чтобы попасть на Кладбище и в Улей.\n\nДружественные существа и части монстров (например, части плюща) при этом на считаются.",
   "In den meisten Ländern werden mehr Schätze generiert, je mehr Feinde du darin tötest. Darüber hinaus benötigst du 100 Kills für den Friedhof und das Nest.\n\nFreundliche Kreaturen und Monsterteile (z.B. Efeu) werden zwar gezeigt, zählen aber nicht zur Gesamt-Punktzahl.",
/*MISSING*/    "In most lands, more treasures are generated with each enemy native to this land you kill. Moreover, 100 kills is a requirement to enter the Graveyard and the Hive.\n\nFriendly creatures and parts of monsters (such as the Ivy) do appear in the list, but are not counted in the total kill count.",
    }},
  {0x443579c7, { // "\"When I visited the Palace, a mouse wanted me to go somewhere.\""
   "\"Kiedy byłem w Pałacu, myszka chciała, bym gdzieś poszedł.\"",
   "\"Sarayı ziyaret ettiğimde, bir fare benden bir yere gitmemi istedi.\"",
   "\"Když jsem navštívil Palác, byla tam myška, co chtěla, abych se někam vydal.\"",
   "\"Когда я был во Дворце, мышь пыталась меня позвать куда-то.\"",
   "\"Als ich den Palast besuchte wollte eine Maus das ich irgendwo hingehe.\"",
/*MISSING*/    "\"When I visited the Palace, a mouse wanted me to go somewhere.\"",
    }},
  {0x44758caa, { // "Rangers take care of the magic mirrors in the Land of Mirrors. They know that rogues like to break these mirrors... so they will attack you!"
   "Strażnicy chronią lustra w Krainie Luster. Wiedzą, że złodzieje lubią rozbijać lustra... także spodziewaj się ataku!",
   "Seyyahların aynalarını kıracağını bildiklerinden sana saldıracaklar!",
   "Strážci se starají o kouzelná zrcadla v Kraji zrcadel. Vědí, že lupiči jako ty rádi tato zrcadla rozbíjejí... a proto na tebe útočí!",
   "Странники охраняют магические зеркала. Они знают, что эти разбойники хотят разбить их зеркала и тут же нападают на Вас!",
   "Ranger beschützen die Zauberspiegel im Spiegelland. Sie wissen, dass Schurken wie du die Spiegel gerne zerbrechen - also greifen sie dich an!",
/*MISSING*/    "Rangers take care of the magic mirrors in the Land of Mirrors. They know that rogues like to break these mirrors... so they will attack you!",
    }},
  {0x44f3a21d, { // "field pattern C"
   "wzór pola C",
/*MISSING*/    "field pattern C",
   "vzor pole C",
   "узор поля C",
/*MISSING*/    "field pattern C",
/*MISSING*/    "field pattern C",
    }},
  {0x44f3a21e, { // "field pattern D"
   "wzór pola D",
/*MISSING*/    "field pattern D",
   "vzor pole D",
   "узор поля D",
/*MISSING*/    "field pattern D",
/*MISSING*/    "field pattern D",
    }},
  {0x44f3a228, { // "field pattern N"
   "wzór pola N",
/*MISSING*/    "field pattern N",
   "vzor pole N",
   "узор поля N",
/*MISSING*/    "field pattern N",
/*MISSING*/    "field pattern N",
    }},
  {0x44f3a22d, { // "field pattern S"
   "wzór pola S",
/*MISSING*/    "field pattern S",
   "vzor pole S",
   "узор поля S",
/*MISSING*/    "field pattern S",
/*MISSING*/    "field pattern S",
    }},
  {0x450c904d, { // "\"It is not possession but the act of getting there, "
   "\"Nie posiadanie, ale proces docieranie na miejsce, ",
/*MISSING*/    "\"It is not possession but the act of getting there, ",
   "\"Ne vlastnictví, ale cesta k němu ",
   "\"Это не владение, но процесс получения знания",
/*MISSING*/    "\"It is not possession but the act of getting there, ",
/*MISSING*/    "\"It is not possession but the act of getting there, ",
    }},
  {0x451f3f10, { // "advanced configuration"
   "zaawansowana konfiguracja",
   "gelişmiş ayarlar",
   "pokročilá konfigurace",
   "продвинутые настройки",
   "Erweiterte Einstellungen",
/*MISSING*/    "advanced configuration",
    }},
  {0x45293820, { // "prepare the line animation"
   "przygotuj animację linii",
/*MISSING*/    "prepare the line animation",
   "připrav přímkovou animaci",
   "приготовить анимацию линии",
   "bereite die Linien-Animation vor",
/*MISSING*/    "prepare the line animation",
    }},
  {0x45350617, { // "Welcome to the HyperRogue tutorial!"
   "Witaj w podręczniku HyperRogue!",
/*MISSING*/    "Welcome to the HyperRogue tutorial!",
   "Vítej v tutoriálu HyperRogue!",
   "Добро пожаловать в руководство HyperRogue!",
/*MISSING*/    "Welcome to the HyperRogue tutorial!",
/*MISSING*/    "Welcome to the HyperRogue tutorial!",
    }},
  {0x453c3739, { // "You got Thorns! Stab monsters by moving around them."
   "Ciernie! Dźgaj potwory chodząc obok nich.",
/*MISSING*/    "You got Thorns! Stab monsters by moving around them.",
   "Získal jsi Trny! Nyní můžeš probodnout netvory tím, že kolem nich projdeš.",
   "Вы получили Шипы! Колите монстров, ходя мимо них.",
/*MISSING*/    "You got Thorns! Stab monsters by moving around them.",
/*MISSING*/    "You got Thorns! Stab monsters by moving around them.",
    }},
  {0x457f1068, { // "Your total treasure has been recorded in the Steam Leaderboards."
   "Twój wynik został zapisany w rankingach Steam.",
   "Toplam hazinen Steam Listelerine kaydedildi.",
   "Tvůj celkový poklad byl zaznamenán na žebříčky na Steamu.",
   "Ваш общий результат занесён в таблицу Steam.",
   "Deine Gesamtschätze wurden in die Steam-Bestenlisten aufgenommen.",
   "Seu total de tesouros foi gravado na Leaderboard da Steam.",
    }},
  {0x461c6331, { // "default"
   "domyślne",
/*MISSING*/    "default",
   "default",
   "по умолчанию",
/*MISSING*/    "default",
/*MISSING*/    "default",
    }},
  {0x461e7611, { // "You cannot attack through the Vine!"
   "Nie możesz atakować poprzez winorośl!",
   "Şarabın içinden saldıramazsın!",
   "Nemůžeš útočit skrz révu!",
   "Вы не можете атаковать сквозь лозу!",
   "Du kannst nicht durch die Reben hindurch angreifen!",
/*MISSING*/    "You cannot attack through the Vine!",
    }},
  {0x463368ad, { // "\"The Holy Grail is in the center of the Round Table.\""
   "\"Święty Graal jest pośrodku Okrągłego Stołu.\"",
   "\"Kutsal Kâse Yuvarlak Masanın merkezinde.\"",
   "\"Svatý grál je ve středu Kulatého stolu.\"",
   "\"Святой Грааль в центре круглого стола.\"",
   "\"Der Heilige Gral ist im Zentrum der Tafelrunde.\"",
/*MISSING*/    "\"The Holy Grail is in the center of the Round Table.\"",
    }},
  {0x4634453e, { // "You can summon these friendly constructs with a magical process."
   "Przyjazna konstrukcja, tworzona poprzez magiczny proces.",
   "Bu dost mekanizmaları büyüsel bir işlemle uyandırabilirsin.",
   "Tyto přátelské konstrukty můžeš vyvolat pomocí magického procesu.",
   "Магия позволяет призывать этих дружелюбных созданий.",
   "Du kannst diese freundlichen Steinriesen mit einem Ritual beschwören.",
/*MISSING*/    "You can summon these friendly constructs with a magical process.",
    }},
  {0x4696bd7f, { // "start a new game"
   "nowa gra",
/*MISSING*/    "start a new game",
   "nová hra",
   "новая игра",
   "neues Spiel starten",
/*MISSING*/    "start a new game",
    }},
  {0x46bbc887, { // "This false princess is immune to the alluring scent of roses."
   "Ta fałszywa księżniczka jest niewrażliwa na zapach róż.",
/*MISSING*/    "This false princess is immune to the alluring scent of roses.",
   "Tato falešná princezna je imunní vůči vábivé vůni růží.",
   "Эта ложная принцесса невосприимчива к ароматам роз.",
   "Diese falsche Prinzessin ist immun gegenüber dem verlockenden Duft der Rosen.",
/*MISSING*/    "This false princess is immune to the alluring scent of roses.",
    }},
  {0x46bd9565, { // "%The1 breathes fire at %the2!"
   "%1 zieje ogniem w %a2!",
/*MISSING*/    "%The1 breathes fire at %the2!",
   "%1 chrlí oheň na %a2!",
   "%1 дышит огнём на %a2!",
   "%Der1 %1 spuckt Feuer auf %den2 %a2!",
/*MISSING*/    "%The1 breathes fire at %the2!",
    }},
  {0x471c20ca, { // "Collecting 25 treasures in a given land may be dangerous, but allows magical Orbs of this land to appear in other places!"
   "Zebranie 25 skarbów w jednej krainie może być niebezpieczne, ale powoduje, że magiczne Sfery z tej krainy pojawiają się w pozostałych krainach!",
/*MISSING*/    "Collecting 25 treasures in a given land may be dangerous, but allows magical Orbs of this land to appear in other places!",
   "Získat 25 pokladù v jednom kraji mùže být nebezpeèné, ale zpùsobí to, že se magické Sféry z tohoto kraje zaènou objevovat i v jiných!",
   "Собирать 25 в одной земле довольно опасно, но это позволяет Сферам из этой земли появляться в других местах!",
/*MISSING*/    "Collecting 25 treasures in a given land may be dangerous, but allows magical Orbs of this land to appear in other places!",
/*MISSING*/    "Collecting 25 treasures in a given land may be dangerous, but allows magical Orbs of this land to appear in other places!",
    }},
  {0x47a46d50, { // "treasure collected"
   "zdobyte skarby",
   "toplanan hazineler",
   "získané poklady",
   "собрано сокровищ",
   "Gesammelte Schätze",
   "tesouro coletado",
    }},
  {0x47bcb239, { // "summon dead orbs"
   "przywołaj martwe sfery",
   "ölü küre çıkar",
   "vyvolej mrtvé sféry",
   "призвать мёртвые сферы",
   "rufe tote Orbs herbei",
/*MISSING*/    "summon dead orbs",
    }},
  {0x48142822, { // "This has been recorded in the Steam Leaderboards."
   "To zostało zapisane w Rankingach Steam.",
   "Bu Steam Listelerine kaydedildi.",
   "Bylo zaznamenáno na žebříčky Steamu.",
   "Это записано в таблицу рекордов Steam.",
   "Dies wurde in den Steam-Bestenlisten vermerkt.",
   "Isso foi salvona Leaderboard da Steam.",
    }},
  {0x483986df, { // "A land filled with huge ivy plants and dangerous animals."
   "Kraina pełna wielkiego pnącego bluszczu i niebezpiecznych bestii.",
   "Sarmaşık bitkileri ve tehlikeli hayvanlarla dolu bir diyar.",
   "Kraj plný obrovských rostlin břečťanu a nebezpečných zvířat.",
   "Земля, полная огромными плющами и опасными животными.",
   "Ein Land voller riesiger Efeupflanzen und gefährlicher Tiere.",
/*MISSING*/    "A land filled with huge ivy plants and dangerous animals.",
    }},
  {0x4895a247, { // "Unavailable in the multiplayer mode.\n"
   "Niedostępne w trybie wielu graczy.\n",
/*MISSING*/    "Unavailable in the multiplayer mode.\n",
   "Nedostupné v módu pro více hráčů.\n",
   "Недоступно в режиме нескольких игроков.\n",
   "Nicht im Mehrspieler-Modus verfügbar.\n",
/*MISSING*/    "Unavailable in the multiplayer mode.\n",
    }},
  {0x489cc548, { // "Periodic patterns"
   "Wzory okresowe",
/*MISSING*/    "Periodic patterns",
   "Periodické vzory",
   "Периодические узоры",
/*MISSING*/    "Periodic patterns",
/*MISSING*/    "Periodic patterns",
    }},
  {0x489faba1, { // "Princess"
   "Księżniczka",
/*MISSING*/    "Princess",
   "Princezna",
   "Принцесса",
   "Prinzessin",
/*MISSING*/    "Princess",
    }},
  {0x48aeb10a, { // "%The1 breaks the mirror!"
   "%1 rozbi%ł1 lustro!",
   "%1 aynayı kırdı!",
   "%1 rozbi%l1 zrcadlo!",
   "%1 разбил%E1 зеркало!",
   "%Der1 %1 zerstört den Spiegel!",
   "%na1 quebra o espelho!",
    }},
  {0x48e37d1a, { // "Wow! %1! This trip should be worth it!"
   "%1! Ta wyprawa powinna się opłacić!",
   "Vay canına! %1! Bu seyahati yaptığına değecek!",
   "Páni! %1! Tenhle výlet by měl stát za to!",
   "%1! А это путешествие стоит того!",
   "Wow, %1! Diese Reise lohnt sich!",
   "Uau! %1! Essa viajem deve compensar!",
    }},
  {0x48f7b4eb, { // "special display modes"
   "specjalne tryby ekranu",
/*MISSING*/    "special display modes",
   "zvláštní zobrazovací módy",
   "специальные режимы отображения",
/*MISSING*/    "special display modes",
/*MISSING*/    "special display modes",
    }},
  {0x4976c551, { // "You drop %the1."
   "Porzuci%łeś0 %a1.",
   "%a1 düşürdün.",
   "Položi%l0 jsi %a1.",
   "Вы выкинули %a1.",
   "Du legst %den1 %a1 ab.",
   "Você derruba %na1.",
    }},
  {0x499be79c, { // "by land"
   "kraina",
/*MISSING*/    "by land",
   "podle kraje",
   "по земле",
/*MISSING*/    "by land",
/*MISSING*/    "by land",
    }},
  {0x49a3ff8e, { // "summon treasure"
   "przywołaj skarby",
   "hazine çıkar",
   "vyvolej poklad",
   "призвать сокровища",
   "rufe Schätze herbei",
/*MISSING*/    "summon treasure",
    }},
  {0x4a1c9024, { // "ESC for menu/quest"
   "ESC - menu/misja",
   "ESC - menü/görev",
   "ESC - menu/úkoly",
   "ESC - меню/миссия",
   "ESC für Menü/Quest",
   "ESC para menu/missão",
    }},
  {0x4ab7c698, { // "turn-based mode"
   "gra turowa",
/*MISSING*/    "turn-based mode",
   "hra na kola",
   "пошаговый режим",
   "zugbasierter Modus",
/*MISSING*/    "turn-based mode",
    }},
  {0x4af3c63d, { // "This Orb makes you stunningly beautiful. Monsters which come next to you will be stunned for one turn. Multi-tile monsters are not affected. Additionally, it makes you immune to beauty."
   "Dzięki tej Sferze stajesz się oszałamiająco piękn%y0. Stworzenia, które wejdą na pole obok Ciebie, zostaną oszołomione na 1 kolejkę. Nie działa to na stworzenia zajmujące wiele pól. Dodatkowo, stajesz się niewrażliw%y0 na piękno.",
/*MISSING*/    "This Orb makes you stunningly beautiful. Monsters which come next to you will be stunned for one turn. Multi-tile monsters are not affected. Additionally, it makes you immune to beauty.",
   "Tato Sféra ti dává omračující krásu. Netvoři, kteří vstoupí na políčko vedle tebe, budou na jedno kolo omráčeni. Netvoří, kteří zabírají více než jedno políčko, nejsou ovlivněni. Kromě toho ti tato Sféra poskytuje imunitu proti kráse.",
   "С этой сферой Вы становитесь ошеломляюще прекрасны. Монстр, подошедший к Вам, застывает на один ход. Не действует на монстров из нескольких клеток. Также Вы теперь можете игнорировать красоту других.",
   "Dieser Orb macht dich atemberaubend schön. Monster, die sich dir nähern, werden für einen Zug betäubt. Monster, die sich über mehrere Zellen erstrecken, sind nicht betroffen. Außerdem macht er dich immun gegenüber 'Schönheit'.",
/*MISSING*/    "This Orb makes you stunningly beautiful. Monsters which come next to you will be stunned for one turn. Multi-tile monsters are not affected. Additionally, it makes you immune to beauty.",
    }},
  {0x4b4fc4fd, { // "move right"
   "idź w prawo",
   "sağa git",
   "doprava",
   "идти направо",
   "Nach rechts bewegen",
/*MISSING*/    "move right",
    }},
  {0x4b9b6984, { // "dist from hyperboloid ctr"
   "odległość od centrum hiperboloidy",
   "hiperboloid merkezinden uzaklık",
   "vzdálenost od středu hyperboloidu",
   "расстояние от центра гиперболоида",
   "Entf. vom hyperbolischen Zentrum",
   "distância do centro hiperbolóide",
    }},
  {0x4bb2ea16, { // "simply resize the window to change resolution"
   "po prostu zmień rozmiar okna",
   "pencerenin boyutunu değiştirerek çözünürlüğü değiştirebilirsin.",
   "rozlišení můžeš změnit prostou změnou velikosti okna",
   "просто поменяйте размер окна чтобы сменить разрешение",
   "Einfach die Fenstergröße anpassen um die Auflösung zu ändern",
   "redimensione a janela para mudar de resolução",
    }},
  {0x4bc7385a, { // "cancel"
   "anuluj",
/*MISSING*/    "cancel",
   "zrušit",
   "отменить",
/*MISSING*/    "cancel",
/*MISSING*/    "cancel",
    }},
  {0x4bf49e02, { // "Equidistants"
   "Ekwidystanty",
/*MISSING*/    "Equidistants",
   "Ekvidistanty",
   "Эквидистанты",
/*MISSING*/    "Equidistants",
/*MISSING*/    "Equidistants",
    }},
  {0x4c302be4, { // "F3 = load"
   "F3 = wczytaj",
   "F3 = yükle",
   "F3 = nahrání",
   "F3 = загрузить",
   "F3 = Laden",
/*MISSING*/    "F3 = load",
    }},
  {0x4ca27502, { // "mark heptagons"
   "oznaczenia na siedmiokątach",
/*MISSING*/    "mark heptagons",
   "označit sedmiúhelníky",
   "отметить семиугольники",
/*MISSING*/    "mark heptagons",
/*MISSING*/    "mark heptagons",
    }},
  {0x4cc25130, { // "You throw fire!"
   "Rzucasz ogniem!",
   "Ateş attın!",
   "Vrhl jsi oheň!",
   "Вы бросили огонь!",
   "Du wirfst Feuer!",
/*MISSING*/    "You throw fire!",
    }},
  {0x4cd01990, { // "Five mines next to you!"
   "Pięć min koło Ciebie!",
   "Yakınında beş mayın var!",
   "Ve tvé blízkosti se nachází pět min!",
   "Пяти мин рядом с Вами!",
   "5 Minen um dich!",
/*MISSING*/    "Five mines next to you!",
    }},
  {0x4d80629f, { // "use Shift to decrease and Ctrl to fine tune "
   "użyj Shift by zmniejszyć, Ctrl by zwiększyć dokładność",
   "Shift'le azalt, Ctrl ile ince ayar yap.",
   "použij Shift ke zmenšení a Ctrl k jemnému",
   "используйте Shift для уменьшения,",
   "Shift zum Verringern, Strg zum feinen",
   "use Shift para diminuir e Ctrl para ajuste fino",
    }},
  {0x4d895a7a, { // "In the pure tactics mode, you concentrate on a specific land. Your goal to obtain as high score as possible, without using features of the other lands. You can then compare your score with your friends!\n\nYou need to be somewhat proficient in the normal game to unlock the given land in this challenge (collect 20 treasure in the given land, or 2 in case of Camelot).\n\nSince getting high scores in some lands is somewhat luck dependent, you play each land N times, and your score is based on N consecutive plays. The value of N depends on how 'fast' the land is to play, and how random it is.\n\nIn the Caribbean, you can access Orbs of Thorns, Aether, and Space if you have ever collected 25 treasure in their native lands.\n\nThe rate of treasure spawn is static in this mode. It is not increased by killing monsters.\n\nGood luck, and have fun!"
   "W trybie taktycznym skupiasz się na jednej krainie. Twoim celem jest zdobycie jak najwyższego wyniku, bez używania elementów pozostałych krain. Możesz potem porównać wyniki ze znajomymi!\n\nMusisz sobie dobrze radzić w normalnej grze, by odblokować daną krainę tutaj (zbierz 20 skarbów w danym świecie, albo 2 w przypadku Camelot).\n\nZdobywanie wysokich wyników w niektórych krainach jest zależne od szczęścia, zatem grasz każdą krainę N razy, a Twój wynik jest oparty na N kolejnych grach. Wartość N zależy od tego, jak 'szybko' gra się w krainie i jak bardzo jest losowa.\n\nNa Karaibach masz dostęp do Sfery Cierni, Eteru i Przestrzeni, jeżeli kiedylkolwiek zebra%łeś0 25 skarbów w ich natywnych krainach.\n\nSzybkość generowania skarbów jest stała w tym trybie, nie powiększa się wraz z zabijaniem potworów.\n\nPowodzenia i miłej zabawy!",
   "Saf taktik modunda sadece tek bir diyara odaklanırsın. Amacın diğer diyarların özelliklerini hiç kullanmadan mümkün olduğunca yüksek puan yapmak. Sonra skorunu arkadaşlarınınkiyle karşılaştırabilirsin!\n\n Normal oyunda maharetini ispatlayarak belli bir diyarı bu modda açabilirsin. (Belli bir diyarda 20 hazine toplamak, Kamelot için 2 hazine toplamak gibi).\n\nYüksek skor almak şöyle böyle şansa bağlı olduğundan, her diyarı N defa oynarsın ve skorun bu ardarda N oyuna göre belirlenir. N'in değeri diyarın ne kadar hızlı oynandığına ve ne derece rassal olduğuna göre değişir.\n\nKarayiplerde, Diken Küresine, Ether'e ve Uzay'a eğer daha önce ilgili diyarlarda 25'er hazine topladıysan erişebilirsin.\n\nHazine üretimi bu modda statiktir ve canavar öldürmekle artmaz.\n\nİyi şanslar ve iyi eğlenceler!",
   "V taktickém módu se soustředíš pouze na jeden specifický kraj. Tvým cílem  je získání co nejvyššího skóre, bez použití jakýchkoli prvků jiných krajů. Potom si své skóre můžeš porovnat s přáteli!\n\nAbys v tomto módu odemkl určitý kraj, musíš být alespoň trochu dobrý ve standardní hře (získat 20 pokladů v daném kraji, 2 v případě Camelotu).\n\nVzhledem k tomu, že získání vysokého skóre v některých krajích poněkud závisí na štěstí, budeš každý kraj hrát N-krát a tvé skóre bude založeno na N po sobě jdoucích hrách. Hodnota N se odvíjí od 'rychlosti' a náhodnosti daného kraje.\n\nV Karibiku budeš mít k dispozici Sféry Trnů, Éteru a Prostoru, pokud jsi někdy získal 25 pokladů v jejich rodných krajích.\n\nGenerování pokladů je v tomto módu statické. Nezvyšuje se zabíjením netvorů.\n\nHodně štěstí, a dobře se bavte!",
   "В тактическом режиме Вы концентрируетесь на одной земле. Цель - максимизировать счёт,не используя особенности других земель. Вы можете сравнить свои результаты с результатами друзей!\n\nВы должны быть достаточно опытным в обычной игре, чтобы открыть землю здесь (соберите в ней 20 сокровищ, или 2 в случае Камелота).\n\nТак как результаты в некоторых землях могут зависеть от случая, Вам нужно сыграть N раз, и счёт зависит от N последовательных игр. Значение N зависит от того, насколько быстро проходит игра, и насколько она зависит от случая.\n\nНа Карибах Вам доступны сферы Шипов, Эфира и Пространства, если Вы собрали 25 сокровищ в их родных землях.\n\nЧастота появления сокровищ здесь постоянна, она не зависит от количества убитых монстров.\n\nУдачи и приятного времяпровождения!",
   "Im Strategiemodus konzentrierst du dich auf ein einziges Land. Das Ziel ist es eine möglichst hohe Punktzahl zu erreichen ohne features anderer Länder zu nutzen. Du kannst deine Punktzahl mit denen deiner Freunde vergleichen!\n\nDu musst im normalen Spiel etwas erfahrener sein um das Land in dieser Herausforderung freizuschalten. (Sammle 20 Schätze im jeweiligem Land, oder 2 im Fall von Camelot).\n\nDa die Punktzahl in manchen Ländern sehr glücksbasierend sein kann, wird jedes Land N mal gespielt und die Punktzahl basiert auf N Durchläufen. Die Anzahl N hängt davon ab wie 'schnell' und zufällig das Land ist.\n\nIn der Karibik hast du Zugriff auf den Orb der Dornen, den Orb des Aether und den Orb des Raumes wenn zuvor 25 Schätze in ihren Heimatländern gesammelt wurden.\n\nDie Anzahl der Schätze in diesem Modus ist festgelegt und wird nicht durch das Töten von Monstern erhöht.\n\nViel Glück und Erfolg!",
/*MISSING*/    "In the pure tactics mode, you concentrate on a specific land. Your goal to obtain as high score as possible, without using features of the other lands. You can then compare your score with your friends!\n\nYou need to be somewhat proficient in the normal game to unlock the given land in this challenge (collect 20 treasure in the given land, or 2 in case of Camelot).\n\nSince getting high scores in some lands is somewhat luck dependent, you play each land N times, and your score is based on N consecutive plays. The value of N depends on how 'fast' the land is to play, and how random it is.\n\nIn the Caribbean, you can access Orbs of Thorns, Aether, and Space if you have ever collected 25 treasure in their native lands.\n\nThe rate of treasure spawn is static in this mode. It is not increased by killing monsters.\n\nGood luck, and have fun!",
    }},
  {0x4d8ebaab, { // "This gate separates the warped area from the normal land."
   "Ta brama oddziela strefę zakrzywioną od normalnych krain.",
/*MISSING*/    "This gate separates the warped area from the normal land.",
   "Tato brána odděluje pokřivenou oblast od normálních krajů.",
   "Эти врата отделяют искривлённые земли от обычных.",
   "Dieses Tor trennt den verzerrten Bereich vom normalen Land.",
/*MISSING*/    "This gate separates the warped area from the normal land.",
    }},
  {0x4da1cb53, { // "Those large seabirds attack you for some reason. At least they are not as fast as Eagles..."
   "Te wielkie morskie ptaki atakują Cię z jakiegoś powodu. Przynajmniej nie są tak szybkie, jak Orły...",
   "Bu iri denizkuşları sana nedense saldırıyor. En azından kartallar kadar hızlı değiller.",
   "Tito velcí mořští ptáci na tebe z nějakého důvodu útočí. Aspoň že nejsou tak rychlí jako Orli...",
   "Эти большие морские птицв почему-то атакуют Вас. Они хотя бы не такие быстрые, как Орлы...",
   "Diese großen Seevögel greifen dich aus irgendeinem Grund an. Wenigstens sind sie langsamer als Adler...",
/*MISSING*/    "Those large seabirds attack you for some reason. At least they are not as fast as Eagles...",
    }},
  {0x4da25981, { // "spin up"
   "obróć w górę",
/*MISSING*/    "spin up",
   "otoč nahoru",
   "вращать вверх",
   "herunterdrehen",
/*MISSING*/    "spin up",
    }},
  {0x4dd131d4, { // "You do not want to lose the game from a single mistake?\nDo you want to use the Orbs strategically?\nTry the Orb Strategy mode!"
   "Nie chcesz ginąć od jednego błędu? Chcesz używać Sfer strategicznie? Wypróbuj tryb strategii sfer!",
/*MISSING*/    "You do not want to lose the game from a single mistake?\nDo you want to use the Orbs strategically?\nTry the Orb Strategy mode!",
   "Nechceš prohrát kvùli jedné chybì?\nChceš používat Sféry strategicky\nZkus Sférostrategický mód!",
   "Не хочешь проигрывать из-за одной ошибки?\nХочешь научить аккуратно использовать Сферы?\nПопробуй режим стратегии сфер!",
/*MISSING*/    "You do not want to lose the game from a single mistake?\nDo you want to use the Orbs strategically?\nTry the Orb Strategy mode!",
/*MISSING*/    "You do not want to lose the game from a single mistake?\nDo you want to use the Orbs strategically?\nTry the Orb Strategy mode!",
    }},
  {0x4dd6967c, { // "next page"
   "kolejna strona",
/*MISSING*/    "next page",
   "další stránka",
   "следующая страница",
   "nächste Seite",
/*MISSING*/    "next page",
    }},
  {0x4de068d5, { // "skip the start menu"
   "pomiń menu startowe",
/*MISSING*/    "skip the start menu",
   "přeskoč menu 'start'",
   "пропустить начальное меню",
/*MISSING*/    "skip the start menu",
/*MISSING*/    "skip the start menu",
    }},
  {0x4dea6545, { // "Many lands in HyperRogue are based around periodic patterns. For example, both Zebra and Windy Plains are based on the pattern shown in the previous slide. Such lands often have tree-like nature."
   "Wiele krain jest opartych na okresowych wzorach. Na przykład Zebra i Wietrzna Równina są obie oparte na wzorze z poprzedniego slajdu. Krainy takie mają zwykle drzewiastą strukturę.",
/*MISSING*/    "Many lands in HyperRogue are based around periodic patterns. For example, both Zebra and Windy Plains are based on the pattern shown in the previous slide. Such lands often have tree-like nature.",
   "HyperRogue obsahuje řadu krajů, které jsou založené na periodických vzorech. Tak například Zebra a Větrné pláně jsou obě založené na vzoru na předešlém snímku. Tyto kraje mívají často stromovou povahu.",
   "Многие земли основаны на периодических узорах. Например, Зебра и Ветреная Равнина содержат узоры с предыдущего слайда. Такие земли часто имеют древовидную структуру.",
/*MISSING*/    "Many lands in HyperRogue are based around periodic patterns. For example, both Zebra and Windy Plains are based on the pattern shown in the previous slide. Such lands often have tree-like nature.",
/*MISSING*/    "Many lands in HyperRogue are based around periodic patterns. For example, both Zebra and Windy Plains are based on the pattern shown in the previous slide. Such lands often have tree-like nature.",
    }},
  {0x4e2aa36d, { // "In the Chaos mode, lands change very often, and there are no walls between them. Some lands are incompatible with this.\n\nYou need to reach Crossroads IV to unlock the Chaos mode."
   "W trybie Chaosu krainy zmieniają się bardzo często i nie ma między nimi ścian. Niektóre krainy są niekompatybilne z tym trybem.\n\nMusisz dotrzeć do Skrzyżowania IV, by odblokować tryb Chaosu.",
/*MISSING*/    "In the Chaos mode, lands change very often, and there are no walls between them. Some lands are incompatible with this.\n\nYou need to reach Crossroads IV to unlock the Chaos mode.",
   "V chaotickém módu se kraje mění velice často a nejsou mezi nimi žádné stěny. Některé kraje nejsou s tímto módem kompatibilní.\n\nChaotický mód musíš odemknout tím, že dojdeš na Křižovatku IV.",
   "В режиме Хаоса земли меняются очень часто, и между ними нет стен. Некоторые земли несовместимы с этим режимом.\n\nВы должны достичь Перекрёстка IV, чтобы открыть режим Хаоса.",
   "Im Chaos-Modus wechseln die Ländern sehr oft und es gibt keine Wände zwischen ihnen. Einige Länder sind damit inkompatibel.\n\nDu musst die Kreuzungen IV erreichen, um den Chaos-Modus freizuschalten.",
/*MISSING*/    "In the Chaos mode, lands change very often, and there are no walls between them. Some lands are incompatible with this.\n\nYou need to reach Crossroads IV to unlock the Chaos mode.",
    }},
  {0x4e47b5a2, { // "0 = asymmetry"
   "0 = asymetria",
   "0 = asimetri",
   "0 = asymetrie",
   "0 = асимметрия",
   "0 = Asymmetrie",
/*MISSING*/    "0 = asymmetry",
    }},
  {0x4e6892d5, { // "Welcome to HyperRogue"
   "Witaj w HyperRogue",
   "HyperRogue'a hoşgeldin",
   "Vítej ve hře HyperRogue",
   "Добро пожаловать в HyperRogue",
   "Willkommen bei HyperRogue",
   "Bem-vindo ao HyperRogue",
    }},
  {0x4eb120ee, { // "player 5 X"
   "gracz 5 X",
/*MISSING*/    "player 5 X",
   "hráč 5 X",
   "игрок 5 X",
   "Spieler 5 X",
/*MISSING*/    "player 5 X",
    }},
  {0x4eb120ef, { // "player 5 Y"
   "gracz 5 Y",
/*MISSING*/    "player 5 Y",
   "hráč 5 Y",
   "игрок 5 Y",
   "Spieler 5 Y",
/*MISSING*/    "player 5 Y",
    }},
  {0x4f0a247c, { // "Equidistants are curves which are at some fixed distance from a straight line. Some lands in HyperRogue are based on equidistants; you should see them after wandering a bit more.\n\nThis tutorial gives you freedom to go wherever you choose, but we do not recommend going deep into the Dungeon or the Ocean -- getting back might be difficult."
   "Ekwidystanty to krzywe w stałej odległości od linii prostej. Niektóre krainy w grze są oparte na ekwidystantach; po krótkiej podróży je zobaczysz.\n\nTen podręcznik pozwala Ci się ruszać dokąd chcesz, ale nie polecamy zagłębiać się w Lochy ani w Ocean -- powrót może być trudny.",
/*MISSING*/    "Equidistants are curves which are at some fixed distance from a straight line. Some lands in HyperRogue are based on equidistants; you should see them after wandering a bit more.\n\nThis tutorial gives you freedom to go wherever you choose, but we do not recommend going deep into the Dungeon or the Ocean -- getting back might be difficult.",
   "Ekvidistanty jsou křivky, které mají určitou danou vzdálenost od nějaké přímky. V HyperRogue jsou kraje, které jsou na ekvidistantách založeny; když budeš putovat dál, měl bys na ně narazit.\n\nV tomto tutoriálu se můžeš vydat, kam se ti zlíbí, ale doporučujeme, abys nechodil příliš hluboko do Žaláře nebo do Oceánu -- mohl bys mít problém vrátit se zpátky.",
   "Эквидистантой называется кривая, состоящая из точек на фиксированном расстоянии от прямой. Некоторые земли основаны на эквидистантах; ты увидишь их через некоторое время.\n\nЭто руководство позволяет идти куда угодно, но мы не рекомендуем забираться глубоко в Подземелье или Океан -- возвращение может оказаться трудным.",
/*MISSING*/    "Equidistants are curves which are at some fixed distance from a straight line. Some lands in HyperRogue are based on equidistants; you should see them after wandering a bit more.\n\nThis tutorial gives you freedom to go wherever you choose, but we do not recommend going deep into the Dungeon or the Ocean -- getting back might be difficult.",
/*MISSING*/    "Equidistants are curves which are at some fixed distance from a straight line. Some lands in HyperRogue are based on equidistants; you should see them after wandering a bit more.\n\nThis tutorial gives you freedom to go wherever you choose, but we do not recommend going deep into the Dungeon or the Ocean -- getting back might be difficult.",
    }},
  {0x4f1793ec, { // " (%1 more digits)"
   " (jeszcze cyfr: %1)",
/*MISSING*/    " (%1 more digits)",
   " (%1 dalších èíslic)",
   " (ещё цифр: %1)",
/*MISSING*/    " (%1 more digits)",
/*MISSING*/    " (%1 more digits)",
    }},
  {0x4f429a41, { // "draw the grid"
   "rysowanie siatki",
/*MISSING*/    "draw the grid",
   "vykreslování møíky",
   "rysowanie siatki",
/*MISSING*/    "draw the grid",
/*MISSING*/    "draw the grid",
    }},
  {0x4f9581a9, { // "hardcore mode"
   "tryb hardcore",
   "Aşmış mod",
   "hardcore mód",
   "режим hardcore",
   "Hardcore Modus",
/*MISSING*/    "hardcore mode",
    }},
  {0x4f9d2e28, { // "gain orb powers"
   "Moce sfer",
   "küre güçleri kazan",
   "získej moc sféry",
   "получить силу сфер",
   "erhalte Orbkräfte",
/*MISSING*/    "gain orb powers",
    }},
  {0x4fbeb090, { // "Trolls of Trollheim are descendants of a bridge Troll, who collected payments from people crossing the bridge. One of them paid with golden eggs. The bridge Troll found the eggs beautiful, but he quickly lost them. Golden eggs are still revered by Trolls, and you can find them in their caves."
   "Trolle z Trollheim są potomkami Trolla, który pobierał myto od osób przechodzących mostem. Jeden z podróżnych zapłacił złotymi jajami. Jaja bardzo spodobały się naszemu Trollowi, ale szybko je zgubił. Trolle z Trollheim wciąż darzą złote jaja wielkim poważaniem i możesz znaleźć je w ich legowiskach.",
/*MISSING*/    "Trolls of Trollheim are descendants of a bridge Troll, who collected payments from people crossing the bridge. One of them paid with golden eggs. The bridge Troll found the eggs beautiful, but he quickly lost them. Golden eggs are still revered by Trolls, and you can find them in their caves.",
   "Trollové z Trollheimu jsou potomky trolla, který hlídal most a vybíral mýtné od těch, kdo po něm chtěli přejít. Jeden z nich zaplatil zlatými vejci. Trollovi tato vejce přišla moc krásná, ale rychle je ztratil. Trollové uctívají zlatá vejce dodnes a můžete je najít v jejich jeskyních.",
   "Все тролли здесь -- потомки Тролля моста, собиравшего плату с людей, пересекавших мост. Один из них заплатил золотыми яйцами. Тролль счёл эти яйца прекрасными, но вскоре потерял. Тролли по-прежнему почитают Золотые яйца, и вы можете найти их в местных пещерах.",
   "Die Trolle von Trollheim sind Nachfahren eines Brückentrolls, der Wegzölle von allen gesammelt hat, die die Brücke überquerten. Einer von ihnen bezahlte mit goldenen Eiern.Der Brückentroll fand die Eier sehr schön, aber verlor sie bald darauf. Die Eier werden von den Trollen noch immer verehrt, und du kannst sie in ihren Höhlen finden.",
/*MISSING*/    "Trolls of Trollheim are descendants of a bridge Troll, who collected payments from people crossing the bridge. One of them paid with golden eggs. The bridge Troll found the eggs beautiful, but he quickly lost them. Golden eggs are still revered by Trolls, and you can find them in their caves.",
    }},
  {0x4fd9ea92, { // "Alternatively: kill a %1 in %the2.\n"
   "Alternatywnie: zabij %a1 %abl2.\n",
/*MISSING*/    "Alternatively: kill a %1 in %the2.\n",
   "Alternativně: zabij %a1 %abl2.\n",
   "Альтернатива: убить %a1 %abl2.\n",
   "Alternativ: töte einen %a1 %d2.\n",
/*MISSING*/    "Alternatively: kill a %1 in %the2.\n",
    }},
  {0x4fdb418c, { // "High detail range"
   "Zasięg wysokiego poziomu szczegółów",
/*MISSING*/    "High detail range",
   "Rozsah vysokých detailů",
   "Зона высокой детализации",
/*MISSING*/    "High detail range",
/*MISSING*/    "High detail range",
    }},
  {0x4ff0f741, { // "z = zoom in"
   "z = zbliżenie",
   "z = yaklaş",
   "z = přiblížení",
   "z = приблизить",
   "z = reinzoomen",
/*MISSING*/    "z = zoom in",
    }},
  {0x50b9b5d3, { // "restart game"
   "gra od nowa",
   "oyunu yeniden başlat",
   "restartovat hru",
   "играть заново",
   "Neustart",
/*MISSING*/    "restart game",
    }},
  {0x514c6466, { // "Killed %1 can be revived with Orb of the Love, after you collect 20 more $$$."
   "Zabit%ya1 %1 może zostać przywrócona do życia Sferą Miłości, po zdobyciu 20 dodatkowych $$$.",
/*MISSING*/    "Killed %1 can be revived with Orb of the Love, after you collect 20 more $$$.",
   "Zabit%ého1 %a1 je možné oživit pomocí Sféry Lásky, ale předtím musíš získat ještě 20 $$$.",
   "Убит%y1 %1 может быть оживлён при помощи Сферы любви, когда Вы соберёте ещё 20 $$$.",
/*MISSING*/    "Killed %1 can be revived with Orb of the Love, after you collect 20 more $$$.",
/*MISSING*/    "Killed %1 can be revived with Orb of the Love, after you collect 20 more $$$.",
    }},
  {0x51679fa7, { // "display the triheptagonal grid"
   "pokaż siatkę triheptagonalną",
/*MISSING*/    "display the triheptagonal grid",
   "zobraz triheptagonální mřížku",
   "показать трёх-семи-угольную сетку",
   "Zeige das tri-heptagonale Gitter an",
/*MISSING*/    "display the triheptagonal grid",
    }},
  {0x517de5e8, { // "random black-and-white"
   "losowy czarno-biały wzorek",
   "rastgle siyah-beyaz",
   "náhodný černobílý vzor",
   "случайный чёрно-белый",
   "Zufalls Schwarz und Weiß",
/*MISSING*/    "random black-and-white",
    }},
  {0x51b360ed, { // "this Orb is never unlocked globally (only hubs)"
   "ta Sfera występuje tylko w centrach",
   "bu Küre hiç tüm dünya için aktif edilmedi (sadece yerelde)",
   "tato Sféra není nikdy odemčena globálně (pouze v centrálních krajích)",
   "эта сфера открывается только ы Центрах",
   "Dieser Orb wird niemals global freigeschaltet",
/*MISSING*/    "this Orb is never unlocked globally (only hubs)",
    }},
  {0x51c52bbd, { // "Halloween mini-game"
   "mini-gra Halloween",
/*MISSING*/    "Halloween mini-game",
   "mini-gra Halloween",
   "мини-игра Хэллоуин",
/*MISSING*/    "Halloween mini-game",
/*MISSING*/    "Halloween mini-game",
    }},
  {0x520faa64, { // "Your total kills"
   "Łączne zabicia",
   "Toplam leşler",
   "Celkový počet zabitých netvorů",
   "Всего убийств",
   "Deine Gesamtkills",
/*MISSING*/    "Your total kills",
    }},
  {0x5250bde7, { // "Orb power (target: mouse)"
   "moc sfery (cel: mysz)",
/*MISSING*/    "Orb power (target: mouse)",
   "Použití Sféry (zaměření myší)",
   "активировать сферу",
   "Orbkraft (Ziel: Maus)",
/*MISSING*/    "Orb power (target: mouse)",
    }},
  {0x52db38db, { // "level to save/load:"
   "mapa do zapisu/wczytania:",
   "kaydedilecek/yüklenecek bölüm",
   "mapa k uložení/nahrání:",
   "карты для загрузки/сохранения:",
   "Level zum speichern/laden:",
/*MISSING*/    "level to save/load:",
    }},
  {0x530987e8, { // "Dragons are powerful monsters. They are slow, but evil, and love to pick on creatures who are even slower than them. They must be stopped!\n\nA Dragon moves each two turns. It may attack with all its segments, or move its whole body forwards or backwards, it may also move a frontal part backwards. To kill a Dragon, you need to hit each of its segments. The head will regenerate on the turns the Dragon is not moving, so you will usually have to hit it with your last attack; otherwise, if the head is healthy, it may breathe fire (at range 3), losing the hitpoint. Killing the Dragon while still in the Dragon Chasms gives you treasure."
   "Smoki to potężne bestie. Są powolne, ale bardzo złe, uwielbiają znęcać się nad stworzeniami, które są od nich jeszcze wolniejsze. Musisz je powstrzymać!\n\nSmok rusza się co dwie kolejki. Może atakować każdym segmentem, ruszyć całe swoje ciało do przodu lub do tyłu albo ruszyć początkowy fragment do tyłu. By pokonać Smoka, musisz uderzyć każdy z jego segmentów. Głowa się regeneruje w kolejkach, podczas których Smok się nie rusza, także zwykle musi być atakowana ostatnia; jeśli głowa jest zdrowa, może zionąć ogniem (zasięg 3), tracąc punkt życia. Za pokonanie Smoka (będąc wciąż w Smoczych Otchłaniach) dostajesz skarby.",
/*MISSING*/    "Dragons are powerful monsters. They are slow, but evil, and love to pick on creatures who are even slower than them. They must be stopped!\n\nA Dragon moves each two turns. It may attack with all its segments, or move its whole body forwards or backwards, it may also move a frontal part backwards. To kill a Dragon, you need to hit each of its segments. The head will regenerate on the turns the Dragon is not moving, so you will usually have to hit it with your last attack; otherwise, if the head is healthy, it may breathe fire (at range 3), losing the hitpoint. Killing the Dragon while still in the Dragon Chasms gives you treasure.",
   "Draci jsou mocní netvoři. Jsou pomalí, ale zlí, a moc rádi si dovolují na ty, kdo jsou ještě pomalejší než oni. Musíš je zastavit!\n\nDrak se pohybuje vždy jednou za dvě kola. Může zaútočit všemi svými segmenty, posunout celé své tělo dopředu nebo dozadu, nebo stáhnout dozadu přední část těla. Abys je zabil, musíš zasáhnout všechny jeho segmenty. Když se Drak nehýbe, jeho hlava automaticky zregeneruje, takže ji budeš obvykle musez zasáhnout posledním útokem; pokud je hlava zdravá, může drak chrlit oheň (na vzdálenost 3 polí), což jí vezme jeden život. Zabitím Draka v Dračích propastech získáš poklady.",
   "Драконы -- очень сильные существа. Они медленные, но злые, и любят издеваться над существами, которые ещё медленнее. Остановите их!\n\nДракон движется каждый второй ход. Он может атаковать любой своей частью, либо передвинуться вперёд или назад, либо передвинуть переднюю часть назад. Чтобы убить дракона, нужно ударить каждую его часть. Голова регенерирует каждый ход, когда дракон не движется, так что обычно её придётся бить последней; если голова здорова, дракон может дышать огнём (дальность 3), тратя одну жизнь. За убийство дракона в Драконьей бездне вы получите награду.",
   "Drachen sind mächtige Monster. Sie sind langsam, aber böse, und lieben es Wesen zu schikanieren, die noch langsamer sind als sie. Sie müssen gestoppt werden!\n\nEin Drachen bewegt sich alle zwei Züge. Er kann mit all seinen Segmenten angreifen, oder seinen ganzen Körper vorwärts oder rückwärts bewegen, oder vordere Segmente rückwärts bewegen. Um einen Drachen zu töten, musst du jedes seiner Segmente treffen. Der Kopf regeneriert sich während der Züge, in denen der Drache sich nicht bewegt, daher solltest du diesen zuletzt treffen; andernfalls kann ein gesunder Kopf Feuer speien (im Radius 3) und dabei seinen Hitpoint verlieren. Den Drachen zu töten wird dir Schätze bescheren.",
/*MISSING*/    "Dragons are powerful monsters. They are slow, but evil, and love to pick on creatures who are even slower than them. They must be stopped!\n\nA Dragon moves each two turns. It may attack with all its segments, or move its whole body forwards or backwards, it may also move a frontal part backwards. To kill a Dragon, you need to hit each of its segments. The head will regenerate on the turns the Dragon is not moving, so you will usually have to hit it with your last attack; otherwise, if the head is healthy, it may breathe fire (at range 3), losing the hitpoint. Killing the Dragon while still in the Dragon Chasms gives you treasure.",
    }},
  {0x53391ee5, { // "The paper model created as papermodel-*.bmp"
   "Model stworzony jako papermodel-*.bmp",
   "Kâğıt model papermodel-*.bmp şeklinde üretildi",
   "Papírový model byl vytvořen jako papermodel-*.bmp",
   "Модель сохранена как papermodel-*.bmp",
   "Das Papiermodell wurde als papermodel-*.bmp erzeugt",
/*MISSING*/    "The paper model created as papermodel-*.bmp",
    }},
  {0x533c543c, { // "play"
   "graj",
/*MISSING*/    "play",
   "hraj",
   "играй",
/*MISSING*/    "play",
/*MISSING*/    "play",
    }},
  {0x537619eb, { // "Welcome to Halloween!"
   "Witaj w Halloween!",
/*MISSING*/    "Welcome to Halloween!",
   "Vítejte v Halloweenu!",
   "Witaj w Halloween!",
/*MISSING*/    "Welcome to Halloween!",
/*MISSING*/    "Welcome to Halloween!",
    }},
  {0x539a39d3, { // "This cell is partially on fire."
   "To pole częściowo się pali.",
   "Bu hücre kısmen alevler içinde.",
   "Toto políčko částečně hoří.",
   "Эта клетка частично в огне.",
   "Diese Zelle brennt teilweise.",
/*MISSING*/    "This cell is partially on fire.",
    }},
  {0x53c8b5d8, { // "Quest status"
   "Stan misji",
   "Görev durumu",
   "Stav úkolů",
   "Статус миссии",
   "Queststatus",
   "Estado da missão",
    }},
  {0x53ca77b3, { // "Collect 10 treasures in various lands to unlock the challenges there"
   "Zbieraj po 10 skarbów w różnych krainach, by odblokować tutejsze misje",
   "Çeşitli diyarlarda 10 hazine toplayarak o diyarın özel görevini aktifleştirebilirsin",
   "Tyto mise odemkneš získáním 10 pokladů v různých krajích",
   "Собирайте 10 сокровищ в разных землях, чтобы разблокировать миссии здесь",
   "Sammle 10 Schätze in mehreren Ländern um diese Herausforderung freizuschalten",
/*MISSING*/    "Collect 10 treasures in various lands to unlock the challenges there",
    }},
#if REPEATED
  {0x53ca77b4, { // "Collect 10 treasures in various lands to unlock the challenges there [repeat]"
   "Zdobywaj 10 skarbów w różnych krainach, by odblokowywać misje",
   "Diyarlardaki ek görevleri aktif etmek için oralarda 10 hazine topla.",
   "Tyto mise odemkneš získáním 10 pokladů v různých krajích",
   "Собирайте 10 сокровищ в разных землях, чтобы открыть миссии здесь",
   "Sammle 10 Schätze in den verschiedenen Ländern um Herausforderungen freizuschalten",
/*MISSING*/    "Collect 10 treasures in various lands to unlock the challenges there [repeat]",
    }},
#endif
  {0x53f26d78, { // "Collect %1x %2 to unlock"
   "Zbierz %1x %2 by odblokować",
   " %1 tane %2 toplayarak aktifleştir",
   "Pro odemknutí získej %1x %2",
   "Соберите %1x %2, чтобы разблокировать",
   "Sammle %1x %2 zum Freischalten",
/*MISSING*/    "Collect %1x %2 to unlock",
    }},
  {0x5466eb5d, { // "c = copy"
   "b = brzeg",
   "b = kopya",
   "w = zdi",
   "w = стены",
   "b = Kopieren",
/*MISSING*/    "c = copy",
    }},
  {0x546da5c3, { // "forbidden to find in %the1"
   "zakazana %abl1",
   "%abl1 bulmak yasak.",
   "zakázaná %abl1",
   "запрещена %abl1",
   "verboten in %der1 etwas zu finden",
/*MISSING*/    "forbidden to find in %the1",
    }},
  {0x546ecfbd, { // "Date: %1 time: %2 s "
   "Data: %1 czas: %2 s ",
   "Tarih: %1 zaman: %2 s ",
   "Datum: %1 čas: %2 s ",
   "Дата: %1 время: %2 s ",
   "Datum: %1 Zeit: %2 s ",
/*MISSING*/    "Date: %1 time: %2 s ",
    }},
  {0x54c2ea87, { // "Fractal landscapes"
   "Fraktalne krajobrazy",
/*MISSING*/    "Fractal landscapes",
   "Fraktální krajinky",
   "Фрактальные пейзажи",
/*MISSING*/    "Fractal landscapes",
/*MISSING*/    "Fractal landscapes",
    }},
  {0x54fb4344, { // "This boat cannot go through the sand. But if you sit inside and wait for the tide, you will be able to use it to travel through the Ocean."
   "Ta łódka nie może pływać przez piasek. Ale jeśli w niej usiądziesz i poczekasz na przypływ, będziesz móc płynąć przez Ocean.",
   "Bu bot kumda hareket edemez. Ama içine oturur ve dalgayı beklersen bunu OKyanus'ta seyahat etmek için kullanabilirsin.",
   "Tento člun nemůže plout po písku. Ale pokud si do něj sedneš a počkáš na příliv, můžeš na něm cestovat přes Oceán.",
   "Эта лодка не может пройти сквозь песок. Но если вы сядете внутрь и дождётесь прилива, Вы сможете ей воспользоваться.",
   "Dieses Boot kann nicht durch den Sand schwimmen. Wartest du allerdings bis zur Flut, kannst du damit durch den Ozean reisen.",
/*MISSING*/    "This boat cannot go through the sand. But if you sit inside and wait for the tide, you will be able to use it to travel through the Ocean.",
    }},
  {0x551495b5, { // "A precious metal from the Dead Caves."
   "Drogocenny metal z Martwej Jaskini.",
   "Ölü mağaralardan çıkan değerli bir metal.",
   "Vzácný kov z Mrtvé jeskyně.",
   "Драгоценный металл из Мёртвых пещер.",
   "Ein wertvolles Metall aus den toten Höhlen.",
/*MISSING*/    "A precious metal from the Dead Caves.",
    }},
  {0x55183080, { // "A being made of stone, who likes high buildings. It becomes normal stone when killed, but only if next to something stable -- otherwise it falls."
   "Stworzenie wykonane z kamienia, lubi wysokie budynki. Zamienia się w normalny kamień, gdy go zabijesz, ale tylko, gdy stanie się to obok czegoś stabilnego -- w przeciwnym razie spada.",
   "Yüksek binaları seven taştan yapılmış bir yaratık. Öldürüldüğünde eğer durağan bir şeyin yanındaysa normal taş haline gelir, yoksa düşer.",
   "Kamenná bytost, která má ráda vysoké budovy. Po smrti se promění v obyčejný kámen, ale pouze tehdy, když je vedle něčeho stabilního -- jinak spadne.",
   "Существо из камня, выглядящее как высокое здание. Становится камнем, когда умирает, если рядом есть что-то стабильное -- иначе падает.",
   "Ein Wesen aus Stein das hohe Gebäude mag. Es wird zu normalem Stein wenn es stirbt, vorausgesetzt es ist neben etwas Stabilem. Ansonsten fällt es.",
/*MISSING*/    "A being made of stone, who likes high buildings. It becomes normal stone when killed, but only if next to something stable -- otherwise it falls.",
    }},
  {0x55361341, { // "There are several species of trolls living in the hyperbolic world. Some of them leave this wall behind them when they die."
   "W hiperbolicznym świecie żyje kilka gatunków trolla. Niektóre z nich zostawiają taką ścianę, gdy się je zabije.",
   "Hiperbolik dünyada pek çok tür trol yaşar.Bazıları öldüğünde bu duvarı geride bırakır.",
   "V hyperbolickém světě žije několik druhů trollů. Někteří z nich se po smrti promění v tuto zeď.",
   "В гиперболическом мире есть несколько видов троллей. Некоторые из них при смерти превращаются в такую стену.",
   "Es gibt mehrere verschiedene Trollgattungen in der hyperbolischen Welt. Manche von ihnen hinterlassen Wände wenn sie sterben.",
/*MISSING*/    "There are several species of trolls living in the hyperbolic world. Some of them leave this wall behind them when they die.",
    }},
  {0x556f6411, { // "This might be very useful for devices with limited memory."
   "To może być bardzo przydatne na urządzeniach z ograniczoną pamięcią.",
/*MISSING*/    "This might be very useful for devices with limited memory.",
   "To může být velmi užitečné při hře na zařízeních s omezenou pamětí.",
   "Это может быть полезно для устройств с ограниченной памятью.",
   "Das kann sehr nützlich für Geräte mit begrenztem Speicher sein.",
/*MISSING*/    "This might be very useful for devices with limited memory.",
    }},
  {0x5578b72d, { // "\"Our Table seats %1 Knights!\""
   "\"Przy naszym stole mieści się %1 Rycerzy!\"",
   "\"Masamızda %1 Şövalye oturuyor!\"",
   "\"U našeho stolu sedí %1 rytířů!\"",
   "\"За нашим столом сидит %1 рыцарей!\"",
   "\"An unserer Tafel sitzen %1 Ritter!\"",
/*MISSING*/    "\"Our Table seats %1 Knights!\"",
    }},
  {0x55879513, { // "item"
   "przedmiot",
   "eşya",
   "předmět",
   "предмет",
   "Item",
/*MISSING*/    "item",
    }},
  {0x55b23682, { // "enable/disable texts"
   "włącz/wyłącz teksty",
/*MISSING*/    "enable/disable texts",
   "zapnout/vypnout texty",
   "включить/выключить тексты",
/*MISSING*/    "enable/disable texts",
/*MISSING*/    "enable/disable texts",
    }},
  {0x55d8e966, { // "\nRanged Orbs can be targeted by shift-clicking the desired location. "
   "\nRobi się to przez kliknięcie na celu z Shiftem.",
/*MISSING*/    "\nRanged Orbs can be targeted by shift-clicking the desired location. ",
   "\nTyto sféry lze zacílovat kliknutím na požadované políčko se stisknutou klávesou Shift.",
   "\nВыбор цели сферы определяется щелчком мыши на цели с зажатым Shift'ом.",
   "\nMit Distanzorbs kann per Shift-Klick auf die gewünschte Position gezielt werden. ",
/*MISSING*/    "\nRanged Orbs can be targeted by shift-clicking the desired location. ",
    }},
  {0x55fa9e37, { // "Known mines may be marked by pressing 'm'. Your allies won't step on marked mines."
   "Znane miny można oznaczać naciskając 'm'. Twoi przyjaciele nie będą stawać na oznaczonych minach.",
/*MISSING*/    "Known mines may be marked by pressing 'm'. Your allies won't step on marked mines.",
   "Políčka, která obsahují miny, je možné označit stiskem klávesy 'm'. Vaši spojenci na označené miny nevstoupí.",
   "Известные мины можно помечать, нажимая 'm'. Твои союзники не будут на них наступать.",
/*MISSING*/    "Known mines may be marked by pressing 'm'. Your allies won't step on marked mines.",
/*MISSING*/    "Known mines may be marked by pressing 'm'. Your allies won't step on marked mines.",
    }},
  {0x562c1c2e, { // "You shoot %the1!"
   "Zastrzeli%łeś0 %a1!",
   "%1 düşmanına ateş ettin!",
   "Zastřeli%l0 jsi %a1!",
   "Вы застрелили %a1!",
   "Du erschießt %den1 %a1!",
/*MISSING*/    "You shoot %the1!",
    }},
  {0x563373a5, { // "Kills required: any Elemental (Living Fjord/Dead Caves).\n"
   "Wymagane zabicia: dowolny Żywiołak (Żywy Fiord/Martwa Jaskinia).",
   "Gereken leşler: Herhangi bir Özüt (Yaşayan Fiyort/Ölü Mağaralar).",
   "Potřebuješ zabít libovolného Elementála (Živoucí fjord/Mrtvé jeskyně).",
   "Требуется убить любого элементаля (Живой фьорд/Мёртвая пещера).",
   "Benötigte Kills: beliebiger Elementar (Lebender Fjord/Tote Höhle).",
/*MISSING*/    "Kills required: any Elemental (Living Fjord/Dead Caves).\n",
    }},
  {0x564507de, { // "%The1 is frozen!"
   "%1 zamarzł%oa1!",
   "%1 dondu!",
   "%1 zamrz%l1!",
   "%1 замерзает!",
   "%Der1 %1 ist eingefroren!",
/*MISSING*/    "%The1 is frozen!",
    }},
  {0x5655aa99, { // "A piece of architecture typical to R'Lyeh."
   "Architektura typowa dla R'Lyeh.",
   "R'Lyeh'e özgü bir mimari eser.",
   "Typický architektonický prvek R'Lyeh.",
   "Типичная архитектура Р'Льеха.",
   "Ein Stück R'Lyeh-typischer Architektur.",
/*MISSING*/    "A piece of architecture typical to R'Lyeh.",
    }},
  {0x565a4f71, { // "%The1 wakes up %the2."
   "%1 budzi %a2.",
/*MISSING*/    "%The1 wakes up %the2.",
   "%1 probudil %a2.",
   "%1 разбудил%E1 %a2.",
/*MISSING*/    "%The1 wakes up %the2.",
/*MISSING*/    "%The1 wakes up %the2.",
    }},
  {0x56675618, { // "summon Sand Worm"
   "przywołaj Czerwia",
   "Kumkurdu çıkar",
   "vyvolej Písečného červa",
   "призвать Песчаного червя",
   "rufe Sandwurm herbei",
/*MISSING*/    "summon Sand Worm",
    }},
  {0x5669c348, { // "\"Have you tried to take a boat and go into the Ocean? Try it!\""
   "\"Próbowa%łeś0 wziąć łódkę i płynąć na Ocean?\"",
   "\"Bir kayığa binip Okyanus'a açılmayı denedin mi? Denemelisin!\"",
   "\"Už jsi zkouše%l0 vzít člun a vyplout na Oceán?\"",
   "\"Вы пробовали взять лодку и уплыть в Океан?\"",
   "\"Hast du schon versucht mit einem Boot auf den Ozean zu fahren?\"",
/*MISSING*/    "\"Have you tried to take a boat and go into the Ocean? Try it!\"",
    }},
  {0x566c3369, { // "Move with mouse, num pad, qweadzxc, or hjklyubn. Wait by pressing 's' or '.'. Spin the world with arrows, PageUp/Down, and Home/Space. To save the game you need an Orb of Safety. Press 'v' for the main menu (configuration, special modes, etc.), ESC for the quest status.\n\n"
   "Ruszasz się myszą, klawiaturą numeryczną, qweadzxc, lub hjklyubn. Czekasz naciskając 's' lub '.'. Obracasz świat strzałkami, PgUp/Dn, lub Home/Space. Naciśnij 'v' by przejść do menu (konfiguracja, tryby specjalne itd.), ESC by zobaczyć stan misji.\n\n",
   "Fareyle, sayı tuşlarıyla, qweadzxc ile, veya hjklyubn ile hareket et.  's' ya da '.' ile bekle. Oklarla, PageUp/Down ile veya  Home/Space ile dünyayı çevirebilirsin. Oyunu kaydetmek için Güvenlik Küresine ihtiyacın var. Ayarlar için 'v'ye görev durumu ve menü için 'ESC'ye bas.\n\n",
   "Pohybuješ se pomocí myši, numerické klávesnice, kláves qweadzxc nebo kláves hjklyubn. Klávesy 's' nebo '.' ti umožňují čekat na místě. Šipky, klávesy PgUp/PgDn a Home nebo mezerník ti umožňují otáčet svět. Nastavení můžeš otevřít klávesou 'v', úkoly a menu klávesou ESC.\n\n",
   "Двигайтесь с помощью мышки, нумпада, qweadzxc или hjklyubn. Ждите, нажимая 's' или '.'. Поворачивайте карту стрелками, PageUp/Down или Home/Space. Чтобы сохраниться, Вам нужна сфера безопасности. Нажмите 'v' для настроек, Esc для статуса квеста или меню.",
   "Bewege dich mit der Maus, dem Numpad, qweadzxc, oder hjklyubn. Warte mit 's' oder '.'. Drehe die Welt mit den Pfeiltasten, Bild auf/ab und Pos1/Space. Um zu speichern, benötigst du einen Orb der Geborgenheit. Drücke V für Einstellungen, ESC für den Quest-Status und das Menü.\n\n",
   "Mova-se com o mouse, teclado numérico, qweadzxc, ou hjklyubn. Espere apertando 's' ou '.'. Gire o mundo com as setas, PageUp/Down, ou Home/Espaço.Para salvar o jogo você precisa de um Orbe da Segurança. Pressione 'v' para ir ao menu principal (configurações, modos especiais, etc.), ESC para o estado da missão.\n\n",
    }},
  {0x568338c0, { // "return"
   "powrót",
/*MISSING*/    "return",
   "návrat",
   "вернуться",
/*MISSING*/    "return",
/*MISSING*/    "return",
    }},
  {0x574701eb, { // "learn about hyperbolic geometry!"
   "dowiedz się więcej o geometrii hiperbolicznej!",
/*MISSING*/    "learn about hyperbolic geometry!",
   "nauč se o hyperbolické geometrii!",
   "узнай о гиперболической геометрии!",
/*MISSING*/    "learn about hyperbolic geometry!",
/*MISSING*/    "learn about hyperbolic geometry!",
    }},
  {0x5789e8a3, { // "Camera level above the plane"
   "Wysokość kamery nad płaszczyzną",
/*MISSING*/    "Camera level above the plane",
   "Výška kamery nad rovinou",
   "Высота камеры над плоскостью",
/*MISSING*/    "Camera level above the plane",
/*MISSING*/    "Camera level above the plane",
    }},
  {0x57b2222c, { // "Those roses smell too nicely. You have to come towards them."
   "Te róże tak ładnie pachną. Musisz iść w ich kierunku.",
/*MISSING*/    "Those roses smell too nicely. You have to come towards them.",
   "Ty růže příliš krásně voní. Musíš jít směrem k nim.",
   "Эти розы пахнут слишком прекрасно. Вы можете пойти лишь в сторону них.",
   "Diese Rosen duften zu gut - du musst zu ihnen gehen.",
/*MISSING*/    "Those roses smell too nicely. You have to come towards them.",
    }},
  {0x57d0af6b, { // "\"Straight lines stay close to each other forever, this is so romantic!\""
   "\"Linie proste, co mogą zawsze być blisko siebie... Jakie to romantyczne!\"",
   "\"Doğrular birbirlerine hep yakın, bu çok romantik!\"",
   "\"Přímky, které navěky zůstávají blízko sebe, to je tak romantické!\"",
   "\"Прямые линии остаются близко друг к другу... Как романтично!\"",
   "\"Grade Linien bleiben einander für immer nah, wie romantisch!\"",
/*MISSING*/    "\"Straight lines stay close to each other forever, this is so romantic!\"",
    }},
  {0x57f5e809, { // "Cultists throw fire at you from distance!"
   "Kultyści rzucają ogniem z odległości!",
/*MISSING*/    "Cultists throw fire at you from distance!",
   "Kultisté na tebe vrhají z dálky oheň!",
   "Культисты кидают огонь издалека!",
/*MISSING*/    "Cultists throw fire at you from distance!",
/*MISSING*/    "Cultists throw fire at you from distance!",
    }},
  {0x5807a2b6, { // "display hints"
   "pokaż wskazówki",
/*MISSING*/    "display hints",
   "zobrazovat tipy",
   "показать подсказки",
/*MISSING*/    "display hints",
/*MISSING*/    "display hints",
    }},
  {0x586df720, { // "\"There are %1 floor tiles inside our Table!\""
   "\"Nasz stół otacza %1 pól podłogi!\"",
   "\"Masamızın içinde %1 zemin hücresi var!\"",
   "\"Uvnitř našeho stolu je %1 dlaždic!\"",
   "\"Внутри стола %1 клеток!\"",
   "\"Es gibt %1 Bodenfliesen in unserer Tafel!\"",
/*MISSING*/    "\"There are %1 floor tiles inside our Table!\"",
    }},
  {0x589f402b, { // "ver"
   "wer",
   "ver",
   "ver",
   "версия",
   "ver",
   "ver",
    }},
  {0x58b8b50f, { // "this Orb appears on floors and is used by witches"
   "ta Sfera pojawia się na ziemi i jest używana przez czarownice",
   "Bu Küre zeminlerde bulunur ve cadılarca kullanılır.",
   "tato Sféra se objevuje na zemi a využívají ji čarodějnice czarownice",
   "эта сфера появляется на земле и используется ведьмами",
   "Dieser Orb erscheint auf dem Boden und wird von Hexen genutzt",
/*MISSING*/    "this Orb appears on floors and is used by witches",
    }},
  {0x58d882b1, { // "player 6 go"
   "gracz 6 idź",
/*MISSING*/    "player 6 go",
   "hráč 6 pohyb",
   "игрок 6 идёт",
   "Spieler 6 los",
/*MISSING*/    "player 6 go",
    }},
  {0x5918f7fc, { // "You have used less real time than ever before. Congratulations!"
   "Zajęło to mniej czasu rzeczywistego niż wcześniej. Gratulacje!",
   "Bu her zamankinden daha az zaman aldı! Tebrikler!",
   "Spotřebova%l0 jsi méně reálného času než kdy předtím. Gratulujeme!",
   "Вы потратили меньше реального времени, чем раньше. Поздравляем!",
   "Du hast weniger Echtzeit als je zuvor benötigt. Glückwunsch!",
   "Você fez seu melhor em menos tempo real. Parabéns!",
    }},
  {0x59236c0a, { // "u = undo"
   "u = cofnij",
   "u = geri al",
   "u = zpět",
   "u = отмена",
   "u = rückgängig",
/*MISSING*/    "u = undo",
    }},
  {0x592fd1b2, { // "You have to escape first!"
   "Musisz najpierw uciec!",
   "Önce kaçman gerekiyor!",
   "Nejprve musíš utéct!",
   "Вы должны бежать первым!",
   "Du musst zuerst entkommen!",
/*MISSING*/    "You have to escape first!",
    }},
  {0x5942af6b, { // "This menu can be also used to configure keys.\n\n"
   "To menu może być też użyte do konfiguracji klawiszy.",
/*MISSING*/    "This menu can be also used to configure keys.\n\n",
   "Toto menu je také možné použít pro konfiguraci kláves.\n\n",
   "В этом меню также можно поменять используемые клавиши\n\n.",
   "Dieses Menü kann auch verwendet werden, um Tasten zu konfigurieren.",
/*MISSING*/    "This menu can be also used to configure keys.\n\n",
    }},
  {0x59589c6e, { // "reset all configuration"
   "reset konfiguracji",
/*MISSING*/    "reset all configuration",
   "vyresetovat konfiguraci",
   "сбросить все настройки",
/*MISSING*/    "reset all configuration",
/*MISSING*/    "reset all configuration",
    }},
  {0x597d1896, { // "You are now a tortoise hero!"
   "Jesteś żółwim bohaterem!",
/*MISSING*/    "You are now a tortoise hero!",
   "Nyní jsi želví hrdina!",
   "Теперь вы -- черепаший герой!",
   "Du bist jetzt ein Held der Schildkröten!",
/*MISSING*/    "You are now a tortoise hero!",
    }},
  {0x59ba4407, { // "the native Orb of %the1"
   "rodzima sfera %abl1",
   "%abl1 doğal olarak bulunan Küre",
   "standardní Sféra %abl1",
   "родная сфера %abl1",
   "der heimische Orb %der1",
/*MISSING*/    "the native Orb of %the1",
    }},
  {0x59d5a94d, { // "zebra lines"
   "zebra: linie",
/*MISSING*/    "zebra lines",
   "zebra: čáry",
   "зебра: линии",
/*MISSING*/    "zebra lines",
/*MISSING*/    "zebra lines",
    }},
  {0x59e16a3b, { // "\"Find the Holy Grail to become one of us!\""
   "\"Znajdź Świętego Graala, by stać się jednym z nas!\"",
   "\"Bizden biri olmak için Kutsal Kâseyi bul!\"",
   "\"Najdi svatý grál a staneš se jedním z nás!\"",
   "\"Найди Святой Грааль, и станешь одним из нас!\"",
   "\"Finde den Heiligen Gral um einer von uns zu werden!\"",
/*MISSING*/    "\"Find the Holy Grail to become one of us!\"",
    }},
  {0x59e46df2, { // "Sleeping bulls wake up when you get into distance of two cells from them."
   "Byki się budzą, gdy się znajdziesz w promieniu dwóch pól od nich.",
/*MISSING*/    "Sleeping bulls wake up when you get into distance of two cells from them.",
   "Spící býci se probudí, když se přiblížíš na dvě políčka od nich.",
   "Спящий бык проснётся, если подойти к нему на расстояние двух клеток.",
/*MISSING*/    "Sleeping bulls wake up when you get into distance of two cells from them.",
/*MISSING*/    "Sleeping bulls wake up when you get into distance of two cells from them.",
    }},
  {0x59e8ecd2, { // "Players cannot get that far away!"
   "Gracze nie mogą się tak oddalać!",
/*MISSING*/    "Players cannot get that far away!",
   "Hráči se nemohou vzdálit tak daleko!",
   "Игроки не могут так сильно отдаляться друг от друга!",
   "Spieler dürfen sich nicht so weit voneinander entfernen!",
/*MISSING*/    "Players cannot get that far away!",
    }},
  {0x59e98e29, { // "summon lots of treasure"
   "przywołaj dużo skarbów",
   "çok hazine çıkar",
   "vyvolej spoustu pokladů",
   "призвать много сокровищ",
   "rufe viele Schätze herbei",
/*MISSING*/    "summon lots of treasure",
    }},
  {0x5a519933, { // "A Witch with a Winter spell. She is able to move through fire."
   "Czarownica z czarem Zimy. Może przechodzić przez ogień.",
   "Kış büyüsü olan bir cadı. Ateşin içinde yürüyebilir.",
   "Čarodějnice s kouzlem Zimy. Dokáže se pohybovat skrz oheň.",
   "Ведьма с заклинанием зимы. Может проходить сквозь огонь.",
   "Eine Frosthexe. Sie kann durch Feuer gehen.",
/*MISSING*/    "A Witch with a Winter spell. She is able to move through fire.",
    }},
  {0x5a54d807, { // "Those roses smell too nicely. You can only target cells closer to them!"
   "Te róże tak ładnie pachną. Możesz celować tylko w pola bliżej nich!",
/*MISSING*/    "Those roses smell too nicely. You can only target cells closer to them!",
   "Ty růže příliš krásně voní. Můžeš zacílovat jenom políčka, která k nim jsou blíž!",
   "Эти розы так прекрасно пахнут. Вы можете выбирать цель только ближе к ним!",
   "Diese Rosen duften zu gut. Du kannst nur Zellen näher bei ihnen anvisieren!",
/*MISSING*/    "Those roses smell too nicely. You can only target cells closer to them!",
    }},
  {0x5a59bb53, { // "%The1 squeaks with hope!"
   "%1 pisn%ął1 z nadzieją!",
   "%1 umutla ciyaklıyor!",
   "%1 nadějně piští!",
   "%1 пищит с надеждой!",
   "%Die1 %1 piepst hoffnungsvoll!",
/*MISSING*/    "%The1 squeaks with hope!",
    }},
  {0x5a5bf35e, { // "This forest is quite dry. Beware the bushfires!\nTrees catch fire on the next turn. The temperature of the grass cells rises once per turn for each fire nearby, and becomes fire itself when its temperature has risen 10 times.\nYou can also chop down the trees. Big trees take two turns to chop down."
   "Ta puszcza jest wyschnięta. Uważaj na pożary!\nSąsiednie drzewa zajmują się ogniem w następnej kolejce. Temperatura pola bez drzewa rośnie o 1 na kolejkę z każdym ogniem w sąsiedztwie i gdy wzrośnie do 10, to pole również staje się ogniem.\nMożesz też ścinać drzewa. Ścięcie dużego drzewa zajmuje dwie kolejki.",
   "Bu orman epey kuru. Çalı yangınlarına dikkat et!\nAğaçlar her tur ateşe yakalanır. Çimen hücrelerinin ısısı her tur yakındaki ateşle bir artar ve sonra ısısı 10 kez yükselirse, kendisi ateş olur.\n Ağaçları kesebilirsin. Büyük ağaçları kesmesi 2 tur alır.\n ",
   "Tento hvozd je velice suchý. Pozor na požáry!\nStromy začnou hořet jedno kolo poté, co se ocitnou v sousedství ohně. Teplota políček s trávou vzrůstá každé kolo o 1 za každé sousední políčko s ohněm a poté, co vzroste desetkrát, začne tráva hořet.\nStromy v Suchém hvozdu můžeš kácet. Pokácení velkých stromů trvá dvě kola.",
   "Этот лес довольно сухой. Берегитесь лесных пожаров!\nДеревья загораются от огня на следующем шагу. Температура травы растёт на каждом шаге, после 10 шагов трава загорается.\nТакже Вы можете рубить деревья. Большие деревья требуют двух ударов.",
   "Dieser Wald ist ziemlich trocken. Achtung - Waldbrände!\nBäume fangen im nächsten Zug Feuer. Die Temperatur von Gras steigt jeden Zug für jedes Feuer in der Nähe, und wird selbst zu Feuer wenn dies 10-mal geschehen ist.\nDu kannst auch Bäume fällen. Große Bäume zu fällen dauert zwei Züge.",
/*MISSING*/    "This forest is quite dry. Beware the bushfires!\nTrees catch fire on the next turn. The temperature of the grass cells rises once per turn for each fire nearby, and becomes fire itself when its temperature has risen 10 times.\nYou can also chop down the trees. Big trees take two turns to chop down.",
    }},
  {0x5a755a19, { // "pan right"
   "przewiń w prawo",
   "sağa kaydır",
   "scrollování doprava",
   "крутить вправо",
   "Nach rechts schwenken",
/*MISSING*/    "pan right",
    }},
  {0x5a90f615, { // "static mode"
   "tryb statyczny",
/*MISSING*/    "static mode",
   "statický mód",
   "статичный режим",
/*MISSING*/    "static mode",
/*MISSING*/    "static mode",
    }},
  {0x5ac56465, { // "Ever wondered how some boardgame would look on the hyperbolic plane? I wondered about Go, so I have created this feature. Now you can try yourself!\nEnter = pick up an item (and score), space = clear an item\nOther keys place orbs and terrain features of various kinds\nIn the periodic editor, press 0-4 to switch walls in different ways\n"
   "Zastanawia%łeś0 się może, jak jakaś gra planszowa wyglądałaby na płaszczyźnie hiperbolicznej? Ja się zastanawiałem nad Go i postanowiłem to sprawdzić. Teraz Ty też możesz sprawdzić!\nEnter = podnieś przedmiot, space = zabierz przedmiot\nPozostałe klawisze kładą różnego rodzaju przedmioty\nW edytorze okresowym, klawisze 0-4 zmieniają ściany na różne sposoby",
   "Hiç bir oyunun hiperbolik düzlemde nasıl görüneceğini merak ettin mi? Ben Go hakkında merak ettim ve bu özelliği geliştirdim. Sen de deneyebilirsin!\nEnter = bir itemi al (ve puan kazan), space = bir itemi temizle\nDiğer tuşlar küreler ve çeşitli arazi özellikleri eklemek için.\n",
   "Přemýšle%l0 jsi někdy o tom, jak by vypadala desková hra na hyperbolické rovině? Mě zajímalo Go, a tak jsem vytvořil tuhle funkci. Teď si to můžeš vyzkoušet i ty!\nEnter = seber předmět (a skóre), mezerník = vymaž předmět\nDalší klávesy umísťují sféry a nejrůznější terénní prvky\nStiskem kláves 0-4 můžete různě přepínat zdi\n",
   "Вы когда-нибудь думали о том, как выглядит настольная игра на гиперболической плоскости? Я думал про Го и создал этот режим. Теперь и Вы можете попробовать!\nEnter = поднять предмет, space = удалить предмет\nОстальные клавиши добавляют разные предметы или элементы ландшафта.\nВ периодическом редакторе клавиши 0-4 меняют местоположение стен разными способами.\n",
   "Hast du dich je gewundert, wie ein Brettspiel auf einer hyperbolischen Ebene aussehen würde? Ich habe mich über Go gewundert und dieses Feature erstellt. Probiere es selbst!\nEnter = Gegenstand aufheben (und punkte), [LEER] = Gegenstand löschen\nAndere Tasten platzieren Orbs, Terrain und Anderes\nIm periodischen Editor kannst du 0-4 verwenden um die Wände zu verändern\n",
/*MISSING*/    "Ever wondered how some boardgame would look on the hyperbolic plane? I wondered about Go, so I have created this feature. Now you can try yourself!\nEnter = pick up an item (and score), space = clear an item\nOther keys place orbs and terrain features of various kinds\nIn the periodic editor, press 0-4 to switch walls in different ways\n",
    }},
  {0x5aecc3d3, { // "--- and teleport there"
   "--- i się teleportuj",
   "--- ve oraya ışınlan.",
   "--- a teleportuj se tam",
   "--- и телепортироваться туда",
   "--- und teleportiere dorthin",
/*MISSING*/    "--- and teleport there",
    }},
  {0x5aeed4d9, { // "You mirror %the1."
   "Odbijasz %a1.",
/*MISSING*/    "You mirror %the1.",
   "Zrcadlíš %a1.",
   "Ты отразил %a1.",
/*MISSING*/    "You mirror %the1.",
/*MISSING*/    "You mirror %the1.",
    }},
  {0x5b5a7df7, { // "TRANSLATIONWARNING2"
/*MISSING*/    "TRANSLATIONWARNING2",
   "çevirmek isterseniz lütfen zeno@attnam.com adresinden benimle iletişime geçin.",
/*MISSING*/    "TRANSLATIONWARNING2",
/*MISSING*/    "TRANSLATIONWARNING2",
/*MISSING*/    "TRANSLATIONWARNING2",
/*MISSING*/    "TRANSLATIONWARNING2",
    }},
  {0x5b8d82af, { // "What is freedom for you? A situation when you can walk wherever you want? Or a situation when you do not have to work, since you have as much tasty food as you want?\n\nWell, this creature has chosen the second option. It won't be happy if you destroy its prison.\n"
   "Czym jest dla Ciebie wolność? Sytuacją, gdy możesz chodzić, gdzie chcesz? Czy sytuacją, gdy nie musisz pracować, bo masz tyle jedzenia, ile zapragniesz?\n\nTo stworzenie wybrało drugą opcję. Nie będzie szczęśliwe, jak je uwolnisz z więzienia.",
/*MISSING*/    "What is freedom for you? A situation when you can walk wherever you want? Or a situation when you do not have to work, since you have as much tasty food as you want?\n\nWell, this creature has chosen the second option. It won't be happy if you destroy its prison.\n",
   "Co je pro tebe svoboda? To, že si můžeš chodit, kam chceš? Nebo to, že nemusíš pracovat, protože máš tolik dobrého jídla, kolik chceš?\n\nInu, tento tvor si vybral tu druhou možnost. Nebude rád, když zničíš jeho vězení.\n",
   "Что для Вас есть свобода? Когда можно идти, куда хочешь? Или когда не надо работать, так как есть столько вкусной еды, сколько хочешь?\n\nЭто существо выбрало второй вариант. Лис не будет рад, если Вы разрушите его тюрьму.\n",
   "Was bedeutet Freiheit für dich? Eine Situation in der du gehen kannst wohin immer du willst? Oder eine Situation in der du nicht arbeiten musst und immer soviel leckeres Essen hast, wie du willst?\n\nWie auch immer, dieses Wesen wählte die zweite Option. Es wird nicht sehr glücklich sein, wenn du sein Gefängnis zerstörst.",
/*MISSING*/    "What is freedom for you? A situation when you can walk wherever you want? Or a situation when you do not have to work, since you have as much tasty food as you want?\n\nWell, this creature has chosen the second option. It won't be happy if you destroy its prison.\n",
    }},
  {0x5bb92408, { // " (E:%1)"
   " (E:%1)",
   " (Ö:%1)",
   " (E:%1)",
   " (E:%1)",
   " (E:%1)",
/*MISSING*/    " (E:%1)",
    }},
  {0x5bd97cf6, { // "Activated Orb of Safety!"
   "Aktywowano Sferę Bezpieczeństwa!",
   "Güvenlik Küresini aktifleştirdin.",
   "Aktivova%l0 jsi Sféru bezpečí!",
   "Активирована сфера безопасности!",
   "Orb der Geborgenheit aktiviert!",
   "Orbe de Segurança ativado!",
    }},
  {0x5bdfa491, { // "A big monster from the Living Caves. A dead Troll will be reunited with the rocks, causing some walls to grow around its body."
   "Duży stwór z Żyjących Jaskiń. Martwy Troll połączy się ze skałą, powodując rozrost skał wokół jego ciała.",
   "Yaşayan Mağaralardan büyük bir canavar. Ölü bir Trol etrafında yükselen duvarlarla kayalarla yeniden bütünleşir. ",
   "Velký netvor z Živoucích jeskyní. Mrtvý Šutroll splyne se skálou a kolem jeho těla se rozrostou zdi.",
   "Огромный монстр из Живых пещер. Мёртвый тролль окружается камнями, вызывая рост скал вокруг себя.",
   "Ein großes Monster aus den Lebenden Höhlen. Ein toter Troll wird eins mit dem Gestein, was dazu führt, dass er von Wänden umgeben wird.",
/*MISSING*/    "A big monster from the Living Caves. A dead Troll will be reunited with the rocks, causing some walls to grow around its body.",
    }},
  {0x5bee3d6e, { // "Two players cannot move/attack the same location!"
   "Dwóch graczy nie może celować w to samo miejsce!",
/*MISSING*/    "Two players cannot move/attack the same location!",
   "Dva hráči nemohou zacílovat stejnou lokaci!",
   "Два игрока не могут ходить в одну клетку!",
   "Zwei Spieler können nicht dieselbe Stelle anvisieren!",
/*MISSING*/    "Two players cannot move/attack the same location!",
    }},
  {0x5bf02e45, { // "On the following slide, the colors change smoothly in the whole infinite world. Again, this works better than in Euclidean geometry."
   "Na kolejnym slajdzie kolory się płynnie zmieniają w nieskończonym świecie. To również działa lepiej niż w geometrii euklidesowej.",
/*MISSING*/    "On the following slide, the colors change smoothly in the whole infinite world. Again, this works better than in Euclidean geometry.",
   "Na následujícím snímku se barvy v celém nekonečném světě hladce mění. I v tomto případě to v hyperbolické geometrii funguje lépe než v eukleidovské.",
   "На следующем слайде цвета изменяются плавно по всему бесконечному миру. Опять же, это работает лучше, чем в евклидовой геометрии.",
/*MISSING*/    "On the following slide, the colors change smoothly in the whole infinite world. Again, this works better than in Euclidean geometry.",
/*MISSING*/    "On the following slide, the colors change smoothly in the whole infinite world. Again, this works better than in Euclidean geometry.",
    }},
  {0x5c032e2e, { // "Welcome to the Euclidean mode!"
   "Witaj w trybie euklidesowym!",
/*MISSING*/    "Welcome to the Euclidean mode!",
   "Vítej v eukleidovském módu!",
   "Добро пожаловать в евклидов режим!",
/*MISSING*/    "Welcome to the Euclidean mode!",
/*MISSING*/    "Welcome to the Euclidean mode!",
    }},
  {0x5c1d1ce9, { // "A kind of Troll native to the Red Rock Valley."
   "Ten rodzaj Trolla mieszka w Dolinie Czerwonej Skały.",
   "Kızılkaya Vadisinde yaşayan bir trol.",
   "Typ trolla, který se vyskytuje v Údolí červených skal.",
   "Этот вид троллей обитает в Долине Красных Скал.",
   "Eine spezielle Trollgattung, die im Rotfelsental lebt.",
/*MISSING*/    "A kind of Troll native to the Red Rock Valley.",
    }},
  {0x5c4fe15c, { // "r: %1"
   "r: %1",
   "r: %1",
   "r: %1",
   "r: %1",
   "r: %1",
/*MISSING*/    "r: %1",
    }},
  {0x5c61b7e8, { // "Two of your images crash and disappear!"
   "Dwa Twoje obrazy wpadły na siebie i znikły!",
   "İki yansıma çarpıştı ve yok oldu!",
   "Dva tvoje obrazy se srazily a zmizely!",
   "Два Ваших двойника столкнулись и уничтожились!",
   "Zwei deiner Abbilder stoßen zusammen und verschwinden!",
   "Duas de suas imagens se chocam e desaparecem!",
    }},
  {0x5c8258dc, { // "Cannot teleport on an item!"
   "Nie możesz teleportować się na przedmiot!",
   "Bir eşyanın üzerine ışınlanamazsın!",
   "Nemůžeš se teleportovat na předmět!",
   "Нельзя телепортироваться на предмет!",
   "Du kannst dich nicht auf Gegenstände teleportieren!",
/*MISSING*/    "Cannot teleport on an item!",
    }},
  {0x5c89219a, { // "switch web display"
   "przestaw strukturę",
/*MISSING*/    "switch web display",
   "přepni strukturu",
   "переключить отображение сетки",
   "wechsle die Web-Anzeige",
/*MISSING*/    "switch web display",
    }},
  {0x5cf476b5, { // "random colors"
   "losowe kolory",
   "rastgele renkler",
   "náhodné barvy",
   "случайные цвета",
   "zufällige Farben",
/*MISSING*/    "random colors",
    }},
  {0x5d483ee2, { // "QUEST"
   "MISJA",
   "GÖREV",
   "ÚKOLY",
   "МИССИЯ",
   "QUEST",
/*MISSING*/    "QUEST",
    }},
  {0x5d4e44de, { // "too dangerous to use in %the1"
   "zbyt niebezpieczna %abl1",
   "zbyt niebezpieczna %abl1",
   "příliš nebezpečná %abl1",
   "слишком опасна %abl1",
   "zu gefährlich um es in %der1 zu nutzen",
/*MISSING*/    "too dangerous to use in %the1",
    }},
  {0x5d4e8ff4, { // "Dead floor, with some rubble."
   "Martwa ziemia, jest tu trochę gruzu.",
   "Ölü zemin, üstünde taş parçaları var.",
   "Mrtvá podlaha se sutinami.",
   "Мёртвая земля с щебнем.",
   "Toter Boden, mit etwas Geröll.",
/*MISSING*/    "Dead floor, with some rubble.",
    }},
  {0x5d8c000c, { // "A typical Graveyard monster."
   "Typowy cmentarny stwór.",
   "Mezarlıkta sık rastlanan bir canavar.",
   "Typický hřbitovní netvor.",
   "Типичный для Кладбища монстр.",
   "Ein typisches Friedhofsmonster.",
/*MISSING*/    "A typical Graveyard monster.",
    }},
  {0x5d93351f, { // "cancel move"
   "anuluj ruch",
/*MISSING*/    "cancel move",
   "zruš pohyb",
   "отмена хода",
/*MISSING*/    "cancel move",
/*MISSING*/    "cancel move",
    }},
  {0x5d94df20, { // "%The1 kisses you, and begs you to bring %him1 away from here."
   "%1 Cię pocałowa%ł1 i poprosi%ł1, byś %go1 stąd zabra%ł0.",
   "%1 seni öpüyor ve sana %go1 buradan uzaklaştırman için yalvarıyor.",
   "%1 tě políbi%l1 a prosí tě, abys ji odsud odved%l0.",
   "%1 поцеловал%E1 тебя, и попросил%E1 увести отсюда.",
   "%Der1 %1 küsst dich und bittet dich %den1 von hier weg zu bringen.",
/*MISSING*/    "%The1 kisses you, and begs you to bring %him1 away from here.",
    }},
  {0x5dee2673, { // "The Elemental Planes are divided into four subzones: Planes of Fire, Water, Air, and Earth. You need to collect a Shard from each Plane to construct an Elemental Gem. It is dangerous to collect too many Shards of the same type without constructing a Gem."
   "Strefa Żywiołów dzieli się na cztery podstrefy: Strefa Ognia, Wody, Powietrza i Ziemi. Musisz znaleźć Okruch każdego żywiołu, by skonstruować Kamień Żywiołów. Jest niebezpiecznie zbierać wiele Okruchów tego samego żywiołu bez konstruowania Kamienia.",
   "Unsur Düzlemleri dört altalana ayrılır: Ateş, Su, Hava ve Toprak Düzlemleri. Bir Unsurtaşı yapmak için her dört düzlemden birer parça toplaman gerekiyor. Bir Unsurtaşı yapmadan aynı cins parçadan çok sayıda toplamak tehlikelidir.",
   "Elementální roviny se dělí na čtyři subzóny: Rovinu Ohně, Vody, Vzduchu a Země. Musíš v každé z Rovin najít Úlomek, ze kterých pak vytvoříš Elementální drahokam. Pozor, je nebezpečné mít příliš mnoho Úlomků stejného typu a nepoužít je k sestavení Drahokamu!",
   "Есть четыре Плоскости Стихий: Плоскость Огня, Воды, Воздуха и Земли. Вам нужно собрать по Осколку из каждой, чтобы собрать камень стихий. Собирать много осколков, не собирая Камень, довольно опасно.",
   "Die Elementare Ebene ist in 4 Zonen unterteilt: Die Feuerebene, die Erdebene, die Luftebene und die Wasserebene. Du brauchst eine Scherbe aus jeder Ebene um ein Elementjuwel herzustellen. Es ist äußerst gefährlich mehrere Scherben desselben Typs zu sammeln ohne ein Elementjuwel herzustellen.",
/*MISSING*/    "The Elemental Planes are divided into four subzones: Planes of Fire, Water, Air, and Earth. You need to collect a Shard from each Plane to construct an Elemental Gem. It is dangerous to collect too many Shards of the same type without constructing a Gem.",
    }},
  {0x5e0a997c, { // "peaceful mode"
   "tryb spokojny",
/*MISSING*/    "peaceful mode",
   "klidný mód",
   "мирный режим",
/*MISSING*/    "peaceful mode",
/*MISSING*/    "peaceful mode",
    }},
  {0x5e1ee8f2, { // "boundary"
   "brzeg",
   "sınır",
   "hranice",
   "граница",
   "Grenze",
/*MISSING*/    "boundary",
    }},
  {0x5e9ce5c0, { // "This orb can be used to move faster for some time."
   "Ta sfera powoduje, że przez pewien czas poruszasz się szybciej.",
   "Bu küre bir süreliğine daha hızlı hareket etmeye yarar.",
   "Tato sféra ti umožňuje pohybovat se po určitou dobu rychleji.",
   "Эта сфера ускорит Вас на некоторое время.",
   "Du kannst diesen Orb verwenden, um dich für einige Zeit schneller zu bewegen.",
/*MISSING*/    "This orb can be used to move faster for some time.",
    }},
  {0x5ef47a97, { // "Your score: %1"
   "Twój wynik: %1",
   "Puanın: %1",
   "Tvé skóre: %1",
   "Ваш счёт: %1",
   "Punkte: %1",
   "Sua pontuação: %1",
    }},
  {0x5f21e2d7, { // "%The1 activates her Flash spell!"
   "%1 rzuci%ł1 swój czar Błysk!",
   "%1 Işık büyüsünü aktifleştirdi!",
   "%1 aktivova%l1 kouzlo Záblesk!",
   "%1 активировал%E1 заклинание вспышки!",
   "%Der1 %1 aktiviert den Blitzzauber!",
/*MISSING*/    "%The1 activates her Flash spell!",
    }},
  {0x5f22ded5, { // "These star-shaped flowers native to Hell are a valuable alchemical component."
   "Te piekielne rośliny w kształcie gwiazdy są cennym składnikiem alchemicznym.",
   "Cehennemde yetişen bu yıldız şekilli çiçekler değerli bir simya bileşeni.",
   "Tyto hvězdám podobné květiny z Pekla jsou cennou alchymistickou surovinou.",
   "Эти звездовидные цветы их Ада - важный алхимический компонент.",
   "Diese sternförmigen Blumen aus der Hölle sind eine wertvolle alchemistische Zutat.",
/*MISSING*/    "These star-shaped flowers native to Hell are a valuable alchemical component.",
    }},
  {0x5f53f87c, { // "press 0 to leave this mode"
   "Wciśnij 0, by opuścić ten tryb",
   "Bu moddan çıkmak için 0'a bas",
   "Pro opuštění tohoto módu stiskni klávesu '0'",
   "Нажмите 0, чтобы выйти из режима",
   "Drücke 0 um diesen Modus zu verlassen",
/*MISSING*/    "press 0 to leave this mode",
    }},
  {0x5f8ef306, { // "Another Dead Orb."
   "Jeszcze jedna Martwa Sfera.",
   "Başka bir Ölü Küre.",
   "Další mrtvá sféra.",
   "Ещё одна мёртвая сфера.",
   "Ein weiterer toter Orb.",
   "Outro Orbe Morto.",
    }},
  {0x5facfa66, { // "zebra pattern"
   "wzór zebry",
   "zebra dokusu",
   "vzor 'zebra'",
   "ущор зебры",
   "Zebramuster",
/*MISSING*/    "zebra pattern",
    }},
  {0x5fc7252e, { // "cat"
   "kot",
/*MISSING*/    "cat",
   "kočka",
   "кот",
   "Katze",
/*MISSING*/    "cat",
    }},
  {0x5ff087bd, { // "You have improved %1 of your specific high scores!"
   "Poprawi%łeś0 %1 z wyników szczegółowych!",
   "Tekil yüksek skorlarından %1 tanesini artırdın !",
   "Zlepši%l0 jsi %1 svých specifických nejvyšších skóre!",
   "Вы увеличили %1 отдельных результатов!!",
   "Du hast %1 deiner spezifischen Highscores verbessert!",
   "Você superou %1 das suas melhores pontuações específicas!",
    }},
  {0x5ff8756f, { // "There are Krakens in your homeland too... huge sea monsters which could easily destroy ships. The geometry of this strange world prevents quick movement of huge objects, so there are no large ships, only small boats, and hyperbolic Krakens are relatively small too. Still, you suppose they might be the widest creatures which could still move at considerable speed...\n\nKraken heads can move only on hexagons. You need to attack all the tentacles to kill the Kraken. A tentacle cannot attack if it has been attacked on the same turn. When a Kraken attacks you while you are in a boat, it destroys the boat, but does not kill you."
   "Tam, skąd pochodzisz, również są Krakeny... wielkie morskie potwory, które z łatwością mogą zatapiać statki. Geometria tego dziwnego świata powoduje, że wielkie obiekty nie mogą się szybko poruszać, nie ma tu zatem dużych statków, tylko małe łódki, i hiperboliczne Krakeny też są względnie małe. Wciąż one są chyba największymi stworzeniami, które wciąż mogą się poruszać z rozsądną prędkością...\n\nGłowa krakena porusza się tylko po sześciokątach. By pokonać Krakena, musisz zaatakować jego wszystkie macki. Macka nie atakuje, jeśli została zaatakowana w tej turze. Jeśli Kraken zaatakuje Cię, gdy jesteś na statku, zniszczy statek, ale Cię nie zabije.",
/*MISSING*/    "There are Krakens in your homeland too... huge sea monsters which could easily destroy ships. The geometry of this strange world prevents quick movement of huge objects, so there are no large ships, only small boats, and hyperbolic Krakens are relatively small too. Still, you suppose they might be the widest creatures which could still move at considerable speed...\n\nKraken heads can move only on hexagons. You need to attack all the tentacles to kill the Kraken. A tentacle cannot attack if it has been attacked on the same turn. When a Kraken attacks you while you are in a boat, it destroys the boat, but does not kill you.",
   "I ve tvé domovině existují Krakeni... obrovské mořské obludy, které dokážou snadno zničit lodě. Geometrie tohoto podivného světa brání rychlému pohybu velkých objektů, a tak tu nejsou žádné velké lodě, jen malé loďky, a i hyperboličtí Krakeni jsou relativně malí. I tak jsou to možná nejširší tvorové, kteří se stále dokážou pohybovat nějakou slušnou rychlostí...\n\nKrakení hlava se může pohybovat pouze po šestiúhelníkových políčcích. Abys Krakena zabil, musíš zasáhnout všechna jeho chapadla. Chapadlo nemůže zaútočit, pokud jsi ve stejném kole zaútočil ty na něj. Pokud jsi na loďce, Krakenův útok ji zničí, ale ty zůstaneš naživu.",
   "На вашей родине тоже есть кракены... огромные морские существа, с лёгкостью уничтожающие большие корабли. Геометрия этого странного мира не позволяет большим объектам быстро двигаться, поэтому здесь нет больших кораблей, только маленькие лодки, и гиперболические кракены тоже довольно малы. Вероятно, это самые большие существа, способные двигаться с приличной скоростью...\n\nГолова кракена движется только по шестиугольникам. Чтобы убить кракена, нужно атаковать все его щупальца. Щупальце не может бить, если оно атаковано в этот же ход. Если кракен атакует вас, когда вы в лодке, он уничтожает лодку, но не бьёт вас.",
   "In deiner Heimat gibt es auch Kraken - riesige Seeungeheuer, die leicht Schiffe zerstören könnten. Die Geometrie dieser seltsamen Welt verhindert schnelle Bewegungen von großen Objekten. Daher gibt es keine großen Schiffe, nur kleine Boote, und hyperbolische Kraken sind auch relativ klein. Dennoch - du nimmst an, das sie die ausgedehntesten Kreaturen sind, die sich noch mit größerer Geschwindigkeit bewegen können...\n\nKöpfe von Kraken können sich nur auf Sechsecken bewegen. Du musst alle Tentakel angreifen, um einen Kraken zu töten. Ein Tentakel kann nicht angreifen, wenn er im gleichen Zug attackiert wurde. Wenn ein Krake dich angreift, während du in einem Boot bist, zerstört er das Boot, tötet dich jedoch nicht.",
/*MISSING*/    "There are Krakens in your homeland too... huge sea monsters which could easily destroy ships. The geometry of this strange world prevents quick movement of huge objects, so there are no large ships, only small boats, and hyperbolic Krakens are relatively small too. Still, you suppose they might be the widest creatures which could still move at considerable speed...\n\nKraken heads can move only on hexagons. You need to attack all the tentacles to kill the Kraken. A tentacle cannot attack if it has been attacked on the same turn. When a Kraken attacks you while you are in a boat, it destroys the boat, but does not kill you.",
    }},
  {0x6038e891, { // "Simply a wall. Mostly."
   "Po prostu ściana. W zasadzie.",
   "Yalnızca bir duvar. Büyük ihtimalle.",
   "Po prostu ściana. W zasadzie.",
   "Обычная стена. В основном.",
   "Eine einfache Wand. Größtenteils.",
/*MISSING*/    "Simply a wall. Mostly.",
    }},
  {0x60b508ab, { // "throw at the pointer"
   "rzut we wskaźnik",
   "mausa doğru at",
   "házení na kurzor",
   "кинуть нож в цель",
   "Zum Zeiger werfen",
/*MISSING*/    "throw at the pointer",
    }},
  {0x60f5c677, { // "A Witch with a Fire spell. She will leave a trail of fire behind her."
   "Czarownica z czarem Ognia. Zostawia za sobą płonącą ścieżkę.",
   "Ateş büyüsü olan bir Cadı. Arkasında ateşten bir yol bırakır.",
   "Čarodějnice s kouzlem Ohně. Zanechává za sebou ohnivou stopu.",
   "Ведьма с заклинанием огня. Оставляет огненный след за собой.",
   "Eine Feuerhexe. Sie hinterlässt eine Spur aus Feuer.",
/*MISSING*/    "A Witch with a Fire spell. She will leave a trail of fire behind her.",
    }},
  {0x61292603, { // "You glance at your precious treasures."
   "Patrzysz na swoje drogocenne skarby.",
   "Değerli hazinelerine şöyle bir baktın.",
   "Díváš se na své převzácné poklady.",
   "Вы смотрите на ваши драгоценные сокровища.",
   "Du betrachtest deine kostbaren Schätze.",
   "Você contempla seus preciosos tesouros.",
    }},
  {0x6133930c, { // "gain Orb of Yendor"
   "zdobądź sferę Yendoru",
/*MISSING*/    "gain Orb of Yendor",
   "získej Yendorskou sféru",
   "получить сферу Йендора",
   "erhalte Orb von Yendor",
/*MISSING*/    "gain Orb of Yendor",
    }},
  {0x616acab2, { // "This Orb allows you to find new lands more easily. Lands where you have already collected less treasure, and especially the Crossroads, are more likely to spawn while you have this. Additionally, Orbs of Safety are more likely to spawn in the Whirlpool."
   "Ta sfera pozwala łatwo znajdować nowe krainy. Gdy ją masz, łatwiej znaleźć jest krainy, w których zebrane zostało mniej skarbów, w szczególności Skrzyżowania. Dodatkowo, Sfery Bezpieczeństwa częściej trafiają się w Wirze.",
   "Bu küre yeni diyarlar bulmanı kolaylaştırır. Daha az hazine topladığın diyarlar ve özellikle Arayollar, bu küreyi elinde tutarken daha yüksek ihtimalle karşına çıkar. Ayrıca, Güvenlik Kürelerinin Tayfun'da ortaya çıkma ihtimalini artırır.",
   "Tato Sféra ti umožňuje snadněji najít nové kraje. Dokud ji máš, budou se s větší pravděpodobností generovat kraje, ve kterých jsi získal méně pokladů, zvláště pak Křižovatky Kromě toho se ve Víru bude generovat více Sfér Bezpečí.",
   "Эта сфера позволяет с лёгкостью находить новые земли. Земли, в которых Вы собрали мало сокровищ, и особенно Перекрёстки, появляются чаще, пока у Вас есть эта сфера. Кроме того, сферы Безопасности чаще находятся в Водовороте.",
   "Dieser Orb erlaubt es dir neue Länder leichter zu entdecken. Länder in denen du bisher weniger Schätze gesammelt hast und besonders Kreuzungen sind wahrscheinlicher anzutreffen so lange du ihn besitzt. Des weiteren sind Orbs der Geborgenheit wahrscheinlicher im Strudel zu finden.",
/*MISSING*/    "This Orb allows you to find new lands more easily. Lands where you have already collected less treasure, and especially the Crossroads, are more likely to spawn while you have this. Additionally, Orbs of Safety are more likely to spawn in the Whirlpool.",
    }},
  {0x61bb06c5, { // "You feel the presence of free saves on the Crossroads."
   "Wyczuwasz bezpieczeństwo na Skrzyżowaniach.",
/*MISSING*/    "You feel the presence of free saves on the Crossroads.",
   "Cítíš, že na Křižovatce si můžeš zdarma uložit hru.",
   "Ты чувствуешь, что можешь безопасно сохраниться на перекрёстке.",
/*MISSING*/    "You feel the presence of free saves on the Crossroads.",
/*MISSING*/    "You feel the presence of free saves on the Crossroads.",
    }},
  {0x61be9d87, { // "or 'r' or F5 to restart"
   "lub 'r' lub F5, by zacząć od początku",
   "'r' veya F5 ile yeniden başlat",
   "'r' nebo F5: začít znovu od začátku",
   "либо 'r' или F5, чтобы начать заново",
   "oder 'r' oder F5 zum Neustarten",
   "ou 'r' ou F5 para reiniciar",
    }},
  {0x61dbe2db, { // "center the map on me"
   "centruj mapę na mnie",
   "centruj mapę na mnie",
   "vycentrování mapy na postavě",
   "центрировать карту по мне",
   "Karte selbstzentrieren",
/*MISSING*/    "center the map on me",
    }},
  {0x61e70a0c, { // "A simple, but effective, missile, used by rogues."
   "Prosta choć efektywna broń rzucana, używana przez poszukiwaczy skarbów.",
   "Seyyahlar tarafından kullanılan basit ama etkili bir silah.",
   "Prostá, ale efektivní vrhací zbraň, kterou používají lovci pokladů.",
   "Простой, но эффективный снаряд, используемый разбойниками.",
   "Eine einfache, aber effektive, Wurfwaffe. Schurken verwenden sie.",
/*MISSING*/    "A simple, but effective, missile, used by rogues.",
    }},
  {0x61fe2a59, { // "sound effects volume"
   "głośność efektów dźwiękowych",
/*MISSING*/    "sound effects volume",
   "hlasitost zvukových efektù",
   "głośność efektów dźwiękowych",
/*MISSING*/    "sound effects volume",
/*MISSING*/    "sound effects volume",
    }},
  {0x624e89f5, { // "%The1 eats %the2!"
   "%1 zjad%ł1 %a2!",
   "%1 %a2 yedi!",
   "%1 sežra%l1 %a2!",
   "%1 съел%E1 %a2!",
   "%Der1 %1 frisst %den2 %a2!",
   "%na1 come %na2!",
    }},
  {0x625c4a3c, { // "PLAY"
   "GRA",
   "OYNA",
   "HRA",
   "ИГРА",
   "START",
   "JOGAR",
    }},
#if REPEATED
  {0x625c4a3d, { // "PLAY [repeat]"
   "GRAJ",
   "OYNA",
   "HRAJ",
   "ИГРА",
   "START",
/*MISSING*/    "PLAY [repeat]",
    }},
#endif
#if REPEATED
  {0x625c4a3e, { // "PLAY [repeat] [repeat]"
   "GRAJ",
/*MISSING*/    "PLAY [repeat] [repeat]",
   "HRÁT",
   "ИГРАЙ",
/*MISSING*/    "PLAY [repeat] [repeat]",
/*MISSING*/    "PLAY [repeat] [repeat]",
    }},
#endif
  {0x62dd148b, { // "The height of walls, in absolute units. For the current values of g and c, wall height of %1 absolute units corresponds to projection value of %2."
   "Wysokość ścian. Przy obecnych wartościach g i c, góra ściany na wysokości %1 odpowiada współczynnikowi projekcji %2.",
/*MISSING*/    "The height of walls, in absolute units. For the current values of g and c, wall height of %1 absolute units corresponds to projection value of %2.",
   "Výška zdí v absolutních jednotkách. Při současných hodnotách g a c odpovídá výška %1 absolutních jednotek projekční hodnotě %2.",
   "Высота стен. При данных значениях g и c стена высоты в %1 единиц соответствует расстоянию на проекции в %2.",
/*MISSING*/    "The height of walls, in absolute units. For the current values of g and c, wall height of %1 absolute units corresponds to projection value of %2.",
/*MISSING*/    "The height of walls, in absolute units. For the current values of g and c, wall height of %1 absolute units corresponds to projection value of %2.",
    }},
  {0x62f1f1e5, { // "Push Skeletons into the holes!"
   "Wpychaj Szkielety do dziur!",
/*MISSING*/    "Push Skeletons into the holes!",
   "Strč Kostlivce do děr!",
   "Сталкивайте скелетов в дырки!",
/*MISSING*/    "Push Skeletons into the holes!",
/*MISSING*/    "Push Skeletons into the holes!",
    }},
  {0x631d6ba2, { // "Revived!"
   "Wskrzeszono!",
/*MISSING*/    "Revived!",
   "Wskrzeszono!",
   "Воскрешено!",
   "Wiederbelebt!",
/*MISSING*/    "Revived!",
    }},
  {0x63355b1b, { // "You will need more experience to defeat the Greater Demon!"
   "Musisz zdobyć doświadczenie, by pokonać Większego Demona!",
/*MISSING*/    "You will need more experience to defeat the Greater Demon!",
   "Na poražení Ďábla budeš potřebovat víc zkušeností!",
   "Вам нужно больше опыта, чтобы победить Старшего демона!",
/*MISSING*/    "You will need more experience to defeat the Greater Demon!",
/*MISSING*/    "You will need more experience to defeat the Greater Demon!",
    }},
  {0x6342c3b9, { // "l = lands"
   "c = kopia",
   "c = diyar",
   "l = kraje",
   "c = копия",
   "c = Länder",
/*MISSING*/    "l = lands",
    }},
  {0x6373be2b, { // "Circles are strange in hyperbolic geometry too. Look for the Castle of Camelot in the Crossroads; the Round Table inside is a circle of radius 28. Finding its center is a difficult challenge.\n\nPress '5' to cheat by seeing the smaller circles too.\n\nNote: Camelot and some other lands are unlocked earlier in the Tutorial than in a real game."
   "Koła również są dziwne w geometrii hiperbolicznej. Szukaj zamku Camelot na Skrzyżowaniu; Okrągły Stół jest kołem o promieniu 28. Znalezienie jego środka to trudne wyzwanie.\n\nNaciśnij '5' by oszukać -- zobaczyć mniejsze okręgi.\n\nUwaga: Camelot i niektóre inne krainy w tym podręczniku są pokazywane wcześniej niż w normalnej grze.",
/*MISSING*/    "Circles are strange in hyperbolic geometry too. Look for the Castle of Camelot in the Crossroads; the Round Table inside is a circle of radius 28. Finding its center is a difficult challenge.\n\nPress '5' to cheat by seeing the smaller circles too.\n\nNote: Camelot and some other lands are unlocked earlier in the Tutorial than in a real game.",
   "I kružnice jsou v hyperbolické geometrii podivné. Zkus na Křižovatce najít hrad Kamelot; Kulatý stůl uvnitř je kružnice o poloměru 28. Najít její střed je obtížný úkol.\n\nStiskem klávesy '5' můžeš podvádět a vidět i menší kružnice.\n\nPoznámka: Camelot a některé další kraje se v Tutoriálu odemykají dříve než ve skutečné hře.",
   "Окружности в гиперболической геометрии тоже странные. Найдите замок Камелот на Перекрёстке; Круглый стол внутри него является кругом радиуса 28. Найти его центр -- трудная задача.\n\nНажмите '5' чтобы получить подсказку -- увидеть меньшие круги.\n\nВнимание: Камелот и некоторые другие земли открываются в Руководстве раньше, чем в обычной игре.",
/*MISSING*/    "Circles are strange in hyperbolic geometry too. Look for the Castle of Camelot in the Crossroads; the Round Table inside is a circle of radius 28. Finding its center is a difficult challenge.\n\nPress '5' to cheat by seeing the smaller circles too.\n\nNote: Camelot and some other lands are unlocked earlier in the Tutorial than in a real game.",
/*MISSING*/    "Circles are strange in hyperbolic geometry too. Look for the Castle of Camelot in the Crossroads; the Round Table inside is a circle of radius 28. Finding its center is a difficult challenge.\n\nPress '5' to cheat by seeing the smaller circles too.\n\nNote: Camelot and some other lands are unlocked earlier in the Tutorial than in a real game.",
    }},
  {0x63981bdd, { // "%The1 is confused!"
   "%1 jest zmieszan%ya1.",
   "%1 sersemledi.",
   "%1 je zmaten%ý1.",
   "%1 в замешательстве",
   "%Der1 %1 ist verwirrt.",
   "%na1 está confuso!",
    }},
  {0x63bd4d6b, { // "You cannot attack diagonally!"
   "Nie możesz atakować po przekątnej!",
/*MISSING*/    "You cannot attack diagonally!",
   "Nemůžeš útočit úhlopříčně!",
   "Вы не можете атаковать по-диагонали!",
   "Du kannst nicht diagonal angreifen!",
/*MISSING*/    "You cannot attack diagonally!",
    }},
  {0x63fd6548, { // "stay in place (left + right)"
   "stój w miejscu (lewo+prawo)",
/*MISSING*/    "stay in place (left + right)",
   "stát na místě (doleva + doprava)",
   "стоять на месте (влево + вправо)",
/*MISSING*/    "stay in place (left + right)",
/*MISSING*/    "stay in place (left + right)",
    }},
  {0x64067fd0, { // "player 7 X"
   "gracz 7 X",
/*MISSING*/    "player 7 X",
   "hráč 7 X",
   "игрок 7 X",
   "Spieler 7 X",
/*MISSING*/    "player 7 X",
    }},
  {0x64067fd1, { // "player 7 Y"
   "gracz 7 Y",
/*MISSING*/    "player 7 Y",
   "hráč 7 Y",
   "игрок 7 Y",
   "Spieler 7 Y",
/*MISSING*/    "player 7 Y",
    }},
  {0x644bdebd, { // "%The1 crashes into %the2!"
   "%1 walnął w %a2!",
/*MISSING*/    "%The1 crashes into %the2!",
   "%1 narazil na %a2!",
   "%1 врезал%c1 в %a2!",
/*MISSING*/    "%The1 crashes into %the2!",
/*MISSING*/    "%The1 crashes into %the2!",
    }},
  {0x645dd0db, { // "Somebody has summoned these evil constructs with a magical process."
   "Wroga konstrukcja, tworzona poprzez magiczny proces.",
   "Birisi büyüyle işleyen bu mekanizmaları harekete geçirmiş.",
   "Tyto zlé konstrukty kdosi vyvolal pomocí magického procesu.",
   "Враждебная конструкция, созданная в результате магического процесса.",
   "Jemand hat diese bösen Konstrukte mit einem magischen Ritual beschworen.",
/*MISSING*/    "Somebody has summoned these evil constructs with a magical process.",
    }},
  {0x645fe546, { // "unlock Orbs of Yendor"
   "otwórz Sfery Yendoru",
/*MISSING*/    "unlock Orbs of Yendor",
   "odemkni Yendorské Sféry",
   "открыть Сферы Йендора",
/*MISSING*/    "unlock Orbs of Yendor",
/*MISSING*/    "unlock Orbs of Yendor",
    }},
  {0x64aa539c, { // "This floor will fall after someone goes there. Go quickly!"
   "Ta podłoga się zapadnie, jak ktoś będzie przechodził. Prechodź szybko!",
   "Bu zemin birisi oradan geçtikten sonra çöker! Çabuk uzaklaş!",
   "Tato podlaha se propadne poté, co přes ni někdo projde. Běž rychle!",
   "Пол провалится, как только на него наступят. Бегите!",
   "Der Boden wird zusammenbrechen sobald jemand über ihn geht. Beeile dich!",
/*MISSING*/    "This floor will fall after someone goes there. Go quickly!",
    }},
  {0x65887dc3, { // "You cannot move between the triangular cells here!"
   "Nie możesz przechodzić między trójkątami!",
/*MISSING*/    "You cannot move between the triangular cells here!",
   "Tady se nemůžeš pohybovat mezi trojúhelníkovými políčky!",
   "Вы не можете ходить между треугольниками!",
   "Du kannst dich nicht zwischen den dreieckigen Zellen bewegen!",
/*MISSING*/    "You cannot move between the triangular cells here!",
    }},
  {0x65a3cf49, { // "You vault over %the1!"
   "Szarżujesz!",
/*MISSING*/    "You vault over %the1!",
   "Přeskoči%l0 jsi %a1!",
   "Вы совершили рывок!",
/*MISSING*/    "You vault over %the1!",
/*MISSING*/    "You vault over %the1!",
    }},
  {0x65c2d2d6, { // "Projection at the ground level"
   "Współczynnik rzutu na poziome podłoża",
/*MISSING*/    "Projection at the ground level",
   "Projekce na úrovni země",
   "Проекция на поверхность земли",
/*MISSING*/    "Projection at the ground level",
/*MISSING*/    "Projection at the ground level",
    }},
  {0x65f9fcde, { // "You push %the1."
   "Popchn%ąłeś0 %a1.",
   "%a1 ittirdin.",
   "Zatlači%l0 jsi %a1.",
   "Вы толкнули %a1.",
   "Du schiebst %den1 %a1.",
   "Você empurra %na1.",
    }},
  {0x66465077, { // "A perfect mirror wall. It is unbreakable and impassable even for aethereal beings, and everything you see inside is just an image of the real world; you can swing your sword at them, but that will not destroy them in the real world. Mirror walls reflect Mimics, lightning bolts, and missiles perfectly."
   "Doskonała lustrzana ściana. Nie da się jej zniszczyć,  i nawet istoty eteryczne nie są w stanie przejść na drugą stronę. Wszystko co widzisz jest tylko obrazem świata rzeczywistego; możesz machnąć mieczem w kierunku tego obrazu, ale to nie wpłynier na rzeczywistość. Lustrzany ściany perfekcyjnie odbijają Mimiki, błyskawice, i pociski.",
/*MISSING*/    "A perfect mirror wall. It is unbreakable and impassable even for aethereal beings, and everything you see inside is just an image of the real world; you can swing your sword at them, but that will not destroy them in the real world. Mirror walls reflect Mimics, lightning bolts, and missiles perfectly.",
   "Dokonalá zrcadlová stěna. Je nerozbitná a neprostupná, a to i pro éterické bytosti. Všechno, co v ní vidíš, je pouhý obraz reálného světa; můžeš do toho bít mečem, ale v reálném světě se nic nestane. Zrcadlové stěny dokonale odrážejí Mimiky, blesky a střely.",
   "Идеальная зеркальная стена. Невозможно сломать её или пройти через неё, даже для эфирных существ, и всё, что ты в ней видишь -- отражения реального мира. Ты можешь ударить по отражению мечом, но это не убьёт его в реальном мире. Стена идеально отражает Мимиков, молнии и ракеты. ",
/*MISSING*/    "A perfect mirror wall. It is unbreakable and impassable even for aethereal beings, and everything you see inside is just an image of the real world; you can swing your sword at them, but that will not destroy them in the real world. Mirror walls reflect Mimics, lightning bolts, and missiles perfectly.",
/*MISSING*/    "A perfect mirror wall. It is unbreakable and impassable even for aethereal beings, and everything you see inside is just an image of the real world; you can swing your sword at them, but that will not destroy them in the real world. Mirror walls reflect Mimics, lightning bolts, and missiles perfectly.",
    }},
  {0x66a55592, { // "%1 turns"
   "kolejek: %1",
/*MISSING*/    "%1 turns",
   "kola: %1",
   "ходов: %1",
/*MISSING*/    "%1 turns",
/*MISSING*/    "%1 turns",
    }},
  {0x66e43c1f, { // "In this mode you can draw your own player characters, floors, monsters, and items. Press 'e' while hovering over an object to edit it. Start drawing shapes with 'n', and add extra vertices with 'a'. Press 0-9 to draw symmetric pictures easily. More complex pictures can be created by using several layers ('l'). See the edges of the screen for more keys."
   "W tym trybie możesz narysować swoje postacie, podłogi, potwory i przedmioty. Naciśnij 'e', gdy przesuwasz mysz ponad obiektem, by go edytować. Zaczynaj nowe kształty od 'n', dodawaj wierzchołki 'a'. Naciśnij 0-9 by łatwo rysować symetryczne obrazy. Bardziej złożone obrazy mogą być tworzone przy użyciu warstw ('l'). Pozostałe klawisze są wymienione na krawędziach ekranu.",
   "Bu modda kendi karakterlerini, zeminlerini, canavarlarını ve eşyalarını çizebilirsin. İşaretçiyi bir nesneye götür ve onu düzenlemek için 'e'ye bas. 'n' ile şekiller çizmeye başlayabilir, 'a' ile yeni köşeler ekleyebilirsin. 0-9'a basarak simetrik resimleri daha kolay çizebilirsin. Daha karmaşık resimler için birden fazla atman gerekebilir ('l'). Ekranın kıyılarına bakara daha fazla tuşu öğrenebilirsin.",
   "V tomto módu si můžeš nakreslit vlastní postavy, podlahu, netvory a předměty. Objekt můžeš editovat tak, že na něj najedeš myší a stiskneš 'e'. 'n' začne kreslit tvar a 'a' přidává nové vrcholy. Klávesy 0-9 umožňují snadnou kresbu symetrických obrázků. Složitější obrázky je možné vytvořit s použitím ně kolika vrstev ('l'). Další ovládací klávesy můžeš najít na okrajích obrazovky.",
   "В этом режиме можно нарисовать своего персонажа, пол, монстров, предметы. Нажмите 'e', держа мышь над объектом, чтобы его изменить. Добавляйте формы с 'n', вершины с 'a'. Нажмите 0-9 чтобы сделать симметричный узор. Более сложные фигуры могут состоять из нескольких слоёв ('l'). Остальные клавиши смотри в углах экрана.",
   "In diesem Modus kannst du deine eigenen Charaktere, Böden, Monster und Items zeichnen. Drücke 'e' während du auf ein Objekt zeigst um es zu bearbeiten. Fang an Formen mit 'n' zu zeichnen und füg mit 'a'. weitere Eckpunkte hinzu. Drücke 0-9 um einfach symmetrisch zu zeichnen. Komplexere Bilder können durch mehrere Lagen ('l') erzeugt werden. Weitere Hotkeys stehen am Rand des Bildschirms.",
/*MISSING*/    "In this mode you can draw your own player characters, floors, monsters, and items. Press 'e' while hovering over an object to edit it. Start drawing shapes with 'n', and add extra vertices with 'a'. Press 0-9 to draw symmetric pictures easily. More complex pictures can be created by using several layers ('l'). See the edges of the screen for more keys.",
    }},
  {0x66eada42, { // "friends killed: %1"
   "przyjaciele zabici: %1",
/*MISSING*/    "friends killed: %1",
   "zabití kamarádi: %1",
   "убито союзников: %1",
   "getötete Freunde: %1",
/*MISSING*/    "friends killed: %1",
    }},
  {0x6729a5eb, { // "do not get checkmated"
   "nie daj się zaszachować",
/*MISSING*/    "do not get checkmated",
   "vyhni se šachmatu",
   "не дай поставить себе мат",
/*MISSING*/    "do not get checkmated",
/*MISSING*/    "do not get checkmated",
    }},
  {0x6731c97f, { // "This Orb is able to bring faraway items to your location, even if there are monsters or barriers on the way. The cost of bringing an item (in charges) equals the square of its distance to you. Contrary to some other Orbs, usage is not allowed if you have not enough power left."
   "Ta Sfera może ściągać do Ciebie odległe przedmioty, nawet, jak po drodze są potwory lub inne przeszkody. Koszt ściągnięcia przedmiotu (w ładunkach) równy jest kwadratowi odległości. W przeciwieństwie do niektórych innych Sfer, nie możesz użyć Sfery Przestrzeni, jeśli nie masz wystarczająco dużo mocy.",
   "Bu küre uzaktaki eşyaları , arada canavar yahut engeller olup olmadığı farketmeksizin, senin olduğun hücreye getirir. Bu işlem eşyanın uzaklığı kadar güç tüketir. Bazı Kürelerin aksine, güç yoksa bu küreyi kullanamazsın.",
   "Tato Sféra vám dokáže přinášet vzdálené předměty, a to i skrz netvory nebo bariéry. Přinesení předmětu stojí tolik síly, kolik je druhá mocnina vzdálenosti předmětu. Na rozdíl od některých jiných Sfér, Sféru Prostoru není možné použít, pokud vám už nezbývá dostatečné množství síly.",
   "Эта сфера может принести Вам предметы издалека, даже если на пути есть монстры или препятствия. Стоимость использования равна квадрату расстояния до предмета. В отличие от некоторых других сфер, она не может быть использована, если ей силы не хватает.",
   "Dieser Orb kann weit entfernte Gegenstände zu dir bringen, auch wenn Monster oder Barrieren im Weg sind. Die Kosten dafür ist die Distanz des Gegenstands zu dir zum Quadrat. Anders als ein paar andere Orbs kannst du diesen nicht benutzen, wenn er nicht genug Kraft hat.",
/*MISSING*/    "This Orb is able to bring faraway items to your location, even if there are monsters or barriers on the way. The cost of bringing an item (in charges) equals the square of its distance to you. Contrary to some other Orbs, usage is not allowed if you have not enough power left.",
    }},
  {0x676ec0fa, { // "When the charges on this Orb expire, you will be automatically returned to the place where you have found it. Extra Orbs of Recall delay this without changing the recall location. Pick up an Orb of Safety causes an immediate recall."
   "Kiedy skończą się ładunki tej Sfery, automatycznie wracasz do miejsca, w którym ją znalazłeś. Dodatkowe Sfery Powrotu opóźniają powrót bez zmiany miejsca powrotu, a Sfera Bezpieczeństwa powoduje natychmiastowy powrót.",
/*MISSING*/    "When the charges on this Orb expire, you will be automatically returned to the place where you have found it. Extra Orbs of Recall delay this without changing the recall location. Pick up an Orb of Safety causes an immediate recall.",
   "Když vyprší náboje této Sféry, automaticky se vrátíš tam, kde jsi ji našel. Pokud sebereš více Sfér Návratu, místo návratu se nezmění -- návrat se pouze oddálí. Pokud sebereš Sféru Bezpečnosti, vrátíš se okamžitě.",
   "Когда заряды этой сферы заканчиваются, вы автоматически возвращаетесь в место, где вы её нашли. Дополнительные сферы откладывают этот момент, не меняя места возвращения, а Сфера Безопасности вызывает немедленное возвращение.",
/*MISSING*/    "When the charges on this Orb expire, you will be automatically returned to the place where you have found it. Extra Orbs of Recall delay this without changing the recall location. Pick up an Orb of Safety causes an immediate recall.",
/*MISSING*/    "When the charges on this Orb expire, you will be automatically returned to the place where you have found it. Extra Orbs of Recall delay this without changing the recall location. Pick up an Orb of Safety causes an immediate recall.",
    }},
  {0x679cde87, { // "The Orb loses its power as it leaves the Land of Power!"
   "Sfera traci swą moc, gdy opuszcza Krainę Mocy!",
   "Küre Güç Diyarından ayrılırken gücünü kaybediyor!",
   "Sféra opustila Kraj Moci a ztratila svou sílu!",
   "Сфера потеряла свою силу, когда покинула Землю силы!",
   "Der Orb verliert seine Kraft wenn du das Land der Macht verlässt!",
/*MISSING*/    "The Orb loses its power as it leaves the Land of Power!",
    }},
  {0x67b00d2a, { // "pan left"
   "przewiń w lewo",
   "sola kaydır",
   "scrollování doleva",
   "крутить влево",
   "Nach links schwenken",
/*MISSING*/    "pan left",
    }},
  {0x67ef03c1, { // "second joystick position (panning)"
   "pozycja drugiego joysticka (przesuwanie)",
/*MISSING*/    "second joystick position (panning)",
   "pozice druhého joysticku (přesouvání)",
   "положение второго джойстика (обзор)",
/*MISSING*/    "second joystick position (panning)",
/*MISSING*/    "second joystick position (panning)",
    }},
  {0x6828d3e3, { // "Rock snakes are similar to Sandworms, but they move each turn. However, they only move on the hexagonal cells. They are also somewhat longer than the Sandworms."
   "Kamienne węże są podobne do Pustynnych Czerwi, ale wykonują ruch w każdej kolejce. Mogą jednak poruszać się tylko po sześciokątach. Są trochę dłuższe niż Pustynne Czerwie.",
   "Kaya yılanları her tur hareket etmesi hariç Kumkurtlarına benzer. Ayrıca, sadece altıgen hücrelerde hareket ederler. Kumkurtlarından bir nebze daha genişlerdir. ",
   "Kamenní hadi jsou podobní Písečným červům, ale pohybují se v každém kole. Mohou se však pohybovat pouze po šestiúhelníkových polích. Kromě toho jsou oproti Písečným červům o něco delší.",
   "Каменные змеи похожи на Пустынных червей, но движутся каждый ход. Однако они могут перемещаться только по шестиугольникам. Также они несколько длиннее червей.",
   "Felsschlangen sind Sandwürmern ähnlich, allerdings bewegen sie sich jeden Zug. Dafür meiden sie allerdings heptagonale Zellen. Sie sind auch länger als Sandwürmer.",
/*MISSING*/    "Rock snakes are similar to Sandworms, but they move each turn. However, they only move on the hexagonal cells. They are also somewhat longer than the Sandworms.",
    }},
  {0x687fbdb7, { // "player 2 spin"
   "gracz 2 obrót",
   "oyuncu 2, dön",
   "hráč 2 otáčení",
   "игрок 2 вертит",
   "Spieler 2 drehen",
/*MISSING*/    "player 2 spin",
    }},
  {0x6890adc7, { // "anti-aliasing disabled"
   "wyłączono anti-aliasing",
   "Anti-aliasing devredışı",
   "anti-aliasing vypnut",
   "сглаживание выключено",
   "Anti-Aliasing deaktiviert",
   "Antiserrilhamento desabilitado",
    }},
  {0x68a71369, { // "You can fight most monsters by moving into their location. The monster could also kill you by moving into your location, but the game automatically cancels all moves which result in that.\n\n"
   "Większość przeciwników zabijasz poprzez wejście na zajmowane przez nie pole. Przeciwnik może Cię zabić w ten sam sposób, ale gra automatycznie zabrania ruchów, które by do tego prowadziły.\n\n",
   "Çoğu canavarın yerine ilerleyerek onu öldürebilirsin. Canavar da seni senin bulunduğun yere ilerleyerek öldürebilir, ama oyun böyle hamlelere zaten izin vermez.\n\n",
   "Na většinu netvorů můžeš zaútočit tím, že se pohneš na jejich políčko. Netvor může zabít tebe, pokud se pohne na tvé políčko, ale hra automaticky zruší každý tah, který by vedl k takové situaci.\n\n",
   "Вы можете атаковать большинство монстров, просто передвигаясь на клетку с ними. Монстры могут атаковать Вас тем же способом, но игра сама блокирует ходы, приводящие к этому.\n\n",
   "Du kannst die meisten Monster bekämpfen, indem du dich an ihre Position begibst. Besagtes Monster könnte dich auch töten, indem es sich an deine Position begibst,  aber das Spiel lässt dies nicht zu.\n\n",
   "Você pode lutar com a maioria dos monstros ao se mover para sua localização. O monstro também pode lhe matar ao se mover para sua localização, mas o jogo cancela automaticamente qualquer movimento que resulte nisso.\n\n",
    }},
  {0x68b8aef8, { // "Welcome to %the1 Challenge!"
   "Uratuj %a1!",
   "%1 görevine hoşgeldin!",
   "Vítej v Misi: %1!",
   "Спаси %a1!",
   "Willkommen zu der %a1-Herausforderung!",
/*MISSING*/    "Welcome to %the1 Challenge!",
    }},
  {0x68c82eb7, { // " (appears here)"
   " (występuje tu)",
/*MISSING*/    " (appears here)",
   " (objevuje se zde)",
   " (появляется здесь)",
/*MISSING*/    " (appears here)",
/*MISSING*/    " (appears here)",
    }},
#if REPEATED
  {0x68c82eb8, { // " (appears here) [repeat]"
   " (występuje tu)",
/*MISSING*/    " (appears here) [repeat]",
   " (objevuje se zde)",
   " (появляется здесь)",
/*MISSING*/    " (appears here) [repeat]",
/*MISSING*/    " (appears here) [repeat]",
    }},
#endif
  {0x68efc6fc, { // "model used"
   "użyty model",
/*MISSING*/    "model used",
   "použitý model",
   "используемая модель",
   "verwendetes Modell",
/*MISSING*/    "model used",
    }},
  {0x691a64f3, { // "r = regular"
   "r = wzorki",
   "r = düzenli",
   "r = vzory",
   "r = узор",
   "r = regulär",
/*MISSING*/    "r = regular",
    }},
  {0x695a10f8, { // "language"
   "język",
   "dil",
   "jazyk",
   "язык",
   "Sprache",
   "idioma",
    }},
  {0x6961cc1d, { // "explore the Palace Pattern (8 colors)"
   "Eksploracja Wzoru Pałac (8 kolorów)",
   "Saray Dokusunu Keşfet (8 renk)",
   "Zkoumání Palácového vzoru (8 barev)",
   "Изучить узор Дворца (8 цветов)",
   "Erkunde die Schlossstruktur (8 Farben)",
/*MISSING*/    "explore the Palace Pattern (8 colors)",
    }},
  {0x699e98b2, { // "Try the Tutorial to help with understanding the geometry of HyperRogue (menu -> special modes).\n\n"
   "Uruchom Podręcznik, by zrozumieć geometrię HyperRogue (menu -> tryby specjalne.\n\n",
/*MISSING*/    "Try the Tutorial to help with understanding the geometry of HyperRogue (menu -> special modes).\n\n",
   "Zkuste Tutoriál, chcete-li pomoci s chápáním geometrie HyperRogue (menu -> speciální módy.\n\n",
   "Попробуйте Руководство, чтобы понять геометрию HyperRogue (меню -> специальные режимы.\n\n",
/*MISSING*/    "Try the Tutorial to help with understanding the geometry of HyperRogue (menu -> special modes).\n\n",
/*MISSING*/    "Try the Tutorial to help with understanding the geometry of HyperRogue (menu -> special modes).\n\n",
    }},
  {0x69b39d48, { // "right to left"
   "w lewo",
/*MISSING*/    "right to left",
   "zprava doleva",
   "влево",
   "von rechts nach links",
/*MISSING*/    "right to left",
    }},
  {0x69dd0ca1, { // "Press F5 or 'o' to try again!"
   "Naciśnij F5 lub 'o' by spróbować jeszcze raz!",
/*MISSING*/    "Press F5 or 'o' to try again!",
   "Stiskem F5 nebo 'o' to mùžeš zkusit znovu!",
   "Нажми F5 или 'o', чтобы попробовать ещё раз!",
/*MISSING*/    "Press F5 or 'o' to try again!",
/*MISSING*/    "Press F5 or 'o' to try again!",
    }},
  {0x6a04fee0, { // "Cannot use %the1 here!"
   "Nie możesz tutaj tego użyć!",
/*MISSING*/    "Cannot use %the1 here!",
   "Tady to nemùe pouít!",
   "Nie możesz tutaj tego użyć!",
/*MISSING*/    "Cannot use %the1 here!",
/*MISSING*/    "Cannot use %the1 here!",
    }},
  {0x6a077dec, { // "%The1 disperses the cloud!"
   "%1 rozprys%ł1 chmurę!",
   "%1 bulutun içerisinde kayboldu!",
   "%1 rozptýli%l1 oblak!",
   "%1 рассеял%E1 облако!",
   "%Der1 %1 löst die Wolke auf!",
   "%na1 dispersa a nuvem!",
    }},
  {0x6a72bef7, { // "Expansion"
   "Ekspansja",
/*MISSING*/    "Expansion",
   "Expanze",
   "Расширение",
/*MISSING*/    "Expansion",
/*MISSING*/    "Expansion",
    }},
  {0x6ab87b7f, { // "render bands automatically"
   "automatycznie twórz wstęgi",
/*MISSING*/    "render bands automatically",
   "renderovat pásy automaticky",
   "рисовать ленту автоматически",
   "rendere Bänder automatisch",
/*MISSING*/    "render bands automatically",
    }},
  {0x6ae460f4, { // "game design, programming, texts and graphics by Zeno Rogue <zeno@attnam.com>\n\n"
   "projekt, programowanie, teksty i grafika: Zeno Rogue <zeno@attnam.com>\n\n",
   "tasarım, programlama, metinler ve grafikler: Zeno Rogue <zeno@attnam.com>\n\n",
   "design, programování, texty a grafika: Zeno Rogue <zeno@attnam.com>\n\n",
   "проект, программирование, тексты и графика: Zeno Rogue <zeno@attnam.com>\n\n",
   "Design, Programmierung, Texte und Grafik: Zeno Rogue \n\n",
   "game design, programação, textos e gráficos feitos por: Zeno Rogue <zeno@attnam.com>\n\n",
    }},
  {0x6b12c064, { // "This orb can be used to invoke a flash spell, which destroys almost everything in radius of 2."
   "Ta sfera pozwala rzucić czar Błysk, który niszczy wszystko w promieniu 2.",
   "Bu küre ışık büyüsünü kullanmaya yarar. Işık büyüsü, 2 hücre uzaklıktaki her şeyi yok eder.",
   "Tuto sféru můžeš použít k seslání kouzla Záblesk, které zničí téměř vše do vzdálenosti 2 políček od tebe.",
   "Эта сфера вызывает заклинание вспышки, уничтожающее почти всё в радиусе 2.",
   "Du kannst diesen Orb verwenden um einen Lichtzauber zu wirken, der nahezu alles um dich in einem Radius von 2 zerstört.",
/*MISSING*/    "This orb can be used to invoke a flash spell, which destroys almost everything in radius of 2.",
    }},
  {0x6b41c58b, { // "You think about possible meanings."
   "Zastanawiasz się nad możliwymi znaczeniami.",
   "Bunun olası anlamlarını düşündün.",
   "Přemýlíš, co by to mohlo znamenat.",
   "Вы думаете о возможном значении.",
   "Du denkst über die Bedeutung nach.",
   "Você pensa em possíveis significados.",
    }},
  {0x6b49cc24, { // "Treasure lost!"
   "Skarb utracony!",
   "Hazine kayboldu!",
   "Ztrati%l0 jsi poklad!",
   "Сокровища потеряны!",
   "Schatz verloren!",
   "Tesouro perdido!",
    }},
  {0x6b4ba313, { // "l = layers: %1"
   "l = warstwy: %1",
   "l = katmanlar: %1",
   "l = vrstvy: %1",
   "l = слой: %1",
   "l = Ebenen: %1",
/*MISSING*/    "l = layers: %1",
    }},
  {0x6b524122, { // "Periodic Editor"
   "Edytor okresowy",
   "Periyodik Editör",
   "Periodický editor",
   "Периодический редактор",
   "Periodischer Editor",
/*MISSING*/    "Periodic Editor",
    }},
  {0x6b63d613, { // "Rock-III to wall ratio"
   "Stosunek skały-III do ścian",
/*MISSING*/    "Rock-III to wall ratio",
   "Poměr skály-III a zdí",
   "Отношение высот Скалы-III и стены",
/*MISSING*/    "Rock-III to wall ratio",
/*MISSING*/    "Rock-III to wall ratio",
    }},
  {0x6b728a57, { // "You collect your first %1!"
   "Znalaz%łeś0 pierwsz%yą1 %a1!",
   "%a1 topladın. Bu senin ilkin!",
   "Naše%l0 jsi sv%ůj1 první %a1!",
   "Вы нашли перв%A1 %a1!",
   "Du hast dein erstes mal %a1 eingesammelt!",
   "Você coletou %seu1 primeir%na1!",
    }},
  {0x6bb668d1, { // "four triangles"
   "cztery trójkąty",
/*MISSING*/    "four triangles",
   "ètyøi trojúhelníky",
   "четыре треугольника",
/*MISSING*/    "four triangles",
/*MISSING*/    "four triangles",
    }},
  {0x6bc18ab5, { // "A magical being which copies your movements."
   "Magiczna istota kopiująca Twoje ruchy.",
   "Hareketlerini taklit eden bir büyülü varlık.",
   "Magická bytost, které kopíruje tvé pohyby.",
   "Это магическое создание копирует Ваши движения.",
   "Ein magisches Wesen, das deine Bewegungen kopiert.",
/*MISSING*/    "A magical being which copies your movements.",
    }},
  {0x6bc2888a, { // "geometry"
   "geometria",
/*MISSING*/    "geometry",
   "geometrie",
   "геометрия",
/*MISSING*/    "geometry",
/*MISSING*/    "geometry",
    }},
  {0x6bdbf93c, { // "exit configuration"
   "wyjdź z konfiguracji",
   "ayarlardan çık",
   "opustit nastavení",
   "выйти из настроек",
   "Einstellungen verlassen",
   "sair das configurações",
    }},
#if REPEATED
  {0x6bdbf93d, { // "exit configuration [repeat]"
   "wyjdź z konfiguracji",
   "ayarlardan çık.",
   "opustit nastavení",
   "выйти из настроек",
   "Einstellungen verlassen",
   "sair das configurações",
    }},
#endif
  {0x6bdecf0b, { // " (mirrored)"
   " (odbite)",
/*MISSING*/    " (mirrored)",
   " (zrcadlená)",
   " (отражено)",
/*MISSING*/    " (mirrored)",
/*MISSING*/    " (mirrored)",
    }},
  {0x6c22a84e, { // "i = items"
   "l = krainy",
   "l = eşya",
   "m = netvoři",
   "l = земли",
   "l = Items",
/*MISSING*/    "i = items",
    }},
  {0x6c6e9f84, { // "next slide"
   "kolejny slajd",
/*MISSING*/    "next slide",
   "další snímek",
   "следующий слайд",
/*MISSING*/    "next slide",
/*MISSING*/    "next slide",
    }},
  {0x6cb36956, { // "Press ESC to view this screen during the game."
   "Naciśnij ESC w czasie gry, by obejrzeć ten ekran.",
/*MISSING*/    "Press ESC to view this screen during the game.",
   "Tuto obrazovku mùžeš bìhem hry zobrazit stiskem ESC.",
   "Нажми ESC, чтобы увидеть этот экран во время игры.",
/*MISSING*/    "Press ESC to view this screen during the game.",
/*MISSING*/    "Press ESC to view this screen during the game.",
    }},
  {0x6cbed3c9, { // "3D"
   "3D",
/*MISSING*/    "3D",
   "3D",
   "3D",
/*MISSING*/    "3D",
/*MISSING*/    "3D",
    }},
  {0x6cc91767, { // "world overview"
   "przegląd krain",
   "dünya önizlemesi",
   "přehled krajů",
   "обзор мира",
   "Weltüberblick",
/*MISSING*/    "world overview",
    }},
  {0x6cfacec4, { // "right"
   "prawo",
/*MISSING*/    "right",
   "doprava",
   "право",
   "rechts",
/*MISSING*/    "right",
    }},
  {0x6d034a95, { // "Hit him by walking away from him."
   "Odejdź od niego, by go trafić.",
   "Ona ondan uzaklaşarak vur.",
   "Zasáhni ho tak, že se pohneš směrem od něj.",
   "Ударьте его, убежав от него.",
   "Triff ihn indem du dich von ihm entfernst.",
/*MISSING*/    "Hit him by walking away from him.",
    }},
  {0x6d1c912c, { // "%The1 is destroyed by the Flash."
   "Błysk zniszczył %a1.",
/*MISSING*/    "%The1 is destroyed by the Flash.",
   "Záblesk zničil %a1.",
   "Вспышка уничтожила %1.",
   "%Der1 %1 wird durch den Lichtblitz zerstört.",
/*MISSING*/    "%The1 is destroyed by the Flash.",
    }},
  {0x6d5e942d, { // "Tiny droplets of magical water. You see images of yourself inside them. Go inside the cloud, to make these images help you."
   "Malutkie kropelki magicznej wody, w których widać Twój obraz. Wejdź do środka chmury, by te obrazy Ci pomogły.",
   "Ufak sihirli su damlaları. Kendi görüntünü içlerinde görüyorsun. Bu görüntülerin sana yardım etmesi için bulutun içine gir. ",
   "Maličké kapičky magické vody. Vidíš v nich své obrazy. Když vejdeš do oblaku, tyto obrazy ti pomohou.",
   "Крошечные капли магической жидкости. Вы видите в них свои отражения. Войдите в облако, чтобы создать двойников, помогающих Вам.",
   "Kleine Tropfen magischen Wassers. Du siehst Abbilder von dir in ihnen. Gehe in die Wolke, um diese Bilder real werden zu lassen.",
/*MISSING*/    "Tiny droplets of magical water. You see images of yourself inside them. Go inside the cloud, to make these images help you.",
    }},
  {0x6d696af0, { // "framerate limit"
   "ograniczenie na fps",
   "kare sayısı sınırı",
   "limit fps",
   "ограничение на fps",
   "Framerate-Limit",
   "limite de qps",
    }},
  {0x6d8c78c2, { // "Usually, you move by touching somewhere on the map; you can also touch one of the four buttons on the map corners to change this (to scroll the map or get information about map objects). You can also touch the numbers displayed to get their meanings.\n"
   "Zazwyczaj poruszasz się przez dotknięcie pola na mapie; możesz również dotknąć jednego z przycisków w rogach, by to zmienić (by przewijać mapę albo zdobyć informacje). Możesz też dotknąć liczb na ekranie, by poznać ich znaczenie.\n",
   "Normalde haritada bir yere dokunarak hareket edersin, haritanın dört köşesindeki tuşlara dokunarak bunu değiştirebilirsin. (haritayı çevirmek ya da haritadaki nesneler hakkında bilgi almak için). Gösterilen numaralara dokunarak da anlamlarını görebilirsin. \n",
   "Pohybuješ se obvykle tak, že se dotkneš nějakého místa na mapě; to můžeš změnit tím, že se dotkneš jednoho ze čtyř tlačítek v rohu mapy (pokud chceš posunovat mapu nebo získat informace o objektech na mapě). Také se můžeš dotknout zobrazených čísel, abys zjistil jejich význam.\n",
   "Обычно Вы перемещаетесь просто кликая мышкой в нужном направлении; на можно использовать 4 кнопки по углам карты чтобы листать карту или получатьинформацию про объекты. Кликая по изображаемым числам, можно смотреть, что они означают.\n",
   "Normalerweise bewegst du dich durch Antippen der Karte, du kannst aber auch einen der vier Knöpfe in den Ecken antippen um dies zu ändern (um die Karte zu scrollen oder Infos über Objekte zu bekommen). Du kannst auch die Zahlen antippen um ihre Bedeutung zu erfahren.\n",
   "Você usualmente se move tocando em algum lugar do mapa; você também pode tocar em um dos quatro botões das bordas do mapa para mudar isso (para navegar pelo mapa ou ter informações sobre os objetos do mapa). Você também pode tocar nos números mostrados para saber seus significados.\n",
    }},
  {0x6da40197, { // "You trick %the1."
   "Pokona%łeś0 %a1.",
   "%a1 kandırdın.",
   "Přelsti%l0 jsi %a1.",
   "Вы победили %a1.",
   "Du überlistest %den1 %a1.",
/*MISSING*/    "You trick %the1.",
    }},
  {0x6df385a2, { // "You need to move to give space to %the1!"
   "Musisz się ruszyć, by zrobić miejsce dla %a1!",
/*MISSING*/    "You need to move to give space to %the1!",
   "Musíš se pohnout, abys uděla%l0 místo pro %a1!",
   "Вам нужно освободить место для %a1!",
   "Du musst dich bewegen, um Platz für %den1 %a1 zu machen!",
/*MISSING*/    "You need to move to give space to %the1!",
    }},
  {0x6dfba684, { // "normal tesselation"
   "zwykła posadzka",
/*MISSING*/    "normal tesselation",
   "normální dělení",
   "обычное замощение",
/*MISSING*/    "normal tesselation",
/*MISSING*/    "normal tesselation",
    }},
  {0x6e1db4ec, { // "drop Dead Orb"
   "połóż Martwą Sferę",
   "Ölü Küreyi bırak",
   "položení Mrtvé sféry",
   "бросить мёртвую сферу",
   "Toten Orb ablegen",
/*MISSING*/    "drop Dead Orb",
    }},
  {0x6e262183, { // "\"Nice castle, eh?\""
   "\"Ładny zamek, co?\"",
   "\"Kalemiz güzel, değil mi?\"",
   "\"Fajn hrad, co?\"",
   "\"Хороший замок, да?\"",
   "\"Nettes Schloss, nicht?\"",
/*MISSING*/    "\"Nice castle, eh?\"",
    }},
  {0x6e34c335, { // "Press F5 to restart the game normally."
   "Wciśnij F5, by zacząć normalną grę.",
   "Oyunu normal başlatmak için F5'e bas",
   "Stisknutím F5 restartuješ hru v normálním módu.",
   "Нажмите F5, чтобы начать нормальную игру.",
   "Drücke F5 um das Spiel neu zu starten.",
/*MISSING*/    "Press F5 to restart the game normally.",
    }},
  {0x6e366778, { // "Affects looks and grammar"
   "Wpływa na wygląd i gramatykę",
   "Görüntüyü ve grameri etkiler",
   "Ovlivňuje vzhled a gramatiku",
   "Влияет на внешний вид и грамматику",
   "Aussehen und Grammatik",
/*MISSING*/    "Affects looks and grammar",
    }},
  {0x6e49d052, { // "move up"
   "idź do góry",
   "yukarı git",
   "nahoru",
   "идти наверх",
   "Hoch bewegen",
/*MISSING*/    "move up",
    }},
  {0x6e5942bf, { // "display the scope"
   "pokaż zasięg",
   "odağı göster",
   "zobraz rozsah",
   "показать охват",
   "Bereich anzeigen",
/*MISSING*/    "display the scope",
    }},
  {0x6e6bfe7c, { // "You would get hurt!"
   "To by bolało!",
   "Orada hasar alırsın!",
   "To by bolelo!",
   "Это было больно!",
   "Du würdest dir wehtun!",
/*MISSING*/    "You would get hurt!",
    }},
  {0x6ead9dc9, { // "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, etc."
   "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, itd.",
   "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, vb.",
   "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, atd.",
   "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, и т. д..",
   "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, etc.",
/*MISSING*/    "Shift+F, Shift+O, Shift+T, Shift+L, Shift+U, etc.",
    }},
  {0x6ec9d34d, { // "firewall lines: Palace"
   "linie ścian ognia: Pałac",
/*MISSING*/    "firewall lines: Palace",
   "čáry ohnivých stěn: Palác",
   "стены огня: Дворец",
/*MISSING*/    "firewall lines: Palace",
/*MISSING*/    "firewall lines: Palace",
    }},
  {0x6f6e8057, { // "clear monster"
   "zabierz potwora",
   "canavarı temizle",
   "vymazat netvore",
   "удалить монстра",
   "Monster entfernen",
/*MISSING*/    "clear monster",
    }},
  {0x6fa9922a, { // "Klein model"
   "model Kleina",
/*MISSING*/    "Klein model",
   "Kleinův model",
   "модель Клейна",
/*MISSING*/    "Klein model",
/*MISSING*/    "Klein model",
    }},
  {0x6fabe1c6, { // "These sinister totems contain valuable gems."
   "Te złowieszcze totemy zawierają cenne klejnoty.",
   "Bu tekinsiz totemler değerli mücevherler içeriyor.",
   "Tyto zlověstné totemy obsahují cenné drahokamy.",
   "Этот зловещий тотем содержит ценные камни.",
   "Diese schaurigen Totems enthalten wertvolle Edelsteine.",
/*MISSING*/    "These sinister totems contain valuable gems.",
    }},
  {0x6fccf858, { // "pics to save/load:"
   "obrazki do zapisu/wczytania:",
   "kaydedilecek/yüklenecek resimler",
   "obrázky k uložení/nahrání:",
   "картинки для загрузки/сохранения:",
   "Bilder zum speichern/laden:",
/*MISSING*/    "pics to save/load:",
    }},
  {0x6fe6194c, { // "You teleport to a new location!"
   "Teleportujesz się do nowego miejsca!",
   "Yeni bir yere ışınlandın!!",
   "Teleportuješ se na nové místo!",
   "Вы телепортировались в новое место!",
   "Du teleportierst dich an einen neuen Ort!",
   "Você teleporta-se para um lugar novo!",
    }},
  {0x70431bf2, { // "save whom"
   "kogo ratujemy",
   "kim kurtarılacak",
   "koho zachraňuješ",
   "кого спасти",
   "wen retten?",
/*MISSING*/    "save whom",
    }},
  {0x704dd087, { // "Reduce the framerate limit to conserve CPU energy"
   "Ogranicz fps, by oszczędzać energię",
   "CPU enerjisinden tasarruf etmek için kare sayısı sınırını azalt.",
   "Sniž fps, abys šetřil energii procesoru",
   "Ограничить fps, чтобы сохранить энергию ЦП",
   "Reduzieren des Framerate Limits spart CPU Energie",
   "Reduzir o limite da taxa de quadros para conservar energia da CPU",
    }},
  {0x70858816, { // "Click a location to teleport there."
   "Kliknij pole, by tam się teleportować.",
/*MISSING*/    "Click a location to teleport there.",
   "Klikni na políčko, kam se chceš teleportovat.",
   "Кликни на поле, чтобы телепортироваться туда.",
/*MISSING*/    "Click a location to teleport there.",
/*MISSING*/    "Click a location to teleport there.",
    }},
  {0x7091f735, { // "leaderboards/achievements"
   "osiągnięcia i rankingi",
/*MISSING*/    "leaderboards/achievements",
   "leaderboardy/achievementy",
   "таблицы рекордов и достижения",
   "Bestenlisten/Errungenschaften",
/*MISSING*/    "leaderboards/achievements",
    }},
  {0x70ae47f3, { // "A Witch with an Aether spell. She is able to move through fire and walls."
   "Czarownica z czarem Eteru. Może przechodzić przez ogień i ściany.",
   "Eter büyüsü olan bir cadı. Duvarların ve ateşin içinden geçebilir.",
   "Čarodějnice s kouzlem Éteru. Dokáže procházet ohněm a zdmi.",
   "Ведьма с заклинанием эфира. Может проходить сквозь огонь и стены.",
   "Eine Ätherhexe. Sie kann durch Feuer und Wände gehen.",
/*MISSING*/    "A Witch with an Aether spell. She is able to move through fire and walls.",
    }},
  {0x70b88165, { // "forward"
   "do przodu",
   "ileri",
   "dopředu",
   "вперёд",
   "Vor",
/*MISSING*/    "forward",
    }},
  {0x70d00f84, { // "This wonderful Orb can only be collected by those who have truly mastered this hyperbolic universe, as you need the right key to unlock it. Luckily, your psychic abilities will let you know where the key is after you touch the Orb."
   "Ta cudowna sfera może być zdobyta tylko przez prawdziwych mistrzów tego świata. By ją otworzyć, potrzebujesz klucza. Na szczęście Twoje moce psychiczne powiedzą Ci, gdzie jest klucz, gdy dotkniesz Sfery.",
   "Bu harika Küre kilidini açmak için doğru anahtarı gerektirdiğinden, sadece bu hiperbolik evrende tamamen ustalaşmış kişilerce toplanabilir.Çok şükür, küreye dokununca psişik yeteneklerin sana anahtarın nerede olduğunu söyler.",
   "Tuto úžasnou sféru mohou sebrat jenom ti, kdo se stali skutečnými pány tohoto hyperbolického vesníru, protože na její odemčení potřebuješ správný klíč. Tvé psychické schopnosti ti naštěstí prozradí, kde se klíč nachází, jakmile se této sféry dotkneš.",
   "Эта прекрасная сфера может быть собрана лишь тем, кто по настоящему освоил гиперболическую вселенную, но Вам нужно найти ключ от неё. К счастью, Вы узнаете, где он лежит, как только дотронетесь до сферы.",
   "Dieser wundervolle Orb kann nur von denen gesammelt werden, die dieses hyperbolische Universum wahrhaft gemeistert haben - denn du benötigst den richtigen Schlüssel, um ihn zu entsperren. Deine psychischen Fähigkeiten erlauben es dir glücklicherweise zu erkennen, wo sich der Schlüssel befindet, nachdem du den Orb berührt hast.",
/*MISSING*/    "This wonderful Orb can only be collected by those who have truly mastered this hyperbolic universe, as you need the right key to unlock it. Luckily, your psychic abilities will let you know where the key is after you touch the Orb.",
    }},
  {0x70d58a62, { // "input"
   "sterowanie",
/*MISSING*/    "input",
   "vstup",
   "ввод",
/*MISSING*/    "input",
/*MISSING*/    "input",
    }},
  {0x70df17ab, { // "This model renders straight lines as straight, but it distorts angles."
   "Ten model pokazuje linie proste jako proste, ale przekłamuje kąty.",
/*MISSING*/    "This model renders straight lines as straight, but it distorts angles.",
   "Tento model zobrazuje přímky jako přímky, ale zkresluje úhly.",
   "В этой модели прямые линии остаются прямыми, но углы искажаются.",
/*MISSING*/    "This model renders straight lines as straight, but it distorts angles.",
/*MISSING*/    "This model renders straight lines as straight, but it distorts angles.",
    }},
  {0x711da86b, { // "\"I would like to congratulate you again!\""
   "\"Pragnę jeszcze raz Ci pogratulować!\"",
   "\"Seni tekrar tebrik etmek isterim!\"",
   "\"Rád bych ti znovu pogratuloval!\"",
   "\"Я хотел бы ещё раз Вас поздравить!\"",
   "\"Ich möchte dir erneut gratulieren!\"",
/*MISSING*/    "\"I would like to congratulate you again!\"",
    }},
  {0x7129ca94, { // "inventory"
   "Twoje sfery",
/*MISSING*/    "inventory",
   "Tvé Sféry",
   "твои сферы",
/*MISSING*/    "inventory",
/*MISSING*/    "inventory",
    }},
  {0x7147a5e0, { // "Saved the Princess!"
   "Uratowano Księżniczkę!",
/*MISSING*/    "Saved the Princess!",
   "Zachránil jsi Princeznu!",
   "Принцесса спасена!",
/*MISSING*/    "Saved the Princess!",
/*MISSING*/    "Saved the Princess!",
    }},
  {0x716d2777, { // "You summon a bonfire!"
   "Przywoła%łeś0 ognisko!",
   "Bir kamp ateşi oluşturdun!",
   "Vyvola%l0 jsi ohniště!",
   "Вы призвали костры!",
   "Du beschwörst ein Leuchtfeuer!",
   "Você invoca uma fogueira!",
    }},
  {0x719ae65a, { // "If you collect too many treasures in a given land, it will become extremely dangerous. Try other lands once you have enough!"
   "Jeśli zbierzesz za dużo skarbów w jednej krainie, stanie się ona bardzo niebezpieczna. Spróbuj pójść do innych krain, gdy masz dość!",
/*MISSING*/    "If you collect too many treasures in a given land, it will become extremely dangerous. Try other lands once you have enough!",
   "Pokud získáš pøíliš mnoho pokladù v jednom kraji, zaène být velmi nebezpeèný. Až budeš mít dost, zkus jiné kraje!",
   "Если ты соберёшь слишком много сокровищ в одной земле, она станет весьма опасной. Попробуй другую землю, когда наберёшь достаточно сокровищ здесь!",
/*MISSING*/    "If you collect too many treasures in a given land, it will become extremely dangerous. Try other lands once you have enough!",
/*MISSING*/    "If you collect too many treasures in a given land, it will become extremely dangerous. Try other lands once you have enough!",
    }},
  {0x71b2d04a, { // "rotate view"
   "obróć widok",
   "görüntüyü çevir",
   "otočení hrací plochy",
   "повернуть вид",
   "Ansicht rotieren",
/*MISSING*/    "rotate view",
    }},
  {0x720fa237, { // "Monsters slain by you in melee are turned into friendly ghosts. Does not affect plants and friends."
   "Potwory zabite przez Ciebie wręcz są zamieniane w przyjazne duchy. Nie wpływa na rośliny i przyjaciół.",
/*MISSING*/    "Monsters slain by you in melee are turned into friendly ghosts. Does not affect plants and friends.",
   "Netvoři, které zabiješ v boji na blízko, se promění v přátelské duchy. Nepůsobí na rostliny a přátelské netvory.",
   "Убитые Вашими руками монстры превращаются в мирных призраков. Не действует на растения и союзников.",
   "Monster, die du im Nahkampf getötet hast, werden in freundliche Geister verwandelt. Es hat keinen Effekt auf Pflanzen und Freunde.",
/*MISSING*/    "Monsters slain by you in melee are turned into friendly ghosts. Does not affect plants and friends.",
    }},
  {0x721a70de, { // "An Air Elemental looks like a live tornado. Once you are three (or less) cells from it, it is impossible to move closer, due to strong winds. You can stand and wait, though. This also affects most monsters."
   "Powietrzny Żywiołak wygląda jak żywa trąba powietrzna. Gdy znajdziesz się trzy lub mniej pól od niego, nie możesz się zbliżyć bardziej, ze względu na silny powiew. Możesz jednak stać w miejscu i czekać. Powietrzny Żywiołak wpływa także na potwory.",
   "Hava Unsuru canlı bir hortuma benziyor. Eğer ondan üç ya da daha az uzaklıktaysan, güçlü rüzgârlardan ötürü daha fazla yakınlaşman mümkün değildir. Geçip gitmesini bekleyebilirsin elbette. Bu çoğu canavarı da etkiler. ",
   "Elementál Vzduchu vypadá jako oživlé tornádo. Pokud jsi od něj tři nebo méně políček, silný vítr ti brání v tom, aby ses k němu přiblížil. Můžeš ovšem stát a čekat. Tento efekt také ovlivňuje většinu netvorů.",
   "Воздушные элементаль похож на живое торнадо. Если Вы находитесь от него на расстоянии 3 или менее Вы не можете подойти ближе из-за сильного ветра. Но можно стоять и ждать. Это действует на всех монстров.",
   "Ein Luftelementar sieht aus wie ein lebender Tornado. Sobald man 3 Zellen oder weniger von ihm entfernt ist, ist es dank der starken Winde unmöglich näher zu kommen. Du kannst allerdings stehen und warten.",
/*MISSING*/    "An Air Elemental looks like a live tornado. Once you are three (or less) cells from it, it is impossible to move closer, due to strong winds. You can stand and wait, though. This also affects most monsters.",
    }},
  {0x72211c18, { // "Come on, this is too easy... find a bigger castle!"
   "Eee, to będzie za łatwe... znajdź większy zamek!",
   "Pöfff, çok kolaydı... daha büyük bir kale bul!",
   "Ale no tak, tohle je příliš jednoduché... najdi si větší hrad!",
   "Ну давай, это слишком просто... Найди больший замок!",
   "Komm schon, das ist viel zu einfach... suche ein größeres Schloss!",
/*MISSING*/    "Come on, this is too easy... find a bigger castle!",
    }},
  {0x722bdd3b, { // "In the peaceful mode, you just explore the world, without any battles; there are also several navigational puzzles available. In the memory game, you have to collect as many Dodecahedra as you can, and return to the starting point -- hyperbolic geometry makes this extremely difficult! Other hyperbolic puzzles include the Burial Grounds (excavate the treasures using your magical sword), Galápagos (try to find an adult tortoise matching the baby), Camelot (find the center of a large hyperbolic circle), and Palace (follow the mouse). Other places listed are for exploration."
   "W trybie spokojnym zwiedzasz świat bez żadnej walki; jest też tu kilka hiperbolicznych zagadek nawigacyjnych. W grze pamięciowej musisz zebrać jak najwięcej Dwunastościanów i wrócić do punktu wyjścia -- geometria hiperboliczna powoduje, że jest to bardzo trudne! Pozostałe zagadki to Kurhany (wykopuj skarby swoim magicznym mieczem), Galapágos (znajdź dużego żółwia odpowiadającego żółwiątku), Camelot (znajdź środek wielkiego hiperbolicznego kółka), i Pałac (idź za myszą). Można też zwiedzić kilka innych miejsc.",
/*MISSING*/    "In the peaceful mode, you just explore the world, without any battles; there are also several navigational puzzles available. In the memory game, you have to collect as many Dodecahedra as you can, and return to the starting point -- hyperbolic geometry makes this extremely difficult! Other hyperbolic puzzles include the Burial Grounds (excavate the treasures using your magical sword), Galápagos (try to find an adult tortoise matching the baby), Camelot (find the center of a large hyperbolic circle), and Palace (follow the mouse). Other places listed are for exploration.",
   "V klidném módu můžeš prostě jen zkoumat svět bez jakýchkoli bitev; je v něm také k dispozici několik navigačních hádanek. V paměťové hře musíš sesbírat co možná nejvíc Dvanáctistěnů a vrátit se do výchozího bodu -- to je v hyperbolické geometrii nesmírně obtížné! Další hyperbolické hádanky zahrnují Pohřebiště (vykopávání pokladů magickým mečem), Galapágy (hledání dospělé želvy, která odpovídá tvé želvičce), Camelot (nalezení středu velkého hyperbolického kruhu) a Palác (sledování myši). Jiná uvedená místa se dají zkoumat.",
   "В мирном режиме ты можешь исследовать мир без битв; здесь также доступно несколько навигационных головоломок. В игре 'память' ты должен собрать как можно больше додекаэдров и вернуться в точку старта -- гиперболическая геометрия делает это крайне сложным! Другие паззлы включают Курганы (выкопай сокровища с помощью магического меча), Галапагосы (найди взрослую черепаху, соответствующую маленькой черепашке), Камелот (найди центр большого гиперболического круга) и Дворец (следуй за мышкой). Остальные места перечислены для изучения.",
/*MISSING*/    "In the peaceful mode, you just explore the world, without any battles; there are also several navigational puzzles available. In the memory game, you have to collect as many Dodecahedra as you can, and return to the starting point -- hyperbolic geometry makes this extremely difficult! Other hyperbolic puzzles include the Burial Grounds (excavate the treasures using your magical sword), Galápagos (try to find an adult tortoise matching the baby), Camelot (find the center of a large hyperbolic circle), and Palace (follow the mouse). Other places listed are for exploration.",
/*MISSING*/    "In the peaceful mode, you just explore the world, without any battles; there are also several navigational puzzles available. In the memory game, you have to collect as many Dodecahedra as you can, and return to the starting point -- hyperbolic geometry makes this extremely difficult! Other hyperbolic puzzles include the Burial Grounds (excavate the treasures using your magical sword), Galápagos (try to find an adult tortoise matching the baby), Camelot (find the center of a large hyperbolic circle), and Palace (follow the mouse). Other places listed are for exploration.",
    }},
  {0x7267229c, { // "Yet another model, which corresponds to orthographic projection of the sphere. Poincaré disk model, Beltrami-Klein model, and the Gans model are all obtained by looking at either the hyperboloid model or an equidistant surface from various distances."
   "Kolejny model, odpowiadający ortograficznemu rzutowi sfery. Model dysku, Kleina i Gansa powstają przez patrzenie na model hiperboloidy lub powierzchnię ekwidystantną z różnych odległości.",
/*MISSING*/    "Yet another model, which corresponds to orthographic projection of the sphere. Poincaré disk model, Beltrami-Klein model, and the Gans model are all obtained by looking at either the hyperboloid model or an equidistant surface from various distances.",
   "A tady máme další model, který odpovídá ortografické projekci koule. Poincerého disk, Beltrami-Kleinův model i Gansův model se získají tak, že se z různých vzdáleností díváte na hyperboloidový model nebo na ekvidistantní plochu.",
   "Ещё одна модель, соответствующая ортогональной проекции сферы. Модели диска, Клейна и Ганса могут быть получены проекциями из модели гиперболоида или из эквидистанты с разных расстояний.",
/*MISSING*/    "Yet another model, which corresponds to orthographic projection of the sphere. Poincaré disk model, Beltrami-Klein model, and the Gans model are all obtained by looking at either the hyperboloid model or an equidistant surface from various distances.",
/*MISSING*/    "Yet another model, which corresponds to orthographic projection of the sphere. Poincaré disk model, Beltrami-Klein model, and the Gans model are all obtained by looking at either the hyperboloid model or an equidistant surface from various distances.",
    }},
  {0x7271ee0d, { // " to submerge"
   " do zanurzenia",
/*MISSING*/    " to submerge",
   " do ponoøení",
   " на погружение",
/*MISSING*/    " to submerge",
/*MISSING*/    " to submerge",
    }},
  {0x72f22ebb, { // "continue game"
   "kontynuuj grę",
   "oyuna devam et",
   "pokračuj ve hře",
   "продолжить игру",
   "Weiter",
/*MISSING*/    "continue game",
    }},
  {0x730c071a, { // "Unbalanced shards in your inventory are dangerous."
   "Niebezpiecznie jest nosić niezrównoważone Okruchy.",
/*MISSING*/    "Unbalanced shards in your inventory are dangerous.",
   "Nevyrovnané Úlomky ve tvém inventáři jsou nebezpečné.",
   "Несбалансированность осколков в Вашем инвентаре опасна.",
   "Unbalancierte Scherben in deinem Inventar sind gefährlich.",
/*MISSING*/    "Unbalanced shards in your inventory are dangerous.",
    }},
  {0x731011ff, { // "The Hyperbugs are fighting!"
   "Hiperinsekty ze sobą walczą!",
   "Aşkınböcekler savaşıyor!",
   "Hyperhmyzáci bojují mezi sebou!",
   "Гипержуки сражаются!",
   "Die Hyperkäfer kämpfen!",
/*MISSING*/    "The Hyperbugs are fighting!",
    }},
  {0x731c0efe, { // "%The1 is destroyed by the forces of Life!"
   "Siły Życia zniszczyły %a1!",
   "Yaşam güçleri %a1 yok etti!",
   "Síla Života zničila %a1!",
   "Сила Жизни уничтожила %a1!",
   "Die Kräfte des Lebens zerstören %den1 %a1!",
/*MISSING*/    "%The1 is destroyed by the forces of Life!",
    }},
  {0x736ac465, { // "a = add v"
   "a = dodaj v",
   "a = ekle v",
   "a = přidej v",
   "a = добавить v",
   "a = hinzufügen v",
/*MISSING*/    "a = add v",
    }},
  {0x73713b3b, { // "You kill %the1 with a mental blast!"
   "Zabi%łeś0 %a1 mocą psychiczną!",
   "%a2 zihin patlamasıyla öldürdün!",
   "Zabi%l1 jsi %a2 mentálním útokem!",
   "Вы убили %a1 психическим взрывом!",
   "Du tötest %den1 %a1 mit einem mentalen Stoß!",
/*MISSING*/    "You kill %the1 with a mental blast!",
    }},
  {0x73962d4d, { // "You may be unable to leave %the1 if you are not careful!\n"
   "Jeśli nie uważasz, możesz nie być w stanie stąd uciec!",
/*MISSING*/    "You may be unable to leave %the1 if you are not careful!\n",
   "Pokud si nedáš pozor, možná už se odsud nedostaneš!",
   "Возможно, вы не сможете отсюда выбраться, если не будете осторожны!",
/*MISSING*/    "You may be unable to leave %the1 if you are not careful!\n",
/*MISSING*/    "You may be unable to leave %the1 if you are not careful!\n",
    }},
  {0x73fd37f2, { // "which coefficient"
   "który współczynnik",
/*MISSING*/    "which coefficient",
   "kterů koeficient",
   "какой коэффициент",
   "welcher Koeffizient",
/*MISSING*/    "which coefficient",
    }},
  {0x73fdab24, { // "Each orb type can be mirrored only once."
   "Każdy typ sfery może być odbity tylko raz.",
/*MISSING*/    "Each orb type can be mirrored only once.",
   "Každý typ Sféry lze zrcadlit pouze jednou.",
   "Каждый тип сфер может быть отражён только один раз.",
/*MISSING*/    "Each orb type can be mirrored only once.",
/*MISSING*/    "Each orb type can be mirrored only once.",
    }},
  {0x741dafec, { // "Customize character"
   "pokoloruj postać",
   "Karakteri düzenle.",
   "upravit postavu",
   "Настройка персонажа",
   "Character anpassen",
/*MISSING*/    "Customize character",
    }},
  {0x7466c9f2, { // "quick mouse"
   "szybka mysz",
/*MISSING*/    "quick mouse",
   "rychlá myš",
   "быстрая мышь",
/*MISSING*/    "quick mouse",
/*MISSING*/    "quick mouse",
    }},
  {0x7466d4c4, { // "With each %1 you collect..."
   "Z każd%ymą1 %abl1, które znajdujesz...",
   "%2 gittikçe daha tehlikeli bir hâl alıyor...",
   "S každ%ým1 %abl1, který sebereš...",
   "С каждым собранным %abl1...",
   "Mit jedem gesammelten %d1...",
   "Com cada %1 que você coleta...",
    }},
  {0x748de98c, { // " (press ESC for some hints about it)."
   " (naciśnij ESC by dostać wskazówki).",
   " (ESC ile hakkında yardıma ulaşabilirsin).",
   " (stiskni ESC pro trochu nápovědy).",
   " (нажмите ESC, чтобы получить подсказки).",
   " (Drücke ESC für Hinweise).",
   " (pressione ESC para dicas a respeito).",
    }},
  {0x74ae0912, { // "leave the game"
   "opuść grę",
/*MISSING*/    "leave the game",
   "vyskoč ze hry",
   "покинуть игру",
/*MISSING*/    "leave the game",
/*MISSING*/    "leave the game",
    }},
  {0x74bc9920, { // "You destroy %the1."
   "Zniszczy%łeś0 %a1.",
   "%a1 yok ettin.",
   "Zniči%l0 jsi %a1.",
   "Вы уничтожили %a1.",
   "Du zerstörst %den1 %a1.",
/*MISSING*/    "You destroy %the1.",
    }},
  {0x74c9faac, { // "%The1 tries to dismount you!"
   "%1 próbuje Cię zrzucić!",
/*MISSING*/    "%The1 tries to dismount you!",
   "%1 se tě snaží shodit!",
   "%1 пробует сбросить вас!",
   "%Der1 %1 versucht dich abzuwerfen!",
/*MISSING*/    "%The1 tries to dismount you!",
    }},
  {0x75123c27, { // "%The1 fills the hole!"
   "%1 zatka%ł1 dziurę!",
   "%1 deliği dolduruyor!",
   "%1 vyplni%l1 díru!",
   "%1 упал%E1 в дыру!",
   "%Der1 %1 füllt das Loch!",
/*MISSING*/    "%The1 fills the hole!",
    }},
  {0x7520df5b, { // "render spiral"
   "rysuj spiralę",
/*MISSING*/    "render spiral",
   "renderování spirály",
   "нарисовать спираль",
/*MISSING*/    "render spiral",
/*MISSING*/    "render spiral",
    }},
  {0x753099b4, { // "enable the Hypersian Rug mode"
   "włącz tryb hiperskiego dywanu",
/*MISSING*/    "enable the Hypersian Rug mode",
   "aktivuj mód hyperského koberce",
   "включить режим гиперсидского ковра",
   "Aktiviere den Hypersischen-Teppich-Modus",
/*MISSING*/    "enable the Hypersian Rug mode",
    }},
  {0x75875396, { // "This parameter affects the ball model the same way as the projection parameter affects the disk model."
   "Ten parametr wpływa na model kuli w ten sam sposób, co projekcja wpływa na model dysku.",
/*MISSING*/    "This parameter affects the ball model the same way as the projection parameter affects the disk model.",
   "Tento parametr ovlivňuje model koule stejným způsobem jako projekce model disku.",
   "Этот параметр влияет на модель сферы также, как параметр проекции влияет на модель диска.",
/*MISSING*/    "This parameter affects the ball model the same way as the projection parameter affects the disk model.",
/*MISSING*/    "This parameter affects the ball model the same way as the projection parameter affects the disk model.",
    }},
  {0x75c48ec5, { // "player 4 spin"
   "gracz 4 obrót",
/*MISSING*/    "player 4 spin",
   "hráč 4 otáčení",
   "игрок 4 вертит",
   "Spieler 4 drehen",
/*MISSING*/    "player 4 spin",
    }},
  {0x75d73de0, { // "azimuthal equidistant"
   "azymutalny, równe d",
/*MISSING*/    "azimuthal equidistant",
   "azimutální délkojevný",
   "азимутальная, сохранение расстояний",
   "mittabstandstreu azimutal",
/*MISSING*/    "azimuthal equidistant",
    }},
  {0x7623ae12, { // "The ivy kills %the1!"
   "Bluszcz zabił %a1!",
   "Sarmaşık %a1 öldürdü!",
   "Břečťan zabil %a1!",
   "Плющ убил %a1!",
   "Der Efeu tötet %den1 %a1!",
/*MISSING*/    "The ivy kills %the1!",
    }},
  {0x763d0f84, { // "The skeleton of a tree."
   "Szkielet drzewa.",
/*MISSING*/    "The skeleton of a tree.",
   "Skelet stromu.",
   "Скелет дерева.",
   "Das Skelett eines Baumes.",
/*MISSING*/    "The skeleton of a tree.",
    }},
  {0x766ddf1f, { // "Bonus land, available only in some special modes.\n"
   "Kraina dodatkowa, dostępna tylko w niektórych trybach.\n",
/*MISSING*/    "Bonus land, available only in some special modes.\n",
   "Bonusový kraj, který je dostupný pouze v některých speciálních módech.\n",
   "Дополнительная земля, доступна только в дополнительных режимах.\n",
   "Bonusland, nur in einigen Spezialmodi verfügbar.\n",
/*MISSING*/    "Bonus land, available only in some special modes.\n",
    }},
  {0x76c3bfe1, { // "lose all treasure"
   "utrata skarbów",
   "tüm hazineyi kaybet",
   "zbav se všech pokladů",
   "потерять все сокровища",
   "verliere alle Schätze",
/*MISSING*/    "lose all treasure",
    }},
  {0x76ddb79b, { // "Mighty warriors from the Fjord, who can use boats."
   "Potężni wojownicy z Fiordu, używający łódek.",
   "Fiyortun kayık kullanabilen kudretli savaşçıları.",
   "Mocní válečníci z Fjordu, kteří se dokážou plavit na člunech.",
   "Могучие воины из Фьорда, способные использовать лодки.",
   "Mächtige Krieger die von den Fjorden stammen und Boote benutzen.",
/*MISSING*/    "Mighty warriors from the Fjord, who can use boats.",
    }},
  {0x771d2159, { // "character"
   "postać",
   "karakter",
   "postava",
   "персонаж",
   "Character",
   "personagem",
    }},
#if REPEATED
  {0x771d215a, { // "character [repeat]"
   "postać",
/*MISSING*/    "character [repeat]",
   "postava",
   "персонаж",
   "Rolle",
/*MISSING*/    "character [repeat]",
    }},
#endif
  {0x77738337, { // "compass size"
   "rozmiar kompasu",
/*MISSING*/    "compass size",
   "velikost kompasu",
   "размер компаса",
/*MISSING*/    "compass size",
/*MISSING*/    "compass size",
    }},
  {0x77837aed, { // "\"I enjoy watching the hyperbug battles.\""
   "\"Lubię oglądać bitwy hiperinsektów.\"",
   "\"Aşkınböcek savaşlarını izlemek hoşuma gidiyor.\"",
   "\"Rád se dívám na bitvy Hyperhmyzáků.\"",
   "\"Мне нравится наблюдать за боями гипержуков.\"",
   "\"Ich mag es, den Hyperkäfern beim Kämpfen zuzusehen.\"",
/*MISSING*/    "\"I enjoy watching the hyperbug battles.\"",
    }},
  {0x77aad383, { // "vector graphics editor -- press F1 for help"
   "edytor grafiki -- wciśnij F1, by zobaczyć pomoc",
   "vektör grafik editörü -- yardım için F1'e bas.",
   "editor vektorové grafiki -- nápovědu zobrazíš stiskem F1",
   "графический редактор -- нажмите F1, чтобы получить помощь",
   "Vektor Grafiken Editor -- drücke F1 um Hilfe zu erhalten",
/*MISSING*/    "vector graphics editor -- press F1 for help",
    }},
  {0x77ac3c91, { // "A piece of a magic mirror, or a mirage cloud, that can be used for magical purposes. Only mirrors and clouds in the Land of Mirrors leave these."
   "Odłamek szkła z magicznego lustra albo fragment mirażowej chmury. Może zostać użyty do magii. Tylko lustra i chmury z Krainy Luster zostawiają odłamki.",
   "Sihirli aynanın veya yansı bulutunun bir parçası. Sihirli amaçlar için kullanılabilir. Sadece aynalar ve bulutlardan düşer.",
   "Kousek kouzelného zrcadla nebo oblaku přeludů, který je možné používat pro magické účely.Zanechávají je po sobě pouze zrcadla a oblaky v Kraji zrcadel.",
   "Часть магического зеркала или облака миражей. Может быть использовано в магических целях. Может быть получено только в Зеркальной земле.",
   "Ein Stück eines Zauberspiegels oder einer Illusionswolke, das für magische Zwecke verwendet werden kann. Ausschließlich Zauberspiegel und Illusionswolken im Land der Spiegel hinterlassen so etwas.",
/*MISSING*/    "A piece of a magic mirror, or a mirage cloud, that can be used for magical purposes. Only mirrors and clouds in the Land of Mirrors leave these.",
    }},
  {0x77b3f296, { // "eight domains"
   "osiem domen",
   "sekiz alan",
   "osm domén",
   "восемь областей",
   "Acht Bereiche",
/*MISSING*/    "eight domains",
    }},
  {0x77ca7f5b, { // "Welcome to the Shoot'em Up mode!"
   "Witaj w trybie strzelanki!",
   "Witaj w trybie strzelanki!",
   "Vítej v módu střílečky!",
   "Добро пожаловать в режим стрельбы!",
   "Willkommen im Ballermodus!",
/*MISSING*/    "Welcome to the Shoot'em Up mode!",
    }},
  {0x77d6ace1, { // "set 3D monsters or walls in basic config first"
   "najpierw ustaw 3D potwory/ściany w konfiguracji podstawowej",
/*MISSING*/    "set 3D monsters or walls in basic config first",
   "nejprve musíš nastavit 3D netvory nebo zdi v základní konfiguraci",
   "сначала выставлять монстров или стены в режиме 3D ",
/*MISSING*/    "set 3D monsters or walls in basic config first",
/*MISSING*/    "set 3D monsters or walls in basic config first",
    }},
  {0x784ca1bd, { // "%The1 is extinguished!"
   "%2 ugasi%ł2 %a1!",
   "%1 söndü!",
   "Elementál Vody uhasil %a1!",
   "%1 гаснет!",
   "%Der1 %a1 wurde ausgelöscht!",
/*MISSING*/    "%The1 is extinguished!",
    }},
  {0x786da209, { // "It would be impolite to land on the table!"
   "Byłoby niegrzecznie wylądować na stole!",
   "Masaya çıkmak çok yakışıksız olur!",
   "Přístát na stole by bylo nezdvořilé!",
   "Было бы невежливо приземляться на стол!",
   "Es wäre unhöflich auf dem Tisch zu landen.",
/*MISSING*/    "It would be impolite to land on the table!",
    }},
  {0x787b6ed1, { // "This wall is quite strong. You will need another way in."
   "Ściana kurhanu jest dosyć mocna, musisz poszukać innego wejścia.",
/*MISSING*/    "This wall is quite strong. You will need another way in.",
   "Tato stěna je velice pevná. Musíš najít jinou cestu dovnitř.",
   "Эта стена довольно крепкая. Вам нужен другой путь.",
   "Diese Wand ist sehr stabil - du brauchst einen anderen Weg hinein.",
/*MISSING*/    "This wall is quite strong. You will need another way in.",
    }},
  {0x78b4341e, { // "These statues of Cthulhu are too large to carry, and they don't look too valuable anyway. Most monsters will never go through them... they probably have their reasons. But you can go! When you move into the cell containing a statue, you push the statue to the cell you left.\n"
   "Te pomniki Cthulhu są za duże, by je nosić, zresztą i tak nie wyglądają na zbyt cenne. Potwory nigdy przez nie nie przechodzą... pewnie mają swoje powody. Ale Ty możesz przejść! Kiedy wchodzisz na pole z pomnikiem, automatycznie przesuwasz pomnik na pole, które opuści%łeś0.",
   "Cthulhu putları taşımak için fazla büyük, zaten çok değerli de gözükmüyorlar. Canavarların çoğu onların içinden geçmiyor, bunun bir sebebi olmalı. Ama sen bunların içinden geçebilirsin! Putun olduğu hücreye girdiğinde putu terkettiğin hücreye itersin. \n",
   "Tyto Cthulhuovy sochy jsou příliš velké na to, abys je unesl, a navíc ani nevypadají příliš cenné. Většina netvorů skrz ně nikdy nepůjde... asi mají své důvody. Ale ty přes ně chodit můžeš! Když vstoupíš na políčko se sochou, odstrčíš ji na to políčko, které jsi opusti%l0.",
   "Эти статуи Ктулху слишком велики, чтобы их забрать, да и не выглядят особо ценными.  Большинство монстров не может ходить через них, у них есть на то свои причины. Но Вы можете! Когда Вы ходите на клетку с памятником, он сдвигается на клетку, которую Вы покинули.",
   "Diese Cthulhu-Statuen sind zu groß um sie zu tragen - und sehen sowieso nicht sonderlich wertvoll aus. Die meisten Monster gehen nie an ihnen vorbei, dafür haben sie wohl ihre Gründe. Du aber schon! Wenn du in die Zelle gehst, in der eine steht, schiebst du sie an deinen Ursprungsort.",
/*MISSING*/    "These statues of Cthulhu are too large to carry, and they don't look too valuable anyway. Most monsters will never go through them... they probably have their reasons. But you can go! When you move into the cell containing a statue, you push the statue to the cell you left.\n",
    }},
  {0x7928529a, { // "\n\nCollect %the1 to gain an extra Orb of the Mirror. You can gain further Orbs of the Mirror by collecting 2, 4, 8..."
   "\n\nZdobądź %a1 by dostać dodatkową Sferę Lustra. Więcej Sfer Lustra dostaniesz przy 2, 4, 8...",
/*MISSING*/    "\n\nCollect %the1 to gain an extra Orb of the Mirror. You can gain further Orbs of the Mirror by collecting 2, 4, 8...",
   "\n\nKdyž získáš %a1, dostaneš další Sféru Zrcadla. Další Sféry Zrcadla pak mùžeš získat za 2, 4, 8...",
   "\n\nСобери %a1, чтобы получить дополнительную Сферу зеркала. Ты получишь следующую сферу, когда соберёшь 2, 4, 8...",
/*MISSING*/    "\n\nCollect %the1 to gain an extra Orb of the Mirror. You can gain further Orbs of the Mirror by collecting 2, 4, 8...",
/*MISSING*/    "\n\nCollect %the1 to gain an extra Orb of the Mirror. You can gain further Orbs of the Mirror by collecting 2, 4, 8...",
    }},
  {0x7933ecb8, { // " for Android"
   " na Android",
   " Android için",
   " na Android",
   " на Android",
   " für Android",
   " para Android",
    }},
  {0x796f67a0, { // "Annoying insects. They can awaken Sleeping Bulls."
   "Irytujące owady. Mogą budzić Śpiące Byki.",
/*MISSING*/    "Annoying insects. They can awaken Sleeping Bulls.",
   "Otravný hmyz. Dokáže probudit Spící Býky.",
   "Надоедливые насекомые. Могут будить Спящих быков.",
/*MISSING*/    "Annoying insects. They can awaken Sleeping Bulls.",
/*MISSING*/    "Annoying insects. They can awaken Sleeping Bulls.",
    }},
  {0x799ed964, { // "This combination is known to be buggy at the moment."
   "Ta kombinacja opcji obecnie działa błędnie.",
/*MISSING*/    "This combination is known to be buggy at the moment.",
   "Tato kombinace možností v současné době nepracuje správně.",
   "Эта комбинация пока что работает с ошибками.",
/*MISSING*/    "This combination is known to be buggy at the moment.",
/*MISSING*/    "This combination is known to be buggy at the moment.",
    }},
  {0x79a71f0c, { // "monsters destroyed: %1"
   "potwory zniszczone: %1",
/*MISSING*/    "monsters destroyed: %1",
   "zničení nepřátelé: %1",
   "монстров уничтожено: %1",
   "zerstörte Monster: %1",
/*MISSING*/    "monsters destroyed: %1",
    }},
  {0x79afcedc, { // "You cannot defeat the Greater Demon yet!"
   "Nie możesz jeszcze pokonać Wielkiego Demona!",
   "Büyük Şeytanı henüz yenemezsiniz!",
   "Ďábla ještě porazit nemůžeš!",
   "Вы пока не можете победить Старшего демона!",
   "Du kannst den Mächtigen Dämon noch nicht besiegen!",
   "Você ainda não pode derrotar o Demônio Maior!",
    }},
  {0x79c0f4f4, { // "%The1 is filled!"
   "%1 zosta%ł1 zalan%1y!",
/*MISSING*/    "%The1 is filled!",
   "%1 je naplněn%1ý!",
   "%1 наполнил%c1!",
/*MISSING*/    "%The1 is filled!",
/*MISSING*/    "%The1 is filled!",
    }},
  {0x79c4823b, { // "hexagonal"
   "sześciokąt",
   "altıgensel",
   "šestiúhelníky",
   "шестиугольник",
   "hexagonal",
/*MISSING*/    "hexagonal",
    }},
  {0x79e5a8d7, { // " kills: %1"
   " zabicia: %1",
   " leşler: %1",
   " zabito nepřátel: %1",
   " убийств: %1",
   " Kills: %1",
/*MISSING*/    " kills: %1",
    }},
  {0x79f64007, { // "You have gained an offensive power!"
   "Zdobywasz moc ofensywną!",
/*MISSING*/    "You have gained an offensive power!",
   "Získal jsi útočnou schopnost!",
   "Ты получил атакующую силу!",
/*MISSING*/    "You have gained an offensive power!",
/*MISSING*/    "You have gained an offensive power!",
    }},
  {0x7a2f868f, { // "configure panning"
   "skonfiguruj przewijanie",
   "kaydırmayı ayarla",
   "konfigurace scrollování",
   "настройки прокрутки",
   "Schwenken konfigurieren",
/*MISSING*/    "configure panning",
    }},
  {0x7a330767, { // "Evil people are mining for emeralds in this living cave. It does not grow naturally, but it is dug out in a regular pattern, which is optimal according to the evil engineers."
   "Źli ludzie szukają szmaragdów w tej żywej jaskini. Nie rośnie ona naturalnie -- ludzie wykopują regularne wzory, które, zgodnie z oceną złych inżynierów, są 'optymalne'. ",
   "Kötü insanlar bu mağarada zümrüt çıkarıyor. Doğal yollarla çıkmıyor, şeytani mühendislerin işine gelecek şekilde düzenli bir şemayla çıkarılıyor.",
   "V této živoucí jeskyni těží zlí lidé smaragdy. Neroste přirozeně, ale je vykopána v pravidelném vzoru, který zlí inženýři pokládají za optimální.",
   "Злые люди искали изумруды в Живой пещере. Она не растёт сама, но выкопана периодически оптимально для нужд злых инженеров. ",
   "Böse Leute bauen Smaragde in dieser lebenden Höhle ab. Sie entsteht nicht natürlich, sondern wird in einem Muster ausgebaut das die bösen Arbeiter als optimal ansehen.",
/*MISSING*/    "Evil people are mining for emeralds in this living cave. It does not grow naturally, but it is dug out in a regular pattern, which is optimal according to the evil engineers.",
    }},
  {0x7a64eea8, { // ", %1"
   ", %1",
/*MISSING*/    ", %1",
   ", %1",
   ", %1",
   ", %1",
/*MISSING*/    ", %1",
    }},
  {0x7a696805, { // "w = walls"
   "w = ściany",
   "w = duvar",
   "b = hranice",
   "m = монстры",
   "w = Wände",
/*MISSING*/    "w = walls",
    }},
  {0x7a727f8b, { // "%The1 bites %the2!"
   "%1 ugryz%ł1 %a2!",
   "%1 %a2 ısırdı!",
   "%1 kous%l1 %2!",
   "%1 ударил%E1 %a2!",
   "%Der1 %1 beißt %den2 %a2!",
/*MISSING*/    "%The1 bites %the2!",
    }},
  {0x7a94214b, { // "In the shoot'em up mode, space and time is continuous. You attack by throwing knives. Very fun with two players!\n\nThere are other special modes too which change the gameplay or focus on a particular challenge."
   "W trybie strzelanki czas i przestrzeń są ciągłe. Atakujesz rzucając nożami. Bardzo fajne na dwóch graczy!\n\nJest wiele trybów specjalnych, zmieniających rozgrywkę lub koncentrujących się na konkretnym wyzwaniu.",
/*MISSING*/    "In the shoot'em up mode, space and time is continuous. You attack by throwing knives. Very fun with two players!\n\nThere are other special modes too which change the gameplay or focus on a particular challenge.",
   "V módu střílečky je prostor i čas spojitý. Útočíš vrháním nožů. Veliká legrace ve dvou hráčích!\n\nHyperRogue také obsahuje další speciální módy, které mění způsob hry nebo jsou zaměřené na určitý úkol.",
   "В режиме стрельбы пространство и время непрерывны. Ты атакуешь, кидая ножи. Очень весело для двоих игроков!\n\nЕсть и другие специальные режимы, изменяющие игру или концентрирующиеся на определённых задачах.",
/*MISSING*/    "In the shoot'em up mode, space and time is continuous. You attack by throwing knives. Very fun with two players!\n\nThere are other special modes too which change the gameplay or focus on a particular challenge.",
/*MISSING*/    "In the shoot'em up mode, space and time is continuous. You attack by throwing knives. Very fun with two players!\n\nThere are other special modes too which change the gameplay or focus on a particular challenge.",
    }},
  {0x7a952f83, { // "Chaos mode"
   "tryb Chaosu",
/*MISSING*/    "Chaos mode",
   "chaotický mód",
   "режим Хаоса",
   "Chaos-Modus",
/*MISSING*/    "Chaos mode",
    }},
  {0x7ad3304d, { // "length of a segment"
   "długość segmentu",
/*MISSING*/    "length of a segment",
   "délka segmentu",
   "длина отрезка",
   "Länge eines Segments",
/*MISSING*/    "length of a segment",
    }},
  {0x7ae6977d, { // "Apparently a prince is kept locked somewhere, but you won't ever find him in this hyperbolic palace. "
   "Gdzieś tu jest uwięziony książę, ale w tym hiperbolicznym pałacu nigdy go nie znajdziesz. ",
   "Görünen o ki, bir şehzade bir yerlerde saklı tutuluyor, ama bu hiperbolik sarayda onu bulman mümkün değil. ",
   "Někde je tu prý uvězněný princ, ale toho v tomhle hyperbolickém paláci nikdy nenajdeš. ",
   "Видимо, принц где-то заперт, но в гиперболическом дворце Вы его не найдёте. ",
   "Scheinbar ist irgendwo ein Prinz eingesperrt, aber du wirst ihn in diesem hyperbolischem Palast nie finden.",
/*MISSING*/    "Apparently a prince is kept locked somewhere, but you won't ever find him in this hyperbolic palace. ",
    }},
  {0x7b2a0f9e, { // "High quality shots not available on this platform"
   "Zrzuty HQ niedostępne na tej platformie",
   "Yüksek kalite resimler bu platformda mevcut değil",
   "HQ snímky nejsou na této platformě k dispozici",
   "Скрины высокого качества недоступны на этой платформе",
   "HQ Screenshots sind auf diesem Gerät nicht verfügbar",
/*MISSING*/    "High quality shots not available on this platform",
    }},
  {0x7b411bc8, { // "This Orb gives you a weapon made of pure magical energy. You do not hold it, it simply floats in the air next to you. When you go, the energy sword moves with you, pointing at the same relative angle it pointed before -- you cannot move or rotate it otherwise. Most monsters can be killed by moving the sword into them, and won't move into the spot with the sword."
   "Ta Sfera daje Ci ostrze zbudowaną z czystej magicznej energii. Nie trzymasz go w ręce, po prostu unosi się ono w powietrzu obok Ciebie. Kiedy się poruszasz, ostrze energii porusza się razem z Tobą, pozostając skierowane pod tym samym kątem względem Twojego ruchu, co wcześniej -- nie możesz nim obracać ani poruszać w inny sposób. Większość potworów ginie, gdy uderzysz w nie ostrzem, nie wejdą one też na pole, na którym ostrze jest.",
/*MISSING*/    "This Orb gives you a weapon made of pure magical energy. You do not hold it, it simply floats in the air next to you. When you go, the energy sword moves with you, pointing at the same relative angle it pointed before -- you cannot move or rotate it otherwise. Most monsters can be killed by moving the sword into them, and won't move into the spot with the sword.",
   "Tato Sféra ti poskytne zbraň vyrobenou z čisté magické energie. Nedržíš ji -- prostě se jen vznáší ve vzduchu vedle tebe. Kamkoli půjdeš, energetický meč půjde s tebou a bude namířený ve stejném relativním úhlu -- jinak s ním nemůžeš pohybovat ani otáčet. Většinu netvorů lze zabít tím, že na jejich políčko posuneš meč, a samotní netvoři na políčko s mečem nevstoupí.",
   "Эта сфера даёт вам меч из чистой магической энергии. Вы не несёте его, он сам летит за вами. Когда вы идёте, он движется за вами, оставаясь под тем же углом относительно вас -- вы не можете двигать или вращать его иначе. Большинство монстров можно убить, передвинув на их клетку меч, и они не могут пойти на клетку с мечом.",
   "Dieser Orb stellt dir eine Waffe aus purer magischer Energie zur Verfügung. Du hältst dieses Schwert nicht; es schwebt einfach neben dir in der Luft. Wenn du dich bewegst, bewegt sich das Energieschwert mit dir, wobei es in dieselbe relative Richtung zeigt wie zuvor; du kannst es nicht in anderer Weise bewegen oder drehen. Die meisten Monster können getötet werden, indem das Schwert in sie hineinbewegt wird, und begeben sich nicht auf das Feld mit dem Schwert.",
/*MISSING*/    "This Orb gives you a weapon made of pure magical energy. You do not hold it, it simply floats in the air next to you. When you go, the energy sword moves with you, pointing at the same relative angle it pointed before -- you cannot move or rotate it otherwise. Most monsters can be killed by moving the sword into them, and won't move into the spot with the sword.",
    }},
  {0x7b5d1002, { // "F4 = extension"
   "F4 = rozszerzenie",
   "F4 = genişletme",
   "F4 = rozšíření",
   "F4 = расширение",
   "F4 = Erweiterung",
/*MISSING*/    "F4 = extension",
    }},
  {0x7b6549fd, { // "In this mode you can draw your own player character and floors. Mostly for the development purposes, but you can have fun too.\n\nf - floor, p - player (repeat 'p' for layers)\n\nn - new shape, u - copy the 'player body' shape\n\n1-9 - rotational symmetries, 0 - toggle axial symmetry\n\npoint with mouse and: a - add point, m - move nearest point, d - delete nearest point, c - nearest point again, b - add after nearest\n\ns - save in C++ format (but cannot be loaded yet without editing source)\n\nz - zoom, o - Poincaré model\n"
   "W tym trybie możesz narysować swoją własną postać albo podłogę. Głównie w celu rozwijania gry, ale Ty też możesz się pobawić.\n\nf - podłoga, p - gracz (powtarzaj, by zmienić warstwę)\n\nn - nowy kształt, u - skopiuj standardową postać gracza\n\n1-9 - symetrie obrotowe, 0 - symetria osiowa\n\nwskaż myszą i: a - dodaj punkt, m - przesuń najbliższy punkt, d - skasuj najbliższy punkt, c - powtórz najbliższy punkt, b - po najbliższym\n\ns - zapisz w formacie C++ (niestety obecnie nie można wczytać bez edytowania źródeł gry)\n\nz - powiększenie, o - model Poincaré\n",
   "Bu modda, kendi karakterini ve zemini çizebilirsin. Aslında geliştirme amaçları için vri ama eğlence için de kullanabilirsin.\n\nf - zemin, p - oyuncu (katman eklemek için 'p'ye tekrar bas)\n\nn - yeni şekil, u - oyuncu şeklini kopyala\n\n1-9 - döngüsel simetriler, 0 - eksensel simetriyi aç/kapa\n\nfareyle göster ve: a - nokta ekle, m - en yakındaki noktaya git, d - en yakındaki noktayı sil, c - tekrar en yakın nokta, b - en yakından sonraya ekle\n\ns - C++ formatında sakla (ancak kaynağı düzenlemeden yüklenemez)\n\nz - yakınlaş, o - Poincaré modeli\n",
   "V tomto módu můžeš nakreslit vlastní postavu a podlahu. Slouží především pro vývoj hry, ale můžeš se s ním i pobavit.\n\nf - podlaha, p - postava (opakovaným stiskem 'p' přepínáš mezi vrstvami)\n\nn - nový tvar, u - zkopíruj tvar 'tělo postavy'\n\n1-9 - rotační symetrie, 0 - zapnout/vypnout osovou symetrii\n\nzamiř myší a: a - přidej bod, m - přesuň nejbližší bod, d - vymaž nejbližší bod, c - znovu nejbližší bod, b - přidej za nejbližším\n\ns - ulož ve formátu C++ (bez editování zdroje však prozatím soubor nelze nahrát)\n\nz - přiblížení/vzdálení, o - Poincarého model\n",
   "В этом режиме Вы можете рисовать персонажа и всё окружение. В основном для нужд разработчиков, но и просто для веселья тоже.\n\nf - пол, p - игрок (повторите для смены слоя)\n\nn - новая форма, u - копировать форму игрока\n\n1-9 - повороты, 0 - осевая симметрия\n\nРасставьте точки мышкой и: a - добавить точку, m - двигать ближайшую точку, d - удалить ближайшую точку, c - снова ближайшая точка, b - после ближайшей\n\ns - запись в формате C++ (пока не может быть загружено без редактирования исходников)\n\nz - увеличение, o - модель Пуанкаре\n",
   "Hier kannst du deine eigenen Spielfiguren/Bodenplatten zeichnen. Primär zur Entwicklung, aber hab Spaß.\n\nf - Boden, p - Spieler (wiederholt für Schichten)\n\nn - neue Form, u - kopiere die Körperform des Spielers\n\n1-9 - Dreh-Symmetrie, 0 - Achsen-Symmetrie umschalten\n\nMaus darüber und: a - Punkt hinzufügen, m - Nächsten Punkt verschieben, d - Nächsten Punkt löschen, c - Nächsten Punkt kopieren, b - Nach Nächstem Punkt hinzufügen\n\ns - im C++-Format speichern (kannst du aber noch nicht laden, ohne Quellcode zu ändern)\n\nz - Zoom, o - Poincaré-Modell\n",
   "Nesse modo você pode desenhar seu próprio personagem e andares. Feito principalmente para motivos de desenvolvimento, mas você pode se divertir também.\n\nf - andar, p - personagem (repetir 'p' para camadas)\n\nn - nova forma, u - copiar a forma do 'corpo do personagem'\n\n1-9 - rotação simétrica, 0 - liga/desliga simetria axial\n\napontar com o mouse e: a - adicionar ponto, m - mover ponto mais próximo, d - apagar ponto mais próximo, c - ponto mais próximo novamente, b - adicionar após o mais perto\n\ns - salvar em formato C++ (não pode ser carregado sem edição do arquivo fonte)\n\nz - zoom, o - modelo Poincaré\n",
    }},
  {0x7b97b391, { // "display pattern codes (simplified)"
   "pokaż kody wzorku (uproszczone)",
   "doku kodlarını göster (basitleştirilmiş)",
   "zobraz kódy vzoru (zjednodušené)",
   "показать коды узоров (упрощённые)",
   "Muster Codes anzeigen (vereinfacht)",
/*MISSING*/    "display pattern codes (simplified)",
    }},
  {0x7b98a273, { // "move down-left"
   "ruch w dół-lewo",
/*MISSING*/    "move down-left",
   "pohyb doleva dolů",
   "ход вниз-влево",
/*MISSING*/    "move down-left",
/*MISSING*/    "move down-left",
    }},
  {0x7bb124ed, { // "You need an Orb of Teleport to teleport."
   "Potrzebujesz Sfery Teleportacji, by to zrobić.",
   "Işınlanmak için Işınlanma Küresine ihtiyacın var.",
   "K teleportaci potřebuješ Sféru teleportace.",
   "Вам нужна сфера телепорта для телепортации.",
   "Du brauchst einen Orb der Teleportation um dich zu teleportieren.",
   "Você precisa de um Orbe do Teleporte para teleportar-se.",
    }},
  {0x7c18932f, { // "%The1 suffocates!"
   "%1 się udusi%ł1!",
/*MISSING*/    "%The1 suffocates!",
   "%1 leknu%l1!",
   "%1 задыхается!",
   "%Der1 %1 erstickt!",
/*MISSING*/    "%The1 suffocates!",
    }},
  {0x7c4d7cd4, { // "Welcome to HyperRogue!"
   "Witaj w HyperRogue!",
/*MISSING*/    "Welcome to HyperRogue!",
   "Vítej v HyperRogue!",
   "Добро пожаловать в HyperRogue!",
/*MISSING*/    "Welcome to HyperRogue!",
/*MISSING*/    "Welcome to HyperRogue!",
    }},
  {0x7c4edcbf, { // "Variants of %the1 are available in the Random Pattern Mode after getting a highscore of at least 10 %2."
   "Warianty krainy %1 są dostępne w trybie losowych wzorków po zdobyciu 10x %2.",
   "%1 çeşitlemeleri Rastgele Doku Modunda en az on tane %2 skoru elde ederek açılır. ",
   "Varianty kraje '%1' jsou dostupné v Módu náhodných vzorů po získání nejméně 10x %2.",
   "Варианты земли %1 доступны в редиме случайного узора при рекорде не менее 10x %2.",
   "Varianten der %1 sind im Zufallsmuster-Modus verfügbar, nachdem eine Punktzahl von 10x %2 erreicht wurde.",
/*MISSING*/    "Variants of %the1 are available in the Random Pattern Mode after getting a highscore of at least 10 %2.",
    }},
  {0x7c4f337d, { // "Unlock this challenge by getting the Orb of Yendor!"
   "Odblokuj Misję Yendor, zdobywając Sferę Yendoru!",
   "Yendor'un Küresi'ni alarak bu ek görevi aktif et!",
   "Tuto misi odemkneš získáním Yendorské sféry!",
   "Откройте этот режим, собрав сферу Йендора!",
   "Finde einen Orb von Yendor um diese Herausforderung freizuschalten!",
/*MISSING*/    "Unlock this challenge by getting the Orb of Yendor!",
    }},
  {0x7c79f1cf, { // "summon Mimics"
   "przywołaj Mimiki",
   "Taklitçiler çıkar",
   "vyvolej Mimiky",
   "призвать двойников",
   "rufe Mimiken herbei",
/*MISSING*/    "summon Mimics",
    }},
  {0x7ce53856, { // "Did you know that the path you take during the game is usually very close to a straight line?\n"
   "Czy wiesz, że droga, którą przebywasz w czasie gry, jest zwykle bardzo zbliżona do linii prostej?",
/*MISSING*/    "Did you know that the path you take during the game is usually very close to a straight line?\n",
   "Vìdìl jsi, že cesta, kterou bìhem hry urazíš, je obvykle velmi blízká pøímce?",
   "Ты знал, что путь, который ты проходишь за игру, обычно очень близок к прямой линии?",
/*MISSING*/    "Did you know that the path you take during the game is usually very close to a straight line?\n",
/*MISSING*/    "Did you know that the path you take during the game is usually very close to a straight line?\n",
    }},
  {0x7ce5604c, { // "player 3 go"
   "gracz 3 idź",
/*MISSING*/    "player 3 go",
   "hráč 3 pohyb",
   "игрок 3 идёт",
   "Spieler 3 los",
/*MISSING*/    "player 3 go",
    }},
  {0x7d054f96, { // "You feel that the souls of slain enemies pull you to the Graveyard..."
   "Czujesz, jak dusze pokonanych wrogów wzywają Cię na Cmentarz...",
   "Katlettiğin düşmanların ruhlarının seni mezarlığa çektiğini hissediyorsun...",
   "Cítíš, že tě duše zabitých nepřátel táhnou na Hřbitov...",
   "Вы чувствуете, как души убитых врагов тянут Вас на Кладбище.",
   "Du hast das Gefühl, dass die Seelen gefallener Gegner dich zum Friedhof führen...",
   "Você sente que as almas dos inimigos abatidos lhe puxam para o Cemitério...",
    }},
  {0x7d05b547, { // "left"
   "lewo",
/*MISSING*/    "left",
   "doleva",
   "лево",
   "links",
/*MISSING*/    "left",
    }},
  {0x7d13171b, { // "Fractal landscapes: application"
   "Zastosowanie fraktalnych krajobrazów",
/*MISSING*/    "Fractal landscapes: application",
   "Fraktální krajinky: aplikace",
   "Приложение фрактальных пейзажей",
/*MISSING*/    "Fractal landscapes: application",
/*MISSING*/    "Fractal landscapes: application",
    }},
  {0x7d13b44b, { // " (next at %1 to %2)"
   " (kolejny przy %1 do %2)",
/*MISSING*/    " (next at %1 to %2)",
   " (další pøi %1 až %2)",
   " (следующий при %1 в %2)",
/*MISSING*/    " (next at %1 to %2)",
/*MISSING*/    " (next at %1 to %2)",
    }},
  {0x7d14a77c, { // "Shift=random, Ctrl=mix"
   "Shift=losowo, Ctrl=miks",
   "Shift=rastgele, Ctrl=karışık",
   "Shift=náhodná, Ctrl=směs",
   "Shift=случайный, Ctrl=смешать",
   "Shift=Zufällig, Ctrl=Mischen",
/*MISSING*/    "Shift=random, Ctrl=mix",
    }},
  {0x7da15cb3, { // "Press F5 to clear the board."
   "Wciśnij F5, by wyczyścić planszę.",
   "Tahtayı temizlemek için F5'e bas",
   "F5 vymaže desku.",
   "Нажмите F5, чтобы очистить доску.",
   "Drücke F5 um das Spielfeld zu leeren.",
/*MISSING*/    "Press F5 to clear the board.",
    }},
  {0x7e199367, { // "Take a revolver, kill outlaws, collect bounties.\n\nNote: since this land is anachronistic, it is not available in normal game. It is only available in special modes."
   "Weź rewolwer, zabijaj bandytów i zgarniaj nagrody!\n\nUwaga: ta kraina jest anachronistyczna, także nie jest dostępna w normalnej grze. Jest za to dostępna w specjalnych trybach.",
   "Altıpatları al, haydutları öldür, başlarına konan ödülleri topla!\n\nNot: Bu diyar anakronistik olduğu için, normal oyunda mevcut değil, sadece özel modlarda oynanabilir.",
   "Vezmi si revolver, zabíjej psance, sbírej odměny.\n\nPoznámka: tento kraj je anachronistický, a proto není k dispozici v normální hře. Je dostupný pouze ve speciálních módech.",
   "Хватай револьвер, убивай бандитов, получай награды!\n\nЗамечание: Так как эта земля анахронистична, она не доступна в нормальной игре. Только в специальных режимах.",
   "Schnapp dir einen Revolver, erschieß Banditen und beanspruche das Kopfgeld!\n\nHinweis: Dieses Land ist nur in Spezialmodi verfügbar.",
/*MISSING*/    "Take a revolver, kill outlaws, collect bounties.\n\nNote: since this land is anachronistic, it is not available in normal game. It is only available in special modes.",
    }},
  {0x7e1a72ab, { // "Collect %1 $$$ to access more worlds"
   "Znajdź %1 $$$, by iść do nowych krain",
   "%1 $$$ toplayarak başka dünyalara eriş",
   "Další kraje zpřístupníš sesbíráním %1 $$$",
   "Соберите %1 $$$, чтобы открыть новые земли",
   "Sammle %1 $$$ um mehr Länder betreten zu können",
   "Colete %1 $$$, para acessar novos mundos",
    }},
  {0x7e9ca1ba, { // "Treasure required: %1 $$$.\n"
   "Wymagane skarby: %1 $$$.\n",
   "Hazine gerekiyor: %1 $$$.\n",
   "Potřebuješ %1 $$$.\n",
   "Сокровищ нужно: %1 $$$.\n",
   "Schätze benötigt: %1 $$$.\n",
/*MISSING*/    "Treasure required: %1 $$$.\n",
    }},
  {0x7ec15097, { // "draw circle around the target"
   "rysuj kółko wokół celu",
/*MISSING*/    "draw circle around the target",
   "nakresli kruh kolem cíle",
   "нарисуйте кружок вокруг цели",
/*MISSING*/    "draw circle around the target",
/*MISSING*/    "draw circle around the target",
    }},
  {0x7f0e5193, { // "three colors rotated"
   "trzy kolory obrócone",
/*MISSING*/    "three colors rotated",
   "tři barvy s otočkou",
   "три цвета повёрнутые",
/*MISSING*/    "three colors rotated",
/*MISSING*/    "three colors rotated",
    }},
  {0x7f368ee4, { // "0-9 = radius (%1)"
   "0-9 = promień (%1)",
   "0-9 = promień (%1)",
   "0-9 = poloměr (%1)",
   "0-9 = радиус (%1)",
   "0-9 = Radius (%1)",
/*MISSING*/    "0-9 = radius (%1)",
    }},
  {0x7f386b21, { // "Scale the displayed model."
   "Skaluj model na ekranie.",
/*MISSING*/    "Scale the displayed model.",
   "Změň měřítko zobrazeného modelu.",
   "Масштаб отображаемой модели.",
/*MISSING*/    "Scale the displayed model.",
/*MISSING*/    "Scale the displayed model.",
    }},
  {0x7f406896, { // "last on top"
   "ostatnie na górze",
/*MISSING*/    "last on top",
   "poslední nahoře",
   "последние сверху",
/*MISSING*/    "last on top",
/*MISSING*/    "last on top",
    }},
  {0x7f729005, { // "football"
   "piłka nożna",
/*MISSING*/    "football",
   "fotbal",
   "футбол",
/*MISSING*/    "football",
/*MISSING*/    "football",
    }},
  {0x7f734e80, { // "Don't be fooled by beauty, or you will be stabbed if you come too close!"
   "Nie daj się zwieść urodzie, bo jak się za bardzo zbliżysz, możesz zostać ugodzon%y0 nożem!",
/*MISSING*/    "Don't be fooled by beauty, or you will be stabbed if you come too close!",
   "Nenech se oklamat krásou -- když se moc přiblížíš, můžeš dostat ránu nožem!",
   "Не будьте обмануты красотой, если не хотите получить удар в спину!",
   "Lass dich von Schönheit nicht täuschen - oder du wirst erstochen, wenn du zu nah kommst!",
/*MISSING*/    "Don't be fooled by beauty, or you will be stabbed if you come too close!",
    }},
  {0x7fef823d, { // "The Overgrown Woods are filled with mutant ivies! These plants grow very fast. Each leaf, after being grown, can grow itself on the next turn. However, each part is only able to grow once in 16 turns. Outside of the Overgrown Woods, the Mutant Ivy may grow only on hexagonal cells.\n\nMaybe such fast growing plants could help you solve the problem of hunger in your world? Kill the Mutant Ivies to collect Mutant Saplings."
   "Przerośnięty Las jest pełen zmutowanego bluszczu! Te rośliny rosną bardzo szybko. Każdy liść może sam rosnąć w następnej turze po wyrośnięciu, ale każda część rośliny może rosnąć tylko raz na 16 tur. Poza Przerośniętym Lasem Zmutowany Bluszcz może rosnąć tylko na sześciokątach.\n\nMoże tak szybko rosnące rośliny pozwoliłyby Ci rozwiązać problem głodu w Twoim świecie? Zabijaj Zmutowane Bluszcze, by zbierać sadzonki.",
   "Azman Orman mutant sarmaşıklarla dolu! Bu bitkiler çok hızlı büyür. Her yaprak, büyüdükten sonra sıradaki tur kendi başına büyüyebilir. Ancak, her parça ancak 16 tur içinde bir kez büyüyebilir. Azman Orman dışındaki diyarlarda, Mutant Sarmaşık sadece altıgen hücrelerde büyüyebilir.\n\nBelki bu kadar hızlı büyüyen bitkiler dünyandaki açlık sorununu çözebilir? Mutant Fidanları toplamak için Mutant Sarmaşıkları öldür.",
   "Prales je plný mutantních břečťanů! Tyto rostliny velice rychle rostou. Jakmile nějaký list vyroste, hned v následujícím kole může sám růst. Každá část však může vyrůst pouze jednou za 16 kol. Mimo Prales může Mutantní břečťan růst pouze na šestiúhelníkových políčkách.\n\nRostliny s takovou rychlostí růstu by ti možná mohly pomoci vyřešit problém s hladem ve tvém světě. Zabíjej Mutantní břečťany, abys získal Mutantní semenáčky.",
   "Заросли полны плющами-мутантами! Они растут очень быстро. Каждый выросший побег может расти на следующем же ходу. Но расти в следующий раз он сможет лишь через 16 ходов. Вне Зарослей плющи-мутанты могут расти только на шестиугольниках.\n\nВозможно, столь быстрый рост решит проблему голода в Вашем мире? Убейте плющ, чтобы собрать саженец.",
   "Der verwilderte Wald ist voll mit mutiertem Efeu! Diese Pflanzen wachsen sehr schnell. Jedes Blatt kann, nachdem es gewachsen ist, im nächsten Zug selbst wachsen. Allerdings kann jeder Teil nur einmal alle 16 Züge austreiben. Außerhalb des verwilderten Walds wächst mutierter Efeu nur auf hexagonalen Zellen.\n\nVielleicht können die schnell wachsenden Pflanzen das Hungerproblem in deiner Welt lösen? Töte mutierten Efeu um mutierte Sätzlinge zu erhalten.",
/*MISSING*/    "The Overgrown Woods are filled with mutant ivies! These plants grow very fast. Each leaf, after being grown, can grow itself on the next turn. However, each part is only able to grow once in 16 turns. Outside of the Overgrown Woods, the Mutant Ivy may grow only on hexagonal cells.\n\nMaybe such fast growing plants could help you solve the problem of hunger in your world? Kill the Mutant Ivies to collect Mutant Saplings.",
    }},
  {0x800ee083, { // "%The1 drowns!"
   "%1 się utopi%ł1!",
   "%1 boğuldu!",
   "%1 se utopi%l1!",
   "%1 утонул%E1!",
   "%Der1 %1 ertrinkt!",
/*MISSING*/    "%The1 drowns!",
    }},
  {0x80512db6, { // "You wonder where all these monsters go, after their death..."
   "Gdzie te wszystkie potwory idą po śmierci?...",
   "Bütün bu canavarların ölümlerinden sonra nereye gittiklerini merak ediyorsun...",
   "Přemýšlíš, kam všichni ti netvoři odejdou, když zemřou...",
   "Вы задаётесь вопросом, куда же деваются эти монстры после смерти?",
   "Du wunderst dich, was mit den Monstern passiert nachdem sie sterben...",
   "Você se pergunta: Para onde vão todos esses monstros depois que morrem?...",
    }},
  {0x806ed3d1, { // "four elements"
   "cztery żywioły",
   "dört element",
   "čtyři živly",
   "четыре элемента",
   "Vier Elemente",
/*MISSING*/    "four elements",
    }},
  {0x807db083, { // "y: %1"
   "y: %1",
   "y: %1",
   "y: %1",
   "y: %1",
   "y: %1",
/*MISSING*/    "y: %1",
    }},
  {0x8084c224, { // "A beatiful purple gem from the Lost Mountain."
   "Piękny fioletowy kamień z Zagubionej Góry.",
/*MISSING*/    "A beatiful purple gem from the Lost Mountain.",
   "Krásný fialový kámen ze Ztracené hory.",
   "Прекрасный фиолетовый камень с Потерянной горы.",
/*MISSING*/    "A beatiful purple gem from the Lost Mountain.",
/*MISSING*/    "A beatiful purple gem from the Lost Mountain.",
    }},
  {0x8085a8a6, { // "Somehow, this cave has not received the spark of Life yet."
   "Dziwnym trafem, ta jaskinia jeszcze nie zdobyła iskry Życia.",
   "Nasılsa bu mağaraya yaşamın ışığı hiç uğramamış.",
   "Tato jeskyně z nějakého důvodu dosud nedostala jiskru Života.",
   "Так или иначе, в эту пещеру ещё не пришла жизнь.",
   "Diese Höhle wurde noch nicht vom Funken des Lebens berührt.",
/*MISSING*/    "Somehow, this cave has not received the spark of Life yet.",
    }},
  {0x80958c9a, { // "spiral on rendering"
   "spirala przy renderowaniu",
/*MISSING*/    "spiral on rendering",
   "spirála při renderování",
   "спираль отрисовки",
   "Spirale beim Rendering",
/*MISSING*/    "spiral on rendering",
    }},
  {0x80a0bf53, { // "\n\nEasy %1 might disappear when you collect more of its kind."
   "\n\nŁatw%ya1 %1 może zniknąć podczas gdy zbierasz je w innych miejscach.",
/*MISSING*/    "\n\nEasy %1 might disappear when you collect more of its kind.",
   "\n\nSnadn%ý1 %1 může zmizet, když jich sebereš více.",
   "\n\nПростые %P1 могут исчезнуть, когда вы соберёте побольше их.",
/*MISSING*/    "\n\nEasy %1 might disappear when you collect more of its kind.",
/*MISSING*/    "\n\nEasy %1 might disappear when you collect more of its kind.",
    }},
  {0x80b2231b, { // "teleport away"
   "uciekaj stąd (teleport)",
/*MISSING*/    "teleport away",
   "teleportuj se pryč",
   "телепортироваться отсюда",
/*MISSING*/    "teleport away",
/*MISSING*/    "teleport away",
    }},
  {0x80b8c15c, { // "y = spin"
   "y = obróć",
   "y = döndür",
   "y = otoč",
   "y = поворот",
   "y = drehen",
/*MISSING*/    "y = spin",
    }},
  {0x80c62d36, { // " for iOS"
   " na iOS",
   " iOS için",
   " na iOS",
   " на iOS",
   " für iOS",
   " para iOS",
    }},
  {0x80cd779a, { // "%The1 punches %the2!"
   "%1 uderzy%ł1 %a2!",
   "%1 %a2 yumrukladı!",
   "%1 udeři%l1 %a2!",
   "%1 ударил%E1 %a2!",
   "%Der1 schlägt %den2 %a2!",
/*MISSING*/    "%The1 punches %the2!",
    }},
  {0x8105667a, { // "Enter = back"
   "Enter = powrót",
   "Enter = geri",
   "Enter = zpět",
   "Enter = назад",
   "Enter = zurück",
/*MISSING*/    "Enter = back",
    }},
  {0x810e69a1, { // "\"I wonder what was there...\""
   "\"Zastanawiam się, co tam było...\"",
   "\"Orada ne olduğunu merak ediyorum...\"",
   "\"Copak tam asi bylo...?\"",
   "\"Интересно, что там было...\"",
   "\"Ich frage mich was da war...\"",
/*MISSING*/    "\"I wonder what was there...\"",
    }},
  {0x81487f61, { // "Press 'r' to choose a regular pattern. When a pattern is on, editing a cell automatically edits all cells which are equivalent according to this pattern. You can choose from several patterns, and choose which symmetries matter for equivalence. Also, you can press Space to switch between the map and graphics editor quickly -- note that editing floors with the graphics editor also adheres to the pattern."
   "Wciśnij 'r', by wybrać regularny wzorek. Gdy wzorek jest włączony, edytowanie pola automatycznie edytuje także wszystkie pola, które są mu równoważne ze względu na ten wzorek. Możesz wybrać jeden z kilku wzorków i wybrać symetrie uwzględniane przy obliczaniu równoważności. Możesz też nacisnąć spację, by szybko się przenosić pomiędzy edytorem map i edytorem grafiki -- edytowanie podłóg w edytorze grafiki również następuje zgodnie ze wzorkiem.",
   "'r'ye basarak düzenli bir doku seçebilirsin. Bir doku aktifken, bir hücreyi değiştirmek o hücreye bu dokuda eşdeğer bütün hücreleri de değiştirir. Birden çok doku seçebilirsin ve hangi simetrilerin eşdeğer olduğunu seçebilirsin. Ayrıca, Space'e basarak harita ve grafik editörü arasında hızlıca gidip gelebilirsin -- zeminin grafik editörüyle değiştirilmesinin dokuyu da değiştireceğini unutma.",
   "Stiskem klávesy 'r' můžeš vybrat pravidelný vzor. Když je vzor zapnutý, editace buňky automaticky ovlivní i všechny ostatní buňky, které jsou podle tohoto vzoru ekvivalentní. Můžeš si vybrat z několika možných vzorů, a také to, které symetrie budou důležité pro ekvivalenci. Také můžeš rychle přepínat mezerníkem mezi editorem mapy a grafiky -- i editování podlahy grafickým editorem se řídí pravidly vzoru.",
   "Нажмите 'r', чтобы выбрать регулярный узор. Когда узор выбран, изменение клетки автоматически изменяет все эквивалентные, ей в этом узоре. Вы можете выбрать один из нескольких узоров и определить, какие симметрии важны для эквивалентности. Нажмите пробел, чтобы переключиться между картой и графическим редактором -- изменение пола влияет на узор.",
   "Drücke 'r' um ein normales Muster zu wählen. Wenn ein Muster ausgewählt ist sorgt das ändern einer Zelle automatisch für die Änderung aller Zellen, die laut dem Muster gleich dieser sind. Du kannst aus mehreren Mustern wählen und bestimmen welche Symmetrien als gleich zählen. Durch das Drücken von LEER wird zwischen dem Karten und Grafikeditor gewechselt. Das ändern von Böden mit dem Grafikeditor wirkt sich auch auf das Muster aus.",
/*MISSING*/    "Press 'r' to choose a regular pattern. When a pattern is on, editing a cell automatically edits all cells which are equivalent according to this pattern. You can choose from several patterns, and choose which symmetries matter for equivalence. Also, you can press Space to switch between the map and graphics editor quickly -- note that editing floors with the graphics editor also adheres to the pattern.",
    }},
  {0x8161e570, { // "always available"
   "zawsze dostępna",
/*MISSING*/    "always available",
   "vždy dostupná",
   "всегда доступна",
/*MISSING*/    "always available",
/*MISSING*/    "always available",
    }},
  {0x817aefaf, { // "c = choose"
   "c = wybór",
/*MISSING*/    "c = choose",
   "c = výbìr",
   "c = выбор",
/*MISSING*/    "c = choose",
/*MISSING*/    "c = choose",
    }},
  {0x81b0724a, { // "hexa floor"
   "podłoga hex",
   "hekzazemin",
   "hexa-podlaha",
   "шестиугольный пол",
   "Hexagonaler Boden",
   "andar hexa",
    }},
  {0x81ba80a7, { // "Retrieving scores from Google Leaderboards..."
   "Ściągam wyniki z rankingów Google...",
   "Google listelerindeki puanlar yükleniyor...",
   "Stahuji skóre z žebříčků Google...",
   "Получение результатов из таблиц Google...",
   "Empfange Rekorde aus Google Leaderboards...",
/*MISSING*/    "Retrieving scores from Google Leaderboards...",
    }},
  {0x81d3a5ba, { // "A castle in the Crossroads..."
   "O zamku na Skrzyżowaniu...",
   "Arayollarda bir Kale...",
   "O hradu na Křižovatce...",
   "О замке на перекрёстке...",
   "Ein Schloss in den Kreuzungen...",
/*MISSING*/    "A castle in the Crossroads...",
    }},
  {0x81d98cfd, { // "Good luck in the elliptic plane!"
   "Powodzenia na płasczyźnie eliptycznej!",
/*MISSING*/    "Good luck in the elliptic plane!",
   "Hodně štěstí v eliptické rovině!",
   "Удачи на эллиптической плоскости!",
/*MISSING*/    "Good luck in the elliptic plane!",
/*MISSING*/    "Good luck in the elliptic plane!",
    }},
  {0x81f5d93f, { // "You will be fighting red rock snakes, too..."
   "Będziesz też walczyć z czerwonymi skalnymi wężami...",
   "Kızıl yılanlarla da savaşacaksın...",
   "Budeš bojovat i s kamennými hady...",
   "Вы тоже будете биться с красными змеями...",
   "Du wirst auch rote Felsschlangen bekämpfen...",
/*MISSING*/    "You will be fighting red rock snakes, too...",
    }},
  {0x82283270, { // "press a key for '%1'"
   "naciśnij klawisz dla '%1'",
   "'%1' için bir tuşa bas",
   "stiskni klávesu pro '%1'",
   "нажмите клавишу для '%1'",
   "Drücke eine Taste für '%1'",
/*MISSING*/    "press a key for '%1'",
    }},
  {0x827238d1, { // "An alternate layout of the Crossroads. It is more dense and more regular, although you won't find the castle of Camelot here."
   "Alternatywny układ Skrzyżowania. Jest gęstszy i bardziej regularny, choć nie znajdziesz tutaj zamku Camelot.",
   "Arayolların alternatif bir versiyonu. Daha yoğun ve daha düzenli, ama Kamelot'un kalesini burada bulamazsın.",
   "Alternativní plán Křižovatky. Je hustší a pravidelnější, ale hrad Camelot tady nenajdete.",
   "Другая схема перекрёстка. Она плотнее и регулярнее обычной, но замок Камелот Вы здесь не найдёте.",
   "Ein alternatives Layout der Kreuzungen. Es ist beständiger und dichter, aber du findest Camelot hier nicht.",
/*MISSING*/    "An alternate layout of the Crossroads. It is more dense and more regular, although you won't find the castle of Camelot here.",
    }},
  {0x827a3936, { // "You can hold this bush to climb the Lost Mountain. Bushes block the movement of birds."
   "Możesz złapać ten krzak, by się wspiąć na Zagubioną Górę. Krzaki blokują drogę ptakom.",
/*MISSING*/    "You can hold this bush to climb the Lost Mountain. Bushes block the movement of birds.",
   "Tohoto keře se můžeš chytit při výstupu na Ztracenou horu. Keře blokují pohyb ptáků.",
   "Вы можете хвататься за этот куст, чтобы лезть на гору. Блокирует перемещение птиц.",
/*MISSING*/    "You can hold this bush to climb the Lost Mountain. Bushes block the movement of birds.",
/*MISSING*/    "You can hold this bush to climb the Lost Mountain. Bushes block the movement of birds.",
    }},
  {0x829b9943, { // "circle/horocycle tree"
   "drzewo bazowe dla kół/horocykli",
/*MISSING*/    "circle/horocycle tree",
   "kruhový/horocyklický strom",
   "основное дерево для окружностей/орициклов",
/*MISSING*/    "circle/horocycle tree",
/*MISSING*/    "circle/horocycle tree",
    }},
  {0x82c80b61, { // "There are many lands in HyperRogue. Collect 10 treasure in the given land type to complete it; this enables you to find the magical Orbs of this land, and in some cases get access to new lands. At 25 treasures this type of Orbs starts appearing in other lands as well. Press 'o' to get the details of all the Lands.\n\n"
   "W świecie HyperRogue jest wiele krain. Kraina jest zdobyta, gdy zbierzesz w niej 10 skarbów; gdy zdobędziesz krainę, możesz znaleźć magiczną Sferę tej krainy, i w niektórych przypadkach uzsykać dostęp do nowych krain. Gdy zbierzesz 25 skarbów, ten typ Sfer będzie się też pojawiał w innych krainach. Naciśnij '0', by obejrzeć szczegóły na temat wszystkich Krain.\n\n",
/*MISSING*/    "There are many lands in HyperRogue. Collect 10 treasure in the given land type to complete it; this enables you to find the magical Orbs of this land, and in some cases get access to new lands. At 25 treasures this type of Orbs starts appearing in other lands as well. Press 'o' to get the details of all the Lands.\n\n",
   "Svět HyperRogue sestává z mnoha krajů. Když v nějakém kraji získáš 10 pokladů, dobyješ ho; pak v něm můžeš najít magické Sféry, a někdy také získáš přístup k jiným krajům. Po získání 25 pokladů se tyto magické Sféry začnou objevovat i v jiných krajích. Detailní informace o všech krajích můžeš zobrazit stisknutím klávesy 'o'.\n\n",
   "В HyperRogue много разных земель. Соберите 10 ценностей в земле, чтобы завершить её; это позволяет находить в ней магические Сферы, и иногда открывает другие земли. После сбора 25 ценностей сферы начинают появляться и в других землях. Нажмите 'o', чтобы узнать подробности о разных землях.\n\n",
/*MISSING*/    "There are many lands in HyperRogue. Collect 10 treasure in the given land type to complete it; this enables you to find the magical Orbs of this land, and in some cases get access to new lands. At 25 treasures this type of Orbs starts appearing in other lands as well. Press 'o' to get the details of all the Lands.\n\n",
/*MISSING*/    "There are many lands in HyperRogue. Collect 10 treasure in the given land type to complete it; this enables you to find the magical Orbs of this land, and in some cases get access to new lands. At 25 treasures this type of Orbs starts appearing in other lands as well. Press 'o' to get the details of all the Lands.\n\n",
    }},
  {0x83095fd3, { // "player 6 spin"
   "gracz 6 obrót",
/*MISSING*/    "player 6 spin",
   "hráč 6 otáčení",
   "игрок 6 вертит",
   "Spieler 6 drehen",
/*MISSING*/    "player 6 spin",
    }},
  {0x830cc4a2, { // "\nRanged Orbs can be targeted by long touching the desired location."
   "\nRobi się to przez długie dotknięcie celu.",
/*MISSING*/    "\nRanged Orbs can be targeted by long touching the desired location.",
   "\nTyto sféry lze zacílovat dlouhým stiskem požadovaného políčka.",
   "\nВыбор цели сферы определяется долгим зажатием мыши на цели.",
   "\nMit Distanzorbs kann durch langes Berühren der gewünschten Position gezielt werden.",
/*MISSING*/    "\nRanged Orbs can be targeted by long touching the desired location.",
    }},
  {0x832fe3e8, { // "%The1 destroys your boat!"
   "%1 zniszczy%ł1 Twój statek!",
/*MISSING*/    "%The1 destroys your boat!",
   "%1 zniči%l1 Tvou loďku!",
   "%1 уничтожил%E1 твою лодку!",
   "%Der1 %1 zerstört dein Boot!",
/*MISSING*/    "%The1 destroys your boat!",
    }},
  {0x834fb047, { // "m = monsters"
   "m = potwory",
   "m = canavar",
   "i = předměty",
   "b = граница",
   "m = Monster",
/*MISSING*/    "m = monsters",
    }},
  {0x838d01ac, { // "Configuration:"
   "Konfiguracja:",
   "AYARLAR:",
   "Nastavení:",
   "Настройки:",
   "Einstellungen:",
   "Configuração:",
    }},
  {0x83e456c6, { // "This orb allows attacking Hedgehog Warriors directly, as well as stabbing other monsters.\n"
   "Ta sfera pozwala bezpośrednio atakować Wojowników-Jeże, a także dźgać pozostałych przeciwników.\n",
   "Bu küre Kirpi Savaşçılara doğrudan saldırmaya ve diğer canavarları bıçaklamaya izin verir.\n",
   "Tato sféra umožňuje útočit na Ježčí válečníky přímo a probodávat i jiné netvory.\n",
   "Эта сфера позволит прямо атаковать Ежей-Воинов и закалывать тех врагов, которых Вы могли убить обычным путём.\n",
   "Dieser Orb erlaubt es dir Igelkrieger direkt anzugreifen, und andere Monster zu 'erstechen'.\n",
/*MISSING*/    "This orb allows attacking Hedgehog Warriors directly, as well as stabbing other monsters.\n",
    }},
  {0x84001e00, { // "%The1 burns!"
   "%1 się spalił%oa1!",
   "%1 yandı!",
   "%1 shoře%l1!",
   "%1 горит!",
   "%Der1 %1 brennt!",
/*MISSING*/    "%The1 burns!",
    }},
  {0x840197e5, { // "There is a flash of thunder!"
   "Błyskawica!",
   "Bir şimşek çaktı!",
   "Zablesklo se!",
   "Молния!",
   "Ein Blitz!",
/*MISSING*/    "There is a flash of thunder!",
    }},
  {0x8447b58f, { // "See more on the website: "
   "Więcej na stronie: ",
   "Daha fazlasını websitesinde görüntüleyin: ",
   "Více informací najdeš na webové stránce: ",
   "Смотрите далее на сайте: ",
   "Mehr auf der Website: ",
   "Veja mais no site: ",
    }},
  {0x84a29ba6, { // "Trolls turn into stone walls when they die. When you have this Orb, this happens to every monster you defeat. Statues created from this Orb have slightly different properties than Trolls who petrify naturally."
   "Gdy Trolle umierają, zamieniają się w kamień. Ta Sfera powoduje, że dzieje się to ze wszystkimi pokonanymi stworzeniami. Pomniki stworzone przez Sferę Kamienia mają inne własności niż Trolle spetryfikowane naturalnie.",
/*MISSING*/    "Trolls turn into stone walls when they die. When you have this Orb, this happens to every monster you defeat. Statues created from this Orb have slightly different properties than Trolls who petrify naturally.",
   "Když Troll zemře, promění se v kámen. Tato Sféra to samé provede s každým netvorem, kterého porazíš. Sochy vytvořené pomocí Sféry Kamene mají trochu jiné vlastnosti než přirozeně zkamenělí Trollové.",
   "Умерший тролль превращается в камень. Если у вас есть та сфера, это происходит с каждым монстром, которого вы убиваете. Статуи, созданные этой сферой, немного отличаются от обычных окаменевших троллей.",
   "Trolle werden zu Steinwänden, wenn sie sterben. Besitzt du diesen Orb, passiert das jedem Monster, das du besiegst. Statuen, die von diesem Orb erschaffen werden, haben etwas andere Eigenschaften als Trolle, die natürlich versteinern.",
/*MISSING*/    "Trolls turn into stone walls when they die. When you have this Orb, this happens to every monster you defeat. Statues created from this Orb have slightly different properties than Trolls who petrify naturally.",
    }},
  {0x854f07d4, { // "A majestic bird, who is able to fly very fast."
   "Ten majestatyczny ptak bardzo szybko lata.",
   "Çok hızlı uçabilen görkemli bir kuş.",
   "Majestátní pták, který dokáže velmi rychle létat.",
   "Величественная птица, летающая очень быстро.",
   "Ein majestätischer Vogel der sehr schnell fliegt.",
/*MISSING*/    "A majestic bird, who is able to fly very fast.",
    }},
  {0x857eedc1, { // "total kills"
   "łączne zabicia",
   "toplam leşler",
   "celkový počet zabitých nepřátel",
   "всего убийств",
   "Gesamt Kills",
   "total de mortes",
    }},
  {0x8590511a, { // "Vampire Bats don't attack normally, but they drain your magical powers if they are at distance at most 2 from you."
   "Nietoperze Wampiry nie atakują jak inne potwory. Zamiast tego, jeśli są w odległości co najwyżej 2, wysysają Twoje magiczne moce.",
/*MISSING*/    "Vampire Bats don't attack normally, but they drain your magical powers if they are at distance at most 2 from you.",
   "Vampýři neútočí normální způsobem, ale pokud se dostanou do vzdálenosti 2 a méně od tebe, vysávají ti tvé magické schopnosti.",
   "Вампиры не атакуют Вас обычным образом, но они вытягивают из Вас магическую силу, если находятся на расстоянии не более 2.",
/*MISSING*/    "Vampire Bats don't attack normally, but they drain your magical powers if they are at distance at most 2 from you.",
/*MISSING*/    "Vampire Bats don't attack normally, but they drain your magical powers if they are at distance at most 2 from you.",
    }},
  {0x85c4a982, { // "GAME OVER"
   "KONIEC GRY",
   "OYUN BİTTİ",
   "KONEC HRY",
   "КОНЕЦ ИГРЫ",
   "GAME OVER",
   "FIM DE JOGO",
    }},
  {0x86314e8f, { // "rotate right"
   "obróć w prawo",
   "sağa döndür",
   "otočení vpravo",
   "повернуть вправо",
   "Nach rechts rotieren",
/*MISSING*/    "rotate right",
    }},
  {0x86332931, { // "Wrong color!"
   "Zły kolor!",
   "Yanlış renk!",
   "Špatná barva!",
   "Не тот цвет!",
   "Falsche Farbe!",
   "Cor errada!",
    }},
  {0x86794019, { // "camera rotation"
   "obróć kamerę",
/*MISSING*/    "camera rotation",
   "rotace kamery",
   "поворот камеры",
/*MISSING*/    "camera rotation",
/*MISSING*/    "camera rotation",
    }},
  {0x86beece5, { // "rotate the character"
   "obróć postać",
   "karakteri döndür",
   "otoč postavu",
   "повернуться",
   "drehe Spielfigur",
/*MISSING*/    "rotate the character",
    }},
  {0x86c654bf, { // "You have to hit Palace Guards several times to kill them. After each hit, they are pushed away and stunned for some time."
   "Musisz uderzyć Strażnika kilka razy, by go zabić. Po każdym trafieniu strażnik jest odpychany i ogłuszony przez pewien czas.",
   "Saray Muhafızlarının ölmesi için onlara bir kaç kere vurmalısın. Her vuruştan sonra geri itilirler ve birkaç turluk sersemlerler.",
   "Na zabití Strážce Paláce potřebuješ několik útoků. Každý zásah Strážce odstrčí dozadu a na chvíli ho omráčí.",
   "Чтобы убить Дворцового стражника, его нужно ударить несколько раз. После каждого удара он отлетает на одну клетку назад и застывает на некоторое время.",
   "Du musst Palastwachen mehrmals treffen um sie zu töten. Nach jedem Treffer werden sie zurückgestoßen und kurz betäubt.",
/*MISSING*/    "You have to hit Palace Guards several times to kill them. After each hit, they are pushed away and stunned for some time.",
    }},
  {0x86c7da3b, { // "Return to the hyperbolic world"
   "powrót na hiperboloidę",
   "Hiperbolik Dünyaya geri dön",
   "Návrat do hyperbolického světa",
   "вернуться на гиперболоид",
   "Zur hyperbolischen Welt zurückkehren",
/*MISSING*/    "Return to the hyperbolic world",
    }},
  {0x86fd7670, { // "do nothing"
   "nic",
   "hiçbir şey yapma",
   "nic",
   "ничего",
   "Tu nichts",
/*MISSING*/    "do nothing",
    }},
  {0x8708cde4, { // " (expired)"
   " (wyczerpany)",
   " (zaman aşımı)",
   " (vypršelo)",
   " (кончился)",
   " (abgelaufen)",
   " (expirado)",
    }},
  {0x8716adcb, { // "SPACE = map/graphics"
   "SPACE = mapa/grafika",
   "SPACE = harita/grafikler",
   "SPACE = mapa/grafika",
   "SPACE = карта/графика",
   "SPACE = Karte/Grafiken",
/*MISSING*/    "SPACE = map/graphics",
    }},
  {0x87432023, { // "%The1 is destroyed!"
   "%1 zosta%ł1 zniszczon%ya1!",
/*MISSING*/    "%The1 is destroyed!",
   "%1 by%l1 zniče%n1!",
   "%1 уничтожен%E1!",
/*MISSING*/    "%The1 is destroyed!",
/*MISSING*/    "%The1 is destroyed!",
    }},
  {0x877ac97d, { // "Cthulhu withdraws his tentacle!"
   "Cthulhu chowa swoją mackę!",
   "Cthulhu dokungacını çekti!",
   "Cthulhu stahuje své chapadlo!",
   "Ктулху забирает свой щупальце!",
   "Cthulhu zieht seinen Tentakel zurück!",
   "Cthulhu retira seu tentáculo!",
    }},
  {0x87a3c451, { // "Witches use these crystal cabinets to protect Powerstones, as well as the more expensive Orbs. They are partially protected from thieves: they are too strong to be smashed by conventional attacks, and if you try to steal the item using an Orb of Aether, your Aether power will be completely drained."
   "Czarownice używają tych kryształowych gablot, by chronić Kamienie Mocy oraz bardziej cenne Sfery. Gabloty są częściowo chronione przed kradzieżą: są zbyt mocne, by je rozbić konwencjonalnymi atakami, a jeśli spróbujesz ukraść przedmiot używając Sfery Eteru, Twoja moc Eteru zostanie całkowicie wyczerpana.",
   "Bu kabinleri Cadılar Güçtaşlarını ve değerli Küreleri korumak için kullanıyorlar. Hırsızlara karşı kısmen korumalıdır, bildik saldırılarla kırılmak için fazlasıyla serttir ve eğer içindeki eşyayı Eter Küresi kullanarak çalmaya çalışırsan kürenin gücü bitecektir.",
   "Čarodějnice používají tyto vitríny k ochraně Mocikamů, a také dražších  typů Sfér. Vitríny jsou částečně chráněné proti zlodějům: jsou příliš odolné na to, aby se daly rozbít běžným útokem, a pokud se pokusíš předmět ukrást pomocí Sféry Éteru, zcela to vyčerpá její sílu.",
   "Ведьмы используют эти кристальные кабинеты для защиты Камней силы, как и самые ценные сферы. Они частично защищают их от воров: они достаточно прочны, чтобы выдерживать прямые атаки, и если Вы попробуете использовать сферу Эфира для проникновения в него, то сила сферы сразу иссякнет.",
   "Hexen verwenden diese Kristallvitrinen um Machtsteine, und wertvolle Orbs zu schützen. Sie sind zu stark um von normalen Angriffen zerschmettert zu werden und solltest du versuchen den Inhalt mit einem Orb des Äthers zu stehlen, wird seine ganze Kraft verbraucht.",
/*MISSING*/    "Witches use these crystal cabinets to protect Powerstones, as well as the more expensive Orbs. They are partially protected from thieves: they are too strong to be smashed by conventional attacks, and if you try to steal the item using an Orb of Aether, your Aether power will be completely drained.",
    }},
  {0x87ba9948, { // "\nRanged Orbs can be targeted by touching the desired location."
   "\nRobi się to przez dotknięcie celu.",
/*MISSING*/    "\nRanged Orbs can be targeted by touching the desired location.",
   "\nTyto sféry můžete zacílovat tak, že se dotknete požadovaného políčka.",
   "\nВыбор цели сферы определяется зажатием мыши на цели.",
   "\nMit Distanzorbs kann durch Berühren der gewünschten Position gezielt werden.",
/*MISSING*/    "\nRanged Orbs can be targeted by touching the desired location.",
    }},
  {0x87c84244, { // "You start chopping down the tree."
   "Zaczynasz ciąć drzewo.",
   "Ağacı kesmeye başladın.",
   "Zača%l0 jsi kácet strom.",
   "Вы начали рубить дерево.",
   "Du fängst an, den Baum zu fällen.",
   "Você começa a cortar a árvore.",
    }},
  {0x87faf958, { // "Accessible only from %the1 (until finished).\n"
   "Kraina dostępna tylko poprzez %a1 (przed ukończeniem).\n",
   "Sadece %a1 erişilebilir. %a1 (bitirilene kadar).\n",
   "Tento kraj je dostupný pouze skrz %a1 (dokud ho nedohraješ).\n",
   "Доступно только из %a1 (пока не окончена).\n",
   "Nur von %der1 %a1 aus erreichbar (bis abgeschlossen).\n",
/*MISSING*/    "Accessible only from %the1 (until finished).\n",
    }},
  {0x880f622a, { // "native to %the1 (collect 10 %2)"
   "rodzima %abl1 (10x %2)",
   "%abl1 bulunuyor (10 tane %2)",
   "standardní %abl1 (získej 10x %2)",
   "родная %abl1 (10x %2)",
   "heimisch in %der1 (10x %2)",
/*MISSING*/    "native to %the1 (collect 10 %2)",
    }},
  {0x8832d31f, { // "Don't center on the player character."
   "Nie centruj na graczu.",
/*MISSING*/    "Don't center on the player character.",
   "Necentrovat na hráče.",
   "Не центрировать на игроке.",
/*MISSING*/    "Don't center on the player character.",
/*MISSING*/    "Don't center on the player character.",
    }},
  {0x8882ea34, { // "This dangerous sea contains nasty sharks and pirates. "
   "Niebezpieczne morze, pełne rekinów i piratów.",
   "Bu tehlikeli deniz vahşi köpekbalıkları ve korsanlar barındırıyor.",
   "V tomto nebezpečném moři se vyskytují nebezpeční žraloci a piráti.",
   "Опасное море, полное акул и пиратов.",
   "Dieser gefährliche Ozean enthält fiese Haie und Piraten.",
/*MISSING*/    "This dangerous sea contains nasty sharks and pirates. ",
    }},
  {0x88a6ca3b, { // "You summon some Mimics!"
   "Przywoła%łeś0 Miraże!",
   "Taklitçiler çağırdın!",
   "Vyvola%l0 jsi několik Přeludů!",
   "Вы призвали миражи!",
   "Du beschwörst ein paar Mimics!",
   "Você invoca algumas Mímicas!",
    }},
  {0x88a77da0, { // "You collect %the1."
   "Znalaz%łeś0 %a1.",
   "%a1 aldın.",
   "Sebra%l0 jsi %a1.",
   "Вы нашли %a1.",
   "Du sammelst %den1 %a1.",
   "Você coleta %na1.",
    }},
  {0x88b058d0, { // "You create an Illusion!"
   "Tworzysz iluzję!",
   "Bir İlüzyon ürettin!",
   "Vytvořil jsi iluzi!",
   "Вы создали иллюзию!",
   "Du erschaffst eine Illusion!",
/*MISSING*/    "You create an Illusion!",
    }},
  {0x88b3c6f4, { // "%The1 squeaks happily!"
   "%1 pisn%ął1 szczęśliwie!",
   "%1 neşeyle ciyaklıyor!",
   "%1 šťastně piští!",
   "%1 пищит радостно!",
   "%Die1 %1 piepst fröhlich!",
/*MISSING*/    "%The1 squeaks happily!",
    }},
  {0x88b7c428, { // "\n\nNote: when pushing %the1 off a heptagonal cell, you can control the pushing direction by clicking left or right half of the heptagon."
   "\n\nUwaga: kiedy spychasz %a1 z siedmiokąta, możesz sterować kierunkiem pchnięcia klikając lewą lub prawą połowę siedmiokąta.",
   "\n\nNot:  %a1 bir yedigensel hücreden iterken, itmenin yönünü yedigenin sol ve sağ yarısına tıklayarak belirleyebilirsin.",
   "\n\nPoznámka: Když tlačíš %a1 ze sedmiúhelníku, můžeš ovládat směr tlačení tím, že klikneš na levou či pravou polovinu tohoto sedmiúhelníku.",
   "\n\nВажно: когда Вы толкаете %a1 с семиугольника, Вы можете выбирать направление движения, нажимая на соответствующую половину семиугольника.",
   "\n\nHinweis: Wenn du %den1 %a1 von einer heptagonalen Zelle schiebst kannst du die Richtung bestimmen, indem du die linke oder rechte Seite des Heptagon anklickst.",
/*MISSING*/    "\n\nNote: when pushing %the1 off a heptagonal cell, you can control the pushing direction by clicking left or right half of the heptagon.",
    }},
  {0x89007bc0, { // "This Orb allows your boat to go against the current, and also to go into the land, creating water on the way."
   "Ta Sfera pozwala Twojemu statkowi płynąć pod prąd, a także płynąć przez ląd, tworząc wodę po drodze.",
   "Bu Küre akıntıya karşı hareket edebilmene imkân verir. Ayrıca bu küreyle karaya doğru hareket edip yolunda su çıkararak ilerleyebilirsin.",
   "Tato Sféra tvému člunu umožňuje plout proti proudu. Také můžeš vyplout na souš a přitom vytvářet vodu, kamkoli se vydáš.",
   "Эта сфера позволяет тебе плыть против течения, а также двигаться по суше, создавая воду по пути.",
   "Dieser Orb erlaubt es dir, gegen die Strömung zu schwimmen. Du kannst damit auch durch Land schwimmen, wobei du Wasser kreierst).",
/*MISSING*/    "This Orb allows your boat to go against the current, and also to go into the land, creating water on the way.",
    }},
  {0x892048aa, { // "Enemies killed: %1 (%2 types)"
   "Zabici przeciwnicy: %1 (%2 typów)",
/*MISSING*/    "Enemies killed: %1 (%2 types)",
   "Počet zabitých nepřátel: %1 (%2 typů)",
   "Существ убито: %1 (%2 типов)",
   "Getötete Feinde: %1 (%2 Arten)",
/*MISSING*/    "Enemies killed: %1 (%2 types)",
    }},
  {0x8920d8ac, { // "Three mines next to you!"
   "Trzy miny koło Ciebie!",
   "Yakınında üç mayın var!",
   "Ve tvé blízkosti se nacházejí tři miny!",
   "Три мины рядом с Вами!",
   "3 Minen um dich!",
/*MISSING*/    "Three mines next to you!",
    }},
  {0x894ae094, { // "Nothing to stand on here!"
   "Nie ma tu na czym stać!",
   "Burada duracağın bir şey yok!",
   "Není tu na čem stát!",
   "Не на чем стоять!",
   "Hier gibt es nichts worauf du stehen kannst!",
/*MISSING*/    "Nothing to stand on here!",
    }},
  {0x894b9946, { // "deplete orb powers"
   "utrata mocy",
   "küre güçlerini tüket",
   "zruš moc sféry",
   "потерять силы сфер",
   "verbrauche Orbkräfte",
/*MISSING*/    "deplete orb powers",
    }},
  {0x89658792, { // "You switch places with %the1."
   "Zamieniasz się miejscami z %abl1.",
   "%abl1 yer değiştirdin.",
   "Vyměni%l0 sis místo s %abl1.",
   "Вы поменялись местами с %abl1.",
   "Du tauschst den Platz mit %dem1 %a1.",
   "Você troca de lugar com %na1.",
    }},
  {0x896b50d1, { // "zoom out"
   "oddalenie",
/*MISSING*/    "zoom out",
   "oddálení",
   "отдалить",
   "herauszoomen",
/*MISSING*/    "zoom out",
    }},
  {0x89842dfc, { // "kill a monster, or collect 25 %2"
   "zabij potwora lub znajdź 25x %2",
   "bir canavar öldür, ya da 25 tane %2 topla.",
   "zabij netvora nebo získej 25x %2",
   "убей монстра, или собери 25x %2",
   "töte ein Monster oder sammle 25x %2",
/*MISSING*/    "kill a monster, or collect 25 %2",
    }},
  {0x899a859f, { // "Don't be fooled by this red-haired boy, or you will be stabbed if you come too close!"
   "Nie daj się zwieść temu rudowłosemu, bo jak się za bardzo zbliżysz, możesz zostać ugodzon%ya1 nożem!",
/*MISSING*/    "Don't be fooled by this red-haired boy, or you will be stabbed if you come too close!",
   "Nenech se oklamat tímto zrzavým chlapcem -- pokud se k němu příliš přiblížíš, bodne tě!",
   "Не будьте обмануты этим рыжеволосым парнем, у него есть острый нож!",
   "Lass dich von diesem rothaarigen Jungen nicht täuschen - oder du wirst erstochen, wenn du zu nah kommst!",
/*MISSING*/    "Don't be fooled by this red-haired boy, or you will be stabbed if you come too close!",
    }},
  {0x89adf1b4, { // "Avoid chopping trees, using Orbs, and non-graveyard monsters in the Haunted Woods."
   "Unikaj ścinania drzew, używania Sfer, i potworów spoza Cmentarza.",
/*MISSING*/    "Avoid chopping trees, using Orbs, and non-graveyard monsters in the Haunted Woods.",
   "Nesmíš ve Strašidelném lese kácet stromy a používat Sféry nebo netvory, kteří nepocházejí ze Hřbitova.",
   "Не рубить деревья, не использовать сферы или посторонних монстров в Лесу с привидениями.",
/*MISSING*/    "Avoid chopping trees, using Orbs, and non-graveyard monsters in the Haunted Woods.",
/*MISSING*/    "Avoid chopping trees, using Orbs, and non-graveyard monsters in the Haunted Woods.",
    }},
  {0x89ca69da, { // "You can collect some valuable Ambers on the coast, but beware the tides!\n\nYou can also take one of the boats and venture into the Ocean, to find other coasts, dangerous whirlpools, and islands inhabited by Pirates."
   "Możesz zbierać cenne bursztyny na plaży, ale uważaj na odpływy!\n\nMożesz też wziąć łódkę i popłynąć nią przez Ocean, by znaleźć inne brzegi, niebezpieczne wiry oraz wyspy zamieszkane przez Piratów.",
   "Kıyıdan çok değerli kehribarlar toplayabilirsin, ama dalgalara dikkat et!\n\nKayıklardan birini alıp , başka kıyılar, tehlikeli tayfunlar vekorsanların egemenliğindeki adalar bulmak için okyanusa açılabilirsin.",
   "Na pobřeží můžeš sbírat cenný Jantar, ale pozor na příliv a odliv!\n\nTaké si můžeš vzít člun a vyplout na Oceán, kde najdeš další pobřeží, nebezpečné víry a ostrovy obývané piráty.",
   "Вы модете найти ценный Янтарь не берегу, но остерегайтесь приливов!\n\nВы модете найти лодку и отправиться в Океан, чтобы найти другие берега, опасные Водовороты и острова, полные пиратов.",
   "Du kannst wertvolle Bernsteine an der Küste sammeln, aber achte auf die Gezeiten!\n\nDu kannst ebenso eines der Boote nehmen und in den Ozean reisen um andere Küsten, gefährliche Strudel, und von Piraten besetzte Inseln zu finden.",
/*MISSING*/    "You can collect some valuable Ambers on the coast, but beware the tides!\n\nYou can also take one of the boats and venture into the Ocean, to find other coasts, dangerous whirlpools, and islands inhabited by Pirates.",
    }},
  {0x89cb1efa, { // "also hold Alt during the game to toggle high contrast"
   "Alt podczas gry przełącza wysoki kontrast",
   "ayrıca oyun sırasında Alt'a basılı tutarak yüksek kontrastı açıp kapatabilirsin.",
   "Podržením klávesy Alt během hry lze přepnout vysoký kontrast",
   "также Alt во время игры переключает высокий контраст",
   "Du kannst auch ALT im Spiel halten, um hohen Kontrast umzuschalten",
/*MISSING*/    "also hold Alt during the game to toggle high contrast",
    }},
  {0x8a08aa94, { // "explore the world, collect treasures"
   "zwiedzaj świat i zbieraj skarby",
/*MISSING*/    "explore the world, collect treasures",
   "zkoumej svět, sbírej poklady",
   "изучай мир, собирай сокровища",
/*MISSING*/    "explore the world, collect treasures",
/*MISSING*/    "explore the world, collect treasures",
    }},
  {0x8a356529, { // "Press F1 or right click for help"
   "Wciśnij F1 lub kliknij prawym, by uzyskać pomoc",
   "F1 ya da sağ tık ile yardıma ulaş.",
   "Pro vyvolání nápovědy stiskni F1 nebo klikni pravým tlačítkem myši",
   "Нажмите F1 или ПКМ для помощи",
   "F1 oder Rechtsklick für Hilfe",
   "Pressione F1 ou clique direito do mouse para ajuda",
    }},
  {0x8a43ee0e, { // "Border between seas."
   "Granica między morzami.",
   "Denizler arasındaki sınır.",
   "Hranice mezi moři.",
   "Граница между морями.",
   "Eine Grenze zwischen Seen.",
/*MISSING*/    "Border between seas.",
    }},
  {0x8a520cd8, { // "cheat mode"
   "tryb oszusta",
   "hile modu",
   "cheat mód",
   "режим обманщика",
   "Cheats",
   "modo de trapaça",
    }},
  {0x8a57eb2e, { // "Your Orb of the Sword can be used to dig here."
   "Możesz tu kopać przy użyciu Sfery Ostrza.",
/*MISSING*/    "Your Orb of the Sword can be used to dig here.",
   "Tvá Sféra Meče se dá použít k prokopání tohoto políčka.",
   "Твоя Сфера меча может копать здесь.",
   "Dein Orb des Schwertes kann verwendet werden, um hier zu graben.",
/*MISSING*/    "Your Orb of the Sword can be used to dig here.",
    }},
  {0x8a74faf6, { // "tutorial"
   "podręcznik",
/*MISSING*/    "tutorial",
   "tutoriál",
   "руководство",
/*MISSING*/    "tutorial",
/*MISSING*/    "tutorial",
    }},
  {0x8afa7d17, { // "An expensive metal from the Living Caves. For some reason gold prevents the living walls from growing close to it."
   "Drogocenny metal z Żyjących Jaskiń. Złoto utrudnia wzrost ścian wokół niego.",
   "Yaşayan Mağaralardan pahalı br metal. Anlaşılamayan bir sebeple altın yaşayan duvarları kendine yaklaştırmıyor.",
   "Valouny drahého kovu z Živoucí jeskyně. Zlato z nějakého důvodu zamezuje tomu, aby v jeho blízkosti rostly živoucí zdi.",
   "Драгоценный металл из Живых пещер. Он препятствует росту живых стен в клетку с ним.",
   "Ein teures Metall aus den Lebenden Höhlen. Aus irgendeinem Grund verhindert Gold, dass lebende Wände in seiner Nähe entstehen.",
/*MISSING*/    "An expensive metal from the Living Caves. For some reason gold prevents the living walls from growing close to it.",
    }},
  {0x8b0e56bc, { // "You have enough treasure now to revive %the1!"
   "Masz dość skarbów, by przywrócić %a1 do życia!",
/*MISSING*/    "You have enough treasure now to revive %the1!",
   "Nyní máš dost pokladů na to, abys mohl oživit %a1!",
   "У вас достаточно сокровищ, чтобы оживить %a1!",
/*MISSING*/    "You have enough treasure now to revive %the1!",
/*MISSING*/    "You have enough treasure now to revive %the1!",
    }},
  {0x8b48ef1f, { // "%The1 destroys %the2!"
   "%1 zniszczy%ł1 %a2!",
   "%1 %a2 yok etti!",
   "%1 zniči%l1 %a2!",
   "%1 уничтожил%E1 %a2!",
   "%Der1 %1 zerstört %den2 %a2!",
   "%na1 destrói %na2!",
    }},
#if REPEATED
  {0x8b48ef20, { // "%The1 destroys %the2! [repeat]"
   "%1 zniszczy%ł1 %a2!",
   "%1  %a2 yok etti!",
   "%1 zniči%l1 %a2!",
   "%1 уничтожил%E1 %a2!",
   "%Der1 %1 zerstört %den2 %a2!",
   "%na1 destrói %na2!",
    }},
#endif
  {0x8b64aa5c, { // "Better find some other place."
   "Lepiej znajdź inne miejsce.",
   "Başka bir yer bulsan daha iyi olacak.",
   "Raději bys mě%l0 najít nějaké jiné místo.",
   "Лучше отправьтесь в другое место.",
   "Du solltest besser einen anderen Ort finden.",
   "É melhor enontrar outro lugar.",
    }},
  {0x8ba2057f, { // "render the texture only once"
   "renderuj teksturę tylko raz",
/*MISSING*/    "render the texture only once",
   "renderuj textury pouze jednou",
   "отрисовать один раз",
   "Rendere die Textur nur einmal",
/*MISSING*/    "render the texture only once",
    }},
  {0x8bf5eda5, { // "Dragon Scales are a prized material for armors. They are also prized by collectors, who would like to boast about how they have killed a Dragon.\n\nDragon Scales disappear after 500 turns."
   "Smocze łuski są cenionym materiałem na zbroje. Są one również poszukiwane przez zbieraczy, którzy lubią się przechwalać tym, ile smoków zabili.\n\nSmocze łuski znikają po 500 kolejkach.",
/*MISSING*/    "Dragon Scales are a prized material for armors. They are also prized by collectors, who would like to boast about how they have killed a Dragon.\n\nDragon Scales disappear after 500 turns.",
   "Dračí šupiny jsou ceněným materiálem pro výrobu zbroje. Také si je cení sběratelé, kteří by se rádi chlubili tím, jak zabili Draka.\n\nDračí šupiny zmizí po uplynutí 500 kol.",
   "Чешуя дракона -- ценный материал для брони. Также они ценятся коллекционерами, которые хотели бы похвастаться, как они убили дракона.\n\nЧешуя дракона исчезнет через 500 ходов.",
   "Drachenschuppen sind ein hochgeschätztes Material für Rüstungen. Sie werden auch von Sammlern begehrt, die gerne damit prahlen, wie sie einen Drachen getötet haben.\n\nDrachenschuppen verschwinden nach 500 Zügen.",
/*MISSING*/    "Dragon Scales are a prized material for armors. They are also prized by collectors, who would like to boast about how they have killed a Dragon.\n\nDragon Scales disappear after 500 turns.",
    }},
  {0x8cf7c8c6, { // "She has flowers in her long fair hair. You could not bring yourself to attack such a beautiful woman."
   "Ta ślicznotka ma kwiaty w swych jasnych, długich lokach. Nie jesteś w stanie zaatakować tak pięknej kobiety.",
/*MISSING*/    "She has flowers in her long fair hair. You could not bring yourself to attack such a beautiful woman.",
   "Tato krasavice má ve svých dlouhých, světlých vlasech květiny. Na tak krásnou ženu se nemůžeš přimět zaútočit.",
   "В её длинные светлые локоны вплетены прекрасные цветы. Вы не можете ударить столь красивую девушку.",
   "Sie trägt Blumen in ihrem langen blonden Haar. Du kannst dich nicht dazu überwinden so eine schöne Frau anzugreifen.",
/*MISSING*/    "She has flowers in her long fair hair. You could not bring yourself to attack such a beautiful woman.",
    }},
  {0x8d0d3879, { // "Hyperbolic pirates do not need huge ships, since so many lands to conquest are so close. These small boats are enough for them.\n\nBoats allow you to go through water. If you are in a boat, you can move into a water cell (and the boat will come with you)."
   "Hiperboliczni piraci nie potrzebują dużych statków, skoro tyle lądów do zdobycia jest tak blisko. Wystarczą im małe łódki.\n\nŁódki pozwalają poruszać się przez wodę. Jeśli jesteś na łódce, możesz przejść na pole z wodą (a łódka przesunie się razem z Tobą).",
   "Hiperbolik korsanlar yakınlarda fethedilecek bir sürü toprak olduğundan devasa gemilere ihtiyaç duymuyor. Bu küçük kayıklar onlar için yeterli.\n\nKayıklar suda ilerleyebilmeni sağlıyor. Eğer bir kayık içindeysen, bir su hücresine ilerleyebilirsin. (Ve kayık da seninle birlikte gelir). ",
   "Hyperboličtí piráti nepotřebují velké lodě, protože tolik zemí, které mohou dobývat, je tak blízko. Tyto malé čluny jim stačí.\n\nČluny ti umožňují pohybovat se po vodě. Pokud jsi ve člunu, můžeš vstoupit na vodní políčko (člun se přesune spolu s tebou).",
   "Гиперболическим пиратам не нужны большие корабли, ведь много земель для завоевания так близки. Достаточно этих маленьких лодок.\n\nЛодки позволяют двигаться по воде. Если Вы в лодке, Вы можете перейти на клетку с водой (и лодка переместится туда же).",
   "Hyperbolische Piraten brauchen keine riesigen Schiffe, da viele Länder zum Erobern so dicht zusammen sind. Diese kleinen Boote reichen ihnen.\n\nBoote lassen dich durch Wasser bewegen. Bist du in einem, kannst du dich auf Wasserzellen bewegen.",
/*MISSING*/    "Hyperbolic pirates do not need huge ships, since so many lands to conquest are so close. These small boats are enough for them.\n\nBoats allow you to go through water. If you are in a boat, you can move into a water cell (and the boat will come with you).",
    }},
  {0x8d11dbf4, { // "Huge, impassable walls which separate various lands."
   "Wielkie ściany, które oddzielają od siebie poszczególne krainy.",
   "Farklı diyarları birbirinden ayıran çok yüksek, aşılamaz duvarlar.",
   "Obrovské, neprostupné zdi, které oddělují různé kraje.",
   "Огромные непроходимые стены, разделяющие разные земли.",
   "Gigantische, undurchdringbare Wände, die die verschiedenen Länder voneinander trennen.",
/*MISSING*/    "Huge, impassable walls which separate various lands.",
    }},
  {0x8d295357, { // "You have collected 10 treasures of each type."
   "Znalaz%łeś0 10 skarbów każdego typu.",
   "her hazine çeşidinden onar tane topladın.",
   "Získa%l0 jsi 10 pokladů od každého typu.",
   "Вы собрали по 10 сокровищ каждого типа.",
   "Du hast 10 Schätze jedes Typs gesammelt.",
   "Você coletou 10 tesouros de cada tipo.",
    }},
  {0x8d81cd48, { // "This might be useful for Android devices with limited memory."
   "To może się przydać, gdy Twoje urządzenie ma mało pamięci.",
   "Bu, sınırlı hafızalı Android cihazları için kullanışlı olabilir.",
   "To se může hodit na zařízeních Android s omezenou pamětí.",
   "Это может быть полезно для Android-устройств с ограниченной памятью.",
   "Könnte für Android-Geräte mit begrenztem Speicher nützlich sein.",
   "Isso pode ser útil para dispositivos Android com memória limitada.",
    }},
  {0x8d9e15f4, { // "Your Aether powers are drained by %the1!"
   "Twoja moc Eteru została wyssana przez %a1!",
   "Güçlerin %1 tarafından tüketiliyor!",
   "%1 vysála tvou sílu Éteru!",
   "Твоя эфирная сила была уменьшена %abl1!",
   "%Der1 %1 saugt deine Ätherkraft auf!",
/*MISSING*/    "Your Aether powers are drained by %the1!",
    }},
  {0x8dc2f051, { // "Showoff mode"
   "Tryb pokazowy",
   "Gösteriş Modu",
   "Předváděcí mód",
   "Режим показа",
   "Angeber Modus",
   "Modo 'exibido'",
    }},
  {0x8e0a8eb4, { // "To learn more about straight lines, wander further, and you should find the Land of Eternal Motion. Try to run in a straight line, with a Running Dog next to you. Even though the Running Dog runs at the same speed as you, it will appear to go slower -- this is because you are running in a straight line, and the Running Dog has to run in a curve called an equidistant.\n\nRemember that you can right click on anything to get more information."
   "By dowiedzieć się więcej o liniach prostych, podróżuj dalej, aż znajdziesz Krainę Wiecznego Ruchu. Biegnij w linii prostej z Psem biegnącym obok. Mimo że Pies biegnie z tą samą prędkością, nie będzie w stanie Cię dogonić -- ponieważ Ty ruszasz się w linii prostej, a Pies biegnie po krzywej zwanej ekwidystantą.\n\nPamiętaj, że możesz klikać prawym przyciskiem na różnych elementach gry, by dowiedzieć się o nich więcej.",
/*MISSING*/    "To learn more about straight lines, wander further, and you should find the Land of Eternal Motion. Try to run in a straight line, with a Running Dog next to you. Even though the Running Dog runs at the same speed as you, it will appear to go slower -- this is because you are running in a straight line, and the Running Dog has to run in a curve called an equidistant.\n\nRemember that you can right click on anything to get more information.",
   "Chceš-li se dozvědět víc o přímkách, běž dál a měl bys najít Zemi věčného pohybu. Pokus se tam běžet po přímce, zatímco vedle tebe běží Běžící pes. I když běží stejně rychle jako ty, bude se zdát, že je pomalejší -- to proto, že ty běžíš po přímce, zatímco Běžící pes musí běžet po křivce zvané 'ekvidistanta'.\n\nNezapomeň, že pokud na cokoli klikneš pravým tlačítkem, můžeš o tom získat víc informací.",
   "Чтобы узнать больше о прямых линиях, найди Землю вечного движения. Попробуй двигаться по прямой вместе с собакой, бегущей рядом. Хотя вы бежите с равной скоростью, она будет отставать -- дело в том, что ты бежишь по прямой, а собака -- по кривой, называемой эквидистантой.\n\n",
/*MISSING*/    "To learn more about straight lines, wander further, and you should find the Land of Eternal Motion. Try to run in a straight line, with a Running Dog next to you. Even though the Running Dog runs at the same speed as you, it will appear to go slower -- this is because you are running in a straight line, and the Running Dog has to run in a curve called an equidistant.\n\nRemember that you can right click on anything to get more information.",
/*MISSING*/    "To learn more about straight lines, wander further, and you should find the Land of Eternal Motion. Try to run in a straight line, with a Running Dog next to you. Even though the Running Dog runs at the same speed as you, it will appear to go slower -- this is because you are running in a straight line, and the Running Dog has to run in a curve called an equidistant.\n\nRemember that you can right click on anything to get more information.",
    }},
  {0x8e157c82, { // "You construct some Elemental Gems!"
   "Skonstruowa%łeś0 Kamień Zywiołów!",
/*MISSING*/    "You construct some Elemental Gems!",
   "Vytvoři%l0 jsi Elementální drahokamy!",
   "Вы создали Элементальный камень!",
   "Du erstellst ein paar Elementjuwelen!",
/*MISSING*/    "You construct some Elemental Gems!",
    }},
  {0x8e282708, { // "Completing the quest in this land is not necessary for the Hyperstone Quest."
   "Wykonanie tego zadania nie jest konieczne do zdobycia Hiperkamieni.",
/*MISSING*/    "Completing the quest in this land is not necessary for the Hyperstone Quest.",
   "Splnění mise v tomto kraji není nezbytné pro Hyperkamovou misi.",
   "Выполнение задания в этой земле не обязательно для квеста Гиперкамней.",
   "Das Bestehen der Herausforderung dieses Landes ist keine Voraussetzung für die Hyperstein-Herausforderung.",
/*MISSING*/    "Completing the quest in this land is not necessary for the Hyperstone Quest.",
    }},
  {0x8e3ccd13, { // "This Orb is not compatible with the Tutorial."
   "Ta Sfera nie jest kompatybilna z podręcznikiem.",
/*MISSING*/    "This Orb is not compatible with the Tutorial.",
   "Tato Sféra není kompatibilní s tutoriálem.",
   "Эта сфера не совместима с Руководством.",
/*MISSING*/    "This Orb is not compatible with the Tutorial.",
/*MISSING*/    "This Orb is not compatible with the Tutorial.",
    }},
  {0x8eb7812c, { // "(includes co-op)"
   "(zawiera co-op)",
/*MISSING*/    "(includes co-op)",
   "(zawiera co-op)",
   "(включая совместный режим)",
   "(enthält co-op)",
/*MISSING*/    "(includes co-op)",
    }},
  {0x8f298833, { // "rotational symmetry"
   "symetria obrotowa",
   "döngüsel simetri",
   "rotační symetrie",
   "вращательная симметрия",
   "drehende Symmetrie",
/*MISSING*/    "rotational symmetry",
    }},
  {0x8f2c66ca, { // "Precious belongings of ancient Viking heroes. Your Orb of the Sword can be used to dig these treasures out of the barrows."
   "Drogocenne przedmioty należące do starożytnych wikińskich bohaterów. Użyj Sfery Ostrza, by wykopać te skarby z kurhanów.",
/*MISSING*/    "Precious belongings of ancient Viking heroes. Your Orb of the Sword can be used to dig these treasures out of the barrows.",
   "Drahocenný majetel dávných vikingských hrdinů. Tvá Sféra Meče může tyto poklady vykopat z mohyl.",
   "Драгоценности древних героев-викингов. Использую Сферу меча, чтобы добыть их из могилы.",
   "Kostbares Eigentum der Helden der alten Wikinger. Dein Orb des Schwertes kann verwendet werden, um diese Schätze aus den Hügelgräbern herauszuholen.",
/*MISSING*/    "Precious belongings of ancient Viking heroes. Your Orb of the Sword can be used to dig these treasures out of the barrows.",
    }},
  {0x8f3b6267, { // "big triangular grid"
   "duża siatka trójkątna",
/*MISSING*/    "big triangular grid",
   "velká trojúhelníková mřížka",
   "большая треугольная сетка",
/*MISSING*/    "big triangular grid",
/*MISSING*/    "big triangular grid",
    }},
  {0x8f57a0ec, { // "Note for mobiles"
   "Notka dla urządzeń mobilnych",
/*MISSING*/    "Note for mobiles",
   "Poznámka pro mobilní telefony",
   "Замечание для мобильных устройств",
/*MISSING*/    "Note for mobiles",
/*MISSING*/    "Note for mobiles",
    }},
  {0x8f624b06, { // "message flash time"
   "czas pokazywania wiadomości",
   "mesaj gösterme süresi",
   "doba blikání zprávy",
   "время показа сообщений",
   "Nachrichten-Zeit",
/*MISSING*/    "message flash time",
    }},
  {0x8f7cc258, { // "fullscreen mode"
   "tryb pełnoekranowy",
   "tam ekran modu",
   "zobrazení na celou obrazovku",
   "полноэкранный режим",
   "Vollbild",
   "modo de tela cheia",
    }},
  {0x8fd09441, { // "This Orb allows you to jump over an adjacent monster, killing or stunning it. You can only vault in a roughly straight line. Target a cell on the other side to use it."
   "Ta Sfera pozwala przeskoczyć sąsiadującego potwora, zabijając lub ogłuszając go. Możesz skakać tylko w linii prostej. Wyceluj w pole po drugiej stronie potwora, by użyć tej Sfery.",
/*MISSING*/    "This Orb allows you to jump over an adjacent monster, killing or stunning it. You can only vault in a roughly straight line. Target a cell on the other side to use it.",
   "Tato Sféra ti umožňuje přeskočit přes sousedního netvora, a tím ho zabít nebo omráčit. Vyběhnout můžeš pouze zhruba po přímce. Sféra se používá zacílováním políčka na druhé straně netvora.",
   "Эта сфера позволяет прыгать через соседнего монстра, убивая или оглушая его. Можно прыгать только по примерно пряной линии. Выберите клетку на другой стороне, чтобы использовать сферу.",
/*MISSING*/    "This Orb allows you to jump over an adjacent monster, killing or stunning it. You can only vault in a roughly straight line. Target a cell on the other side to use it.",
/*MISSING*/    "This Orb allows you to jump over an adjacent monster, killing or stunning it. You can only vault in a roughly straight line. Target a cell on the other side to use it.",
    }},
  {0x8fd8fcf3, { // "Beltrami-Klein model"
   "Model Beltramiego-Kleina",
/*MISSING*/    "Beltrami-Klein model",
   "Beltrami-Kleinův model",
   "Модель Бельтрами-Клейна",
/*MISSING*/    "Beltrami-Klein model",
/*MISSING*/    "Beltrami-Klein model",
    }},
  {0x8fdb2403, { // "Herds of these Bulls are running long distances for some reason. They become Raging Bulls if something stops them."
   "Stada tych Byków biegną długie dystansy z jakiegoś powodu. Gdy coś je zatrzyma, stają się Wściekłymi Bykami.",
/*MISSING*/    "Herds of these Bulls are running long distances for some reason. They become Raging Bulls if something stops them.",
   "Sáda těchto Býků z nějakého důvodu běhají na dlouhé vzdálenosti. Pokud je něco zastaví, promění se ve Vzteklé Býky.",
   "Стада этих быков зачем-то пробегают огромные расстояния. Они становятся Бешеными быками, если что-то их остановит.",
/*MISSING*/    "Herds of these Bulls are running long distances for some reason. They become Raging Bulls if something stops them.",
/*MISSING*/    "Herds of these Bulls are running long distances for some reason. They become Raging Bulls if something stops them.",
    }},
  {0x900ef316, { // "pan up"
   "przewiń w górę",
   "yukarı kaydır",
   "scrollování nahoru",
   "крутить вверх",
   "Hochschwenken",
/*MISSING*/    "pan up",
    }},
  {0x9032b3d6, { // "INFO"
   "INFO",
   "BİLGİ",
   "INFORMACE",
   "ИНФО",
   "INFO",
/*MISSING*/    "INFO",
    }},
  {0x906a1466, { // "Your inventory is empty."
   "Nie masz żadnych skarbów.",
   "Envanterin boş.",
   "Nic u sebe nemáš.",
   "Ваш инвентарь пуст.",
   "Dein Inventar ist leer.",
   "Seu inventário está vazio.",
    }},
  {0x90719f73, { // "Cannot create temporary matter on an item!"
   "Nie można tworzyć tymczasowej materii na przedmiocie!",
   "Bir eşyanın üzerinde geçici madde oluşturamazsın!",
   "Dočasnou hmotu nelze vytvořit na předmětu!",
   "Нельзя создать временную материю на предмете!",
   "Temporäre Materie kann nicht auf einem Item erzeugt werden!",
/*MISSING*/    "Cannot create temporary matter on an item!",
    }},
  {0x90acac78, { // "There must be some temples of Cthulhu in R'Lyeh..."
   "Muszą być jakieś świątynie Cthulhu w R'Lyeh...",
   "R'Lyeh'te Cthulhu'ya adanmış bazı tapınaklar olmalı...",
   "Cthulhu musí mít V R'Lyeh nějaké Chrámy...",
   "В Р'Льехе должны быть храмы Ктулху...",
   "Es muss doch auch Tempel des Cthulhu in R'Lyeh geben...",
/*MISSING*/    "There must be some temples of Cthulhu in R'Lyeh...",
    }},
  {0x90df12d0, { // "images"
   "obrazki",
/*MISSING*/    "images",
   "obrázky",
   "изображения",
/*MISSING*/    "images",
/*MISSING*/    "images",
    }},
  {0x90e25cd6, { // "anti-aliasing enabled"
   "włączono anti-aliasing",
   "Anti=alising devrede/",
   "anti-aliasing zapnut",
   "сглаживание включено",
   "Anti-Aliasing aktiviert",
   "Antiserrilhamento habilitado",
    }},
  {0x90e37bd4, { // "A weirdly colored rock. Hyperentomologists claim that the Hyperbug armies use these rocks to navigate back home after a victorious battle."
   "Ta skała ma dziwny kolor. Hiperentomolodzy twierdzą, że roje Hiperinsektów używają tych skał do nawigacji, by wrócić do domu po wygranej bitwie.",
   "Garip renkli bir kaya. Aşkınböcekbilimcilere göre aşkın böcek orduları bu kayaları zaferlerinden sonra evlerini bulmak için kullanıyorlar.",
   "Podivně zbarvená skála. Hyperentomologové tvrdí, že armády Hyperhmyzáků je používají k navigaci, když se vracejí domů po vítězné bitvě.",
   "Странно окрашенный камень. Гиперэнтомологи считают, что армии гипержуков используют их, чтобы вернуться после победоносной битвы домой.",
   "Ein seltsam gefärbter Stein. Hyperentomologen behaupten, die Hyperkäfer-Armeen verwenden sie, um nach einer siegreichen Schlacht nach Hause zu finden.",
/*MISSING*/    "A weirdly colored rock. Hyperentomologists claim that the Hyperbug armies use these rocks to navigate back home after a victorious battle.",
    }},
  {0x90e70e66, { // "Branches here will bear you weight, but if you use them to move (not fall) to an unstable place, they will break."
   "Te gąłęzie udźwigną Twój ciężar, ale jeśli użyjesz ich, by się przemieścić (nie spaść) na niestabilną pozycję, złamią się.",
/*MISSING*/    "Branches here will bear you weight, but if you use them to move (not fall) to an unstable place, they will break.",
   "Tyto větve tvou váhu unesou, ale pokud se z nich pohneš (kromě pádu) na nestabilní políčko, zlomí se.",
   "Эти ветви выдерживают ваш вес, но если из использовать, чтобы пойти (не упасть) на что-то неустойчивое, они сломаются.",
   "Die Äste hier werden dein Gewicht tragen, aber wenn du sie nutzt um dich zu einem instabilen Platz zu bewegen (nicht fallen), werden sie brechen.",
/*MISSING*/    "Branches here will bear you weight, but if you use them to move (not fall) to an unstable place, they will break.",
    }},
  {0x90ee6489, { // "This Orb lets you ride Dragons and other worm-like creatures. Simply move onto such a creature to ride them; while riding, you are protected from dangerous terrains and partially from attacks (they cause you to lose half of your Domination power), but you cannot collect items. When only one charge is left, you have to dismount this turn -- be very careful to make this possible, as your mount could attack you immediately!\n\nWhile riding, click on a location to order your mount to move or attack there."
   "Ta Sfera pozwala Ci jeździć Smokami i innymi wężopodobnymi stworami. Po prostu wykonaj ruch na pole z takim stworem, by go dosiąść. Podczas jazdy jesteś chronion%y0 przed niebezpiecznym terenem i częściowo przed atakami (które powodują utratę połowy Twojej mocy Dominacji), ale nie możesz zbierać przedmiotów. Kiedy zostanie tylko 1 ładunek, musisz natychmiast zsiąść -- zadbaj o to, by to było możliwe bez zostania zaatakowanym przez Twojego dawnego wierzchowca!\n\nPodczas jazdy kliknij dowolne pole, by nakazać Twojemu wierzchowcowi tam się ruszyć lub zaatakować.",
/*MISSING*/    "This Orb lets you ride Dragons and other worm-like creatures. Simply move onto such a creature to ride them; while riding, you are protected from dangerous terrains and partially from attacks (they cause you to lose half of your Domination power), but you cannot collect items. When only one charge is left, you have to dismount this turn -- be very careful to make this possible, as your mount could attack you immediately!\n\nWhile riding, click on a location to order your mount to move or attack there.",
   "Tato Sféra ti umožňuje jezdit na Dracích a jiných červovitých netvorech. Stačí se na ně pohnout, a můžeš nasednout; během jízdy jsi chráněn před nebezpečným terénem a částečně i před útoky (ty tě připraví o polovinu zbývající síly Dominance), nemůžeš ale sbírat předměty. Ve chvíli, kdy ti zbývá pouze jeden náboj, musíš v tomto kole sesednout -- dej si dobrý pozor, aby to bylo možné,  protože tvůj oř na tebe může okamžitě zaútočit!\n\n Během jízdy můžeš kliknutím nařizovat svému oři, kam se má pohnout nebo zaútočit.",
   "Эта сфера позволяет вас ездить на драконах и других червеобразных существах. Просто пойди на клетку с таким монстром, чтобы оседлать его. Пока ты едешь на существе, на тебя не действуют никакие опасные объекты и, частично, атаки (от которых вы теряете половину силы этой сферы), но вы не можете собирать предметы. Когда остаётся только один заряд, вы должны слезть с существа -- будьте осторожны, ведь оно может сразу же атаковать вас!\n\nНажимайте на клетки, чтобы заставить существо туда пойти или атаковать.",
   "Diser Orb erlaubt es dir Drachen oder ondere wurmartige Wesen zu reiten. Tritt einfach auf solch ein Wesen um es zu reiten; während des Ritts bist du vor gefährlichen Gebieten und teilweise vor Angriffen (sie verursachen den Verlust der Hälfte deiner Kraft der Dominanz) geschützt. Jedoch kannst du keine Gegenstände sammeln. Sobald nur noch eine Ladung übrig ist, musst du in diesem Zug absteigen -- achte darauf dies zu ermöglichen, denn dein Reittier könnte dich unmittelbar angreifen!\n\nWährend du reitest, klicke auf eine Zelle um deinem Reittier zu befehlen dorthin zu gehen oder dort anzugreifen.",
/*MISSING*/    "This Orb lets you ride Dragons and other worm-like creatures. Simply move onto such a creature to ride them; while riding, you are protected from dangerous terrains and partially from attacks (they cause you to lose half of your Domination power), but you cannot collect items. When only one charge is left, you have to dismount this turn -- be very careful to make this possible, as your mount could attack you immediately!\n\nWhile riding, click on a location to order your mount to move or attack there.",
    }},
  {0x91033a83, { // "You can stand here."
   "Możesz na niej stanąć.",
   "Burada durabilirsin.",
   "Na této plošince můžeš stát.",
   "Можно встать сюда.",
   "Du kannst hier stehen.",
/*MISSING*/    "You can stand here.",
    }},
  {0x9149140f, { // "Different kills required: %1.\n"
   "Różnych zabitych przeciwników wymaganych: %1.\n",
/*MISSING*/    "Different kills required: %1.\n",
   "Požadovaný počet typů zabitých nepřátel: %1.\n",
   "Необходимо убить различных существ: %1.\n",
   "Verschiedene Tötungen erforderlich: %1.\n",
/*MISSING*/    "Different kills required: %1.\n",
    }},
  {0x91ad66d1, { // "This monster has come from another world, presumably to steal our treasures. Not as fast as an Eagle, not as resilient as the guards from the Palace, and not as huge as the Mutant Ivy from the Clearing; however, they are very dangerous because of their intelligence, and access to magical powers.\n\n"
   "Ten potwór przybył z innego świata, prawdopodobnie po nasze skarby. Nie taki szybki jak Orzeł, nie taki odporny jak strażnicy Pałacu, i nie taki wielki jak Zmutowany Bluszcz z Polany; jest jednak bardzo niebezpieczny ze względu na inteligencję i dostęp do mocy magicznych.\n\n",
/*MISSING*/    "This monster has come from another world, presumably to steal our treasures. Not as fast as an Eagle, not as resilient as the guards from the Palace, and not as huge as the Mutant Ivy from the Clearing; however, they are very dangerous because of their intelligence, and access to magical powers.\n\n",
   "Tento netvor přišel z jiného světa, asi proto, aby nám ukradl poklady. Není tak rychlý jako Orel, není tak odolný jako stráže z Paláce, a není tak obrovský jako Mutantní břečtan z Mýtiny; díky své inteligenci a přístupu k magickým schopnostem je však velmi nebezpečný.\n\n",
   "Этот монстр прибыл из другого мира, видимо, чтобы украсть наши сокровища. Не такой быстрый, как орёл, не такой живучий, как стражники из Дворца, не такой живучий, как плющ-мутант с Поляны; однако они опасны из-за их интеллекта и доступа к магическим силам.\n\n",
/*MISSING*/    "This monster has come from another world, presumably to steal our treasures. Not as fast as an Eagle, not as resilient as the guards from the Palace, and not as huge as the Mutant Ivy from the Clearing; however, they are very dangerous because of their intelligence, and access to magical powers.\n\n",
/*MISSING*/    "This monster has come from another world, presumably to steal our treasures. Not as fast as an Eagle, not as resilient as the guards from the Palace, and not as huge as the Mutant Ivy from the Clearing; however, they are very dangerous because of their intelligence, and access to magical powers.\n\n",
    }},
  {0x91cc33b3, { // "Hyperbolic geometry can be shown in many ways."
   "Geometria hiperboliczna może być pokazana na wiele sposobów...",
/*MISSING*/    "Hyperbolic geometry can be shown in many ways.",
   "Hyperbolickou geometrii je možné ukázat mnoha zpùsoby...",
   "На гиперболическую геометрию можно смотреть с разных сторон.",
/*MISSING*/    "Hyperbolic geometry can be shown in many ways.",
/*MISSING*/    "Hyperbolic geometry can be shown in many ways.",
    }},
  {0x91f0ec0a, { // "clockwise"
   "zegarowo",
/*MISSING*/    "clockwise",
   "po směru hodinových ručiček",
   "по часовой стрелке",
   "im Uhrzeigersinn",
/*MISSING*/    "clockwise",
    }},
  {0x920e4875, { // "Hint: press 1 2 3 4 to change the projection."
   "Wskazówka: 1 2 3 4 zmienia rzut.",
/*MISSING*/    "Hint: press 1 2 3 4 to change the projection.",
   "Nápověda: Klávesy 1 2 3 4 mění projekci.",
   "Подсказка: 1 2 3 4 -- сменить проекцию.",
/*MISSING*/    "Hint: press 1 2 3 4 to change the projection.",
/*MISSING*/    "Hint: press 1 2 3 4 to change the projection.",
    }},
  {0x921e4397, { // "You hit %the1."
   "Trafi%łeś0 %a1.",
   "%1 düşmanına vurdun.",
   "Zasáhl jsi %a1.",
   "Вы ударили %a1.",
   "Du triffst %den1 %a1.",
/*MISSING*/    "You hit %the1.",
    }},
  {0x92345333, { // "This frozen lake is a hellish version of the Icy Land. Now, your body heat melts the floor, not the walls."
   "Zamarznięte jezioro. Piekielna wersja Krainy Lodu, w której ciepło Twojego ciała topi ziemię pod Tobą, a nie ściany.",
   "Bu donmuş göl, Buz Diyarlarının korkunç bir versiyonu gibi. Şimdi, vücut ısın duvarları değil, yeri eritiyor.",
   "Toto zamrzlé jezero je pekelnou verzí Ledového kraje. Tady tvým tělesným teplem netají zdi, ale podlaha.",
   "Замёрзшее озеро - адский вариант Ледяной земли. Но сейчас тепло плавит пол, а не стены.",
   "Dieser gefrorene See ist eine höllische Version des Eisigen Landes. Hier schmilzt deine Körperwärme den Boden anstatt der Wände.",
/*MISSING*/    "This frozen lake is a hellish version of the Icy Land. Now, your body heat melts the floor, not the walls.",
    }},
  {0x924212cc, { // "five players"
   "pięciu graczy",
/*MISSING*/    "five players",
   "pět hráčů",
   "пять игроков",
   "fünf Spieler",
/*MISSING*/    "five players",
    }},
  {0x92642dd1, { // "A Troll without the power of Life."
   "Troll bez mocy Życia.",
   "Hayat gücünden mahrum bir trol.",
   "Troll bez síly Života.",
   "Тролль без жизненной силы.",
   "Ein Troll ohne die Kraft des Lebens.",
/*MISSING*/    "A Troll without the power of Life.",
    }},
  {0x92790eef, { // "or ESC to see how it ended"
   "lub ESC by zobaczyć, jak się skończyło",
   "veya nasıl bittiğini görmek için ESC'ye bas",
   "nebo ESC, abys vidě%l0, jak to skončilo",
   "либо ESC, чтобы посмотреть, как всё закончилось",
   "oder ESC um zu sehen, wie alles endete",
/*MISSING*/    "or ESC to see how it ended",
    }},
  {0x92859386, { // "This orb can protect you from damage."
   "Ta sfera chroni Cię przed obrażeniami od potworów.",
   "Bu küre seni hasardan korur.",
   "Tato sféra tě dokáže chránit před zraněním.",
   "Эта сфера защитит Вас от всех атак.",
   "Dieser Orb kann dich vor Schaden schützen.",
/*MISSING*/    "This orb can protect you from damage.",
    }},
  {0x9297e135, { // "use your Orbs in tough situations"
   "używaj Sfer w trudnych sytuacjach",
/*MISSING*/    "use your Orbs in tough situations",
   "v obtížných situacích používej Sféry",
   "используй Сферы в трудных ситуациях",
/*MISSING*/    "use your Orbs in tough situations",
/*MISSING*/    "use your Orbs in tough situations",
    }},
  {0x92b415c8, { // "You push %the1 behind you!"
   "Popchn%ąłeś0 %a1 za siebie.",
   "%a1 arkana ittin.",
   "Odstrčil jsi %a1 za sebe.",
   "Вы передвинули %a1 за себя.",
   "Du schiebst %den1 %a1 hinter dich.",
/*MISSING*/    "You push %the1 behind you!",
    }},
  {0x92c84a71, { // "move up-right"
   "ruch w górę-prawo",
/*MISSING*/    "move up-right",
   "pohyb doprava nahoru",
   "ход вверх-вправо",
/*MISSING*/    "move up-right",
/*MISSING*/    "move up-right",
    }},
  {0x92e9560d, { // "help for keyboard users"
   "pomoc dla użytkowników klawiatury",
/*MISSING*/    "help for keyboard users",
   "pomoc pro uživatele klávesnice",
   "помощь для пользователей клавиатуры",
/*MISSING*/    "help for keyboard users",
/*MISSING*/    "help for keyboard users",
    }},
  {0x92e9c33d, { // "Euclidean mode"
   "tryb euklidesowy",
   "Öklid modu",
   "Eukleidovský mód",
   "режим евклидовой плоскости",
   "Euklidischer Modus",
/*MISSING*/    "Euclidean mode",
    }},
  {0x931ed7e1, { // "This Orb summons a friendly Bomberbird."
   "Ta sfera przywołuje przyjaznego Bombardiera.",
   "Bu Küre barışçıl bir Bombacıkuş çıkarır.",
   "Tato sféra vyvolá přátelského Bombarďáka.",
   "Эта сфера вызывает ручного Бомбардира.",
   "Dieser Orb ruft einen verbündeten Bombenvogel herbei.",
/*MISSING*/    "This Orb summons a friendly Bomberbird.",
    }},
  {0x935a45ec, { // "Map saved to %1"
   "Mapa zapisana do %1",
   "Harita %1 konumuna kaydedildi",
   "Mapa uložena do %1",
   "Карта сохранена в %1",
   "Karte unter %1 gespeichert",
/*MISSING*/    "Map saved to %1",
    }},
  {0x935ce07a, { // "useless in %the1"
   "bezużyteczna %abl1",
   "%abl1 işe yaramaz",
   "k ničemu %abl1",
   "бесполезна %abl1",
   "nutzlos in %der1",
/*MISSING*/    "useless in %the1",
    }},
  {0x9373d1a6, { // "summon Orb of Yendor"
   "przywołaj sferę Yendoru",
   "Yendorun Küresini çıkar",
   "vyvolej Yendorskou sféru",
   "призвать сферы Йендора",
   "rufe einen Orb von Yendor herbei",
/*MISSING*/    "summon Orb of Yendor",
    }},
  {0x93e87d87, { // "A rare and expensive substance found in the Desert. It is believed to extend life and raise special psychic powers."
   "Rzadka i droga substancja z Pustyni. Podobno wydłuża życie i daje moce psychiczne.",
   "Çölde bulunan ender ve pahalı bir katkı maddesi. Yaşamı uzattığına ve bazı özel psişik güçleri açığa çıkardığına inanılıyor.",
   "Drahá a vzácná látka, která se nachází v Poušti. Věří se, že prodlužuje život a zlepšuje zvláštní psychické schopnosti.",
   "Редкое и дорогое вещество из Пустыни. Говорят, продлевает срок жизни и даёт сверхспособности.",
   "Eine seltene und teure Substanz aus der Wüste. Man sagt, sie verlängert das Leben und verleiht besondere psychische Kräfte.",
/*MISSING*/    "A rare and expensive substance found in the Desert. It is believed to extend life and raise special psychic powers.",
    }},
  {0x93fa3f18, { // "Defeat %1 enemies to access the Graveyard"
   "Pokonaj %1 przeciwników, by trafić na Cmentarz",
   "Mezarlığa erişmek için %1 düşman öldür.",
   "Hřbitov zpřístupníš poražením %1 nepřátel",
   "Убейте %1 врагов, чтобы открыть Кладбище",
   "Töte %1 Gegner um Zugang zum Friedhof zu erhalten",
   "Derrote %1 inimigos para acessar o Cemitério",
    }},
  {0x93fe8846, { // "You chop down %the1."
   "Ści%ąłeś0 %a1.",
/*MISSING*/    "You chop down %the1.",
   "Pokáce%l0 jsi %a1.",
   "Вы срубили %a1.",
   "Du hackst %den1 %a1 ab.",
/*MISSING*/    "You chop down %the1.",
    }},
  {0x940873d4, { // "This beautiful flower is greatly prized by wizards, as it allows them to cast powerful magical spells without preparation.\n"
   "Ten piękny kwiat ma wysoką cenę wśród czarodziejów, bo pozwala im rzucać silne czary bez przygotowania.\n",
/*MISSING*/    "This beautiful flower is greatly prized by wizards, as it allows them to cast powerful magical spells without preparation.\n",
   "Tento nádherný květ si velmi cení čarodějové, neboť jim umožňuje sesílat silná kouzla bez jakékoli přípravy.\n",
   "Этот прекрасный цветок очень ценится колдунами, так как позволяет создавать мощные чары без подготовки.\n",
   "Diese schöne Blume ist hochgeschätzt bei Zauberern, denn sie erlaubt es ihnen, ohne Vorbereitung machtvolle magische Zauber zu wirken.\n",
/*MISSING*/    "This beautiful flower is greatly prized by wizards, as it allows them to cast powerful magical spells without preparation.\n",
    }},
  {0x9447d99f, { // "A rare unliving construct from the Dead Caves. It instantly destroys cave walls next to its path, and also leaves an impassable wall behind it. You suppose that this impassable wall helps it to escape from some threats. You hope you won't meet these threats..."
   "Rzadka konstrukcja z Martwych Jaskiń. Natychmiast niszczy skalne ściany przy jego trasie i tworzy za sobą ścianę skały. Przypuszczasz, że ta ściana pomaga mu uciekać przed jakimiś zagrożeniami. Masz nadzieję, że nie spotkasz tych zagrożeń...",
   "Ölü Mağaralardan cansız bir mekanizma. Yoluna çıkan tüm mağara duvarlarını yıkar ve arkasında geçilemeyen bir duvar bırakır.Bu duvarları birtakım tehditlerden kullanabileceğini düşünüyorsun. Aynı zamanda bu tehditlerle hiç karşılaşmayacağını da umuyorsun..",
   "Vzácný neživý konstrukt z Mrtvých jeskyní. Okamžitě ničí stěny jeskyně kolem své dráhy a zanechává za sebou neprostupnou zeď. Tato neprostupná zeď mu zřejmě umožňuje uniknout před nebezpečím. Doufáš, že na tato nebezpečí nenarazíš...",
   "Редкое неживое создание из Мёртвых пещер. Мгновенно разрушает стены на своём пути и оставляет стены за собой.",
   "Ein seltenes untotes Geschöpf aus den Toten Höhlen. Es zerstört unmittelbar alle Höhlenwände vor und neben sich und hinterlässt wiederum undurchdringbare Wände. Du vermutest, dass diese Wände ihm helfen sollen irgendwelchen Gefahren zu entkommen. Hoffentlich wirst du nicht sehen, wovor es flieht...",
/*MISSING*/    "A rare unliving construct from the Dead Caves. It instantly destroys cave walls next to its path, and also leaves an impassable wall behind it. You suppose that this impassable wall helps it to escape from some threats. You hope you won't meet these threats...",
    }},
  {0x9455bd4b, { // "Y shift"
   "przesunięcie Y",
/*MISSING*/    "Y shift",
   "Y posuv",
   "сдвиг по Y",
/*MISSING*/    "Y shift",
/*MISSING*/    "Y shift",
    }},
  {0x9490fa41, { // "This tutorial is different than most other game tutorials -- you are not forced to do anything, and you can go wherever you want.\n\nHowever, %the1 is not what we are talking about now. We will not explain this land at the moment, and you could ponentially get lost there.\n\nRemember that you can get to the next slide by pressing Enter."
   "Ten podręcznik jest inny niż większość innych -- nie masz obowiązku robić tego co tutorial chce, możesz iść dokąd chcesz.\n\nAle %1 nie jest miejscem, o którym teraz mówimy. Nie wyjaśnimy go, i potencjalnie możesz się zgubić.\n\nNaciśnij Enter by przejść do kolejnego slajdu.",
/*MISSING*/    "This tutorial is different than most other game tutorials -- you are not forced to do anything, and you can go wherever you want.\n\nHowever, %the1 is not what we are talking about now. We will not explain this land at the moment, and you could ponentially get lost there.\n\nRemember that you can get to the next slide by pressing Enter.",
   "Tento tutoriál funguje jinak, než ve většině her -- k ničemu tě nenutíme a můžeš si jít, kam chceš.\n\nAle %1 není to, o čem teď mluvíme. Tento kraj momentálně tutoriál nevysvětluje a je možné, že se v něm ztratíš.\n\nNezapomeň, že stiskem klávesy Enter můžeš přejít na další snímek.",
   "Это руководство отличается от большинства других -- ты не обязан ничего делать, иди куда хочешь.\n\nОднако %1 -- не то, о чём мы сейчас говорим. Мы не будем сейчас объяснять эту землю, и ты можешь заблудиться в ней.\n\nНажми Enter, чтобы перейти на следующий слайд.",
/*MISSING*/    "This tutorial is different than most other game tutorials -- you are not forced to do anything, and you can go wherever you want.\n\nHowever, %the1 is not what we are talking about now. We will not explain this land at the moment, and you could ponentially get lost there.\n\nRemember that you can get to the next slide by pressing Enter.",
/*MISSING*/    "This tutorial is different than most other game tutorials -- you are not forced to do anything, and you can go wherever you want.\n\nHowever, %the1 is not what we are talking about now. We will not explain this land at the moment, and you could ponentially get lost there.\n\nRemember that you can get to the next slide by pressing Enter.",
    }},
  {0x949e4d00, { // "\"The Red Rock Valley is dangerous, but beautiful.\""
   "\"Dolina Czerwonej Skały jest piękna, choć niebezpieczna.\"",
   "\"KızılKaya Vadisi tehlikeli ama manzara çok güzel.\"",
   "\"Údolí červených skal je nebezpečné, ale krásné.\"",
   "\"Долина красных камней опасна, но прекрасна.\"",
   "\"Das Rotfelsental ist gefährlich aber wunderschön.\"",
/*MISSING*/    "\"The Red Rock Valley is dangerous, but beautiful.\"",
    }},
  {0x949f3d3c, { // "Euclidean geometry"
   "geometria euklidesowa",
/*MISSING*/    "Euclidean geometry",
   "eukleidovská geometrie",
   "евклидова геометрия",
/*MISSING*/    "Euclidean geometry",
/*MISSING*/    "Euclidean geometry",
    }},
  {0x94f9c468, { // "Love takes time, but it heals all wounds, and transcends time and space.\n\nThe Orb of Love is worth 30$$$, if you end the game with it.\n"
   "Miłość potrzebuje czasu, ale leczy wszelkie rany; czas i przestrzeń nie są dla niej przeszkodami.\n\nSfera Miłości jest warta 30$$$, jeśli skończysz z nią grę.\n",
   "Aşk zaman alır, ama her yarayı iyileştirir ve zaman ve mekâna galip gelir.\n\nAşk Küresi oyunu onunla bitirebilirsen 30$$$ eder.\n",
   "Láska chce čas, ale zahojí všechny rány a překoná čas i prostor.\n\nSféra lásky má cenu 30$$$, pokud ji máš u sebe na konci hry.\n",
   "Любовь требует времени, но лечит все раны, и проходит сквозь время и пространство.\n\nСфера любви стоит 30$$$, если закончить игру с ней.\n",
   "Liebe braucht Zeit, aber sie heilt alle Wunden und überschreitet die Grenzen von Raum und Zeit.\n\nDer Orb der Liebe ist 30$$$ wert wenn du das Spiel mit ihm beendest.\n",
/*MISSING*/    "Love takes time, but it heals all wounds, and transcends time and space.\n\nThe Orb of Love is worth 30$$$, if you end the game with it.\n",
    }},
  {0x95334578, { // "Escher"
   "Escher",
   "Escher",
   "escherovský",
   "Эшер",
   "Escher",
   "Escher",
    }},
  {0x957837bd, { // "You can touch the Dead Orb in your inventory to drop it."
   "Dotknij Martwej Sfery na liście, by ją porzucić.",
   "Ölü Küreye dokunarak onu bırakabilirsin.",
   "Mrtvou sféru ve svém inventáři můžeš položit tím, že se jí dotkneš.",
   "Коснитесь мёртвой сферы, чтобы собрать её.",
   "Du kannst den toten Orb in deinem Inventar berühren, um ihn abzulegen.",
   "Você pode tocar o Orbe Morto no seu inventário para descartá-lo.",
    }},
  {0x95cf0169, { // "exit menu"
   "wyjdź z menu",
/*MISSING*/    "exit menu",
   "opusť menu",
   "выйти из меню",
/*MISSING*/    "exit menu",
/*MISSING*/    "exit menu",
    }},
  {0x96103279, { // "F4 = file"
   "F4 = plik",
   "F4 = dosya",
   "F4 = soubor",
   "F4 = файл",
   "F4 = Datei",
/*MISSING*/    "F4 = file",
    }},
  {0x96433cdb, { // "key Orb power"
   "moc sfery jednym klawiszem",
/*MISSING*/    "key Orb power",
   "moc sfery jednym klawiszem",
   "клавиша для силы сферы",
   "Hauptkraft des Orb",
/*MISSING*/    "key Orb power",
    }},
  {0x9653c1d0, { // "Wow! That was close."
   "Uff! Było blisko.",
/*MISSING*/    "Wow! That was close.",
   "Uf! To bylo o chlup!",
   "Ох! Это было близко.",
   "Wow! Das war knapp.",
/*MISSING*/    "Wow! That was close.",
    }},
  {0x9657dce2, { // "elliptic"
   "eliptyczna",
/*MISSING*/    "elliptic",
   "eliptická",
   "эллиптическая",
/*MISSING*/    "elliptic",
/*MISSING*/    "elliptic",
    }},
  {0x966f306e, { // "Failed to load map from %1"
   "Nie udało się wczytać %1",
   "Harita %1 konumundan yüklenemedi",
   "Nahrání mapy z %1 se nezdařilo",
   "Не удалось загрузить карту из %1",
   "Karte konnte nicht von %1 geladen werden",
/*MISSING*/    "Failed to load map from %1",
    }},
  {0x9698c7bc, { // "This Orb creates a warped zone of radius 5 around you, and also allows you to move diagonally in warped zones."
   "Ta sfera tworzy wokół Ciebie w promieniu zakrzywioną przestrzeń, a także pozwala Ci się poruszać po przekątnej w zakrzywionej przestrzeni.",
/*MISSING*/    "This Orb creates a warped zone of radius 5 around you, and also allows you to move diagonally in warped zones.",
   "Tato Sféra kolem tebe vytvoří pokřivenou zónu o poloměru 5, a kromě toho ti také umožňuje úhlopříčný pohyb v pokřivených zónách.",
   "Эта сфера создаёт искривление в радиусе 5 клеток вокруг Вас, и позволяет Вам ходить по-диагонали в искривлённых зонах.",
   "Dieser Orb erzeugt einen verzerrten Bereich vom Radius 5 um dich herum und erlaubt es dir dich diagonal in verzerrten Bereichen zu bewegen.",
/*MISSING*/    "This Orb creates a warped zone of radius 5 around you, and also allows you to move diagonally in warped zones.",
    }},
  {0x96c5e7f7, { // "Static mode disabled."
   "Tryb statyczny wyłączony.",
/*MISSING*/    "Static mode disabled.",
   "Statický mód vypnut.",
   "Статичный режим выключен.",
/*MISSING*/    "Static mode disabled.",
/*MISSING*/    "Static mode disabled.",
    }},
  {0x96d4d8a4, { // "(v) menu"
   "(v) menu",
   "(v) menü",
   "(v) menu",
   "(v) меню",
   "Menü - v",
/*MISSING*/    "(v) menu",
    }},
  {0x97043e71, { // "Press Enter or F10 to quit"
   "Wciśnij Enter lub F10, by wyjść z gry",
   "Enter ya da F10 ile çık",
   "Enter nebo F10: ukončit hru",
   "Нажмите Enter или F10, чтобы выйти",
   "Drücke Enter oder F10 zum Verlassen des Spiels",
   "Pressione Enter ou F10 para sair",
    }},
  {0x9718515d, { // "%The1 squeaks in despair."
   "%1 pisn%ął1 rozpaczliwie.",
   "%1 karamsarlıkla ciyaklıyor.",
   "%1 zoufale piští.",
   "%1 пищит в отчаянии.",
   "%Die1 %1 piepst voll Verzweiflung.",
/*MISSING*/    "%The1 squeaks in despair.",
    }},
  {0x971c306d, { // "coefficient"
   "współczynnik",
/*MISSING*/    "coefficient",
   "koeficient",
   "коэффициент",
   "Koeffizient",
/*MISSING*/    "coefficient",
    }},
  {0x971caf7e, { // "light"
   "lekki",
   "hafif",
   "lehký",
   "легкий",
   "einfach",
   "leve",
    }},
  {0x9754bf46, { // "Unavailable in the shmup mode.\n"
   "Niedostępne w trybie strzelanki.\n",
/*MISSING*/    "Unavailable in the shmup mode.\n",
   "Nedostupné v módu střílečky.\n",
   "Недоступно в режиме стрельбы.\n",
   "Nicht im shmup-Modus verfügbar.\n",
/*MISSING*/    "Unavailable in the shmup mode.\n",
    }},
  {0x975cfefc, { // "A fresh grave. Necromancers like those."
   "Świeży grób. Nekromanci to lubią.",
   "Taze bir mezar. Ölüçağıranlar bunu sever.",
   "Čerstvý hrob. Ty mají rádi nekromanceři.",
   "Свежая могила. Некроманты их любят.",
   "Ein frisches Grab. Totenbeschwörer mögen diese.",
/*MISSING*/    "A fresh grave. Necromancers like those.",
    }},
  {0x975dd65d, { // "Orbs summoned!"
   "Sfery mocy przywołane!",
   "Küreler çağırdın!",
   "Vyvola%l0 jsi sféry!",
   "Сферы призваны!",
   "Orbs beschworen!",
   "Orbes invocados!",
    }},
  {0x97b380ef, { // "Hardness frozen at %1."
   "Trudność zamrożona: %1.",
   "Trudność zamrożona: %1.",
   "Obtížnost nastavená na: %1.",
   "Сложность заморожена на %1.",
   "Festgefroren: %1.",
/*MISSING*/    "Hardness frozen at %1.",
    }},
  {0x97b82241, { // "Cannot move into the current location of another player!"
   "Nie możesz wejść na miejsce innego gracza!",
/*MISSING*/    "Cannot move into the current location of another player!",
   "Nemůžeš vstoupit na políčko obsazené jiným hráčem!",
   "Нельзя пойти в клетку, занятую другим игроком!",
   "Kann nicht auf die momentane Position eines anderen Spielers gehen!",
/*MISSING*/    "Cannot move into the current location of another player!",
    }},
  {0x97bc983b, { // "%The1 shoots you!"
   "%1 do Ciebie strzeli%ł!",
   "%1 sana ateş etti!",
   "%1 tě střeli%l!",
   "%1 застрелит тебя!",
   "%Der1 %1 erschießt dich!",
/*MISSING*/    "%The1 shoots you!",
    }},
  {0x97cc96e0, { // "A large bird who likes strong winds. Just as you, it can fly quickly in the wind."
   "Wielki ptak, lubiący silne wiatry. Tak jak Ty może szybko latać z wiatrem.",
/*MISSING*/    "A large bird who likes strong winds. Just as you, it can fly quickly in the wind.",
   "Welký pták, který má rád silný vítr. Dokáže v něm rychle létat, stejně jako ty.",
   "Большая птица, которая любит сильный ветер. Как и Вы, движется быстрее с ветром.",
   "Ein großer Vogel, der starken Wind liebt. Wie du kann er schnell mit dem Wind fliegen.",
/*MISSING*/    "A large bird who likes strong winds. Just as you, it can fly quickly in the wind.",
    }},
  {0x97e25bca, { // "change the alpha parameter to show the lines"
   "zmień parametr alfa by linie były widoczne",
/*MISSING*/    "change the alpha parameter to show the lines",
   "čáry můžeš zviditelnit změnou parametru alfa",
   "измените параметр альфа, чтобы увидеть эти линии",
/*MISSING*/    "change the alpha parameter to show the lines",
/*MISSING*/    "change the alpha parameter to show the lines",
    }},
  {0x97e2cdae, { // "\n\nThis Orb is triggered on your first attack or illegal move."
   "\n\nTa Sfera się aktywuje automatycznie, gdy wykonasz atak lub nielegalny ruch.",
/*MISSING*/    "\n\nThis Orb is triggered on your first attack or illegal move.",
   "\n\nTato sféra se aktivuje prvním útokem nebo ilegálním tahem.",
   "\n\nЭта сфера срабатывает при атаке или невозможном ходе.",
   "\n\nDieser Orb wird bei deiner ersten Attacke oder deinem ersten unerlaubten Zug ausgelöst.",
/*MISSING*/    "\n\nThis Orb is triggered on your first attack or illegal move.",
    }},
  {0x98012938, { // "\n\nFast flying creatures may attack or go against gravity only in their first move."
   "\n\nSzybko latające stworzenia mogą atakować i poruszać się wbrew grawitacji tylko w swoim pierwszym ruchu.",
/*MISSING*/    "\n\nFast flying creatures may attack or go against gravity only in their first move.",
   "\n\nRychlí létající tvorové mohou útočit nebo se protivit gravitaci pouze při svém prvním pohybu.",
   "\n\nБыстро летающие существа могут двигаться или атаковать против гравитации только первым своим ходом.",
/*MISSING*/    "\n\nFast flying creatures may attack or go against gravity only in their first move.",
/*MISSING*/    "\n\nFast flying creatures may attack or go against gravity only in their first move.",
    }},
  {0x982b5c26, { // "Zebra Pattern"
   "Wzór Zebra",
   "Zebra Dokusu",
   "Vzor 'Zebra'",
   "узор Зебра",
   "Zebramuster",
/*MISSING*/    "Zebra Pattern",
    }},
  {0x982ba756, { // "Ancient Viking heroes were buried here. Their graves have barrows raised over them, and are guarded by Draugar, animated corpses who are immune to mundane weapons. You will need to use a magical weapon spell to defeat them, and to rob the ancient jewelry buried in the graves."
   "Tu pochowano starożytnych wikińskich bohaterów. Nad ich grobami wzniesiono kurhany. Kurhany są strzeżone przez Draugry, ożywione zwłoki odporne na zwyczajne bronie. Potrzebujesz magicznej broni, by je pokonać i zrabować starożytną biżuterię zakopaną w kurhanach.",
/*MISSING*/    "Ancient Viking heroes were buried here. Their graves have barrows raised over them, and are guarded by Draugar, animated corpses who are immune to mundane weapons. You will need to use a magical weapon spell to defeat them, and to rob the ancient jewelry buried in the graves.",
   "Tady jsou pohřbeni dávní vikingští hrdinové. Nad jejich hroby jsou postaveny mohyly a střeží je Draugarové, oživlé mrtvoly imunní proti obyčejným zbraním. Pokud je chceš porazit a uloupit prastaré klenoty pohřbené v jejich hrobech, budeš si muset nějak vyčarovat magickou zbraň.",
   "Древние герои Викингов похоронены здесь. Над их могилами построены курганы. охраняемые Драуграми -- живыми мертвецами, невосприимчивыми к обычному оружию. Вам нужно магическое оружие, чтобы победить их и добраться до драгоценностей, закопанных в курганах.",
   "Die Helden der alten Wikinger wurden hier begraben. Über ihren Gräbern erheben sich Hügel, die von Draugar bewacht werden, Untote, die immun gegen irdische Waffen sind. Du wirst einen Zauber für eine magische Waffe brauchen, um sie zu besiegen und die in den Gräbern verborgenen antiken Juwelen zu rauben.",
/*MISSING*/    "Ancient Viking heroes were buried here. Their graves have barrows raised over them, and are guarded by Draugar, animated corpses who are immune to mundane weapons. You will need to use a magical weapon spell to defeat them, and to rob the ancient jewelry buried in the graves.",
    }},
  {0x983cfb96, { // "dress color II"
   "kolor sukienki II",
/*MISSING*/    "dress color II",
   "barva šatů II",
   "цвет одежды II",
   "Kleiderfarbe II",
/*MISSING*/    "dress color II",
    }},
  {0x98941047, { // "The ratio of Rock III to walls is %1, so Rock III are %2 absolute units high. Length of paths on the Rock III level is %3 of the corresponding length on the ground level."
   "Iloraz Skały III do ściany to %1, czyli Skała III ma wysokość %2. Drogi na poziomie Skały III są %3 odpowiednich odległości na poziomie podłoża.",
/*MISSING*/    "The ratio of Rock III to walls is %1, so Rock III are %2 absolute units high. Length of paths on the Rock III level is %3 of the corresponding length on the ground level.",
   "Poměr Skály III a zdí je %1, takže Skála III je vysoká %2 absolutních jednotek. Délka cest na úrovni Skály III je %3-násobkem odpovídající délky na úrovni země..",
   "Отношение высот скалы-III и стены равно %1, так что скала-III имеет высоту %2. Длина пути на такой высоте в %3 раз больше длины на уровне земли.",
/*MISSING*/    "The ratio of Rock III to walls is %1, so Rock III are %2 absolute units high. Length of paths on the Rock III level is %3 of the corresponding length on the ground level.",
/*MISSING*/    "The ratio of Rock III to walls is %1, so Rock III are %2 absolute units high. Length of paths on the Rock III level is %3 of the corresponding length on the ground level.",
    }},
  {0x98d5754f, { // "As you leave, your powers are drained!"
   "Opuszczasz Krainę Mocy... Twoja moc znikła!",
   "Ayrılırken, tüm güçlerinin tükendiğini farkettin!",
   "Opouštíš Kraj Moci... Tvá síla mizí!",
   "Покидая землю силы, Вы теряете магию сфер!",
   "Deine Kraft verblasst, als du gehst.",
/*MISSING*/    "As you leave, your powers are drained!",
    }},
  {0x98f031b2, { // "Note: currently scores are saved only in the normal mode on Android"
   "Uwaga: wyniki na Androidzie są zapisywane tylko w normalnym trybie",
/*MISSING*/    "Note: currently scores are saved only in the normal mode on Android",
   "Poznámka: na Androidu se skóre v tuto chvíli ukládají pouze v normálním módu",
   "Важно: на Андроиде счёт записывается только в обычном режиме",
   "Hinweis: aktuell werden bei Android Punktestände nur im Normalmodus gespeichert",
/*MISSING*/    "Note: currently scores are saved only in the normal mode on Android",
    }},
  {0x98f9d752, { // "three colors"
   "trzy kolory",
/*MISSING*/    "three colors",
   "tři barvy",
   "три цвета",
/*MISSING*/    "three colors",
/*MISSING*/    "three colors",
    }},
  {0x9908e6b5, { // "You have improved both your real time and knife count. Congratulations!"
   "Poprawi%łeś0 swój najlepszy czas rzeczywisty i liczbę noży. Gratulacje!",
   "Hem toplam zamanını hem de bıçak sayını geliştirdin. Tebrikler!",
   "Zlepšil jsi jak svůj čas, tak počet nožů. Gratulujeme!",
   "Вы улучшили результаты по реальному времени и количеству ножей. Поздравляем!",
   "Du hast sowohl weniger Zeit sowie Messer als zuvor benutzt. Glückwunsch!",
/*MISSING*/    "You have improved both your real time and knife count. Congratulations!",
    }},
  {0x996d58ad, { // "Could not create an image of that size."
   "Nie da się stworzyć obrazu tej wielkości.",
/*MISSING*/    "Could not create an image of that size.",
   "Obrázek této velikosti nebylo možné vytvořit.",
   "Невозможно создать изображение такого размера.",
   "Konnte ein Bild dieser Größe nicht erzeugen.",
/*MISSING*/    "Could not create an image of that size.",
    }},
  {0x9972c6a9, { // "PICK"
   "TEN",
/*MISSING*/    "PICK",
   "VYBRAT",
   "ВЫБИРАЙ",
/*MISSING*/    "PICK",
/*MISSING*/    "PICK",
    }},
  {0x99887770, { // "create a baby tortoise"
   "stwórz żółwika",
/*MISSING*/    "create a baby tortoise",
   "vytvoř želvičku",
   "создай черепашку",
/*MISSING*/    "create a baby tortoise",
/*MISSING*/    "create a baby tortoise",
    }},
  {0x9990ffba, { // "You command %the1!"
   "%1 przyjmuje Twój rozkaz.",
/*MISSING*/    "You command %the1!",
   "%1 přijímá tvůj rozkaz.",
   "Вы оседлали %a1!.",
   "Du befehligst %den1 %1!",
/*MISSING*/    "You command %the1!",
    }},
  {0x99efd678, { // "switch ghost timer"
   "przestaw czaso-duchy",
/*MISSING*/    "switch ghost timer",
   "přepni časové duchy",
   "переключить таймер призраков",
   "Geister-Zähler umschalten",
/*MISSING*/    "switch ghost timer",
    }},
  {0x9a0d6e13, { // "\n\nIn the Orb Strategy Mode, Orbs of Yendor appear in Hell after you collect 25 Demon Daisies in Hell, in Crossroads/Ocean after you collect 50, and everywhere after you collect 100."
   "\n\nW trybie strategii sfer Sfery Yendoru się pojawiają w Piekle po zebraniu 25 Czarcich Ziel, na Skrzyżowaniu/Oceanie po zebraniu 50, wszędzie po zebraniu 100.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, Orbs of Yendor appear in Hell after you collect 25 Demon Daisies in Hell, in Crossroads/Ocean after you collect 50, and everywhere after you collect 100.",
   "\n\nVe sférostrategickém módu se Yendorské Sféry objevují v Pekle poté, co tam získáš 25 Èertových kvítek. Po získání 50 Èertových kvítek se zaènou objevovat na Køižovatce a v Oceánu, a po získání 100 Èertových kvítek všude.",
   "\n\nВ режим стратегии сфер Сфера Йендора появляется в Аду, когда ты соберёшь 25 Адских Ромашек, на Перекрёстке и а Океана, когда ты соберёшь 50, и везде, когда ты соберёшь 100.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, Orbs of Yendor appear in Hell after you collect 25 Demon Daisies in Hell, in Crossroads/Ocean after you collect 50, and everywhere after you collect 100.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, Orbs of Yendor appear in Hell after you collect 25 Demon Daisies in Hell, in Crossroads/Ocean after you collect 50, and everywhere after you collect 100.",
    }},
  {0x9a327390, { // "unassign a key"
   "odwiąż klawisz",
   "bir tuş atamasını iptal et",
   "zruš klávesu",
   "отменить",
   "Taste entbinden",
/*MISSING*/    "unassign a key",
    }},
  {0x9a357317, { // "Return to the normal game"
   "Powrót do normalnej gry",
/*MISSING*/    "Return to the normal game",
   "Návrat do normální hry",
   "Вернуться в обычную игру",
   "Kehre zum normalen Spiel zurück",
/*MISSING*/    "Return to the normal game",
    }},
  {0x9aa13108, { // "Kills gained!"
   "Zdobyto trupy!",
   "Leşler kazandınız!",
   "Získa%l0 jsi zabité netvory!",
   "Убийства получены!",
   "Kills erhalten!",
   "Mortes obtidas!",
    }},
  {0x9ab34e3f, { // "\"I want my revenge. Stun a guard and leave him for me!\""
   "\"Chcę zemsty. Ogłusz strażnika i mi go zostaw!\"",
   "İntikam istiyorum! Bir muhafızı sersemlet ve onu bana bırak!\"",
   "Chci se pomstít. Omrač nějakého strážce a nech mi ho!\"",
   "Я хочу мстить! Оглуши стража и оставь его мне!\"",
   "Ich will Rache. Betäube eine Wache und überlasse sie mir!\"",
/*MISSING*/    "\"I want my revenge. Stun a guard and leave him for me!\"",
    }},
  {0x9ab47466, { // "You can choose one of the several modes"
   "Możesz wybrać jeden z kilku trybów",
   "Farklı modlardan birini seçebilirsin.",
   "Můžeš si vybrat jeden z několika módů",
   "Вы можете выбрать один из нескольких режимов",
   "Du kannst einen der verschiedenen Modi auswählen",
   "Você pode escolher um, dos vários modos disponíveis",
    }},
  {0x9ab8eee4, { // "return to the turn-based mode"
   "powrót do trybu turowego",
   "tur bazlı moda dön",
   "vrať se do módu na kola",
   "вернуться в пошаговый режим",
   "Zurück zum turn based Modus",
/*MISSING*/    "return to the turn-based mode",
    }},
  {0x9ab8f250, { // "objects found: %1"
   "znalezione przedmioty: %1",
   "bulunan eşyalar: %1",
   "předmětů nalezeno: %1",
   "найдено предметов: %1",
   "Objekte gefunden: %1",
   "objetos encontrados: %1",
    }},
  {0x9ac7e8ae, { // "Cannot create temporary matter on a monster!"
   "Nie można tworzyć tymczasowej materii na potworze!",
   "Geçici maddenin üzerinde bir canavar çıkaramazsın!",
   "Dočasnou hmotu nelze vytvořit na netvorovi!",
   "Нельзя создать временную материю на монстре!",
   "Temporäre Materie kann nicht auf einem Monster erzeugt werden!",
/*MISSING*/    "Cannot create temporary matter on a monster!",
    }},
  {0x9b2b3f7c, { // "play the game!"
   "graj!",
   "oyna!",
   "hraj!",
   "играть!",
   "Spielen!",
/*MISSING*/    "play the game!",
    }},
  {0x9b4277d8, { // "Kills required: %1.\n"
   "Wymagani pokonani przeciwnicy: %1.\n",
   "Gereken leşler: %1 $$$.\n",
   "Potřebuješ zabít %1 nepřátel.\n",
   "УБийств нужно: %1 $$$.\n",
   "Benötigte Kills: %1.\n",
/*MISSING*/    "Kills required: %1.\n",
    }},
  {0x9b894de7, { // "Curvature"
   "Krzywizna",
/*MISSING*/    "Curvature",
   "Zakřivení",
   "Кривизна",
/*MISSING*/    "Curvature",
/*MISSING*/    "Curvature",
    }},
  {0x9b9260e5, { // "\"Thank you very much for talking, and have a great rest of your day!\""
   "\"Dziękuję bardzo za rozmowę, życzę wspaniałej reszty dnia!\"",
/*MISSING*/    "\"Thank you very much for talking, and have a great rest of your day!\"",
   "\"Moc ti děkuju za rozhovor a hezký zbytek dne!\"",
   "\"Большое спасибо за беседу, удачного дня!\"",
   "\"Vielen Dank für die Unterhaltung - hab einen schönen Tag!\"",
/*MISSING*/    "\"Thank you very much for talking, and have a great rest of your day!\"",
    }},
  {0x9b9bc969, { // "Abandon all hope, the gates of Hell are opened!"
   "Porzuć wszelką nadzieję, bramy Piekła są otwarte!",
   "Umutlar tükendi, Cehennemin kapıları açıldı!",
   "Zanech vší naděje, brány Pekla jsou otevřeny!",
   "Оставь надежду, врата Ада открылись!",
   "Gib jegliche Hoffnung auf, die Pforte der Hölle ist geöffnet!",
   "Abandone toda a esperança, os portões do Inferno estão abertos!",
    }},
  {0x9bbf0d6e, { // "A black gem with white stripes. It is beautiful."
   "Piękny, drogocenny czarny kamień z białymi paskami.",
   "Beyaz şeritli siyah bir mücevher. Çok güzel.",
   "Černý drahokam s bílými pásky. Velice krásný.",
   "Драгоценный чёрный камень с белыми полосками.",
   "Ein schwarzer Edelstein mit weißen Streifen. Er ist wunderschön.",
/*MISSING*/    "A black gem with white stripes. It is beautiful.",
    }},
  {0x9c0fa751, { // "%The1 moves his finger downwards."
   "%1 przesun%ął1 swój palec w dół.",
   "%1 parmağını aşağı doğru hareket ettirdi.",
   "%1 pohnu%l1 prstem dolů.",
   "%1 опустил%E1 свои пальцы.",
   "%Der1 %1 bewegt seinen Finger nach unten.",
   "%na1 move seu dedo para baixo.",
    }},
  {0x9c38236e, { // "A Witch with a Flash spell. Very dangerous!\n\nLuckily, she never uses the spell if it would kill her friends. She could destroy an Evil Golem, though."
   "Czarownica z czarem Błysku. Bardzo niebezpieczna!\n\nNa szczęście, nie użyje swojego czaru, jeśli by to zabiło jej przyjaciółki.",
   "Işık büyüsü olan bir cadı. Çok tehlikeli!\n\n Çok şükür, büyüsünü büyüsü arkadaşlarını da öldürecekse hiç kullanmaz. Ama Şeytani Golemi de umursayacak değil tabi.",
   "Čarodějnice s kouzlem Záblesk. Velmi nebezpečná!\n\nNaštěstí toto kouzlo nepoužije,  pokud by to zabilo její přátele. Zlé golemy ale ničit může.",
   "Ведьма с заклинанием вспышки. Очень опасна!\n\nК счастью, она никогда не использует заклинание, ели оно убьёт других ведьм. Но големов она может и уничтожить.",
   "Eine Donnerhexe. Sehr gefährlich!\n\nGlücklicherweise setzt sie den Zauber nicht ein, wenn dies ihre Freunde gefährden würde. Sie kann allerdings böse Golems zerstören.",
/*MISSING*/    "A Witch with a Flash spell. Very dangerous!\n\nLuckily, she never uses the spell if it would kill her friends. She could destroy an Evil Golem, though.",
    }},
  {0x9c5df18c, { // "\n\nOnce you collect 10 Bomberbird Eggs, stepping on a cell with no adjacent mines also reveals the adjacent cells. Collecting even more Eggs will increase the radius. Additionally, collecting 25 Bomberbird Eggs will reveal adjacent cells even in your future games."
   "\n\nGdy zdobędziesz 10 Jaj Bombardiera, ruch na pole bez sąsiadujących min pokazuje także liczby na sąsiednich polach; jeszcze więcej jaj to większy promień. Dodatkowo, zebranie 25 Jaj Bombardiera powoduje pokazywanie sąsiednich pól nawet w kolejnych grach.",
/*MISSING*/    "\n\nOnce you collect 10 Bomberbird Eggs, stepping on a cell with no adjacent mines also reveals the adjacent cells. Collecting even more Eggs will increase the radius. Additionally, collecting 25 Bomberbird Eggs will reveal adjacent cells even in your future games.",
   "\n\nJakmile získáš 10 Vajec Bombarďáků, odhalí vstup na políčko bez sousedních min i všechna sousední políčka. Sebráním ještě většího počtu Vajec, se rozsah odhalování zvyšuje. Pokud sebereš 25 Vajec Bombarďáků, budou se navíc ve všech dalších hrách sousední políčka odhalovat automaticky.",
   "\n\nЕсли вы собрали 10 яиц Бомбардира, наступая на клетку, вы получаете информацию обо всех соседних клетках; собирая больше яиц вы увеличиваете радиус обзора. Также, если собрать 25 яиц Бомбардира, то во всех следующих играх вы будете автоматически получать информацию о соседних клетках.",
   "\n\nSobald du 10 Bombenvogeleier gesammelt hast, werden beim Betreten jeder Zelle, die nicht an eine Miene grenzt, auch alle Nachbarzellen aufgedeckt. Sammeln von noch mehr Eiern vergrößert den Radius.  Das Sammeln von 25 Bombenvogeleiern schaltet sogar das Aufdecken von Nachbarzellen in zukünftigen Spielen frei.",
/*MISSING*/    "\n\nOnce you collect 10 Bomberbird Eggs, stepping on a cell with no adjacent mines also reveals the adjacent cells. Collecting even more Eggs will increase the radius. Additionally, collecting 25 Bomberbird Eggs will reveal adjacent cells even in your future games.",
    }},
  {0x9ca1dbd1, { // "cross display mode"
   "tryb pokazywania osi",
   "yıldız görünümü",
   "mód zobrazování os",
   "режим показа осей",
   "Anzeigemodus (Kreuz)",
   "modo de exibição cruzada",
    }},
  {0x9ce514f1, { // "ASCII"
   "ASCII",
   "ASCII",
   "ASCII",
   "ASCII",
   "ASCII",
   "ASCII",
    }},
  {0x9cee54cf, { // "\"Only the stupid hyperbugs do not understand this.\""
   "\"Tylko głupie hiperinsekty tego nie rozumieją.\"",
/*MISSING*/    "\"Only the stupid hyperbugs do not understand this.\"",
   "\"Jenom ti hloupí hyperhmyzáci tomu nerozumějí.\"",
   "\"Только глупые гипержуки не понимают этого.\"",
/*MISSING*/    "\"Only the stupid hyperbugs do not understand this.\"",
/*MISSING*/    "\"Only the stupid hyperbugs do not understand this.\"",
    }},
  {0x9d168012, { // "%The1 is crushed!"
   "%1 zosta%ł1 przygniecion%ya1!",
   "%1 ezildi!",
   "Brána rozdrtila %a1!",
   "%1 разрушается",
   "%Der1 wurde zerquetscht!",
/*MISSING*/    "%The1 is crushed!",
    }},
  {0x9d2b039f, { // "These huge monsters normally live below the sand, but your movements have disturbed them. They are too big to be slain with your weapons, but you can defeat them by making them unable to move. This also produces some Spice. They move two times slower than you."
   "Te wielkie stworzenia normalnie żyją pod pustynnym piaskiem, ale Twoja obecność na pustyni spowodowała ich wyjście na powierzchnię. Są za duże, byś m%ógł0 je zaatakować swoją bronią, ale zginą, jeśli nie będą mogły się ruszyć -- to produkuje także trochę Przyprawy. Czerwie ruszają się 2 razy wolniej od Ciebie.",
   "Normalde bu canavarlar yer altında yaşarlar ama hareketlerin onları rahatsız etti. Bıçağınla kesilmek için fazla iriler, ama onları hareket edemez hale getirerek yenebilirsin. Bu biraz Baharat da üretir. Senden iki kat yavaş hareket ederler.",
   "Tito obrovští netvoři normálně žijí pod pískem, ale tvé pohyby je vyrušily. Jsou tak velcí, že je nemůžeš zabít svými zbraněmi, ale můžeš je porazit tím, že jim znemožníš pohyb. Tím také vznikne nějaké Koření. Píseční červi se pohybují dvakrát pomaleji než ty.",
   "Эти огромные монстры обычно живут под огромным слоем песка, но Вы их потревожили. Они слишком велики, чтобы их можно было убить Вашим оружием, но они умирают, если не могут двигаться. Также они производят пряности. Они движутся в 2 раза медленнее Вас.",
   "Diese riesigen Monster leben normalerweise unter dem Sand, aber deine Bewegungen haben sie geweckt. Sie sind zu groß, als dass du sie mit deinen Waffen erschlagen könntest. Du kannst sie besiegen, indem du dafür sorgst, dass sie sich nicht mehr bewegen können. Dies produziert auch etwas Spice. Sie bewegen sich halb so schnell wie du.",
   "Esses monstros enormes vivem normalmente abaixo da areia, mas os seus movimentos os perturbaram. Eles são muito grandes para serem mortos pelas suas armas, mas você pode derrotá-los ao deixá-los sem possibilidades de moverem-se. Isso também produz Especiaria. Eles se movem duas vezes mais lento que você.",
    }},
  {0x9d2f19ca, { // "-- use the Android menu instead"
   "zapiszą -- użyj menu Androida",
/*MISSING*/    "-- use the Android menu instead",
   "-- použij menu Android",
   "используйте меню Андроида",
   "-- verwende dafür das Android-Menü",
/*MISSING*/    "-- use the Android menu instead",
    }},
  {0x9d7dc7ad, { // "Conformal square model"
   "Konforemny kwadrat",
/*MISSING*/    "Conformal square model",
   "Konformální čtvercový model",
   "Конформный квадрат",
/*MISSING*/    "Conformal square model",
/*MISSING*/    "Conformal square model",
    }},
  {0x9dd81e7b, { // "\"I have once talked to a Yendorian researcher... he was only interested in infinite trees.\""
   "\"Odwiedził mnie raz badacz Yendoriański... interesowały go tylko nieskończone drzewa.\"",
/*MISSING*/    "\"I have once talked to a Yendorian researcher... he was only interested in infinite trees.\"",
   "\"Jednou se mnou mluvil jeden Yendorský výzkumník... zajímal se jenom o nekonečné stromy.\"",
   "\"Со мной беседовал Йендорский исследователь... его интересовали лишь бесконечные деревья.\"",
/*MISSING*/    "\"I have once talked to a Yendorian researcher... he was only interested in infinite trees.\"",
/*MISSING*/    "\"I have once talked to a Yendorian researcher... he was only interested in infinite trees.\"",
    }},
  {0x9e0f4028, { // "(in the MENU). You can reduce the sight range, this should make the animations smoother."
   "(w MENU). Możesz zmniejszyć zasięg widzenia, dzięki temu animacje będą płynniejsze.",
   "(MENÜDE). Görüş aralığını seçebilirsin. Bu animasyonları daha akıcı hâle getirebilir.",
   "(v MENU). Můžeš snížit viditelnost - to by mělo vést k hladším animacím.",
   "(в MENU). Вы можете уменьшить радиус обзора, что сделает анимацию более гладкой. ",
   "(im Menü) kannst du die Sichtweite reduzieren, dies sollte die Animation flüssiger machen.",
   "(no MENU). Você pode reduzir o limite de visão, isso deve fazer com que as animações fiquem mais suaves.",
    }},
#if REPEATED
  {0x9e0f4029, { // "(in the MENU). You can reduce the sight range, this should make the animations smoother. [repeat]"
   "(w MENU). Możesz zmniejszyć zasięg wzroku, co powinno poprawić płynność animacji.",
/*MISSING*/    "(in the MENU). You can reduce the sight range, this should make the animations smoother. [repeat]",
   "(v MENU). Můžete zredukovat dohled, což by mělo udělat animace hladšími.",
   "(в МЕНЮ). Вы можете уменьшить радиус обзора, чтобы сделать анимацию более гладкой.",
   "(im MENÜ). Du kannst die Sichtweite reduzieren, das sollte die Animationen glatter machen.",
/*MISSING*/    "(in the MENU). You can reduce the sight range, this should make the animations smoother. [repeat]",
    }},
#endif
  {0x9e11d347, { // "Remember that you can right click almost anything for more information."
   "Pamiętaj, że prawie wszystko możesz kliknąć prawym przyciskiem, by dowiedzieć się czegoś na dany temat.",
/*MISSING*/    "Remember that you can right click almost anything for more information.",
   "Nezapomeò, že témìø na všechno mùžeš kliknout pravým tlaèítkem a získat o tom více informací.",
   "Помни, что правая кнопка мыши даст информацию практически обо всём.",
/*MISSING*/    "Remember that you can right click almost anything for more information.",
/*MISSING*/    "Remember that you can right click almost anything for more information.",
    }},
  {0x9e7df098, { // "This is your first victory!"
   "To Twoje pierwsze zwycięstwo!",
   "Bu senin ilk zaferin!",
   "Toto je tvé první vítězství!",
   "Это ваша первая победа!",
   "Dies ist dein erster Sieg!",
   "Essa é sua primeira vitória!",
    }},
  {0x9e87899b, { // "save a Princess"
   "uratuj Księżniczkę",
/*MISSING*/    "save a Princess",
   "zachraň Princeznu",
   "спасти Принцессу",
/*MISSING*/    "save a Princess",
/*MISSING*/    "save a Princess",
    }},
  {0x9e8f56ee, { // "You can use this ladder to climb the Tower."
   "Użyj tej drabiny, by wspiąć się na Wieżę.",
   "Kuleye çıkmak için bu merdiveni kullanabilirsin.",
   "Po tomto žebříku můžeš lézt na Věž.",
   "Используйти лестницу, чтобы забраться на Башню.",
   "Du kannst diese Leiter nutzen um den Turm zu erklimmen.",
/*MISSING*/    "You can use this ladder to climb the Tower.",
    }},
  {0x9ee91778, { // "move down-right"
   "ruch w dół-prawo",
/*MISSING*/    "move down-right",
   "pohyb doprava dolů",
   "ход вниз-вправо",
/*MISSING*/    "move down-right",
/*MISSING*/    "move down-right",
    }},
  {0x9f0283f2, { // "line width"
   "szerokość linii",
/*MISSING*/    "line width",
   "šíøka èar",
   "широта линии",
/*MISSING*/    "line width",
/*MISSING*/    "line width",
    }},
  {0x9f821b2e, { // "Your power is drained by %the1!"
   "Twoja moc jest wyssana przez %a1!",
/*MISSING*/    "Your power is drained by %the1!",
   "%1 vysává tvou sílu!",
   "%1 уменьшил Вашу мощь!",
   "Deine Kraft wird von %dem1 %d1 ausgesaugt!",
/*MISSING*/    "Your power is drained by %the1!",
    }},
  {0x9fcff1d6, { // "This orb lets you instantly move to a safe faraway location. Knowing the nature of this strange world, you doubt that you will ever find the way back...\n\nYour game will be saved if you quit the game while the Orb of Safety is still powered.\n\nTechnical note: as it is virtually impossible to return, this Orb recycles memory used for the world so far (even if you do not use it to save the game). "
   "Ta sfera pozwala Ci natychmiast przenieść się do bezpiecznego miejsca. Znając naturę tego świata, zapewne nigdy nie trafisz z powrotem...\n\nJeśli wyjdziesz z gry podczas gdy Sfera Bezpieczeństwa wciąż ma moc, gra zostanie zapisana.\n\nUwaga techniczna: nawet jeśli nie zapiszesz gry, ta sfera czyści całą pamięć o świecie gry. ",
   "Bu küre seni güvenli bir uzak yere ışınlar. Bu dünyanın garip yapısını düşününce geri dönüp dönemeyeceğin belli değil...\n\nEğer Güvenlik Küresi hâlâ açıkken oyundan çıkarsan oyunun kaydedilecek.\n\nTeknik not: geri dönmek aslında tamamen imkânsız olduğundan, bu Küre sadece şimdiye kadar bu dünya için kullanılan hafızayı kullanır. (onu ounu kaydetmek için kullanmasanız da.) ",
   "Tato sféra ti umožňuje se okamžitě přemístit na bezpečné, vzdálené místo. Vzhledem k povaze tohoto podivného světa pochybuješ, že ještě někdy najdeš cestu zpátky...\n\nPokud před vypršením efektu Sféry bezpečnosti ukončíš hru, uloží se.\n\nTechnická poznámka: vzhledem k tomu, že návrat do jakéhokoli již navštíveného místa je stejně prakticky nemožný, použití Sféry bezpečnosti uvolní veškerou paměť, kterou herní svět až dosud využíval, a to i tehdy, pokud ji nepoužiješ k uložení hry.",
   "Эта сфера перемещает Вас в далёкое безопасное место, откуда Вы не вернётесь назад...\n\nПока сфера активна, при выходе из игры Ваша игра будет сохранена.\n\nТехнически, старый мир просто удаляется. ",
   "Dieser Orb versetzt dich augenblicklich an einen weit entfernten, sicheren Ort. In Anbetracht der Natur dieser Welt bezweifelst du, dass du je einen Weg zurückfindest...\n\nDein Spiel wird gespeichert wenn du es beendest während der Orb der Geborgenheit noch Kraft besitzt.\n\nTechnische Anmerkung: Da es wirklich unmöglich ist zurückzukehren, gibt dieser Orb den Speicher, den die Welt bis dahin belegt hat, frei (selbst wenn du das Spiel nicht speicherst.)",
/*MISSING*/    "This orb lets you instantly move to a safe faraway location. Knowing the nature of this strange world, you doubt that you will ever find the way back...\n\nYour game will be saved if you quit the game while the Orb of Safety is still powered.\n\nTechnical note: as it is virtually impossible to return, this Orb recycles memory used for the world so far (even if you do not use it to save the game). ",
    }},
  {0x9fe084e4, { // "You cannot attack %the1 directly!"
   "Nie możesz atakować %a1 wprost!",
   "%a1 doğrudan öldüremezsiniz!",
   "Na %a1 nemůžeš útočit přímo!",
   "Вы не можете напасть на %a1!",
   "Du kannst %den1 %a1 nicht direkt angreifen!",
   "Você não pode atacar %na1 diretamente!",
    }},
  {0x9ff28df6, { // "manual"
   "ręczny",
   "kılavuz",
   "ruční",
   "ручной",
   "manuell",
   "manual",
    }},
  {0xa0360507, { // "Friendly ghosts are friendly beings who can go through any obstacles. However, unlike most friends, they tend to fly away from you."
   "Przyjazne duchy mogą przechodzić przez wszelkie przeszkody. Jednak, w przeciwieństwie do większości pozostałych przyjaciół, mają zwyczaj odlatywać od Ciebie.",
/*MISSING*/    "Friendly ghosts are friendly beings who can go through any obstacles. However, unlike most friends, they tend to fly away from you.",
   "Přátelští duchové jsou přátelské bytosti, kteří dokážou procházet jakýmikoli překážkami. Na rozdíl od většiny tvých kamarádů ale mají tendenci létat pryč od tebe.",
   "Дружественные духи могут проходить сквозь любые препятствия. В отличие от других союзников, склонны улетать от Вас.",
   "Freundliche Geister sind freundliche Wesen, die sich durch alle Hindernisse bewegen können. Aber anders als deine meisten Freunde neigen sie dazu von dir wegzufliegen.",
/*MISSING*/    "Friendly ghosts are friendly beings who can go through any obstacles. However, unlike most friends, they tend to fly away from you.",
    }},
  {0xa0a30d4c, { // "openGL mode"
   "tryb OpenGL",
/*MISSING*/    "openGL mode",
   "mód OpenGL",
   "режим OpenGL",
/*MISSING*/    "openGL mode",
/*MISSING*/    "openGL mode",
    }},
  {0xa0df3717, { // "expansion"
   "rozszerzanie",
/*MISSING*/    "expansion",
   "rozšíření",
   "расширение",
/*MISSING*/    "expansion",
/*MISSING*/    "expansion",
    }},
  {0xa0ecd1c7, { // "and not in a chaos without norms.\""
   "a nie w chaosie bez reguł.\"",
/*MISSING*/    "and not in a chaos without norms.\"",
   "nikoli v chaosu, kde neplatí žádné normy.\"",
   "а не в хаосе без правил.\"",
/*MISSING*/    "and not in a chaos without norms.\"",
/*MISSING*/    "and not in a chaos without norms.\"",
    }},
  {0xa11852fa, { // "\nYou can never target cells which are adjacent to the player character, or ones out of the sight range."
   "\nNie możesz celować w pola obok postaci, ani poza zasięgiem wzroku.",
/*MISSING*/    "\nYou can never target cells which are adjacent to the player character, or ones out of the sight range.",
   "\nPolíčka, která sousedí s vaší postavou nebo jsou mimo váš dohled, nelze nikdy zacílit.",
   "\nНельзя выбрать в качестве цели соседние с вами клетки или те, которые вы не видите.",
   "\nDu kannst niemals auf Zellen, die der Spielfigur benachbart oder die nicht in Sichtweite sind, zielen.",
/*MISSING*/    "\nYou can never target cells which are adjacent to the player character, or ones out of the sight range.",
    }},
  {0xa16a5b13, { // "p = paint"
   "p = kolory",
   "p = boya",
   "p = obarvit",
   "p = раскраска",
   "p = Farbe",
/*MISSING*/    "p = paint",
    }},
  {0xa16ba0c3, { // "\"I am lost...\""
   "\"Zgubiłem się...\"",
   "Kayboldum...",
   "\"I am lost...\"",
   "\"Я потерялся...\"",
   "\"Ich bin verloren...\"",
/*MISSING*/    "\"I am lost...\"",
    }},
  {0xa1aeb8a7, { // "You are killed by %the1!"
   "Zabi%ł1 Cię %1!",
   "%1 tarafından öldürüldün!",
   "Zabi%l1 tě %1!",
   "Вас убил %1!",
   "%Der1 %1 hat dich getötet!",
/*MISSING*/    "You are killed by %the1!",
    }},
  {0xa1cb3d38, { // ", m - mode: shoot'em up"
   "m - tryb strzelanki",
   "m - SHMUP modu",
   "m - mód střílečky",
   "m - режим стрельбы",
   "m - Modus: Ballern",
/*MISSING*/    ", m - mode: shoot'em up",
    }},
  {0xa1fd06c1, { // "%The1 scares %the2!"
   "%1 wystraszy%ł1 %a2!",
   "%1 %a2 korkuttu!",
   "%1 vyděsi%l1 %a2!",
   "%1 пугает %a2!",
   "%Der1 %1 erschreckt %den2 %a2!",
   "%na1 assusta %na2!",
    }},
  {0xa2298e28, { // "player 7 go"
   "gracz 7 idź",
/*MISSING*/    "player 7 go",
   "hráč 7 pohyb",
   "игрок 7 идёт",
   "Spieler 7 los",
/*MISSING*/    "player 7 go",
    }},
  {0xa25132c0, { // "Welcome to Spherogue!"
   "Witaj w Sferogue!",
/*MISSING*/    "Welcome to Spherogue!",
   "Vítej ve Sférogue!",
   "Добро пожаловать в Spherogue!",
/*MISSING*/    "Welcome to Spherogue!",
/*MISSING*/    "Welcome to Spherogue!",
    }},
  {0xa257a57d, { // "Press 'c' for credits."
   "Naciśnij 'c' by obejrzeć autorów",
   "Yapımcılar için 'c'ye bas.",
   "Stiskni 'c' pro zobrazení autorů",
   "Нажмите 'c', чтобы узнать авторов.",
   "Drücke 'c' für die Credits",
   "Pressione 'c' para créditos",
    }},
  {0xa267f7d9, { // "\"By now, you should have your own formula, you know?\""
   "\"Wiesz co, m%ógł0byś już mieć swój własny wzór!\"",
   "\"Şimdiye kadar kendi formülünü bulmalıydın değil mi?\"",
   "\"Teď už bys mě%l0 mít svůj vlastní vzorec, ne?\"",
   "\"Сейчас у Вас должна быть своя формула!\"",
   "\"Mittlerweile solltest du deine eigene Formel haben weißt du?\"",
/*MISSING*/    "\"By now, you should have your own formula, you know?\"",
    }},
  {0xa2b6bab2, { // "Trees in this forest can be chopped down. Big trees take two turns to chop down."
   "Drzewa w tym lesie można ścinać. Ścięcie dużego drzewa zajmuje dwie kolejki.",
   "Bu ormandaki ağaçlar kesilebilir. Büyük ağaçların kesilmesi iki tur alır.",
   "Stromy v tomto lese je možné kácet. Kácení velkých stromů trvá dvě kola.",
   "Вы можете рубить деревья в лесу. Большие деревья рубятся за два хода.",
   "Bäume in diesem Wald können gefällt werden. Große Bäume zu fällen dauert zwei Züge.",
/*MISSING*/    "Trees in this forest can be chopped down. Big trees take two turns to chop down.",
    }},
  {0xa2dc468c, { // "Aww, poor %1... where is your family?"
   "Auu, biedny %1... gdzie Twoja rodzina?",
/*MISSING*/    "Aww, poor %1... where is your family?",
   "Á, chudáčku Želvičko... kdepak máš rodinu?",
   "Ох, бедный %1... Где его семья?",
   "Oh, du arme %1... wo ist deine Familie?",
/*MISSING*/    "Aww, poor %1... where is your family?",
    }},
  {0xa2ea7202, { // "%The1 squeaks excitedly!"
   "%1 się ekscytuje!",
   "%1 heyecanla ciyaklıyor!",
   "%1 vzrušeně piští!",
   "%1 пищит азартно!",
   "%Die1 %1 piepst aufgeregt!",
/*MISSING*/    "%The1 squeaks excitedly!",
    }},
  {0xa2eb0db6, { // "Variants of %the1 are always available in the Random Pattern Mode."
   "Warianty krainy %1 są zawsze dostępne w trybie losowych wzorków.",
   "%1 çeşitlemeleri Rastgele Doku Modunda her zaman mevcuttur.",
   "Varianty kraje '%1' jsou vždy k dispozici v Módu náhodných vzorů.",
   "Варианты земли %1 всегда доступны в режиме случайного узора.",
   "Varianten der %1 sind im Zufallsmuster-Modus verfügbar.",
/*MISSING*/    "Variants of %the1 are always available in the Random Pattern Mode.",
    }},
  {0xa367f3a9, { // "The hyperbolic pirates have no use for treasure maps. However, they have found out that a compass points to the center of the island. So they just go as far towards the center as they can, and hide their treasure there."
   "Mapy są bezużyteczne dla hiperbolicznych piratów, ale zauważyli oni, że kompas zawsze wskazuje w kierunku środka wyspy. Zatem piraci po prostu idą tak daleko w stronę środka, jak mogą i tam chowają swoje skarby.",
   "Hazine haritalarının hiperbolik korsanlar için bir değeri yok. Ancak, adanın ortasını gösteren bir pusula keşfettiler. Böylece, ortaya doğru olabildiğince ilerler ve hazinelerini orada saklarlar.",
   "Mapy jsou hyperbolickým pirátům k ničemu. Zjistili ale, že kompas vždy ukazuje do středu ostrova. A tak se prostě vydávají tak daleko do středu, jak to jen jde, a tam ukrývají své poklady.",
   "Карты бесполезны для гиперболических пиратов, но они знают, что компас всегда показывает в центр острова. Так что они просто идут как можно дальше к центру острова и прячут там сокровища.",
   "Die hyperbolischen Piraten haben für Schatzkarten keinen Gebrauch. Jedoch haben sie erfahren, dass ein Kompass in die Mitte einer Insel zeigt. Also gehen sie so weit dort hin, wie sie können, und verstecken dort ihre Schätze.",
/*MISSING*/    "The hyperbolic pirates have no use for treasure maps. However, they have found out that a compass points to the center of the island. So they just go as far towards the center as they can, and hide their treasure there.",
    }},
  {0xa3c20464, { // "Hypersian Rug model"
   "Model Hiperskiego Dywanu",
/*MISSING*/    "Hypersian Rug model",
   "Model hyperského koberce",
   "Модель Гиперсидского ковра",
/*MISSING*/    "Hypersian Rug model",
/*MISSING*/    "Hypersian Rug model",
    }},
  {0xa3ced252, { // "movement color"
   "kolor ruchu",
/*MISSING*/    "movement color",
   "barva pohybu",
   "цвет хода",
   "Richtungsmarker-Farbe",
/*MISSING*/    "movement color",
    }},
  {0xa3de7c50, { // "Escher/3D"
   "Escher/3D",
/*MISSING*/    "Escher/3D",
   "Escher/3D",
   "Escher/3D",
/*MISSING*/    "Escher/3D",
/*MISSING*/    "Escher/3D",
    }},
  {0xa3f46380, { // "Activated the Hyperstone Quest!"
   "Aktywowałeś Misję Hiperkamień!",
   "Aşkıntaş Görevini etkinleştirdin!",
   "Aktivoval jsi Hyperkamový úkol!",
   "Активирована миссия Гиперкамней!",
   "Hyperstein-Herausforderung aktiviert!",
   "Missão da Pedra do Ímpeto ativada!",
    }},
  {0xa3f7444f, { // "You cannot target that close!"
   "Nie możesz celować tak blisko!",
   "O kadar yakını hedefleyemezsin!",
   "Nemůžeš zamířit tak blízko!",
   "Nie możesz celować tak blisko!",
   "Zu nah!",
/*MISSING*/    "You cannot target that close!",
    }},
  {0xa499a90d, { // "Your brain is steaming."
   "Twój mózg paruje.",
   "Beynin kaynıyor.",
   "Ze tvého mozku se kouří.",
   "Ваш мозг дымится.",
   "Dein Hirn dampft!",
   "Seu cérebro está cozinhando.",
    }},
  {0xa4a3cb42, { // "These Cultists can push the statues, just like you."
   "Ci kultyści mogą przesuwać pomniki tak samo jak Ty.",
   "Bu Müritler putları tıpkı senin gibi itebilir.",
   "Tito kultisté dokážou strkat sochy, stejně jako ty.",
   "Эти культисты могут двигать статуи,как и Вы.",
   "Diese Kultanhänger können die Statuen schieben, genau wie du.",
/*MISSING*/    "These Cultists can push the statues, just like you.",
    }},
  {0xa51aefb4, { // "You collect %the1. (%2)"
   "Znalaz%łeś0 %a1. (%2)",
/*MISSING*/    "You collect %the1. (%2)",
   "Získa%l0 jsi %a1. (%2)",
   "Вы собрали %a1. (%2)",
   "Du sammelst %den1 %a1. (%2)",
/*MISSING*/    "You collect %the1. (%2)",
    }},
  {0xa52c28a0, { // "This land contains high rock formations. Most of the valley is at level 0, while items are found at level 3. It is impossible to gain two or more levels, or to lose three levels, in a single move, (attacks are possible at any difference, though). Kill Red Trolls and Rock Snakes to make a cell higher."
   "W tej krainie są wysokie formacje skalne. Większość doliny jest na poziomie 0, a przedmioty można znaleźć na poziomie 3. Niemożliwe jest wzniesienie się na dwa lub więcej poziomów lub obniżenie się o trzy poziomy w jednym ruchu (ale atakować można przy każdej różnicy poziomów). Zabij Czerwonego Trolla albo Skalnego Węża, by podnieść poziom pola.",
   "Bu diyar kaya oluşumları barındırır. Vadinin çoğu yükseklik 0'dadır, eşyalar ise yükseklik 3'te bulunabilir. Tek hamlede iki veya daha fazla yükseklik kazanamaz yahut üç veya daha fazla yükseklik kazanamazsın. (ama yükseklikler arası saldırılar mümkün). Kızıl Trolleri ve Kaya Yılanlarını öldürerek bir hücreyi yükseltebilirsin. ",
   "V tomto kraji se nacházejí vysoké skalní útvary. Většina údolí je na úrovni 0, ale předměty se nacházejí pouze na úrovni 3. Jedním pohybem se můžete dostat maximálně o jednu úroveň výš nebo o dvě úrovně níž, ale útočit je možné bez ohledu na výškový rozdíl. Zabíjením Červených trollů a Skalních hadů můžete pole posouvat nahoru.",
   "Эта земля содержт участки высоких скал. Большая часть долины находится на уровне 0, а все предметы лежат на уровне 3. Невозможно подняться на 2 или больше уровня или опуститься на 3 уровня за один ход, хотя атаковать можно всегда. Убейте Красного тролля или Каменную змею, чтобы сделать клетку выше.",
   "Dieses Land enthält hohe Felsformationen. Der Großteil des Tals ist auf Ebene 0, aber die meisten Gegenstände finden sich auf Ebene 3 (III). Es ist unmöglich mehr als eine Ebene aufzusteigen sowie alle 3 in einem Zug zu verlieren. Angriffe sind auf jeder Höhe möglich. Töte Rote Trolle und Felsschlangen, um eine Zelle zu erhöhen.",
/*MISSING*/    "This land contains high rock formations. Most of the valley is at level 0, while items are found at level 3. It is impossible to gain two or more levels, or to lose three levels, in a single move, (attacks are possible at any difference, though). Kill Red Trolls and Rock Snakes to make a cell higher.",
    }},
  {0xa58e0943, { // "Welcome to the Yendor Challenge %1!"
   "Witaj w Misji Yendor %1!",
/*MISSING*/    "Welcome to the Yendor Challenge %1!",
   "Vítej v Yendorské misi %1!",
   "Добро пожаловать в миссию Йендора %1!",
/*MISSING*/    "Welcome to the Yendor Challenge %1!",
/*MISSING*/    "Welcome to the Yendor Challenge %1!",
    }},
  {0xa5b5aaa6, { // "People worshipping Cthulhu. They are very dangerous."
   "Wyznawcy Cthulhu, bardzo niebiezpieczni.",
   "Cthulhu'ya tapan insanlar. Çok tehlikeliler.",
   "Lidé uctívající Cthulhua. Jsou velmi nebezpeční.",
   "Люди, поклоняющиеся Ктулху. Они довольно опасны.",
   "Verehrer Cthulhus. Sie sind sehr gefährlich.",
/*MISSING*/    "People worshipping Cthulhu. They are very dangerous.",
    }},
  {0xa5d1bc77, { // "Branches here could bear your weight easily."
   "Te gałęzie z łatwością udźwigną Twój ciężar.",
/*MISSING*/    "Branches here could bear your weight easily.",
   "Tyhle větve by snadno unesly tvou váhu.",
   "Эти ветви легко выдержат ваш вес.",
   "Die Äste hier tragen dein Gewicht mit Leichtigkeit.",
/*MISSING*/    "Branches here could bear your weight easily.",
    }},
  {0xa605f9d2, { // "\nSpecial conduct failed:\n"
   "\nDodatkowe wyzwanie (przegrane):",
/*MISSING*/    "\nSpecial conduct failed:\n",
   "\nZvláštní výzva (neúspěšná):",
   "\nДополнительное условие (нарушено):",
/*MISSING*/    "\nSpecial conduct failed:\n",
/*MISSING*/    "\nSpecial conduct failed:\n",
    }},
  {0xa61bdb9e, { // "You are playing %the1 in the Pure Tactics mode."
   "Grasz %a1 w trybie taktycznym.",
/*MISSING*/    "You are playing %the1 in the Pure Tactics mode.",
   "Právě hraješ %a1 v čistě taktickém módu.",
   "Вы играете в %a1 в Тактическом режиме.",
/*MISSING*/    "You are playing %the1 in the Pure Tactics mode.",
/*MISSING*/    "You are playing %the1 in the Pure Tactics mode.",
    }},
  {0xa63c8d96, { // "These warriors of the Forest wield exotic weapons called hedgehog blades. These blades protect them from a frontal attack, but they still can be 'stabbed' easily by moving from one place next to them to another."
   "Ci wojownicy z Puszczy walczą egzotyczną bronią, Jeżowym Ostrzem. Ostrza te bronią przed atakiem z frontu, ale za to możesz ich łatwo 'dźgnąć' poprzez przejście z jednego pola sąsiadującego z nimi na inne.",
   "Ormandaki bu savaşçılar, adına kirpi bıçağı denen ilginç bıçaklar taşıyorlar. Bu bıçaklar onları önden gelen saldırılardan korur, ama hâlâ yanlarından geçerek öldürülebilirler.",
   "Tito válečníci z Hvozdu vládnou exotickými zbraněmi zvanými 'ježčí čepele'. Tyto čepele je chrání proti útoku zepředu, ale můžeš je snadno 'probodnout' tím, že se pohneš z jednoho políčka, které s nimi sousedí, na druhé.",
   "Эти лесные воители вооружены экзотическим оружием, называемым ежиным клинком. Клинки защищают их от прямых атак, но они могут быть 'заколоты', достаточно перейти с одной клетки рядом с нм на другую.",
   "Diese Krieger des Waldes besitzen exotische Waffen, die Igelklingen. Diese schützen sie vor frontalen Angriffen - aber sie können dennoch 'erstochen' werden indem man sich direkt neben ihnen von einer Zelle auf eine andere bewegt.",
/*MISSING*/    "These warriors of the Forest wield exotic weapons called hedgehog blades. These blades protect them from a frontal attack, but they still can be 'stabbed' easily by moving from one place next to them to another.",
    }},
  {0xa690f718, { // "%The1 shows you a finger."
   "%1 pokaza%ł1 Ci palec.",
   "%1 sana parmak hareketi yaptı.",
   "%1 ti ukáz%l1 jeden prst.",
   "%1 показал%E1 Вам палец.",
   "%Der1 %1 zeigt dir einen Finger.",
   "%na1 lhe mostra um dedo.",
    }},
  {0xa6b27a7b, { // "You can right click any element to get more information about it.\n\n"
   "Kliknij obiekt prawym przyciskiem myszy, by zobaczyć więcej informacji na jego temat.\n\n",
   "Herhangi bir şeye sağ tıklayarak hakkında daha çok bilgiye ulaşabilirsin.\n\n",
   "Informace o jakémkoli herním prvku můžeš zobrazit tak, že na něj klikneš pravým tlačítkem myši.\n\n",
   "Кликните правой кнопкой мыши, чтобы узнать больше.\n\n",
   "Du kannst auf alles rechtsklicken, um Informationen darüber zu erhalten.\n\n",
/*MISSING*/    "You can right click any element to get more information about it.\n\n",
    }},
  {0xa6ec396d, { // "Normally, the power of most Orbs slowly fades away, even when you are not actively using them. This Orb prevents this.\n\nWhen you have the Orb of Time, Orbs which are not active won't lose their power. Orbs are considered active if they have a continuous power which has actually affected something in the last turn.\n\nOrbs of Shielding remain active after being activated (even if you are no longer attacked), and Orbs of Time have a bigger cap inside their native Caribbean than outside."
   "Zazwyczaj moc Sfer powoli wygasa, nawet gdy ich nie używasz aktywnie. Ta Sfera temu zapobiega.\n\nJeśli masz Sferę Czasu, nieaktywne sfery nie będą traciły mocy. Sfery są uważane za aktywne, jeśli mają ciągły efekt, który rzeczywiście na coś wpłynął w ostatniej turze.\n\nSfery Tarczy pozostają aktywne po aktywacji (nawet jeżeli nikt więcej nie atakuje), a Sfery Czasu mają większy limit na Karaibach niż poza nimi.",
   "Normalde Kürelerin güçleri ,onları aktif olarak kullanmadığında bile, yavaşça azalır. Bu Küre bunu engeller.\n\nZaman Küresine sahip olduğunda aktif olmayan küreler güçlerini kaybetmez. Sürekli gücü olan küreler eğer son turda bir şey etkiledilerse aktif sayılır.\n\nKalkan Küresi aktifleştirdikten sonra da (saldıraya uğramasan bile) aktif kalır.  Ayrıca Zaman Küresinin sınırı onların bulunduğu Karayiplerde daha fazladır.",
   "Síla většiny Sfér se většinou pomalu vytrácí, a to i tehdy, když je aktivně nepoužíváte. Tato sféra tomu brání.\n\nKdyž máš Sféru Času, nebudou neaktivní Sféry ztrácet sílu. 'Aktivní Sféra' je taková Sféra, která má stálý efekt a která v minulém kole něco skutečně ovlivnila.\n\nSféry Štítu zůstávají trvale aktivní, jakmile jednou zablokují útok (i když na tebe právě nikdo neútočí a Sféry Času mohou ve svém domovském Karibiku dosahovat vyšších hodnot síly než mimo něj.",
   "Обычно сила сфер угасает, даже если Вы не используете их. Сфера Времени препятствует этому.\n\nЕсли у Вас есть сфера Времени, неактивные сферы не теряют свою силу. Сфера считается активной, если она имеет постоянный эффект, проявившийся на последнем ходе.\n\nСфера Щита считается активной всегда (даже если Вы никого не атаковали). Сфера Времени имеет на Карибах больший эффект, чем в других местах.",
   "Normalerweise schwindet die Kraft der meisten Orbs, sogar wenn du sie nicht einsetzt. Der Orb der Zeit verhindert dies.\n\nBesitzt du einen, verlieren inaktive Orbs keine Kraft. Orbs zählen als aktiv, wenn sie eine andauernde Kraft besitzen, die im letzten Zug etwas beeinflusst hat.\n\nOrbs der Abschirmung bleiben nach Aktivierung aktiv und Orbs der Zeit halten in der Karibik länger als an anderen Orten.",
/*MISSING*/    "Normally, the power of most Orbs slowly fades away, even when you are not actively using them. This Orb prevents this.\n\nWhen you have the Orb of Time, Orbs which are not active won't lose their power. Orbs are considered active if they have a continuous power which has actually affected something in the last turn.\n\nOrbs of Shielding remain active after being activated (even if you are no longer attacked), and Orbs of Time have a bigger cap inside their native Caribbean than outside.",
    }},
  {0xa7088e66, { // " (won!)"
   " (wygrane!)",
   " (kazandın!)",
   " (vítězství!)",
   " (победа!)",
   " (Gewonnen!)",
/*MISSING*/    " (won!)",
    }},
  {0xa7134b44, { // "CONGRATULATIONS!"
   "GRATULACJE!",
   "TEBRIKLER!",
   "GRATULUJEME!",
   "ПОЗДРАВЛЯЕМ!",
   "GLÜCKWUNSCH!",
   "PARABÉNS!",
    }},
  {0xa7313705, { // "special thanks to the following people for their bug reports, feature requests, porting, and other help:\n\n%1\n\n"
   "Szczególne podziękowania dla poniższych osób za zgłoszone przez nich błędy, pomysły, porty i inną pomoc:\n\n%1\n\n",
   "Hata bildirimleri, özellik istekleri, portlama ve diğer yardımları için şu kişilere teşekkürler: %1  \n\n",
   "zvláštní poděkování následujícím lidem za hlášení chyb, žádosti o nové funkce, porty a další pomoc:\n\n%1\n\n",
   "отдельное спасибо следующим людям за багрепорты, предложения и другую помощь:\n\n%1\n\n",
   "Besonderer Dank gebührt diesen Leuten für ihre Bug Reports, Feature-Anfragen, Ports, und andere Hilfe:\n\n%1\n\n",
   "Agradecimentos especiais às seguintes pessoas pelos seus relatórios de bugs, pedidos de melhorias, porting e outras ajudas:\n\n%1\n\n",
    }},
  {0xa741c82e, { // "Band model"
   "Model wstęgi",
/*MISSING*/    "Band model",
   "Pásový model",
   "Модель ленты",
/*MISSING*/    "Band model",
/*MISSING*/    "Band model",
    }},
  {0xa757d525, { // "first joystick position (movement)"
   "pozycja pierwszego joysticka (ruch)",
/*MISSING*/    "first joystick position (movement)",
   "pozice prvního joysticku (pohyb)",
   "положение первого джойстика (движение)",
/*MISSING*/    "first joystick position (movement)",
/*MISSING*/    "first joystick position (movement)",
    }},
  {0xa768dbfc, { // "A land for people wanting to experiment with cellular automata in the HyperRogue grid. Rules can be given on the command line; the default rules are:\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(-c0 or -c1 can be given if the same rule is to be used for hexagonal and heptagonal cells)."
   "Kraina dla tych, co chcą eksperymentować z automatami komórkowymi w siatce HyperRogue. Reguły można podawać w linii poleceń; domyślne to\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(można podać -c0 lub -c1 by dać te same reguły dla sześciokątów i siedmiokątów).",
/*MISSING*/    "A land for people wanting to experiment with cellular automata in the HyperRogue grid. Rules can be given on the command line; the default rules are:\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(-c0 or -c1 can be given if the same rule is to be used for hexagonal and heptagonal cells).",
   "Kraj pro ty, kdo by chtěli experimentovat s buněčnými automaty na mřížce HyperRogue. Pravidla lze zadávat v příkazové řádce; standardně jsou to\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(také lze zadat -c0 nebo -c1, pokud se má použít stejné pravidlo pro šestiúhelníky a sedmiúhelníky).",
   "Земля для желающих поэкспериментировать с клеточными автоматами на сетке HyperRogue. Правила могут быть заданы в командной строке; правила по умолчанию:\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(можно использовать -c0 или -c1, если правила для шестиугольников и семиугольников совпадают).",
/*MISSING*/    "A land for people wanting to experiment with cellular automata in the HyperRogue grid. Rules can be given on the command line; the default rules are:\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(-c0 or -c1 can be given if the same rule is to be used for hexagonal and heptagonal cells).",
/*MISSING*/    "A land for people wanting to experiment with cellular automata in the HyperRogue grid. Rules can be given on the command line; the default rules are:\n-c07 00100000 -c06 0010000 -c17 00011000 -c16 0001100 -caprob 0.3\n(-c0 or -c1 can be given if the same rule is to be used for hexagonal and heptagonal cells).",
    }},
  {0xa7741462, { // "%1 knives (%2)"
   "noży: %1 (%2)",
   "bıçaklar: %1 (%2)",
   "nožů: %1 (%2)",
   "ножи: %1 (%2)",
   "%1 Messer (%2)",
/*MISSING*/    "%1 knives (%2)",
    }},
  {0xa7b0f3df, { // "Welcome to the Random Pattern mode!"
   "Witaj w trybie losowych wzorów!",
/*MISSING*/    "Welcome to the Random Pattern mode!",
   "Vítej v módu náhodných vzorů!",
   "Добро пожаловать в режим случайных узоров!",
/*MISSING*/    "Welcome to the Random Pattern mode!",
/*MISSING*/    "Welcome to the Random Pattern mode!",
    }},
  {0xa7f306ff, { // "You have improved your '%1' high score on Google!"
   "Poprawi%łeś0 swój wynik w kategorii '%P1' na Google!",
   "Google'daki '%P1'puanını artırdın!",
   "Zlepši%l0 jsi své nejvyšší skóre na Googleu v kategorii '%P1'!",
   "Вы повысили свой результат '%1' в таблицах Google",
   "Du hast deinen '%P1'-Rekord auf Google verbessert!",
/*MISSING*/    "You have improved your '%1' high score on Google!",
    }},
  {0xa7ff7222, { // "Uses to gain: %1"
   "Dostaniesz użyć: %1",
/*MISSING*/    "Uses to gain: %1",
   "Získáš použití: %1",
   "Получишь применений: %1",
/*MISSING*/    "Uses to gain: %1",
/*MISSING*/    "Uses to gain: %1",
    }},
  {0xa82244d9, { // "Straight Lines"
   "Linie proste",
/*MISSING*/    "Straight Lines",
   "Přímky",
   "Прямые линии",
/*MISSING*/    "Straight Lines",
/*MISSING*/    "Straight Lines",
    }},
  {0xa82d5bf4, { // "Failed to save map to %1"
   "Nie udało się zapisać do %1",
   "Harita %1 konumuna kaydedilemedi.",
   "Uložení mapy do %1 se nezdařilo",
   "Не удалось созранить карту в %1",
   "Karte konnte nicht nach %1 gespeichert werden",
/*MISSING*/    "Failed to save map to %1",
    }},
  {0xa833d93c, { // "summon a Monster"
   "przywołaj potwora",
   "Bir canavar çıkar",
   "vyvolej netvora",
   "призвать монстра",
   "rufe ein Monster herbei",
/*MISSING*/    "summon a Monster",
    }},
  {0xa83688cc, { // "Cannot cast illusion on a monster!"
   "Nie możesz rzucić iluzji na potwora!",
   "Bir canavarın üzerinde ilüzyon oluşturamazsın!",
   "Nemůžeš seslat iluzi na netvora!",
   "Nie możesz rzucić ilzuji na potwora!",
   "Du kannst Illusionen nicht auf Monstern erschaffen!",
/*MISSING*/    "Cannot cast illusion on a monster!",
    }},
  {0xa86c6c39, { // "You feel the Orbs of Yendor nearby..."
   "Czujesz w pobliżu Sfery Yendoru...",
/*MISSING*/    "You feel the Orbs of Yendor nearby...",
   "Cítíš blízkost Yendorských Sfér...",
   "Ты чувствуешь Сферы Йендора поблизости...",
/*MISSING*/    "You feel the Orbs of Yendor nearby...",
/*MISSING*/    "You feel the Orbs of Yendor nearby...",
    }},
  {0xa8713d1d, { // "You already have this Grimoire! Seek new tomes in the inner circles."
   "Już masz taką Księgę! Szukaj nowych tomów w wewnętrznych kręgach.",
   "Bu Kara Kitaptan elinde zaten var! Yeni ciltleri iç dairelerde ara.",
   "Tenhle Grimoár už máš! Hledej nové Grimoáry ve vnitřních kruzích.",
   "У Вас уже есть этот Гримуар! Ищите новые тома в более глубоких кругах.",
   "Du hast dieses Grimoire schon! Suche neue in den inneren Kreisen.",
/*MISSING*/    "You already have this Grimoire! Seek new tomes in the inner circles.",
    }},
  {0xa8739e67, { // "This orb is dead..."
   "Ta sfera jest martwa...",
   "Bu küre ölü...",
   "Tato sféra je mrtvá...",
   "Эта сфера мертва...",
   "Dieser Orb ist tot...",
   "Esse orbe está morto...",
    }},
  {0xa87b3042, { // "quit the game"
   "koniec gry",
   "oyundan çık",
   "ukončit hru",
   "выйти из игры",
   "Beenden",
/*MISSING*/    "quit the game",
    }},
  {0xa87d300f, { // "An alternate layout of the Crossroads, without walls."
   "Alternatywny układ Skrzyżowań, bez ścian.",
/*MISSING*/    "An alternate layout of the Crossroads, without walls.",
   "Alternativní verze Křižovatky bez zdí.",
   "Другой вариант перекрёстка, без стен.",
/*MISSING*/    "An alternate layout of the Crossroads, without walls.",
/*MISSING*/    "An alternate layout of the Crossroads, without walls.",
    }},
  {0xa8918f7b, { // "\n\nIn the Orb Strategy Mode, dead orbs are available once you collect 10 Necromancer Totems in the Graveyard."
   "\n\nW trybie strategii sfer martwe sfery są dostępne po zdobyciu 10 Totemów Nekromanty na Cmentarzu.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, dead orbs are available once you collect 10 Necromancer Totems in the Graveyard.",
   "\n\nMrtvé Sféry se ve sférostrategickém módu objevují poté, co získáš 10 Nekromancerových totemù na Høbitovì.",
   "\n\nВ режиме стратегии сфер Мёртвые сферы доступны после сбора десяти Тотемов некроманта на Кладбище.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, dead orbs are available once you collect 10 Necromancer Totems in the Graveyard.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, dead orbs are available once you collect 10 Necromancer Totems in the Graveyard.",
    }},
  {0xa91e5c94, { // "An ancient grave."
   "Stary grób.",
   "Antik bir mezar.",
   "Starý hrob.",
   "Старая могила.",
   "Ein altes Grab.",
/*MISSING*/    "An ancient grave.",
    }},
  {0xa9244759, { // "Horocycles"
   "Horocykle",
/*MISSING*/    "Horocycles",
   "Horocykly",
   "Орициклы",
/*MISSING*/    "Horocycles",
/*MISSING*/    "Horocycles",
    }},
  {0xa93109d6, { // "HyperRogue %1"
   "HyperRogue %1",
   "HyperRogue %1",
   "HyperRogue %1",
   "HyperRogue %1",
   "HyperRogue %1",
/*MISSING*/    "HyperRogue %1",
    }},
  {0xa9555d15, { // "A vine is growing here... but only on a half of the cell. How is that even possible?!Most monsters cannot move from this cell to the cell containing the other half. Vine spirits can move only to the adjacent cells which are also adjacent to the other half."
   "Tu rośnie winorośl... ale tylko na połowie pola. Jak to w ogóle jest możliwe?!Większość stworzeń nie może ruszyć się z tego pola na pole zawierające drugą połowę winorośli. Winne duszki mogą się ruszyć tylko na te sąsiednie pola, które sąsiadują także z drugą połową.",
   "Burada şarap oluşuyor... ama sadece hücrenin yarısında. Bu nasıl mümkün olabilir?! Çoğu canavar bu hücreden diğer yarıyı içeren hücreye hareket edemez. Şarap özleri sadece diğer yarıya da bitişik olan hücrelere gidebilir.",
   "Tady roste réva... ale pouze na půlce políčka. Jak je to vůbec možné?!Většina netvorů nemůže přecházet z tohoto políčka na políčko s druhou polovinou révy. Révoví duchové se mohou pohybovat pouze na ta sousední políčka, která sousedí i s druhou polovinou.",
   "Здесь растёт лоза... но только на половине клетки. Как это возможно?! Большинство монстров не могут перейти с этой клетки на соседнюю с ней по стороне, содержащей лозу; винный дух, напротив, на может перейти с этой клетки на лозу, соседнюю с ней по свободной половине.",
   "Hier wächst eine Rebe... aber nur auf der halben Zelle. Wie kann das sein?!Die meisten Monster können sich nicht von dieser Zelle aus auf die bewegen, die die andere Hälfte enthält. Rebenseelen können sich nur auf angrenzende Zellen bewegen, die auch der anderen Hälfte anliegen. ",
/*MISSING*/    "A vine is growing here... but only on a half of the cell. How is that even possible?!Most monsters cannot move from this cell to the cell containing the other half. Vine spirits can move only to the adjacent cells which are also adjacent to the other half.",
    }},
  {0xa9a41ee3, { // "Game loaded."
   "Gra odczytana.",
   "Oyun yüklendi.",
   "Hra nahrána.",
   "Игра сохранена.",
   "Spiel geladen.",
   "Carregado.",
    }},
  {0xa9c9c4f0, { // "This fire never burns out."
   "Ten ogień pali się wiecznie.",
   "Bu ateş hiç sönmüyor.",
   "Tento oheň nikdy neuhasne.",
   "Этот огонь никогда не гаснет.",
   "Dieses Feuer erlischt niemals.",
/*MISSING*/    "This fire never burns out.",
    }},
  {0xa9feaa6b, { // "You fall into a wormhole!"
   "Wpad%łeś0 w tunel przestrzenny!",
   "Bir solucan deliğine düştün!",
   "Spad%l0 jsi do červí díry!",
   "Вы упали в червоточину!",
   "Du fällst in ein Wurmloch!",
   "Você cai num túnel de minhoca!",
    }},
  {0xaa0edc30, { // "animals killed: %1"
   "zabitych zwierząt: %1",
/*MISSING*/    "animals killed: %1",
   "zabitých zvířat: %1",
   "животных убито: %1",
   "getötete Tiere: %1",
/*MISSING*/    "animals killed: %1",
    }},
  {0xaa94065f, { // "SHARE"
   "SHARE",
   "PAYLAŞ",
   "SDÍLEJ",
   "SHARE",
   "TEILE",
/*MISSING*/    "SHARE",
    }},
  {0xab15ace4, { // "female"
   "kobieta",
   "kadın",
   "žena",
   "женский",
   "weiblich",
   "feminino",
    }},
  {0xab621991, { // "zebra triangles"
   "zebra: trójkąty",
/*MISSING*/    "zebra triangles",
   "zebra: trojúhelníky",
   "зебра: треугольники",
/*MISSING*/    "zebra triangles",
/*MISSING*/    "zebra triangles",
    }},
  {0xab720893, { // "You would be killed by %the1!"
   "Zabi%ł1by Cię %1!",
   "%1 tarafından öldürülürsün!",
   "To by tě zabi%l1 %1!",
   "Вас убьёт %1!",
   "%Der1 %1 würde dich töten!",
   "Você seria morto %abl1!",
    }},
  {0xab7a0e7d, { // "With this Elixir, your life should be long and prosperous..."
   "Ten Eliksir pozwoli Ci na długie i szczęśliwe życie...",
   "Bu İksirle hayatın uzun ve refah olacak...",
   "S tímto elixírem bys měl žít dlouho a blaze...",
   "С этим эликсиром Ваша жизнь будет долгой и процветающей...",
   "Mit diesem Elixir sollte dein Leben lang und wohlhabend sein...",
   "Com esse Elixir, sua vida será longa e próspera...",
    }},
  {0xab81bd79, { // "F1 - help"
   "F1 - pomoc",
/*MISSING*/    "F1 - help",
   "F1 - nápověda",
   "F1 - помощь",
/*MISSING*/    "F1 - help",
/*MISSING*/    "F1 - help",
    }},
  {0xaba3926a, { // "This orb is activated if you are unable to escape (radius 4) without making illegal moves or going through cells which are currently adjacent to enemy monsters. Most game over situations are covered by this, but generally, this orb is oversensitive...\n\nWhen activated, it creates a Flash effect of radius 5."
   "Ta sfera aktywuje się, gdy nie możesz uciec (promień 4) bez wykonywania nielegalnych ruchów lub przechodzenia przez komórki nie sąsiadujące z przeciwnikami. Pokrywa większość sytuacji 'koniec gry', ale zazwyczaj ta sfera jest nadwrażliwa...\n\nAktywacja powoduje efekt czaru Błysk w promieniu 5.",
/*MISSING*/    "This orb is activated if you are unable to escape (radius 4) without making illegal moves or going through cells which are currently adjacent to enemy monsters. Most game over situations are covered by this, but generally, this orb is oversensitive...\n\nWhen activated, it creates a Flash effect of radius 5.",
   "Tato Sféra se aktivuje, pokud nemůžeš utéct (do vzdálenosti 4 políček), aniž bys provedl ilegální pohyby nebo prošel políčky, která momentálně sousedí s nepřátelskými netvory. To pokrývá většinu situací 'konec hry', ale obvykle tato Sféra reaguje poněkud přehnaně...\n\nPři aktivaci vytvoří efekt Záblesku s poloměrem 5 políček.",
   "Эта сфера активируется, если Вам некуда идти (в радиусе 4), не делая запрещённых ходов и не ходя на клетки, рядом с которыми стоит монстр. Обычно игра кончается именно в такой ситуации, хотя обратное верно далеко не всегда...\n\nПри активации создаёт Вспышку радиуса 5.",
   "Dieser Orb wird aktiviert, wenn du nicht mehr entkommen kannst (Radius 4), ohne unerlaubte Züge zu machen oder Zellen zu überqueren, die sich neben einem feindlichen Monster befinden. Die meisten game-over-Situationen erfüllen diese Bedingung, aber im Allgemeinen ist dieser Orb überempfindlich...\n\nWenn er aktiviert wird, erzeugt er einen Lichtblitz vom Radius 5.",
/*MISSING*/    "This orb is activated if you are unable to escape (radius 4) without making illegal moves or going through cells which are currently adjacent to enemy monsters. Most game over situations are covered by this, but generally, this orb is oversensitive...\n\nWhen activated, it creates a Flash effect of radius 5.",
    }},
  {0xaba70922, { // "inventory/kill mode"
   "tryb rzeczy/zabić",
/*MISSING*/    "inventory/kill mode",
   "mód předmětů/zabitých netvorů",
   "режим предметов/убийств",
/*MISSING*/    "inventory/kill mode",
/*MISSING*/    "inventory/kill mode",
    }},
  {0xabc14003, { // "parameters set correctly"
   "parametry prawidłowe",
/*MISSING*/    "parameters set correctly",
   "správně nastavené parametry",
   "параметры корректны",
/*MISSING*/    "parameters set correctly",
/*MISSING*/    "parameters set correctly",
    }},
  {0xac03fe5f, { // "Extras:"
   "Dodatkowe:",
/*MISSING*/    "Extras:",
   "Extra:",
   "Дополнительно:",
/*MISSING*/    "Extras:",
/*MISSING*/    "Extras:",
    }},
  {0xac1227ad, { // "pan down"
   "przewiń w dół",
   "aşağı kaydır",
   "scrollování dolů",
   "крутить вниз",
   "Runterschwenken",
/*MISSING*/    "pan down",
    }},
  {0xac1e3d03, { // "Your Aether power has expired! RUN!"
   "Twoja moc Eteru się wyczerpała! UCIEKAJ!",
   "Eter gücünün süresi bitti! KAÇ!",
   "Tvá síla Éteru vyprchala. UTÍKEJ!",
   "Ваша эфирная сила иссякает! БЕГИТЕ!",
   "Deine Ätherkraft ist zu Ende! Lauf!!!",
/*MISSING*/    "Your Aether power has expired! RUN!",
    }},
  {0xac35a37d, { // "You activate %the1."
   "Aktywowa%łeś0 %a1.",
   "%a1 aktifleştirdin.",
   "Aktivova%l0 jsi %a1.",
   "Вы активировали %a1.",
   "Du aktivierst %den1 %a1.",
   "Você ativa %na1.",
    }},
  {0xac47aca2, { // "\"Infinite trees are boring. I prefer other graphs.\""
   "\"Nieskończone drzewa są nudne. Wolę inne grafy.\"",
/*MISSING*/    "\"Infinite trees are boring. I prefer other graphs.\"",
   "\"Nekonečné stromy jsou nudné. Mám raději jiné grafy.\"",
   "\"Бесконечные деревья скучные. Предпочитаю другие графы.\"",
/*MISSING*/    "\"Infinite trees are boring. I prefer other graphs.\"",
/*MISSING*/    "\"Infinite trees are boring. I prefer other graphs.\"",
    }},
  {0xac4f826d, { // "kills"
   "zab",
   "leşler",
   "zab",
   "убийств",
   "kills",
   "mortes",
    }},
  {0xad05121c, { // "Always available.\n"
   "Kraina zawsze dostępna.\n",
   "Her zaman mevcut.\n",
   "Tento kraj je vždy k dispozici.\n",
   "Земля всегда доступна.\n",
   "Immer verfügbar.\n",
/*MISSING*/    "Always available.\n",
    }},
  {0xad1a09a5, { // "openGL mode enabled"
   "włączono tryb OpenGL",
   "OpenGL modu devrede",
   "mód openGL zapnut",
   "режим OpenGL включен",
   "OpenGL aktiviert",
   "Modo OpenGL habilitado",
    }},
  {0xad56c229, { // "time elapsed"
   "czas",
   "geçen zaman",
   "čas",
   "время",
   "Zeit",
   "tempo passado",
    }},
  {0xadc5accd, { // "%The1 hits %the2."
   "%1 uderzy%ł1 %a2.",
/*MISSING*/    "%The1 hits %the2.",
   "%1 udeři%l1 %a2.",
   "%1 ударил%E1 %a2.",
   "%Der1 %1 schlägt %den2 %a2.",
/*MISSING*/    "%The1 hits %the2.",
    }},
  {0xaded0207, { // "Don't be fooled by this red-haired girl, or you will be stabbed if you come too close!"
   "Nie daj się zwieść tej rudowłosej, bo jak się za bardzo zbliżysz, możesz zostać ugodzon%ya1 nożem!",
/*MISSING*/    "Don't be fooled by this red-haired girl, or you will be stabbed if you come too close!",
   "Nenech se oklamat touto zrzavou dívkou -- pokud se k ní příliš přiblížíš, bodne tě!",
   "Не будьте обмануты этой рыжеволосой девушкой, у неё есть острый нож!",
   "Lass dich von diesem rothaarigen Mädchen nicht täuschen - oder du wirst erstochen, wenn du zu nah kommst!",
/*MISSING*/    "Don't be fooled by this red-haired girl, or you will be stabbed if you come too close!",
    }},
  {0xae23f5e8, { // "configure input"
   "konfiguracja sterowania",
/*MISSING*/    "configure input",
   "konfigurace vstupu",
   "настроить ввод",
/*MISSING*/    "configure input",
/*MISSING*/    "configure input",
    }},
  {0xae245bf3, { // "TRANSLATIONWARNING"
/*MISSING*/    "TRANSLATIONWARNING",
   "ÇEVİRİ UYARISI: Türkçe çevirisi güncel değil -- 8.1'den sonra eklenen özellikleri",
/*MISSING*/    "TRANSLATIONWARNING",
/*MISSING*/    "TRANSLATIONWARNING",
/*MISSING*/    "TRANSLATIONWARNING",
/*MISSING*/    "TRANSLATIONWARNING",
    }},
  {0xae2cf6e1, { // "The temple of Cthulhu consists of many concentric circles of columns. You will surely encounter many Cultists there, who believe that a pilgrimage to the inner circles will bring them closer to Cthulhu himself, and Grimoires which surely contain many interesting secrets.\n\nThe circles in the temple of Cthulhu are actually horocycles. They are infinite, and there is an infinite number of them."
   "Świątynia Cthulhu składa się z wielu koncentrycznych kręgów kolumn. Spotkasz tu wielu Kultystów, dla których pielgrzymka do wewnętrznych kręgów to szansa na bycie bliżej Cthulhu, a także Księgi zawierające wiele interesujących sekretów.\n\nKręgi w Świątyni Cthulhu są właściwie horocyklami. Są nieskończone i jest ich nieskończenie wiele.",
   "Cthulhu tapınağı çok sayıda eş merkezli dairesel sütunlardan oluşur. Burada iç dairelere yapılan bir haccın kendilerini Cthulhu'ya daha çok yaklaştırdığına inanan bir çok Müritle ve çok ilginç sırlar içeren Kara Kitaplarla karşılaşacaksın. \n\nCthulhu Tapınağının merkezindeki daireler aslında horoçemberler. Onlar sonsuz ve onlardan sonsuz tane var.",
   "Cthulhuův chrám sestává z mnoha soustředných kruhů sloupů. Jistě tu narazíš na mnoho Kultistů, kteří věří, že je vykonání pouti do vnitřních kruhů přivede blíž k samotnému Cthulhuovi a ke Grimoárům, které jistě obsahují mnoho zajímavých tajemství.\n\nKruhy v Cthulhuově chtámu jsou ve skutečnosti horocykly. Jsou nekonečně velké a je jich nekonečně mnoho.",
   "Храм Ктулху содержит много концентрических кругов колонн. Здесь Вы встретите культистов, верящих, что путешествие во внутренние круги храма приблизит их ко Ктулху, и найдёте Гримуары, содержащие весьма интересные секреты.\n\nКруги в храме на самом деле орициклы. Они бесконечны, и их число бесконечно.",
   "Der Tempel des Cthulhu. Er besteht aus vielen konzentrischen Kreisen von Säulen. Du wirst hier sicherlich vielen Anhängern Cthulhus begegnen. Sie glauben, dass eine Pilgerreise in die inneren Kreise sie näher zu Cthulhu bringt und man mächtige Grimoires finden kann, die viele Geheimnisse in sich bergen.\n\nDie Kreise hier sind eigentlich Horozykeln. Sie sind unendlich groß, und es gibt unendlich viele davon.",
/*MISSING*/    "The temple of Cthulhu consists of many concentric circles of columns. You will surely encounter many Cultists there, who believe that a pilgrimage to the inner circles will bring them closer to Cthulhu himself, and Grimoires which surely contain many interesting secrets.\n\nThe circles in the temple of Cthulhu are actually horocycles. They are infinite, and there is an infinite number of them.",
    }},
  {0xae2d6177, { // "\"That felt great. Thanks!\""
   "\"Wspaniałe uczucie. Dzięki!\"",
   "\"Çok iyi hissettim! Sağol!\"",
   "\"To byl skvělý pocit. Děkuji!\"",
   "\"Это было прекрасно. Спасибо!\"",
   "\"Danke, das hat sich gut angefühlt.\"",
/*MISSING*/    "\"That felt great. Thanks!\"",
    }},
  {0xae6f4a8d, { // "copying"
   "kopiuję",
   "kopyalıyor",
   "kopíruje",
   "копирование",
   "kopieren",
/*MISSING*/    "copying",
    }},
  {0xaea25fcd, { // "Accessible only from %the1.\n"
   "Kraina dostępna tylko poprzez %a1.\n",
   "Sadece %a1 erişilebilir.\n",
   "Tento kraj je dostupný pouze skrz %a1.\n",
   "Доступно только из %a1.\n",
   "Nur von %der1 %a1 aus erreichbar.\n",
/*MISSING*/    "Accessible only from %the1.\n",
    }},
  {0xaeb1129b, { // "player 2 X"
   "gracz 2 X",
   "oyuncu 2, yatay",
   "hráč 2 X",
   "игрок 2 X",
   "Spieler 2 X",
/*MISSING*/    "player 2 X",
    }},
  {0xaeb1129c, { // "player 2 Y"
   "gracz 2 Y",
   "oyuncu 2, dikey",
   "hráč 2 Y",
   "игрок 2 Y",
   "Spieler 2 Y",
/*MISSING*/    "player 2 Y",
    }},
  {0xaef79f07, { // "firewall lines"
   "linie ścian ognia",
/*MISSING*/    "firewall lines",
   "čáry ohnivých stěn",
   "стены огня",
/*MISSING*/    "firewall lines",
/*MISSING*/    "firewall lines",
    }},
  {0xaf02ccc6, { // "Mid detail range"
   "Zasięg średniego poziomu szczegółów",
/*MISSING*/    "Mid detail range",
   "Rozsah středních detailů",
   "Зона средней детализации",
/*MISSING*/    "Mid detail range",
/*MISSING*/    "Mid detail range",
    }},
  {0xaf542723, { // "backward"
   "do tyłu",
   "geri",
   "dozadu",
   "назад",
   "Zurück",
/*MISSING*/    "backward",
    }},
  {0xaf5bf681, { // "You cannot defeat the Shadow!"
   "Nie możesz pokonać Cienia!",
   "gölgeyi henüz yenemezsiniz!",
   "Stín nemůžeš porazit!",
   "Вы не можете убить свою тень!",
   "Du kannst den Schatten nicht besiegen!",
   "Você não pode derrotar a Sombra!",
    }},
  {0xaf704be6, { // "Vampire Bats drain your magical powers!"
   "Wampiry wysysają Twoje magiczne moce!",
/*MISSING*/    "Vampire Bats drain your magical powers!",
   "Vampýři ti vysávají magické schopnosti!",
   "Вампиры высасывают из Вас силу!",
/*MISSING*/    "Vampire Bats drain your magical powers!",
/*MISSING*/    "Vampire Bats drain your magical powers!",
    }},
  {0xaf9479a6, { // "The treasure gives your magical powers!"
   "Skarby dają Ci moce magiczne!",
/*MISSING*/    "The treasure gives your magical powers!",
   "Poklady ti dávají magickou sílu!",
   "Сокровище дало тебе магическую силу!",
/*MISSING*/    "The treasure gives your magical powers!",
/*MISSING*/    "The treasure gives your magical powers!",
    }},
  {0xaf9f8e73, { // "An alternate layout of the Crossroads. Great Walls cross here at right angles."
   "Alternatywny układ Skrzyżowania. Wielkie Ściany przecinają się tu pod kątami prostymi.",
   "Arayollar için bir alternatif düzen. Büyük Duvarlar burada dik açı yaparak kesişir.",
   "Alternativní uspořádání Křižovatky. Velké zdi se tu protínají v pravých úhlech.",
   "Альтернативный вариант Перекрёстка. Великие стены пересекаются под прямым углом.",
   "Ein alternatives Layout der Kreuzungen. Große Mauern kreuzen sich hier in rechten Winkeln.",
/*MISSING*/    "An alternate layout of the Crossroads. Great Walls cross here at right angles.",
    }},
  {0xafa28a3b, { // "Congratulations! Your score is %1."
   "Gratulacje! Twój wynik to %1.",
   "Tebrikler! Puanın: %1.",
   "Gratulujeme! Tvé skóre je %1.",
   "Поздравляем! Ваш результат: %1.",
   "Glückwunsch! Deine Punktzahl beträgt %1.",
/*MISSING*/    "Congratulations! Your score is %1.",
    }},
  {0xafbe27dc, { // "A precious green gem from the Emerald Mines."
   "Cenny, zielony klejnot z Kopalni Szmaragdów.",
   "Zümrüt Madenlerinden eğerli yeşil bir mücevher.",
   "Cenný zelený drahokam ze Smaragdového dolu.",
   "Ценный зелёный камень из изумрудных шахт.",
   "Ein wertvoller grüner Edelstein aus den Smaragdminen.",
/*MISSING*/    "A precious green gem from the Emerald Mines.",
    }},
  {0xb071059d, { // "The mirror shatters!"
   "Lustro pęka!",
   "Ayna kırıldı!",
   "Zrcadlo se roztříštilo!",
   "Зеркало разбилось!",
   "Der Spiegel zerbricht",
   "O espelho se despedaça!",
    }},
  {0xb1009491, { // "player"
   "gracz",
/*MISSING*/    "player",
   "hráč",
   "игрок",
   "Spieler",
/*MISSING*/    "player",
    }},
  {0xb185b26f, { // "Skeletons work similar to Palace Guards, but they won't die no matter how many times you hit them. Well, you have to be more creative...\n\nSkeletons attacked outside of their native land, Palace, are stunned for a longer time."
   "Kościotrupy działają podobnie do Strażników Pałacu, ale nie zabijesz ich, ilekolwiek byś ich atakowa%ł0. Wymyśl coś bardziej twórczego...\n\nKościotrup zaatakowany poza Pałacem będzie ogłuszony przez dłuższy czas.",
   "İskeletler Saray Muhafızlarına benzerdir, ama kaç sefer vurulurlarsa vurulsunlar ölmezler. Onları öldürmek için daha yaratıcı olmanız gerekiyor... \n\n Sarayın dışında darbe alan iskeletler daha uzun süre için sersemlerler.",
   "Kostlivci fungují podobně jako Strážci Paláce, ale ať na ně zaútočíš, kolikrát chceš, nikdy je nezabiješ. No, musíš být vynalézavější...\n\nMimo Palác, jejich rodný kraj, budou Kostlivci útokem omráčení na delší dobu.",
   "Скелет реагирует на удары как Дворцовый стражник, но сколько раз бы Вы его не ударили, он не умрёт. Хитрее надо быть...\n\nЕсли скелет атакован вне его родной земли, Дворца, он застывает на более долгое время.",
   "Skelette ähneln den Palastwachen. Sie sterben zwar nicht egal wie oft du sie triffst, aber wenn du sie außerhalb des Palastes angreifst sind sie für lange Zeit betäubt \n\n",
/*MISSING*/    "Skeletons work similar to Palace Guards, but they won't die no matter how many times you hit them. Well, you have to be more creative...\n\nSkeletons attacked outside of their native land, Palace, are stunned for a longer time.",
    }},
  {0xb1bbbfea, { // "You feel that a magical weapon is waiting for you..."
   "Czujesz, że magiczna broń na Ciebie czeka...",
/*MISSING*/    "You feel that a magical weapon is waiting for you...",
   "Cítíš, že na tebe čeká magická zbraň...",
   "Ты чувствуешь, что магическое оружие ждёт тебя...",
/*MISSING*/    "You feel that a magical weapon is waiting for you...",
/*MISSING*/    "You feel that a magical weapon is waiting for you...",
    }},
  {0xb1c9f938, { // "Six mines next to you!"
   "Sześć min koło Ciebie!",
   "Yakınında altı mayın var!",
   "Ve tvé blízkosti se nachází šest min!",
   "Шесть мин рядом с Вами!",
   "6 Minen um dich!",
/*MISSING*/    "Six mines next to you!",
    }},
  {0xb1dbbe26, { // "rotation"
   "obrót",
/*MISSING*/    "rotation",
   "rotace",
   "поворот",
   "Rotation",
/*MISSING*/    "rotation",
    }},
  {0xb21f6237, { // "You feel more experienced in demon fighting!"
   "Jesteś bardziej doświadczony w walce z demonami!",
   "Şeytanlarla savaşmada daha deneyimli hissediyorsun!",
   "Cítíš se zkušenější v boji proti pekelným démonům!",
   "Теперь Вы более опытны в борьбе с демонами!",
   "Du fühlst dich erfahrener im Kampf gegen Dämonen!",
   "Você se sente mais experiente matando demônios!",
    }},
  {0xb226a49b, { // "one player"
   "jeden gracz",
   "tek kişilik",
   "jeden hráč",
   "один игрок",
   "Ein Spieler",
/*MISSING*/    "one player",
    }},
  {0xb22b7504, { // "Kill a Vizier in the Palace to access Emerald Mine"
   "Zabij Wezyra w Pałacu, by iść do Kopalni Szmaragdów",
   "Zümrüt Madenine ulaşmak için Saray'da bir Vezir öldür.",
   "Zabij Vezíra v Paláci a dostaneš se do Smaragdového dolu",
   "Убей Визиря во дворце, чтобы открыть Изумрудную шахту.",
   "Töte einen Wisier im Palast um Zugang zu den Smaragdminen zu erhalten",
/*MISSING*/    "Kill a Vizier in the Palace to access Emerald Mine",
    }},
  {0xb274d5bf, { // "A small reward for braving the Hell."
   "Drobna nagroda za zmierzenie się z Piekłem.",
/*MISSING*/    "A small reward for braving the Hell.",
   "Malá odměna za to, že ses postavil Peklu.",
   "Небольшая награда для прошедших сквозь Ад.",
/*MISSING*/    "A small reward for braving the Hell.",
/*MISSING*/    "A small reward for braving the Hell.",
    }},
  {0xb2c04442, { // "A land filled with demons and molten sulphur. Abandon all hope ye who enter here!"
   "Kraina pełna demonów i stopionej siarki. Ty, który wchodzisz, żegnaj się z nadzieją!",
   "Şeytanlarla ve erimiş sülfürle dolu bir diyar. Buraya giriyorsan tüm ümitlerini terket!",
   "Kraj plný démonů a roztavené síry. Zanech vší naděje, ty, kdož vstupuješ!",
   "Земля, полная демонов и расплавленной серы. Оставь надежду всяк сюда входящий!",
   "Ein Land voller Dämonen und geschmolzenem Schwefel. Gib alle Hoffnung auf, wenn du eintrittst!",
/*MISSING*/    "A land filled with demons and molten sulphur. Abandon all hope ye who enter here!",
    }},
  {0xb2ccbf52, { // "You jump!"
   "Skaczesz!",
   "Zıpladın!",
   "Skočil jsi!",
   "Вы прыгнули!",
   "Du springst!",
/*MISSING*/    "You jump!",
    }},
  {0xb2e991dc, { // "towards orthographic"
   "w stronę rzutu ortogonalnego",
/*MISSING*/    "towards orthographic",
   "směr k ortografické projekci",
   "в сторону ортогональной",
/*MISSING*/    "towards orthographic",
/*MISSING*/    "towards orthographic",
    }},
  {0xb2f43970, { // "Actually, their powers appear god-like...\n\n"
   "Jego moce wyglądają właściwie na boskie...\n\n",
/*MISSING*/    "Actually, their powers appear god-like...\n\n",
   "Vlastně se zdá, že má až božskou moc...\n\n",
   "Его силы кажутся богоподобными...\n\n",
/*MISSING*/    "Actually, their powers appear god-like...\n\n",
/*MISSING*/    "Actually, their powers appear god-like...\n\n",
    }},
  {0xb309a7f9, { // "Your mundane weapon cannot hurt %the1!"
   "Twoja zwykła broń nie działa na %a1!",
/*MISSING*/    "Your mundane weapon cannot hurt %the1!",
   "%1 se nedá zranit tvou obyčejnou zbraní!",
   "Обычное оружие не берёт %a1!",
   "Deine irdische Waffe kann %den1 %a1 nicht verletzen!",
/*MISSING*/    "Your mundane weapon cannot hurt %the1!",
    }},
  {0xb3145edc, { // "%The1 falls!"
   "%1 spad%ł1!",
   "%1 düştü!",
   "%1 spad%l1!",
   "%1 упал%E1!",
   "%Der1 %1 fällt!",
/*MISSING*/    "%The1 falls!",
    }},
  {0xb33b4797, { // "home"
   "centruj na graczy",
   "ortaya geri dön",
   "vycentrování",
   "центр на игроках",
   "Home",
/*MISSING*/    "home",
    }},
  {0xb33df516, { // "render texture without OpenGL"
   "renderuj teksturę bez OpenGL",
/*MISSING*/    "render texture without OpenGL",
   "renderuj textury bez OpenGL",
   "не использовать OpenGL",
   "Rendere die Textur ohne OpenGL",
/*MISSING*/    "render texture without OpenGL",
    }},
  {0xb36abaf8, { // "The world can be mapped conformally to a square too."
   "Świat można też konforemnie odwzorować na kwadrat.",
/*MISSING*/    "The world can be mapped conformally to a square too.",
   "Svět je také možné konformálně namapovat na čtverec.",
   "Также можно конформно отобразить мир на квадрат.",
/*MISSING*/    "The world can be mapped conformally to a square too.",
/*MISSING*/    "The world can be mapped conformally to a square too.",
    }},
  {0xb3826405, { // "%The1 throws fire at you!"
   "%1 rzuca w Ciebie ogniem!",
   "%1 sana ateş attı!",
   "%1 po tobě vystřelil oheň!",
   "%1 бросил в Вас огнём!",
   "%Der1 %1 wirft Feuer nach dir!",
   "%na1 joga fogo em você!",
    }},
  {0xb3c80e9a, { // "%The1 is immune to mental blasts!"
   "%1 jest odporn%ya1 na atak psychiczny!",
   "%1 zihin patlamasından etkilenmez!",
   "%1 je imunní proti mentálnímu útoku!",
   "%1 защищён%E1 от психических атак!",
   "%Der1 %1 ist immun gegen mentale Stöße!",
/*MISSING*/    "%The1 is immune to mental blasts!",
    }},
  {0xb3ce009a, { // "second joystick: panning speed"
   "dżojstik 2: prędkość przewijania",
   "ikinci oyun kolu: gezdirme hızı",
   "joystick 2: rychlost posouvání",
   "джойстик 2: скорость панорамы",
   "Joystick 2: Scrollgeschwindigkeit",
/*MISSING*/    "second joystick: panning speed",
    }},
  {0xb3f93924, { // "These warped humanoids are skilled warriors and sailors, and they feel at home at the Warped Coast. Their battle experience has taught them that enemies who wait without moving or attacking anything are the most deadly. If they see such an enemy, they become extremely suspicious, and they also wait."
   "Te człekokształtne stworzenia są zręczne na morzu i w walce, a na Zakrzywionym Brzegu czują się jak w domu. Ich doświadczenie bitewne nauczyło je, że ci przeciwnicy, którzy czekają bez ruchu i bez ataku, są najbardziej niebezpieczni. Kiedy widzą takiego przeciwnika, stają się bardzo podejrzliwe i również czekają.",
/*MISSING*/    "These warped humanoids are skilled warriors and sailors, and they feel at home at the Warped Coast. Their battle experience has taught them that enemies who wait without moving or attacking anything are the most deadly. If they see such an enemy, they become extremely suspicious, and they also wait.",
   "Tito pokřivení humanoidé jsou dobrými válečníky i námořníky a na Pokřiveném pobřeží jsou jako doma. Jejich bojové zkušenosti je naučily, že nepřátelé, kteří čekají -- nehýbou se a neútočí -- jsou ti nejvražednější. Když takového nepřítele uvidí, začnou být velmi podezíraví a budou čekat také.",
   "Эти искривлённые гуманоиды -- опытные воины и мореходы, и чувствуют себя как дома на Искривлённом берегу. Их боевой опыт говорит, что враги, которые стоят на месте и никого не атакуют, наиболее опасны. Когда они видят такого врага, Они становятся подозрительными и тоже стоят на месте.",
   "Diese verzerrten Humanoiden sind geschickte Krieger und Segler, die sich an der Verzerrten Küste zu Hause fühlen. Ihre Kampferfahrung hat sie gelehrt, das Feinde, die warten ohne sich zu bewegen oder anzugreifen, die tödlichsten sind. Wenn sie solch einen Feind treffen, werden sie extrem misstrauisch und warten ebenfalls.",
/*MISSING*/    "These warped humanoids are skilled warriors and sailors, and they feel at home at the Warped Coast. Their battle experience has taught them that enemies who wait without moving or attacking anything are the most deadly. If they see such an enemy, they become extremely suspicious, and they also wait.",
    }},
  {0xb405f402, { // "g = grid"
   "g = siatka",
   "g = şablon",
   "g = síť",
   "g = сетка",
   "g = Raster",
/*MISSING*/    "g = grid",
    }},
  {0xb40f0c4d, { // "\"I would love to come to your world with you!\""
   "\"Chcia%ł1bym wrócić z Tobą do Twojego świata!\"",
   "Senle beraber dünyana gelmeyi çok isterdim!\"",
   "Chtě%l1 bych se s tebou vrátit do tvého světa!\"",
   "Я хотел%E1 бы отправиться в Ваш мир вместе с Вами!\"",
   "Ich würde gerne mit dir in deine Welt kommen!\"",
/*MISSING*/    "\"I would love to come to your world with you!\"",
    }},
  {0xb4167b8c, { // "%The1 squeaks in a confused way."
   "Zmieszan%ya1 %1 pisn%ął1.",
   "%1 şaşkın bir şekilde ciyaklıyor.",
   "%1 zmateně piští.",
   "%1 пищит, запутавшись.",
   "%Die1 %1 piepst verwirrt.",
/*MISSING*/    "%The1 squeaks in a confused way.",
    }},
  {0xb4276c46, { // "A bird who hunts in the treetops of the Yendorian Forest."
   "Ptak polujący w koronach drzew w lesie yendoriańskim.",
/*MISSING*/    "A bird who hunts in the treetops of the Yendorian Forest.",
   "Pták, který loví v korunách Yendorského lesa.",
   "Эта птица охотится в Йендорском лесу.",
   "Ein Vogel welcher in den Baumwipfeln des Yendorianischen Waldes jagt.",
/*MISSING*/    "A bird who hunts in the treetops of the Yendorian Forest.",
    }},
  {0xb435ae8b, { // "Circles"
   "Koła",
/*MISSING*/    "Circles",
   "Kružnice",
   "Окружности",
/*MISSING*/    "Circles",
/*MISSING*/    "Circles",
    }},
  {0xb45264c8, { // "Palace Pattern"
   "Pałacowy Wzór",
   "Saray Dokusu",
   "Palácový vzor",
   "Дворцовый узор",
   "Palastmuster",
/*MISSING*/    "Palace Pattern",
    }},
  {0xb495e92b, { // "Use Orb of the Mirror to gain copies of one of your Orbs; mirroring weaker Orbs usually yields more copies. It can only be used once per Orb type, and only when you are next to a mirror."
   "Użyj Sfery Lustra by skopiować jedną z Twoich sfer; odbijanie słabszych Sfer zwykle daje więcej kopii. Możesz użyć tylko raz na każdy typ Sfery, i tylko stojąc obok lustra.",
/*MISSING*/    "Use Orb of the Mirror to gain copies of one of your Orbs; mirroring weaker Orbs usually yields more copies. It can only be used once per Orb type, and only when you are next to a mirror.",
   "Sféru Zrcadla mùžeš použít ke kopírování ostatních Sfér; ze slabších Sfér obvykle získáš více kopií. Toto lze s každým typem Sféry udìlat pouze jednou, a pouze tehdy, když stojíš vedle zrcadla.",
   "Используй Сферу зеркала, чтобы получить копии одной из твоих сфер; отражение слабых сфер обычно даёт больше копий. Ты можешь отражать каждый тип сферы лишь один раз, и только когда ты рядом с зеркалом.",
/*MISSING*/    "Use Orb of the Mirror to gain copies of one of your Orbs; mirroring weaker Orbs usually yields more copies. It can only be used once per Orb type, and only when you are next to a mirror.",
/*MISSING*/    "Use Orb of the Mirror to gain copies of one of your Orbs; mirroring weaker Orbs usually yields more copies. It can only be used once per Orb type, and only when you are next to a mirror.",
    }},
  {0xb4b7757a, { // "HyperRogue uses the Minkowski hyperboloid model internally. Klein and Poincaré models can be obtained by perspective, and the Gans model is obtained by orthogonal projection. See also the conformal mode (in the special modes menu) for more models."
   "HyperRogue wewnętrznie używa modelu hiperboloidy Minkowskiego. Modele Kleina i Poincaré są uzyskiwane przez perspektywę, a model Gansa przez rzut ortogonalny. Więcej modeli w trybie konforemnym (w menu trybów specjalnych).",
/*MISSING*/    "HyperRogue uses the Minkowski hyperboloid model internally. Klein and Poincaré models can be obtained by perspective, and the Gans model is obtained by orthogonal projection. See also the conformal mode (in the special modes menu) for more models.",
   "HyperRogue interně používá model Minkowského hyperboloidu. Kleinův a Poincarého model lze získat pomocí perspektivy a Gansův model pomocí ortogonální projekce. Více modelů můžeš najít v konformálním módu (menu speciálních módů).",
   "HyperRogue использует для расчётов модель Минковского. Модели Клейна и Пуанкаре получаются из неё перспективой, а модель Ганса -- ортогональной проекцией. Также смотрите конформный режим (в меню специальных режимов) для других моделей.",
/*MISSING*/    "HyperRogue uses the Minkowski hyperboloid model internally. Klein and Poincaré models can be obtained by perspective, and the Gans model is obtained by orthogonal projection. See also the conformal mode (in the special modes menu) for more models.",
/*MISSING*/    "HyperRogue uses the Minkowski hyperboloid model internally. Klein and Poincaré models can be obtained by perspective, and the Gans model is obtained by orthogonal projection. See also the conformal mode (in the special modes menu) for more models.",
    }},
  {0xb4f15d44, { // "Demons of Hell do not drown when they fall into the lake in Cocytus. They turn into demonic sharks, enveloped in a cloud of steam."
   "Demony nie topią się, gdy wpadną w jezioro Kocyt. Zamieniają się w demoniczne rekiny, otoczone chmurą pary.",
   "Cehennem Şeytanları Kokitos'ta göle düştüklerinde boğulmazlar.Bir buhar bulutuyla kaplı şeytanî köpekbalıklarına dönüşürler.",
   "Když pekelní démoni spadnou do jezera v Cocytu, neutopí se. Místo toho se promění v Ďábelské žraloky zahalené v oblaku páry.",
   "Демоны из Ада не тонут в озёрах Коцита. Они превращаются в демонических акул, окутанных облаками пара.",
   "Dämonen aus der Hölle ertrinken nicht, wenn sie in den See von Kokytos fallen. Stattdessen werden sie zu Dämonenhaien, eingehüllt in eine Wolke aus Dampf.",
/*MISSING*/    "Demons of Hell do not drown when they fall into the lake in Cocytus. They turn into demonic sharks, enveloped in a cloud of steam.",
    }},
  {0xb4fadd62, { // "A strange land filled with mirrors. Break magic mirrors and enter clouds of mirage to gain treasures and helpful Mimics."
   "Dziwna kraina wypełniona lustrami. Rozbijaj magiczne lustra i wchodź w chmury mirażowe, by zdobywać skarby i przyjazne Mimiki.",
/*MISSING*/    "A strange land filled with mirrors. Break magic mirrors and enter clouds of mirage to gain treasures and helpful Mimics.",
   "Podivná země plná zrcadel. Rozbíjením kouzelných zrcadel a procházením oblaků přeludů získáš poklady a užitečné Mimiky.",
   "Странная земля, наполненная зеркалами. Разбивай магические зеркала и входи в облака миражей, чтобы найти сокровища и дружелюбных Мимиков.",
/*MISSING*/    "A strange land filled with mirrors. Break magic mirrors and enter clouds of mirage to gain treasures and helpful Mimics.",
/*MISSING*/    "A strange land filled with mirrors. Break magic mirrors and enter clouds of mirage to gain treasures and helpful Mimics.",
    }},
  {0xb5149a1e, { // "You have found the Key! Now unlock this Orb of Yendor!"
   "Znalaz%łeś0 Klucz! Otwórz Sferę Yendoru!",
   "Anahtarı buldun! Bu Yendorun Küresinin kilidini açabilirsin!",
   "Naše%l0 jsi Klíč! Teď odemkni tuto Yendorskou sféru!",
   "Вы нашли ключ! Откройте им сферу Йендора!",
   "Du hast den Schlüssel gefunden! Zeit den Orb von Yendor zu öffnen!",
   "Você achou a chave! Agora desbloqueie esse Orbe de Yendor!",
    }},
  {0xb51f5809, { // "heptagonal"
   "siedmiokąt",
   "yedigensel",
   "sedmiúhelníky",
   "семиугольник",
   "heptagonal",
/*MISSING*/    "heptagonal",
    }},
  {0xb5565061, { // "Enemies killed: %1"
   "Potwory pokonane: %1",
   "Öldürülen Düşmanlar: %1",
   "Zabitých netvorů: %1",
   "Врагов убито: %1",
   "Getötete Gegner: %1",
   "Inimigos mortos: %1",
    }},
  {0xb56b6bad, { // "The 'world overview' shows all the lands in HyperRogue."
   "'Przegląd krain' pokazuje wszystkie krainy w HyperRogue.",
/*MISSING*/    "The 'world overview' shows all the lands in HyperRogue.",
   "'Pøehled krajù' ti ukáže všechny kraje v HyperRogue.",
   "'Обзор мира' покажет все земли в HyperRogue.",
/*MISSING*/    "The 'world overview' shows all the lands in HyperRogue.",
/*MISSING*/    "The 'world overview' shows all the lands in HyperRogue.",
    }},
  {0xb612caa4, { // "\"Maps... Just like the world, but smaller... how is that even possible?!\""
   "\"Mapy... jak świat, tylko mniejsze... jak to w ogóle możliwe?!\"",
   "\"Haritalar... Gerçek dünya gibi, ama daha küçük... bu nasıl mümkün olabilir?!\"",
   "\"Mapy... stejné jako svět, ale menší... jak je to vůbec možné?!\"",
   "\"Карты... как мир, только меньше... как это возможно?!\"",
   "\"Karten... Genau wie die Welt, nur kleiner... wie ist das möglich?!\"",
/*MISSING*/    "\"Maps... Just like the world, but smaller... how is that even possible?!\"",
    }},
  {0xb620161f, { // "display only heptagons"
   "pokaż tylko siedmiokąty",
   "sadece yedigenleri göster",
   "zobraz pouze sedmiúhelníky",
   "показать только семиугольники",
   "Nur Heptagone anzeigen",
/*MISSING*/    "display only heptagons",
    }},
  {0xb658c19a, { // ", m - mode: normal"
   ", m - tryb normalny",
   ", m - normal mod",
   ", m - normální mód",
   ", m - нормальный режим",
   ", m - Modus: Normal",
/*MISSING*/    ", m - mode: normal",
    }},
  {0xb6689dff, { // "Let's see more models of the hyperbolic plane. This model uses a hyperboloid in the Minkowski geometry; it is used internally by HyperRogue."
   "Pokażemy teraz więcej modeli. To jest hiperboloida w geometrii Minkowskiego; HyperRogue wewnętrznie używa tego modelu.",
/*MISSING*/    "Let's see more models of the hyperbolic plane. This model uses a hyperboloid in the Minkowski geometry; it is used internally by HyperRogue.",
   "Podívejme se na další modely hyperbolické roviny. Tento model používá hyperboloid v Minkowského geometrii; právě tento model HyperRogue používá interně.",
   "Давай посмотрим на другие модели. Эта модель использует гиперболоид в геометрии Минковского; HyperRogue внутренне использует эту модель.",
/*MISSING*/    "Let's see more models of the hyperbolic plane. This model uses a hyperboloid in the Minkowski geometry; it is used internally by HyperRogue.",
/*MISSING*/    "Let's see more models of the hyperbolic plane. This model uses a hyperboloid in the Minkowski geometry; it is used internally by HyperRogue.",
    }},
  {0xb74fae26, { // "Only thin twigs and leaves here. They may bear fruits, but for you, these cells count as unstable."
   "Tu są jedynie cieniutkie gałązki i liście. Mogą tu rosnąć owoce, ale dla Ciebie te pola liczą się jako niestabilne.",
/*MISSING*/    "Only thin twigs and leaves here. They may bear fruits, but for you, these cells count as unstable.",
   "Tady jsou jen tenké větvičky a listy. Mohou nést ovoce, ale pro tebe se počítají jako nestabilní.",
   "Здесь лишь тонкие веточки и листья. Они могут выдержать фрукт, но не ваш вес.",
   "Hier sind nur dünne Zweige und Blätter. Sie mögen Früchte tragen, aber für dich sind diese Zellen instabil.",
/*MISSING*/    "Only thin twigs and leaves here. They may bear fruits, but for you, these cells count as unstable.",
    }},
  {0xb7760445, { // "\"In this world there is plenty of space for everyone. We do not need wars.\""
   "\"W tym świecie jest dość miejsca dla każdego. Nie potrzebujemy wojen.\"",
/*MISSING*/    "\"In this world there is plenty of space for everyone. We do not need wars.\"",
   "\"V tomhle světě je dost místa pro všechny. Nepotřebujeme války.\"",
   "\"В этом мире так много места для каждого. Вам не нужны войны.\"",
/*MISSING*/    "\"In this world there is plenty of space for everyone. We do not need wars.\"",
/*MISSING*/    "\"In this world there is plenty of space for everyone. We do not need wars.\"",
    }},
  {0xb7879863, { // "Hint: use arrow keys to scroll."
   "Wskazówka: przewijasz strzałkami.",
/*MISSING*/    "Hint: use arrow keys to scroll.",
   "Nápověda: šipkami můžeš scrollovat.",
   "Подсказка: листайте экран стрелками.",
/*MISSING*/    "Hint: use arrow keys to scroll.",
/*MISSING*/    "Hint: use arrow keys to scroll.",
    }},
  {0xb792f8be, { // "The Grimoires contain many secrets of the Great Old Ones. Each new inner circle in the Temple of Cthulhu contains new Grimoires, with new secrets. You hope to read them when you return home, and to learn many things. The knowledge is valuable to you, but it is rather pointless to try to get several copies of the same Grimoire..."
   "Te Księgi zawierają tajemnice Wielkich Przedwiecznych. Każdy krąg wewnętrzny w Świątyni Cthulhu zawiera nowe Księgi, z nowymi sekretami. Zamierzasz je przeczytać po powrocie do domu. Wiedza jest dla Ciebie wartościowa, ale nie ma sensu, by brać kilka kopii tej samej Księgi...",
   "Kara Kitaplar Eski Yücelere dair pek çok sır içerir. Cthulhu Tapınağındaki her iç daire yeni sırlar içeren yeni Kara Kitaplar içerir.Eve geri döndüğünde bunları okuyup pek çok şey öğrenmeyi umuyorsun. Bilgi senin için önemli, bu yüzden aynı Kara Kitaptan birden fazla almak çok anlamsız.",
   "Grimoáry obsahují mnoho tajemství Starých Velkých. Každý další vnitřní kruh Cthulhuova chrámu obsahuje nové Grimoáry s novými tajemstvími. Doufáš, že si je všechny přečteš, až se vrátíš domů, a že se naučíš mnoho věcí. Znalosti z Grimoárů jsou pro tebe cenné, ale nemá smysl snažit se získat několik kopií stejného Grimoáru...",
   "Гримуары содержат много секретов Великих предков Каждый новый круг Храма Ктулху содержит новые тома с новыми секретами. Вы можете принести их домой, прочитать и узнать много вещей.Знание ценно для Вас, но Вам не нужны копии одного Гримуара... ",
   "Die Grimoires enthalten viele Geheimnisse der weisen Alten. Jeder neue innere Kreis des Tempels des Cthulhu enthält neue Grimoires mit mehr Geheimnissen. Das Wissen ist dir sehr kostbar, aber es wäre natürlich sinnlos, mehrmals das gleiche Grimoire aufzuheben...",
/*MISSING*/    "The Grimoires contain many secrets of the Great Old Ones. Each new inner circle in the Temple of Cthulhu contains new Grimoires, with new secrets. You hope to read them when you return home, and to learn many things. The knowledge is valuable to you, but it is rather pointless to try to get several copies of the same Grimoire...",
    }},
  {0xb7c225da, { // "The Necromancer's Totem contains hellish incantations..."
   "Totem Nekromanty zawiera piekielne inkantacje...",
   "Ölüçağıranın Totemi şeytani işlemelerle doluydu...",
   "Nekromancerův totem obsahuje pekelná zaklínání...",
   "Тотем некроманта содержит адские заклинания...",
   "Das Totenbeschwörer-Totem enthält diabolische Zaubersprüche...",
   "O Totem do Necromante contém encantamentos infernais...",
    }},
  {0xb8041ec3, { // "t = shift"
   "t = przesuń",
   "t = kaydır",
   "t = přesuň",
   "t = сдвиг",
   "t = verlagern",
/*MISSING*/    "t = shift",
    }},
  {0xb8043afb, { // "%The1 fights with %the2!"
   "%1 zaatakował %a2!",
   "%1 %a2 savaşıyor!",
   "%1 se pustil do boje s %a2!",
   "%1 атаковал%E1 %a2!",
   "%Der1 %1 bekämpft %den2 %a2!",
/*MISSING*/    "%The1 fights with %the2!",
    }},
  {0xb80aeaae, { // "\n\nOrb unlocked: %1"
   "\n\nOdblokowana Sfera: %1",
/*MISSING*/    "\n\nOrb unlocked: %1",
   "\n\nByla odemèena Sféra: %1",
   "\n\nОткрыта сфера: %1",
/*MISSING*/    "\n\nOrb unlocked: %1",
/*MISSING*/    "\n\nOrb unlocked: %1",
    }},
  {0xb836c2fa, { // "joystick mode: manual (press a button to move)"
   "tryb dżojstika: ręczny (naciśnij przycisk)",
   "oyun kolu modu: manuel (bir düğmeye basarak hareket edersin)",
   "mód joysticku: ruční (pohybuješ se stiskem tlačítka)",
   "Режим джойстика: ручной (нажмите на клавишу для перемещения)",
   "Joystick: Manuell (Knopf drücken zum bewegen)",
   "modo joystick: manual (pressione os botões para mover-se)",
    }},
  {0xb842e685, { // "Kill monsters and collect treasures, and you may get access to Hell..."
   "Zabijaj potwory, zdobywaj skarby, może trafisz do Piekła...",
   "Canavarlar öldür, hazineler topla ve belki Cehenneme erişebilirsin...",
   "Zabíjej netvory a sbírej poklady a možná najdeš cestu do Pekla...",
   "Убивайте монстров и собирайте сокровища, и Вы сможете открыть Ад...",
   "Wenn du Monster tötest und Schätze sammelst erhältst du vielleicht Zutritt zur Hölle...",
   "Mate monstros e colete tesouros e você poderá acessar o Inferno...",
    }},
  {0xb8492a60, { // "Necromancers can raise ghosts and zombies from fresh graves."
   "Nekromanci wzbudzają duchy i zombie ze świeżych grobów.",
   "Ölüçağıranlar mezarlardan hayaletler ve hortlaklar çağırabilirler.",
   "Nekromanceři dokážou probouzet duchy a zombie z čerstvých hrobů.",
   "Некромант может призывать призраков и зомби из свежих могил.",
   "Totenbeschwörer können Geister und Zombies aus frischen Gräbern auferwecken.",
/*MISSING*/    "Necromancers can raise ghosts and zombies from fresh graves.",
    }},
  {0xb872da2c, { // "THE END"
   "KONIEC",
/*MISSING*/    "THE END",
   "KONEC",
   "КОНЕЦ",
/*MISSING*/    "THE END",
/*MISSING*/    "THE END",
    }},
  {0xb8868a94, { // "This is applied in HyperRogue to create landscapes, such as the chasms in the land of Reptiles or the Dragon Chasms, which you should find quickly. Also in the Dragon Chasms, you can find a Baby Tortoise, and try to find a matching adult tortoise in the Galápagos. There are over two millions of species, but since there is so much space in hyperbolic geometry, finding a matching tortoise is possible. The brighter the color in Galápagos is, the more aspects of the tortoises in the given area are matching."
   "Jest to używane w grze do stworzenia krajobrazów, jak przepaści w Jaszczurkach lub Smoczych Otchłaniach, które niedługo znajdziesz. W Smoczych Otchłaniach możesz znaleźć Żółwika, i poszukać pasującego dorosłego żółwia w Galápagos. Jest ponad 2 miliony gatunków, ale ze względu na ilość miejsca w geometrii hiperbolicznej, znalezienie pasującego żółwia jest możliwe. Im jaśniejszy kolor w Galápagos, tym więcej aspektów żółwika pasuje do żółwii z danego terenu.",
/*MISSING*/    "This is applied in HyperRogue to create landscapes, such as the chasms in the land of Reptiles or the Dragon Chasms, which you should find quickly. Also in the Dragon Chasms, you can find a Baby Tortoise, and try to find a matching adult tortoise in the Galápagos. There are over two millions of species, but since there is so much space in hyperbolic geometry, finding a matching tortoise is possible. The brighter the color in Galápagos is, the more aspects of the tortoises in the given area are matching.",
   "HyperRogue tohoto využívá pro vytváření krajin, jako jsou propasti v zemi Ještěrek nebo v Dračích propastech, které bys měl rychle najít. V Dračích propastech můžeš také najít Želvičku, a potom se pokusit na Galapágách najít dospělou želvu, která jí odpovídá. Existuje více než dva milióny druhů želv, ale protože v hyperbolické geometrii je tolik místa, je možné najít tu, která patří ke tvé Želvičce. Čím jasnější barva, tím více aspektů želv v dané oblasti odpovídá tvé Želvičce.",
   "Фракталы применяются для создания пропастей в земле Рептилий и в Драконьей бездне, которые ты быстро найдёшь. Также в Драконьей бездне ты можешь найти Черепашку и попробовать найти её родителей на Галапагосах. Там живёт более двух миллионов видов, но в гиперболической геометрии очень много места, и найти её вид вполне возможно. Чем ярче цвет клетки, тем больше параметров совпадает с тем видом, что ты ищешь.",
/*MISSING*/    "This is applied in HyperRogue to create landscapes, such as the chasms in the land of Reptiles or the Dragon Chasms, which you should find quickly. Also in the Dragon Chasms, you can find a Baby Tortoise, and try to find a matching adult tortoise in the Galápagos. There are over two millions of species, but since there is so much space in hyperbolic geometry, finding a matching tortoise is possible. The brighter the color in Galápagos is, the more aspects of the tortoises in the given area are matching.",
/*MISSING*/    "This is applied in HyperRogue to create landscapes, such as the chasms in the land of Reptiles or the Dragon Chasms, which you should find quickly. Also in the Dragon Chasms, you can find a Baby Tortoise, and try to find a matching adult tortoise in the Galápagos. There are over two millions of species, but since there is so much space in hyperbolic geometry, finding a matching tortoise is possible. The brighter the color in Galápagos is, the more aspects of the tortoises in the given area are matching.",
    }},
  {0xb8a07c50, { // "This Orb halves the power usage of orbs which cost some charges with each activation. It even affects the one-shot orbs such as Flash or Teleport. If such an activation normally costs x charges, it costs only x/2 (rounded up) if you have an Orb of Energy."
   "Ta Sfera dwukrotnie zmniejsza zużycie mocy przez te Sfery, których każda aktywacja kosztuje pewną liczbę ładunków. Działa nawet na Sfery jednokrotnego użytku jak Błysk czy Teleport. Jeśli normalnie taka aktywacja kosztuje x ładunków, a masz Sferę Energii, to będzie ona kosztowała x/2 ładunków (zaokrąglone w górę).",
/*MISSING*/    "This Orb halves the power usage of orbs which cost some charges with each activation. It even affects the one-shot orbs such as Flash or Teleport. If such an activation normally costs x charges, it costs only x/2 (rounded up) if you have an Orb of Energy.",
   "Některé Sféry spotřebují každou aktivací nějaké množství nábojů. Sféra Energie snižuje toto množství na polovinu. Funguje dokonce i s jednorázovými Sférami, jako je Sféra Záblesku nebo Teleportace. Aktivace, která by normálně stála x nábojů bude se Sférou Energie stát pouze x/2 nábojů (zaokrouhleno nahoru).",
   "Эта сфера уменьшает энергию, используемую другими сферами при активации. Это работает даже для одноразовых сфер, например, Вспышки или Телепорта. Если обычно использование стоило x единиц энергии, то с этой сферой оно будет стоить x/2 единиц (округлённое вверх).",
   "Dieser Orb halbiert den Kraftverbrauch von Orbs, die mit jeder Aktivierung Aufladungen verbrauchen. Es beeinflusst auch die einmalig aktivierbaren Orbs wie Lichtblitz oder Teleportation. Verbraucht eine Aktivierung normalerweise x Ladungen, dann verbraucht sie nur x/2 Ladungen (aufgerundet),  wenn du einen Orb der Energie besitzt.",
/*MISSING*/    "This Orb halves the power usage of orbs which cost some charges with each activation. It even affects the one-shot orbs such as Flash or Teleport. If such an activation normally costs x charges, it costs only x/2 (rounded up) if you have an Orb of Energy.",
    }},
  {0xb8cfc315, { // "spherical"
   "sferyczna",
/*MISSING*/    "spherical",
   "sférická",
   "сферическая",
/*MISSING*/    "spherical",
/*MISSING*/    "spherical",
    }},
  {0xb8fdc440, { // "A petrified creature."
   "Stworzenie zamienione w kamień.",
/*MISSING*/    "A petrified creature.",
   "Jakýsi tvor proměněný v kámen.",
   "Окаменевшее существо.",
   "Eine versteinerte Kreatur.",
/*MISSING*/    "A petrified creature.",
    }},
  {0xb93299c8, { // "You have improved your total high score on Google. Congratulations!"
   "Poprawi%łeś0 swój ogólny najlepszy wynik na Google. Gratulacje!",
   "Google'daki toplam yüksek puanını artırdın. Tebrikler!",
   "Zlepši%l0 jsi své celkové nejvyšší skóre na Googleu. Gratulujeme!",
   "Вы поставили рекорд по общему счёту в таблицах Google. Поздравляем!",
   "Du hast deinen Gesamtrekord auf Google verbessert! Glückwunsch!",
/*MISSING*/    "You have improved your total high score on Google. Congratulations!",
    }},
  {0xb957e9ce, { // "You join %the1."
   "Połączy%łeś0 się z %abl1.",
   "%abl1 bütünleştin.",
   "Spoji%l0 ses %s1 %abl1.",
   "Вы встретились с %abl1.",
   "Du führst die %P1 zusammen.",
   "Você se junta %d1.",
    }},
  {0xb9aa3ece, { // "take me back"
   "powrót do menu",
/*MISSING*/    "take me back",
   "návrat do menu",
   "вернуться в меню",
   "Bring mich zurück",
/*MISSING*/    "take me back",
    }},
  {0xb9ce32c3, { // "Map loaded from %1"
   "Mapa wczytana z %1",
   "Harita %1 konumundan yüklendi",
   "Mapa načtena z %1",
   "Карта загружена из %1",
   "Karte von %1 geladen",
/*MISSING*/    "Map loaded from %1",
    }},
  {0xb9e45a0e, { // "%The1 will be revivable at %2 $$$"
   "%1 będzie mogła ożyć przy %2 $$$",
/*MISSING*/    "%The1 will be revivable at %2 $$$",
   "%a1 bude možné oživit na %2 $$$",
   "%1 будет пригоден для возрождения при %2 $$$",
/*MISSING*/    "%The1 will be revivable at %2 $$$",
/*MISSING*/    "%The1 will be revivable at %2 $$$",
    }},
  {0xb9eae632, { // "%1 takes %his1 revenge on %the2!"
   "%1 się mści!",
   "%1 intikamını alıyor!",
   "%1 se mstí!",
   "%1 будет мстить!",
   "%1 nimmt Rache an %dem2 %d2!",
/*MISSING*/    "%1 takes %his1 revenge on %the2!",
    }},
  {0xb9f1b54e, { // "counterclockwise"
   "antyzegarowo",
/*MISSING*/    "counterclockwise",
   "proti směru hodinových ručiček",
   "против часовой стрелки",
   "gegen den Uhrzeigersinn",
/*MISSING*/    "counterclockwise",
    }},
  {0xba0df93c, { // "Cannot push into another player!"
   "Nie można pchać na innego gracza!",
/*MISSING*/    "Cannot push into another player!",
   "Nemůžeš tlačit do jiného hráče!",
   "Нельзя толкать в другого игрока!",
   "Kann nicht auf einen anderen Spieler ziehen!",
/*MISSING*/    "Cannot push into another player!",
    }},
  {0xba80959f, { // "This orb can be used to summon friendly golems. It is used instantly when you pick it up."
   "Ta sfera przywołuje przyjazne golemy. Jest natychmiast używana w momencie podniesienia.",
   "Bu küre size dost golemler çağırmak için kullanılır. Onu eline aldığında anında kullanılır.",
   "Tuto sféru můžeš použít k vyvolání přátelských golemů. Použije se automaticky, jakmile ji sebereš.",
   "Эта сфера призывает дружественных големов. Она срабатывает, как только Вы её поднимаете.",
   "Dieser Orb dient dazu verbündete Golems herbeizurufen. Er wird aktiviert, sobald du ihn aufhebst.",
/*MISSING*/    "This orb can be used to summon friendly golems. It is used instantly when you pick it up.",
    }},
  {0xbad9db6a, { // "You damage %the1 with a mental blast!"
   "Zrani%łeś0 %a1 mocą psychiczną!",
/*MISSING*/    "You damage %the1 with a mental blast!",
   "Zrani%l0 jsi %a1 mentálním útokem!",
   "Вы атаковали %a1 энергией мысли!",
   "Du schädigst %den1 %a1 mit einem mentalen Stoß!",
/*MISSING*/    "You damage %the1 with a mental blast!",
    }},
  {0xbae8597e, { // "An alternative version of Orb of the Sword. If you have both of them, you have two energy swords, facing in opposite directions."
   "Alternatywna wersja Sfery Ostrza. Jeśli masz obie, masz dwa ostrza energii, skierowane w przeciwnych kierunkach.",
/*MISSING*/    "An alternative version of Orb of the Sword. If you have both of them, you have two energy swords, facing in opposite directions.",
   "Alternativní verze Sféry Meče. Pokud máš obě zároveň, budeš mít dva energetické meče namířené do opačných směrů..",
   "Другая версия Сферы меча. Если у вас есть обе, то вы имеете два меча, направленные в противоположные стороны.",
   "Eine alternative Variante des Orbs des Schwertes. Wenn du beide Orbs besitzt, hast du zwei Energieschwerter, die in entgegengesetzte Richtungen zeigen.",
/*MISSING*/    "An alternative version of Orb of the Sword. If you have both of them, you have two energy swords, facing in opposite directions.",
    }},
  {0xbb2abf97, { // "seven players"
   "siedmiu graczy",
/*MISSING*/    "seven players",
   "sedm hráčů",
   "семь игроков",
   "sieben Spieler",
/*MISSING*/    "seven players",
    }},
  {0xbb5ccdb9, { // "You can go inside the Magic Mirror, and produce some mirror images to help you."
   "Możesz wejść w Magiczne Lustro, by Twoje odbicia z niego wyszły i Ci pomogły.",
   "Sihirli aynanın içine girerek sana yardım etmesi için bazı yansımalar üretebilirsin.",
   "Do Kouzelného zrcadla můžeš vstoupit a vytvořit několik zrcadlových obrazů, které ti pomohou.",
   "Вы можете пройти в Магическое зеркало и создать отражения, которые будут Вам помогать.",
   "Du kannst den Zauberspiegel betreten, um Spiegelbilder zu erzeugen, die dir helfen.",
/*MISSING*/    "You can go inside the Magic Mirror, and produce some mirror images to help you.",
    }},
  {0xbb82f716, { // "The tactics of fighting simple monsters, such as the Yetis from the Icy Lands, might appear shallow, but hyperbolic geometry is essential even there. In the next slide, you are attacked by two monsters at once. You can make them line up simply by running away in a straight line. Press '2' to try the same in the Euclidean world -- it is impossible."
   "Taktyka walki z prostymi potworami, jak Yeti z Krainy Lodu, może się wydawać płytka, ale geometria hiperboliczna jest istotna nawet tu. W następnym slajdzie atakują Cię 2 potwory naraz. Możesz uciec po prostu odchodząc od nich w linii prostej. Naciśnij '2', by spróbować tego w świecie euklidesowym -- jest to niemożliwe.",
/*MISSING*/    "The tactics of fighting simple monsters, such as the Yetis from the Icy Lands, might appear shallow, but hyperbolic geometry is essential even there. In the next slide, you are attacked by two monsters at once. You can make them line up simply by running away in a straight line. Press '2' to try the same in the Euclidean world -- it is impossible.",
   "Taktika boje proti jednoduchým netvorům, jako jsou například Yetiové z Ledového kraje, může vypadat jednoduše, ale i tam je hyperbolická geometrie zásadní. Na dalším snímku na tebe útočí dva netvoři najednou. Pokud budeš utíkat po přímce, přinutíš je, aby se seřadili. Stiskni klávesu '2', abys to samé zkusil v eukleidovském světě -- je to nemožné.",
   "Тактика боя с простыми монстрами, такими как Йети из Ледяной земли, может показаться обычной, но геометрия важна и здесь. На следующем слайде тебя атакуют сразу два монстра. Ты можешь справиться с ними, просто отходя по прямой. Нажми '2', чтобы попробовать сделать то же самое в евклидовом мире -- ничего не выйдет.",
/*MISSING*/    "The tactics of fighting simple monsters, such as the Yetis from the Icy Lands, might appear shallow, but hyperbolic geometry is essential even there. In the next slide, you are attacked by two monsters at once. You can make them line up simply by running away in a straight line. Press '2' to try the same in the Euclidean world -- it is impossible.",
/*MISSING*/    "The tactics of fighting simple monsters, such as the Yetis from the Icy Lands, might appear shallow, but hyperbolic geometry is essential even there. In the next slide, you are attacked by two monsters at once. You can make them line up simply by running away in a straight line. Press '2' to try the same in the Euclidean world -- it is impossible.",
    }},
  {0xbbbed4ae, { // "Multiplayer: Play cooperatively (locally); treasures, kills, and deaths are calculated for each player too, for more competitive play. Orbs and treasures are shared, orbs drain faster, knives recharge slower, and player characters are not allowed to separate.\n\n"
   "Wielu graczy: współpracujecie (lokalnie); można też współzawodniczyć o skarby i zabicia, które są liczone dla każdego gracza. Sfery i skarby są dzielone, sfery wyczerpują się szybciej, noże wracają wolniej, i gracze nie mogą się rozdzielać.\n\n",
/*MISSING*/    "Multiplayer: Play cooperatively (locally); treasures, kills, and deaths are calculated for each player too, for more competitive play. Orbs and treasures are shared, orbs drain faster, knives recharge slower, and player characters are not allowed to separate.\n\n",
   "Více hráčů: Zde může hrát několik hráčů kooperativně (lokálně); poklady, zabití netvoři a úmrtí se počítají pro každého hráče zvlášť, aby se dalo hrát kompetitivněji. Sféry a poklady se sdílejí, sféry se vyčerpávají rychleji, nože se dobíjejí pomaleji a postavy se nesmějí příliš vzdálit.\n\n",
   "Мультиплеер: Играйте совместно (локально); сокровища, убийства и смерти считаются отдельно, сферы и сокровища складываются, сферы угасают быстрее, ножи восстанавливаются медленнее, игроки не могут сильно разделяться.\n\n",
   "Mehrspieler: Spiele gemeinsam (lokal); Schätze, Tötungen und Tode werden für ein mehr wettstreitorientiertes Spiel zusätzlich für jeden Spieler einzeln berechnet.Orbs und Schätze werden miteinander geteilt; Orbs werden schneller verbraucht, Messer werden langsamer wiederaufgeladen, und die Spielfiguren dürfen sich nicht weit voneinander entfernen.\n\n",
/*MISSING*/    "Multiplayer: Play cooperatively (locally); treasures, kills, and deaths are calculated for each player too, for more competitive play. Orbs and treasures are shared, orbs drain faster, knives recharge slower, and player characters are not allowed to separate.\n\n",
    }},
  {0xbccfd5d0, { // "Static mode enabled."
   "Tryb statyczny włączony.",
/*MISSING*/    "Static mode enabled.",
   "Statický mód zapnut.",
   "Статичный режим включен.",
/*MISSING*/    "Static mode enabled.",
/*MISSING*/    "Static mode enabled.",
    }},
  {0xbcd35cbc, { // "These people study gravity and infinite trees. They have no special features, other than wearing a strange hat."
   "Ci ludzie badają grawitację i drzewa nieskończone. Nie mają szczególnych własności poza swoim dziwnym nakryciem głowy.",
/*MISSING*/    "These people study gravity and infinite trees. They have no special features, other than wearing a strange hat.",
   "Tito lidé studují gravitaci a nekonečné stromy. Kromě svých podivných čepic nemají žádné zvláštní schopnosti.",
   "Эти люди изучают гравитацию и бесконечные деревья. Они ничем не отличаются от обычных людей, кроме этих странных шляп.",
   "Diese Leute studieren die Gravitation und unendliche Bäume. Sie haben keine besonderen Eigenschaften außer dem Tragen eines seltsamen Hutes.",
/*MISSING*/    "These people study gravity and infinite trees. They have no special features, other than wearing a strange hat.",
    }},
  {0xbcf4a764, { // "There is not enough space!"
   "Nie ma miejsca!",
   "Yeterince yer yok!",
   "Není tady dost místa!",
   "Нет места!",
   "Zu wenig Platz!",
/*MISSING*/    "There is not enough space!",
    }},
  {0xbcf9f5e6, { // "You feel something strange about gravity here..."
   "Czujesz coś dziwnego z grawitacją tutaj...",
/*MISSING*/    "You feel something strange about gravity here...",
   "Cítíš, že na zdejší gravitaci je cosi podivného...",
   "Вы чувствуете, что с гравитацией здесь что-то не так...",
/*MISSING*/    "You feel something strange about gravity here...",
/*MISSING*/    "You feel something strange about gravity here...",
    }},
  {0xbcfaeaff, { // "A Witch with a Speed spell. She moves twice as fast as you. Unless you have an Orb of Speed too, of course!"
   "Czarownica z czarem Szybkości. Rusza się dwa razy szybciej od Ciebie. Chyba, że Ty też masz Sferę Szybkości!",
   "Hız büyüsü olan bir Cadı. Senden iki kat hızlı hareket eder. Tabi senin de bir Hız Küren yoksa!",
   "Čarodějnice s kouzlem Rychlosti. Pohybuje se dvakrát rychleji než ty. Samozřejmě, pokud náhodou také nemáš Sféru Rychlosti!",
   "Ведьма с заклинанием скорость. Она движется в 2 раза быстрее Вас. Если у Вас нет того же заклинания, конечно!",
   "Eine Geschwindigkeitshexe, sie bewegt sich doppelt so schnell wie du. Außer natürlich, du hast auch einen Orb der Geschwindigkeit!",
/*MISSING*/    "A Witch with a Speed spell. She moves twice as fast as you. Unless you have an Orb of Speed too, of course!",
    }},
  {0xbd7bf660, { // "You cannot attack Tentacles directly!"
   "Nie możesz atakować Macek wprost!",
   "Dokungaçlara doğrudan saldıramazsın!",
   "Na Chapadla nemůžeš útočit přímo!",
   "Вы не можете атаковать щупальца Ктулху!",
   "Du kannst Tentakel nicht direkt angreifen!",
   "Você não pode atacar Tentáculos diretamente!",
    }},
  {0xbdca60db, { // "0 - return"
   "0 - powrót",
/*MISSING*/    "0 - return",
   "0 - návrat",
   "0 - вернуться",
/*MISSING*/    "0 - return",
/*MISSING*/    "0 - return",
    }},
  {0xbe010ea9, { // "only native Orbs allowed in %the1"
   "tylko rodzime sfery %abl1",
   "%abl1 sadece oraya özgü Küreler kullanılabilir.",
   "%abl1 jsou povoleny pouze náležící Sféry",
   "только родные сферы доступны %abl1",
   "Nur heimische Orbs in %der1 erlaubt",
/*MISSING*/    "only native Orbs allowed in %the1",
    }},
  {0xbe125228, { // "You gain your protective Shell back!"
   "Masz Skorupę z powrotem!",
/*MISSING*/    "You gain your protective Shell back!",
   "Získal jsi zpátky svůj ochranný Krunýř!",
   "Вы вернули свою раковину!",
/*MISSING*/    "You gain your protective Shell back!",
/*MISSING*/    "You gain your protective Shell back!",
    }},
  {0xbe458555, { // "You have a vision of the future, fighting demons in Hell..."
   "Masz wizję przyszłości, w której walczysz z demonami w Piekle...",
   "Geleceğini gördün, Cehennemde şeytanlarla savaşıyordun...",
   "Máš vizi budoucnosti; vidíš se, jak bojuješ s pekelnými démony...",
   "Вы видите себя в будущем, убивающего демонов в Аду...",
   "Du hast eine Vision, du wirst in der Zukunft Dämonen in der Hölle bekämpfen...",
   "Você tem uma visão do futuro, lutando com demônios no Inferno...",
    }},
  {0xbeae1fd0, { // "One of few things that does not cause the floor in the Land of Eternal Motion to collapse. Obviously they are quite valuable."
   "Pióro Feniksa jest tak lekie, że podłoga z Krainy Wiecznego Ruchu pod nim się nie zapada. Takie pióra muszą być bardzo cenne.",
   "Hareketler Diyarında yerlerin çökmemesini sağlayan ender şeylerden biri. Bunlar, tabi ki epey değerli.",
   "Jedna z mála věcí, pod kterou se v Kraji věčného pohybu nepropadá země. Tato pírka jsou samozřejmě velice cenná.",
   "Единственная вещь, которую выдерживает пол в Земле вечного движения. Очевидно, довольно ценно.",
   "Eines der wenigen Dinge, die den Boden im Land der ewigen Bewegung nicht einstürzen lassen. Sie sind offensichtlich sehr wertvoll.",
/*MISSING*/    "One of few things that does not cause the floor in the Land of Eternal Motion to collapse. Obviously they are quite valuable.",
    }},
  {0xbeb77adf, { // "You can walk on it... but beware."
   "Możesz tu przejść, ale uważaj!",
   "Üzerinde yürüyebilirsin. Ama dikkatli ol!",
   "Můžeš po něm chodit... ale pozor.",
   "Вы можете пойти сюда, но будьте осторожны!",
   "Du kannst darauf gehen... aber Vorsicht.",
/*MISSING*/    "You can walk on it... but beware.",
    }},
  {0xbec253f3, { // "You have been trapped in a strange, non-Euclidean world. Collect as much treasure as possible before being caught by monsters. The more treasure you collect, the more monsters come to hunt you, as long as you are in the same land type. The Orbs of Yendor are the ultimate treasure; get at least one of them to win the game!"
   "Trafi%łeś0 do dziwnej, nieeuklidesowej krainy. Zdobądź jak najwięcej skarbów, zanim potwory Cię dopadną. Im więcej zdobędziesz skarbów, tym więcej potworów w danej krainie na Ciebie poluje. Sfery Yendoru są skarbem ostatecznym: zdobądź Sferę Yendoru, by wygrać grę!",
   "Garip ve Ökliddışı bir dünyada hapsoldun. Canavarlar tarafından yakalanmadan önce olabildiğince hazine topla.Aynı arazi tipinde daha fazla hazine topladıkça,daha fazla canavar sana saldıracak.Yendor'un küreleri nihai hazinelerdir, onlardan bir tane ile oyunu kazanabilirsin!",
   "Jsi uvězněn v podivném, neeukleidovském světě. Sesbírej co nejvíce pokladů, než tě dostanou netvoři. Čím víc pokladů sebereš, tím víc netvorů po tobě půjde, dokud budeš ve stejném typu kraje. Největším pokladem jsou Yendorské sféry: získáš-li alespoň jednu z nich, vyhraješ hru!",
   "Вы оказались заперты в странном неевклидовом мире. Соберите как можно больше сокровищ перед тем как Вас поймают монстры. Чем больше сокровищ Вы собираете, тем больше монстров на Вас охотится, пока Вы находитесь на одной территории. Сферы Йендора - уникальные сокровища; соберите хотя бы одну, чтобы выиграть!",
   "Du wurdest in einer seltsamen, nicht-euklidischen Welt gefangen... Sammle so viele Schätze wie du kannst, bevor dich die Monster fangen. Je mehr Schätze du hast, desto mehr Monster werden dich jagen bis du das Land verlässt. Die Orbs von Yendor sind der ultimative Schatz - sammle mindestens eine um zu gewinnen!",
   "Você está preso em um estranho mundo não-Euclidiano. Colete o máximo de tesouro possível antes de ser pego pelos monstros. Quanto mais tesouro você coletar, mais monstros virão lhe caçar, desde que você esteja no mesmo tipo de terreno. Os Orbes de Yendor são o tesouro final; pegue ao menos um para vencer o jogo!",
    }},
  {0xbf0ab4a0, { // "You just cannot stand in place, those roses smell too nicely."
   "Nie możesz stać w miejscu, róże zbyt ładnie pachną.",
/*MISSING*/    "You just cannot stand in place, those roses smell too nicely.",
   "Nemůžeš zůstat na místě, ty růže příliš krásně voní.",
   "Вы не можете устоять на месте, эти розы пахнут слишком прекрасно.",
   "Du kannst einfach nicht stehen bleiben - diese Rosen duften zu gut.",
/*MISSING*/    "You just cannot stand in place, those roses smell too nicely.",
    }},
  {0xbf1a2063, { // "Warnings are issued when you try to do something that appears dangerous, like stepping on a known mine, or getting your boat destroyed by a Kraken without having Orb of the Fish. In some cases the action might actually be safe -- so you can ignore the warning and do it anyway, simply by repeating the action."
   "Ostrzeżenia pokazują się, gdy próbujesz zrobić coś niebezpiecznego: stanąć na minie, spowodować zniszczenie łódki przez Krakena bez Sfery Ryby, itp. W niektórych przypadkach może to być bezpieczne -- także możesz zignorować ostrzeżenie, po prostu próbując wykonać akcję jeszcze raz.",
/*MISSING*/    "Warnings are issued when you try to do something that appears dangerous, like stepping on a known mine, or getting your boat destroyed by a Kraken without having Orb of the Fish. In some cases the action might actually be safe -- so you can ignore the warning and do it anyway, simply by repeating the action.",
   "Varování se objeví, když se pokusíte udělat něco, co vypadá nebezpečně, jako je šlápnutí na políčko, kde víte, že je mina, nebo riskování, že vám Kraken zničí loďku, když nemáte Sféru Ryby. V některých případech může být taková věc ve skutečnosti bezpečná -- a proto je možné opakováním akce varování ignorovat a stejně to udělat.",
   "Предупреждения возникают, когда вы пытаетесь сделать что-то опасное: наступить на известную мину, позволить Кракену уничтожить вашу лодку без Сферы Рыбы, и т. п. Иногда эти действия на самом деле безопасны -- вы можете проигнорировать предупреждение, просто повторив действие.",
   "Es werden Warnungen ausgegeben wenn du versuchst etwas zu tun das gefährlich scheint wieauf eine bekannte Mine treten oder wenn ein Krake dein Boot zerstören würde während du keinen Orb des Fischs besitzt. In einigen Fällen mag deine Aktion tatsächlich ungefährlich sein. Dann kannst du die Warnung ignorieren und es trotzdem tun, indem du die Aktion einfach wiederholst.",
/*MISSING*/    "Warnings are issued when you try to do something that appears dangerous, like stepping on a known mine, or getting your boat destroyed by a Kraken without having Orb of the Fish. In some cases the action might actually be safe -- so you can ignore the warning and do it anyway, simply by repeating the action.",
    }},
  {0xbf2d0f78, { // "You feel that the stars are right, and you can access R'Lyeh!"
   "Gwiazdy są na miejscu, R'Lyeh czeka!",
   "Yıldızlar sana, R'Lyeh'e ulaşabileceğini söylüyor!",
   "Cítíš, že postavení hvězd je správné, a ty můžeš vstoupit do R'Lyeh!",
   "Вы чувствуете, что звёзды правы, Вы можете открыть Р'Льех!",
   "Die Sterne stehen günstig, du kannst R´Lyeh erreichen!",
   "Você sente que as estrelas estão alinhadas e você pode acessar o R'Lyeh!",
    }},
  {0xbf3fc51e, { // "Screenshot saved to %1"
   "Zrzut ekranu zapisany do %1",
   "Ekran görüntüsü %1 konumuna kaydedildi.",
   "Snímek obrazovky uložen do %1",
   "Скриншот сохранён в %1",
   "Screenshot nach %1 gespeichert",
   "Captura de tela salva em %1",
    }},
  {0xbf582b3c, { // "last messages:"
   "ostatnie wiadomości: ",
   "son mesajlar: ",
   "poslední zprávy: ",
   "последние сообщения: ",
   "Letzte Nachrichten: ",
   "últimas mensagens: ",
    }},
  {0xbf7d0fe1, { // "Cold white gems, found in the Icy Land."
   "Zimne białe kamienie z Krainy Lodu.",
   "Buz Diyarlarında bulunan soğuk beyaz mücevherler.",
   "Chladné bílé drahokamy, které se nacházejí v Ledovém kraji.",
   "Холодные белые камни, найденные в Ледяной земле.",
   "Kalte weiße Edelsteine aus dem Eisigen Land.",
/*MISSING*/    "Cold white gems, found in the Icy Land.",
    }},
  {0xbfd31f6f, { // "You'll need your magical sword against the Draugar!"
   "Potrzebujesz magicznego miecza przeciwko Draugrom!",
/*MISSING*/    "You'll need your magical sword against the Draugar!",
   "Proti Draugarům budeš potřebovat kouzelný meč!",
   "Против Драугра Вам понадобится магический меч!",
/*MISSING*/    "You'll need your magical sword against the Draugar!",
/*MISSING*/    "You'll need your magical sword against the Draugar!",
    }},
  {0xbfdce89a, { // "b = reorder"
   "b = kolejność",
   "b = yeniden sırala",
   "b = změň uspořádání",
   "b = порядок",
   "b = neu anordnen",
/*MISSING*/    "b = reorder",
    }},
  {0xbfe4ef66, { // "You are %1 cells away from the starting point, or the place where you used an Orb of Safety last time. There are %2 such cells.\n"
   "Jesteś %1 kroków od punktu startu, lub miejsca ostatniego użycia Sfery Bezpieczeństwa. Takich pól jest %2.\n",
/*MISSING*/    "You are %1 cells away from the starting point, or the place where you used an Orb of Safety last time. There are %2 such cells.\n",
   "Nacházíš se %1 políèek od výchozího bodu nebo od místa, kde jsi naposledy použil Sféru Bezpeènosti. Takových políèek je celkem %2.\n",
   "Ты на расстоянии %1 клеток от начала пути или от того места, где ты в последний раз использовал Сферу безопасности. Всего таких клеток %2.\n",
/*MISSING*/    "You are %1 cells away from the starting point, or the place where you used an Orb of Safety last time. There are %2 such cells.\n",
/*MISSING*/    "You are %1 cells away from the starting point, or the place where you used an Orb of Safety last time. There are %2 such cells.\n",
    }},
  {0xc01131cf, { // "This false prince is immune to the alluring scent of roses."
   "Ten fałszywy książę jest niewrażliwy na zapach róż.",
/*MISSING*/    "This false prince is immune to the alluring scent of roses.",
   "Tento falešný princ je imunní vůči vábivé vůni růží.",
   "Этот ложный принц невосприимчив к ароматам роз.",
   "Dieser falsche Prinz ist immun gegenüber dem verlockenden Duft der Rosen.",
/*MISSING*/    "This false prince is immune to the alluring scent of roses.",
    }},
  {0xc03bfc26, { // "Teleported to %1!"
   "Przeniesiono w %abl1",
   "%loc1 e ışınlandın.",
   "Teleportova%l0 ses; nyní jsi %abl1",
   "Вы телепортированы %abl1",
   "Nach %1 teleportiert",
   "Teleportado para %na1",
    }},
  {0xc065f905, { // "Extremely narrow Crossroads layout.\n"
   "Skrajnie wąski układ Skrzyżowań.",
/*MISSING*/    "Extremely narrow Crossroads layout.\n",
   "Velmi úzká verze Křižovatky.",
   "Очень узкий вариант Перекрёстка.",
/*MISSING*/    "Extremely narrow Crossroads layout.\n",
/*MISSING*/    "Extremely narrow Crossroads layout.\n",
    }},
  {0xc0eb5e0d, { // "\n\nIn the Orb Strategy Mode, the Orb of Luck also significantly increases the frequency of Great Walls, Crossroads IV, and sub-lands."
   "\n\nW trybie strategii sfer Sfera Szczęścia dodatkowo znacznie zwiększa częstotliwość wielkich ścian, Skrzyżowań IV, i podkrain.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, the Orb of Luck also significantly increases the frequency of Great Walls, Crossroads IV, and sub-lands.",
   "\n\nVe sférostrategickém módu Sféra Štìstí také výraznì zvyšuje frekvenci Velkých zdí, Køižovatek IV a podkrajù.",
   "\n\nВ режиме стратегии сфер Сфера удачи также заметно увеличивает частоту Великих стен, Перекрёстков IV и под-земель.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, the Orb of Luck also significantly increases the frequency of Great Walls, Crossroads IV, and sub-lands.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, the Orb of Luck also significantly increases the frequency of Great Walls, Crossroads IV, and sub-lands.",
    }},
  {0xc123b9e6, { // "A big and quite intelligent monster living in the Icy Land."
   "Duża i całkiem inteligentna bestia z Krainy Lodu.",
   "Buz Diyarlarında yaşayan büyük ve epey zeki bir canavar.",
   "Velký a dosti inteligentní netvor, který žije v Ledovém kraji.",
   "Большой и довольно умный монстр из Ледяной земли.",
   "Ein großes und äußerst intelligentes Monster das im Eisigem Land lebt.",
/*MISSING*/    "A big and quite intelligent monster living in the Icy Land.",
    }},
  {0xc1405e48, { // "You need to stand next to a magic mirror or cloud to use %the1."
   "Musisz stać przy magicznym lustrze, by odbić %a1.",
/*MISSING*/    "You need to stand next to a magic mirror or cloud to use %the1.",
   "Abys mohl použít %a1, musíš stát vedle kouzelného zrcadla nebo oblaku.",
   "Ты должен стоять рядом с магическим зеркалом или облаком, чтобы использовать %a1.",
/*MISSING*/    "You need to stand next to a magic mirror or cloud to use %the1.",
/*MISSING*/    "You need to stand next to a magic mirror or cloud to use %the1.",
    }},
  {0xc1661595, { // "How long should the messages stay on the screen."
   "Jak długo wiadomości zostają na ekranie",
/*MISSING*/    "How long should the messages stay on the screen.",
   "Jak dlouho zùstanou zprávy na obrazovce",
   "Jak długo wiadomości zostają na ekranie",
/*MISSING*/    "How long should the messages stay on the screen.",
/*MISSING*/    "How long should the messages stay on the screen.",
    }},
  {0xc167cfca, { // "down"
   "dół",
/*MISSING*/    "down",
   "dolů",
   "вниз",
   "runter",
/*MISSING*/    "down",
    }},
  {0xc1731760, { // "You take the Illusion away."
   "Zabierasz iluzję!",
   "İllüzyonu yok ettin.",
   "Zrušil jsi iluzi.",
   "Вы забрали иллюзию!",
   "Du entfernst die Illusion.",
/*MISSING*/    "You take the Illusion away.",
    }},
  {0xc1916916, { // "This is what Hyperbug Queens eat. Very tasty and healthy."
   "Pożywienie Królowej Hiperinsektów, bardzo smaczne i zdrowe.",
   "Bu Aşkınböcek Kraliçelerinin yediği şey. Çok lezzetli ve sağlıklı.",
   "Tohle je potrava, kterou jedí Královny hyperhmyzáků. Velmi chutná a zdravá.",
   "Это - еда королевы гипержуков. Очень вкусно и полезно.",
   "Dieses Gelee wird von Hyperkäfer-Königinnen gegessen. Sehr schmackhaft und gesund.",
/*MISSING*/    "This is what Hyperbug Queens eat. Very tasty and healthy.",
    }},
  {0xc1a86dba, { // "Hyperstone Quest: collect at least %3 %1 in %the2"
   "Misja alternatywna: znajdź co najmniej %3 skarbów %abl2",
   "Aşkıntaş Görevi: %2'de en az %3 hazine topla.",
   "Hyperkamový úkol: sesbírej nejméně %3 pokladů %abl2",
   "Миссия Гиперкамня: соберите %3 драгоценностей в %abl2",
   "Hyperstein-Herausforderung: Sammle %3 %P1 %a2",
   "Missão da Pedra do Ímpeto: colete pelo menos %3 %1 no-a %abl1",
    }},
  {0xc1d694c6, { // "This monster leaves a trail of fire behind."
   "Ten żywiołak zostawia za sobą ścianę ognia.",
   "Bu canavar arkasında ateşten bir iz bırakıyor.",
   "Tento netvor za sebou zanechává ohnivou stopu.",
   "Этот Элементаль оставляет за собой след Огня.",
   "Dieses Monster hinterlässt einen brennenden Pfad.",
/*MISSING*/    "This monster leaves a trail of fire behind.",
    }},
  {0xc22de0db, { // "This would only move you deeper into the trap!"
   "To tylko przeniesie Cię w głąb pułapki!",
/*MISSING*/    "This would only move you deeper into the trap!",
   "Tím by ses jen dostal dál do pasti!",
   "Это только уведёт тебя глубже в ловушку!",
/*MISSING*/    "This would only move you deeper into the trap!",
/*MISSING*/    "This would only move you deeper into the trap!",
    }},
  {0xc23cf11e, { // "Apparently a princess is kept locked somewhere, but you won't ever find her in this hyperbolic palace. "
   "Gdzieś tu jest uwięziona księżniczka, ale w tym hiperbolicznym pałacu nigdy jej nie znajdziesz. ",
   "Görünen o ki, bir prenses bir yerlerde saklı tutuluyor, ama bu hiperbolik sarayda onu bulman mümkün değil. ",
   "Někde je tu prý uvězněná princezna, ale tu v tomhle hyperbolickém paláci nikdy nenajdeš. ",
   "Видимо, принцесса где-то заперта, но в гиперболическом дворце Вы её не найдёте. ",
   "Scheinbar ist irgendwo eine Prinzessin eingesperrt, Aber du wirst sie in diesem hyperbolischem Palast nie finden. ",
/*MISSING*/    "Apparently a princess is kept locked somewhere, but you won't ever find her in this hyperbolic palace. ",
    }},
  {0xc24ac2cd, { // "Scores retrieved."
   "Wyniki ściągnięte.",
   "Puanlar yüklendi.",
   "Skóre stažena.",
   "Результаты получены.",
   "Rekorde empfangen.",
/*MISSING*/    "Scores retrieved.",
    }},
  {0xc24cb3b5, { // "conformal/history mode"
   "tryb konforemny/historii",
/*MISSING*/    "conformal/history mode",
   "konformální mód / historie",
   "режим конформный/истории",
   "konformer/historischer Modus",
/*MISSING*/    "conformal/history mode",
    }},
  {0xc298f3eb, { // "Hyperstone Quest: collect at least %3 points in %the2"
   "Misja Hiperkamień: zdobądź co najmniej %3 punktów %abl2",
/*MISSING*/    "Hyperstone Quest: collect at least %3 points in %the2",
   "Hyperkamový úkol: získej nejméně %3 bodů %abl2",
   "Миссия Гиперкамней: собери не меньше %3 очков %abl2",
   "Hyperstein-Herausforderung: sammle mindestens %3 Punkte %abl2",
/*MISSING*/    "Hyperstone Quest: collect at least %3 points in %the2",
    }},
  {0xc29bc729, { // "Which orb to use?"
   "Której Sfery użyć?",
/*MISSING*/    "Which orb to use?",
   "Kterou Sféru chceš použít?",
   "Какую сферу использовать?",
/*MISSING*/    "Which orb to use?",
/*MISSING*/    "Which orb to use?",
    }},
  {0xc2b0798a, { // "big triangles: rings"
   "duże trójkąty: pierścenie",
/*MISSING*/    "big triangles: rings",
   "velké trojúhelníky: prstence",
   "большие треугольники: кольца",
/*MISSING*/    "big triangles: rings",
/*MISSING*/    "big triangles: rings",
    }},
  {0xc2d58d16, { // "already mirrored"
   "już było odbijane",
/*MISSING*/    "already mirrored",
   "již zrcadleno",
   "уже была отражена",
/*MISSING*/    "already mirrored",
/*MISSING*/    "already mirrored",
    }},
  {0xc333fdc3, { // "\"I hate roses.\""
   "\"Nienawidzę róż.\"",
/*MISSING*/    "\"I hate roses.\"",
   "\"Nenávidím růže.\"",
   "\"Ненавижу розы.\"",
/*MISSING*/    "\"I hate roses.\"",
/*MISSING*/    "\"I hate roses.\"",
    }},
  {0xc34d6cef, { // "The next slide shows the number of cells in distance 1, 2, 3, ... from you. It grows exponentially: there are more than 10^100 cells in radius 1000 around you, and you will move further away during the game!\n\nThis is extremely important in the design of HyperRogue. HyperRogue has many navigational puzzles -- what would be simple in Euclidean world is extremely tricky in hyperbolic geometry (you want to reach a specific location 20 cells away, which of the thousands of possible directions should you take?); however, other things virtually impossible in Euclidean world become easy in HyperRogue. HyperRogue had to be specially designed so that it is impossible to grind the infinite world. There are almost no permanent upgrades; collecting treasures brings you benefits, but trying to get too many of the same kind is extremely dangerous."
   "Kolejny slajd pokazuje liczbę pól w odległości 1, 2, 3, ... od Ciebie. Rośnie ona wykładniczo: jest ponad 10^100 pól w promieniu 1000 od Ciebie, i w trakcie gry będziesz odchodzić dalej!\n\nJest to bardzo ważne dla designu HyperRogue. Są tu zagadki nawigacyjne -- coś co by było proste w świecie euklidesowej wymaga przemyślenia w geometrii hiperbolicznej (chcesz dostać się do miejsca 20 pól od Ciebie, który z tysiąca możliwych kierunków wybrać?); z kolei inne rzeczy stają się tu łatwe. Gra HyperRogue jest zaprojektowana tak, by nie dało się 'grindować' nieskończonego świata -- prawie nie ma permanentnych ulepszeń, zbieranie skarbów przynosi pożytki, ale zdobycie wielu jednego rodzaju jest skrajnie niebezpieczne.",
/*MISSING*/    "The next slide shows the number of cells in distance 1, 2, 3, ... from you. It grows exponentially: there are more than 10^100 cells in radius 1000 around you, and you will move further away during the game!\n\nThis is extremely important in the design of HyperRogue. HyperRogue has many navigational puzzles -- what would be simple in Euclidean world is extremely tricky in hyperbolic geometry (you want to reach a specific location 20 cells away, which of the thousands of possible directions should you take?); however, other things virtually impossible in Euclidean world become easy in HyperRogue. HyperRogue had to be specially designed so that it is impossible to grind the infinite world. There are almost no permanent upgrades; collecting treasures brings you benefits, but trying to get too many of the same kind is extremely dangerous.",
   "Další snímek ukazuje počet políček ve vzdálenosti 1, 2, 3, ... od tebe. Roste exponenciálně: v kruhu o poloměru 1000 kolem tebe je více než 10^100 políček, a během hry dojdeš ještě dál!\n\nTento fakt je pro design HyperRogue velmi důležitý. Hra obsahuje mnoho navigačních hádanek -- to, co by v eukleidovském světě bylo jednoduché, je v hyperbolické geometrii extrémně složité (chceš se dostat na určité místo 20 políček daleko -- kterým z tisíců možných směrů se máš vydat?); na druhou stranu, jiné věci, které jsou v eukleidovském světě prakticky nemožné, jsou v HyperRogue snadné. HyperRogue musel být speciálně navržen tak, aby se v jeho nekonečném světě nedalo grindovat. Nejsou v něm téměř žádné trvalé upgrady; sbírání pokladů ti přináší výhody, ale pokud se pokusíš získat příliš mnoho pokladů jednoho druhu, bude to nesmírně nebezpečné.",
   "Следующий слайд покажет количество клеток на расстоянии 1, 2, 3, ... от тебя. Оно растёт экспоненциально: в радиусе 1000 клеток от тебя находится более 10^100 клеток, а ты за время игры уйдёшь гораздо дальше!\n\nЭто очень важно для HyperRogue. Здесь есть много загадок навигации -- то, что очень просто в евклидовой геометрии, становится невероятно сложным в гиперболической геометрии (ты хочешь попасть в определённую клетку на расстоянии 20 от тебя, какое из тысяч направлений выбрать?); однако другие вещи, практически невозможные в евклидовом мире, здесь очень просты. Игра HyperRogue сделана так, чтобы было невозможно просто захватывать бесконечный мир -- практически нет постоянных улучшений, сбор сокровищ даёт преимущества, но собирать много сокровищ одного вида весьма опасно.",
/*MISSING*/    "The next slide shows the number of cells in distance 1, 2, 3, ... from you. It grows exponentially: there are more than 10^100 cells in radius 1000 around you, and you will move further away during the game!\n\nThis is extremely important in the design of HyperRogue. HyperRogue has many navigational puzzles -- what would be simple in Euclidean world is extremely tricky in hyperbolic geometry (you want to reach a specific location 20 cells away, which of the thousands of possible directions should you take?); however, other things virtually impossible in Euclidean world become easy in HyperRogue. HyperRogue had to be specially designed so that it is impossible to grind the infinite world. There are almost no permanent upgrades; collecting treasures brings you benefits, but trying to get too many of the same kind is extremely dangerous.",
/*MISSING*/    "The next slide shows the number of cells in distance 1, 2, 3, ... from you. It grows exponentially: there are more than 10^100 cells in radius 1000 around you, and you will move further away during the game!\n\nThis is extremely important in the design of HyperRogue. HyperRogue has many navigational puzzles -- what would be simple in Euclidean world is extremely tricky in hyperbolic geometry (you want to reach a specific location 20 cells away, which of the thousands of possible directions should you take?); however, other things virtually impossible in Euclidean world become easy in HyperRogue. HyperRogue had to be specially designed so that it is impossible to grind the infinite world. There are almost no permanent upgrades; collecting treasures brings you benefits, but trying to get too many of the same kind is extremely dangerous.",
    }},
  {0xc370670a, { // "Collect %1 $$$ to access even more lands"
   "Znajdź %1 $$$ by iść do kolejnych krain",
   "Daha da fazla diyara ulaşmak için %1 $$$ toplayın.",
   "Sesbíráním %1 $$$ získáš přístup do dalších krajů",
   "Соберите %1 $$$, чтобы открыть новые земли",
   "Sammle %1 $$$ um noch mehr Länder zu besuchen.",
/*MISSING*/    "Collect %1 $$$ to access even more lands",
    }},
  {0xc3cbae01, { // "draw the heptagons darker"
   "siedmiokąty ciemniejsze",
   "yedigenleri daha koyu çiz",
   "tmavší sedmiúhelníky",
   "семиугольники темнее",
   "Heptagone verdunkeln",
/*MISSING*/    "draw the heptagons darker",
    }},
  {0xc3ee86d1, { // "reset the special game modes"
   "zresetuj specjalne tryby gry",
/*MISSING*/    "reset the special game modes",
   "vyresetovat speciální herní módy",
   "сбросить специальные режимы",
/*MISSING*/    "reset the special game modes",
/*MISSING*/    "reset the special game modes",
    }},
  {0xc406717d, { // "player 4 X"
   "gracz 4 X",
/*MISSING*/    "player 4 X",
   "hráč 4 X",
   "игрок 4 X",
   "Spieler 4 X",
/*MISSING*/    "player 4 X",
    }},
  {0xc406717e, { // "player 4 Y"
   "gracz 4 Y",
/*MISSING*/    "player 4 Y",
   "hráč 4 Y",
   "игрок 4 Y",
   "Spieler 4 Y",
/*MISSING*/    "player 4 Y",
    }},
  {0xc4c15ce0, { // "firewall lines: Power"
   "linie ścian ognia: Moc",
/*MISSING*/    "firewall lines: Power",
   "čáry ohnivých stěn: Moc",
   "стены огня: Сила",
/*MISSING*/    "firewall lines: Power",
/*MISSING*/    "firewall lines: Power",
    }},
  {0xc4c1e40c, { // "F/;/Space/Enter/KP5 = fire, WASD/IJKL/Numpad = move"
   "F/;/Space/Enter/KP5 = strzał, WASD/IJKL/Numpad = ruch",
   "F/;/Space/Enter/KP5 = strzał, WASD/IJKL/Numpad = ruch",
   "F/;/Space/Enter/KP5 = střelba, WASD/IJKL/Numpad = pohyb",
   "F/;/Space/Enter/KP5 = стрелять, WASD/IJKL/Numpad = ходить",
   "F/[;]/[LEER]/[ENTER]/[Numpad 5] = schießen, WASD/IJKL/[Numpad] = bewegen",
/*MISSING*/    "F/;/Space/Enter/KP5 = fire, WASD/IJKL/Numpad = move",
    }},
  {0xc4cb4c3f, { // "0 = symmetry"
   "0 = symetria",
   "0 = simetri",
   "0 = symetrie",
   "0 = симметрия",
   "0 = Symmetrie",
/*MISSING*/    "0 = symmetry",
    }},
  {0xc5882c72, { // "monster"
   "potwór",
   "canavar",
   "netvor",
   "монстр",
   "Monster",
/*MISSING*/    "monster",
    }},
  {0xc58bcb69, { // "Ice Walls melt after some time has passed."
   "Lodowe ściany topią się po pewnym czasie.",
   "Buz Duvarları biraz zaman geçince erirler.",
   "Ledové zdi časem tají.",
   "Ледяные стены со временем могут таять.",
   "Eiswände schmelzen nach gewisser Zeit.",
/*MISSING*/    "Ice Walls melt after some time has passed.",
    }},
  {0xc599df67, { // "Want to understand the geometry in HyperRogue? Try the Tutorial!"
   "Chcesz zrozumieć geometrię w HyperRogue? Obejrzyj Podręcznik!",
/*MISSING*/    "Want to understand the geometry in HyperRogue? Try the Tutorial!",
   "Chceš porozumìt geometrii v HyperRogue? Zkus Tutoriál!",
   "Хочешь понять геометрию в HyperRogue? Воспользуйся Руководством!",
/*MISSING*/    "Want to understand the geometry in HyperRogue? Try the Tutorial!",
/*MISSING*/    "Want to understand the geometry in HyperRogue? Try the Tutorial!",
    }},
  {0xc5cc6605, { // "Zebra quotient"
   "przestrzeń ilorazowa zebry",
/*MISSING*/    "Zebra quotient",
   "kvocient Zebra",
   "фактор Зебры",
/*MISSING*/    "Zebra quotient",
/*MISSING*/    "Zebra quotient",
    }},
  {0xc6366bc3, { // "player 4 go"
   "gracz 4 idź",
/*MISSING*/    "player 4 go",
   "hráč 4 pohyb",
   "игрок 4 идёт",
   "Spieler 4 los",
/*MISSING*/    "player 4 go",
    }},
  {0xc644756c, { // "Green pressure plates open all gates in radius 3 around them. Similarly, red pressure plates close all gates in this radius."
   "Zielone płyty otwierają wszystkie bramy w promieniu 3 wokół nich. Podobnie, czerwone płyty zamykają wszystkie bramy w tym promieniu.",
   "Yeşil baskı levhaları 3 uzaklıktaki bütün kapıları açar. Aynı şekilde, kırmızı baskı levhaları bu komşuluktaki kapıları kapatır.",
   "Zelené desky otevřou všechny brány v okruhu 3 polí, když na ně někdo šlápne. Červené desky naopak všechny brány v tomto okruhu zavřou.",
   "Зелёная плита открывает все проходы в радиусе 3 клеток. Аналогично, красная плита закрывает проходы в том же радиусе.",
   "Grüne Druckplatten öffnen alle Tore in einem Radius von 3. Rote Druckplatten hingegen schließen die Tore.",
/*MISSING*/    "Green pressure plates open all gates in radius 3 around them. Similarly, red pressure plates close all gates in this radius.",
    }},
  {0xc6bc4da8, { // "Yendor Challenge"
   "Misja Yendor",
   "Yendor Ek Görevi",
   "Mise: Yendor",
   "Миссия Йендора",
   "Yendor Herausforderung",
/*MISSING*/    "Yendor Challenge",
    }},
  {0xc6e3d53a, { // "Hub Land: orbs appear here if unlocked in their native land"
   "Centrum: sfery pojawiają się tu przy 10 skarbach",
   "Merkez: Ana diyarında aktifleştirilen küreler burada gözükür.",
   "Centrální kraj: Sféry se zde objevují, pokud jsi je odemkl v jejich rodné krajině",
   "Центр: сфера появвляется здесь, когда она разблокировна в своей родной земле",
   "Zentrum: Orbs erscheinen hier wenn sie in ihrem Heimatland freigeschaltet wurden",
/*MISSING*/    "Hub Land: orbs appear here if unlocked in their native land",
    }},
  {0xc70ad7ca, { // "%The1 is protected from this kind of magic!"
   "Tego typu magia nie działa na %1!",
/*MISSING*/    "%The1 is protected from this kind of magic!",
   "%1 má ochranu proti tomuto typu magie!",
   "%1 защищён от этого вида магии!",
/*MISSING*/    "%The1 is protected from this kind of magic!",
/*MISSING*/    "%The1 is protected from this kind of magic!",
    }},
  {0xc71c7d9f, { // "Unlock the Orb of Yendor!"
   "Otwórz Sferę Yendoru!",
   "Yendor'un küresinin kilidini aç!",
   "Odemkni Yendorskou sféru!",
   "Откройте сферу Йендора!",
   "Öffne den Orb von Yendor!",
   "Desbloqueie o Orbe de Yendor!",
    }},
  {0xc7362d82, { // "easy"
   "łatwe",
   "kolay",
   "snadná",
   "легко",
   "einfach",
/*MISSING*/    "easy",
    }},
  {0xc743921a, { // "Galápagos is the land of Tortoises. They are very slow, which allows the Dragons to pick on them by stealing and eating their young. Bring the Baby Tortoises back, but, there is a catch: the Tortoises come in many varieties, depending on the part of Galápagos they live in -- there are 21 binary environmental factors, and thus 2097152 varieties. You'll have to find a Tortoise which matches the baby exactly!\n\nTortoises move each 3 turns, and attacks only stun them.\n\nBringing back a Baby Tortoise counts as 5 $$$. The more factors agree in the given location of Galápagos, the brighter it is shown on your screen."
   "Galápagos to kraina Żółwi. Żółwie są bardzo powolne, co pozwala Smokom znęcac się nad nimi -- kradnąc i zjadając ich młode. Zwróć Żółwiki ich rodzinom, ale jest jeden haczyk: w różnych częściach Galápagos żyje wiele odmian Żółwi -- jest 21 binarnych czynników środowiskowych, co prowadzi do 2097152 odmian. Musisz znaleźć Żółwia dokładnie pasującego do Żółwika!\n\nŻówie ruszają się co trzy kolejki. Ataki powodują jedynie ogłuszenie.\n\nZwrócenie Żółwika liczy się jako 5 $$$. Im więcej czynników zgadza się w danym miejscu, tym jaśniej jest ono narysowane na Twoim ekranie.",
/*MISSING*/    "Galápagos is the land of Tortoises. They are very slow, which allows the Dragons to pick on them by stealing and eating their young. Bring the Baby Tortoises back, but, there is a catch: the Tortoises come in many varieties, depending on the part of Galápagos they live in -- there are 21 binary environmental factors, and thus 2097152 varieties. You'll have to find a Tortoise which matches the baby exactly!\n\nTortoises move each 3 turns, and attacks only stun them.\n\nBringing back a Baby Tortoise counts as 5 $$$. The more factors agree in the given location of Galápagos, the brighter it is shown on your screen.",
   "Galapágy jsou zemí Želv. Želvy jsou velice pomalé, takže si na ně mohou dovolovat Draci, kteří kradou a požírají jejich mláďata. Ty jim můžeš jejich Želvičky přinést zpátky, ale je tu jeden háček: existuje mnoho typů Želv podle toho, ve které části Galapág žijí -- je tu celkem 21 binárních faktorů prostředí, a tedy 2 097 152 variant. Musíš najít takovou Želvu, která přesně odpovídá tvé Želvičce!\n\nŽelvy se pohybují vždy jednou za 3 kola a útoky je mohou pouze omráčit.\n\nVrácení Želvičky se počítá za 5 $$$. Čím více faktorů se v dané části Galapág shoduje, tím jasněji tato část vypadá na obrazovce.",
   "Галапагосы -- край черепах. Черепахи очень медленны, что позволяет драконам издеваться над ними -- красть и поедать их детей. Отведи черепашку на родину! Но есть проблема: здесь есть много разновидностей черепах, зависящих от места проживания -- есть 21 различных параметров, что даёт 2097152 вариантов. Найдите черепаху, точно соответствующую этому малышу!\n\nЧерепахи движутся каждый третий ход. Атаки только оглушают их.\n\nКаждый приведённый черепашонок считается за 5 $$$. Чем больше параметров совпадают в данной местности, тем светлее она показывается.",
   "Galápagos ist das Land der Schildkröten -- Sie sind sehr langsam, was es den Drachen ermöglicht sie zu schikanieren, indem sie die jungen Schilkröten stehlen oder fressen. Bring die Babyschildkröte zurück, aber es gibt einen Haken: es gibt viele Schildkrötenarten, abhängig von der Gegend Galápagos in der sie leben -- es gibt 21 binäre regionale Faktoren und somit 2097152 verschiedene Arten. Du musst eine Schildkröte finden, die exakt mit der Babyschildkröte übereinstimmt!\n\nSchildkröten bewegen sich alle 3 Züge und Angriffe betäuben sie nur.\n\nEine Babyschildkröte zurückzubringen zählt 5 $$$. Je mehr Faktoren in der aktuellen Region Galápagos übereinstimmen, umso heller erscheint sie auf dem Bildschirm.",
/*MISSING*/    "Galápagos is the land of Tortoises. They are very slow, which allows the Dragons to pick on them by stealing and eating their young. Bring the Baby Tortoises back, but, there is a catch: the Tortoises come in many varieties, depending on the part of Galápagos they live in -- there are 21 binary environmental factors, and thus 2097152 varieties. You'll have to find a Tortoise which matches the baby exactly!\n\nTortoises move each 3 turns, and attacks only stun them.\n\nBringing back a Baby Tortoise counts as 5 $$$. The more factors agree in the given location of Galápagos, the brighter it is shown on your screen.",
    }},
  {0xc829b82e, { // "monsters killed: %1"
   "pokonane potwory: %1",
   "öldürülen canavarlar: %1",
   "netvorů zabito: %1",
   "монстров убито: %1",
   "Kills: %1",
   "monstros mortos: %1",
    }},
  {0xc83f16e9, { // "display only hexagons"
   "pokaż tylko sześciokąty",
   "sadece altıgenleri göster",
   "zobraz pouze šestiúhelníky",
   "показать только шестиугольники",
   "Nur Hexagone anzeigen",
/*MISSING*/    "display only hexagons",
    }},
  {0xc864fee0, { // "(press v) and change the wall/monster mode to ASCII, or change the resolution."
   "(naciśnij 'v') i zmień tryb ścian/potworów na ASCII albo zmień rozdzielczość.",
   "('v'ye bas) ve duvar ve canavar modunu ASCII'ye çevir veya çözünnürlüğü değiştir.",
   "(stiskni 'v') a změň mód zdí/netvorů na ASCII, nebo změň rozlišení.",
   "(нажмите 'v') и поменяйте режим отображения стен/монстров на ASCII, или измените разрешение. ",
   "(drücke V) um den 'Wand/Monster'-Modus auf ASCII zu schalten, oder die  Auflösung zu ändern.",
   "(pressione v) e mude o modo parede/monstro para ASCII, ou troque a resolução.",
    }},
  {0xc89e01f8, { // "Parrots feel safe in the forests of Caribbean, so they never leave them. But they will help the Pirates by attacking the intruders."
   "Papugi czują się bezpiecznie w lasach Karaibów i ich nie opuszczają, ale chętnie pomogą Piratom walczyć z najeźdźcami.",
   "Papağanlar Karayip Ormanlarında güvende hissederler, bu yüzden buraları hiç terketmezler. İçeriye girenlere saldırarak korsanlara da yardım etmiş olurlar. ",
   "Papoušci se v karibských lesích cítí bezpečně, a tak je nikdy neopouštějí. Pomáhají ale pirátům tím, že útočí na vetřelce.",
   "Попугаи чувствуют себя в безопасности в Карибских лесах и никогда их не покидают. Но они помогают пиратам, нападая на чужаков.",
   "Papageien fühlen sich sicher in den karibischen Wäldern, daher verlassen sie diese nie. Allerdings helfen sie den Piraten beim Angreifen der Eindringlinge.",
/*MISSING*/    "Parrots feel safe in the forests of Caribbean, so they never leave them. But they will help the Pirates by attacking the intruders.",
    }},
  {0xc8a18138, { // "A gem from the Red Rock Valley."
   "Kamień szlachetny z Doliny Czerwonej Skały.",
   "Kızılkaya Vadisinden bir kızıl mücevher.",
   "Drahokam z Údolí červených skal.",
   "Камень из Долины Красных Скал.",
   "Ein Edelstein aus dem Rotfelsental.",
/*MISSING*/    "A gem from the Red Rock Valley.",
    }},
  {0xc8a94b77, { // "Better not to let your greed make you stray from your path."
   "Nie pozwól, by zgubiła Cię chciwość!",
/*MISSING*/    "Better not to let your greed make you stray from your path.",
   "Raději nedopusť, aby tě tvá chamtivost svedla z cesty!",
   "Лучше не идти на поводу у своей жадности!",
   "Lass dich besser nicht durch Gier von deinem Weg abbringen!",
/*MISSING*/    "Better not to let your greed make you stray from your path.",
    }},
  {0xc8c9f40a, { // "The gate is closing right on you! RUN!"
   "Brama leci na Ciebie! UCIEKAJ!",
   "Kapı üzerine kapanıyor! KAÇ!",
   "Brána se zavírá! UTÍKEJ!",
   "Проход закрывается! БЕГИТЕ!",
   "Das Tor über dir schließt sich! Renne!",
/*MISSING*/    "The gate is closing right on you! RUN!",
    }},
  {0xc8ceb645, { // "Causes most monsters to attack other monsters, not only you and your friends."
   "Powoduje, że większość potworów będzie też atakowała inne potwory, nie tylko Ciebie i Twoich przyjaciół.",
   "Çoğu canavarın sen ve dostların yanında diğer canavarlara da saldırmasına sebep olur.",
   "Díky této Sféře bude většina netvorů útočit nejen na tebe a na tvé kamarády, ale i na jiné netvory.",
   "Монстры начинают атаковать других монстров, а не только Вас и ваших союзников.",
   "Dieser Orb bringt die meisten Monster dazu auch andere Monster und nicht nur dich anzugreifen.",
/*MISSING*/    "Causes most monsters to attack other monsters, not only you and your friends.",
    }},
  {0xc900c1da, { // "or 'o' to see the world overview"
   "lub 'o', by obejrzeć przegląd krain",
   "veya dünya önizlemesini görmek için 'o'ya basın.",
   "nebo stiskni 'o' pro přehled krajů",
   "или 'o', чтобы посмотреть обзор",
   "oder 'o' um einen Überblick über die Welt zu sehen",
/*MISSING*/    "or 'o' to see the world overview",
    }},
  {0xc92c7dd1, { // "A coastal area, from where you can get both to the inland worlds and to the Ocean. Each turn, each cell will become water or earth, based on the majority of cells around it. Contrary to the Living Caves, this process is not affected by most items and monsters, but elementals, dead Trolls, and cells outside of the Living Fjord have a very powerful effect."
   "Nadbrzeżny rejon, z którego możesz dostać się zarówno do krain wewnątrzlądowych jak i na Ocean. W każdej kolejce, każde pole zmienia się w wodę lub ziemię w zależności od większości pól wokół. W przeciwieństwie do Żywej Jaskini na ten proces nie wpływa większość przedmiotów i stworzeń, ale żywiołaki, martwe Trolle i pola poza granicą Żywego Fiordu mają bardzo silny efekt.",
   "Sahil şeridinde bir bölge, buradan hem iç diyarlara hem de Okyanus'a geçebilirsin. Her tur, her hücre etrafındaki hücrelerin çoğunluğuna göre su veya toprak olacak. Yaşayan Mağaraların aksine, bu süreç eşyalar veya canavarlardan etkilenmez. Ancak Özütler, Ölü Troller ve Yaşayan Fiyortların dışındaki hücrelerin etkisi çok güçlüdür.",
   "Pobřežní oblast, odkud se můžeš dostat jak do vnitrozemských světů, tak do Oceánu. V každém kole se každé pole změní na vodu nebo na souš podle toho, kterého typu je většina polí kolem něj. Na rozdíl od Živoucí jeskyně nemá většina předmětů a netvorů na tento proces žádný vliv, ale elementálové, mrtví trollové a pole mimo Živoucí fjord mají velmi silný efekt.",
   "Прибрежный регион, из которого можно попасть как в наземные области, так и в Океан. Каждый ход каждая клетка становится сушей или водой, основываясь на том, каких соседей больше. В отличие от Живой пещеры этот процесс не зависит от объектов и монстров, за исключением Элементалей, мёртвых Троллей и клеток вне Живого фьорда - все они имеют очень сильный эффект.",
   "Ein Küstengebiet von dem aus du ins Inland sowie zum Ozean kommst. Jeden Zug wird aus jeder Zelle Wasser oder Erde basierend auf der Mehrzahl der Zellen um sie herum. Im Gegensatz zu den lebenden Höhlen wird dieser Prozess von den meisten Monstern und Items nicht beeinflusst, Elementare, tote Trolle und Zellen außerhalb des Fjords haben jedoch mächtige Effekte. ",
/*MISSING*/    "A coastal area, from where you can get both to the inland worlds and to the Ocean. Each turn, each cell will become water or earth, based on the majority of cells around it. Contrary to the Living Caves, this process is not affected by most items and monsters, but elementals, dead Trolls, and cells outside of the Living Fjord have a very powerful effect.",
    }},
  {0xc97ce988, { // "You pierce %the1."
   "Bodziesz %a1.",
/*MISSING*/    "You pierce %the1.",
   "Probodl jsi %a1.",
   "Вы проткнули %a1.",
/*MISSING*/    "You pierce %the1.",
/*MISSING*/    "You pierce %the1.",
    }},
  {0xc982a2f2, { // "You hear a distant squeak!"
   "Słyszysz odległy pisk!",
   "Uzaktan gelen bir ciyaklama duydun!",
   "V dálce slyšíš zapištění!",
   "Вы слышите писк!",
   "Du hörst ein Piepsen in der Ferne!",
/*MISSING*/    "You hear a distant squeak!",
    }},
  {0xc9a03358, { // "Saved the band image as: "
   "Zapisano obraz jako: ",
/*MISSING*/    "Saved the band image as: ",
   "Pásový obrázek uložen jako: ",
   "Изображение сохранено как: ",
   "Bänderbild gespeichert als: ",
/*MISSING*/    "Saved the band image as: ",
    }},
  {0xc9ca07d4, { // "F7 = player on/off"
   "F7 = gracz wł/wył",
   "F7 = oyuncu açık/kapalı",
   "F7 = zapnout/vypnout hráče",
   "F7 = игрок вкл/выкл",
   "F7 = Spieler an/aus",
/*MISSING*/    "F7 = player on/off",
    }},
  {0xc9e54ff0, { // "collect 3 %2 to unlock it in %the1"
   "znajdź 3x %2 by udostępnić %abl1",
   "%abl1 onu aktif etmek için 3 tane %2 topla.",
   "získej 3x %2, abys ji odemkl %abl1",
   "collect 3x %2, чтобы открыть её %abl1",
   "Sammle 3x %2 um es in  %der1 freizuschalten",
/*MISSING*/    "collect 3 %2 to unlock it in %the1",
    }},
  {0xca147bdd, { // "coefficient (imaginary)"
   "współczynnik (część urojona)",
/*MISSING*/    "coefficient (imaginary)",
   "koeficient (imaginární část)",
   "коэффициент (мнимый)",
/*MISSING*/    "coefficient (imaginary)",
/*MISSING*/    "coefficient (imaginary)",
    }},
  {0xca6326b7, { // "Press Enter or F10 to save"
   "Wciśnij Enter lub F10, by zapisać i wyjść",
   "Enter ya da F10 ile kayıt et",
   "Enter nebo F10: uložit hru",
   "Нажмите Enter или F10, чтобы сохраниться",
   "Drücke Enter oder F10 zum Speichern",
   "Pressione Enter ou F10 para salvar",
    }},
  {0xca67eb2d, { // "extra graphical effects"
   "dodatkowe efekty graficzne",
/*MISSING*/    "extra graphical effects",
   "extra grafické efekty",
   "dodatkowe efekty graficzne",
/*MISSING*/    "extra graphical effects",
/*MISSING*/    "extra graphical effects",
    }},
  {0xca7a1de8, { // "Cannot create temporary matter here!"
   "Nie można tu stworzyć tymczasowej materii!",
   "Burada geçici madde oluşturamazsın!",
   "Zde nelze vytvořit dočasnou hmotu!",
   "Нельзя создать временную материю здесь!",
   "Hier kann keine temporäre Materie erzeugt werden!",
/*MISSING*/    "Cannot create temporary matter here!",
    }},
  {0xca836329, { // "\nRanged Orbs can be targeted by clicking the desired location. "
   "\nRobi się to przez kliknięcie na celu.",
/*MISSING*/    "\nRanged Orbs can be targeted by clicking the desired location. ",
   "\nTyto sféry lze zacílovat kliknutím na požadované políčko.",
   "\nВыбор цели сферы определяется щелчком мыши на цели.",
   "\nMit Distanzorbs kann per Klick auf die gewünschte Position gezielt werden. ",
/*MISSING*/    "\nRanged Orbs can be targeted by clicking the desired location. ",
    }},
  {0xcaabb2e2, { // "The game starts in the Icy Lands. Collect the Ice Diamonds (press F1 if you do not know how to move). After you collect many of them, monsters will start to pose a challenge.\nAs is typical in roguelikes and other games based on tactical skill rather than story, if you lose, you have to start a new one from the start. However, in this tutorial, you can simply press '4' to teleport away from a bad situation.In general, the tutorial is rigged to show you what it wants -- for example, in this slide, you can press '5' to get lots of Ice Diamonds quickly."
   "Gra rozpoczyna się w Lodowej Krainie. Zbieraj Lodowe Diamenty (naciśnij F1 jeśli nie wiesz jak się ruszać). Po zebraniu wielu Diamentów ataki potworów staną się wyzwaniem. Zgodnie z konwencjami gier roguelike, jeśli przegrasz, zaczynasz od początku; w tym podręczniku jednak wystarczy nacisnąć '4', by uciec ze złej sytuacji. Podręcznik jest podkręcony tak, by pokazać to co chce -- na przykład tutaj możesz nacisnąć '5' by zdobyć szybko dużo diamentów.",
/*MISSING*/    "The game starts in the Icy Lands. Collect the Ice Diamonds (press F1 if you do not know how to move). After you collect many of them, monsters will start to pose a challenge.\nAs is typical in roguelikes and other games based on tactical skill rather than story, if you lose, you have to start a new one from the start. However, in this tutorial, you can simply press '4' to teleport away from a bad situation.In general, the tutorial is rigged to show you what it wants -- for example, in this slide, you can press '5' to get lots of Ice Diamonds quickly.",
   "Hra začíná v Ledovém kraji. Sbírej Ledové diamanty (pokud nevíš, jak se pohybovat, stiskni F1). Když jich nasbíráš mnoho, netvoři začnou být nebezpeční.\nJak je zvykem ve hrách žánru roguelike a v jiných, které jsou založené na taktických dovednostech a nikoli na příběhu, pokud prohraješ, budeš muset začít hrát od začátku. V tomto tutoriálu se ale můžeš z ošklivé situace jednoduše teleportovat stisknutím klávesy '4'. Tutoriál je obecně dělaný tak, aby ti mohl snadno ukázat, co chce -- v tomto snímku můžeš například stisknout klávesu '5', abys rychle získal velké množství Ledových diamantů.",
   "Игра начинается в Ледяной земле. Собирай Ледяные алмазы (нажми F1, если не знаешь, как ходить). Когда ты соберёшь много алмазов, монстры начнут становиться проблемой. Как и в других играх жанра roguelike, после проигрыша ты начнёшь играть с начала; но в пределах этого руководства ты можешь нажать '4', чтобы телепортироваться из плохой ситуации. Руководство устроено так, чтобы показать тебе игру -- например, на этом слайде ты можешь нажать '5', чтобы получить сразу много алмазов.",
/*MISSING*/    "The game starts in the Icy Lands. Collect the Ice Diamonds (press F1 if you do not know how to move). After you collect many of them, monsters will start to pose a challenge.\nAs is typical in roguelikes and other games based on tactical skill rather than story, if you lose, you have to start a new one from the start. However, in this tutorial, you can simply press '4' to teleport away from a bad situation.In general, the tutorial is rigged to show you what it wants -- for example, in this slide, you can press '5' to get lots of Ice Diamonds quickly.",
/*MISSING*/    "The game starts in the Icy Lands. Collect the Ice Diamonds (press F1 if you do not know how to move). After you collect many of them, monsters will start to pose a challenge.\nAs is typical in roguelikes and other games based on tactical skill rather than story, if you lose, you have to start a new one from the start. However, in this tutorial, you can simply press '4' to teleport away from a bad situation.In general, the tutorial is rigged to show you what it wants -- for example, in this slide, you can press '5' to get lots of Ice Diamonds quickly.",
    }},
  {0xcabd677d, { // "\n\nThis is a part of a monster. It does not count for your total kills."
   "\n\nTo jest część potwora. Nie liczy się do łącznych zabić.",
/*MISSING*/    "\n\nThis is a part of a monster. It does not count for your total kills.",
   "\n\nToto je část netvora. Nepočítá se do celkového počtu zabitých netvorů.",
   "\n\nЭто лишь часть монстра. Не учитывается в количестве убийств.",
   "\n\nDas ist ein Teil eines Monsters. Es wird der Gesamtzahl deiner Tötungen nicht eingerechnet.",
/*MISSING*/    "\n\nThis is a part of a monster. It does not count for your total kills.",
    }},
  {0xcae30e00, { // "+5 = move instantly"
   "+5 = ruch natychmiastowy",
/*MISSING*/    "+5 = move instantly",
   "+5 = okamitý pohyb",
   "+5 = ruch natychmiastowy",
/*MISSING*/    "+5 = move instantly",
/*MISSING*/    "+5 = move instantly",
    }},
  {0xcafe26a4, { // "This bomberbird will follow you at some distance, and attack your enemies. You can kill it to place a mine."
   "Ten bombardier będzie latał za Tobą i atakował Twoich wrogów. Możesz też go zabić, by podłożyć minę.",
   "Bu bombacıkuş seni belli bir mesafeden takip eder ve düşmanlarına saldırır. Onu öldürerek bir mayın döşeyebilirsin.",
   "Tento Bombarďák tě bude sledovat v určité vzdálenosti a útočit na tvé nepřátele. Kromě toho ho můžeš zabít, a tím položit minu.",
   "Этот бомбардир будет следовать за Вами и атаковать Ваших врагов. Можете его убить, чтобы оставить мину.",
   "Dieser Bombenvogel folgt dir auf Distanz, und attackiert deine Feinde. Du kannst ihn töten, um eine Mine zu legen.",
/*MISSING*/    "This bomberbird will follow you at some distance, and attack your enemies. You can kill it to place a mine.",
    }},
  {0xcb093da9, { // "The axes help with keyboard movement"
   "Osie pomagają, gdy gramy na klawiaturze",
   "Eksenler klavye hareketine yardım eder.",
   "Osy pomáhají při pohybu pomocí klávesnice",
   "Оси помогают перемещаться с помощью клавиатуры",
   "Die Achsen helfen bei der Tastatursteuerung",
   "Os eixos ajudam o movimento do teclado",
    }},
  {0xcb0f63b8, { // "The higher the number, the smoother the animations in the game. If you find that animations are not smooth enough, you can try to change the options "
   "Im większa liczba, tym płynniejsze są animacje. Jeśli animacja nie jest dostatecznie płynna, spróbuj zmienić opcje",
   "Daha yüksek sayı, daha akıcı animasyonlr demektir. Eğer animasyonlar senin için yeterince akıcı değilse, ayarları değiştirmeyi düşünebilirsin",
   "Čím vyšší číslo, tím hladší budou animace ve hře. Pokud se ti zdá, že animace nejsou dostatečně hladké, můžeš zkusit změnit nastavení ",
   "Чем выше это число, тем более гладкой будет анимация в игре. Если она недостаточно гладкая, поменяйте настройки. ",
   "Je höher die Zahl, desto flüssiger sind die Animationen. Sind sie nicht flüssig genug, versuche die Optionen zu ändern ",
   "Quanto maior o número, mais suave serão as animações no jogo. Se você acha que as animações não estão suficientemente suaves, você pode tentar mudar as opções ",
    }},
  {0xcb24fed4, { // "collect 25 %2 to unlock it in %the1"
   "znajdź 25x %2 by udostępnić %abl1",
   "%abl1 onu aktif etmek için 25 tane %2 topla.",
   "získej 25x %2, abys ji odemkl %abl1",
   "collect 25x %2, чтобы открыть её %abl1",
   "Sammle 25x %2 um es in %der1 freizuschalten",
/*MISSING*/    "collect 25 %2 to unlock it in %the1",
    }},
  {0xcb577ecd, { // "Many clans of Trolls spend their lives in this kingdom. You can find many statues of Trolls here. You suppose that they are not actually statues, but simply elderly Trolls, who have petrified upon death. Or maybe you have killed these Trolls yourself?"
   "Wiele klanów Trolli żyje w tym królestwie. Można tu znaleźć mnóstwo pomników trolli, albo raczej nie pomników, tylko po prostu stare Trolle, które umierając zamieniły się kamień. A może to Trolle zabite przez Ciebie?",
/*MISSING*/    "Many clans of Trolls spend their lives in this kingdom. You can find many statues of Trolls here. You suppose that they are not actually statues, but simply elderly Trolls, who have petrified upon death. Or maybe you have killed these Trolls yourself?",
   "V tomto království tráví svůj život řada klanů Trollů. Také tu můžeš najít mnoho jejich soch. Říkáš si, že to asi nejsou sochy, ale prostě jen staří Trollové, kteří se po smrti proměnili v kámen. Nebo že by to byli Trollové, které jsi zabil ty?",
   "Много кланов Троллей жило в этом королевстве. Вы можете найти много статуй Троллей здесь. Вы предполагаете, что это не статуи, а просто старые тролли, окаменевшие после смерти. Или это вы их убили?",
   "Viele Troll-Clans verbringen ihr Leben in diesem Königreich. Du kannst hier viele Troll-Statuen finden. Du vermutest, dass es nicht wirklich Statuen sind, sondern einfach ältere Trolle, die bei ihrem Tode versteinert sind. Oder hast du diese Trolle selbst getötet?",
/*MISSING*/    "Many clans of Trolls spend their lives in this kingdom. You can find many statues of Trolls here. You suppose that they are not actually statues, but simply elderly Trolls, who have petrified upon death. Or maybe you have killed these Trolls yourself?",
    }},
  {0xcb66465f, { // "The band model is the hyperbolic analog of the Mercator projection of the sphere: a given straight line is rendered as a straight line, and the rest of the world is mapped conformally, that is, angles are not distorted. Here, we take the straight line connecting your starting point and your current position -- usually the path taken by the player is surprisingly close to a straight line. Press '8' to see this path.\n\nIf you want, press '5' to see it rendered as a spiral, although it takes lots of time and memory."
   "Model wstęgi to hiperboliczny analog rzutu Mercatora: pewna linia prosta jest rysowana jaka linia prosta, a reszta świata jest odwzorowana konforemnie, tzn. kąty nie są przekłamywane. Tutaj jako centralną linię prostą bierzemy linię łączącą punkt początkowy i Twoją obecną pozycję -- zazwyczaj droga wybrana przez gracza jest niespodzianie zbliżona do linii prostej. Naciśnij '8', by zobaczyć ścieżkę.\n\nMożesz też nacisnąć '5' by zobaczyć to w postaci spirali, ale wymaga to bardzo dużo czasu i pamięci.",
/*MISSING*/    "The band model is the hyperbolic analog of the Mercator projection of the sphere: a given straight line is rendered as a straight line, and the rest of the world is mapped conformally, that is, angles are not distorted. Here, we take the straight line connecting your starting point and your current position -- usually the path taken by the player is surprisingly close to a straight line. Press '8' to see this path.\n\nIf you want, press '5' to see it rendered as a spiral, although it takes lots of time and memory.",
   "Pásový model je hyperbolickým analogem Mercatorovy projekce na kouli: jedna daná přímka se zobrazí jako přímka a zbytek světa je namapován konformálně, tedy bez zkreslení úhlů. V tomto případě použijeme přímku, která spojuje bod, kde jsi začal hru, a tvou současnou polohu -- dráha, po které hráč jde, je obvykle překvapivě blízká přímce. Tuto cestu můžeš vidět stisknutím klávesy '8'.\n\nPokud chceš, můžeš ji také stisknutím klávesy '5' vyrenderovat jako spirálu; vyžaduje to však hodně času a paměti.",
   "Модель ленты -- гиперболический аналог проекции Меркатора сферы: данная прямая отображается в прямую, а оставшаяся часть отображается конформно, то есть углы не искажаются. Здесь в качестве центральной прямой используется прямая, соединяющая точку старта и твою текущую позицию -- обычно твой путь оказывается неожиданно близко с прямой. Нажми '8', чтобы увидеть свой путь.\n\nЕсли хочешь, нажми '5', чтобы отрисовать спираль, хотя это займёт много времени и памяти.",
/*MISSING*/    "The band model is the hyperbolic analog of the Mercator projection of the sphere: a given straight line is rendered as a straight line, and the rest of the world is mapped conformally, that is, angles are not distorted. Here, we take the straight line connecting your starting point and your current position -- usually the path taken by the player is surprisingly close to a straight line. Press '8' to see this path.\n\nIf you want, press '5' to see it rendered as a spiral, although it takes lots of time and memory.",
/*MISSING*/    "The band model is the hyperbolic analog of the Mercator projection of the sphere: a given straight line is rendered as a straight line, and the rest of the world is mapped conformally, that is, angles are not distorted. Here, we take the straight line connecting your starting point and your current position -- usually the path taken by the player is surprisingly close to a straight line. Press '8' to see this path.\n\nIf you want, press '5' to see it rendered as a spiral, although it takes lots of time and memory.",
    }},
  {0xcb78f60d, { // "rhombic tesselation"
   "posadzka rombowa",
/*MISSING*/    "rhombic tesselation",
   "kosočtvercové dělení",
   "замощение ромбами",
/*MISSING*/    "rhombic tesselation",
/*MISSING*/    "rhombic tesselation",
    }},
  {0xcb8232b7, { // "visit the website"
   "idź na stronę HyperRogue",
/*MISSING*/    "visit the website",
   "navštiv webovou stránku",
   "перейти на страницу HyperRogue",
   "Webseite besuchen",
/*MISSING*/    "visit the website",
    }},
  {0xcbd211bb, { // "\n\nTortoises are not monsters! They are just annoyed. They do not count for your total kills."
   "\n\nŻółwie to nie potwory! Są tylko rozwścieczone. Nie liczą się do Twoich zabić.",
/*MISSING*/    "\n\nTortoises are not monsters! They are just annoyed. They do not count for your total kills.",
   "\n\nŽelvy nejsou netvoři! Jsou jenom podrážděné. Proto se nepočítají do celkového počtu zabitých nepřátel.",
   "\n\nЧерепашки -- не монстры! Хоть и раздражают. Они не считаются в количестве убийств.",
   "\n\nSchildkröten sind keine Monster! Sie sind nur verärgert. Sie werden nicht der Gesamtzahl von Tötungen angerechnet.",
/*MISSING*/    "\n\nTortoises are not monsters! They are just annoyed. They do not count for your total kills.",
    }},
  {0xcbe3ae13, { // ", you"
   ", Ty",
   ", Sen",
   ", ty",
   ", Вы",
   ", du",
   ", você",
    }},
  {0xcbf72793, { // "This part of the world is warped, restricting the movement somewhat. \"Diagonal\" movement and attacking between triangular cells is not allowed. Flash, Storms, and Freedom spells ignore this, and Ghosts can move, attack, and be attacked diagonally."
   "Ta część świata jest zakrzywiona, co powoduje zmniejszenie możliwości ruchu. Ruch i atakowanie \"po przekątnej\", czyli między trójkątami, są zabronione. Czary Błysk, Burza i Wolność ignorują to. Duchy mogą się ruszać, atakować i być atakowane po przekątnej.",
/*MISSING*/    "This part of the world is warped, restricting the movement somewhat. \"Diagonal\" movement and attacking between triangular cells is not allowed. Flash, Storms, and Freedom spells ignore this, and Ghosts can move, attack, and be attacked diagonally.",
   "Tato část světa je pokřivená, což poněkud omezuje pohyb. \"Úhlopříčný\" pohyb a útok mezi trojúhelníkovými políčky není možný. Kouzla Záblesku, Bouře a Svobody toto pravidlo ignorují a Duchové se mohou pohybovat a útočit úhlopříčně nebo být sami úhlopříčně napadeni.",
   "Это земля искривлена, и движения немного ограничены. Ходы и атаки \"по-диагонали\" между треугольниками невозможны. Заклинания Вспышки, Бури и Свободы игнорируют это. Духи также могут ходить, атаковать и быть атакованными по-диагонали.",
   "Dieser Teil der Welt ist verzerrt, was die Bewegung etwas einschränkt. \"Diagonale\" Bewegung und Angriffe zwischen dreieckigen Zellen sind nicht erlaubt. Lichtblitz-, Sturm- und Freiheitszauber ignorieren das. Geister können sich diagonal bewegen und diagonal angegriffen werden.",
/*MISSING*/    "This part of the world is warped, restricting the movement somewhat. \"Diagonal\" movement and attacking between triangular cells is not allowed. Flash, Storms, and Freedom spells ignore this, and Ghosts can move, attack, and be attacked diagonally.",
    }},
  {0xcc012b4d, { // "secondary in %the1 (collect 10 %3, or 25 %2)"
   "podrzędna %abl1 (10x %3, lub 25x %2)",
   "%abl1 ikincil olarak bulunuyor (10 tane %3, veya 25 tane %2)",
   "sekundární %abl1 (získej 10x %3 nebo 25x %2)",
   "дополнительная %abl1 (10x %3, или 25x %2)",
   "sekundär in %der1 (10x %3, oder 25x %2)",
/*MISSING*/    "secondary in %the1 (collect 10 %3, or 25 %2)",
    }},
  {0xcc095f49, { // "Follow the Mouse and escape with %the1!"
   "Idź za Myszką i ucieknij z %abl1!",
   "Fareyi takip et ve %abl1 kaç!",
   "Následuj Myšku a uteč s %abl1!",
   "Следуq за Мышкой и выйди с %abl1!",
   "Folge der Maus und entkomme mit %dem1!",
/*MISSING*/    "Follow the Mouse and escape with %the1!",
    }},
  {0xcc0fa034, { // "dog"
   "pies",
/*MISSING*/    "dog",
   "pes",
   "пёс",
   "Hund",
/*MISSING*/    "dog",
    }},
  {0xcc13ece6, { // "Ghosts can move through chasms!"
   "Duchy przelatują nad dziurami!",
/*MISSING*/    "Ghosts can move through chasms!",
   "Duchové se mohou pohybovat přes propasti!",
   "Призраки могут перелетать пропасти!",
/*MISSING*/    "Ghosts can move through chasms!",
/*MISSING*/    "Ghosts can move through chasms!",
    }},
  {0xcc3fc29f, { // "configure joystick axes"
   "skonfiguruj osie joysticka",
   "joystick eksenlerini ayarla",
   "konfigurace os joysticku",
   "настройка джойстика",
   "Joystick Achsen konfigurieren",
/*MISSING*/    "configure joystick axes",
    }},
  {0xcc49e6af, { // "The Knights laugh at your failure!"
   "Rycerze śmieją się z Twojej porażki!",
   "Şövalyeler hezimetine katıla katıla gülüyor!",
   "Rytíři se smějí tvému neúspěchu!",
   "Рыцари смеются над Вашей неудачей!",
   "Die Ritter lachen dich aus weil du versagt hast!",
/*MISSING*/    "The Knights laugh at your failure!",
    }},
  {0xccb41803, { // "These reptiles are quite strange creatures. They spend most of their lives sleeping as floors that other creatures can walk on. Sometimes they wake up to hunt their prey, but they will happily go back to sleep if they happen to move into a hole on their way. Your attacks do not kill the Reptiles, but you can push and stun them."
   "Te jaszczurki to dosyć dziwne stwory. Większość życia spędzają śpiąc, tworząc podłogi, po których można chodzić. Czasami się budzą by zapolować, ale bardzo chętnie położą się znów do snu, jeśli natrafią na swojej drodze na dziurę. Twoje ataki nie zabijają jaszczurek, ale możesz je pchać i ogłuszać.",
/*MISSING*/    "These reptiles are quite strange creatures. They spend most of their lives sleeping as floors that other creatures can walk on. Sometimes they wake up to hunt their prey, but they will happily go back to sleep if they happen to move into a hole on their way. Your attacks do not kill the Reptiles, but you can push and stun them.",
   "Tyto Ještěrky jsou opravdu podivní tvorové. Vštšinu života tráví tím, že spí ve formě podlahy, po které jiní tvorové chodí. Někdy se probudí, aby lovily kořist, ale pokud po cestě náhodou šlápnou do díry, ochotně zase usnou. Tvé útoky nemohou Ještěrky zabít, můžeš je ale strkat a omračovat.",
   "Эти ящерицы -- довольно странные существа. Они проводят большую часть жизни во сне, служа полом, по которому другие могут ходить. Иногда они просыпаются, чтобы поохотиться, но они с радостью вернутся ко сну, если найдут отверстие в полу на своём пути. Ваги атаки их не убивают, а только толкают и оглушают.",
/*MISSING*/    "These reptiles are quite strange creatures. They spend most of their lives sleeping as floors that other creatures can walk on. Sometimes they wake up to hunt their prey, but they will happily go back to sleep if they happen to move into a hole on their way. Your attacks do not kill the Reptiles, but you can push and stun them.",
/*MISSING*/    "These reptiles are quite strange creatures. They spend most of their lives sleeping as floors that other creatures can walk on. Sometimes they wake up to hunt their prey, but they will happily go back to sleep if they happen to move into a hole on their way. Your attacks do not kill the Reptiles, but you can push and stun them.",
    }},
  {0xcccdfec3, { // "You rejoin %the1."
   "Połączy%łeś0 się z %abl1.",
   "%abl1 bütünleştin.",
   "Znovu ses spoji%l0 %s1 %abl1.",
   "Вы воссоединились с %abl1.",
   "Du stößt wieder zu %den1 %a1 hinzu",
   "Você se reencontra com %na1.",
    }},
  {0xcdd4911e, { // "Euclidean"
   "euklidesowa",
/*MISSING*/    "Euclidean",
   "eukleidovská",
   "евклидова",
/*MISSING*/    "Euclidean",
/*MISSING*/    "Euclidean",
    }},
  {0xcddadaf6, { // "view your achievements"
   "pokaż osiągnięcia",
/*MISSING*/    "view your achievements",
   "tvé achievementy",
   "смотреть достижения",
   "deine Errungenschaften ansehen",
/*MISSING*/    "view your achievements",
    }},
  {0xce3381e8, { // "spin down"
   "obróć w dół",
/*MISSING*/    "spin down",
   "otoč dolů",
   "вращать вниз",
   "hochdrehen",
/*MISSING*/    "spin down",
    }},
  {0xce7f0820, { // "add credits for your translation here"
   "Teksty i tłumaczenie: Zeno i Tehora Rogue\n\n",
   "Çeviri: zulmetefza & Seyacim\n\n",
   "Český překlad: Marek Čtrnáct <marek14@seznam.cz>\n\n",
   "Перевод: green orange <khadaev98@gmail.com>\n\n",
   "Deutsche Übersetzung: Jeffrey Smith, Arriator und Phillip 'TheGons' Gonsior \n\n",
   "Tradução para Português do Brasil: Fernando Antonio <fernandojr.semrhfj@hotmail.com>\n\n",
    }},
  {0xce87854f, { // "It was a floor... until something walked on it."
   "Tu kiedyś była podłoga, ale coś po niej przeszło.",
   "Burası bir zemin parçasıydı...ta ki bir şey üzerinden geçene kadar.",
   "Tady byla podlaha... než po ní něco přešlo.",
   "Здесь был пол... пока кто-то не наступил не него.",
   "Das war einmal Boden... bis etwas darauf gegangen ist.",
/*MISSING*/    "It was a floor... until something walked on it.",
    }},
  {0xce8db891, { // "panning X"
   "przewiń X",
   "yatay kaydır",
   "scrollování ve směru X",
   "крутить X",
   "X schwenken",
/*MISSING*/    "panning X",
    }},
  {0xce8db892, { // "panning Y"
   "przewiń Y",
   "dikey kaydır",
   "scrollování ve směru Y",
   "крутить Y",
   "Y schwenken",
/*MISSING*/    "panning Y",
    }},
  {0xce9a291e, { // "configure panning and general keys"
   "skonfiguruj klawisze ogólne",
/*MISSING*/    "configure panning and general keys",
   "konfigurace pøejíždìní a obecných kláves",
   "настроить панораму и общие клавиши",
/*MISSING*/    "configure panning and general keys",
/*MISSING*/    "configure panning and general keys",
    }},
  {0xcea95257, { // "Turn-based multiplayer: Turns are executed in parallel. A player can leave the game by pressing a designated key (useful when about to get killed or lost). The following Orbs work to bring such players back: "
   "Turowo na wielu graczy: tury wykonujecie równolegle. Gracz może opuścić grę naciskając wyznaczony klawisz (użyteczne, gdy się zgubi lub by zaraz został zabity). Następujących Sfer można użyć, by taki gracz wrócił:",
/*MISSING*/    "Turn-based multiplayer: Turns are executed in parallel. A player can leave the game by pressing a designated key (useful when about to get killed or lost). The following Orbs work to bring such players back: ",
   "Hra pro více hráčů na kola: Všichni hráči odehrají své kolo současně. Hráči mohou opustit hru stisknutím specifické klávesy (hodí se, když jim hrozí, že se ztratí nebo je něco zabije). Tyto hráče je možné vrátit zpátky použitím následujících Sfér:",
   "Пошаговый мультиплеер: ходы совершаются параллельно. Игрок может покинуть игру, нажав определённую клавишу (полезно, когда почти умер или потерялся). Следующие сферы могут вернуть игроков обратно:",
   "Mehrspieler zugbasiert: Züge werden simultan ausgeführt. Ein Spieler kann das Spiel durch Drücken einer bestimmten Taste verlassen (nützlich wenn Gefahr droht getötet zu werden oder verlorenzugehen). Die folgenden Orbs können dazu verwendet werden solche Spieler zurückzubringen: ",
/*MISSING*/    "Turn-based multiplayer: Turns are executed in parallel. A player can leave the game by pressing a designated key (useful when about to get killed or lost). The following Orbs work to bring such players back: ",
    }},
  {0xcecd6064, { // "z: %1"
   "z: %1",
   "z: %1",
   "z: %1",
   "z: %1",
   "z: %1",
/*MISSING*/    "z: %1",
    }},
  {0xced30d4e, { // "%The1 attacks your shell!"
   "%1 zaatakowa%ł1 Twoją skorupę!",
/*MISSING*/    "%The1 attacks your shell!",
   "%1 zaútoči%l1 na tvůj krunýř!",
   "%1 атакует вашу раковину!",
   "%Der1 %1 attackiert deinen Panzer!",
/*MISSING*/    "%The1 attacks your shell!",
    }},
  {0xcede212a, { // "heptagonal mode"
   "tryb siedmiokątów",
/*MISSING*/    "heptagonal mode",
   "sedmiúhelníkový mód",
   "режим семиугольников",
   "heptagonaler Modus",
/*MISSING*/    "heptagonal mode",
    }},
  {0xcee13684, { // "%The1 is washed away!"
   "%2 zmy%ł2 %a1!",
   "%1 suya karışıp gitti!",
   "Elementál Vody spláchl %a1!",
   "%a1 смыло водой!",
   "%Der1 %a1 wurde weggeschwemmt!",
/*MISSING*/    "%The1 is washed away!",
    }},
  {0xcef0ea3b, { // "%The1 pierces %the2!"
   "%1 przebi%ł1 %a2!",
   "%1 %a2 delip geçiyor!",
   "%1 probod%l1 %a2!",
   "%1 проколол%E1 %a2!",
   "%Der1 durchbohrt %den2 %a2!",
/*MISSING*/    "%The1 pierces %the2!",
    }},
  {0xcf3aac29, { // "t/left/right - change display, up/down - scroll, s - sort by"
   "t/lewo/prawo - zmiana, góra/dół - przewijanie, s - sortowanie",
   "t/sol/sağ - görünümü değiştir, yukarı/aşağı - kaydır, s - sırala",
   "t/doleva/doprava - změna zobrazení, nahoru/dolů - posunování, s - řazení",
   "t/левая/правая стрелки - изменить, вверх/вниз - листать, s - сортировать",
   " T/Links/Rechts - Anzeige ändern, Oben/Unten - Scrollen, S - Sortieren nach",
   "t/esquerda/direita - mudar display, cima/baixo - rolar, s - organizar por",
    }},
  {0xcf65d6a2, { // "In the Windy Plains, you can let the wind carry you, causing you to move two cells with the wind in a single turn. This cannot be done if you are standing at distance at most 2 from the Air Elemental, or if any of the three cells on the way has two wind directions.\n\nPress 't' or click the destination to activate."
   "Na Wietrznej Równinie możesz pozwolić dać się nieść wiatrowi, dzięki czemu ruszasz się 2 pola z wiatrem w jednej kolejce. Jest to możliwe tylko, jeśli nie stoisz w odległości co najwyżej 2 od Powietrznego Żywiołaka i żadne z trzech pól po drodze nie ma wiatru w dwóch kierunkach.\n\nNaciśnij 't' albo kliknij pole docelowe, by wykonać ten ruch.",
/*MISSING*/    "In the Windy Plains, you can let the wind carry you, causing you to move two cells with the wind in a single turn. This cannot be done if you are standing at distance at most 2 from the Air Elemental, or if any of the three cells on the way has two wind directions.\n\nPress 't' or click the destination to activate.",
   "Na Větrné pláni se můžeš nechat unášet větrem a pohybovat se o dvě políčka v jednom kole. Toto není možné, pokud se nacházíš ve vzdálenosti 2 nebo méně od Elementála vzduchu nebo pokud jsou na kterémkoli ze tří políček, kterými bys prošel, dva směry větru.\n\nPro unášení větrem stiskni klávesu 't' nebo klikni na cílové políčko.",
   "На Ветреной равнине Вы можете позволить ветру унести Вас, подвинувшись сразу на 2 клетки за 1 ход. Это невозможно, если Вы стоите на расстоянии не более 2 от Воздушного элементаля, или на одной из проходимых клеток более одного направления ветра.\n\nНажмите 't' или кликните на клетку, чтобы это использовать.",
   "In den windigen Ebenen kannst du dich vom Wind tragen lassen, dann bewegst du dich mit dem Wind zwei Zellen in einem Zug. Dies gilt nicht, wenn du maximal 2 Zellen von einem Luftelementar entfernt bist, oder eine der drei Zellen in deinem Weg zwei Windrichtungen hat.\n\nDrücke 't' oder klicke auf dein Ziel zur Aktivierung.",
/*MISSING*/    "In the Windy Plains, you can let the wind carry you, causing you to move two cells with the wind in a single turn. This cannot be done if you are standing at distance at most 2 from the Air Elemental, or if any of the three cells on the way has two wind directions.\n\nPress 't' or click the destination to activate.",
    }},
  {0xcfa35fb1, { // "The islands of Caribbean are infinite and circular, and the pirates probably have hidden their treasures somewhere far from the coast."
   "Wyspy na Karaibach są okrągłe i nieskończone... Skarby piratów są pewnie schowane gdzieś daleko od brzegu.",
   "Karayip Adaları sonsuz ve dairesel. Korsanlar hazinelerini büyük ihtimalle kıyılardan uzak yerlere saklıyordur.",
   "Karibské ostrovy jsou nekonečné a kulaté a piráti pravděpodobně ukryli své poklady kdesi daleko od pobřeží.",
   "Карибские острова бесконечные и круглые. Возможно, пираты прячут свои сокровища где-то вдали от берега.",
   "Die Inseln der Karibik sind unzählig und kreisförmig, die Piraten haben ihre Schätze sicher fernab der Küste versteckt.",
/*MISSING*/    "The islands of Caribbean are infinite and circular, and the pirates probably have hidden their treasures somewhere far from the coast.",
    }},
  {0xcfc5e0f7, { // "gain Orb of the Sword"
   "zdobądź Sferę Ostrza",
/*MISSING*/    "gain Orb of the Sword",
   "získej Sféru meče",
   "получи Сферу меча",
/*MISSING*/    "gain Orb of the Sword",
/*MISSING*/    "gain Orb of the Sword",
    }},
  {0xd033df57, { // "triangle grid: rings"
   "trójkątna siatka: pierścienie",
/*MISSING*/    "triangle grid: rings",
   "trojúhelníková mřížka: s prstenci",
   "треугольная сетка: кольца",
/*MISSING*/    "triangle grid: rings",
/*MISSING*/    "triangle grid: rings",
    }},
  {0xd0863104, { // "Congratulations! You have found the Holy Grail!"
   "Gratulacje! Znalaz%łeś0 Świętego Graala!",
   "Tebrikler! Kutsal Kâseyi buldun!",
   "Gratulujeme! Naš%el0 jsi Svatý grál!",
   "Поздравляем! Вы нашли святой Грааль!",
   "Glückwunsch! Du hast den Heiligen Gral gefunden!",
/*MISSING*/    "Congratulations! You have found the Holy Grail!",
    }},
  {0xd09668f7, { // "Press 'o' to see all the lands in HyperRogue."
   "Naciśnij 'o', by zobaczyć wszystkie krainy w HyperRogue.",
/*MISSING*/    "Press 'o' to see all the lands in HyperRogue.",
   "Stiskem 'o', zobrazíš všechny kraje v HyperRogue.",
   "Нажми 'o', чтобы увидеть все земли в HyperRogue.",
/*MISSING*/    "Press 'o' to see all the lands in HyperRogue.",
/*MISSING*/    "Press 'o' to see all the lands in HyperRogue.",
    }},
  {0xd0ab00e0, { // "Illusions are targeted by most monsters, just like yourself, Thumpers, and your friends."
   "Większość potworów celuje w iluzje, podobnie jak w Ciebie, Dudniki i Twoich przyjaciół.",
   "İlüzyonlar tıpkı sen, Gümleyenler ve dostların gibi pek çok canavar tarafından saldırıya uğrayabilir.",
   "Většina netvorů půjde po tvých iluzích, stejně jako jdou po tobě, po Tloucích a po tvých přátelích.",
   "Их будут атаковать почти все монстры, так же как Вас, Тамперы и ваших союзников.",
   "Illusionen werden von den meisten Monstern verfolgt.",
/*MISSING*/    "Illusions are targeted by most monsters, just like yourself, Thumpers, and your friends.",
    }},
  {0xd0b045b5, { // "heptagonal grid"
   "siatka siedmiokątna",
/*MISSING*/    "heptagonal grid",
   "sedmiúhelníková mřížka",
   "семиугольная сетка",
/*MISSING*/    "heptagonal grid",
/*MISSING*/    "heptagonal grid",
    }},
  {0xd0d94ce6, { // "Nice, a hyperbolic rug! It will not fit very well on your flat Euclidean floor, but who cares?"
   "Jaki ładny hiperboliczny dywan! Nie będzie dobrze przylegał do Twojej płaskiej, euklidesowej podłogi, ale co z tego?",
   "Harika! Bir hiperbolik kilim! Evindeki öklidyen zemine pek yakışmayacak, ama kimin umrunda!",
   "Prima, hyperbolický koberec! Na tvoji placatou eukleidovskou podlahu moc dobře pasovat nebude, ale co?",
   "Прекрасный гиперболический ковёр! На плоский пол его не положишь, ну и пусть.",
   "Ein hyperbolischer Teppich! Er wird nicht gut auf deinen flachen euklidischen Boden passen aber wen interessiert's?",
/*MISSING*/    "Nice, a hyperbolic rug! It will not fit very well on your flat Euclidean floor, but who cares?",
    }},
  {0xd0e49b9a, { // "field pattern"
   "wzór pola",
/*MISSING*/    "field pattern",
   "vzor pole",
   "узор поля",
/*MISSING*/    "field pattern",
/*MISSING*/    "field pattern",
    }},
  {0xd10a4538, { // "You swing your sword at the mirror."
   "Wywijasz mieczem w kierunku lustra.",
/*MISSING*/    "You swing your sword at the mirror.",
   "Udeøil jsi meèem do zrcadla.",
   "Ты размахиваешь мечом в сторону зеркала.",
/*MISSING*/    "You swing your sword at the mirror.",
/*MISSING*/    "You swing your sword at the mirror.",
    }},
  {0xd10ed1c0, { // "pure tactics mode"
   "tryb taktyczny",
   "saf taktik modu",
   "taktický mód",
   "режим тактики",
   "Strategiemodus",
/*MISSING*/    "pure tactics mode",
    }},
  {0xd18ab005, { // "band"
   "wstęga",
/*MISSING*/    "band",
   "pás",
   "лента",
   "Band",
/*MISSING*/    "band",
    }},
  {0xd1fc64ef, { // "You activate your demonic powers!"
   "Aktywowałeś moce demoniczne!",
   "Şeytani güçlerini aktifleştirdin!",
   "Aktivoval jsi svou démonickou moc!",
   "Вы активировали свои демонические силы!",
   "Du aktivierst deine dämonischen Kräfte!",
   "Você ativa seus poderes demoníacos!",
    }},
  {0xd2223067, { // "Your eyes shine like gems."
   "Twoje oczy błyszczą jak klejnoty.",
   "Gözlerin mücevherler gibi ışıldıyor.",
   "Tvé oči září jako drahokamy.",
   "Ваши глаза сияют, как драгоценные камни.",
   "Deine Augen funkeln wie Edelsteine.",
   "Seus olhos brilham como gemas.",
    }},
  {0xd250e502, { // "This orb allows one to pass through all kinds of walls and chasms."
   "Ta sfera pozwala przechodzić przez wszelkie ściany i przepaście.",
   "Bu küre her türlü duvarların ve çatlakların içinden geçmene izin verir.",
   "Tato sféra ti umožňuje procházet skrz libovolné zdi a přes propasti.",
   "Эта сфера позволяет ходить через все стены и пропасти.",
   "Dieser Orb erlaubt es dir, alle Arten von Wänden und Abgründen zu passieren.",
/*MISSING*/    "This orb allows one to pass through all kinds of walls and chasms.",
    }},
  {0xd298b8d5, { // "A strange land filled with mirrors. Break magic mirrors and mirage clouds to gain treasures and helpful Mimics."
   "Dziwna kraina pełna luster. Zbijaj magiczne lustra i mirażowe chmury, by zdobywać skarby i przyjazne Mimiki.",
/*MISSING*/    "A strange land filled with mirrors. Break magic mirrors and mirage clouds to gain treasures and helpful Mimics.",
   "Podivný kraj plný zrcadel. Rozbíjej magická zrcadla a oblaky přeludů, abys získal poklady a užitečné Mimiky.",
   "Странная земля, полная зеркал. Разбивай волшебные зеркала и облака миражей, чтобы найти сокровища и дружелюбных Мимиков.",
/*MISSING*/    "A strange land filled with mirrors. Break magic mirrors and mirage clouds to gain treasures and helpful Mimics.",
/*MISSING*/    "A strange land filled with mirrors. Break magic mirrors and mirage clouds to gain treasures and helpful Mimics.",
    }},
  {0xd2995d4e, { // "This Orb allows you to throw fire, just like the Fire Cultists.\n\nEach fire drains 5 charges. You are not allowed to throw fire into adjacent cells."
   "Ta sfera pozwala Ci rzucać ogniem tak jak Ogniści Kultyści. \n\nKażdy rzucony ogień wymaga 5 ładunków. Nie możesz rzucać ogniem do sąsiednich komórek.",
   "Bu Küre tıpkı Ateş Müritleri gibi ateş atabilmeni sağlar.\n\nHer ateş 5 şarj tüketir, bitişik hücrelere ateş atamazsın.",
   "Tato sféra ti umožňuje vrhat oheň jako Kultisté ohně.\n\nKaždé použití sníží náboj sféry o 5. Oheň nelze vrhnout na sousední políčko.",
   "Эта сфера позволит Вам кидать огонь, прямо как огненные культисты.\n\nКаждое использование тратит 5 зарядов. Вы не можете кинуть огонь на соседнюю клетку.",
   "Dieser Orb erlaubt es dir, Feuer zu werfen.\n\nJedes Feuer verbraucht 5 Ladungen. Du kannst kein Feuer auf angrenzende Zellen werfen.",
/*MISSING*/    "This Orb allows you to throw fire, just like the Fire Cultists.\n\nEach fire drains 5 charges. You are not allowed to throw fire into adjacent cells.",
    }},
  {0xd2c6cd03, { // "Scores and achievements are not"
   "Wyniki i osiągnięcia nie są",
   "Başarımlar ve puanlar Öklid",
   "V Eukleidovském módu se",
   "Очки и достижения не",
   "Im euklidischem Modus werden",
/*MISSING*/    "Scores and achievements are not",
    }},
  {0xd2dabc99, { // "you are looking through it!"
   "patrzysz przez hiperboloidę!",
   "Hiperboloidin içinden bakıyorsun!",
   "díváš se skrz něj!",
   "взгляните на гиперболоид!",
   "Du fängst an zu verstehen!",
   "Você está olhando através!",
    }},
  {0xd305b3e9, { // "Avoid escaping from a discharge (\"That was close\")."
   "Unikaj uciekania przed wyładowaniem (\"było blisko\").",
/*MISSING*/    "Avoid escaping from a discharge (\"That was close\").",
   "Nesmíš se vyhnout výboji (\"To bylo těsně\").",
   "Не убегать из-под разряда (\"Это было близко\").",
/*MISSING*/    "Avoid escaping from a discharge (\"That was close\").",
/*MISSING*/    "Avoid escaping from a discharge (\"That was close\").",
    }},
  {0xd30ccbb8, { // "A simple servant of the master of the Ivory Tower."
   "Prosty sługa mistrza Wieży z Kości Słoniowej.",
   "Fildişi Kuledeki basit bir hizmetkâr.",
   "Prostý služebník pána Věže ze slonové kosti.",
   "Обычный слуга Хозяина башни из слоновой кости.",
   "Ein einfacher Diener des Meisters des Elfenbeinturms.",
/*MISSING*/    "A simple servant of the master of the Ivory Tower.",
    }},
  {0xd33a9d85, { // "disk"
   "koło",
/*MISSING*/    "disk",
   "disk",
   "диск",
   "Scheibe",
/*MISSING*/    "disk",
    }},
  {0xd3566de2, { // "Despite this intelligence, Rogues appear extremely surprised by the most basic facts about geometry. They must come from some really strange world.\n\n"
   "Mimo ich inteligencji Cwaniacy wyglądają na szalenie zdziwionych nawet najprostszymi faktami geometrycznymi. Muszą pochodzić z jakiegoś bardzo dziwnego świata.",
/*MISSING*/    "Despite this intelligence, Rogues appear extremely surprised by the most basic facts about geometry. They must come from some really strange world.\n\n",
   "Navzdory jejich inteligenci se zdá, že Chytráky dokážou nesmírně překvapit i ta nejzákladnější geometrická fakta. Musejí pocházet z nějakého opravdupodivného světa.",
   "Несмотря на их интеллект, Разбойники весьма удивлены простейшими фактами из геометрии. Наверное, они пришли из очень странного мира\n\n.",
/*MISSING*/    "Despite this intelligence, Rogues appear extremely surprised by the most basic facts about geometry. They must come from some really strange world.\n\n",
/*MISSING*/    "Despite this intelligence, Rogues appear extremely surprised by the most basic facts about geometry. They must come from some really strange world.\n\n",
    }},
  {0xd3c0dbc4, { // "This attack is likely to hit the attacker."
   "Ten atak pewnie trafi w atakującego.",
   "Bu saldırı büyük ihtimalle saldırganı vuracaktır.",
   "Tento útok může velmi snadno zasáhnout i samotného útočníka.",
   "Эта атака, вероятно, попадёт в атакующего.",
   "Dieser Angriff trifft wahrscheinlich den Angreifer.",
/*MISSING*/    "This attack is likely to hit the attacker.",
    }},
  {0xd3c2afef, { // "It is a shame to cheat!"
   "Wstyd oszukiwać!",
   "Seni Hilekâr!",
   "Cheatovat je hanba!",
   "Как Вам не стыдно!",
   "Cheaten ist eine Schande!",
   "É vergonhoso trapacear!",
    }},
  {0xd4319f18, { // "A beast made of vines!\n\nVine Beasts turn into vines when they die."
   "Stworzenie zrobione z winorośli!\n\nWinne Stwory zmieniają się w winorośle, gdy giną.",
   "Şaraptan yaratılmış bir canavar!\n\nŞarap Canavarı öldüğünde şaraba dönüşür.",
   "Netvor vytvořený z révy!\n\nRévoví netvoři se po smrti promění v révu.",
   "Зверь, свитый из виноградных лоз!\n\nВинный зверь превращается в лозу при гибели.",
   "Eine Bestie aus Reben!\n\nRebenbestien werden zu Reben wenn sie sterben.",
/*MISSING*/    "A beast made of vines!\n\nVine Beasts turn into vines when they die.",
    }},
  {0xd4420f18, { // "summon Bonfire"
   "przywołaj ognisko",
   "Kampateşi çıkar",
   "vyvolej ohniště",
   "призвать Костры",
   "rufe Leuchtfeuer herbei",
/*MISSING*/    "summon Bonfire",
    }},
  {0xd47953f4, { // "save the configuration"
   "zapis konfiguracji",
   "ayarları kaydet",
   "ulož konfiguraci",
   "сохранить настройки",
   "Konfiguration speichern",
/*MISSING*/    "save the configuration",
    }},
  {0xd49712c9, { // "exit 3D configuration"
   "wyjdź z konfiguracji 3D",
/*MISSING*/    "exit 3D configuration",
   "opustit 3D konfiguraci",
   "выйти из настроек 3D",
/*MISSING*/    "exit 3D configuration",
/*MISSING*/    "exit 3D configuration",
    }},
  {0xd4d45743, { // "The sandworm explodes in a cloud of Spice!"
   "Czerw eksplodował! Chmura Przyprawy!",
   "Kumkurdu patladı ve etrafa baharat saçtı!",
   "Písečný červ vybuchl v oblaku Koření!",
   "Червь взорвался и стал облаком специй",
   "Der Sandwurm explodiert in einer Wolke aus Spice",
   "A minhoca de areia explode numa nuvem de Especiarias!",
    }},
  {0xd4e089bc, { // "\n\n(For the heptagonal mode, the radius has been reduced to 2 for closing plates.)"
   "\n\n(W trybie siedmiokątów promień działania dla płyt zamykających jest zmniejszony do 2.)",
/*MISSING*/    "\n\n(For the heptagonal mode, the radius has been reduced to 2 for closing plates.)",
   "\n\n(V sedmiúhelníkovém módu mají zavírací desky dosah pouze 2.)",
   "\n\n(В режиме семиугольников радиус действия закрывающих плит уменьшен до 2.)",
   "\n\n(Für den heptagonalen Modus wurde der Wirkradius von Schließmechanismen auf 2 reduziert.)",
/*MISSING*/    "\n\n(For the heptagonal mode, the radius has been reduced to 2 for closing plates.)",
    }},
  {0xd53e31f8, { // "This spot will be burning soon! RUN!"
   "To się zaraz zapali! UCIEKAJ!",
   "Bu nokta yakında yanacak! KAÇ!",
   "Za chvíli to tu začne hořet! UTÍKEJ!",
   "Земля под Вами загорается! БЕГИТЕ!",
   "Dieser Ort wird bald in Flammen stehen - RENN!",
   "Esse lugar estará em chamas em breve! CORRA!",
    }},
  {0xd5612bae, { // "You stab %the1."
   "Dźgn%ąłeś0 %a1.",
   "%a1 bıçakladın.",
   "Probod%l0 jsi %a1.",
   "Вы закололи %a1.",
   "Du erstichst %den1 %a1.",
   "Você apunhala %na1.",
    }},
  {0xd568e544, { // "rotate left"
   "obróć w lewo",
   "sola döndür",
   "otočení vlevo",
   "повернуть влево",
   "Nach links rotieren",
/*MISSING*/    "rotate left",
    }},
  {0xd57b179f, { // "Collected the keys!"
   "Zebrano klucze!",
/*MISSING*/    "Collected the keys!",
   "Získal jsi klíče!",
   "Ключи собраны!",
/*MISSING*/    "Collected the keys!",
/*MISSING*/    "Collected the keys!",
    }},
  {0xd5a03aa4, { // "\"Have you visited a temple in R'Lyeh?\""
   "\"By%łeś0 w świątyni w R'Lyeh?\"",
   "\"R'Lyeh'teki tapınağı ziyaret ettin mi?\"",
   "\"Navštívi%l0 jsi už chrám v R'Lyeh?\"",
   "\"Вы уже были в храме в Р'Льехе?\"",
   "\"Warst du schon in einem der Tempel in R'Lyeh?\"",
/*MISSING*/    "\"Have you visited a temple in R'Lyeh?\"",
    }},
  {0xd5bce324, { // "This mode allows you to edit the map.\n\nNOTE: Use at your own risk. Combinations which never appear in the real game may work in an undefined way (do not work, look strangely, give strange messages, or crash the game).\n\nTo get the most of this editor, some knowledge of inner workings of HyperRogue is required. Each cell has four main fields: land type, wall type, monster type, item type. The same wall type (especially \"none\", \"sea\", or \"bonfire\") may look or work a bit differently, based on the land it is in. Sometimes an object may appear twice on the list due to subtle differences (for example, Demons could move next turn or not).\n\nPress w, i, l, or m to choose which aspect of cells to change, then just click on the cells and they will change. Press 'c' while hovering over a cell to copy that cell, this copies all information about it. When copying large areas or placing multi-tile monsters, it might be important where on the cell you are clicking.\n\nYou can also press 0-9 to apply your changes to a greater radius. This also affects the copy/paste feature, allowing to copy a larger area.\n\nPress F2 to save the current map (and F3 to load it). If you try this after a long game of HyperRogue (without using Orbs of Safety), the filesize will be very large! Note however that large structures, such as Great Walls, large circles and horocycles, are destroyed by this.\n\nPress 'b' to mark cells as boundaries. Such cells, and cells beyond them, are not copied by the copy/paste feature, nor saved by the save feature.\n\n"
   "Ten tryb pozwala edytować mapę.\n\nUWAGA: używasz na własne ryzyko! Kombinacje, które nigdy nie pojawiają się w prawdziwej grze, mogą działać w sposób nieokreślony (nie działać, wyglądać dziwnie, produkować dziwne teksty albo wywalać grę).\n\nBy wynieść najwięcej z tego edytora, przydatna jest pewna wiedza o działaniu HyperRogue. Każde pole ma cztery atrybuty: typ lądu, typ ściany, typ potwora, typ przedmiotu. Ten sam typ ściany (szczególnie \"nic\", \"morze\" albo \"ognisko\") może wyglądać albo działać różnie w zależności od lądu, na którym się znajduje. Czasami obiekt może występować dwa razy na liście ze względu na subtelne różnice (na przykład, Demony mogą się ruszać w następnej kolejce albo nie).\n\nWciśnij w, i, l, m by wybrać, który atrybut zmienić, następnie klikaj pola i będą się zmieniać. Naciśnij 'c', gdy myszka jest nad polem, by kopiować to pole -- to kopiuje całą informację o tym polu. Kiedy kopiujesz albo stawiasz potwory wielopolowe, może mieć znaczenie punkt pola, w którym klikasz.\n\nMożesz też nacisnąć 0-9 by zastosować zmianę w większym promieniu. To wpływa również na kopiowanie/wklejanie, pozwalając na skopiowanie większego obszaru.\n\nWciśnij F2, by zapisać obecną mapę (i F3, by naładować). Jak spróbujesz to zrobić po długiej grze w HyperRogue (nie używającej Sfer Bezpieczeństwa), rozmiar pliku będzie bardzo duży! Uwaga: duże struktury, jak Wielkie Ściany i horocykle, są przez to niszczone.\n\nNaciśnij 'b' by zaznaczyć pole jako brzeg. Takie pola i pola za nimi nie są kopiowane funkcją copy/paste, ani też nie są zapisywane funkcją save.\n\n",
   "Bu mod haritayı düzenlemeni sağlar..\n\nNOT: sorumluluk sende! Gerçek oyunda hiç karşına çıkmayacak kombinasyonlar bilinmeyen bir şekilde çalışabilir. (çalışmayabilir, garip gözükebilir, oyunu çökertebilir.).\n\nEditörü layığıyla kullanabilmek için HyperRogue'un iç dinamiklerine dair bilgi sahibi olman beklenir. Her hücre dört temel alana sahip: diyar tipi, duvar tipi, canavar tipi, eşya tipi.Aynı duvar tipi (özellikle \"hiç\", \"deniz\", ve \"kampateşi\") içinde bulunduğu diyara göre farklı gözükebilir ve farklı işlev edinebilir. Bazen bir eşya özel durumundan ötürü listede iki kez gözükebilir.(Meselâ şeytanın bir sonraki tur hareket edip edemeyeceği durumu).\n\nHücrelerin hangi özelliğini değiştirmek istedğinizi w, i, l ve m'ye basarak seç. İşaretçinin olduğu hücreyi başka yere kopyalamak için c'ye bas. Bu, hücreye dair bilgiyi de kopyalar. Çok hücreli canavarları kopyalarken hücrenin neresine bastığın sonucu değiştirebilir.\n\n Değişikliklerini daha büyük bir çapta uygulamak için 0-9'a basabilirsin.. Bu aynı zamanda, kopyala/yapıştır özelliğinin de çapını genişletir.\n\nMevcut haritayı kaydetmek için F2'ye bas (yüklemek için de F3'e bas). Eğer bunu Güvenlik Küresi kullanmadığınız bir oyunda denersen, dosya boyutu çok büyük olacaktır! Dikkat: Büyük yapılar, meselâ horosikıllar, büyük duvarlar, geniş çemberler bu özelliği kullanınca yok olurlar. \n\nHücreleri sınır olarak belirlemek için b'ye basın. Bu hücreler ve bu hücrelerin ötesindeki hücreler, kopyala/yapıştır ve kaydetme özelliğine dahil edilmezler. \n\n",
   "V tomto módu můžeš editovat mapup.\n\nPOZOR: Používej na vlastní nebezpečí. Kombinace, které se v normální hře nikdy nevyskytují, mohou fungovat nedefinovaným způsobem (nefungovat, divně vypadat, generovat podivné zprávy nebo způsobit spadnutí hry).\n\nAbys z tohoto editoru získal co nejvíce, musíš vědět něco o tom, jak HyperRogue funguje \"pod kapotou\". Každé políčko má čtyři atributy: typ kraje, typ zdi, typ netvora, a typ předmětu. Ten samý typ zdi (především \"nic\", \"moře\" nebo \"oheň\") může vypadat nebo se chovat trochu jinak podle toho, ve kterém kraji se nachází. Některé položky mohou být v seznamu dvakrát díky drobným rozdílům (například Čerti mají dvě formy podle toho, zda se v příštím kole mají pohnout nebo ne).\n\nKlávesami 'w', 'i', 'l' a 'm' udáváš, který aspekt políček chceš změnit; pak na políčko jednoduše klikni, a ono se změní. Stiskem klávesy 'c' zkopíruješ políčko pod kurzorem (zkopírují se všechny informace o tomto políčku). Při kopírování nebo vkládání vícepolíčkových netvorů může být důležité, kam přesně v políčku klikáš.\n\nSvé změny můžeš také aplikovat na větší oblast pomocí kláves 0-9, které nastavují poloměr. To také ovlivňuje kopírování/vkládání, což ti umožní kopírovat větší oblast.\n\nF2 uloží současnou mapu (a F3 ji nahraje). Pokud to uděláš po dlouhé hře HyperRogue (kde jsi nepoužil Sféru Bezpečnosti), bude vzniklý soubor hodně velký! Ovšem pozor -- velké struktury, jako jsou Velké zdi, velké kruhy a horocykly budou uložením a nahráním mapy zničeny.\n\nKlávesou 'b' můžeš políčka označit jako hraniční. Tato políčka a políčka za nimi nebudou při operacích kopírování a vkládání kopírovány a nebudou ani uloženy.\n\n",
   "В этом режиме можно редактировать карту.\n\nВНИМАНИЕ: используйте его на свой риск! Комбинации, которые не могут возникнуть в настоящей игре, могут иметь неопределённое поведение (не работать, выглядеть странно, выдавать странные сообщения, вызывать вылет игры).\n\nЧтобы пользоваться редактором, нужно немного представлять, как устроена игра HyperRogue. Каждая клетка содержит четыре основных элемента: тип земли, тип стены, тип монстратип предмета. некоторые типы стен (особенно \"ничего\", \"море\" и \"костёр\") могут выглядеть или работать немного по-разному в зависимости от земли, в которой они находятся. Иногда объект может находиться в списке в двух экземплярах в зависимости от разных вещей (например, будет ли Демон двигаться на следующий ход).\n\nНажмите w, i, l, m чтобы выбрать, какой атрибут клетки будете менять, нажмите на клетку, и она изменится. Нажмите 'c', когда мышка находится над клетко       , чтобы копировать клетку -- копируется вся информация о ней. Если Вы копируете или размещаете монстров из нескольких клеток, может иметь значение часть клетки, на которую Вы нажмёте.\n\nНажмите 0-9, чтобы применить изменения с большим радиусом. Это также влияет на копирование/вставку, копируя большую область.\n\nНажмите F2, чтобы сохранить текущую карту (и F3, чтобы загрузить). Если вы попробуете сделат это после долгой игры в HyperRogue (без сфер Безопасности), размер файла может быть очень велик! Большие структуры, такие как Великие стены и ороциклы, будут уничтожены.\n\nНажмите 'b', чтобы обозначить клетку как границу. Такие клетки, как и клетки за ними, не будут копироваться, вставляться и сохраняться.\n\n",
   "Dieser Modus erlaubt es dir die Map zu bearbeiten.\n\nWARNUNG: Benutzen auf eigene Gefahr! Kombinationen, die im Spiel selbst nicht vorkommen, können in skurrilen Situationen enden (funktioniert nicht, sieht komisch aus, erzeugt komische Meldungen, führt zum Absturz.\n\nUm das beste aus dem Editor rauszuholen ist etwas Wissen über Abläufe in HyperRogue nötig. Jede Zelle hat vier Hauptarten: Landtyp, Wandtyp, Monstertyp und Itemart. Ein und dieselbe Wand kann unterschiedlich aussehen und funktionieren, je nachdem auf welchem Land sie platziert ist. Manche Objekte können auch mehrmals Meldungen erzeugen. Drücke w, i, l oder m um zu bestimmen, welchen Aspekt der Zelle du änderst. Dann klick einfach auf die Zelle und sie wird sich ändern. Drücke 'c' während du dich über einer Zelle befindest um diese zu kopieren. Beim Kopieren von großen Gebieten und dem Platzieren von großen Monstern kann es sehr wichtig sein wo man klickt.\n\nDurch das Drücken von 0-9 werden die Änderungen in einem größerem Radius ausgeführt. Das Ganze beeinflusst auch das Kopieren und Einfügen, wodurch ein größerer Bereich kopiert werden kann.\n\nDrücke F2 um die Map zu speichern (und F3 um sie zu laden). Nach einer langen HyperRogue Sitzung (ohne den Orb der Geborgenheit zu nutzen) kann die Datei sehr groß sein! Allerdings werden große Strukturen, wie Große Mauern, große Kreise und Horozykeln dadurch zerstört.\n\nDrücke 'b' um Zellen als Grenzen zu markieren. Solche Zellen, sowie Zellen hinter diesen können nicht kopiert oder gespeichert werden.\n\n",
/*MISSING*/    "This mode allows you to edit the map.\n\nNOTE: Use at your own risk. Combinations which never appear in the real game may work in an undefined way (do not work, look strangely, give strange messages, or crash the game).\n\nTo get the most of this editor, some knowledge of inner workings of HyperRogue is required. Each cell has four main fields: land type, wall type, monster type, item type. The same wall type (especially \"none\", \"sea\", or \"bonfire\") may look or work a bit differently, based on the land it is in. Sometimes an object may appear twice on the list due to subtle differences (for example, Demons could move next turn or not).\n\nPress w, i, l, or m to choose which aspect of cells to change, then just click on the cells and they will change. Press 'c' while hovering over a cell to copy that cell, this copies all information about it. When copying large areas or placing multi-tile monsters, it might be important where on the cell you are clicking.\n\nYou can also press 0-9 to apply your changes to a greater radius. This also affects the copy/paste feature, allowing to copy a larger area.\n\nPress F2 to save the current map (and F3 to load it). If you try this after a long game of HyperRogue (without using Orbs of Safety), the filesize will be very large! Note however that large structures, such as Great Walls, large circles and horocycles, are destroyed by this.\n\nPress 'b' to mark cells as boundaries. Such cells, and cells beyond them, are not copied by the copy/paste feature, nor saved by the save feature.\n\n",
    }},
  {0xd5bf9cc8, { // "explore the Emerald Pattern"
   "Eksploracja Szmaragdowego Wzoru",
   "Zümrüt Dokusunu keşfet",
   "Zkoumání Smaragdového vzoru",
   "Изучить Изумрудный узор",
   "Erkunde das Smaragdmuster",
/*MISSING*/    "explore the Emerald Pattern",
    }},
  {0xd5dfb865, { // "%1 vertices"
   "wierzchołki: %1",
   "köşeler: %1",
   "vrcholy: %1",
   "вершины: %1",
   "Eckpunkte: %1",
/*MISSING*/    "%1 vertices",
    }},
  {0xd6623e59, { // "A Stone from the Land of Power. You are not sure what it is exactly, but as the Powerstones are kept in crystal cabinets, they are surely valuable."
   "Kamień z Krainy Mocy. W sumie to nie wiesz, co to jest, ale trzymają te kamienie w kryształowych gablotach, także muszą być bardzo cenne.",
   "Güç Diyarından bir taş. Ne olduğuna tam olarak emin değilsin, ama bu taşlar kristal kabinlerde saklandığına göre çok değerli olmalılar.",
   "Kámen z Kraje Moci. Nejsi si jistý, co je přesně zač, ale když dávají Mocikamy do křišťálových vitrín, musejí mít jistě velkou cenu.",
   "Камень из Земли силы. Вы точно не знаете, что это такое, но раз они спрятаны в кристальных кабинетах, они довольно ценны.",
   "Ein Stein aus dem Land der Macht. Du bist dir nicht ganz sicher was er tut, aber da er in einer Kristallvitrine ist, ist er sicher wertvoll!",
/*MISSING*/    "A Stone from the Land of Power. You are not sure what it is exactly, but as the Powerstones are kept in crystal cabinets, they are surely valuable.",
    }},
  {0xd66b2b2a, { // "Poincaré Ball model"
   "Model kuli",
/*MISSING*/    "Poincaré Ball model",
   "Model Poincarého koule",
   "Сферическая модель Пуанкаре",
/*MISSING*/    "Poincaré Ball model",
/*MISSING*/    "Poincaré Ball model",
    }},
  {0xd673c1e3, { // "challenge"
   "wyzwanie",
   " ek görev",
   "obtížná",
   "испытание",
   "herausfordernd",
/*MISSING*/    "challenge",
    }},
  {0xd687021e, { // "e = edit this"
   "e = edytuj to",
   "e = bunu düzenle",
   "e = editace",
   "e = изменить",
   "e = editieren",
/*MISSING*/    "e = edit this",
    }},
  {0xd68ec8e1, { // "Demons are slow, but you'll need the experience against stronger ones..."
   "Mniejsze Demony są powolne, ale musisz zdobyć doświadczenie przeciwko Większym...",
/*MISSING*/    "Demons are slow, but you'll need the experience against stronger ones...",
   "Čerti jsou pomalí, ale proti Ďáblům budeš potřebovat zkušenosti...",
   "Демоны медленны, но Вам нужен опыт, чтобы сражаться против более сильных...",
/*MISSING*/    "Demons are slow, but you'll need the experience against stronger ones...",
/*MISSING*/    "Demons are slow, but you'll need the experience against stronger ones...",
    }},
  {0xd693326f, { // "Forest Trolls create an impassable wall when they die."
   "Zabite Leśne Trolle pozostawiają ścianę.",
   "Orman Trolleri öldüklerinde artlarında aşılmaz bir duvar bırakır.",
   "Lesní trollové se po smrti promění v neprostupnou zeď.",
   "Лесные тролли превращаются в непроходимые стены, когда умирают.",
   "Waldtrolle erschaffen eine unpassierbare Wand wenn sie sterben.",
/*MISSING*/    "Forest Trolls create an impassable wall when they die.",
    }},
  {0xd6a28707, { // "Cannot teleport on a monster!"
   "Nie możesz teleportować się na potwora!",
   "Bir canavarın üzerine ışınlanamazsın!",
   "Nemůžeš se teleportovat na netvora!",
   "Нельзя телепортироваться на монстра!",
   "Du kannst nicht auf Monster teleportieren!",
/*MISSING*/    "Cannot teleport on a monster!",
    }},
  {0xd6f55214, { // " (used %1 times)"
   " (użyte %1 razy)",
/*MISSING*/    " (used %1 times)",
   " (použitá %1krát)",
   " (использовано %1 раз)",
/*MISSING*/    " (used %1 times)",
/*MISSING*/    " (used %1 times)",
    }},
  {0xd707820a, { // "In the Shoot'em Up mode, you are armed with thrown Knives."
   "W trybie strzelanki jesteś uzbrojony w noże do rzucania.",
   "SHMUP modunda silahın fırlatılabilen bıçaklar.",
   "V módu střílečky jsi vyzbrojený vrhacími noži.",
   "В режиме стрельбы Вы вооружены метательными ножами.",
   "Im Ballermodus bist du mit Wurfmessern bewaffnet.",
/*MISSING*/    "In the Shoot'em Up mode, you are armed with thrown Knives.",
    }},
  {0xd7297e6d, { // "This statue is made of materials which cannot be found in your world."
   "Ta statua jest zrobiona z materiałów, których nie ma w Twoim świecie.",
   "Bu heykel, senin geldiğin dünyada bulunamayacak maddelerden yapılmış.",
   "Tato soška je vytvořena z materiálů, které se ve tvém světě nenacházejí.",
   "Эти статуи сделаны из материала, который не найдёшь в Вашем мире.",
   "Diese Statue besteht aus Materialien, die in deiner Welt nicht existieren.",
/*MISSING*/    "This statue is made of materials which cannot be found in your world.",
    }},
  {0xd75fc967, { // "Poincaré model"
   "model Poincaré",
/*MISSING*/    "Poincaré model",
   "Poincarého model",
   "модель Пуанкаре",
/*MISSING*/    "Poincaré model",
/*MISSING*/    "Poincaré model",
    }},
  {0xd786c69d, { // "A beautiful figurine, made of ivory. Figurines close to the base of the Tower tend do disappear after you have collected many of them."
   "Piękna figurka z kości słoniowej. Figurki blisko podstawy Wieży znikają, gdy zbierzesz dużo figurek.",
   "Fildişinden yapılmış güzel bir biblo. Bunlardan çok sayıda topladığında, kulenin temeline yakın biblolar yok olmaya başlar.",
   "Nádherná figurka ze slonové kosti. Figurky poblíž základů Věže mají tendenci mizet, pokud jich získáš hodně.",
   "Чудесная фигурка из слоновой кости. Фигурки рядом с основанием башни пропадают, когда Вы соберёте много их.",
   "Eine hübsche Figur aus Elfenbein. Figuren nah am Fuße des Turms neigen dazu zu verschwinden wenn du viele gesammelt hast.",
/*MISSING*/    "A beautiful figurine, made of ivory. Figurines close to the base of the Tower tend do disappear after you have collected many of them.",
    }},
  {0xd7d9ba22, { // "plain"
   "prosty",
   "düz",
   "obyčejný",
   "простой",
   "schlicht",
   "plano",
    }},
  {0xd853202b, { // "Dead orbs gained!"
   "Zdobyte Martwe Sfery!",
   "Ölü Küreler kazandın!",
   "Získa%l0 jsi Mrtvé sféry!",
   "Мёртвые сферы получены!",
   "Tote Orbs erhalten!",
   "Orbes mortos obtidos!",
    }},
  {0xd85c8c1c, { // "three stripes"
   "trzy paski",
   "üç şerit",
   "tři pásy",
   "три полосы",
   "Drei Streifen",
/*MISSING*/    "three stripes",
    }},
  {0xd867f5a7, { // "Safety (quick save)"
   "Bezpieczeństwo (szybki save)",
   "Güvenlik (hızlı kayıt)",
   "Bezpečnost (rychlé uložení)",
   "Сохраниться",
   "Sicherheit (schnell speichern)",
/*MISSING*/    "Safety (quick save)",
    }},
  {0xd86b25e9, { // "or another key to continue"
   "lub inny klawisz, by kontynuować",
   "veya devam etmek için başka bir tuşa bas.",
   "jakákoli jiná klávesa: pokračovat",
   "либо другую клавишу, чтобы продолжить",
   "oder eine beliebige andere Taste zum Fortfahren.",
   "ou outra tecla para continuar",
    }},
  {0xd87a76e8, { // "cells generated"
   "wygenerowane pola",
   "üretilen hücreler",
   "počet vygenerovaných políček",
   "клеток сгенерировано",
   "Generierte Zellen",
   "celulas geradas",
    }},
  {0xd8a00f35, { // "letters"
   "literki",
/*MISSING*/    "letters",
   "písmena",
   "буквы",
/*MISSING*/    "letters",
/*MISSING*/    "letters",
    }},
  {0xd8b71c12, { // "These creatures are slow, but they often appear in large numbers."
   "Te demony są powolne, ale często pojawiają się w dużych grupach.",
   "Bu yaratıklar yavaştır ama genellikle sürü halinde gezerler.",
   "Tito tvorové jsou pomalí, ale často se objevují ve velkém počtu.",
   "Эти существа медленны, но они берут количеством.",
   "Diese Monster sind langsam aber sie erscheinen oft in großen Gruppen.",
/*MISSING*/    "These creatures are slow, but they often appear in large numbers.",
    }},
  {0xd8c4577a, { // "Your Orb of Recall is blocked by something big!"
   "Coś dużego blokuje Twoją Sferę Powrotu!",
/*MISSING*/    "Your Orb of Recall is blocked by something big!",
   "Tvoji Sféru Návratu blokuje něco velikého!",
   "Что-то большое блокирует вашу Сферу Возвращения!",
/*MISSING*/    "Your Orb of Recall is blocked by something big!",
/*MISSING*/    "Your Orb of Recall is blocked by something big!",
    }},
  {0xd8ca7331, { // "Collect more treasures, there are still more lands waiting..."
   "Zbieraj skarby, nowe krainy czekają...",
   "Daha fazla hazine topla, hâlâ seni bekleyen diyarlar var...",
   "Sbírej další poklady, další kraje stále čekají...",
   "Собирайте больше сокровищ, ещё много земель Вас ждут...",
   "Sammle mehr Schätze, es warten weitere Länder auf dich...",
   "Colete mais tesouros, ainda mais terras lhe aguardam...",
    }},
  {0xd8cc6c5b, { // "Now, go to the Burial Grounds and find an Orb of the Sword. The Sword appears to always be facing in the same direction whatever you do, and it appears that you have to rotate the sword to excavate the treasures; yet, it is possible to excavate them! You migth have already noticed that the world rotates after you move around a loop and return to an old place.\n\nThis is related to the fact that the world of HyperRogue is curved, and the sum of angles in a triangle is not equal to 180 degrees."
   "Teraz idź do Kurhanów i znajdź Sferę Ostrza. Ostrze wygląda jakby zawsze było skierowane w tą samą stronę, a wydaje się, że do wykopania skarbów trzeba je obrócić. Jednak wykopanie skarbów jest możliwe! Być może już zauważy%łeś0, że świat się obraca, gdy zrobisz pętlę.\n\nJest to związane z zakrzywieniem świata HyperRogue, i tym, że suma kątów trójkąta nie jest równa 180 stopni.",
/*MISSING*/    "Now, go to the Burial Grounds and find an Orb of the Sword. The Sword appears to always be facing in the same direction whatever you do, and it appears that you have to rotate the sword to excavate the treasures; yet, it is possible to excavate them! You migth have already noticed that the world rotates after you move around a loop and return to an old place.\n\nThis is related to the fact that the world of HyperRogue is curved, and the sum of angles in a triangle is not equal to 180 degrees.",
   "Teď se vydej na Pohřebiště a najdi Sféru meče. Ať děláš, co děláš, Meč bude vždy otočený stejným směrem, a přitom se zdá, že abys mohl vykopat poklady, musíš Meč natočit; přesto je však možné je vykopat! Možná už sis všiml, že když se pohybuješ po uzavřené křivce a vrátíš se tam, odkud jsi vyšel, bude svět pootočený.\n\nTo souvisí s faktem, že svět HyperRogue je zakřivený a součet úhlů v trojúhelníku se nerovná 180 stupňům.",
   "Теперь иди в Курганы и найди Сферу меча. Кажется, что меч всегда направлен в одну и ту же сторону, что бы ты не делал%E0, а для того, чтобы добыть сокровища, нужно его повернуть. И его можно повернуть! Может быть, ты уже заметил%E0, что мир повернётся, если пройти по петле и вернуться на то же место.\n\nЭто связано с тем, что мир HyperRogue искривлён, и сумма углов треугольника не равна 180 градусам.",
/*MISSING*/    "Now, go to the Burial Grounds and find an Orb of the Sword. The Sword appears to always be facing in the same direction whatever you do, and it appears that you have to rotate the sword to excavate the treasures; yet, it is possible to excavate them! You migth have already noticed that the world rotates after you move around a loop and return to an old place.\n\nThis is related to the fact that the world of HyperRogue is curved, and the sum of angles in a triangle is not equal to 180 degrees.",
/*MISSING*/    "Now, go to the Burial Grounds and find an Orb of the Sword. The Sword appears to always be facing in the same direction whatever you do, and it appears that you have to rotate the sword to excavate the treasures; yet, it is possible to excavate them! You migth have already noticed that the world rotates after you move around a loop and return to an old place.\n\nThis is related to the fact that the world of HyperRogue is curved, and the sum of angles in a triangle is not equal to 180 degrees.",
    }},
  {0xd8e7b5df, { // "quit"
   "koniec",
/*MISSING*/    "quit",
   "konec",
   "выход",
/*MISSING*/    "quit",
/*MISSING*/    "quit",
    }},
  {0xd91463d5, { // "The sandworm explodes!"
   "Czerw eksplodował!",
   "Kumkurdu patladı!",
   "Písečný červ vybuchl!",
   "Червь взорвался!",
   "Der Sandwurm explodiert!",
/*MISSING*/    "The sandworm explodes!",
    }},
  {0xd919e5b0, { // "skin color"
   "kolor skóry",
   "deri rengi",
   "barva pleti",
   "цвет кожи",
   "Hautfarbe",
/*MISSING*/    "skin color",
    }},
  {0xd95bd05f, { // "player 6 X"
   "gracz 6 X",
/*MISSING*/    "player 6 X",
   "hráč 6 X",
   "игрок 6 X",
   "Spieler 6 X",
/*MISSING*/    "player 6 X",
    }},
  {0xd95bd060, { // "player 6 Y"
   "gracz 6 Y",
/*MISSING*/    "player 6 Y",
   "hráč 6 Y",
   "игрок 6 Y",
   "Spieler 6 Y",
/*MISSING*/    "player 6 Y",
    }},
  {0xd96191e7, { // "line patterns"
   "wzory linii",
/*MISSING*/    "line patterns",
   "vzory čar",
   "шаблоны линий",
/*MISSING*/    "line patterns",
/*MISSING*/    "line patterns",
    }},
  {0xd97c7d02, { // "This Orb allows to temporarily create impassable matter, either to block paths or to build bridges across chasms and waters."
   "Ta sfera pozwala tymczasowo stworzyć nieprzekraczalną barierę. Można to wykorzystać do blokowania tras lub do budowy mostów nad przepaściami i wodami.",
   "Bu Küre, yolları tıkamak veya boşluklar ve su üzerine köprü kurmak için geçilemez madde oluşturmaya yarar. ",
   "Tato Sféra ti umožňuje dočasně vytvářet neprostupnou hmotu, kterou můžeš zatarasit cestu nebo využít jako most přes propasti nebo přes vodu.",
   "Эта сфера позволяет временно создавать непроходимое вещество, чтобы блокировать пути или строить мосты через пропасть или воду.",
   "Dieser Orb erlaubt es dir, unpassierbare temporäre Materie zu erzeugen, um Pfade zu blockieren oder Brücken über Wasser und Schluchten zu bauen.",
/*MISSING*/    "This Orb allows to temporarily create impassable matter, either to block paths or to build bridges across chasms and waters.",
    }},
  {0xd987ce32, { // "All the monsters you kill are carried to this strange land, and buried. Careless Rogues are also carried here..."
   "Pochowane są tu wszystkie potwory, które zabijasz. A także poszukiwacze skarbów, którzy nie uważali...",
   "Öldürdüğünüz bütün canavarlar buraya taşındı ve gömüldü. Dikkatsiz Seyyahlar da buraya taşındı...",
   "Všichni netvoři, které zabiješ, jsou odneseni do tohoto podivného kraje, a tam pohřbeni. A také neopatrní lupiči...",
   "Все убитые Вами монстры похоронены здесь. И неосторожные охотники за сокровищами тоже...",
   "Alle Monster, die du tötest, werden hier begraben. Auch unvorsichtige Schurken landen hier...",
/*MISSING*/    "All the monsters you kill are carried to this strange land, and buried. Careless Rogues are also carried here...",
    }},
  {0xd9a6e92a, { // "Cannot drop %the1 here!"
   "Nie możesz tu położyć %a1!",
   "Ölü Küreyi buraya bırakamazsın!",
   "Tady nelze položit Mrtvou Sféru!",
   "Сюда нельзя положить мёртвую сферу!",
   "Du kannst %den1 %a1 hier nicht ablegen!",
/*MISSING*/    "Cannot drop %the1 here!",
    }},
  {0xd9b57a91, { // "turn right"
   "skręć w prawo",
   "sağa dön",
   "otočení vpravo",
   "повернуть направо",
   "Nach rechts drehen",
/*MISSING*/    "turn right",
    }},
  {0xd9c1c404, { // "first on top"
   "pierwsze na górze",
/*MISSING*/    "first on top",
   "první nahoře",
   "первые сверху",
/*MISSING*/    "first on top",
/*MISSING*/    "first on top",
    }},
  {0xd9c3c871, { // "This flower brings fortune to the person who finds it.\n"
   "Te rośliny dają szczęście temu, kto je znajdzie.\n",
   "Bu çiçek kendisini bulana şans getirir.\n",
   "Tato rostlina přináší štěstí tomu, kdo ji najde.\n",
   "Этот цветок приносит удачу тому, кто найдёт его.\n",
   "Diese Blume bringt ihrem Finder Glück.\n",
/*MISSING*/    "This flower brings fortune to the person who finds it.\n",
    }},
  {0xda09920d, { // "'n' to start"
   "n = start",
   "n = başlat",
   "n = start",
   "n = начать",
   "n = start",
/*MISSING*/    "'n' to start",
    }},
  {0xda0cf1c2, { // "sight range"
   "zasięg wzroku",
/*MISSING*/    "sight range",
   "viditelnost",
   "радиус зрения",
/*MISSING*/    "sight range",
/*MISSING*/    "sight range",
    }},
  {0xda2f2292, { // "You summon %the1!"
   "Przywoła%łeś0 %a1!",
   "%a1 çağırdın!",
   "Vyvola%l1 jsi %a1!",
   "Вы призвали %a1!",
   "Du beschwörst %den1 %a1!",
/*MISSING*/    "You summon %the1!",
    }},
  {0xda37711f, { // "heavy"
   "gruby",
   "ağır",
   "těžký",
   "тяжёлый",
   "schwierig",
   "pesado",
    }},
  {0xda55b584, { // "first on bottom"
   "pierwsze na dole",
/*MISSING*/    "first on bottom",
   "první dole",
   "первые снизу",
/*MISSING*/    "first on bottom",
/*MISSING*/    "first on bottom",
    }},
  {0xda753c45, { // "summon Ivy"
   "przywołaj Bluszcz",
   "Sarmaşık çıkar",
   "vyvolej Břečťan",
   "призвать Плющ",
   "rufe Efeu herbei",
/*MISSING*/    "summon Ivy",
    }},
  {0xda924f7c, { // "The result of a collaboration of the Great Vizier and the Wizard of the Ivory Tower."
   "Efekt współpracy Wielkiego Wezyra i Czarodzieja z Wieży z Kości Słoniowej.",
/*MISSING*/    "The result of a collaboration of the Great Vizier and the Wizard of the Ivory Tower.",
   "Výsledek spolupráce Velkovezíra a Čaroděje z Věže ze slonové kosti.",
   "Совместная разработка Великого визиря и Мага из Башни из слоновой кости.",
/*MISSING*/    "The result of a collaboration of the Great Vizier and the Wizard of the Ivory Tower.",
/*MISSING*/    "The result of a collaboration of the Great Vizier and the Wizard of the Ivory Tower.",
    }},
  {0xdac95cba, { // "Treasure gained!"
   "Skarb zdobyty!",
   "Hazine kazanıldı!",
   "Získa%l0 jsi poklad!",
   "Сокровища получены!",
   "Schatz erhalten!",
   "Tesouro obtido!",
    }},
  {0xdb18151b, { // "Could not save the paper model data"
   "Nie udało się zapisać danych do modelu",
   "Kâğıt modeli verisi kaydedilemedi",
   "Nepodařilo se uložit data pro papírový model",
   "Не удалось сохранить модель",
   "Papiermodell konnte nicht gespeichert werden",
/*MISSING*/    "Could not save the paper model data",
    }},
  {0xdb6a8542, { // "old style joystick configuration"
   "konfiguracja joysticka w starym stylu",
/*MISSING*/    "old style joystick configuration",
   "starý styl konfigurace joysticku",
   "старый режим настройки джойстика",
/*MISSING*/    "old style joystick configuration",
/*MISSING*/    "old style joystick configuration",
    }},
  {0xdb6beab6, { // "main menu"
   "menu główne",
/*MISSING*/    "main menu",
   "hlavní menu",
   "главное меню",
/*MISSING*/    "main menu",
/*MISSING*/    "main menu",
    }},
  {0xdb7f948a, { // "joystick configuration"
   "konfiguracja joysticka",
/*MISSING*/    "joystick configuration",
   "konfigurace joysticku",
   "настройка джойстика",
/*MISSING*/    "joystick configuration",
/*MISSING*/    "joystick configuration",
    }},
  {0xdb8c4eb4, { // "%The1 squeaks hopelessly."
   "%1 pisn%ął1 bez nadziei.",
   "%1 umutsuzca ciyaklıyor.",
   "%1 zdrceně piští.",
   "%1 пищит без надежды.",
   "%Die1 %1 piepst hoffnungslos.",
/*MISSING*/    "%The1 squeaks hopelessly.",
    }},
  {0xdb96b407, { // "When you have this Orb, most monsters won't see you, unless you are standing still, attacking, or picking up items."
   "Kiedy masz Sferę Niewidzialności, większość potworów Cię nie widzi, chyba że stoisz w miejscu, atakujesz, albo podnosisz przedmioty.",
   "Bu küren varsa çoğu canavar seni beklemediğin, saldırmadığın yahut yerden bir şey almadığın sürece görmez.",
   "Dokud máš tuto Sféru, většina netvorů tě neuvidí, pokud nebudeš stát na místě, nezaútočíš nebo nesebereš předmět.",
   "Пока у Вас есть эта сфера, большинство монстров не видят Вас, если Вы не стоите рядом, не атакуете и не поднимаете предметы.",
   "Während du diesen Orb besitzt, sehen dich die meisten Monster nicht, außer du stehst still, greifst an, oder hebst etwas auf.",
/*MISSING*/    "When you have this Orb, most monsters won't see you, unless you are standing still, attacking, or picking up items.",
    }},
  {0xdbba0c19, { // " Hyperstone: %1/%2"
   " Hiperkamienie: %1/%2",
   " Aşkıntaş: %1/%2",
   " Hyperkamy: %1/%2",
   " Гиперкамней: %1/%2",
   " Hypersteine: %1/%2",
/*MISSING*/    " Hyperstone: %1/%2",
    }},
  {0xdbe7dca1, { // "Whenever after your move there is a connection between a charged and a grounded cell, there is a short circuit. All cells on any path connecting a charged and a grounded cell (without repeated cells, or two consecutive grounded/charged cells) become damaged.\n\nSandstone walls and most creatures are conductive. Great Walls are isolators, but lands beyond them count as grounded.\n\nFulgurite, the treasure, is created when you manage to short circuit a sandstone wall, or a Rich Metal Beast.\n\nTrolls leave conductive rocks when killed, and Metal Beasts can only be killed by electricity -- your attacks only stun them, or push them away if already stunned."
   "Za każdym razem, gdy po Twoim ruchu jest połączenie między naładowanym a uziemionym polem, następuje zwarcie. Wszystkie pola na dowolnej ścieżce łączące pole naładowane z uziemionym (bez powtórzonych pól i bez dwóch sąsiednich naładowanych/uziemionych komórek) zostają uszkodzone.\n\nPiaskowe ściany i większość potworów przewodzi prąd. Wielkie ściany są izolatorami, ale krainy za nimi liczą się jako uziemione.\n\nFulguryt to skarb, który powstaje, gdy zwarcie zniszczy piaskową ścianę albo Bogatego Stwora.\n\nZabite trolle pozostawiają przewodzące ściany. Metalowe i Bogate Stwory można zabić tylko elektrycznością -- Twoje ataki je ogłuszają albo odpychają, jeśli są już ogłuszone.",
   "Hareketinden sonra, eğer yüklü ve topraklı bir hücre arasında bağlantı varsa, kısa devre olur.Yüklü ve topraklı hücrenin arasındaki bütün hücreler (tekrarlayan hücreler, yahut iki ardışık topraklı yahut yüklü hücre hariç) hasar görür.\n\nKumtaşı duvarları ve çoğu yaratık iletkendir. Büyük Duvarlar yalıtkandır, ancak onlardan sonraki hücreler topraklanmış sayılır. \n\nFulgurit, buradaki hazine, bir kumtaşı duvarını veya Zengin Metal Canavarını kısa devreye uğratmanla ortaya çıkar.\n\nTroller öldüklerinde artlarında iletken kayalar bırakırlar ve Metal Canavarlar sadece elektrik ile öldürülebilirler. Saldırıların metal canavarları sersemletir, yahut zaten sersemlemişlerse ittirir. ",
   "Po každém tvém tahu se zkontroluje, zda nevzniklo spojení mezi nějakým nabitým a uzemněným políčkem; pokud ano, dojde ke zkratu. Všechna políčka na jakékoli dráze spojující nabité a uzemněné políčko (pokud se na ní žádné políčko nevyskytuje dvakrát a pokud na ní nejsou dvě uzemněná/nabitá políčka za sebou, dostanou zásah.\n\nPískovcové zdi a většina bytostí je vodivá. Velké zdi jsou izolanty, ale kraje za nimi se počítají jako uzemněné.\n\nZdejší poklad, Fulgurit, vznikne, když se ti podaří zkratovat pískovcovou zeď nebo Bohatou bestii.\n\nTrollové se po smrti promění ve vodivé kameny a Kovové a Bohaté bestie je možné zabít pouze elektřinou -- tvé útoky je pouze omráčí nebo, pokud už omráčené jsou, odstrčí.",
   "После каждого Вашего хода, если есть соединение между заряженной и заземлённой клеткой, возникает короткое замыкание. По всем клеткам на пути между заряженной и заземлённой (без повторов, и без двух соседних заряженных/заземлённых) проходит ток.\n\nСтены из песчаника и большинство существ проводят ток. Великие стены -- изоляторы, а все клетки между ними считаются заземлёнными.\n\nФульгурит, местное сокровище, появляется, если разрушить песчаник или убить Богатого железного зверя.\n\nТролль при гибели превращается в проводящую стену. Железных зверей можно убить только электричеством -- атаки только оглушают их, или толкают, если уже оглушены.",
   "Nach jedem Zug gibt es einen Kurzschluss, vorausgesetzt es gibt eine Verbindung zwischen einer geladenen und einer geerdeten Zelle. Alle Zellen auf einem Pfad der eine geladene mit einer geerdeten Zelle verbindet werden beschädigt.\n\nSandsteinwände und die meisten Monster sind leitfähig. Große Mauern sind Isolatoren aber Zellen hinter ihnen zählen als geerdet.\n\nDer Schatz Fulgurit entsteht wenn eine Sandsteinwand oder ein ergiebiges Metallbiest kurzgeschlossen wird.\n\nTrolle hinterlassen leitfähige Steine wenn sie sterben und Metallbiester können nur durch Elektrizität getötet werden -- Deine Angriffe können sie nur betäuben oder wegstoßen wenn sie bereits betäubt sind.",
/*MISSING*/    "Whenever after your move there is a connection between a charged and a grounded cell, there is a short circuit. All cells on any path connecting a charged and a grounded cell (without repeated cells, or two consecutive grounded/charged cells) become damaged.\n\nSandstone walls and most creatures are conductive. Great Walls are isolators, but lands beyond them count as grounded.\n\nFulgurite, the treasure, is created when you manage to short circuit a sandstone wall, or a Rich Metal Beast.\n\nTrolls leave conductive rocks when killed, and Metal Beasts can only be killed by electricity -- your attacks only stun them, or push them away if already stunned.",
    }},
  {0xdc1433e4, { // "A wonderful beverage, apparently obtained by mixing red and blue slime. You definitely feel more healthy after drinking it, but you still feel that one hit of a monster is enough to kill you."
   "Cudowny napój, powstały z mieszania czerwonej i niebieskiej mazi. Po jego wypiciu czujesz się zdecydowanie lepiej, ale wciąż jedno uderzenie potwora może Cię powalić.",
   "Şahane bir içki, görünen o ki kırmızı ve mavi balçığın karıştırılmasıyla elde edilmiş. İçince daha sağlıklı hissediyorsun ama bir canavarın tek vuruşunun seni öldüreceğine de eminsin.",
   "Úžasný nápoj, prý vytvořený smísením červeného a modrého slizu. Po jeho vypití se rozhodně cítíš zdravější, ale pořád máš pocit, že by tě dokázal zabít jediný úder netvora.",
   "Чудесный напиток, полученный из синей и красной слизи. Вы чувствуете себя лучше, выпив его, но одного удара Вам всё ещё хватит, чтобы погибнуть.",
   "Ein wunderbares Getränk, anscheinend entstanden aus der Mischung von blauem und rotem Schleim. Du fühlst dich definitiv gesünder nach seinem Genuss, aber trotzdem ahnst du, dass ein Schlag eines Monsters ausreichen würde, dich zu töten.",
/*MISSING*/    "A wonderful beverage, apparently obtained by mixing red and blue slime. You definitely feel more healthy after drinking it, but you still feel that one hit of a monster is enough to kill you.",
    }},
  {0xdc4798ed, { // "Seven mines next to you!"
   "Siedem min koło Ciebie!",
   "Yakınında yedi mayın var!",
   "Ve tvé blízkosti se nachází sedm min!",
   "Семь мин рядом с Вами!",
   "7 Minen um dich!",
/*MISSING*/    "Seven mines next to you!",
    }},
  {0xdc55616f, { // "edit all three colors"
   "edytuj wszystkie kolory",
/*MISSING*/    "edit all three colors",
   "edituj všechny tři barvy",
   "редактировать три цвета",
/*MISSING*/    "edit all three colors",
/*MISSING*/    "edit all three colors",
    }},
  {0xdc5d5d33, { // "face the pointer"
   "celuj we wskaźnik",
   "mausa bak",
   "otočení ke kurzoru",
   "лицом к цели",
   "Zum Zeiger drehen",
/*MISSING*/    "face the pointer",
    }},
  {0xdc7303f4, { // "Orb power (target: facing)"
   "moc sfery (cel: na wprost)",
/*MISSING*/    "Orb power (target: facing)",
   "Použití Sféry (zaměření před postavou)",
   "клавиша для силы сферы",
   "Orbkraft (Ziel: Blickrichtung)",
/*MISSING*/    "Orb power (target: facing)",
    }},
  {0xdc994855, { // "Lots of treasure gained!"
   "Mnóstwo skarbów zdobyte!",
   "Bir sürü hazine kazandın!",
   "Získa%l0 jsi spoustu pokladů!",
   "Много сокровищ получено!",
   "Eine Menge Schätze erhalten!",
   "Vários tesouros obtidos!",
    }},
  {0xdc9aef4a, { // "The total value of the treasure you have collected.\n\nEvery world type contains a specific type of treasure, worth 1 $$$; your goal is to collect as much treasure as possible, but every treasure you find causes more enemies to hunt you in its native land.\n\nOrbs of Yendor are worth 50 $$$ each.\n\n"
   "Łączna wartość Twoich skarbów.\n\nKażda kraina zawiera specyficzny dla niej skarb, wart 1 $$$; Twoim celem jest zebranie jak najwięcej skarbów, ale każdy znaleziony skarb powoduje, że w jego krainie więcej potworów na Ciebie poluje.\n\nSfery Yendoru są warte po 50 $$$.\n\n",
   "Topladığınız hazinelerin toplam değeri.\n\nHer dünya 1 $$$ değerinde belli bir çeşit hazine içerir;  amacın olabildiğince çok hazine toplamak, ama her bulduğun hazine, daha fazla düşmanın sana saldırmasına sebep olur.\n\n Her Yendorun Küresi 50 $$$ değerindedir.\n\n",
   "Celková hodnota pokladů, které jsi sesbíral.\n\nV každém typu kraje se nachází specifický typ pokladů, který má vždy cenu 1 $$$; tvým cílem je sesbírat co nejvíce pokladů, ale každý poklad, který získáš, způsobí, že tě v krajích příslušného typu bude pronásledovat více nepřátel.\n\nKaždá Yendorská sféra má cenu 50 $$$.\n\n",
   "Общая стоимость найденных Вами сокровищ.\n\nКаждая земля содержит свой тип ценности, стоящей 1 $$$; Ваша цель - собрать как можно больше ценностей, но каждая найденная ценность увеличивает количество врагов в этой земле.\n\nСфера Йендора стоит 50 $$$.\n\n",
   "Der Gesamtwert deiner Schätze.\n\nJedes Land hat einen eigenen Schatztyp, der 1 $$$ wert ist. Dein Ziel ist soviel Schätze wie möglich zu sammeln, aber jeder Schatz    den du findest erzeugt mehr Feinde in seinem Herkunftsland.\n\nOrbs von Yendor sind 50 $$$ pro Stück wert.\n\n",
   "O valor total do tesouro que você coletou.\n\nCada mundo contém um tipo específico de tesouro , que vale 1 $$$; seu objetivo é coletar o máximo de tesouro possível, mas cada tesouro que você encontra faz com que mais inimigos lhe cacem na sua terra nativa.\n\nOrbes de Yendor valem 50 $$$ cada.\n\n",
    }},
  {0xdcb255ed, { // "Miners have special tools for dealing with the Living Cave. When they die, these tools activate, destroying the living cave around them."
   "Górnicy mają specjalne narzędzia do kopania w Żywych Jaskiniach. Kiedy giną, te narzędzia się uruchamiają, powodując, że żywa jaskinia wokół nich obumiera.",
   "Madenciler Yaşayan Mağarayla baş etmek için özel gereçlere sahipler. Öldüklerinde, bu gereçler aktifleşiyor ve etraflarındaki Yaşayan Mağarayı yok ediyor.",
   "Horníci mají speciální nástroje pro práci v Živoucí jeskyni. Když zemřou, jejich nástroje se aktivují a zničí živoucí jeskyni kolem nich.",
   "Шахтёры имеют специальные инструменты для копания Живой пещеры. При гибели они автоматически используются, уничтожая пещеру вокруг них.",
   "Minenarbeiter haben spezielle Werkzeuge für den Umgang mit der Lebenden Höhle. Sterben sie, aktivieren sich diese Werkzeuge, was die Höhle um sie herum zerstört.",
/*MISSING*/    "Miners have special tools for dealing with the Living Cave. When they die, these tools activate, destroying the living cave around them.",
    }},
  {0xdcebb09e, { // "So, you have killed a Ratling on the unwarped sea? You will be punished for this! Luckily, if you run away from the Warped Sea quickly, the Ratling Avengers will lose track of you."
   "Zabi%łeś0 Szczuraka na niezakrzywionym morzu? Mściciele Cię dosięgną! Na szczęście, jeśli szybko uciekniesz z Zakrzywionego Morza, Szczuraki-Mściciele zgubią Twój ślad.",
/*MISSING*/    "So, you have killed a Ratling on the unwarped sea? You will be punished for this! Luckily, if you run away from the Warped Sea quickly, the Ratling Avengers will lose track of you.",
   "Ty jsi zabi%l0 Krysáka na nepokřiveném moři? Za to zaplatíš! Krysáci mstitelé o tebe naštěstí ztratí zájem, pokud z Pokřiveného moře rychle utečeš.",
   "Вы убили Крысолюда в неискривлённом море? Мститель до Вас доберётся!К счастью, если достаточно далеко уйти от Искривлённого моря, он потеряет Ваш след.",
   "Du hast also einen Rattenmenschen auf der unverzerrten See getötet? Dafür wirst du bestraft werden! Wenn du Glück hast und schnell von der Verzerrten See verschwindest, werden die Rattenmenschen-Rächer deine Spur verlieren.",
/*MISSING*/    "So, you have killed a Ratling on the unwarped sea? You will be punished for this! Luckily, if you run away from the Warped Sea quickly, the Ratling Avengers will lose track of you.",
    }},
  {0xdcec3a85, { // "You have improved your total high score and %1 specific high scores!"
   "Poprawi%łeś0 swój ogólny najlepszy wynik oraz %1 z wyników szczegółowych!",
   "Toplam yüksek puanını ve %1 farklı tekil puanını artırdın!",
   "Zlepši%l0 jsi své celkové nejvyšší skóre a %1 specifických nejvyšších skóre!",
   "Вы подняли свой общий результат и %1 отдельных результатов!",
   "Du hast deinen Gesamthighscore und %1 spezifische Highscores verbessert!",
   "Você superou sua melhor pontuação e mais %1 melhores pontuações específicas!",
    }},
  {0xdcf852ca, { // "half-plane"
   "półpłaszczyzna",
/*MISSING*/    "half-plane",
   "polorovina",
   "полуплоскость",
   "Halbebene",
/*MISSING*/    "half-plane",
    }},
  {0xdd0527d1, { // "help"
   "help",
   "yardım",
   "nápověda",
   "помощь",
   "Hilfe",
/*MISSING*/    "help",
    }},
  {0xdd22a807, { // "This being radiates an aura of wisdom. It is made of a beautiful crystal, you would love to take it home. But how is it going to defend itself? Better not to think of it, thinking causes your brain to go hot...\n\nCrystal Sages melt at -30 °C, and they can rise the temperature around you from afar."
   "Ta istota promieniuje mądrością. Jest zrobiona z pięknego kryształu, na pewno bardzo cennego. Ciekawe, jak zamierza się przed Tobą bronić? Lepiej o tym nie myśleć, myślenie za bardzo Cię rozgrzewa...\n\nKryształowi Mędrcy topią się w -30 °C i powodują wzrost temperatury wokół Ciebie.",
   "Muhteşem bir kristalden yapılmış, eve götürmeyi çok isterdin.Ama kendini nasıl savunacak? Düşünmesen daha iyi, sadece düşünmek bile beynini zorluyor. ...\n\nKristal Pirleri -30 °C'de erir ve etrfındaki ısıyı uzaktan artırabilir.",
   "Tato bytost vyzařuje auru moudrosti. Je vytvořená z krásného krystalu -- moc rád%d0 by sis ji vza%l0 domů. Ale jak se bude bránit? Raději na to nemyslet, z myšlení se ti zahřívá mozek...\n\nKřišťáloví mudrci tají při teplotě -30 °C a dokážou kolem tebe na dálku zvyšovat teplotu.",
   "Они излучают ауру мудрости. Они сделаны из красивых кристаллов, и Вы хотели бы забрать их домой. Но как он собирается защищать себя? Лучше не думайте об этом, мышление нагревает Ваш мозг...\n\nКристальные мудрецы тают при -30 °C, и они могут повысить температуру вокруг Вас издалека.",
   "Dieses Wesen strahlt eine Aura der Weisheit aus. Es besteht aus einem schönen Kristall, am liebsten würdest du es mit nach Hause nehmen... Nur, wie verteidigt es sich? Besser nicht daran denken... Denken lässt dein Gehirn heiß werden...\n\nKristallweisen schmelzen bei -30 °C. Sie können die Temperatur um dich herum aus der Ferne ansteigen lassen.",
/*MISSING*/    "This being radiates an aura of wisdom. It is made of a beautiful crystal, you would love to take it home. But how is it going to defend itself? Better not to think of it, thinking causes your brain to go hot...\n\nCrystal Sages melt at -30 °C, and they can rise the temperature around you from afar.",
    }},
  {0xdd31dce5, { // "That was not a challenge. Find a larger castle!"
   "To nie było wyzwanie. Znajdź większy zamek!",
   "Bu seni heyecanlandırmadı bile. Daha büyük bir kale bul!",
   "To nebylo nic těžkého. Najdi si větší hrad!",
   "Это было несложно. Найдите больший замок!",
   "Das war keine Herausforderung. Finde ein größeres Schloss!",
/*MISSING*/    "That was not a challenge. Find a larger castle!",
    }},
  {0xdd6d9cd4, { // "settings set here won't be saved"
   "ustawienia ustawione tutaj się nie",
/*MISSING*/    "settings set here won't be saved",
   "nastavení provedená zde nebudou uložena",
   "установленные настройки не будут сохранены",
   "hier gemachte Einstellungen werden nicht gespeichert",
/*MISSING*/    "settings set here won't be saved",
    }},
  {0xdd78d3c0, { // "symmetry 0-1"
   "symetria 0-1",
   "simetri 0-1",
   "symetrie 0-1",
   "симметрия 0-1",
   "Symmetrie 0-1",
/*MISSING*/    "symmetry 0-1",
    }},
  {0xdd78d3c1, { // "symmetry 0-2"
   "symetria 0-2",
   "simetri 0-2",
   "symetrie 0-2",
   "симметрия 0-2",
   "Symmetrie 0-2",
/*MISSING*/    "symmetry 0-2",
    }},
  {0xdd78d3c2, { // "symmetry 0-3"
   "symetria 0-3",
   "simetri 0-3",
   "symetrie 0-3",
   "симметрия 0-3",
   "Symmetrie 0-3",
/*MISSING*/    "symmetry 0-3",
    }},
  {0xde221b93, { // "These dodecahedra made of a mysterious material are the Reptiles' favorite toy."
   "Te dwunastościany zrobione z tajemniczego materiału to ulubiona zabawka Jaszczurek.",
/*MISSING*/    "These dodecahedra made of a mysterious material are the Reptiles' favorite toy.",
   "Tyto dvanáctistěny z jakéhosi tajemného materiálu jsou oblíbenou hračkou Ještěrek.",
   "Эти додекаэдры из странного материала -- любимые игрушки ящериц.",
/*MISSING*/    "These dodecahedra made of a mysterious material are the Reptiles' favorite toy.",
/*MISSING*/    "These dodecahedra made of a mysterious material are the Reptiles' favorite toy.",
    }},
  {0xdec03d1b, { // "In this mode, HyperRogue is played on a 3D model of a part of the hyperbolic plane, similar to one you get from the 'paper model creator' or by hyperbolic crocheting.\n\nThis requires some OpenGL extensions and may crash or not work correctly -- enabling the 'render texture without OpenGL' options may be helpful in this case. Also the 'render once' option will make the rendering faster, but the surface will be rendered only once, so you won't be able to play a game on it.\n\nUse arrow keys to rotate, Page Up/Down to zoom."
   "W tym trybie można grać w HyperRogue na trójwymiarowym modelu części płaszczyzny hiperbolicznej, podobnym do tego, który można otrzymać w kreatorze modeli papierowych albo poprzez hiperboliczne szydełkowanie.\n\nDziałanie wymaga pewnych rozszerzeń OpenGL i może działać nieprawidłowo lub powodować zatrzymanie gry -- w takim przypadku może być przydatne włączenie opcji 'renderuj bez OpenGL'. Opcja 'renderuj raz' spowoduje, że animacja będzie szybsza, ale tekstura powierzchni zostanie narysowana tylko raz, zatem nie będzie możliwe granie na niej na żywo.\n\nMożesz obracać model strzałkami, a przybliżać i oddalać go klawiszami Page Up/Down.",
/*MISSING*/    "In this mode, HyperRogue is played on a 3D model of a part of the hyperbolic plane, similar to one you get from the 'paper model creator' or by hyperbolic crocheting.\n\nThis requires some OpenGL extensions and may crash or not work correctly -- enabling the 'render texture without OpenGL' options may be helpful in this case. Also the 'render once' option will make the rendering faster, but the surface will be rendered only once, so you won't be able to play a game on it.\n\nUse arrow keys to rotate, Page Up/Down to zoom.",
   "V tomto módu můžete hrát HyperRogue na 3D modelu části hyperbolické roviny podobném tomu, co vytváří generátor papírových modelů nebo hyperbolické háčkování.\n\nJe k tomu zapotřebí OpenGL a je možné, že v tomto módu hra spadne nebo nepoběží správně -- v takovém případě může pomoci zapnout možnost 'renderuj textury bez OpenGL'. Kromě toho je možné renderování urychlit volbou 'renderuj jednou', ale v takovém případě bude povrch vygenerován pouze jednou, takže na něm není možné hrát.\n\nModel se otáčí šipkami a lze ho zoomovat klávesami Page Up/Down.",
   "В этом режиме Вы играете в HyperRogue на трёхмерной модели гиперболической плоскости, похожей на те, что можно получить в 'создателе бумажных моделей'\n\nЭто требует некоторых расширений OpenGL, может работать некорректно или вызывать падение игры -- включение опции 'не использовать OpenGL' может помочь. Опция 'отрисовать один раз' ускоряет отрисовку, но играть в таком режиме невозможно.\n\nИспользуйте стрелки для вращения, Page Up/Down для изменения масштаба.",
   "In diesem Modus wird HyperRogue auf einem 3D-Modell eines Teils einer hyperbolischen Ebene gespielt, ähnlich zu dem was man beim 'Papiermodell-Schöpfer' oder beim hyperbolischen Häkeln erhält.\n\nDieser Modus erfordert einige OpenGL-Erweiterungen und kann abstürzen oder fehlerhaft funktionieren -- das Einschalten der Option 'Rendere Textur ohne OpenGL' kann hilfreich sein. Auch die Option 'Rendere die Textur nur einmal' beschleunigt das Rendern, aber die Oberfläche wird nur einmal gerendert und du kannst darauf nicht spielen.\n\nVerwende die Pfeiltasten zum Rotieren, Bild auf/ab zum Zoomen.",
/*MISSING*/    "In this mode, HyperRogue is played on a 3D model of a part of the hyperbolic plane, similar to one you get from the 'paper model creator' or by hyperbolic crocheting.\n\nThis requires some OpenGL extensions and may crash or not work correctly -- enabling the 'render texture without OpenGL' options may be helpful in this case. Also the 'render once' option will make the rendering faster, but the surface will be rendered only once, so you won't be able to play a game on it.\n\nUse arrow keys to rotate, Page Up/Down to zoom.",
    }},
  {0xdec89004, { // "scrolling speed"
   "prędkość przewijania",
/*MISSING*/    "scrolling speed",
   "rychlost scrollování",
   "prędkość przewijania",
/*MISSING*/    "scrolling speed",
/*MISSING*/    "scrolling speed",
    }},
  {0xded7510d, { // "The Knights of the Round Table are the greatest warriors of these lands. They are not very inventive with names though, as they call each of their castles Camelot. You are probably worthy of joining them, but they will surely give you some quest to prove yourself...\n\nEach castle contains a single treasure, the Holy Grail, in the center. The radius of the Round Table is usually 28, but after you find a Holy Grail successfully, each new castle (and each Round Table) you find will be bigger."
   "Rycerze Okrągłego Stołu to najwięksi wojownicy tych krain. Niestety, nie są zbyt pomysłowi w nazywaniu i każdy swój zamek nazywają Camelot. Prawdopodobnie nadajesz się, by stać się jednym z nich, ale na pewno dadzą Ci jakąś misję, byś pokaza%ł0 swoją wartość...\n\nKażdy zamek ma jeden skarb (Święty Graal) w samym środku. Promień Okrągłego Stołu to normalnie 28, ale z każdym znalezionym Śwętym Graalem każdy nowy zamek (i jego Okrągły Stół) będzie większy.",
   "Yuvarlak masa Şövalyeleri bu toprakların en müthiş savaşçıları. Kalelerine Kamelot demelerine bakılırsa, isimlerle araları pek iyi değil herhalde. Aralarına katılmaya büyük ihtimalle layıksın, ama yine de sana kendini kanıtlamak için bir görev verecekler... \n\n Her kale merkezinde tek bir hazineyi, Kutsal Kâseyi barındırıyor. Yuvarlak masanın çapı genelde 28'dir, ama Kutsal Kâseyi bir kez bulduğunda, karşılaştığın her yeni kale (ve her yuvarlak masa) eskisinden daha büyük olacak.",
   "Rytíři Kulatého stolu jsou největší bojovníci těchto krajů. Ale nemají zrovna talent na vymýšlení jmen, protože všem svým hradům říkají Camelot. Pravděpodobně jsi hoden toho, aby ses k nim připojil, ale určitě budeš muset splnit nějaký úkol, abys jim to dokázal...\n\nNa každém hradě je jeden poklad (Svatý grál), který se nachází v jeho středu. Poloměr Kulatého stolu je obvykle 28, ale poté, co najdeš Svatý grál, bude každý další hrad (a každý další Kulatý stůl) větší a větší.",
   "Рыцари Круглого стола - величайшие воины этой земли. Они не очень изобретательны и называют каждый свой замок Камелот. Возможно, Вам стоит встретиться с ними, и они дадут Вам задание.\n\nКаждый замок содержит одно сокровище (Святой Грааль) в центре. Обычно радиус круглого стола равен 28, но с каждым новым Граалем радиус становится всё больше.",
   "Die Ritter der Tafelrunde sind die größten Krieger dieses Landes. Sie sind nur nicht sehr einfallsreich was Namen angeht, alle ihre Schlösser heißen Camelot. Bevor du ihnen beitreten kannst, wirst du ihnen beweisen müssen dass du ihrer würdig bist.\n\nIn der Mitte jedes Schlosses befindet sich ein einziger Schatz, der Heilige Gral. Der Radius einer Tafelrunde ist normalerweise 28, allerdings wächst dieser sowie das Schloss mit jedem Heiligem Gral den du sammelst.",
/*MISSING*/    "The Knights of the Round Table are the greatest warriors of these lands. They are not very inventive with names though, as they call each of their castles Camelot. You are probably worthy of joining them, but they will surely give you some quest to prove yourself...\n\nEach castle contains a single treasure, the Holy Grail, in the center. The radius of the Round Table is usually 28, but after you find a Holy Grail successfully, each new castle (and each Round Table) you find will be bigger.",
    }},
  {0xdeed1b81, { // "paper model creator"
   "kreator modeli papierowych",
   "kâğıt modeli üreticisi",
   "generátor papírových modelů",
   "создатель бумажных моделей",
   "Papiermodell-Schöpfer",
/*MISSING*/    "paper model creator",
    }},
  {0xdef3113d, { // "hair color"
   "kolor włosów",
   "saç rengi",
   "barva vlasů",
   "цвет волос",
   "Haarfarbe",
/*MISSING*/    "hair color",
    }},
  {0xdf145984, { // "The ivy attacks %the1!"
   "Bluszcz zaatakował %a1!",
/*MISSING*/    "The ivy attacks %the1!",
   "Břečťan zaútočil na %a1!",
   "Плющ атаковал %a1!",
   "Der Efeu greift %den1 %a1 an!",
/*MISSING*/    "The ivy attacks %the1!",
    }},
  {0xdf6b66d1, { // "frames per second"
   "klatki na sekundę",
   "saniye başına kare",
   "snímky za sekundu",
   "кадров в секунду",
   "FPS",
   "quadros por segundo",
    }},
  {0xdf88bdd4, { // "shift+O to switch anti-aliasing"
   "shift+O by przełączyć antialiasing",
/*MISSING*/    "shift+O to switch anti-aliasing",
   "Stiskem Shift+O se přepíná antialiasing",
   "shift+O включает сглаживание",
/*MISSING*/    "shift+O to switch anti-aliasing",
/*MISSING*/    "shift+O to switch anti-aliasing",
    }},
  {0xdf9236a7, { // "This magical missile burns whatever it hits."
   "Ten magiczny pocisk pali to, w co trafi.",
   "Bu büyülü saldırı değdiği şeyi yakar.",
   "Tato magická střela spálí všechno, co zasáhne.",
   "Этот магический снаряд поджигает всё, во что попадает.",
   "Dieses magische Geschoss verbrennt alles, was es trifft.",
/*MISSING*/    "This magical missile burns whatever it hits.",
    }},
  {0xdfaed299, { // "Could not open the score file: "
   "Nie udało się otworzyć pliku: ",
   "Puan dosyası açılamadı: ",
   "Nepodařilo se otevřít soubor se skóre: ",
   "Невозможно открыть файл с результатами: ",
   "Highscores konnten nicht geladen werden: ",
   "Não foi possível abrir o arquivo de pontuação: ",
    }},
  {0xdfcb4b36, { // "F6 = HQ shot"
   "F6 = zrzut HQ",
   "F6 = HQ resim çek",
   "F6 = HQ snímek",
   "F6 = скриншот HQ",
   "F6 = HQ Screenshot",
/*MISSING*/    "F6 = HQ shot",
    }},
  {0xdff093bf, { // "When you have this Orb, you will leave a trail of fire behind you."
   "Kiedy masz Sferę Ognia, zostawiasz za sobą ognistą ścieżkę.",
   "Bu Küreye sahip olduğunda arkanda bir ateşten yol bırakırsın.",
   "Když máš tuto Sféru, zanecháváš za sebou ohnivou stopu.",
   "Пока у Вас есть эта сфера, Вы оставляете огненный след за собой.",
   "Mit diesem Orb hinterlässt du eine Spur aus Feuer.",
/*MISSING*/    "When you have this Orb, you will leave a trail of fire behind you.",
    }},
  {0xe0214c82, { // "Drawing %1 (layer %2), F1 for help"
   "Rysujemy %1 (warstwa %2), F1 - pomoc",
   "%1 çiziliyor (katman %2), F1 - yardım",
   "Kreslím %1 (vrstva %2), F1 - nápověda",
   "Рисуем %1 (слой %2), F1 - помощь",
   "Zeichne %1 (Schicht %2), F1 für Hilfe",
   "Desenhando %1 (camada %2), F1 para ajuda",
    }},
  {0xe029a2ec, { // "This person loves to look at their own reflection in the mirror. He believes himself to be one of the most important creatures in this world, and hates those who do not admire him."
   "Ten osobnik kocha patrzeć na własne lustrzane odbicie. Uważa się za jedną z najważniejszych istot na świecie, i nienawidzi tych, którzy go nie podziwiają.",
/*MISSING*/    "This person loves to look at their own reflection in the mirror. He believes himself to be one of the most important creatures in this world, and hates those who do not admire him.",
   "Tenhle èlovìk se rád kochá svým odrazem v zrcadle. Myslí si, že je jedním z nejdùležitìjších tvorù na svìtì a nenávidí ty, kteøí ho neobdivují.",
   "Этот человек любит глядеть на своё отражение в зеркале. Он верит, что он -- один из самых важных в существ этого мира, и ненавидит любого, кто не восхищается им.",
/*MISSING*/    "This person loves to look at their own reflection in the mirror. He believes himself to be one of the most important creatures in this world, and hates those who do not admire him.",
/*MISSING*/    "This person loves to look at their own reflection in the mirror. He believes himself to be one of the most important creatures in this world, and hates those who do not admire him.",
    }},
  {0xe076f634, { // " %1"
   " %1",
/*MISSING*/    " %1",
   " %1",
   " %1",
   " %1",
/*MISSING*/    " %1",
    }},
  {0xe08be549, { // "Still confused? Read the FAQ on the HyperRogue website!\n\n"
   "Coś jest niejasne? Przeczytaj FAQ na stronie HyperRogue! (po angielsku)\n\n",
   "Hâlâ karışık mı geliyor? HyperRogue vebsitesinden SSS'i okuyun! \n\n",
   "Stále zmatení? Přečtěte si FAQ na webové stránce HyperRogue (prozatím pouze v angličtině)\n\n",
   "Что-т неясно? Прочитайте FAQ на сайте HyperRogue! (по-английски)\n\n",
   "Immer noch verwirrt? Lies das FAQ auf der HyperRogue-Webseite.\n\n",
/*MISSING*/    "Still confused? Read the FAQ on the HyperRogue website!\n\n",
    }},
  {0xe09e0a49, { // "Alternatively: kill a %1.\n"
   "Alternatywnie: zabij %a1.\n",
   "Ya da: Bir %1 öldür.\n",
   "Alternativně: zabij %a1.\n",
   "Альтернатива: убить %a1.\n",
   "Alternativ: töte einen %a1.\n",
/*MISSING*/    "Alternatively: kill a %1.\n",
    }},
  {0xe0be7240, { // "previous slide"
   "poprzedni slajd",
/*MISSING*/    "previous slide",
   "předešlý snímek",
   "предыдущий слайд",
/*MISSING*/    "previous slide",
/*MISSING*/    "previous slide",
    }},
  {0xe0c51c51, { // "The slime reacts with %the1!"
   "Maź reaguje z %abl1!",
   "Balçık %abl1 etkileşime girdi!",
   "Sliz reaguje s %abl1!",
   "Слизь реагирует с %abl1!",
   "Der Schleim reagiert mit %dem %d1!",
   "O lodo reage com %na1!",
    }},
  {0xe0dd1448, { // "This land is filled with beautiful, but dangerous, creatures and plants."
   "Ta kraina pełna jest pięknych, lecz niebezpiecznych, stworzeń i roślin.",
/*MISSING*/    "This land is filled with beautiful, but dangerous, creatures and plants.",
   "Tento kraj je plný krásných, ale nebezpečných tvorů a rostlin.",
   "Эта земля полна прекрасными, но опасными существами и растениями.",
   "Dieses Land ist voller schöner, aber gefährlicher Wesen und Pflanzen.",
/*MISSING*/    "This land is filled with beautiful, but dangerous, creatures and plants.",
    }},
  {0xe15c3d8f, { // "You have used less knives than ever before. Congratulations!"
   "Zużyłeś mniej noży niż kiedykolwiek wcześniej. Gratulacje!",
   "Her zamankinden daha az bıçak kullandın. Tebrikler!",
   "Použil jsi méně nožů než kdykoli předtím. Gratulujeme!",
   "Вы использовали меньше ножей, чем раньше. Поздравляем!",
   "Du hast weniger Messer als zuvor benutzt. Glückwunsch!",
/*MISSING*/    "You have used less knives than ever before. Congratulations!",
    }},
  {0xe16601a8, { // "This is the biggest Hypersian Rug you have ever seen! Unfortunately, it is too large to take it as a trophy."
   "To jest największy hiperski dywan, jaki kiedykolwiek widzia%łeś0! Niestety, jest za duży, by wziąć go jako zdobycz.",
   "Bu, gördüğün en büyük Aşkınlı Kilim! Maalesef yanında götüremeyeceğin kadar büyük.",
   "Tohle je ten největší Hyperský koberec, jaký jsi kdy vidě%l0!  Bohužel je ale příliš velký na to, aby sis ho moh%l0 vzít jako trofej.",
   "Это самый большой гиперсидский ковёр, который Вы когда-либо видели! К сожалению, он слишком большой, чтобы забрать его.",
   "Das ist der größte Hypersische Teppich den du je gesehen hast! Leider ist er zu groß um ihn als Trophäe mitzunehmen.",
/*MISSING*/    "This is the biggest Hypersian Rug you have ever seen! Unfortunately, it is too large to take it as a trophy.",
    }},
  {0xe1d73b07, { // "(locked)"
   "(zablokowane)",
   "(aktif değil)",
   "(uzamčeno)",
   "(закрыто)",
   "(Gesperrt)",
/*MISSING*/    "(locked)",
    }},
  {0xe1dd5530, { // "player 1 spin"
   "gracz 1 obrót",
   "oyuncu 1, dön",
   "hráč 1 otáčení",
   "игрок 1 вертит",
   "Spieler 1 drehen",
/*MISSING*/    "player 1 spin",
    }},
  {0xe1ffed98, { // "Orb of Yendor gained!"
   "Zdobyta Sfera Yendoru!",
   "Yendorun Orbu kazandın!",
   "Získa%0l jsi Yendorskou sféru!",
   "Сферы Йендора получены!",
   "Orb von Yendor erhalten!",
   "Orbe de Yendor obtido!",
    }},
  {0xe21bf3fa, { // "Aethereal powers let you go through the holes."
   "Moc Eteru pozwala Ci przechodzić przez dziury.",
/*MISSING*/    "Aethereal powers let you go through the holes.",
   "Síla Éteru ti umožňuje pohybovat se přes díry.",
   "Силы эфира позволяют ходить сквозь пропасти.",
/*MISSING*/    "Aethereal powers let you go through the holes.",
/*MISSING*/    "Aethereal powers let you go through the holes.",
    }},
  {0xe21cf668, { // "shoot'em up mode"
   "tryb strzelanki",
   "tryb strzelanki",
   "mód střílečky",
   "режим стрельбы",
   "Ballermodus",
/*MISSING*/    "shoot'em up mode",
    }},
#if REPEATED
  {0xe21cf669, { // "shoot'em up mode [repeat]"
   "strzelanka",
/*MISSING*/    "shoot'em up mode [repeat]",
   "střílečka",
   "режим стрельбы",
   "Ballermodus",
/*MISSING*/    "shoot'em up mode [repeat]",
    }},
#endif
  {0xe23e1677, { // "%1 Challenge"
   "Misja %1",
   "%1 Görevi",
   "Mise: %1",
   "Миссия %1",
   "Herausforderung %1",
/*MISSING*/    "%1 Challenge",
    }},
  {0xe24ffa92, { // "config"
   "ustawione",
/*MISSING*/    "config",
   "konfigurace",
   "настройки",
/*MISSING*/    "config",
/*MISSING*/    "config",
    }},
  {0xe2626fb8, { // "press a key to unassign"
   "naciśnij klawisz do odwiązania",
   "atamayı iptal için bir tuşa bas ",
   "stiskni klávesu, kterou chceš zrušit",
   "нажмите, чтобы отменить назначение",
   "Drücke eine Taste zum entbinden",
/*MISSING*/    "press a key to unassign",
    }},
  {0xe270cb42, { // "You summon an Ivy!"
   "Przywoła%łeś0 Bluszcz!",
   "Bir sarmaşık oluşturdun!",
   "Vyvola%l0 jsi Břečťan!",
   "Вы призвали плющ!",
   "Du beschwörst Efeu!",
   "Você invoca uma Hera!",
    }},
  {0xe28e78c4, { // "You cannot target that far away!"
   "Nie możesz celować tak daleko!",
   "O kadar uzağı hedefleyemezsin.",
   "Nemůžeš zamířit tak daleko!",
   "Nie możesz celować tak daleko!",
   "Zu weit weg!",
/*MISSING*/    "You cannot target that far away!",
    }},
  {0xe2943d19, { // "explore the Palace Pattern (4 colors)"
   "Eksploracja Wzoru Pałac (4 kolory)",
   "Saray Dokusunu Keşfet (4 renk)",
   "Zkoumání Palácového vzoru (4 barvy)",
   "Изучить узор Дворца (4 цветов)",
   "Erkunde die Schlossstruktur (4 Farben)",
/*MISSING*/    "explore the Palace Pattern (4 colors)",
    }},
  {0xe2bd90fe, { // "The game is normally displayed in the so called Poincaré disk model, which is a kind of a map of the infinite hyperbolic world. There are many projections of Earth, but since Earth is curved, all of them have to distort distances or angles in some way -- the same is true in hyperbolic geometry. The next slide shows another model, called the Poincaré upper half-plane model. In this model, horocycles centered at one specific ideal point are drawn as straight lines."
   "Gra jest normalnie w tak zwanym modelu Poincaré, rodzaju mapy nieskończonego hiperbolicznego świata. Istnieje wiele rodzajów rzutu Ziemi, ale ze względu na jej krzywiznę, każdy z nich przekłamuje odległości czy kąty w jakiś sposób -- to samo dzieje się w geometrii hiperbolicznej. Następny slajd pokazuje inny model, model półpłaszczyzny. W tym modelu horocykle o środku w danym punkcie idealnym wyglądają jak linie poziome.",
/*MISSING*/    "The game is normally displayed in the so called Poincaré disk model, which is a kind of a map of the infinite hyperbolic world. There are many projections of Earth, but since Earth is curved, all of them have to distort distances or angles in some way -- the same is true in hyperbolic geometry. The next slide shows another model, called the Poincaré upper half-plane model. In this model, horocycles centered at one specific ideal point are drawn as straight lines.",
   "Hra se normálně zobrazuje v modelu nazývaném 'Poincarého disk', což je něco jako mapa nekonečného hyperbolického světa. Země má také mnoho projekcí, ale protože je zakřivená, musejí tyto projekce vždy nějakým způsobem zkreslovat vzdálenosti nebo úhly -- a v hyperbolické geometrii platí totéž. Další snímek ti ukáže jiný model, zvaný 'Poincarého model horní poloroviny'. V tomto modelu vypadají horocykly se středem v jednom specifickém ideálním bodě jako přímky.",
   "Обычно игра отображается в так называемой модели Пуанкаре в диске, карте бесконечного гиперболического мира. Есть много проекций Земли, но, так как она искривлённая, все они искажают расстояния или углы каким-то образом -- то же самое верно и здесь. Следующий слайд изображает другую модель, называемую моделью верхней полуплоскости. В этой модели орициклы с центром в одной определённой идеальной точке изображаются прямыми линиями.",
/*MISSING*/    "The game is normally displayed in the so called Poincaré disk model, which is a kind of a map of the infinite hyperbolic world. There are many projections of Earth, but since Earth is curved, all of them have to distort distances or angles in some way -- the same is true in hyperbolic geometry. The next slide shows another model, called the Poincaré upper half-plane model. In this model, horocycles centered at one specific ideal point are drawn as straight lines.",
/*MISSING*/    "The game is normally displayed in the so called Poincaré disk model, which is a kind of a map of the infinite hyperbolic world. There are many projections of Earth, but since Earth is curved, all of them have to distort distances or angles in some way -- the same is true in hyperbolic geometry. The next slide shows another model, called the Poincaré upper half-plane model. In this model, horocycles centered at one specific ideal point are drawn as straight lines.",
    }},
  {0xe2d2f9d7, { // "auto"
   "auto",
   "otomatik",
   "auto",
   "авто",
   "auto",
   "automático",
    }},
  {0xe307b485, { // "time"
   "czas",
   "zaman",
   "čas",
   "время",
   "zeit",
   "tempo",
    }},
  {0xe31fd91e, { // "You glance at your precious treasure."
   "Patrzysz na swój drogocenny skarb.",
   "Hazinene şöyle bir baktın.",
   "Díváš se na svůj převzácný poklad.",
   "Вы смотрите на ваше драгоценное сокровище.",
   "Du betrachtest deine kostbaren Schätze.",
   "Você contempla seu precioso tesouro.",
    }},
  {0xe32a6ab7, { // "page"
   "strona",
/*MISSING*/    "page",
   "stránka",
   "страница",
/*MISSING*/    "page",
/*MISSING*/    "page",
    }},
  {0xe3441b20, { // "Euclidean/elliptic mode"
   "tryb euklidesowy/eliptyczny",
/*MISSING*/    "Euclidean/elliptic mode",
   "Eukleidovský/eliptický mód",
   "евклидов/эллиптический режим",
/*MISSING*/    "Euclidean/elliptic mode",
/*MISSING*/    "Euclidean/elliptic mode",
    }},
  {0xe34c8953, { // "about "
   "około ",
/*MISSING*/    "about ",
   "zhruba ",
   "примерно ",
/*MISSING*/    "about ",
/*MISSING*/    "about ",
    }},
  {0xe35613f4, { // "floor"
   "podłoga",
   "zemin",
   "podlaha",
   "пол",
   "Boden",
/*MISSING*/    "floor",
    }},
  {0xe35992ba, { // " With the keyboard, you can rotate the view for a similar effect (Page Up/Down)."
   " Na klawiaturze możesz obrócić widok, by uzyskać podobny efekt (Page Up/Down).",
   " Klavyeyle görüntüyü döndürerek benzer bir etki yakalayabilirsin. (Page Up/Down).",
   " Na klávesnici můžeš podobného efektu dosáhnout otočením hrací plochy (Page Up/Down).",
   " На клавиатуре пользуйтесь поворотом поля для того же эффекта (Page Up/Down).",
   " Du kannst die Ansicht mit der Tastatur rotieren; dies erzeugt einen ähnlichen Effekt (Bild auf/ab).",
/*MISSING*/    " With the keyboard, you can rotate the view for a similar effect (Page Up/Down).",
    }},
  {0xe3c7403d, { // "gain Ice Diamonds"
   "daj Lodowe Diamenty",
/*MISSING*/    "gain Ice Diamonds",
   "získej Ledové diamanty",
   "собери Ледяные алмазы",
/*MISSING*/    "gain Ice Diamonds",
/*MISSING*/    "gain Ice Diamonds",
    }},
  {0xe3fc1d5c, { // "configure player 1"
   "skonfiguruj gracza 1",
   "1. oyuncuyu ayarla",
   "konfigurace hráče 1",
   "настройки игрока 1",
   "Spieler 1 konfigurieren",
/*MISSING*/    "configure player 1",
    }},
  {0xe3fc1d5d, { // "configure player 2"
   "skonfiguruj gracza 2",
   "2. oyuncuyu ayarla",
   "konfigurace hráče 2",
   "настройки игрока 2",
   "Spieler 2 konfigurieren",
/*MISSING*/    "configure player 2",
    }},
  {0xe3fc1d5e, { // "configure player 3"
   "skonfiguruj gracza 3",
/*MISSING*/    "configure player 3",
   "konfigurace hráče 3",
   "настройки игрока 3",
   "Spieler 3 konfigurieren",
/*MISSING*/    "configure player 3",
    }},
  {0xe3fc1d5f, { // "configure player 4"
   "skonfiguruj gracza 4",
/*MISSING*/    "configure player 4",
   "konfigurace hráče 4",
   "настройки игрока 4",
   "Spieler 4 konfigurieren",
/*MISSING*/    "configure player 4",
    }},
  {0xe3fc1d60, { // "configure player 5"
   "konfiguruj gracza 5",
/*MISSING*/    "configure player 5",
   "konfigurace hráče 5",
   "настройки игрока 5",
   "Spieler 5 konfigurieren",
/*MISSING*/    "configure player 5",
    }},
  {0xe3fc1d61, { // "configure player 6"
   "konfiguruj gracza 6",
/*MISSING*/    "configure player 6",
   "konfigurace hráče 6",
   "настройки игрока 6",
   "Spieler 6 konfigurieren",
/*MISSING*/    "configure player 6",
    }},
  {0xe3fc1d62, { // "configure player 7"
   "konfiguruj gracza 7",
/*MISSING*/    "configure player 7",
   "konfigurace hráče 7",
   "настройки игрока 7",
   "Spieler 7 konfigurieren",
/*MISSING*/    "configure player 7",
    }},
  {0xe4388b4f, { // "continue"
   "kontynuuj",
/*MISSING*/    "continue",
   "pokračuj",
   "продолжить",
/*MISSING*/    "continue",
/*MISSING*/    "continue",
    }},
  {0xe45854fb, { // "You can find safety in some places in the Prairie, but if you want treasures, they can be found only on the other side of a giant herd of bulls."
   "Są na Prerii miejsca bezpieczne, ale jeśli szukasz skarbów, możesz je znaleźć tylko po drugiej stronie wielkiego stada biegnących byków.",
/*MISSING*/    "You can find safety in some places in the Prairie, but if you want treasures, they can be found only on the other side of a giant herd of bulls.",
   "Na některých místech Prérie je bezpečno, ale pokud chceš poklady, musíš je hledat na druhé straně obrovského stáda býků.",
   "В некоторых местах Прерии весьма безопасно, но если вы хотите сокровищ, их можно найти лишь на другой стороне от огромных стад быков.",
/*MISSING*/    "You can find safety in some places in the Prairie, but if you want treasures, they can be found only on the other side of a giant herd of bulls.",
/*MISSING*/    "You can find safety in some places in the Prairie, but if you want treasures, they can be found only on the other side of a giant herd of bulls.",
    }},
  {0xe469319b, { // "which grants the greatest enjoyment.\""
   "sprawia największą przyjemność.\"",
/*MISSING*/    "which grants the greatest enjoyment.\"",
   "poskytuje největší potěšení.\"",
   "который приносит большое удовольствие.\"",
/*MISSING*/    "which grants the greatest enjoyment.\"",
/*MISSING*/    "which grants the greatest enjoyment.\"",
    }},
  {0xe4aad319, { // "triheptagonal tesselation"
   "posadzka trójkątno-siedmiokątna",
/*MISSING*/    "triheptagonal tesselation",
   "troj-sedmiúhelníkové dělení",
   "замощение треугольниками и семиугольниками",
/*MISSING*/    "triheptagonal tesselation",
/*MISSING*/    "triheptagonal tesselation",
    }},
  {0xe50afb4c, { // "Gameboard"
   "Plansza",
   "Oyun Tahtası",
   "Herní desky",
   "Игровая доска",
   "Spielbrett",
/*MISSING*/    "Gameboard",
    }},
  {0xe523dc37, { // "Not available in the pure tactics mode!"
   "Niedostępne w trybie taktycznym!",
   "Taktik modunda kullanılamaz!",
   "Toto není v taktickém módu k dispozici!",
   "Недоступно в тактическом режиме!",
   "Nicht verfügbar im Strategiemodus!",
/*MISSING*/    "Not available in the pure tactics mode!",
    }},
  {0xe53ef0f2, { // "Press 'i' to access your magical powers."
   "Naciśnij 'i', by użyć mocy.",
/*MISSING*/    "Press 'i' to access your magical powers.",
   "Stiskem klávesy 'i' pøejdeš na svou magickou sílu.",
   "Нажми 'i' для доступа к своим магическим силам.",
/*MISSING*/    "Press 'i' to access your magical powers.",
/*MISSING*/    "Press 'i' to access your magical powers.",
    }},
  {0xe5512986, { // "(e.g. Shift+Ctrl+Z)"
   " (na przykład Shift+Ctrl+Z)",
   " (Meselâ Shift+Ctrl+Z)",
   "ladění (například Shift+Ctrl+Z)",
   "Ctrl для тонкой настройки",
   "Ändern (z.B. Strg+Shift+Z)",
   " (ex: Shift+Ctrl+Z)",
    }},
  {0xe560137d, { // "A huge plant growing in the Jungle. Each Ivy has many branches, and one branch grows per each of your moves. Branches grow in a clockwise order. The root itself is vulnerable."
   "Wielka roślina z dżungli. Bluszcz zazwyczaj ma wiele gałęzi, po każdym Twoim ruchu jedna z nich rośnie. Gałęzie rosną w kolejnosci zegarowej. Sam korzeń nie atakuje, zatem goły korzeń łatwo ściąć.\n",
   "Vahşi Ormanda büyüyen devasa bir bitki. Her sarmaşığın bir sürü dalı vardır ve senin her hareketinle bir dal büyür. Dallar saatyönü sırasında büyür. Kökün kendisi savunmasızdır. \n",
   "Obrovská rostlina, která roste v Džungli. Každý Břečťan má mnoho větví, a po každém tvém pohybu jedna z nich povyroste. Větve rostou v pořadí po směru hodinových ručiček. Kořen Břečťanu je také zranitelný.",
   "Огромное растение, растущее в Джунглях. Каждый плющ имеет много побегов, и один из них растёт каждый ход. Побеги растут по часовой стрелке. Корень плюща - самое уязвимое место.\n",
   "Eine riesige Pflanze, die im Dschungel wächst. Jeder Efeu hat viele Zweige, und jedes Mal, wenn du dich bewegst, wächst einer. Sie wachsen im Uhrzeigersinn. Die Wurzel selbst ist verwundbar.",
   "Uma grande planta que cresce na Selva. Cada Hera tem vários galhos, e cada galho cresce para cada um dos seus movimentos. Galhos crescem em sentido horário. A raiz é vulnerável.\n",
    }},
  {0xe5905e88, { // "Roughly 42% cells are on the edge of your sight range. Reducing the sight range makes HyperRogue work faster, but also makes the game effectively harder."
   "Około 42% pól jest na brzegu Twojego pola widzenia. Zmniejszenie zasięgu sprawia, że HyperRogue działa szybciej, ale rozgrywka jest trudniejsza.",
/*MISSING*/    "Roughly 42% cells are on the edge of your sight range. Reducing the sight range makes HyperRogue work faster, but also makes the game effectively harder.",
   "Zhruba 42% políček se nachází na poloměru tvého dohledu. Pokud snížíš dohled, bude HyperRogue běhat rychleji, ale také se tím ztíží.",
   "Примерно 42% клеток находятся в радиусе видимости. Уменьшение радиуса обзора ускоряет работу HyperRogue, но заметно усложняет игру.",
/*MISSING*/    "Roughly 42% cells are on the edge of your sight range. Reducing the sight range makes HyperRogue work faster, but also makes the game effectively harder.",
/*MISSING*/    "Roughly 42% cells are on the edge of your sight range. Reducing the sight range makes HyperRogue work faster, but also makes the game effectively harder.",
    }},
  {0xe5953994, { // "You feel the Orbs of Yendor everywhere..."
   "Wyczuwasz Sfery Yendoru, wszędzie...",
/*MISSING*/    "You feel the Orbs of Yendor everywhere...",
   "Cítíš Yendorské Sféry všude...",
   "Ты чувствуешь Сферы Йендора везде...",
/*MISSING*/    "You feel the Orbs of Yendor everywhere...",
/*MISSING*/    "You feel the Orbs of Yendor everywhere...",
    }},
  {0xe59fd741, { // "azimuthal equi-area"
   "azymutalny, równe pola",
/*MISSING*/    "azimuthal equi-area",
   "azimutální plochojevný",
   "азимутальная, сохранение площадей",
   "flächentreu azimutal",
/*MISSING*/    "azimuthal equi-area",
    }},
  {0xe5d1b8b3, { // "This tutorial is mostly aimed to show what is special about the geometry used by HyperRogue. It also shows the basics of gameplay, and how is it affected by geometry.\n\nYou decide when you want to stop playing with the current \"slide\" and go to the next one, by pressing Enter. You can also press ESC to see a menu with other options."
   "Zadaniem tego podręcznika jest pokazanie szczególnych własności geometrii używanej przez HyperRogue; są też pokazane podstawy rozgrywki, i jak geometria na nie wpływa.\n\nTy decydujesz, kiedy chcesz skończyć bawić się obecnym \"slajdem\" i przejść do kolejnego, naciskając Enter. Naciśnij ESC by zobaczyć pozostałe opcje.",
/*MISSING*/    "This tutorial is mostly aimed to show what is special about the geometry used by HyperRogue. It also shows the basics of gameplay, and how is it affected by geometry.\n\nYou decide when you want to stop playing with the current \"slide\" and go to the next one, by pressing Enter. You can also press ESC to see a menu with other options.",
   "Tento tutoriál je převážně zaměřen na to, aby ukázal, co je zvláštního na geometrii, kterou HyperRogue používá. Také ukazuje základy hry a to, jak ji geometrie ovlivňuje.\n\nKdyž si budeš chtít přestat hrát se současným \"snímkem\" a přejít na další, stiskni Enter. Stiskem ESC můžeš také zobrazit menu s dalšími možnostmi.",
   "Главная цель этого руководства -- показать особенности геометрии в HyperRogue. Также здесь объясняются основы игры, и как геометрия на них влияет.\n\nТы можешь перейти на следующий \"слайд\", нажав Enter. Нажми ESC, чтобы открыть меню с другими опциями.",
/*MISSING*/    "This tutorial is mostly aimed to show what is special about the geometry used by HyperRogue. It also shows the basics of gameplay, and how is it affected by geometry.\n\nYou decide when you want to stop playing with the current \"slide\" and go to the next one, by pressing Enter. You can also press ESC to see a menu with other options.",
/*MISSING*/    "This tutorial is mostly aimed to show what is special about the geometry used by HyperRogue. It also shows the basics of gameplay, and how is it affected by geometry.\n\nYou decide when you want to stop playing with the current \"slide\" and go to the next one, by pressing Enter. You can also press ESC to see a menu with other options.",
    }},
  {0xe61f191b, { // "You overhear miners talking about a castle..."
   "Podsłucha%łeś0 górników rozmawiających o zamku...",
   "Madencilerin bir kale hakkında konuştuğunu duydun...",
   "Zaslechl jsi Horníky, jak mluví o nějakém hradu...",
   "Вы слышите шахтёров, говорящих о замке...",
   "Du hörst Minenarbeiter über ein Schloss sprechen...",
/*MISSING*/    "You overhear miners talking about a castle...",
    }},
  {0xe629d1bd, { // "You summon a sandworm!"
   "Przywoła%łeś0 Czerwia!",
   "Bir kumkurdu çağırdın!",
   "Vyvola%l0 jsi Písečného červa!",
   "Вы призвали червя!",
   "Du beschwörst einen Sandwurm!",
   "Você invoca uma minhoca de areia!",
    }},
  {0xe686d5dc, { // "Cannot jump through %the1!"
   "Nie możesz skakać przez %a1!",
   "%1 hücresinden zıplayamazsın!",
   "Nemůžeš skočit přes %a1!",
   "Нельзя прыгать через %a1!",
   "Du kannst nicht durch %den1 %a1 springen!",
/*MISSING*/    "Cannot jump through %the1!",
    }},
  {0xe6a912ff, { // "activate Orb power"
   "aktywuj moc Sfery",
   "Küre gücünü aktif et",
   "aktivace schopnosti Sféry",
   "активировать сферу",
   "Orb aktivieren",
/*MISSING*/    "activate Orb power",
    }},
  {0xe74836ff, { // ", m - mode: hardcore only"
   ", m - tryb hardcore",
   ", m - aşmış mod",
   ", m - pouze hardcore mód",
   ", m - режим hardcore",
   ", m - Modus: Hardcore",
/*MISSING*/    ", m - mode: hardcore only",
    }},
  {0xe775008a, { // "Prove yourself here, then find new lands, with new quests..."
   "Pokaż, co potrafisz tutaj, potem znajdź nowe krainy i misje...",
   "Kendini burada kanıtla, sonra yeni görevlerle yeni diyarlar keşfet...",
   "Ukaž tady, co umíš, a pak najdi nové kraje a nové úkoly...",
   "Проверь себя здесь, а потом ищи новые земли с новыми миссиями",
   "Beweise hier deine Fähigkeiten und finde dann neue Länder mit neuen Aufgaben....",
   "Prove-se aqui, então encontre novas terras, com novas missões...",
    }},
  {0xe7819731, { // "You do not know exactly why, but there are valuable pearls on many boats in the whirlpool."
   "Właściwie nie wiesz, dlaczego, ale na wielu statkach w Wirze można znaleźć drogocenne perły.",
   "Nedeni bilinmez ama, pek tayfunun içindeki pek çok botta değerli inciler var.",
   "Nevíš tak úplně jistě proč, ale na mnoha člunech ve Víru jsou cenné perly.",
   "Неизвестно, почему, но этих жемчужин довольно много в лодках в этом водовороте.",
   "Du weißt nicht genau warum, aber es gibt schöne Perlen auf vielen Booten im Strudel.",
/*MISSING*/    "You do not know exactly why, but there are valuable pearls on many boats in the whirlpool.",
    }},
  {0xe7ab2a66, { // "Gravitational anomalies in the Jungle create mountains overgrown with ivies and bushes. Will you dare to climb the ivies to get the amethysts hidden above?\n\nCells adjacent to Ivies count as stable (but Ivies cannot climb themselves or other Ivies)."
   "Anomalia grawitacyjna w Dżungli stworzyła górę porośniętą bluszczem i krzakami. Czy odważysz się wspiąć po bluszczu, by zdobyć ukryte na niej ametysty?\n\nPola sąsiadujące z Bluszczem liczą się jako stabilne (ale Bluszcz nie może się wspinać po samym sobie ani po innych bluszczach).",
/*MISSING*/    "Gravitational anomalies in the Jungle create mountains overgrown with ivies and bushes. Will you dare to climb the ivies to get the amethysts hidden above?\n\nCells adjacent to Ivies count as stable (but Ivies cannot climb themselves or other Ivies).",
   "Gravitační anomálie v Džungli vytvářejí hory zarostlé břečťanem a keři. Odvážíš se vylézt na břečťany, abys získal ametysty ukryté nahoře?\n\nVšechny políčka, která sousedí s Břečťanem, se berou jako stabilní (Břečťan ale nemůže šplhat po sobě ani po jiných Břečťanech).",
   "Гравитационная аномалия в Джунглях создала гору, заросшую плющом и кустами. Рискнёте ли вы продраться через плющ и найти спрятанные здесь аметисты?\n\nКлетки рядом с плющом считаются стабильными (однако плющ не может расти вдоль себя или другого плюща.",
/*MISSING*/    "Gravitational anomalies in the Jungle create mountains overgrown with ivies and bushes. Will you dare to climb the ivies to get the amethysts hidden above?\n\nCells adjacent to Ivies count as stable (but Ivies cannot climb themselves or other Ivies).",
/*MISSING*/    "Gravitational anomalies in the Jungle create mountains overgrown with ivies and bushes. Will you dare to climb the ivies to get the amethysts hidden above?\n\nCells adjacent to Ivies count as stable (but Ivies cannot climb themselves or other Ivies).",
    }},
  {0xe7b428d4, { // "F2 = save"
   "F2 = zapisz",
   "F2 = kaydet",
   "F2 = uložení",
   "F2 = записать",
   "F2 = Speichern",
/*MISSING*/    "F2 = save",
    }},
  {0xe819ab65, { // "more curved hyperbolic geometry"
   "większa krzywizna",
/*MISSING*/    "more curved hyperbolic geometry",
   "zakřivenější hyperbolická geometrie",
   "более искривлённая геометрия",
/*MISSING*/    "more curved hyperbolic geometry",
/*MISSING*/    "more curved hyperbolic geometry",
    }},
  {0xe86acfd5, { // "\n\nThis Orb protects you from attacks, scents, and insulates you from electricity. It does not let you go through deadly terrain, but if you are attacked with fire, it lets you stay in place in it."
   "\n\nTa Sfera chroni Cię przed atakami, zapachami, i izoluje od elektryczności. Nie pozwala iść przez zabójczy teren, ale jeśli zostaniesz zaatakowan%y0 ogniem, możesz stać w miejscu.",
/*MISSING*/    "\n\nThis Orb protects you from attacks, scents, and insulates you from electricity. It does not let you go through deadly terrain, but if you are attacked with fire, it lets you stay in place in it.",
   "\n\nTato sféra te chrání před útoky a pachy a izoluje tě před elektřinou. Neumožňuje procházet smrtícím terénem, ale pokud na tebe někdo zaútočí ohněm, můžeš v něm zůstat stát.",
   "\n\nЭта сфера защищает вас от атак, запахов, изолирует от электричества. Она не позволяет вам пойти в опасную клетку, но если вы атакованы огнём, вы можете стоять на месте.",
   "\n\nDieser Orb schützt dich vor Angriffen, Düften, und isoliert dich gegenüber Elektrizität. Er lässt dich nicht durch tödliches Terrain laufen, aber wenn du mit Feuer angegriffen wirst, kannst du darin stehen bleiben.",
/*MISSING*/    "\n\nThis Orb protects you from attacks, scents, and insulates you from electricity. It does not let you go through deadly terrain, but if you are attacked with fire, it lets you stay in place in it.",
    }},
  {0xe89063bd, { // "You need to dismount %the1!"
   "Musisz opuścić %a1!",
/*MISSING*/    "You need to dismount %the1!",
   "Musíš seskočit!",
   "Вы должны слезть с %a1!",
   "Du musst von %dem1 %a1 absteigen!",
/*MISSING*/    "You need to dismount %the1!",
    }},
  {0xe8dc5933, { // "\"We live in a beautiful and orderly world, "
   "\"Żyjemy w pięknym i uporządkowanym świecie, ",
/*MISSING*/    "\"We live in a beautiful and orderly world, ",
   "\"Žijeme v krásném a uspořádaném světě, ",
   "\"Мы живём в прекрасном и упорядоченном мире, ",
/*MISSING*/    "\"We live in a beautiful and orderly world, ",
/*MISSING*/    "\"We live in a beautiful and orderly world, ",
    }},
  {0xe8f470be, { // "Your total wealth"
   "Twoje łączne bogactwo",
   "Toplam servetiniz",
   "Tvé celkové bohatství",
   "Ваше богатство",
   "Dein Gesamtvermögen",
   "Sua riqueza total",
    }},
  {0xe966825b, { // "A beautiful gem from the Jungle."
   "Piękny klejnot z Dżungli.",
   "Vahşi Ormandan elde edilmiş bir mücevher.",
   "Nádherný drahokam z Džungle.",
   "Прекрасный камень из Джунглей.",
   "Ein schöner Edelstein aus dem Dschungel.",
/*MISSING*/    "A beautiful gem from the Jungle.",
    }},
  {0xe96dc1b1, { // "all"
   "wszystko",
/*MISSING*/    "all",
   "vše",
   "все",
/*MISSING*/    "all",
/*MISSING*/    "all",
    }},
  {0xe970675e, { // "date"
   "data",
   "tarih",
   "datum",
   "дата",
   "Datum",
   "data",
    }},
  {0xe990670b, { // "background music"
   "muzyka w tle",
/*MISSING*/    "background music",
   "hudba",
   "фоновая музыка",
   "Hintergrundmusik",
/*MISSING*/    "background music",
    }},
  {0xe9914d4e, { // "polygon sides"
   "boki wielokąta",
/*MISSING*/    "polygon sides",
   "strany mnohoúhelníku",
   "стороны многоугольника",
   "Polygon-Seiten",
/*MISSING*/    "polygon sides",
    }},
  {0xe9e88caa, { // "A clearing in the Overgrown Woods. Obviously, this gives the Mutant Ivies an infinite space to grow...\n\nMutant Fruits rot if they are not adjacent to a Mutant Ivy."
   "Polana w Przerośniętym Lesie. Oczywiście, pozwala Zmutowanym Bluszczom rosnąć w nieskończoność...\n\nZmutowane Owoce gniją, gdy nie sąsiadują ze Zmutowanym Bluszczem.",
/*MISSING*/    "A clearing in the Overgrown Woods. Obviously, this gives the Mutant Ivies an infinite space to grow...\n\nMutant Fruits rot if they are not adjacent to a Mutant Ivy.",
   "Mýtina v Pralese. Mutantní břečťan tu samozřejmě má nekonečný prostor k růstu...\n\nMutantní ovoce shnije, pokud se nenachází vedle Mutantního břečťanu.",
   "Поляна в Зарослях. Представляет собой неограниченное пространство для роста плющей-мутантов...\n\nМутировавшие фрукты гниют, если они не находятся рядом с плющом.",
   "Eine Lichtung in den verwilderten Wäldern. Offensichtlich gibt sie den mutierten Efeuen unbegrenzten Platz zu wachsen...\n\nMutierte Früchte verfaulen, wenn sie sich nicht direkt neben einem mutierten Efeu befinden.",
/*MISSING*/    "A clearing in the Overgrown Woods. Obviously, this gives the Mutant Ivies an infinite space to grow...\n\nMutant Fruits rot if they are not adjacent to a Mutant Ivy.",
    }},
  {0xe9f539aa, { // "display pattern codes (full)"
   "pokaż kody wzorku (pełne)",
   "doku kodlarını göster (tam)",
   "zobraz kódy vzoru (úplné)",
   "показать коды узоров (полные)",
   "Muster Codes anzeigen (komplett)",
/*MISSING*/    "display pattern codes (full)",
    }},
  {0xea43495e, { // "player 1 go"
   "gracz 1 idź",
   "oyuncu 1, git",
   "hráč 1 pohyb",
   "игрок 1 идёт",
   "Spieler 1 los",
/*MISSING*/    "player 1 go",
    }},
  {0xea811d40, { // "This Orb lets you dive into the water. While diving, you are able to see and collect underwater treasures, while being invisible to monsters yourself. You are afraid of jumping into the water straight from a boat, so usually you won't do it."
   "Ta Sfera pozwala Ci nurkować. Podczas nurkowania możesz widzieć i zbierać podwodne skarby, a potwory Cię nie widzą. Boisz się wyskakiwać do wody z łódki, także raczej tego nie zrobisz.",
   "Bu Küre suya dalabilmeni sağlar. Suyun altındayken suyun altını görebilir ve sualtı hazinelerini toplayabilirsin. Bunu yaparken canavarlar seni göremez.Korkun sebebiyle bir kayıktan suya genelde doğrudan atlayamıyorsun.",
   "Tato Sféra ti umožňuje potopit se pod vodu. Když jsi potopený, můžeš vidět a získávat podvodní poklady, zatímco netvoři tě nevidí. Obvykle ale nebudeš skákat do vody přímo ze člunu -- to tě trochu děsí.",
   "Эта сфера позволяет нырять. Когда Вы ныряете, Вы видите и можете собирать подводные ценности, а враги Вас не видят. Но Вы боитесь прягать из лодки прямо в воду.",
   "Dieser Orb erlaubt es dir zu tauchen. Während du tauchst kannst du versunkene Schätze sammeln, dabei bist du für Monster unsichtbar. Du hast Angst von einem Boot aus ins Meer zu springen daher tust du es meist nicht.",
/*MISSING*/    "This Orb lets you dive into the water. While diving, you are able to see and collect underwater treasures, while being invisible to monsters yourself. You are afraid of jumping into the water straight from a boat, so usually you won't do it.",
    }},
  {0xeaaeb0a1, { // "+5 = center instantly, -5 = do not center the map"
   "+5 = centruj natychmiast, -5 = nie centruj",
   "+5 = anında merkeze al, -5 = merkezleme",
   "+5 = okamžité vycentrování, -5 = žádné centrování",
   "+5 = центрировать сразу, -5 = не центрировать",
   "+5 = sofort zentrieren, -5 = nicht zentrieren",
   "+5 = centralizar instantaneamente, -5 = não centralizar o mapa",
    }},
  {0xeb30aac0, { // "A fruit from the Yendorian Forest."
   "Owoc z Yendoriańskiego Lasu.",
/*MISSING*/    "A fruit from the Yendorian Forest.",
   "Ovoce z Yendorského lesa.",
   "Фрукт из Йендорского леса.",
   "Eine Frucht des Yendorianischen Waldes.",
/*MISSING*/    "A fruit from the Yendorian Forest.",
    }},
  {0xeb9ca8cf, { // "see the help screen"
   "obejrzyj ekran pomocy",
   "yardım ekranına bak.",
   "zobrazit obrazovku s nápovědou",
   "смотри экран помощи",
   "siehe Hilfe Bildschirm",
   "veja a tela de ajuda",
    }},
  {0xebaeb90b, { // "The Hive is filled with Hyperbugs. They are huge insects which look a bit like ants, a bit like bees, and a bit like roaches. They live in circular nests, and an army of Hyperbugs will attack any intruder, including you and enemy Hyperbugs. Will you manage to get to the heart of such a nest, and get the precious Royal Jelly?"
   "Mrowisko jest wypełnione Hiperinsektami. Te wielkie owady wyglądają trochę jak mrówki, trochę jak pszczoły, trochę jak karaluchy. Żyją w okrągłych gniazdach, i atakują każdego najeźdźcę, łącznie z Tobą i wrogimi hiperinsektami. Czy uda Ci się dotrzeć do serca roju i zdobyć cenne królewskie mleczko?",
   "Kovan aşkın böceklerle dolu. Aşkınböcekler biraz arıya biraz karıncaya biraz da hamamböceğine benzeyen kocaman böcekler. Dairesel kovanlarda yaşıyorlar ve yaklaşan düşman aşkınböceklerineve sana saldıracaklardır. Kovanın merkezine ulaşıp Müthiş Jöleyi ele geçirebilecek misin ?",
   "Úl je plný Hyperhmyzáků. Tento obří hmyz vypadá trochu jako mravenci, trochu jako včely a trochu jako švábi. Žijí v kruhových úlech a jejich armáda zaútočí na každého vetřelce včetně tebe a nepřátelských Hyperhmyzáků. Dokážeš se dostat do srdce takového kruhového úlu a získat vzácnou Mateří kašičku?",
   "Улей населён гипержуками. Эти огромные насекомые выглядят немного как муравьи, немного как пчёлы и немного как тараканы. Они живут в круглых гнёздах, и их армия готова напасть на любого нарушителя, включая Вас и жуков другого цвета. Хотите ли Вы добраться до центра такого гнезда и забрать ценное маточное молочко?",
   "Das Nest ist voller Hyperkäfer. Riesige Insekten, die ein bisschen wie Ameisen, Bienen oder Kakerlaken aussehen. Sie leben in kreisförmigen Nestern, und eine Armee von ihnen greift sämtliche Eindringlinge an - auch dich und feindliche Hyperkäfer. Schaffst du es zum Zentrum eines Nests, um das wertvolle Gelée Royale zu ergattern?",
/*MISSING*/    "The Hive is filled with Hyperbugs. They are huge insects which look a bit like ants, a bit like bees, and a bit like roaches. They live in circular nests, and an army of Hyperbugs will attack any intruder, including you and enemy Hyperbugs. Will you manage to get to the heart of such a nest, and get the precious Royal Jelly?",
    }},
  {0xebc24350, { // "Your %1 activates!"
   "Uruchomi%ł1 się Twoja %1!",
/*MISSING*/    "Your %1 activates!",
   "Tvoje %1 se aktivovala!",
   "Ваша %1 сработала!",
   "Dein %1 aktiviert sich!",
/*MISSING*/    "Your %1 activates!",
    }},
  {0xebd2779b, { // "this Orb never appears in %the1"
   "ta Sfera nie pojawia się %abl1",
   "Bu Küre %abl1 hiç bulunmaz",
   "tato Sféra, se %abl1 nikdy neobjevuje",
   "эта сфера никогда не появляется %abl1",
   "Dieser Orb erscheint nie in %der1",
/*MISSING*/    "this Orb never appears in %the1",
    }},
  {0xebdeec21, { // "Still, even greater treasures lie ahead..."
   "Jeszcze większe skarby przed Tobą...",
   "İleride daha bile çok hazine seni bekliyor...",
   "Ale dál leží ještě větší poklady...",
   "Большая часть сокровищ ещё впереди...",
   "Dich erwarten größere Schätze auf deinem Weg",
   "Maiores tesouros ainda estão à frente...",
    }},
  {0xebff281a, { // "No info about this..."
   "Brak informacji o tym...",
   "Hakkında bir bilgi yok...",
   "O tomhle nemám žádné informace...",
   "Нет информации об этом...",
   "Keine Info hierüber...",
   "Sem informações a respeito...",
    }},
  {0xec03df26, { // "This land is a quick gateway to other lands. It is very easy to find other lands from the Crossroads. Which means that you find monsters from most other lands here!\n\nAs long as you have found enough treasure in their native lands, you can find magical items in the Crossroads. Mirror Land brings mirrors and clouds, and other land types bring magical orbs.\n\nA special treasure, Hyperstone, can be found on the Crossroads, but only after you have found 10 of every other treasure."
   "Ta kraina jest szybkim przejściem do pozostałych krain.\n\nBardzo łatwo stąd dostać się do większości miejsc, ale można też tu spotkać potwory z różnych stron!\n\nMożesz znaleźć magiczne przedmioty na Skrzyżowaniu, jeśli zdoby%łeś0 wystarczająco wiele skarbów w ich rodzinnych krainach. Są to sfery, magiczne lustra i chmury.\n\nSpecyficzne skarby Skrzyżowań, Hiperkamienie, można znaleźć na Skrzyżowaniu, ale tylko po znalezieniu 10 jednostek każdego innego skarbu.",
   "Bu diyar diğer diyarlara kolay geçiş için var. Arayollardan başka diyarları bulması çok kolay. Bu demektir ki, burada başka diyarlardan canavarlar bulman çok olası!\n\nEğer hazineleri kendi yerlerinde yeterince topladıysan, Arayollarda sihirli eşyalar bulabilirsin. Ayanalar Diyarından yansımalar ve bulutlar, diğer diyarlardan sihirli küreler gelebilir.\n\nÖzel bir hazine, Aşkıntaş, sadece Arayollarda bulnabilir, ama öncesinde diğer her hazine çeşidinden 10'ar tane bulman gerekiyor.",
   "Tento kraj je rychlou branou do jiných krajů. Na Křižovatce je velmi snadné najít jiné kraje. Což znamená, že tu najdeš i netvory z většiny jiných krajů!\n\nNa Křižovatce můžeš najít magické předměty, pokud jsi v jejich krajích nalez%l0 dost pokladů. Můžeš tu najít zrcadla a oblaky z Kraje zrcadel a magické sféry z jiných typů krajů.\n\nNa Křižovatce můžeš najít zvláštní poklad -- Hyperkam -- ale pouze poté, co najdeš 10 kusů od všech ostatních pokladů.",
   "Эта земля - кратчайший путь в любую другую. Здесь очень легко найти любую другую землю, и здесь можно встретить монстров изо всех земель!\n\nКак только Вы набираете достаточно сокровищ в некоторой земле, её магические предметы становятся доступны здесь. В Зеркальной земле это зеркала и миражи, в остальных - сферы.\n\nСвой сокровище, Гиперкамень, появится лишь после того, как Вы соберёте по 10 сокровищ каждого типа.",
   "Dieses Land ist eine schnelle Verbindung zu anderen Ländern. Es ist sehr einfach, an den Kreuzungen andere Länder zu finden. Das heißt, dass du hier auch Monstern aus den meisten anderen Ländern begegnen wirst!\n\nSobald du genug Schätze in den jeweiligen Ländern gefunden hast, erscheinen deren magische Gegenstände auf den Kreuzungen. Spiegel und Wolken stammen aus dem Spiegelland, andere magische Orbs aus anderen Ländern.\n\nEin besonderer Schatz, der Hyperstein, kann auf den Kreuzungen gefunden werden. Aber du musst zuerst 10 von allen anderen Schätzen gesammelt haben.",
/*MISSING*/    "This land is a quick gateway to other lands. It is very easy to find other lands from the Crossroads. Which means that you find monsters from most other lands here!\n\nAs long as you have found enough treasure in their native lands, you can find magical items in the Crossroads. Mirror Land brings mirrors and clouds, and other land types bring magical orbs.\n\nA special treasure, Hyperstone, can be found on the Crossroads, but only after you have found 10 of every other treasure.",
    }},
  {0xec251dd1, { // "HELP"
   "POMOC",
   "YARDIM",
   "NÁPOVĚDA",
   "ПОМОШЬ",
   "HILFE",
/*MISSING*/    "HELP",
    }},
  {0xec29dcbd, { // "New players think that the action of HyperRogue takes place on a sphere. This is not true -- the next slide will show the surface HyperRogue actually takes place on.\n\nUse arrow keys to rotate the model, and Page Up/Down to zoom.\n\nIf you do not see anything, press '5' to try a safer renderer."
   "Nowi gracze myślą, że akcja HyperRogue toczy się na sferze. To nieprawda -- kolejny slajd pokazuje powierzchnię, na której toczy się gra. Obracaj modelem strzałkami, Page Up/Down by skalować.\n\nJeśli nic nie widzisz, naciśnij '5' dla bezpieczniejszego renderera.",
/*MISSING*/    "New players think that the action of HyperRogue takes place on a sphere. This is not true -- the next slide will show the surface HyperRogue actually takes place on.\n\nUse arrow keys to rotate the model, and Page Up/Down to zoom.\n\nIf you do not see anything, press '5' to try a safer renderer.",
   "Noví hráči si často myslí, že se HyperRogue odehrává na povrchu koule. Není to pravda -- další snímek ti ukáže povrch, na jakém se HyperRogue opravdu odehrává.\n\nModel můžeš otáčet šipkami a zoomovat klávesami Page Up/Down.\n\nPokud nic nevidíš, stiskni klávesu '5' pro aktivaci bezpečnějšího rendereru.",
   "Новички иногда думают, что действие в HyperRogue происходит на сфере. Это неправда -- на следующем слайде показана настоящая поверхность HyperRogue.\n\nИспользуйте стрелки, чтобы поворачивать модель, и Page Up/Down, чтобы менять размер.\n\nЕсли ничего не видно, нажми '5' для безопасного визуализатора.",
/*MISSING*/    "New players think that the action of HyperRogue takes place on a sphere. This is not true -- the next slide will show the surface HyperRogue actually takes place on.\n\nUse arrow keys to rotate the model, and Page Up/Down to zoom.\n\nIf you do not see anything, press '5' to try a safer renderer.",
/*MISSING*/    "New players think that the action of HyperRogue takes place on a sphere. This is not true -- the next slide will show the surface HyperRogue actually takes place on.\n\nUse arrow keys to rotate the model, and Page Up/Down to zoom.\n\nIf you do not see anything, press '5' to try a safer renderer.",
    }},
  {0xec5728d2, { // "A nasty creature native to the Living Caves. They don't like you for some reason."
   "Brzydki stwór z Żywych Jaskiń. Jakoś Cię nie lubi.",
   "Yaşayan Mağaralarda yaşayan iğrenç bir yaratık. Nedense seni hiç sevmiyor.",
   "Hnusná potvora žijící v Živoucích jeskyních. Z nějakého důvodu tě nemají rády.",
   "Противное существо из Живых пещер. По каким-то причинам не любит людей.",
   "Ein grässliches Geschöpf, das in den Lebenden Höhlen haust. Irgendwie können diese Kreaturen dich nicht leiden.",
/*MISSING*/    "A nasty creature native to the Living Caves. They don't like you for some reason.",
    }},
  {0xec6ad5de, { // "Two mines next to you!"
   "Dwie miny koło Ciebie!",
   "Yakınında iki mayın var!",
   "Ve tvé blízkosti se nacházejí dvě miny!",
   "Две мины рядом с Вами!",
   "2 Minen um dich!",
/*MISSING*/    "Two mines next to you!",
    }},
  {0xec94383d, { // "This orb can be used to invoke a wall of ice. It also protects you from fires."
   "Ta sfera zostawia za Tobą ścianę lodu, a także chroni Cię przed ogniem.",
   "Bu küre bir buz duvarı oluşturmaya yarar. Aynı zamanda seni ateşten korur",
   "Tuto sféru můžeš použít k vytvoření ledové zdi. Kromě toho tě chrání před ohněm.",
   "Эта сфера создаёт за Вами стену льда. Также защищает Вас от огня.",
   "Du kannst diesen Orb verwenden um einen Eiszauber zu wirken, der eine Wand aus Eis erschafft. Er schützt dich auch vor Feuer.",
/*MISSING*/    "This orb can be used to invoke a wall of ice. It also protects you from fires.",
    }},
  {0xecb8ebbf, { // "Nowhere to blow %the1!"
   "Nie ma gdzie zdmuchnąć %a1!",
   "%a1! ittirecek bir yer yok!",
   "%a1 není kam odfouknout!",
   "Некуда сдуть %a1!",
   "Kein Platz um %den1 %a1 wegzuwehen!",
/*MISSING*/    "Nowhere to blow %the1!",
    }},
  {0xece1dc90, { // "Joystick %1, axis %2"
   "Joystick %1, oś %2",
   "Joystick %1, eksen %2",
   "Joystick %1, osa %2",
   "Джойстик %1, ось %2",
   "Joystick %1, Achse %2",
/*MISSING*/    "Joystick %1, axis %2",
    }},
  {0xecef3333, { // "That was easy, but groups could be dangerous."
   "To było proste, z grupami może być trudniej.",
   "Bu kolaydı, ama gruplar daha tehlikeli olabilir.",
   "To bylo snadné, ale skupiny by mohly být nebezpečné.",
   "Они слабы, но в стае могут быть опасны.",
   "Das war einfach, aber mehrere könnten gefährlich sein.",
   "Isso foi fácil, mas grupos podem ser perigosos.",
    }},
  {0xed1ff286, { // "ϕ: %1°"
   "ϕ: %1°",
   "ϕ: %1°",
   "ϕ: %1°",
   "ϕ: %1°",
   "ϕ: %1°",
/*MISSING*/    "ϕ: %1°",
    }},
  {0xed2fcf78, { // "Preparing the line (%1/1000)..."
   "Przygotowanie linii (%1/1000)...",
/*MISSING*/    "Preparing the line (%1/1000)...",
   "Připravuji přímku (%1/1000)...",
   "Приготовление линии (%1/1000)...",
   "Vorbereitung der Strecke (%1/1000)...",
/*MISSING*/    "Preparing the line (%1/1000)...",
    }},
  {0xed45567c, { // "\"I do not like butterflies. They are treacherous.\""
   "\"Nie lubię motyli. Są zdradzieckie.\"",
/*MISSING*/    "\"I do not like butterflies. They are treacherous.\"",
   "\"Nesnáším motýly. Jsou zrádní.\"",
   "\"Не люблю бабочек. Они предательницы.\"",
/*MISSING*/    "\"I do not like butterflies. They are treacherous.\"",
/*MISSING*/    "\"I do not like butterflies. They are treacherous.\"",
    }},
  {0xed5e21dc, { // "items only"
   "tylko przedmioty",
   "tylko przedmioty",
   "pouze předměty",
   "только предметы",
   "nur Items",
   "apenas itens",
    }},
  {0xed787b62, { // "last on bottom"
   "ostatnie na dole",
/*MISSING*/    "last on bottom",
   "poslední dole",
   "последние снизу",
/*MISSING*/    "last on bottom",
/*MISSING*/    "last on bottom",
    }},
  {0xedc17179, { // "Stab them by walking around them."
   "Przejdź wokół, by dźgnąć.",
   "Etrafında yürüyerek bıçaklayın.",
   "Probodni ho tím, že ho obejdeš.",
   "Нападите на него сбоку.",
   "Erstich sie, indem du um sie herumgehst.",
   "Apunhale-os andando em volta deles.",
    }},
  {0xedd232b6, { // "\n\nIn the Orb Strategy Mode, each 25 Necromancer's Totems you are given a random offensive Orb."
   "\n\nW trybie strategii sfer każde 25 Totemów Nekromanty daje dodatkową ofensywną sferę.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, each 25 Necromancer's Totems you are given a random offensive Orb.",
   "\n\nVe sférostrategickém módu ti každých 25 Nekromancerových totemù pøidá jednu náhodnou útoènou Sféru.",
   "\n\n режиме стратегии сфер каждый 25 Тотемов некроманта дают тебе случайную атакующую сферу.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, each 25 Necromancer's Totems you are given a random offensive Orb.",
/*MISSING*/    "\n\nIn the Orb Strategy Mode, each 25 Necromancer's Totems you are given a random offensive Orb.",
    }},
  {0xee0610ec, { // "This minefield has been created by the Bomberbirds to protect their valuable eggs. Mines explode when they are stepped on, creating fire all around them (this fire never expires in the Minefield). If you don't step on a mine, you know the number of mines in cells around you. These numbers are marked with color codes on the cells you have visited."
   "To pole minowe zostało stworzone przez ptaki, Bombardiery, w celu ochrony ich jaj. Miny eksplodują, gdy ktoś na nich stanie, tworząc ogień wokół (ten ogień nigdy nie wygasa na Polu Minowym). Jeśli uda Ci się nie stanąć na minę, to poznajesz liczbę min na polach wokół siebie. Te liczby są oznaczone kolorowymi kropkami na odwiedzonych polach.",
   "Bu mayıntarlası Bombacıkuşlar tarafından değerli yumurtalarını korumak için düzenlenmiş.Mayınlar üzerine basınca etrafında ateş çıkararak patlıyor(Mayıntarlasında bu ateş zamanla sönmez). Eğer bir mayına basmazsan, etrafındaki mayın sayısını bilebilirsin.Bu sayılar daha önce ziyaret ettiğin hücrelerde farklı renk kodlarıyla işaretlidir.",
   "Toto minové pole vytvořili ptáci Bombarďáci, aby ochránili svá cenná vejce. Mina exploduje, když na ni někdo nebo něco šlápne, čímž zapálí oheň na svém poli a na všech sousedních polích (v Minovém poli tento oheň nikdy neuhasne). A samozřejmě ji také odpálí jakýkoli oheň! Pokud nešlápneš na minu, budeš vždy vědět, kolik min se nachází na polích okolo tebe. Tento počet bude označený barevnými kódy na polích, která jsi už navštívil.",
   "Минное поле было создано птицами Бомбардирами, чтобы охранять их ценные яйца. Мина взрывается, когда Вы наступаете на неё, и создаёт огонь вокруг Вас (огонь здесь горит вечно). Если Вы не наступили на мину, то Вы знаете, сеолько мин рядом с Вами. Эти числа кодируются цветами клеток, на которых Вы были.",
   "Dieses Minenfeld wurde von den Bombenvögeln erschaffen, um ihre wertvollen Eier zu schützen. Minen explodieren, wenn sie betreten werden, was Feuer um sie herum erzeugt (dieses erlischt im Minenfeld nicht). Wenn du auf keine Mine trittst, weißt du, wie viele Minen um dich herum sind. Diese Zahlen werden auf Zellen, die du besucht hast farblich angezeigt.",
/*MISSING*/    "This minefield has been created by the Bomberbirds to protect their valuable eggs. Mines explode when they are stepped on, creating fire all around them (this fire never expires in the Minefield). If you don't step on a mine, you know the number of mines in cells around you. These numbers are marked with color codes on the cells you have visited.",
    }},
  {0xee08f272, { // "You cannot go against the current!"
   "Nie możesz płynąć pod prąd!",
   "Akıntıya karşı hareket edemezsin!",
   "Nemůžeš plout proti proudu!",
   "Нельзя плыть против течения!",
   "Du kannst nicht gegen die Strömung schwimmen!",
/*MISSING*/    "You cannot go against the current!",
    }},
  {0xee489efb, { // "Your eyes shine as you glance at your precious treasures."
   "Twoje oczy błyszczą, gdy patrzysz na swoje skarby.",
   "Değerli hazinelerine bakarken gözlerin ışıldadı.",
   "Tvé oči září, když se díváš na své převzácné poklady.",
   "Ваши глаза сверкают, когда Вы смотрите на свои драгоценности.",
   "Deine Augen strahlen als du die Schätze erblickst.",
   "Seus olhos brilham quando você contempla seus preciosos tesouros.",
    }},
  {0xee6664b5, { // " (local treasure)"
   " (lokalny skarb)",
/*MISSING*/    " (local treasure)",
   " (místní poklad)",
   " (местное сокровище)",
/*MISSING*/    " (local treasure)",
/*MISSING*/    " (local treasure)",
    }},
  {0xee761f21, { // "%The1 says, \"I die, but my clan in Trollheim will avenge me!\""
   "%1 mówi: \"Ja umieram, ale mój klan w Trollheim mnie pomści!\"",
/*MISSING*/    "%The1 says, \"I die, but my clan in Trollheim will avenge me!\"",
   "1 říká: \"Já umírám, ale můj klan v Trollheimu mě pomstí!\"",
   "%1 говорит: \"Я умираю, но мой клан в Тролльхейме отомстит за меня!\"",
/*MISSING*/    "%The1 says, \"I die, but my clan in Trollheim will avenge me!\"",
/*MISSING*/    "%The1 says, \"I die, but my clan in Trollheim will avenge me!\"",
    }},
  {0xeebcc20e, { // "Someone has told you that one can get battle experience safely by killing tons of essentially harmless creatures, such as Bats. But does this make any sense?...\n\nIt does not. Bats cannot hurt you, but may block your movement, or toggle switches if they fall on them."
   "Ktoś powiedział, że można bezpiecznie zdobyć doświadczenie w walce poprzez zabicie dużej liczby właściwie niegroźnych stworzeń, jak na przykład Nietoperzy. Ale czy to ma jakiś sens?...\n\nNie ma. Nietoperze nie mogą Ci nic zrobić, ale blokują Twój ruch, i mogą przestawiać przełączniki, gdy na nie spadną.",
/*MISSING*/    "Someone has told you that one can get battle experience safely by killing tons of essentially harmless creatures, such as Bats. But does this make any sense?...\n\nIt does not. Bats cannot hurt you, but may block your movement, or toggle switches if they fall on them.",
   "Někdo ti řekl, že zabitím spousty v podstatě neškodných tvorů, jako jsou Netopýři, můžeš bezpečně získat bojové zkušenosti. Ale dává to nějaký smysl...?\n\nNe, nedává. Netopýři ti nemohou ublížit, ale mohou blokovat tvůj pohyb nebo přepínat vypínače, když na ně spadnou.",
   "Кто-то сказал, что можно получить боевой опыт абсолютно безопасно, убивая тысячи абсолютно безобидных существ, например, Летучих мышей. Но какой в этом смысл?...\n\nДа никакого. Летучие мыши не наносят вам вреда, но могут мешать движению, а также падать на переключатели.",
/*MISSING*/    "Someone has told you that one can get battle experience safely by killing tons of essentially harmless creatures, such as Bats. But does this make any sense?...\n\nIt does not. Bats cannot hurt you, but may block your movement, or toggle switches if they fall on them.",
/*MISSING*/    "Someone has told you that one can get battle experience safely by killing tons of essentially harmless creatures, such as Bats. But does this make any sense?...\n\nIt does not. Bats cannot hurt you, but may block your movement, or toggle switches if they fall on them.",
    }},
  {0xeedc6223, { // "parts destroyed: %1"
   "części zniszczone: %1",
/*MISSING*/    "parts destroyed: %1",
   "zničené části: %1",
   "уничтожено частей: %1",
   "zerstörte Teile: %1",
/*MISSING*/    "parts destroyed: %1",
    }},
  {0xef22263e, { // "player 3 spin"
   "gracz 3 obrót",
/*MISSING*/    "player 3 spin",
   "hráč 3 otáčení",
   "игрок 3 вертит",
   "Spieler 3 drehen",
/*MISSING*/    "player 3 spin",
    }},
  {0xef2df1f4, { // "! (version %1)\n\n"
   "! (wersja %1)\n\n",
   "! (versiyon %1)\n\n",
   "! (verze %1)\n\n",
   "! (версия %1)\n\n",
   "! (Version %1)\n\n",
   "! (versão %1)\n\n",
    }},
  {0xef8ca48d, { // "star factor"
   "współczynnik gwiazdki",
/*MISSING*/    "star factor",
   "zhvězdění",
   "коэффициент звезды",
   "Sternfaktor",
/*MISSING*/    "star factor",
    }},
  {0xefe5ded9, { // "triangular game board"
   "plansza do gry z trójkątów",
   "üçgensel tahta",
   "Trojúhelníková herní deska",
   "треугольная игровая доска",
   "Triangulares Spielbrett",
   "tabuleiro de jogo triangular",
    }},
  {0xf01af9bf, { // "joystick mode"
   "tryb dżojstika",
   "oyun kolu modu",
   "mód joysticku",
   "режим джойстика",
   "Joystick",
   "modo joystick",
    }},
  {0xf01c6ce1, { // "targetting ranged Orbs Shift+click only"
   "celowanie sfer na odgległość wymaga Shift+klik",
   "menzilli küreler için sadece Shift+sol tık",
   "zaměřování dálkových Sfér pouze kombinací Shift + kliknutí",
   "выбирать цель сфер только через Shift+клик",
   "Zielen mit Distanzorbs per Shift-Klick",
/*MISSING*/    "targetting ranged Orbs Shift+click only",
    }},
  {0xf08d9520, { // "Everything in this Land has black and white stripes."
   "Wszystko w tej Krainie ma czarne i białe paski.",
   "Bu diyardaki her şey siyah ve beyaz şeritlere sahip.",
   "Všechno v tomto kraji má černé a bílé pruhy.",
   "Всё в этой Земли имеет белые и чёрные полосы.",
   "Alles in diesem Land hat schwarze und weiße Streifen.",
/*MISSING*/    "Everything in this Land has black and white stripes.",
    }},
  {0xf0901c78, { // "other"
   "inne",
   "diğer",
   "další",
   "другое",
   "andere",
/*MISSING*/    "other",
    }},
  {0xf0b074f1, { // "You have found %the1!"
   "Znalaz%łeś0 %a1!",
   "%a1 buldun!",
   "Naše%l0 jsi %a1!",
   "Вы нашли %a1!",
   "Du hast %den1 %a1 gefunden!",
   "Você encontrou %na1!",
    }},
  {0xf0b17da7, { // "You can also scroll to the desired location and then press 't'."
   "Możesz też przewinąć do miejsca docelowego i wcisnąć 't'.",
   "İstersen görüntüyü isteidğin yere çekip sonra 't'ye basabilirsin.",
   "Také můžeš posunout obraz do požadované lokace a stisknout 't'.",
   "Вы также можете пролистать в нужное месть и нажать 't'.",
   "Du kannst auch an die gewünschte Position scrollen und 't' drücken.",
   "Você também pode rolar para o local desejado e então pressionar 't'.",
    }},
  {0xf0e3a522, { // "Periodic patterns: application"
   "Zastosowanie wzorów okresowych",
/*MISSING*/    "Periodic patterns: application",
   "Periodické vzory: aplikace",
   "Приложение периодических узоров",
/*MISSING*/    "Periodic patterns: application",
/*MISSING*/    "Periodic patterns: application",
    }},
  {0xf0fb16c4, { // "Left the game."
   "Opuszczono grę.",
/*MISSING*/    "Left the game.",
   "Opuštění hry.",
   "Выйти из игры.",
   "Spiel verlassen.",
/*MISSING*/    "Left the game.",
    }},
  {0xf130cd50, { // "Cargo of a ship which was once destroyed by a Kraken."
   "Ładunek statku zatopionego dawno temu przez Krakena.",
/*MISSING*/    "Cargo of a ship which was once destroyed by a Kraken.",
   "Náklad z lodi, kterou kdysi zničil Kraken.",
   "Груз корабля, когда-то уничтоженного Кракеном.",
   "Ladung eines Schiffes, das einst von einem Kraken zerstört wurde.",
/*MISSING*/    "Cargo of a ship which was once destroyed by a Kraken.",
    }},
  {0xf15919d5, { // "You have used less turns than ever before. Congratulations!"
   "Zajęło to mniej kolejek niż kiedykolwiek wcześniej. Gratulacje!",
   "Her zamankinden daha az sayıda tur kullandın! Tebrikler! ",
   "Spotřebova%l0 jsi méně kol než kdy předtím. Gratulujeme!",
   "Вы потратили меньше ходов, чем раньше. Поздравляем!",
   "Du hast sowohl weniger Züge als je zuvor benötigt. Glückwunsch!",
   "Você fez seu melhor em menos turnos. Parabéns!",
    }},
  {0xf1a528d4, { // "Emerald Pattern"
   "Szmaragdowy Wzór",
   "Zümrüt Dokusu",
   "Smaragdový vzor",
   "Изумрудный узор",
   "Smaragdmuster",
/*MISSING*/    "Emerald Pattern",
    }},
  {0xf1c15b62, { // "You activate the Lightning spell!"
   "Aktywowa%łeś0 czar Burza!",
   "Şimşek büyüsünü etkinleştirdin!",
   "Aktivova%l0 jsi kouzlo Blesk!",
   "Вы активировали заклинание молнии!",
   "Du aktivierst den Donner Zauber!",
   "Você ativa o feitiço de Relâmpago!",
    }},
  {0xf1e162a7, { // "stereographic projection"
   "rzut stereograficzny",
/*MISSING*/    "stereographic projection",
   "stereografická projekce",
   "стереографическая проекция",
/*MISSING*/    "stereographic projection",
/*MISSING*/    "stereographic projection",
    }},
  {0xf201e5f2, { // "Orb power depleted!"
   "Zabrano sfery mocy!",
   "Küre gücü tükendi!",
   "Síla sféry se vyčerpala!",
   "Сила сферы исчерпана!",
   "Du hast die Kraft des Orbs aufgebraucht!",
   "Poder do Orbe esgotado!",
    }},
  {0xf201fbc0, { // "You will now start your games in %1"
   "Będziesz zaczynał grę %abl1",
   "Şimdi oyunlarını %abl1 başlatacaksın.",
   "Odteď budeš začínat hru %abl1",
   "Теперь вы начинаете игру %abl1",
   "Du startest deine Spiele von nun an in %a1",
   "Você iniciará suas partidas %l1",
    }},
  {0xf22050cb, { // "Collect at least 10 Demon Daisies to find the Orbs of Yendor"
   "Znajdź 10 Czarciego Ziela, by znaleźć Sfery Yendoru",
   "10 Şeytan Papatyası toplayarak Yendorun Kürelerini bulabilirsin.",
   "Yendorské sféry zpřístupníš sesbíráním nejméně 10 Čertových kvítek",
   "Соберите 10 адских ромашек, чтобы найти сферу Йендора",
   "Sammle mindestens 10 Dämonenblümchen um die Orbs von Yendor zu finden",
   "Colete pelo menos 10 Margaridas Demônio para encontrar os Orbes de Yendor",
    }},
  {0xf223aa46, { // "summon a Golem"
   "przywołaj Golema",
   "Golem çıkar",
   "vyvolej golema",
   "призвать Голема",
   "rufe einen Golem herbei",
/*MISSING*/    "summon a Golem",
    }},
  {0xf265ee4b, { // "\"There is no royal road to geometry.\""
   "W geometrii nie ma specjalnych dróg dla królów.\"",
/*MISSING*/    "\"There is no royal road to geometry.\"",
   "Není žádná královská cesta ke geometrii.\"",
   "\"Нет царских путей к геометрии.\"",
/*MISSING*/    "\"There is no royal road to geometry.\"",
/*MISSING*/    "\"There is no royal road to geometry.\"",
    }},
  {0xf28650d8, { // "The faster you get here, the better you are!"
   "Im szybciej to osiągniesz, tym lepiej!",
   "Daha hızlı demek, daha iyi demek !",
   "Čím rychleji se sem dostaneš, tím lepší jsi!",
   "Чем быстрее, тем лучше!",
   "Je schneller du es hierher schaffst, desto besser bist du!",
   "Quanto mais rápido chegar aqui, melhor você se torna!",
    }},
  {0xf2d2d33a, { // "Defeat Flail Guards by moving away from them."
   "Wojownika z Cepem pokonujesz odchodząc od niego.",
/*MISSING*/    "Defeat Flail Guards by moving away from them.",
   "Strážce s řemdihem porazíš tím, že se pohneš směrem od něj.",
   "Поразите Воина с цепом, убегая от него.",
/*MISSING*/    "Defeat Flail Guards by moving away from them.",
/*MISSING*/    "Defeat Flail Guards by moving away from them.",
    }},
  {0xf2d8259d, { // "[+] keep the window size, [-] use the screen resolution"
   "[+] zachowaj rozmiar okna, [-] użyj rozdzielczości ekranu",
   "[+] pencere boyutunu koru, [-] ekran çözünürlüğünü kullan",
   "[+] zachovej rozměr okna, [-] použij rozlišení obrazovky",
   "[+] фиксирует размер окна, [-] использует разрешение экрана",
   "[+] Fenstergröße behalten, [-] Bildschirmauflösung verwenden",
   "[+] manter o tamanho da janela, [-] usar a resolução da tela",
    }},
  {0xf308e57a, { // "For now, collect as much treasure as possible..."
   "Szukaj skarbów, znajdź, ile tylko możesz...",
   "Şimdilik mümkün olduğunca fazla hazine topla...",
   "Prozatím sesbírej co nejvíc pokladů......",
   "Соберите столько сокровищ, сколько сможете...",
   "Aber sammle vorerst so viele Schätze wie möglich.",
   "Por ora, colete o máximo de tesouro possível...",
    }},
  {0xf351c0a0, { // "left to right"
   "w prawo",
/*MISSING*/    "left to right",
   "zleva doprava",
   "вправо",
   "von links nach rechts",
/*MISSING*/    "left to right",
    }},
  {0xf36f20c6, { // "sort"
   "sortuj",
/*MISSING*/    "sort",
   "tøídìní",
   "сортируй",
/*MISSING*/    "sort",
/*MISSING*/    "sort",
    }},
  {0xf385bf50, { // "random pattern mode"
   "tryb losowych wzorów",
   "rastgele doku modu",
   "mód náhodných vzorů",
   "режим случайных узоров",
   "Zufallsmuster Modus",
/*MISSING*/    "random pattern mode",
    }},
  {0xf3914c9c, { // "Accessible only from %the1 or %the2.\n"
   "Kraina dostępna tylko poprzez %a1 i %a2.\n",
   "Sadece %a1 veya %a2 erişilebilir.\n",
   "Tento kraj je dostupný pouze skrz %a1 nebo %a2.\n",
   "Доступно только из %a1 и %a2.\n",
   "Nur von %der1 %a1 oder %der %a2 aus erreichbar.\n",
/*MISSING*/    "Accessible only from %the1 or %the2.\n",
    }},
  {0xf3b20816, { // "\"Train in the Desert first!\""
   "\"Potrenuj najpierw na Pustyni!\"",
   "\"Önce Çöl'de kendini yetiştir!\"",
   "\"Nejdřív se natrénuj v Poušti!\"",
   "\"Потренируйтесь сначала в Пустыне!\"",
   "\"Trainier zuerst in der Wüste!\"",
/*MISSING*/    "\"Train in the Desert first!\"",
    }},
  {0xf3f04786, { // "%The1 throws fire at %the2!"
   "%1 rzuca ogniem w %a2!",
/*MISSING*/    "%The1 throws fire at %the2!",
   "%1 hází oheň na %a2!",
   "%1 кидает огонь в %a2!",
   "%Der1 %1 wirft Feuer auf %den2 %a2!",
/*MISSING*/    "%The1 throws fire at %the2!",
    }},
  {0xf3ffe25e, { // "turns"
   "kol",
/*MISSING*/    "turns",
   "kol",
   "ходы",
/*MISSING*/    "turns",
/*MISSING*/    "turns",
    }},
  {0xf40355a7, { // "dark rainbow landscape"
   "ciemna tęcza",
/*MISSING*/    "dark rainbow landscape",
   "temná duha",
   "тёмная радуга",
/*MISSING*/    "dark rainbow landscape",
/*MISSING*/    "dark rainbow landscape",
    }},
  {0xf472ee24, { // "(You can also use right Shift)\n\n"
   "(Możesz też użyć prawego Shifta)\n\n",
/*MISSING*/    "(You can also use right Shift)\n\n",
   "(Také lze použít pravý Shift)\n\n",
   "(Также можно использовать правый Shift)\n\n",
/*MISSING*/    "(You can also use right Shift)\n\n",
/*MISSING*/    "(You can also use right Shift)\n\n",
    }},
  {0xf4ca290b, { // "A hot land, full of sand dunes, mysterious Spice, and huge and dangerous sand worms."
   "Gorąca ziemia, pełna wydm, tajemniczej Przyprawy i wielkich Pustynnych Czerwi.",
   "Kum tepeleri, gizemli Baharat ve devasa Kumkurtlarıyla dolu sıcak bir diyar.",
   "Horký kraj plný písečných dun, tajemného Koření a obrovských a nebezpečných Písečných červů.",
   "Жаркая земля, полная песчаных дюн, загадочных специй и огромных опасных червей.",
   "Ein heißes Land voller Sanddünen, geheimnisvollem Spice, und riesigen und gefährlichen Sandwürmern.",
/*MISSING*/    "A hot land, full of sand dunes, mysterious Spice, and huge and dangerous sand worms.",
    }},
  {0xf4d6bb2d, { // "\n\nSecondary orb: %1"
   "\n\nDodatkowa Sfera: %1",
/*MISSING*/    "\n\nSecondary orb: %1",
   "\n\nSekundární Sféra: %1",
   "\n\nДополнительная сфера: %1",
/*MISSING*/    "\n\nSecondary orb: %1",
/*MISSING*/    "\n\nSecondary orb: %1",
    }},
  {0xf4e796f2, { // "You summon a golem!"
   "Przywoła%łeś0 Golema!",
   "Bir golem çağırdın!",
   "Vyvola%l0 jsi golema!",
   "Вы призвали голема!",
   "Du beschwörst einen Golem!",
   "Você invoca um golem!",
    }},
  {0xf4fb4209, { // "reset per-player statistics"
   "zresetuj statystyki graczy",
/*MISSING*/    "reset per-player statistics",
   "Vyresetovat hráčské statistiky",
   "очистить статистику по игрокам",
/*MISSING*/    "reset per-player statistics",
/*MISSING*/    "reset per-player statistics",
    }},
  {0xf541b622, { // "or 'v' to see the main menu"
   "lub 'v' by zobaczyć menu",
   "veya ana menüyü görmek için 'v'ye bas.",
   "nebo stiskni 'v' pro hlavní menu",
   "или 'v', чтобы открыть главное меню",
   "oder v um das Menü einzublenden",
/*MISSING*/    "or 'v' to see the main menu",
    }},
  {0xf5440591, { // "joystick mode: automatic (release the joystick to move)"
   "tryb dżojstika: automatyczny (puść by ruszyć)",
   "oyun kolu modu: otomatik (kolu bırakarak hareket edersin)",
   "mód joysticku: automatický (pohybuješ se uvolněním joysticku)",
   "Режим джойстика: авто (опустите джойстик для перемещения)",
   "Joystick: Automatisch (loslassen zum bewegen)",
   "modo joystick: automático (solte o joystick para mover-se)",
    }},
  {0xf54446a7, { // "projection"
   "rzut",
/*MISSING*/    "projection",
   "projekce",
   "проекция",
/*MISSING*/    "projection",
/*MISSING*/    "projection",
    }},
  {0xf549cbfc, { // "Basics of gameplay"
   "Podstawy rozgrywki",
/*MISSING*/    "Basics of gameplay",
   "Základy hry",
   "Начала игры",
/*MISSING*/    "Basics of gameplay",
/*MISSING*/    "Basics of gameplay",
    }},
  {0xf565d694, { // "Challenges do not get harder"
   "Misje nie stają się trudniejsze",
   "Ek görevler zorlaşmaz.",
   "Obtížnost misí zůstává stále stejná",
   "Сложность не меняется",
   "Herausforderungen werden nicht schwieriger",
/*MISSING*/    "Challenges do not get harder",
    }},
  {0xf58afff6, { // "You have improved your total and '%1' high score!"
   "Poprawi%łeś0 swój wynik ogólny oraz swój wynik w kategorii '%P1!'",
   "Toplam yüksek puanını ve '%P1!' puanını artırdın!",
   "Zlepši%l0 jsi své celkové nejvyšší skóre a své nejvyšší skóre v kategorii '%P1!'",
   "Вы подняли свой общий результат и '%1'!",
   "Du hast deinen Gesamthighscore und '%1' Highscore verbessert!",
   "Você superou seu total e mais '%1!' melhores pontuações!",
    }},
  {0xf5af26f3, { // "predesigned patterns"
   "zaprojektowane wzorki",
   "hazır dokular",
   "předem vytvořené vzory",
   "спроектированные узоры",
   "Vorgefertigte Muster",
/*MISSING*/    "predesigned patterns",
    }},
  {0xf5eadca0, { // "Coastal region -- connects inland and aquatic regions.\n"
   "Kraina przybrzeżna -- łączy krainy lądowe i morskie.\n",
/*MISSING*/    "Coastal region -- connects inland and aquatic regions.\n",
   "Pobřežní kraj -- spojuje vnitrozemské a mořské kraje.\n",
   "Прибрежный регион -- соединяет наземные и морские регионы.\n",
/*MISSING*/    "Coastal region -- connects inland and aquatic regions.\n",
/*MISSING*/    "Coastal region -- connects inland and aquatic regions.\n",
    }},
  {0xf5fc270a, { // "f = flip %1"
   "f = odbicie %1",
   "f = çevir %1",
   "f = překlopit %1",
   "f = перевернуть %1",
   "f = kippen %1",
/*MISSING*/    "f = flip %1",
    }},
  {0xf6297cf7, { // "These creatures are slow, but very powerful... more powerful than you. You need some more experience in demon fighting before you will be able to defeat them. Even then, you will be able to slay this one, but more powerful demons will come...\n\nEach 10 lesser demons you kill, you become powerful enough to kill all the greater demons on the screen, effectively turning them into lesser demons."
   "Te demony są powolne, ale bardzo silne... silniejsze od Ciebie. Potrzebujesz zdobyć trochę doświadczenia, zanim będziesz w stanie je pokonać. Nawet wtedy przyjdą jeszcze silniejsze demony...\n\nZa każdym razem, gdy pokonasz 10 mniejszych demonów, stajesz się dostatecznie silny, by pokonać wszystkie Wielkie Demony na ekranie. Stają się one w tym momencie Mniejszymi Demonami.",
   "Bu yaratıklar yavaştır ama çok güçlüdür. Senden bile. Bunları yenebilmek için öncesinde şeytanlarla savaşma konusunda tecrübeli olman gerekir. O zaman bile, bunu yenebilirsin, ama daha güçlü şeytanlar gelecektir...\n\nÖldürdüğün her 10 küçük şeytan için, ekrandaki bütün Büyük Şeytanları -sanki onlar da küçükmüş gibi- öldürebilirsin.",
   "Tito tvorové jsou pomalí, ale velmi silní... silnější než ty. Abys je mohl porazit, musíš mít větší zkušenosti v boji proti démonům. I pak budeš sice schopný zabít tohoto Ďábla, ale přijdou ještě silnější...\n\nZa každých 10 Čertů, které zabiješ, získáš dost síly na to, abys mohl zabít všechny Ďábly na obrazovce -- v podstatě se promění v Čerty.",
   "Эти существа медленны, но очень сильны... гораздо сильнее Вас. Вам нужно набраться опыта в битве с демонами.\n\nКогда Вы убьёте 10 младших демонов, Вы станете достаточно сильны, чтобы побеждать старших, которые тут же сами станут младшими, но потом придут новые старшие демоны...",
   "Diese Kreaturen sind langsam, aber sehr mächtig... mächtiger als du. Du brauchst Erfahrung im Kampf gegen Dämonen, bevor du sie besiegen kannst. Trotzdem bist du in der Lage, diesen zu töten, aber es werden stärkere Dämonen kommen...\n\nJedes Mal, wenn du 10 gewöhnliche Dämonen getötet hast, wirst du stark genug, um alle mächtigen Dämonen auf dem Bildschirm zu töten, indem du sie in gewöhnliche Dämonen verwandelst.",
/*MISSING*/    "These creatures are slow, but very powerful... more powerful than you. You need some more experience in demon fighting before you will be able to defeat them. Even then, you will be able to slay this one, but more powerful demons will come...\n\nEach 10 lesser demons you kill, you become powerful enough to kill all the greater demons on the screen, effectively turning them into lesser demons.",
    }},
  {0xf67f9ac8, { // "You cannot attack %the1!"
   "Nie możesz zaatakować %a1!",
/*MISSING*/    "You cannot attack %the1!",
   "Nemůžeš útočit na %a1!",
   "Вы не можете атаковать %a1!",
/*MISSING*/    "You cannot attack %the1!",
/*MISSING*/    "You cannot attack %the1!",
    }},
  {0xf698e17d, { // "You cannot attack Rock Snakes directly!"
   "Nie możesz atakować Skalnych Węży bezpośrednio!",
   "Kaya Yılanlarına doğrudan saldıramazsın!",
   "Na Kamenné hady nemůžeš útočit přímo!",
   "Вы не можете атаковать Каменную змею!",
   "Du kannst Felsschlangen nicht direkt angreifen!",
/*MISSING*/    "You cannot attack Rock Snakes directly!",
    }},
  {0xf6a96efd, { // "inventory/kill sorting"
   "sortowanie skarbów/potworów",
/*MISSING*/    "inventory/kill sorting",
   "řazení inventáře/netvorů",
   "сортировать предметы/монстров",
/*MISSING*/    "inventory/kill sorting",
/*MISSING*/    "inventory/kill sorting",
    }},
  {0xf6c1527a, { // "treasure collected: %1"
   "zdobyte skarby: %1",
   "toplanan hazine: %1",
   "pokladů získáno: %1",
   "собрано сокровищ: %1",
   "Schätze gesammelt: %1",
   "tesouro coletado: %1",
    }},
  {0xf71db775, { // "You jump over the table!"
   "Skaczesz przez stół!",
   "Masanın üstüne çıktın!",
   "Přeskočil jsi stůl!",
   "Вы запрыгнули на стол!",
   "Du springst über den Tisch!",
/*MISSING*/    "You jump over the table!",
    }},
  {0xf7467711, { // "A heap of wood that can be used to start a fire. Everything is already here, you just need to touch it to fire it."
   "Stos drewna. Wszystko gotowe, wystarczy podpalić.",
   "Ateş yakmaya uygun bir yığın odun. Her şey zaten burada, ateş için sadece dokunman yetecek.",
   "Hromada dřeva, která se dá použít k zapálení ohně. Všechno už tu je, na podpálení se jí stačí jen dotknout.",
   "Куча брёвен. Всё готово для костра, Вам нужно лишь поджечь его.",
   "Ein Haufen Holz, den du verwenden kannst, um ein Feuer zu machen. Alles ist schon da, du musst es nur berühren um das Feuer zu entfachen.",
/*MISSING*/    "A heap of wood that can be used to start a fire. Everything is already here, you just need to touch it to fire it.",
    }},
  {0xf74b7514, { // "emerald pattern"
   "szmaragdowy wzór",
   "zümrüt dokusu",
   "smaragdový vzor",
   "изумрудный узор",
   "Smaragdmuster",
/*MISSING*/    "emerald pattern",
    }},
  {0xf751afd7, { // "A strange land which contains mirrors and mirages, protected by Mirror Rangers."
   "Dziwna kraina, gdzie można znaleźć magiczne lustra i chmury, bronione przez Strażników Luster.",
   "Aynalar ve yansımalarla dolu garip bir diyar, ayna korucuları tarafından korunuyor.",
   "Podivný kraj se zrcadly a přeludy, který chrání Zrcadloví strážci.",
   "Странная земля с зеркалами и миражами, защищёнными Странниками.",
   "Ein seltsames Land voller Spiegel und Illusionswolken, die von den Rangern geschützt werden.",
/*MISSING*/    "A strange land which contains mirrors and mirages, protected by Mirror Rangers.",
    }},
  {0xf7705951, { // "graphics configuration"
   "konfiguracja grafiki",
/*MISSING*/    "graphics configuration",
   "konfigurace grafiky",
   "настройка графики",
/*MISSING*/    "graphics configuration",
/*MISSING*/    "graphics configuration",
    }},
  {0xf7b278b8, { // "\"There is no branch of mathematics, however abstract, "
   "\"Nie ma na tyle abstrakcyjnej gałęzi matematyki, by się nie dało jej ",
/*MISSING*/    "\"There is no branch of mathematics, however abstract, ",
   "\"Není žádné matematické odvětví, jakkoli abstraktní, ",
   "\"Нет ни одной области математики, как бы абстрактна она ни была, ",
/*MISSING*/    "\"There is no branch of mathematics, however abstract, ",
/*MISSING*/    "\"There is no branch of mathematics, however abstract, ",
    }},
  {0xf7d5e5f2, { // "turn left"
   "skręć w lewo",
   "sola dön",
   "otočení vlevo",
   "повернуть налево",
   "Nach links drehen",
/*MISSING*/    "turn left",
    }},
  {0xf83ed08e, { // "configure keys/joysticks"
   "konfiguracja klawiszy/joysticka",
/*MISSING*/    "configure keys/joysticks",
   "konfigurace kláves/joysticku",
   "настройки клавиш/джойстика",
/*MISSING*/    "configure keys/joysticks",
/*MISSING*/    "configure keys/joysticks",
    }},
  {0xf8472f7f, { // "move through walls"
   "przechodź przez ściany",
/*MISSING*/    "move through walls",
   "procházení zdmi",
   "проходить сквозь стены",
/*MISSING*/    "move through walls",
/*MISSING*/    "move through walls",
    }},
  {0xf86c2b63, { // "F5 = restart"
   "F5 = restart",
   "F5 = yeniden başlat",
   "F5 = restart",
   "F5 = перезапуск",
   "F5 = Neustart",
/*MISSING*/    "F5 = restart",
    }},
  {0xf86d1171, { // "You summon some Thumpers!"
   "Przywoła%łeś0 dudniki!",
   "Gümleyenler oluşturdun",
   "Vyvola%l0 jsi několikl Tlouků!",
   "Вы призвали тамперы!",
   "Du beschwörst ein paar Stampfer!",
   "Você invoca alguns Batedores!",
    }},
  {0xf88f8da1, { // "You gain a protective Shell!"
   "Zdobył%eś0 ochronną Skorupę!",
/*MISSING*/    "You gain a protective Shell!",
   "Získa%l0 jsi ochranný Krunýř!",
   "Вы получили защитную раковину!",
/*MISSING*/    "You gain a protective Shell!",
/*MISSING*/    "You gain a protective Shell!",
    }},
  {0xf893311c, { // "%The1 is killed by thorns!"
   "%1 wpad%ł1 na kolce i się zabi%ł1!",
/*MISSING*/    "%The1 is killed by thorns!",
   "%1 se nabod%l na trny!",
   "%1 убит%E1 шипами!",
   "%Der1 %1 wurde von den Dornen getötet!",
/*MISSING*/    "%The1 is killed by thorns!",
    }},
  {0xf89cf7e3, { // "A dark forest filled with ghosts and graves. But there is also treasure hidden deep within... But don't let greed make you stray from your path, as you can get lost!\n\nThe Haunted Woods are bounded by a single equidistant curve. It is not a circle or horocycle.\n\n"
   "Mroczny las, pełen duchów i grobów. Głęboko w środku są również skarby... Ale nie pozwól, by zgubiła Cię chciwość!\n\nNawiedzony Las jest ograniczony jedną ekwidystantą. Nie jest kołem ani horocyklem.\n\n",
/*MISSING*/    "A dark forest filled with ghosts and graves. But there is also treasure hidden deep within... But don't let greed make you stray from your path, as you can get lost!\n\nThe Haunted Woods are bounded by a single equidistant curve. It is not a circle or horocycle.\n\n",
   "Temný hvozd plný duchů a hrobů. Ale hluboko uvnitř se skrývá poklad... Nedopusť však, aby tě tvá chamtivost svedla z cesty, protože tu můžeš zabloudit!\n\nStrašidelný hvozd je ohraničený jednou ekvidistantou. Není to kruh ani horocyklus.\n\n",
   "Мрачный лес, полный духов и надгробий. Но в глубине спрятаныi сокровища... Но не будьте слишком жадными, а то потеряетесь!\n\nЛес с привидениями ограничен одной эквидистантой.\n\n",
   "Ein dunkler Wald voller Geister und Gräber. Doch in seinen Tiefen sind auch Schätze versteckt... Aber lass dich nicht durch Gier von deinem Weg abbringen, du könntest dich verlaufen!\n\nDer verwunschene Wald ist begrenzt durch eine einzige äquidisdante Kurve. Diese ist kein Kreis und kein Horozykel.\n\n",
/*MISSING*/    "A dark forest filled with ghosts and graves. But there is also treasure hidden deep within... But don't let greed make you stray from your path, as you can get lost!\n\nThe Haunted Woods are bounded by a single equidistant curve. It is not a circle or horocycle.\n\n",
    }},
  {0xf89f4e7f, { // "yes"
   "tak",
/*MISSING*/    "yes",
   "ano",
   "да",
/*MISSING*/    "yes",
/*MISSING*/    "yes",
    }},
  {0xf8abf46f, { // "The Knights congratulate you on your success!"
   "Rycerze gratulują Ci sukcesu!",
   "Şövalyeler başarın için seni tebrik ediyor!",
   "Rytíři ti gratulují ke tvému úspěchu!",
   "Рыцари поздравляют Вас с успехом!",
   "Die Ritter gratulieren dir zu deinem Erfolg!",
/*MISSING*/    "The Knights congratulate you on your success!",
    }},
  {0xf8d5e789, { // "MOVE"
   "RUCH",
   "OYNAT",
   "POHYB",
   "ХОД",
   "VOR",
/*MISSING*/    "MOVE",
    }},
  {0xf927dd7c, { // "Someone has put air fans in these plains, causing strong winds everywhere. You think that the purpose is to harness the magical power of Air Elementals, but you are not sure.\n\nAll cells except fans are grouped into three colors according to a pattern. Wind blows counterclockwise around each group of cells of a single color. Cells which are blocked by walls, or at distance at most 2 from an Air Elemental, do not count for this.\n\nIt is illegal to move in a direction which is closer to incoming wind than to outcoming wind. However, you can move two cells with the wind in a single turn, and so can the birds."
   "Ktoś porozstawiał wentylatory na tych równinach, powodując wszędzie silne wiatry. Celem jest zapewne uzyskanie magicznych mocy Powietrznych Żywiołaków, ale nie masz pewności.\n\nWszystkie komórki poza wentylatorami są podzielone na trzy kolory, zgodnie z regularnym wzorem. Wiatr wieje przeciwnie do ruchu wskazówek zegara wokół każdej grupy pól jednego koloru. Pola zablokowane przez ściany lub w odległości co najwyżej 2 od Powietrznego Żywiołaka, nie liczą się tutaj.\n\nNie można ruszać się w kierunku który jest bliżej przychodzącego niż odchodzącego wiatru. Możesz jednak się ruszyć o dwa pola z wiatrem w jednej kolejce. Ptaki również tak mogą zrobić.",
/*MISSING*/    "Someone has put air fans in these plains, causing strong winds everywhere. You think that the purpose is to harness the magical power of Air Elementals, but you are not sure.\n\nAll cells except fans are grouped into three colors according to a pattern. Wind blows counterclockwise around each group of cells of a single color. Cells which are blocked by walls, or at distance at most 2 from an Air Elemental, do not count for this.\n\nIt is illegal to move in a direction which is closer to incoming wind than to outcoming wind. However, you can move two cells with the wind in a single turn, and so can the birds.",
   "Někdo umístil na tuto pláň větráky, díky kterým tu všude vanou silné větry. Myslíš si, že jejich účelem je zkrotit magickou moc Elementálů vzduchu, ale s jistotou to nevíš.\n\nVšechna políčka s výjimkou větráků jsou rozdělena na tři barvy podle určitého vzoru. Vítr vane proti směru hodinových ručiček kolem každé skupiny stejnobarevných políček. Políčka zablokovaná zdmi nebo políčka ve vzdálenosti 2 nebo méně od Elementála vzduchu se nepočítají.\n\nNení možné pohybovat se směrem, který je víc proti větru než po větru. Při pohybu po větru se však můžeš (stejně jako libovolný typ ptáka) pohnout o dvě políčka v jediném kole.",
   "Кто-то расставил вентиляторы на этой равнине и вызвал сильные ветры. Вам кажется, что кто-то хотел использовать силу Воздушных Элементалей, но не уверены.\n\nВсе клетки, кроме вентиляторов, раскрашены в 3 цвета регулярным образом. Ветер дует против часовой стрелки вокруг каждой группы одноцветных клеток. Клетки, занятые стенами, или находящиеся на  расстоянии не более 2 от Воздушного Элементаля, не учитываются.\n\nНевозможно пойти в направлении, которое ближе ко входящему ветру, чем к выходящему. Зато Вы можете пройти сразу 2 клетки в  направлении ветра, и птицы тоже могут.",
   "Jemand hat Ventilatoren in diese Ebenen gebracht, die überall starke Winde hervorrufen. Du vermutest, sie sollen die magische Kraft der Luftelementare zügeln, aber du bist dir nicht sicher.\n\nAlle Zellen außer den Ventilatoren unterteilen sich einem Muster entsprechend in drei Farben. Der Wind weht gegen den Uhrzeigersinn in jeder Gruppe von Zellen mit einer Farbe. Zellen, die durch Wände blockiert oder nicht mehr als 2 Zellen von einem Luftelementar entfernt sind, sind dabei ausgenommen.\n\nEs ist unzulässig in eine Richtung zu gehen, die näher am einströmenden als am ausströmenden Wind liegt. Aber du kannst in einem Zug zwei Zellen mit dem Wind gehen, genau wie die Vögel.",
/*MISSING*/    "Someone has put air fans in these plains, causing strong winds everywhere. You think that the purpose is to harness the magical power of Air Elementals, but you are not sure.\n\nAll cells except fans are grouped into three colors according to a pattern. Wind blows counterclockwise around each group of cells of a single color. Cells which are blocked by walls, or at distance at most 2 from an Air Elemental, do not count for this.\n\nIt is illegal to move in a direction which is closer to incoming wind than to outcoming wind. However, you can move two cells with the wind in a single turn, and so can the birds.",
    }},
  {0xf9555116, { // "No mines next to you."
   "Nie ma min obok Ciebie.",
   "Yakınında hiç mayın yok.",
   "Ve tvé blízkosti se nenacházejí žádné miny.",
   "Нет мин рядом с Вами.",
   "Keine Minen um dich.",
/*MISSING*/    "No mines next to you.",
    }},
  {0xf98b5600, { // "Note: you can play, but scores won't be saved on Android"
   "Uwaga: możesz grać, ale wyniki nie będą zapisywane",
/*MISSING*/    "Note: you can play, but scores won't be saved on Android",
   "Poznámka: můžeš hrát, ale tvé skóre se neuloží",
   "Важно: вы можете играть, но счёт не будет сохранён",
   "Hinweis: du kannst spielen, aber bei Android werden keine Punktestände gespeichert",
/*MISSING*/    "Note: you can play, but scores won't be saved on Android",
    }},
  {0xfa0c015f, { // "use at your own risk!"
   "używasz na własne ryzyko!",
   "sorumluluk sende!",
   "použití na vlastní nebezpečí!",
   "используйте это на свой риск!",
   "Benutzen auf eigene Gefahr!",
/*MISSING*/    "use at your own risk!",
    }},
  {0xfa396010, { // "The Land of Power is filled with everburning fire, magical Orbs, and guarded by witches and golems. There are basic orbs lying everywhere, and more prized ones are kept in crystal cabinets.\n\nWitches are allowed to use all the powers of the basic orbs against intruders. These powers never expire, but a Witch can use only one power at a time (not counting Orbs of Life).\n\nWitches and Golems don't pursue you into other Lands. Also, most Orb powers are drained when you leave the Land of Power."
   "Kraina Mocy jest wypełniona wiecznie płonącym ogniem, magicznymi Sferami, i strzeżona przez Czarownice i golemy. Podstawowe Sfery walają się wszędzie, a te bardziej cenne są trzymane w kryształowych gablotach.\n\nCzarownice używają wszystkich mocy podstawowych Sfer przeciwko intruzom. Ich moce nigdy nie przemijają, ale Czarownica może używać naraz tylko jednej mocy (nie licząc Sfer Życia).\n\nCzarownice i Golemy nie gonią Cię do innych krain. Większość Sfer traci swoją moc, gdy opuszczasz Krainę Mocy.",
   "Güç Diyarı sönmeyen ateşle, büyülü Kürelerle dolu ve cadılar ve golemler ile korunuyor. Her yerde temel Küreler var ve daha değerliler de kristal kabinlerde korunuyor. \n\nCadılar bu kürelerin gücünü olası işgalcilere karşı kullanabilirler. Bu güçlerin süresi bitmez ama Yaşam Küresi hariç cadılar aynı anda sadece birini kullanabilir. \n\nCadılar ve Golemler seni diğer diyarlara kadar takip etmezler. Ayrıca çoğu kürenin gücü Güç Diyarını terkettiğinde biter.",
   "Kraj Moci je plný věčně planoucích ohňů a magických Sfér a střeží ho čarodějnice a golemové. Základní Sféry se válejí všude kolem a ty cennější jsou schované v křišťálových vitrínách.\n\nČarodějnice mohou v boji proti vetřelcům používat všechny základní sféry. Jejich síla se nikdy nevyčerpá, ale každá Čarodějnice může mít vždy pouze jednu schopnost (Sféry Života se nepočítají).\n\nČarodějnice a golemové tě nikdy nebudou pronásledovat do jiných Krajů. Kromě toho většina sfér při opuštění Kraje Moci svou sílu ztratí.",
   "Земля силы полна вечно горящими огнями и магическими сферами и защищена ведьмами и големами. Все обычные сферы есть здесь, и наиболее ценные из них защищены кристальными кабинетами.\n\nВедьмы способны использовать силы сфер против нарушителей. Эти силы действуют бесконечно, но каждая ведьма может использовать лишь Одну (не считая сферы жизни).\n\nВедьмы и големы не покидают эту землю. Когда вы уходите из неё, силы некоторых Ваших сфер покидают Вас.",
   "Das Land der Macht ist voll mit ewigen Flammen, magischen Orb sowie  Hexen und Golems die es bewachen. Überall liegen einfache Orbs herum, und wertvollere werden in Kristallvitrinen aufbewahrt.\n\nHexen dürfen alle Orbkräfte gegen Angreifer einsetzen. Diese laufen nie aus, allerdings kann eine Hexe jederzeit nur einen Zauber nutzen. (mit Ausnahme der Orbs des Lebens)\n\nHexen und Golems verfolgen dich nicht in andere Länder. Außerdem verlieren viele Zauber die Kraft sobald du das Land verlässt.",
/*MISSING*/    "The Land of Power is filled with everburning fire, magical Orbs, and guarded by witches and golems. There are basic orbs lying everywhere, and more prized ones are kept in crystal cabinets.\n\nWitches are allowed to use all the powers of the basic orbs against intruders. These powers never expire, but a Witch can use only one power at a time (not counting Orbs of Life).\n\nWitches and Golems don't pursue you into other Lands. Also, most Orb powers are drained when you leave the Land of Power.",
    }},
  {0xfacc6c34, { // "This feather is truly beautiful and strong."
   "Pióro białego gołębia. Piękne i silne.",
/*MISSING*/    "This feather is truly beautiful and strong.",
   "Pero bílé holubice. Opravdu nádherné a silné.",
   "Красивое и сильное голубиное перо.",
   "Diese Feder ist wahrlich schön und stark.",
/*MISSING*/    "This feather is truly beautiful and strong.",
    }},
  {0xfafa0222, { // "%The1 would kill you there!"
   "Zabi%ł1by Ciebie %1!",
   "%1 orada seni hemencecik öldürür!",
   "Tam by tě zabi%l1 %1!",
   "%1 убьёт Вас здесь!",
   "%Der1 %1 würde dich dort töten!",
   "%na1 lhe mataria ali!",
    }},
  {0xfb1e1e81, { // "Cannot teleport here!"
   "Nie możesz teleportować się tutaj!",
   "Buraya ışınlanamazsın!",
   "Sem se teleportovat nemůžeš!",
   "Нельзя телепортироваться сюда!",
   "Du kannst dich hierhin nicht teleportieren!",
/*MISSING*/    "Cannot teleport here!",
    }},
  {0xfbb3ff91, { // "Thus, it is potentially useful for extremely long games, which would eat all the memory on your system otherwise.\n"
   "Jest więc to użyteczne w bardzo długich grach, które inaczej by zjadły całą pamięć.",
   "Yani, bu olası çok uzun, sistemin tüm hafızasını tüketecek oyunlarında yardımcı olabilir.",
   "To je tedy potenciálně užitečné pro extrémně dlouhé hry, které by jinak sežraly všechnu paměť tvého počítače.",
   "Это может быть полезно для крайне долгих игр, если у Вас заканчивается память.",
   "Dies ist nützlich für sehr lange Spiele, welche ansonsten viel Speicher nutzen würden.",
   "Portanto, é potencialmente útil para jogos extremamente longos, que de outra forma consumiria toda a memória do sistema.",
    }},
  {0xfbe924dd, { // " (click to drop)"
   " (naciśnij by porzucić)",
   " (bırakmak için tıkla)",
   " (polož kliknutím)",
   " (нажмите, чтобы бросить)",
   " (Klicke zum Ablegen)",
   " (clique para derrubar)",
    }},
  {0xfbe9cdc2, { // "To access Hell, collect %1 treasures each of 9 kinds..."
   "By dostać się do Piekła, znajdź po %1 skarbów każdego z 9 rodzajów...",
   "Cehenneme erişmek için her 9 hazine tipinden %1'ar tane toplamalısın...",
   "Aby ses dostal do Pekla, sesbírej %1 pokladů od každého z 9 různých typů...",
   "Чтобы открыть Ад, соберите по %1 сокровищ из 9 различных мест...",
   "Um die Hölle zu erreichen sammle je %1 von 9 verschiedenen Schätzen...",
   "Para acessar o Inferno, colete %1 tesouros de cada um dos 9 tipos...",
    }},
  {0xfbeae111, { // "Do you think you are playing on a ball? This is far from the truth!\n"
   "Myślisz, że grasz na sferze? Jest to dalekie od prawdy!\n",
/*MISSING*/    "Do you think you are playing on a ball? This is far from the truth!\n",
   "Myslíš, že hraješ na kouli? Samá voda!\n",
   "Думаешь, ты играешь на сфере? Как же это далеко от правды!\n",
/*MISSING*/    "Do you think you are playing on a ball? This is far from the truth!\n",
/*MISSING*/    "Do you think you are playing on a ball? This is far from the truth!\n",
    }},
  {0xfbf4ee95, { // "Collect at least %1 treasures in each of 9 types to access Hell"
   "Znajdź po %1 skarbów w 9 typach, by się dostać do Piekła",
   "9 hazine çeşidinin hepsinden %1'ar tane toplayarak Cehenneme erişebilirsin.",
   "Peklo zpřístupníš sesbíráním %1 pokladů od každého z 9 typů",
   "Соберите хотя бы по %1 сокровищ 9 разных типов, чтобы попасть в Ад",
   "Sammle 9 verschiedene Schätze mindestens %1x um Zugang zur Hölle zu erhalten",
   "Colete pelo menos %1 tesouros de cada um dos 9 tipos para acessar o Inferno",
    }},
  {0xfc0a5900, { // "synchronize net and map"
   "synchronizuj siatkę z mapą",
   "haritayı ve ağı eşle",
   "synchronizuj síť a mapu",
   "синхронизовать сеть и карту",
   "Synchronisier Netz und Karte",
/*MISSING*/    "synchronize net and map",
    }},
  {0xfc0edbfc, { // "You summon a monster!"
   "Przywoła%łeś0 potwora!",
   "Bir canavar oluşturdun!",
   "Vyvola%l0 jsi netvora!",
   "Вы призвали монстра!",
   "Du beschwörst ein Monster!",
   "Você invoca um monstro!",
    }},
  {0xfc2d9158, { // "flash"
   "błysk",
/*MISSING*/    "flash",
   "záblesk",
   "вспышка",
/*MISSING*/    "flash",
/*MISSING*/    "flash",
    }},
  {0xfc349b88, { // "Shoot'em up mode"
   "Tryb strzelanki",
/*MISSING*/    "Shoot'em up mode",
   "Mód střílečky",
   "режим стрельбы",
/*MISSING*/    "Shoot'em up mode",
/*MISSING*/    "Shoot'em up mode",
    }},
  {0xfc66f74c, { // "player 5 spin"
   "gracz 5 obrót",
/*MISSING*/    "player 5 spin",
   "hráč 5 otáčení",
   "игрок 5 вертит",
   "Spieler 5 drehen",
/*MISSING*/    "player 5 spin",
    }},
  {0xfc89b511, { // "EN"
   "PL",
   "TR",
   "CZ",
   "RU",
   "DE",
   "PT-BR",
    }},
  {0xfc9e712a, { // "yes, and delete the config file"
   "tak, i skasuj plik konfiguracyjny",
/*MISSING*/    "yes, and delete the config file",
   "ano, a vymaž i konfigurační soubor",
   "да, и удалить файл с настройками",
/*MISSING*/    "yes, and delete the config file",
/*MISSING*/    "yes, and delete the config file",
    }},
  {0xfcb63ab4, { // "Kills required: %1 (%2).\n"
   "Wymagani pokonani przeciwnicy: %1 (%2).\n",
/*MISSING*/    "Kills required: %1 (%2).\n",
   "Nutno zabít nepřátel: %1 (%2).\n",
   "Убить монстров необходимо: %1 (%2).\n",
   "Notwendige Tötungen: %1 (%2).\n",
/*MISSING*/    "Kills required: %1 (%2).\n",
    }},
  {0xfcd24c71, { // "If we are viewing an equidistant g absolute units below a plane, from a point c absolute units above the plane, this corresponds to viewing a Minkowski hyperboloid from a point tanh(g)/tanh(c) units below the center. This in turn corresponds to the Poincaré model for g=c, and Klein-Beltrami model for g=0."
   "Jeśli ekwidystantną powierzchnię g jednostek pod płaszczyzną oglądamy z punktu na wysokości c, odpowiada to oglądaniu hiperboloidy Monkowskiego z punktu tanh(g)/tanh(c) poniżej jej środka. To z kolei odpowiada modelowi Poincaré dla g=c, i Kleina-Beltramiego dla g=0.",
/*MISSING*/    "If we are viewing an equidistant g absolute units below a plane, from a point c absolute units above the plane, this corresponds to viewing a Minkowski hyperboloid from a point tanh(g)/tanh(c) units below the center. This in turn corresponds to the Poincaré model for g=c, and Klein-Beltrami model for g=0.",
   "Díváme-li se na ekvidistantní plochu, která se nachází g absolutních jednotek pod rovinou, z bodu, který je c absolutních jednotek nad rovinou, je to ekvivalentní tomu, když se díváme na Minkowskiho hyperboloid z bodu, který je tanh(g)/tanh(c) jednotek pod jeho středem. To pak odpovídá Poincarého modelu, když je g=c, a Klein-Beltramiho modelu, když je g=0.",
   "Если мы смотрим на эквидистанту на глубине g единиц из точки на высоте c единиц, это соответствует гиперболоиду Минковского, на который мы смотрим из точки на tanh(g)/tanh(c) единиц ниже центра. То есть это соответствует модели Пуанкаре при g=c, и Клейна-Бельтрами -- при g=0.",
/*MISSING*/    "If we are viewing an equidistant g absolute units below a plane, from a point c absolute units above the plane, this corresponds to viewing a Minkowski hyperboloid from a point tanh(g)/tanh(c) units below the center. This in turn corresponds to the Poincaré model for g=c, and Klein-Beltrami model for g=0.",
/*MISSING*/    "If we are viewing an equidistant g absolute units below a plane, from a point c absolute units above the plane, this corresponds to viewing a Minkowski hyperboloid from a point tanh(g)/tanh(c) units below the center. This in turn corresponds to the Poincaré model for g=c, and Klein-Beltrami model for g=0.",
    }},
  {0xfd264e48, { // "Watch the Minkowski hyperboloid or the hypersian rug mode with the red/cyan 3D glasses."
   "Oglądaj hiperboloidę Minkowskiego albo tryb hiperskiego dywanu przez czerwono-turkusowe okulary 3D.",
/*MISSING*/    "Watch the Minkowski hyperboloid or the hypersian rug mode with the red/cyan 3D glasses.",
   "Sleduj Minkowského hyperboloid nebo mód hyperského koberce s pomocí červenomodrých 3D brýlí.",
   "Смотреть гиперболоид Минковского или режим гиперсидского через красно-голубые 3D очки.",
/*MISSING*/    "Watch the Minkowski hyperboloid or the hypersian rug mode with the red/cyan 3D glasses.",
/*MISSING*/    "Watch the Minkowski hyperboloid or the hypersian rug mode with the red/cyan 3D glasses.",
    }},
  {0xfd2f1f8a, { // "This move appears dangerous -- are you sure?"
   "Ten ruch wygląda na niebezpieczny -- na pewno?",
/*MISSING*/    "This move appears dangerous -- are you sure?",
   "Tento tah vypadá nebezpečně -- určitě ho chceš udělat?",
   "Этот ход выглядит опасным -- вы уверены?",
   "Dieser Zug scheint gefährlich zu sein - bist du sicher?",
/*MISSING*/    "This move appears dangerous -- are you sure?",
    }},
  {0xfd351e0b, { // "This giant ape thinks that you are an enemy."
   "Ta małpa uważa, że jesteś jej przeciwnikiem.",
   "Bu devasa maymun senin bir düşman olduğunu düşünüyor.",
   "Tato obrovská opice si myslí, že jsi její nepřítel.",
   "Эта огромная обезьяна считает Вас своим врагом.",
   "Dieser riesige Affe glaubt, dass du ein Feind bist.",
/*MISSING*/    "This giant ape thinks that you are an enemy.",
    }},
  {0xfd4f8051, { // "The Knights stare at you!"
   "Rycerze gapią się na Ciebie ze zdziwieniem!",
   "Şövalyeler sana dik dik bakıyor!",
   "Rytíři na tebe zírají!",
   "Рыцари смотрят на Вас!",
   "Die Ritter starren dich an...",
/*MISSING*/    "The Knights stare at you!",
    }},
  {0xfd51d9ba, { // "Saved the high quality shot to %1"
   "Zapisano zrzut HQ do %1",
   "HQ resim %1 konumuna kaydedildi",
   "HQ snímek uložen do %1",
   "Скрин высокого качества сохранён в %1",
   "HQ Screenshot nach %1 gespeichert",
/*MISSING*/    "Saved the high quality shot to %1",
    }},
  {0xfd8f6cfa, { // "player character"
   "postać gracza",
   "oyuncunun karakteri",
   "tvoje postava",
   "Ваш персонаж",
   "Charakter",
   "personagem do jogador",
    }},
  {0xfe55d537, { // "You feel the Orbs of Yendor in the Crossroads..."
   "Wyczuwasz Sfery Yendoru na Skrzyżowaniach...",
/*MISSING*/    "You feel the Orbs of Yendor in the Crossroads...",
   "Cítíš Yendorské Sféry na Křižovatce...",
   "Ты чувствуешь Сферы Йендора на перекрёстке...",
/*MISSING*/    "You feel the Orbs of Yendor in the Crossroads...",
/*MISSING*/    "You feel the Orbs of Yendor in the Crossroads...",
    }},
  {0xfe662a68, { // "Cannot jump on %the1!"
   "Nie możesz skoczyć na %1!",
   "%1 hücresine atlayamazsın!",
   "Nemůžeš skočit na %1!",
   "Нельзя прыгнуть на %1!",
   "Du kannst nicht auf %den1 %a1 springen!",
/*MISSING*/    "Cannot jump on %the1!",
    }},
  {0xfe6ce188, { // "Each challenge gets harder after each victory"
   "Każda misja staje się trudniejsza z każdym zwycięstwem",
   "Her zaferinden sonra ek görevler zorlaşır.",
   "Obtížnost každé mise se zvýší, když ji splníš",
   "Миссия становится сложнее с каждой победой",
   "Die Schwierigkeit einer Herausforderung steigt nach jedem Sieg.",
/*MISSING*/    "Each challenge gets harder after each victory",
    }},
  {0xfe879321, { // "A land where you cannot stop, because every piece of floor is extremely unstable. Only monsters who can run forever are able to survive there, and only phoenix feathers are so light that they do not disturb the floor.\n"
   "Kraina, w której nie możesz przestać się ruszać, bo grunt jest wszędzie bardzo niestabilny. Jedynie stworzenia będące w stanie wiecznie się ruszać mogą tu przeżyć i jedynie pióra feniksa są na tyle lekkie, by nie zaburzyć podłogi.",
   "Duramayacağınız bir diyar...Çünkü zeminin her noktası oldukça kırılgan. Sadece sonsuza kadar koşabilen canavarlar burada yaşayabilirler ve anka telekleri öyle hafiftir ki, zemine zarar vermezler.",
   "Kraj, ve kterém se nemůžeš zastavit, protože každý kousek podlahy je nesmírně nestabilní. Mohou tu přežít pouze netvoři, kteří dokážou věčně běžet, a pouze fénixova pírka jsou tak lehká, že nenaruší podlahu.",
   "Земля, в которой нельзя остановиться, потому что пол крайне неустойчив. Лишь монстры, способные всегда бежать, могут здесь выжить, и лишь перья феникса столь легки, что не тревожат землю.\n",
   "Ein Land, in dem du nicht stehen bleiben darfst, denn jedes Stück Boden ist extrem instabil. Nur Monster, die ewig laufen können, sind in der Lage hier zu überleben. Nur Phönixfedern sind leicht genug, den Boden nicht zu beeinflussen.\n",
/*MISSING*/    "A land where you cannot stop, because every piece of floor is extremely unstable. Only monsters who can run forever are able to survive there, and only phoenix feathers are so light that they do not disturb the floor.\n",
    }},
  {0xfe8aaa55, { // "continue playing"
   "kontynuuj grę",
   "oyuna devam et",
   "pokračuj ve hře",
   "продолжить игру",
   "Weiterspielen",
/*MISSING*/    "continue playing",
    }},
#if REPEATED
  {0xfe8aaa56, { // "continue playing [repeat]"
   "kontynuuj grę",
/*MISSING*/    "continue playing [repeat]",
   "pokračuj ve hře",
   "продолжить игру",
   "Spiel fortsetzen",
/*MISSING*/    "continue playing [repeat]",
    }},
#endif
  {0xfea7d9b1, { // "You are playing in the Orb Strategy Mode. Collecting treasure gives you access to magical Orb powers. In this mode, unlocking requirements are generally higher, and several quests and lands give you extremely powerful Orbs of the Mirror.\n"
   "Grasz w trybie strategii sfer. Zebrane skarby dają Ci dostęp do magicznych mocy. W tym trybie wymagania są generalnie wyższe, i  niektóre krainy i misje dają Ci bardzo potężne Sfery Lustra.\n",
/*MISSING*/    "You are playing in the Orb Strategy Mode. Collecting treasure gives you access to magical Orb powers. In this mode, unlocking requirements are generally higher, and several quests and lands give you extremely powerful Orbs of the Mirror.\n",
   "Právì hraješ ve sférostrategickém módu. Sbírání pokladù ti dá pøístup k magickým silám Sfér. V tomto módu jsou požadavky na odemykání nových vìcí obecnì vyšší a nìkolik úkolù a krajù ti dává nesmírnì silné Sféry Zrcadla.\n",
   "Ты играешь в режим стратегии сфер. Собирай сокровища и получай доступ к силе магических сфер. В этом режиме условия открытия доступа, как правило, выше, а некоторые земли и миссии дадут тебе очень сильные Сферы зеркала.\n",
/*MISSING*/    "You are playing in the Orb Strategy Mode. Collecting treasure gives you access to magical Orb powers. In this mode, unlocking requirements are generally higher, and several quests and lands give you extremely powerful Orbs of the Mirror.\n",
/*MISSING*/    "You are playing in the Orb Strategy Mode. Collecting treasure gives you access to magical Orb powers. In this mode, unlocking requirements are generally higher, and several quests and lands give you extremely powerful Orbs of the Mirror.\n",
    }},
  {0xfeafb644, { // "shoot'em up and multiplayer"
   "strzelanka / wielu graczy",
/*MISSING*/    "shoot'em up and multiplayer",
   "střílečka / více hráčů",
   "стрельба / игроки",
   "shmup und Mehrspieler",
/*MISSING*/    "shoot'em up and multiplayer",
    }},
  {0xfebfb8a6, { // "move left"
   "idź w lewo",
   "sola git",
   "doleva",
   "идти налево",
   "Nach links bewegen",
/*MISSING*/    "move left",
    }},
  {0xfec7e1e9, { // "(press ESC during the game to review your quest)"
   "(naciśnij ESC w trakcie gry, by zobaczyć stan swojej misji)",
   "(Görevine bakmak için oyun sırasında ESC'ye bas.)",
   "(své úkoly můžeš během hry zobrazit stiskem ESC)",
   "(нажмите ESC во время игры, чтобы посмотреть ваши миссии)",
   "(Drücke ESC während des Spiels, um deine aktuelle Quest anzuzeigen)",
   "(pressione ESC durante o jogo para revisar suas missões)",
    }},
  {0xfec941a9, { // "\n\nAfter the Trolls leave, you have 750 turns to collect %the1, or it gets stolen."
   "\n\nGdy Trolle wyruszą, masz 750 kolejek na zebranie skarbów -- jeśli tego nie zrobisz, %P1 zostaną ukradzione.",
/*MISSING*/    "\n\nAfter the Trolls leave, you have 750 turns to collect %the1, or it gets stolen.",
   "\n\nKdyž Trollové odejdou, máš 750 kol na získání pokladu -- jinak ho něco ukradne.",
   "\n\nКогда тролли уйдут, у вас будет 750 ходов, чтобы собрать %a1, иначе украдут.",
/*MISSING*/    "\n\nAfter the Trolls leave, you have 750 turns to collect %the1, or it gets stolen.",
/*MISSING*/    "\n\nAfter the Trolls leave, you have 750 turns to collect %the1, or it gets stolen.",
    }},
  {0xfefe0ac5, { // "This cave contains walls which are somehow living. After each turn, each cell counts the number of living wall and living floor cells around it, and if it is currently of a different type than the majority of cells around it, it switches. Items count as three floor cells, and dead Trolls count as five wall cells. Some foreign monsters also count as floor or wall cells.\n"
   "Ściany tej jaskini sa żywe. W każdej kolejce każde pole liczy, ile wokół niego jest pól ściany i pól ziemi i jeśli jest innego typu niż większość pól wokół niego, to zmienia typ na przeciwny. Przedmioty liczą się jako 3 pola ziemi, martwe Trolle jako 5 pól ściany. Niektóre potwory z innych krain również liczą się jako ziemia lub ściana.\n",
   "Bu mağara nasılsa yaşayan duvarlarla kaplıdır. Her tur her hücre çevresindeki yaşayan zemin ve yaşayan duvar hücrelerini sayar ve eğer çevresindeki çoğunluktan farklı bir türdeyse çoğunluğa dönüşür. Eşyalar üç zemin hücresi, ölü Troller beş duvar hücresi olarak sayılır.\n",
   "Zdi této jeskyně jsou, kdovíjak, živé. Po každém kole každé políčko spočítá počet políček s živoucími zdmi a políček s živoucími podlahami kolem sebe a pokud má jiný typ než většina jeho sousedů, změní se ze zdi na podlahu nebo opačně. Předměty se počítají jako tři políčka podlahy, mrtví Šutrollové jako pět políček zdí. Někteří cizí netvoři se také počítají jako podlaha nebo zeď.\n",
   "В этой пещере даже стены живые. Каждый ход каждая клетка считает количество клеток живого пола и живых стен среди соседних, и если клеток другого типа больше, она меняет тип. Предметы считаются за 3 клетки пола, мёртвые тролли - за 5 клеток стен.Некоторые монстры из других земель тоже могут считаться за пол или стены.\n",
   "Diese Höhle enthält Wände, die auf irgendeine Art leben. Nach jedem Zug zählt jede Zelle die Zahl von lebenden Wänden und Böden um sich herum. Wenn sie von anderer Art als die Mehrzahl der umgebenden Zellen ist, wandelt sie sich um. Gegenstände zählen als drei Böden; tote Trolle zählen als fünf Wände. Manche fremde Monster zählen auch als Böden oder Wände.\n",
/*MISSING*/    "This cave contains walls which are somehow living. After each turn, each cell counts the number of living wall and living floor cells around it, and if it is currently of a different type than the majority of cells around it, it switches. Items count as three floor cells, and dead Trolls count as five wall cells. Some foreign monsters also count as floor or wall cells.\n",
    }},
  {0xff26e9bf, { // "You have improved both your real time and turn count. Congratulations!"
   "Poprawi%łeś0 swój najlepszy czas rzeczywisty i liczbę ruchów. Gratulacje!",
   "Hem gerçek zamanını hem de tur sayını geliştirdin! Tebrikler! ",
   "Zlepši%l0 jsi svůj reálný čas i svůj počet kol. Gratulujeme!",
   "Вы улучшили результаты по реальному времени и по числу ходов! Поздравляем",
   "Du hast sowohl weniger Echtzeit als auch Züge als je zuvor benötigt. Glückwunsch!",
   "Você superou tanto seu contador de turnos quanto seu tempo real. Parabéns!",
    }},
  {0xffa845cd, { // "You slash %the1."
   "Poci%ąłeś0 %a1.",
/*MISSING*/    "You slash %the1.",
   "Sek%l0 jsi %a1.",
   "Вы разрубили %a1.",
   "Du schlägst %den1 %a1.",
/*MISSING*/    "You slash %the1.",
    }},
  {0xffb51405, { // "review your quest"
   "przegląd misji",
   "görevine gözat",
   "přehled úkolů",
   "смотреть квест",
   "Quest anzeigen",
/*MISSING*/    "review your quest",
    }},
  {0xffb8423c, { // "basic configuration"
   "podstawowa konfiguracja",
   "temel ayarlar",
   "základní konfigurace",
   "базовые настройки",
   "Einstellungen",
/*MISSING*/    "basic configuration",
    }},
  {0xffcd55dd, { // " (next at %1)"
   " (kolejny przy %1)",
/*MISSING*/    " (next at %1)",
   " (další pøi %1)",
   " (следующий при %1)",
/*MISSING*/    " (next at %1)",
/*MISSING*/    " (next at %1)",
    }},
  };

fullnoun all_nouns[386] = {
/*MISSING*/   {0x0, 0, { // """"
    {2/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""},
    {2/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""},
    {2/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""},
    {2/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""},
    {2/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""},
    {0/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""/*MISSING*/ , ""},
    }},
  {0x3c73fa, 0, { // ""Plane of Water""
    {1, "Strefa Wody", "Strefy Wody", "Strefę Wody", "w Strefie Wody"},
    {1, "Su Düzlemi", "Su Düzlemleri", "Su Düzlemini", "Su Düzleminde"},
    {1, "Rovina Vody", "Roviny Vody", "Rovinu Vody", "v Rovině Vody"},
    {1, "Плоскость Воды", "Плоскости Воды", "Плоскость Воды", "в Плоскости Воды"},
    {1, "Wasserebene", "Wasserebenen", "Wasserebene", "Wasserebene"},
    {0/*MISSING*/ , "Plane of Water"/*MISSING*/ , "Plane of Water"/*MISSING*/ , "Plane of Water"/*MISSING*/ , "Plane of Water"},
    }},
  {0x64ad71, 0, { // ""Wine""
    {2, "Wino", "Wina", "Wino", "Winem"},
    {2, "Şarap", "Şaraplar", "Şarabı", "Şarapla"},
    {2, "Láhev vína", "Lahve vína", "Láhev vína", "Lahví vína"},
    {2, "Вино", "Вина", "Вино", "Вином"},
    {0, "Wein", "Weine", "Wein", "Wein"},
    {0/*MISSING*/ , "Wine"/*MISSING*/ , "Wine"/*MISSING*/ , "Wine"/*MISSING*/ , "Wine"},
    }},
  {0x11132be, 0, { // ""Water Elemental""
    {0, "Wodny Żywiołak", "Wodne Żywiołaki", "Wodnego Żywiołaka", "Wodnym Żywiołakiem"},
    {0, "Su Özütü", "Su Özütleri", "Su Özütünü", "Su Özütüyle"},
    {0, "Elementál Vody", "Elementálové Vody", "Elementála Vody", "Elementálem Vody"},
    {0, "Водный элементаль", "Водные элементали", "Водного элементаля", "Водным элементалем"},
    {0, "Wasserelementar", "Wasserelementare", "Wasserelementar", "Wasserelementar"},
    {0/*MISSING*/ , "Water Elemental"/*MISSING*/ , "Water Elemental"/*MISSING*/ , "Water Elemental"/*MISSING*/ , "Water Elemental"},
    }},
  {0x1e6876e, 0, { // ""False Prince""
    {0, "Fałszywy Książę", "Fałszywi Książęta", "Fałszywego Księcia", "Fałszywym Księciem"},
    {0/*MISSING*/ , "False Prince"/*MISSING*/ , "False Prince"/*MISSING*/ , "False Prince"/*MISSING*/ , "False Prince"},
    {0, "Falešný princ", "Falešní princové", "Falešného prince", "Falešným princem"},
    {0, "Ложный принц", "Ложные принцы", "Ложного принца", "Ложным принцем"},
    {0, "Falscher Prinz", "Falsche Prinzen", "Falschen Prinz", "Falschen Prinz"},
    {0/*MISSING*/ , "False Prince"/*MISSING*/ , "False Prince"/*MISSING*/ , "False Prince"/*MISSING*/ , "False Prince"},
    }},
  {0x2d68a1c, 0, { // ""Orb of Thorns""
    {1, "Sfera Cierni", "Sfery Cierni", "Sferę Cierni", "Sferą Cierni"},
    {1, "Diken Küresi", "Diken Küreleri", "Diken Küresini", "Diken Küresiyle"},
    {1, "Sféra Trnů", "Sféry Trnů", "Sféru Trnů", "Sférou Trnů"},
    {1, "Сфера Шипов", "Сферы Шипов", "Сферу Шипов", "Сферой Шипов"},
    {0, "Orb der Dornen", "Orbs der Dornen", "Orb der Dornen", "Orb der Dornen"},
    {0, "Orbe de Espinhos", "Orbes de Espinhos"/*MISSING*/ , "Orb of Thorns", "Orbe de Espinhos"},
    }},
  {0x3bf4992, 0, { // ""Cellular Automaton""
    {0, "Automat Komórkowy", "Automaty Komórkowe", "Automat Komórkowy", "w Automacie Komórkowym"},
    {0/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"},
    {0, "Buněčný automat", "Buněčné automaty", "Buněčný automat", "v Buněčném automatu"},
    {3, "Клеточный автомат", "Клеточные автоматы", "Клеточный автомат", "в Клеточном автомате"},
    {0/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"},
    {0/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"/*MISSING*/ , "Cellular Automaton"},
    }},
  {0x594962b, 0, { // ""Platform""
    {1, "Platforma", "Platformy", "Platformę", "Platformą"},
    {1, "Platform", "Platformlar", "Platformu", "Platformla"},
    {1, "Plošinka", "Plošinky", "Plošinku", "Plošinkou"},
    {1, "Платформа", "Платформы", "Платформу", "Платформой"},
    {0/*MISSING*/ , "Platform"/*MISSING*/ , "Platform"/*MISSING*/ , "Platform"/*MISSING*/ , "Platform"},
    {0/*MISSING*/ , "Platform"/*MISSING*/ , "Platform"/*MISSING*/ , "Platform"/*MISSING*/ , "Platform"},
    }},
  {0x59bc9da, 0, { // ""Galápagos""
    {0, "Galápagos", "Galápagos", "Galápagos", "na Galápagos"},
    {0/*MISSING*/ , "Galápagos"/*MISSING*/ , "Galápagos"/*MISSING*/ , "Galápagos"/*MISSING*/ , "Galápagos"},
    {1, "Galapágy", "Galapágy", "Galapágy", "na Galapágách"},
    {0, "Галапагосы", "Галапагосы", "Галапагосы", "на Галапагосах"},
    {0, "Galápagos", "Galápagos", "in Galápagos", "in Galápagos"},
    {0/*MISSING*/ , "Galápagos"/*MISSING*/ , "Galápagos"/*MISSING*/ , "Galápagos"/*MISSING*/ , "Galápagos"},
    }},
  {0x72878ed, 0, { // ""dead troll""
    {0, "martwy troll", "martwe trolle", "martwego trolla", "martwym trollem"},
    {0, "Ölü Trol", "Ölü Troller", "Ölü Trolü", "Ölü Trolle"},
    {0, "mrtvý troll", "mrtví trollové", "mrtvého trolla", "mrtvým trollem"},
    {0, "мёртвый тролль", "мёртвые тролли", "мёртвого тролля", "мёртвым троллем"},
    {0, "toter Troll", "toten Trolle", "toten Troll", "toten Troll"},
    {0, "troll morto", "trolls mortos"/*MISSING*/ , "dead troll", "troll morto"},
    }},
  {0x73dabf9, 0, { // ""Crossroads V""
    {3, "Skrzyżowanie V", "Skrzyżowania V", "Skrzyżowanie V", "na Skrzyżowaniu V"},
    {0/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"},
    {3, "Křižovatka V", "Křižovatky V", "Křižovatku V", "na Křižovatce V"},
    {3, "Перекрёсток V", "Перекрёстки V", "Перекрёсток V", "на Перекрёстке V"},
    {0/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"},
    {0/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"/*MISSING*/ , "Crossroads V"},
    }},
  {0x7486caa, 0, { // ""Ruby""
    {3, "Rubin", "Rubiny", "Rubin", "Rubinem"},
    {3, "Yakut", "Yakutlar", "Yakutu", "Yakutla"},
    {3, "Rubín", "Rubíny", "Rubín", "Rubínem"},
    {3, "Рубин", "Рубины", "Рубин", "Рубином"},
    {0, "Rubin", "Rubine", "Rubin", "Rubin"},
    {0, "Rubi", "Rubis"/*MISSING*/ , "Ruby", "Rubi"},
    }},
  {0x7ae3b95, 0, { // ""Mouse""
    {1, "Myszka", "Myszki", "Myszkę", "Myszką"},
    {1, "Fare", "Fareler", "Fareyi", "Fareyle"},
    {1, "Myška", "Myška", "Myšku", "Myškou"},
    {1, "Мышка", "Мышки", "Мышку", "Мышкой"},
    {1, "Maus", "Mäuse", "Maus", "Maus"},
    {0/*MISSING*/ , "Mouse"/*MISSING*/ , "Mouse"/*MISSING*/ , "Mouse"/*MISSING*/ , "Mouse"},
    }},
  {0x7d0776e, 0, { // ""Orb of Fire""
    {1, "Sfera Ognia", "Sfery Ognia", "Sferę Ognia", "Sferą Ognia"},
    {1, "Ateş Küresi", "Ateş Küreleri", "Ateş Küresini", "Ateş Küresiyle"},
    {1, "Sféra Ohně", "Sféry Ohně", "Sféru Ohně", "Sférou Ohně"},
    {1, "Сфера Огня", "Сферы Огня", "Сферу Огня", "Сферой Огня"},
    {0, "Orb des Feuers", "Orbs des Feuers", "Orb des Feuers", "Orb des Feuers"},
    {0/*MISSING*/ , "Orb of Fire"/*MISSING*/ , "Orb of Fire"/*MISSING*/ , "Orb of Fire"/*MISSING*/ , "Orb of Fire"},
    }},
  {0x982eac3, 0, { // ""Reptile""
    {1, "Jaszczurka", "Jaszczurki", "Jaszczurkę", "Jaszczurką"},
    {0/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"},
    {1, "Ještěrka", "Ještěrky", "Ještěrku", "Ještěrce"},
    {1, "Ящерица", "Ящерицы", "Ящерицу", "Ящерицей"},
    {0/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"},
    {0/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"/*MISSING*/ , "Reptile"},
    }},
  {0xa2dba6e, 0, { // ""Onyx""
    {3, "Onyks", "Onyksy", "Onyks", "Onyksem"},
    {3, "Akik", "Akikler", "Akiği", "Akikle"},
    {3, "Onyx", "Onyxy", "Onyx", "Onyxem"},
    {3, "Оникс", "Ониксы", "Оникс", "Ониксом"},
    {0, "Onyx", "Onyxe", "Onyx", "Onyx"},
    {0/*MISSING*/ , "Onyx"/*MISSING*/ , "Onyx"/*MISSING*/ , "Onyx"/*MISSING*/ , "Onyx"},
    }},
  {0xacc4f88, 0, { // ""Shadow""
    {0, "Cień", "Cienie", "Cień", "Cieniem"},
    {0, "Gölge", "Gölgeler", "Gölgeyi", "Gölgeyle"},
    {0, "Stín", "Stínové", "Stína", "Stínem"},
    {1, "Тень", "Тени", "Тень", "Тенью"},
    {0, "Schatten", "Schatten", "Schatten", "Schatten"},
    {1, "Sombra", "Sombras"/*MISSING*/ , "Shadow", "Sombra"},
    }},
  {0xc15c294, 0, { // ""Mirror Land""
    {1, "Kraina Luster", "Kraina Luster", "Krainę Luster", "w Krainie Luster"},
    {1, "Aynalar Diyarı", "Aynalar Diyarları", "Aynalar Diyarını", "Aynalar Diyarında"},
    {3, "Kraj zrcadel", "Kraje zrcadel", "Kraj zrcadel", "v Kraji zrcadel"},
    {1, "Зеркальная земля", "Зеркальные земли", "Зеркальную землю", "в Зеркальной земле"},
    {0, "Spiegelland", "Spiegelländer", "im Spiegelland", "im Spiegelland"},
    {1, "Terra dos Espelhos", "Terras dos Espelhos"/*MISSING*/ , "Mirror Land", "Terra dos Espelhos"},
    }},
  {0xc23bfef, 0, { // ""Minefield""
    {2, "Pole Minowe", "Pola Minowe", "Pole Minowe", "na Polu Minowym"},
    {2, "Mayıntarlası", "Mayıntarlaları", "Mayıntarlasını", "Mayıntarlasında"},
    {2, "Minové pole", "Minová pole", "Minové pole", "v Minovém poli"},
    {2, "Минное поле", "Минные поля", "Минное поле", "на Минном поле"},
    {2, "Minenfeld", "Minenfelder", "im Minenfeld", "im Minenfeld"},
    {0/*MISSING*/ , "Minefield"/*MISSING*/ , "Minefield"/*MISSING*/ , "Minefield"/*MISSING*/ , "Minefield"},
    }},
  {0xc271d6d, 0, { // ""Silver""
    {2, "Srebro", "Srebra", "Srebro", "Srebrem"},
    {2, "Gümüş", "Gümüşler", "Gümüşü", "Gümüşle"},
    {2, "Kus stříbra", "Kusy stříbra", "Kus stříbra", "Kusem stříbra"},
    {2, "Серебро", "Серебро", "Серебро", "Серебром"},
    {2, "Silber", "Silber", "Silber", "Silber"},
    {0/*MISSING*/ , "Silver"/*MISSING*/ , "Silver"/*MISSING*/ , "Silver"/*MISSING*/ , "Silver"},
    }},
  {0xc9d9e0c, 0, { // ""Eagle""
    {0, "Orzeł", "Orły", "Orła", "Orłem"},
    {0, "Kartal", "Kartallar", "Kartalı", "Kartalla"},
    {0, "Orel", "Orli", "Orla", "Orlem"},
    {0, "Орёл", "Орлы", "Орла", "Орлом"},
    {0, "Adler", "Adler", "Adler", "Adler"},
    {1, "Águia", "Águias"/*MISSING*/ , "Eagle", "Águia"},
    }},
  {0xd0f6f4b, 0, { // ""Red Rock Valley""
    {1, "Dolina Czerwonej Skały", "Dolina Czerwonej Skały", "Dolinę Czerwonej Skały", "w Dolinie Czerwonej Skały"},
    {1, "Kızılkaya Vadisi", "Kızılkaya Vadileri", "Kızılkaya Vadisini", "Kızılkaya Vadisinde"},
    {2, "Údolí červených skal", "Údolí červených skal", "Údolí červených skal", "v Údolí červených skal"},
    {1, "Долина Красных Скал", "Долины Красных Скал", "Далину Красных Скал", "в Долине Красных Скал"},
    {2, "Rotfelsental", "Rotfelsentäler", "im Rotfelsental", "im Rotfelsental"},
    {0/*MISSING*/ , "Red Rock Valley"/*MISSING*/ , "Red Rock Valley"/*MISSING*/ , "Red Rock Valley"/*MISSING*/ , "Red Rock Valley"},
    }},
  {0xdf3c837, 0, { // ""Key""
    {3, "Klucz", "Klucze", "Klucz", "Kluczem"},
    {3, "Anahtar", "Anahtarlar", "Anahtarı", "Anahtarla"},
    {3, "Klíč", "Klíče", "Klíč", "Klíčem"},
    {3, "Ключ", "Ключи", "Ключ", "Ключом"},
    {0, "Schlüssel", "Schlüssel", "Schlüssel", "Schlüssel"},
    {1, "Chave", "Chaves"/*MISSING*/ , "Key", "Chave"},
    }},
  {0xeb45f25, 0, { // ""Dragon Scale""
    {1, "Smocza Łuska", "Smocze Łuski", "Smoczą Łuskę", "Smoczą Łuską"},
    {0/*MISSING*/ , "Dragon Scale"/*MISSING*/ , "Dragon Scale"/*MISSING*/ , "Dragon Scale"/*MISSING*/ , "Dragon Scale"},
    {1, "Dračí šupina", "Dračí šupiny", "Dračí šupinu", "Dračí šupinou"},
    {1, "чешуя дракона", "чешуи дракона", "чешую дракона", "чешуёй дракона"},
    {1, "Drachenschuppe", "Drachenschuppen", "Drachenschuppe", "Drachenschuppe"},
    {0/*MISSING*/ , "Dragon Scale"/*MISSING*/ , "Dragon Scale"/*MISSING*/ , "Dragon Scale"/*MISSING*/ , "Dragon Scale"},
    }},
  {0xf6089ad, 0, { // ""Draugr""
    {0, "Draugr", "Draugry", "Draugra", "Draugrem"},
    {0/*MISSING*/ , "Draugr"/*MISSING*/ , "Draugr"/*MISSING*/ , "Draugr"/*MISSING*/ , "Draugr"},
    {0, "Draugr", "Draugrové", "Draugra", "Draugrem"},
    {0, "Драугр", "Драугры", "Драугра", "Драугром"},
    {0, "Draugr", "Draugar", "Draugr", "Draugr"},
    {0/*MISSING*/ , "Draugr"/*MISSING*/ , "Draugr"/*MISSING*/ , "Draugr"/*MISSING*/ , "Draugr"},
    }},
  {0x10ddbfc9, 0, { // ""Orb of the Mind""
    {1, "Sfera Psychiki", "Sfery Psychiki", "Sferę Psychiki", "Sferą Psychiki"},
    {1, "Psychiki Küresi", "Psychiki Küreleri", "Psychiki Küresini", "Psychiki Küresiyle"},
    {1, "Sféra Mysli", "Sféry Mysli", "Sféru Mysli", "Sférou Mysli"},
    {1, "Сфера Разума", "Сферы Разума", "Сферу Разума", "Сферой Разума"},
    {0, "Orb der Psyche", "Orbs der Psyche", "Orb der Psyche", "Orb der Psyche"},
    {0/*MISSING*/ , "Orb of the Mind"/*MISSING*/ , "Orb of the Mind"/*MISSING*/ , "Orb of the Mind"/*MISSING*/ , "Orb of the Mind"},
    }},
  {0x10ff2cb0, 0, { // ""Hive""
    {2, "Mrowisko", "Mrowisko", "Mrowisko", "w Mrowisku"},
    {2, "Kovan", "Kovanlar", "Kovanı", "Kovanda"},
    {3, "Úl", "Úly", "Úl", "v Úlu"},
    {3, "Улей", "Ульи", "Улей", "в Улье"},
    {2, "Nest", "Nest", "im Nest", "im Nest"},
    {0/*MISSING*/ , "Hive"/*MISSING*/ , "Hive"/*MISSING*/ , "Hive"/*MISSING*/ , "Hive"},
    }},
  {0x11b70d9f, 0, { // ""Red Hyperbug""
    {0, "Czerwony Hiperinsekt", "Czerwone Hiperinsekty", "Czerwonego Hiperinsekta", "Czerwonym Hiperinsektem"},
    {0, "Kırmızı Aşkınböcek", "Kırmızı Aşkınböcekler", "Kırmızı Aşkınböceği", "Kırmızı Aşkınböcekle"},
    {0, "Červený hyperhmyzák", "Červení hyperhmyzáci", "Červeného hyperhmyzáka", "Červeným hyperhmyzákem"},
    {0, "Красный гипержук", "Красные гипержуки", "Красного гипержука", "Красным гипержуком"},
    {0, "Roter Hyperkäfer", "Rote Hyperkäfer", "Roten Hyperkäfer", "Roten Hyperkäfer"},
    {0/*MISSING*/ , "Red Hyperbug"/*MISSING*/ , "Red Hyperbug"/*MISSING*/ , "Red Hyperbug"/*MISSING*/ , "Red Hyperbug"},
    }},
  {0x13551589, 0, { // ""Mirage""
    {0, "Miraż", "Miraże", "Miraż", "Mirażem"},
    {0, "Yansı", "Yansılar", "Yansıyı", "Yansıyla"},
    {3, "Přelud", "Přeludy", "Přelud", "Přeludem"},
    {3, "Мираж", "Миражи", "Мираж", "Миражом"},
    {1, "Spiegelung", "Spiegelungen", "Spiegelung", "Spiegelung"},
    {1, "Miragem", "Miragens"/*MISSING*/ , "Mirage", "Miragem"},
    }},
  {0x13f7d362, 0, { // ""Viking""
    {0, "Wiking", "Wikingowie", "Wikinga", "Wikingiem"},
    {0, "Viking", "Vikingler", "Vikingi", "Vikingle"},
    {0, "Viking", "Vikingové", "Vikinga", "Vikingem"},
    {0, "Викинг", "Викинги", "Викинга", "Викингом"},
    {0, "Wikinger", "Wikinger", "Wikinger", "Wikinger"},
    {0/*MISSING*/ , "Viking"/*MISSING*/ , "Viking"/*MISSING*/ , "Viking"/*MISSING*/ , "Viking"},
    }},
  {0x1422459d, 0, { // ""Green Hyperbug""
    {0, "Zielony Hiperinsekt", "Zielone Hiperinsekty", "Zielonego Hiperinsekta", "Zielonym Hiperinsektem"},
    {0, "Yeşil Aşkınböcek", "Yeşil Aşkınböcekler", "Yeşil Aşkınböceği", "Yeşil Aşkınböcekle"},
    {0, "Zelený hyperhmyzák", "Zelení hyperhmyzáci", "Zeleného hyperhmyzáka", "Zeleným hyperhmyzákem"},
    {0, "Зелёный гипержук", "Зелёные гипержуки", "Зелёного гипержука", "Зелёным гипержуком"},
    {0, "Grüner Hyperkäfer", "Grüne Hyperkäfer", "Grünen Hyperkäfer", "Grünen Hyperkäfer"},
    {0/*MISSING*/ , "Green Hyperbug"/*MISSING*/ , "Green Hyperbug"/*MISSING*/ , "Green Hyperbug"/*MISSING*/ , "Green Hyperbug"},
    }},
  {0x144bd658, 0, { // ""big bush""
    {3, "duży krzak", "duże krzaki", "duży krzak", "dużym krzakiem"},
    {0/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"},
    {3, "velký keř", "velké keře", "velký keř", "velkým keřem"},
    {3, "большой куст", "большие кусты", "большой куст", "большим кустом"},
    {0/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"},
    {0/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"/*MISSING*/ , "big bush"},
    }},
  {0x145a3ea0, 0, { // ""wall of Camelot""
    {1, "ściana Camelot", "ściany Camelot", "ścianę Camelot", "ścianą Camelot"},
    {1, "Kamelot Duvarı", "Kamelot Duvarları", "Kamelot Duvarını", "Kamelot Duvarıyla"},
    {1, "hradba Camelotu", "hradby Camelotu", "hradbu Camelotu", "hradbou Camelotu"},
    {1, "стена Камелота", "стены Камелота", "стену Камелота", "стеной Камелота"},
    {1, "Wand von Camelot", "Wände von Camelot", "Wand von Camelot", "Wand von Camelot"},
    {0/*MISSING*/ , "wall of Camelot"/*MISSING*/ , "wall of Camelot"/*MISSING*/ , "wall of Camelot"/*MISSING*/ , "wall of Camelot"},
    }},
  {0x146a6b1a, 0, { // ""Tentacle""
    {1, "Macka", "Macki", "Mackę", "Macką"},
    {0, "Dokunaç", "Dokunaçlar", "Dokunacı", "Dokunaçla"},
    {3, "Konec chapadla", "Konce chapadel", "Konec chapadla", "Koncem chapadla"},
    {2, "Щупальце", "Щупальца", "Щупальце", "Щупальцем"},
    {1, "Tentakel", "Tentakel", "Tentakel", "Tentakel"},
    {0, "Tentáculo", "Tentáculos"/*MISSING*/ , "Tentacle", "Tentáculo"},
    }},
  {0x14f61189, 0, { // ""small bush""
    {3, "krzaczek", "krzaczki", "krzaczek", "krzaczkiem"},
    {0/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"},
    {3, "małý keřík", "małé keříky", "małý keřík", "małým keříkem"},
    {3, "маленький куст", "маленькие кусты", "маленький куст", "маленьким кустом"},
    {0/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"},
    {0/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"/*MISSING*/ , "small bush"},
    }},
  {0x151e041d, 0, { // ""island""
    {1, "wyspa", "wyspy", "wyspę", "wyspą"},
    {1, "ada", "adalar", "adayı", "adayla"},
    {3, "ostrov", "ostrovy", "ostrov", "ostrovem"},
    {3, "остров", "острова", "остров", "островом"},
    {1, "Insel", "Inseln", "Insel", "Insel"},
    {0/*MISSING*/ , "island"/*MISSING*/ , "island"/*MISSING*/ , "island"/*MISSING*/ , "island"},
    }},
  {0x15d403e8, 0, { // ""Reptiles""
    {3, "Jaszczurki", "Jaszczurki", "Jaszczurki", "na Jaszczurkach"},
    {0/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"},
    {3, "Ještěrky", "Ještěrky", "Ještěrky", "na Ještěrkách"},
    {3, "Ящерицы", "Ящерицы", "Ящериц", "на Ящерицах"},
    {0/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"},
    {0/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"/*MISSING*/ , "Reptiles"},
    }},
  {0x1648dfe3, 0, { // ""Orb of Teleport""
    {1, "Sfera Teleportacji", "Sfery Teleportacji", "Sferę Teleportacji", "Sferą Teleportacji"},
    {1, "Işınlanma Küresi", "Işınlanma Küreleri", "Işınlanma Küresini", "Işınlanma Küresiyle"},
    {1, "Sféra Teleportace", "Sféry Teleportace", "Sféru Teleportace", "Sférou Teleportace"},
    {1, "Сфера Телепорта", "Сферы Телепорта", "Сферу Телепорта", "Сферой Телепорта"},
    {0, "Orb der Teleportation", "Orbs der Teleportation", "Orb der Teleportation", "Orb der Teleportation"},
    {0, "Orbe de Teleporte", "Orbes de Teleporte"/*MISSING*/ , "Orb of Teleport", "Orbe de Teleporte"},
    }},
  {0x16a1c106, 0, { // ""Orb of Space""
    {1, "Sfera Przestrzeni", "Sfery Przestrzeni", "Sferę Przestrzeni", "Sferą Przestrzeni"},
    {1, "Boşluk Küresi", "Boşluk Küreleri", "Boşluk Küresini", "Boşluk Küresiyle"},
    {1, "Sféra Prostoru", "Sféry Prostoru", "Sféru Prostoru", "Sférou Prostoru"},
    {1, "Сфера Пространства", "Сферы Пространства", "Сферу Пространства", "Сферой Пространства"},
    {0, "Orb des Raumes", "Orbs des Raumes", "Orb des Raumes", "Orb des Raumes"},
    {0/*MISSING*/ , "Orb of Space"/*MISSING*/ , "Orb of Space"/*MISSING*/ , "Orb of Space"/*MISSING*/ , "Orb of Space"},
    }},
  {0x17931c8f, 0, { // ""Yendorian Researcher""
    {0, "Yendoriański Badacz", "Yendoriańscy Badacze", "Yendoriańskiego Badacza", "Yendoriańskim Badaczem"},
    {0/*MISSING*/ , "Yendorian Researcher"/*MISSING*/ , "Yendorian Researcher"/*MISSING*/ , "Yendorian Researcher"/*MISSING*/ , "Yendorian Researcher"},
    {0, "Yendorský badatel", "Yendorští badatelé", "Yendorského badatele", "Yendorským badatelem"},
    {0, "Йендорский исследователь", "Йендорские исследователи", "Йендорского исследователя", "Йендорским исследователем"},
    {0, "Yendorianischer Forscher", "Yendorianische Forscher", "Yendorianischen Forscher", "Yendorianischen Forscher"},
    {0/*MISSING*/ , "Yendorian Researcher"/*MISSING*/ , "Yendorian Researcher"/*MISSING*/ , "Yendorian Researcher"/*MISSING*/ , "Yendorian Researcher"},
    }},
  {0x18630997, 0, { // ""Orb of Speed""
    {1, "Sfera Szybkości", "Sfery Szybkości", "Sferę Szybkości", "Sferą Szybkości"},
    {1, "Hız Küresi", "Hız Küreleri", "Hız Küresini", "Hız Küresiyle"},
    {1, "Sféra Rychlosti", "Sféry Rychlosti", "Sféru Rychlosti", "Sférou Rychlosti"},
    {1, "Сфера Скорости", "Сферы Скорости", "Сферу Скорости", "Сферой Скорости"},
    {0, "Orb der Geschwindigkeit", "Orbs der Geschwindigkeit", "Orb der Geschwindigkeit", "Orb der Geschwindigkeit"},
    {0, "Orbe de Velocidade", "Orbes de Velocidade"/*MISSING*/ , "Orb of Speed", "Orbe de Velocidade"},
    }},
  {0x1930f25c, 0, { // ""sand dune""
    {1, "wydma", "wydmy", "wydmę", "wydmą"},
    {1, "Kumul", "Kumullar", "Kumulu", "Kumulda"},
    {1, "písečná duna", "písečné duny", "písečnou dunu", "písečnou dunou"},
    {1, "дюна", "дюны", "дюну", "дюной"},
    {1, "Sanddüne", "Sanddünen", "Sanddüne", "Sanddüne"},
    {1, "duna de areia", "dunas de areia"/*MISSING*/ , "sand dune", "duna de areia"},
    }},
  {0x19468c49, 0, { // ""Lost Mountain""
    {1, "Zagubiona Góra", "Zagubione Góry", "Zagubioną Górą", "na Zagubionej Górze"},
    {0/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"},
    {1, "Ztracená hora", "Ztracené hory", "Ztracenou horu", "na Ztracené hoře"},
    {1, "Потерянная гора", "Потерянные горы", "Потерянную гору", "на Потерянной горе"},
    {0/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"},
    {0/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"/*MISSING*/ , "Lost Mountain"},
    }},
  {0x194752fa, 0, { // ""Friendly Ghost""
    {3, "Przyjazny Duch", "Przyjazne Duchy", "Przyjaznego Ducha", "Przyjaznym Duchem"},
    {0/*MISSING*/ , "Friendly Ghost"/*MISSING*/ , "Friendly Ghost"/*MISSING*/ , "Friendly Ghost"/*MISSING*/ , "Friendly Ghost"},
    {0, "Přátelský duch", "Přátelští duchové", "Přátelského ducha", "Přátelským duchem"},
    {0, "Дружественный дух", "Дружественные духи", "Дружественного духа", "Дружественным духом"},
    {0, "Freundlicher Geist", "Freundliche Geister", "Freundlichen Geist", "Freundlichen Geist"},
    {0/*MISSING*/ , "Friendly Ghost"/*MISSING*/ , "Friendly Ghost"/*MISSING*/ , "Friendly Ghost"/*MISSING*/ , "Friendly Ghost"},
    }},
  {0x19640701, 0, { // ""Raging Bull""
    {0, "Wściekły Byk", "Wściekłe Byki", "Wściekłego Byka", "Wściekłym Bykiem"},
    {0/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"},
    {0, "Vzteklý Býk", "Vzteklí Býci", "Vzteklého Býka", "Vzteklým Býkem"},
    {0, "Бешеный бык", "Бешеные быки", "Бешеного быка", "Бешеным быком"},
    {0/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"},
    {0/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"/*MISSING*/ , "Raging Bull"},
    }},
  {0x1972ecde, 0, { // ""Orb of the Mirror""
    {1, "Sfera Lustra", "Sfery Lustra", "Sferę Lustra", "Sferą Lustra"},
    {0/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"},
    {1, "Sféra Zrcadla", "Sféry Zrcadla", "Sféru Zrcadla", "Sférou Zrcadla"},
    {1, "Сфера Зеркала", "Сферы Зеркала", "Сферу Зеркала", "Сферой Зеркала"},
    {0/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"},
    {0/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"/*MISSING*/ , "Orb of the Mirror"},
    }},
  {0x1a404e9f, 0, { // ""Vine Spirit""
    {0, "Winny Duszek", "Winne Duszki", "Winnego Duszka", "Winnym Duszkiem"},
    {0, "Şarap Özü", "Şarap Özleri", "Şarap Özünü", "Şarap Özüyle"},
    {0, "Révový duch", "Révoví duchové", "Révového ducha", "Révovým duchem"},
    {0, "Винный дух", "Винные духи", "Винного духа", "Винным духом"},
    {1, "Rebenseele", "Rebenseelen", "Rebenseele", "Rebenseele"},
    {0/*MISSING*/ , "Vine Spirit"/*MISSING*/ , "Vine Spirit"/*MISSING*/ , "Vine Spirit"/*MISSING*/ , "Vine Spirit"},
    }},
  {0x1c2a0990, 0, { // ""Ivy D""
    {0, "Bluszcz D", "Bluszcze D", "Bluszcz D", "Bluszczem D"},
    {0, "Sarmaşık D", "sarmaşıklar D", "Sarmaşığı D", "Sarmaşıkla D"},
    {3, "Břečťan D", "Břečťany D", "Břečťan D", "Břečťanem D"},
    {3, "Плющ D", "Плющи D", "Плющ D", "Плющом D"},
    {0, "Efeu D", "Efeue D", "Efeu D", "Efeu D"},
    {1, "Hera D", "Heras D"/*MISSING*/ , "Ivy D", "Hera d"},
    }},
  {0x1c2a099a, 0, { // ""Ivy N""
    {0, "Bluszcz N", "Bluszcze N", "Bluszcz N", "Bluszczem N"},
    {0, "sarmaşık N", "Sarmaşıklar N", "Sarmaşığı N", "Sarmaşıkla N"},
    {3, "Břečťan N", "Břečťany N", "Břečťan N", "Břečťanem N"},
    {3, "Плющ N", "Плющи N", "Плющ N", "Плющом N"},
    {0, "Efeu N", "Efeue N", "Efeu N", "Efeu N"},
    {1, "Hera N", "Heras N"/*MISSING*/ , "Ivy N", "Hera N"},
    }},
  {0x1d4ed559, 0, { // ""Pirate""
    {0, "Pirat", "Piraci", "Pirata", "Piratem"},
    {0, "Korsan", "Korsanlar", "Korsanı", "Korsanla"},
    {0, "Pirát", "Piráti", "Piráta", "Pirátem"},
    {0, "Пират", "Пираты", "Пирата", "Пиратом"},
    {0, "Pirat", "Piraten", "Pirat", "Pirat"},
    {0/*MISSING*/ , "Pirate"/*MISSING*/ , "Pirate"/*MISSING*/ , "Pirate"/*MISSING*/ , "Pirate"},
    }},
  {0x1ddb40e6, 0, { // ""column""
    {1, "kolumna", "kolumny", "kolumnę", "kolumną"},
    {1, "Sütun", "Sütunlar", "Sütunu", "Sütunla"},
    {3, "sloup", "sloupy", "sloup", "sloupem"},
    {1, "колонна", "колонны", "колонну", "колонной"},
    {1, "Säule", "Säulen", "Säule", "Säule"},
    {1, "coluna", "colunas"/*MISSING*/ , "column", "coluna"},
    }},
  {0x1e704a9a, 0, { // ""Red Troll""
    {0, "Czerwony Troll", "Czerwone Trolle", "Czerwonego Trolla", "Czerwonym Trollem"},
    {0, "Kızıl Trol", "Kızıl Troller", "Kızıl Trolü", "Kızıl Trolle"},
    {0, "Červený troll", "Červení trollové", "Červeného trolla", "Červeným trollem"},
    {0, "Красный тролль", "Красные тролли", "Красного тролля", "Красным троллем"},
    {0, "Roter Troll", "Rote Trolle", "Roten Troll", "Roten Troll"},
    {0/*MISSING*/ , "Red Troll"/*MISSING*/ , "Red Troll"/*MISSING*/ , "Red Troll"/*MISSING*/ , "Red Troll"},
    }},
  {0x1eadd9b1, 0, { // ""Land of Power""
    {1, "Kraina Mocy", "Kraina Mocy", "Krainą Mocy", "w Krainie Mocy"},
    {1, "Güç Diyarı", "Güç Diyarları", "Güç Diyarını", "Güç Diyarında"},
    {3, "Kraj Moci", "Kraje Moci", "Kraj Moci", "v Kraji Moci"},
    {1, "Земля силы", "Земли силы", "Землю силы", "в Земле силы"},
    {2, "Land der Macht", "Länder der Macht", "im Land der Macht", "im Land der Macht"},
    {0/*MISSING*/ , "Land of Power"/*MISSING*/ , "Land of Power"/*MISSING*/ , "Land of Power"/*MISSING*/ , "Land of Power"},
    }},
  {0x1eeb8385, 0, { // ""Amethyst""
    {3, "Ametyst", "Ametysty", "Ametyst", "Ametystem"},
    {0/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"},
    {3, "Ametyst", "Ametysty", "Ametyst", "Ametystem"},
    {3, "Аметист", "Аметисты", "Аметист", "Аметистом"},
    {0/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"},
    {0/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"/*MISSING*/ , "Amethyst"},
    }},
  {0x214d9df6, 0, { // ""Ivy Branch""
    {1, "Gałąź Bluszczu", "Gałęzie Bluszczu", "Gałąź Bluszczu", "Gałęzią Bluszczu"},
    {1, "Sarmaşık Dalı", "Sarmaşık Dalları", "Sarmaşık Dalını", "Sarmaşık Dalıyla"},
    {1, "Větev břečťanu", "Větve břečťanu", "Větev břečťanu", "Větví břečťanu"},
    {1, "Ветвь плюща", "Ветви плюща", "Ветвь плюща", "Ветвью плюща"},
    {0, "Efeuzweig", "Efeuzweige", "Efeuzweig", "Efeuzweig"},
    {0, "Galho de Hera", "Galhos de Hera"/*MISSING*/ , "Ivy Branch", "Galho de Hera"},
    }},
  {0x22d86d83, 0, { // ""Tame Bomberbird""
    {3, "Oswojony Bombardier", "Oswojone Bombardiery", "Oswojonego Bombardiera", "Oswojonym Bombardierem"},
    {3, "Sakin Bombacıkuş", "Sakin Bombacıkuşlar", "Sakin Bombacıkuşunu", "Sakin Bombacıkuşuyla"},
    {0, "Ochočený Bombarďák", "Ochočení Bombarďáci", "Ochočeného Bombarďáka", "Ochočeným Bombarďákem"},
    {0, "Ручной бомбардир", "Ручные бомбардиры", "Ручного бомбардира", "Ручным бомбардиром"},
    {0, "Zahmer Bombenvogel", "Zahme Bombenvögel", "Zahmen Bombenvogel", "Zahmen Bombenvogel"},
    {0/*MISSING*/ , "Tame Bomberbird"/*MISSING*/ , "Tame Bomberbird"/*MISSING*/ , "Tame Bomberbird"/*MISSING*/ , "Tame Bomberbird"},
    }},
  {0x22e16aae, 0, { // ""Pirate Treasure""
    {0, "Skarb Piratów", "Skarby Piratów", "Skarb Piratów", "Skarbem Piratów"},
    {0, "Korsan Hazinesi", "Korsan Hazineleri", "Korsan Hazinesini", "Korsan Hazinesiyle"},
    {0, "Pirátský poklad", "Pirátské poklady", "Pirátský poklad", "Pirátským pokladem"},
    {2, "Сокровище пиратов", "Сокровища пиратов", "Сокровище пиратов", "Сокровищем пиратов"},
    {0, "Piratenschatz", "Piratenschätze", "Piratenschatz", "Piratenschatz"},
    {0/*MISSING*/ , "Pirate Treasure"/*MISSING*/ , "Pirate Treasure"/*MISSING*/ , "Pirate Treasure"/*MISSING*/ , "Pirate Treasure"},
    }},
  {0x239a26a6, 0, { // ""Land of Storms""
    {1, "Kraina Burzy", "Krainy Burzy", "Krainę Burzy", "w Krainie Burzy"},
    {1, "Fırtına Diyarı", "Fırtına Diyarları", "Fırtına Diyarını", "Fırtına Diyarında"},
    {1, "Kraj Bouří", "Kraje Bouří", "Kraj Bouří", "v Kraji Bouří"},
    {1, "Земля Бури", "Земли Бури", "Землю Бури", "в Земле Бури"},
    {2, "Land der Stürme", "Länder der Stürme", "im Land der Stürme", "im Land der Stürme"},
    {0/*MISSING*/ , "Land of Storms"/*MISSING*/ , "Land of Storms"/*MISSING*/ , "Land of Storms"/*MISSING*/ , "Land of Storms"},
    }},
  {0x25e264cf, 0, { // ""Orb of Water""
    {1, "Sfera Wody", "Sfery Wody", "Sferę Wody", "Sferą Wody"},
    {1, "Su Küresi", "Su Küreleri", "Su Küresini", "Su Küresiyle"},
    {1, "Sféra vody", "Sféry vody", "Sféru vody", "Sférou vody"},
    {1, "Сфера Воды", "Сферы Воды", "Сферу Воды", "Сферой Воды"},
    {0, "Orb des Wassers", "Orbs des Wassers", "Orb des Wassers", "Orb des Wassers"},
    {0/*MISSING*/ , "Orb of Water"/*MISSING*/ , "Orb of Water"/*MISSING*/ , "Orb of Water"/*MISSING*/ , "Orb of Water"},
    }},
  {0x26f959f7, 0, { // ""Alchemist Lab""
    {2, "Laboratorium", "Laboratorium", "Laboratorium", "w Laboratorium"},
    {2, "Laboratuar", "Laboratuarlar", "Laboratuarı", "laboratuarda"},
    {1, "Alchymistická laboratoř", "Alchymistické laboratoře", "Alchymistickou laboratoř", "v Alchymistické laboratoři"},
    {1, "Лаборатория", "Лаборатории", "Лабораторию", "в Лаборатории"},
    {0, "Alchemielabor", "Alchemielabore", "im Alchemielabor", "im Alchemielabor"},
    {0, "Laboratório do Alquimista", "Laboratórios dos Alquimistas"/*MISSING*/ , "Alchemist Lab", "Laboratório do Alquimista"},
    }},
  {0x27373108, 0, { // ""Orb of Energy""
    {1, "Sfera Energii", "Sfery Energii", "Sferę Energii", "Sferą Energii"},
    {0/*MISSING*/ , "Orb of Energy"/*MISSING*/ , "Orb of Energy"/*MISSING*/ , "Orb of Energy"/*MISSING*/ , "Orb of Energy"},
    {1, "Sféra Energie", "Sféry Energie", "Sféru Energie", "Sférou Energie"},
    {1, "Сфера Energii", "Сферы Energii", "Сферу Energii", "Сферой Energii"},
    {0, "Orb der Energie", "Orbs der Energie", "Orb der Energie", "Orb der Energie"},
    {0/*MISSING*/ , "Orb of Energy"/*MISSING*/ , "Orb of Energy"/*MISSING*/ , "Orb of Energy"/*MISSING*/ , "Orb of Energy"},
    }},
  {0x27627e19, 0, { // ""stone gargoyle floor""
    {1, "podłoga z rzygacza", "podłogi z rzygacza", "podłogę z rzygacza", "podłogą z rzygacza"},
    {1, "taş çörten zemini", "taş çörten zeminleri", "taş çörten zeminini", "taş çörten zeminiyle"},
    {1, "podlaha z chrliče", "podlahy z chrličů", "podlahu z chrliče", "podlahou z chrliče"},
    {3, "пол с горгульей", "полы с горгульями", "пол с горгульей", "полом с горгульей"},
    {0, "Steingargoyle-Boden", "Steingargoyle-Böden", "Steingargoyle-Boden", "Steingargoyle-Boden"},
    {0/*MISSING*/ , "stone gargoyle floor"/*MISSING*/ , "stone gargoyle floor"/*MISSING*/ , "stone gargoyle floor"/*MISSING*/ , "stone gargoyle floor"},
    }},
  {0x295eca8f, 0, { // ""Orb of the Fish""
    {1, "Sfera Ryby", "Sfery Ryby", "Sferę Ryby", "Sferą Ryby"},
    {1, "Balık Küresi", "Balık Küreleri", "Balık Küresini", "Balık Küresiyle"},
    {1, "Sféra Ryby", "Sféry Ryby", "Sféru Ryby", "Sférou Ryby"},
    {1, "Сфера Рыбы", "Сферы Рыбы", "Сферу Рыбы", "Сферой Рыбы"},
    {0, "Orb des Fischs", "Orbs des Fischs", "Orb des Fischs", "Orb des Fischs"},
    {0/*MISSING*/ , "Orb of the Fish"/*MISSING*/ , "Orb of the Fish"/*MISSING*/ , "Orb of the Fish"/*MISSING*/ , "Orb of the Fish"},
    }},
  {0x2aa339b0, 0, { // ""Evil Golem""
    {0, "Zły Golem", "Złe Golemy", "Złego Golema", "Złym Golemem"},
    {0, "Şeytani Golem", "Şeytani Golemler", "Şeytani Golemi", "Şeytani Golemle"},
    {0, "Zlý golem", "Zlí golemové", "Zlého golema", "Zlým golemem"},
    {0, "Злой голем", "Злые големы", "Злого голема", "Злым големом"},
    {0, "Böser Golem", "Böse Golems", "Bösen Golem", "Bösen Golem"},
    {0/*MISSING*/ , "Evil Golem"/*MISSING*/ , "Evil Golem"/*MISSING*/ , "Evil Golem"/*MISSING*/ , "Evil Golem"},
    }},
  {0x2bab10d5, 0, { // ""Knife""
    {3, "Nóż", "Noże", "Nóż", "Nożem"},
    {0, "Bıçak", "Bıçaklar", "Bıçağı", "Bıçakla"},
    {3, "Nůž", "Nože", "Nůž", "Nožem"},
    {3, "Нож", "Ножи", "Нож", "Ножом"},
    {2, "Messer", "Messer", "Messer", "Messer"},
    {0/*MISSING*/ , "Knife"/*MISSING*/ , "Knife"/*MISSING*/ , "Knife"/*MISSING*/ , "Knife"},
    }},
  {0x2be7948a, 0, { // ""no item""
    {3, "brak przedmiotu", "braki przedmiotu", "braku przedmiotu", "brakiem przedmiotu"},
    {0/*MISSING*/ , "no item"/*MISSING*/ , "no item"/*MISSING*/ , "no item"/*MISSING*/ , "no item"},
    {3, "žádný předmět", "žádné předměty", "žádný předmět", "žádným předmětem"},
    {3, "нет предмета", "нет предмета", "нет предмета", "нет предмета"},
    {0, "kein Gegenstand", "keine Gegenstände", "keinen Gegenstand", "keinen Gegenstand"},
    {0/*MISSING*/ , "no item"/*MISSING*/ , "no item"/*MISSING*/ , "no item"/*MISSING*/ , "no item"},
    }},
  {0x2c05e0f6, 0, { // ""Sand Worm W""
    {0, "Pustynny Czerw W", "Pustynne Czerwie W", "Pustynnego Czerwia W", "Pustynnym Czerwiem W"},
    {0, "Kumkurdu W", "Kumkurtları W", "Kumkurdunu W", "Kumkurduyla W"},
    {0, "Písečný červ W", "Píseční červi W", "Písečného červa W", "Písečným červem W"},
    {0, "Пустынный червь W", "Пустынные черви W", "Пустынного червя W", "Пустынным червём W"},
    {0, "Sandwurm W", "Sandwürmer W", "Sandwurm W", "Sandwurm W"},
    {1, "Minhoca de Areia W", "Minhocas de Areia W"/*MISSING*/ , "Sand Worm W", "Minhoca de Areia W"},
    }},
  {0x2c125756, 0, { // ""Fire Cultist""
    {0, "Ognisty Kultysta", "Ogniści Kultyści", "Ognistego Kultystę", "Ognistym Kultystą"},
    {0, "Ateş Müridi", "Ateş Müritleri", "Ateş Müridini", "Ateş Müritle"},
    {0, "Kultista ohně", "Kultisté ohně", "Kultistu ohně", "Kultistou ohně"},
    {0, "Огненный культист", "Огненные культисты", "Огненного культиста", "Огненным культистом"},
    {0, "Feuerkultist", "Feuerkultisten", "Feuerkultisten", "Feuerkultisten"},
    {0, "Cultista do Fogo", "Cultistas do Fogo"/*MISSING*/ , "Fire Cultist", "Cultista do Fogo"},
    }},
  {0x2c23322d, 0, { // ""Royal Jelly""
    {2, "Królewskie Mleczko", "Królewskie Mleczka", "Królewskie Mleczko", "Królewskim Mleczkiem"},
    {2, "Müthiş Jöle", "Müthiş Jöleler", "Müthiş Jöleyi", "Müthiş Jöleyle"},
    {1, "Mateří kašička", "Mateří kašičky", "Mateří kašičku", "Mateří kašičkou"},
    {2, "маточное молочко", "маточное молочко", "маточное молочко", "маточным молочком"},
    {2, "Gelée Royale", "Gelée Royale", "Gelée Royale", "Gelée Royale"},
    {0/*MISSING*/ , "Royal Jelly"/*MISSING*/ , "Royal Jelly"/*MISSING*/ , "Royal Jelly"/*MISSING*/ , "Royal Jelly"},
    }},
  {0x2ff35477, 0, { // ""Necromancer""
    {0, "Nekromanta", "Nekromanci", "Nekromantę", "Nekromantą"},
    {0, "Ölüçağıran", "Ölüçağıranlar", "Ölüçağıranı", "Ölüçağıranla"},
    {0, "Nekromancer", "Nekromanceři", "Nekromancera", "Nekromancerem"},
    {0, "Некромант", "Некроманты", "Некроманта", "Некромантом"},
    {0, "Totenbeschwörer", "Totenbeschwörer", "Totenbeschwörer", "Totenbeschwörer"},
    {0, "Necromante", "Necromantes"/*MISSING*/ , "Necromancer", "Necromante"},
    }},
  {0x3063d2f6, 0, { // ""Orb of Domination""
    {1, "Sfera Dominacji", "Sfery Dominacji", "Sferę Dominacji", "Sferą Dominacji"},
    {0/*MISSING*/ , "Orb of Domination"/*MISSING*/ , "Orb of Domination"/*MISSING*/ , "Orb of Domination"/*MISSING*/ , "Orb of Domination"},
    {1, "Sféra Dominance", "Sféry Dominance", "Sféru Dominance", "Sférou Dominance"},
    {1, "Сфера Доминации", "Сферы Доминации", "Сферу Доминации", "Сферой Доминации"},
    {0, "Orb der Dominanz", "Orbs der Dominanz", "Orb der Dominanz", "Orb der Dominanz"},
    {0/*MISSING*/ , "Orb of Domination"/*MISSING*/ , "Orb of Domination"/*MISSING*/ , "Orb of Domination"/*MISSING*/ , "Orb of Domination"},
    }},
  {0x309f6294, 0, { // ""grounded wall""
    {1, "uziemiona ściana", "uziemione ściany", "uziemioną scianę", "uziemioną ścianą"},
    {1, "topraklı duvar", "topraklı duvarlar", "topraklı duvarı", "topraklı duvarla"},
    {1, "uzemněná zeď", "uzemněné zdi", "uzemněnou zeď", "uzemněnou zdí"},
    {1, "заземлённая стена", "заземлённые стены", "заземлённую стену", "заземлённой стеной"},
    {1, "geerdete Wand", "geerdete Wände", "geerdete Wand", "geerdete Wand"},
    {0/*MISSING*/ , "grounded wall"/*MISSING*/ , "grounded wall"/*MISSING*/ , "grounded wall"/*MISSING*/ , "grounded wall"},
    }},
  {0x30ade9f9, 0, { // ""Prince""
    {0, "Książę", "Książęta", "Księcia", "Księciem"},
    {0, "Şehzade", "Şehzadeyi", "Şehzadeler", "Şehzadeyle"},
    {0, "Princ", "Princové", "Prince", "Princem"},
    {0, "Принц", "Принцы", "Принца", "Принцем"},
    {0, "Prinz", "Prinzen", "Prinz", "Prinz"},
    {0/*MISSING*/ , "Prince"/*MISSING*/ , "Prince"/*MISSING*/ , "Prince"/*MISSING*/ , "Prince"},
    }},
  {0x31586dfd, 0, { // ""Desert""
    {1, "Pustynia", "Pustynia", "Pustynię", "na Pustyni"},
    {1, "Çöl", "Çöller", "Çölü", "Çölde"},
    {1, "Poušť", "Pouště", "Poušť", "na Poušti"},
    {1, "Пустыня", "Пустыни", "Пустыню", "в Пустыне"},
    {1, "Wüste", "Wüsten", "in der Wüste", "in der Wüste"},
    {0, "Deserto", "Desertos"/*MISSING*/ , "Desert", "Deserto"},
    }},
  {0x3181b219, 0, { // ""Earth Elemental""
    {0, "Żywiołak", "Żywiołaki", "Żywiołaka", "Żywiołakiem"},
    {0, "Toprakadam", "Toprakadamlar", "Toprakadamı", "Toprakadamla"},
    {0, "Elementál Země", "Elementálové Země", "Elementála Země", "Elementálem Země"},
    {0, "Земной элементаль", "Земные элементали", "Земного элементаля", "Земным элементалем"},
    {0, "Erdelementar", "Erdelementare", "Erdelementar", "Erdelementar"},
    {0/*MISSING*/ , "Earth Elemental"/*MISSING*/ , "Earth Elemental"/*MISSING*/ , "Earth Elemental"/*MISSING*/ , "Earth Elemental"},
    }},
  {0x318d9d8d, 0, { // ""Land of Eternal Motion""
    {2, "Kraina Wiecznego Ruchu", "Kraina Wiecznego Ruchu", "Krainę Wiecznego Ruchu", "w Krainie Wiecznego Ruchu"},
    {2, "Sonsuz Devinim Diyarı", "Sonsuz Devinim Diyarları", "Sonsuz Devinim Diyarını", "Sonsuz Devinim Diyarında"},
    {3, "Kraj věčného pohybu", "Kraje věčného pohybu", "Kraj věčného pohybu", "v Kraji věčného pohybu"},
    {1, "Земля Вечного Движения", "Земли Вечного Движения", "Землю Вечного Движения", "в Земле Вечного Движения"},
    {2, "Land der ewigen Bewegung", "Länder der ewigen Bewegung", "im Land der ewigen Bewegung", "im Land der ewigen Bewegung"},
    {1, "Terra do Movimento Eterno", "Terras do Movimento Eterno"/*MISSING*/ , "Land of Eternal Motion", "Terra do Movimento Eterno"},
    }},
  {0x331fcbb5, 0, { // ""stone gargoyle""
    {0, "kamienny rzygacz", "kamienne rzygacze", "kamiennego rzygacza", "kamiennym rzygaczem"},
    {0, "taş çörten", "taş çörtenler", "taş çörteni", "taş çörtenle"},
    {3, "kamenný chrlič", "kamenné chrliče", "kamenného chrliče", "kamenným chrličem"},
    {1, "каменная горгулья", "каменные горгульи", "каменную горгулью", "каменной горгульей"},
    {0, "Steingargoyle", "Steingargoyles", "Steingargoyle", "Steingargoyle"},
    {0/*MISSING*/ , "stone gargoyle"/*MISSING*/ , "stone gargoyle"/*MISSING*/ , "stone gargoyle"/*MISSING*/ , "stone gargoyle"},
    }},
  {0x33878aca, 0, { // ""Ancient Jewelry""
    {1, "Starożytna Biżuteria", "Starożytna Biżuteria", "Starożytną Biżuterię", "Starożytną Biżuterią"},
    {0/*MISSING*/ , "Ancient Jewelry"/*MISSING*/ , "Ancient Jewelry"/*MISSING*/ , "Ancient Jewelry"/*MISSING*/ , "Ancient Jewelry"},
    {3, "Dávný šperk", "Dávné šperky", "Dávný šperk", "Dávným šperkem"},
    {1, "Древняя драгоценность", "Древние драгоценности", "Древнюю драгоценность", "Древней драгоценностью"},
    {1, "Antike Juwelen", "Antike Juwelen", "Antiken Juwelen", "Antiken Juwelen"},
    {0/*MISSING*/ , "Ancient Jewelry"/*MISSING*/ , "Ancient Jewelry"/*MISSING*/ , "Ancient Jewelry"/*MISSING*/ , "Ancient Jewelry"},
    }},
  {0x338f38ab, 0, { // ""Rock Snake Tail""
    {3, "Ogon Węża", "Ogony Węża", "Ogon Węża", "Ogonem Węża"},
    {3, "Kaya Yılanı Kuyruğu", "Kaya Yılanı Kuyrukları", "Kaya Yılan Kuyruğunu", "Kaya Yılan Kuyruğuyla"},
    {3, "Ocas Kamenného hada", "Ocasy Kamenných hadů", "Ocas Kamenného hada", "Ocasem Kamenného hada"},
    {3, "Хвост змеи", "Хвосты змей", "Хвост змеи", "Хвостом змеи"},
    {0, "Felsschlangenschweif", "Felsschlangenschweifs", "Felsschlangenschweif", "Felsschlangenschweif"},
    {0/*MISSING*/ , "Rock Snake Tail"/*MISSING*/ , "Rock Snake Tail"/*MISSING*/ , "Rock Snake Tail"/*MISSING*/ , "Rock Snake Tail"},
    }},
  {0x34f98b1f, 0, { // ""Bonfire""
    {2, "Ognisko", "Ogniska", "Ognisko", "Ogniskiem"},
    {2, "Kamp Ateşi", "Kamp Ateşleri", "Kamp Ateşini", "Kamp Ateşiyle"},
    {2, "Ohniště", "Ohniště", "Ohniště", "Ohništěm"},
    {3, "Костёр", "Костры", "Костёр", "Костром"},
    {2, "Leuchtfeuer", "Leuchtfeuer", "Leuchtfeuer", "Leuchtfeuer"},
    {1, "Fogueira", "Fogueiras"/*MISSING*/ , "Bonfire", "Fogueira"},
    }},
  {0x356918f6, 0, { // ""Living Fjord""
    {3, "Żywy Fiord", "Żywe Fiordy", "Żywy Fiord", "w Żywym Fiordzie"},
    {3, "Yaşayan Fiyort", "Yaşayan Fiyortlar", "Yaşayan Fiyortu", "Yaşayan Fiyortta"},
    {3, "Živoucí fjord", "Živoucí fjordy", "Živoucí Fjord", "v Živoucím Fjordu"},
    {3, "Живой фьорд", "Живые фьорды", "Живой фьорд", "в Живом фьорде"},
    {0, "Lebender Fjord", "Lebende Fjorde", "Lebender Fjord", "Lebender Fjord"},
    {0/*MISSING*/ , "Living Fjord"/*MISSING*/ , "Living Fjord"/*MISSING*/ , "Living Fjord"/*MISSING*/ , "Living Fjord"},
    }},
  {0x36e98812, 0, { // ""palace wall""
    {1, "mur Pałacu", "mury Pałacu", "mur Pałacu", "murem Pałacu"},
    {1, "Saray Duvarı", "Saray Duvarları", "Saray Duvarını", "Saray Duvarı"},
    {1, "zeď paláce", "zdi paláce", "zeď paláce", "zeď paláce"},
    {1, "стена дворца", "стены дворца", "стену дворца", "стеной дворца"},
    {1, "Palastwand", "Palastwände", "Palastwand", "Palastwand"},
    {0/*MISSING*/ , "palace wall"/*MISSING*/ , "palace wall"/*MISSING*/ , "palace wall"/*MISSING*/ , "palace wall"},
    }},
  {0x3807f344, 0, { // ""Fire Shard""
    {3, "Okruch Ognia", "Okruchy Ognia", "Okruch Ognia", "Okruchem Ognia"},
    {0, "Ateş Parçası", "Ateş Parçaları", "Ateş Parçasını", "Ateş Parçasıyla"},
    {3, "Úlomek Ohně", "Úlomky Ohně", "Úlomek Ohně", "Úlomkem Ohně"},
    {3, "Осколок Огня", "Осколки Огня", "Осколок Огня", "Осколком Огня"},
    {1, "Feuerscherbe", "Feuerscherben", "Feuerscherbe", "Feuerscherbe"},
    {0/*MISSING*/ , "Fire Shard"/*MISSING*/ , "Fire Shard"/*MISSING*/ , "Fire Shard"/*MISSING*/ , "Fire Shard"},
    }},
  {0x38384004, 0, { // ""Giant Fox""
    {3, "Wielki Lis", "Wielkie Lisy", "Wielkiego Lisa", "Wielkim Lisem"},
    {0/*MISSING*/ , "Giant Fox"/*MISSING*/ , "Giant Fox"/*MISSING*/ , "Giant Fox"/*MISSING*/ , "Giant Fox"},
    {1, "Obří liška", "Obří lišky", "Obří lišku", "Obří liškou"},
    {0, "Огромный лис", "Огромные лисы", "Огромного лиса", "Огромным лисом"},
    {0, "Riesenfuchs", "Riesenfüchse", "Riesenfuchs", "Riesenfuchs"},
    {0/*MISSING*/ , "Giant Fox"/*MISSING*/ , "Giant Fox"/*MISSING*/ , "Giant Fox"/*MISSING*/ , "Giant Fox"},
    }},
  {0x3ab53b1b, 0, { // ""Orb of Undeath""
    {1, "Sfera Nieumarłych", "Sfery Nieumarłych", "Sferę Nieumarłych", "Sferą Nieumarłych"},
    {0/*MISSING*/ , "Orb of Undeath"/*MISSING*/ , "Orb of Undeath"/*MISSING*/ , "Orb of Undeath"/*MISSING*/ , "Orb of Undeath"},
    {1, "Sféra Nemrtvých", "Sféry Nemrtvých", "Sféru Nemrtvých", "Sférou Nemrtvých"},
    {1, "Сфера Нежити", "Сферы Нежити", "Сферу Нежити", "Сферой Нежити"},
    {0, "Orb der Untoten", "Orbs der Untoten", "Orb der Untoten", "Orb der Untoten"},
    {0/*MISSING*/ , "Orb of Undeath"/*MISSING*/ , "Orb of Undeath"/*MISSING*/ , "Orb of Undeath"/*MISSING*/ , "Orb of Undeath"},
    }},
  {0x3b4c9ee0, 0, { // ""weird rock""
    {2, "dziwna skała", "dziwne skały", "dziwną skałę", "dziwną skałą"},
    {2, "garip kaya", "garip kayalar", "garip kayayı", "garip kayayla"},
    {1, "podivná skála", "podivné skály", "podivnou skálu", "podivnou skálou"},
    {3, "странный камень", "странные камни", "странный камень", "странным камнем"},
    {0, "Seltsamer Stein", "Seltsame Steine", "Seltsamen Stein", "Seltsamen Stein"},
    {0/*MISSING*/ , "weird rock"/*MISSING*/ , "weird rock"/*MISSING*/ , "weird rock"/*MISSING*/ , "weird rock"},
    }},
  {0x3bb978a5, 0, { // ""moat of Camelot""
    {1, "fosa Camelot", "fosy Camelot", "fosę Camelot", "fosą Camelot"},
    {1, "Kamelot hendeği", "Kamelot hendekleri", "Kamelot hendeğini", "Kamelot hendeğiyle"},
    {3, "vodní příkop", "vodní příkopy", "vodní příkop", "vodním příkopem"},
    {3, "ров Камелота", "рвы Камелота", "ров Камелота", "рвом Камелота"},
    {0, "Graben von Camelot", "Gräben von Camelot", "Graben von Camelot", "Graben von Camelot"},
    {0/*MISSING*/ , "moat of Camelot"/*MISSING*/ , "moat of Camelot"/*MISSING*/ , "moat of Camelot"/*MISSING*/ , "moat of Camelot"},
    }},
  {0x3be3878f, 0, { // ""Thornless Rose""
    {1, "Róża bez Kolców", "Róże bez Kolców", "Różę bez Kolców", "Różą bez Kolców"},
    {0/*MISSING*/ , "Thornless Rose"/*MISSING*/ , "Thornless Rose"/*MISSING*/ , "Thornless Rose"/*MISSING*/ , "Thornless Rose"},
    {1, "Růže bez trnů", "Růže bez trnů", "Růži bez trnů", "Růží bez trnů"},
    {1, "роза без шипов", "розы без шипов", "розу без шипов", "розой без шипов"},
    {1, "Rose ohne Dornen", "Rosen ohne Dornen", "Rose ohne Dornen", "Rose ohne Dornen"},
    {0/*MISSING*/ , "Thornless Rose"/*MISSING*/ , "Thornless Rose"/*MISSING*/ , "Thornless Rose"/*MISSING*/ , "Thornless Rose"},
    }},
  {0x3d400c71, 0, { // ""Thumper""
    {3, "Dudnik", "Dudniki", "Dudnik", "Dudnikiem"},
    {3, "Gümleyen", "Gümleyenler", "Gümleyeni", "Gümleyenle"},
    {3, "Tlouk", "Tlouky", "Tlouk", "Tloukem"},
    {3, "Тампер", "Тамперы", "Тампер", "Тампером"},
    {0, "Stampfer", "Stampfer", "Stampfer", "Stampfer"},
    {0, "Batedor", "Batedores"/*MISSING*/ , "Thumper", "Batedor"},
    }},
  {0x3dd9fb5f, 0, { // ""barrow""
    {3, "kurhan", "kurhany", "kurhan", "kurhanem"},
    {0/*MISSING*/ , "barrow"/*MISSING*/ , "barrow"/*MISSING*/ , "barrow"/*MISSING*/ , "barrow"},
    {1, "mohyla", "mohyly", "mohylu", "mohylou"},
    {3, "курган", "курганы", "курган", "курганом"},
    {3, "Hügelgrab", "Hügelgräber", "Hügelgrab", "Hügelgrab"},
    {0/*MISSING*/ , "barrow"/*MISSING*/ , "barrow"/*MISSING*/ , "barrow"/*MISSING*/ , "barrow"},
    }},
  {0x3ec5a2f2, 0, { // ""Orb of Safety""
    {1, "Sfera Bezpieczeństwa", "Sfery Bezpieczeństwa", "Sferę Bezpieczeństwa", "Sferą Bezpieczeństwa"},
    {1, "Güvenlik Küresi", "Güvenlik Küreleri", "Güvenlik Küresini", "Güvenlik Küresiyle"},
    {1, "Sféra Bezpečnosti", "Sféry Bezpečnosti", "Sféru Bezpečnosti", "Sférou Bezpečnosti"},
    {1, "Сфера Безопасности", "Сферы Безопасности", "Сферу Безопасности", "Сферой Безопасности"},
    {0, "Orb der Geborgenheit", "Orbs der Geborgenheit", "Orb der Geborgenheit", "Orb der Geborgenheit"},
    {0, "Orbe de Segurança", "Orbes de Segurança"/*MISSING*/ , "Orb of Safety", "Orbe de Segurança"},
    }},
  {0x3ee3571b, 0, { // ""Dead Orb""
    {1, "Martwa Sfera", "Martwe Sfery", "Martwą Sferę", "Martwą Sferą"},
    {1, "Ölü Küre", "Ölü Küreler", "Ölü Küreyle", "Ölü Küreyi"},
    {1, "Mrtvá sféra", "Mrtvé sféry", "Mrtvou sféru", "Mrtvou sférou"},
    {1, "Мёртвая сфера", "Мёртвые сферы", "Мёртвую сферу", "Мёртвой сферой"},
    {0, "toter Orb", "toten Orbs", "toten Orb", "toten Orb"},
    {0, "Orbe Morto", "Orbes Mortos"/*MISSING*/ , "Dead Orb", "Orbe Morto"},
    }},
  {0x3ee8eeb4, 0, { // ""Orb of Discord""
    {1, "Sfera Niezgody", "Sfery Niezgody", "Sferę Niezgody", "Sferą Niezgody"},
    {1, "İhtilaf Küresi", "İhtilaf Küreleri", "İhtilaf Küresini", "İhtilaf Küresiyle"},
    {1, "Sféra Nesváru", "Sféry Nesváru", "Sféru Nesváru", "Sférou Nesváru"},
    {1, "Сфера Раздора", "Сферы Раздора", "Сферу Раздора", "Сферой Раздора"},
    {0, "Orb des Zwiespalts", "Orbs des Zwiespalts", "Orb des Zwiespalts", "Orb des Zwiespalts"},
    {0/*MISSING*/ , "Orb of Discord"/*MISSING*/ , "Orb of Discord"/*MISSING*/ , "Orb of Discord"/*MISSING*/ , "Orb of Discord"},
    }},
  {0x4065e362, 0, { // ""limestone wall""
    {1, "wapienna ściana", "wapienne ściany", "wapienną ścianę", "wapienną ścianą"},
    {1, "kireçtaşı duvar", "kireçtaşı duvarlar", "kireçtaşı duvarı", "kireçtaşı duvarla"},
    {1, "vápencová stěna", "vápencové stěny", "vápencovou stěnu", "vápencovou stěnou"},
    {1, "известняковая стена", "известняковые стены", "известняковую стену", "известняковой стеной"},
    {1, "Kalkstein Wand", "Kalkstein Wände", "Kalkstein Wand", "Kalkstein Wand"},
    {0/*MISSING*/ , "limestone wall"/*MISSING*/ , "limestone wall"/*MISSING*/ , "limestone wall"/*MISSING*/ , "limestone wall"},
    }},
  {0x40e4a983, 0, { // ""Fulgurite""
    {3, "Fulguryt", "Fulguryty", "Fulguryt", "Fulgurytem"},
    {0/*MISSING*/ , "Fulgurite"/*MISSING*/ , "Fulgurite"/*MISSING*/ , "Fulgurite"/*MISSING*/ , "Fulgurite"},
    {3, "Fulgurit", "Fulgurity", "Fulgurit", "Fulguritem"},
    {3, "Фульгурит", "Фульгуриты", "Фульгурит", "Фульгуритом"},
    {0, "Fulgurit", "Fulgurite", "Fulgurit", "Fulgurit"},
    {0/*MISSING*/ , "Fulgurite"/*MISSING*/ , "Fulgurite"/*MISSING*/ , "Fulgurite"/*MISSING*/ , "Fulgurite"},
    }},
  {0x418c6203, 0, { // ""Baby Tortoise""
    {0, "Żółwik", "Żółwiki", "Żółwika", "Żółwikiem"},
    {0/*MISSING*/ , "Baby Tortoise"/*MISSING*/ , "Baby Tortoise"/*MISSING*/ , "Baby Tortoise"/*MISSING*/ , "Baby Tortoise"},
    {1, "Želvička", "Želvičky", "Želvičku", "Želvičkou"},
    {1, "Черепашка", "Черепашки", "Черепашку", "Черепашкой"},
    {0, "Babyschildkröte", "Babyschildkröten", "Babyschildkröte", "Babyschildkröte"},
    {0/*MISSING*/ , "Baby Tortoise"/*MISSING*/ , "Baby Tortoise"/*MISSING*/ , "Baby Tortoise"/*MISSING*/ , "Baby Tortoise"},
    }},
  {0x4290380f, 0, { // ""minefield""
    {2, "pole minowe", "pola minowe", "pole minowe", "polem minowym"},
    {2, "mayıntarlası", "mayıntarlaları", "mayıntarlasını", "mayıntarlasıyla"},
    {2, "minové pole", "minová pole", "minové pole", "minovým polem"},
    {2, "минное поле", "минные поля", "минное поле", "на минном поле"},
    {1, "Mine?", "Minen?", "Mine?", "Mine?"},
    {0/*MISSING*/ , "minefield"/*MISSING*/ , "minefield"/*MISSING*/ , "minefield"/*MISSING*/ , "minefield"},
    }},
  {0x43dcc29c, 0, { // ""rubble""
    {3, "gruz", "gruzy", "gruz", "gruzem"},
    {3, "gruz", "gruzy", "gruz", "gruzem"},
    {1, "sutiny", "sutiny", "sutiny", "sutinami"},
    {3, "щебень", "щебни", "щебень", "щебнем"},
    {2, "Geröll", "Geröll", "Geröll", "Geröll"},
    {0/*MISSING*/ , "rubble"/*MISSING*/ , "rubble"/*MISSING*/ , "rubble"/*MISSING*/ , "rubble"},
    }},
  {0x4427f2e9, 0, { // ""red slime""
    {1, "czerwona maź", "czerwone mazie", "czerwoną maź", "czerwoną mazią"},
    {1, "Kırmızı Balçık", "Kırmızı Balçıklar", "Kırmızı Balçığı", "Kırmızı Balçıkta"},
    {3, "červený sliz", "červené slizy", "červený sliz", "červeným slizem"},
    {1, "красная слизь", "красная слизь", "красную слизь", "красной слизью"},
    {0, "Roter Schleim", "Roten Schleime", "Roten Schleim", "Roten Schleim"},
    {0, "lodo vermelho", "lodos vermelhos"/*MISSING*/ , "red slime", "lodo vermelho"},
    }},
  {0x4435c8e0, 0, { // ""Orb of Horns""
    {1, "Sfera Rogów", "Sfery Rogów", "Sferę Rogów", "Sferą Rogów"},
    {0/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"},
    {1, "Sféra Rohů", "Sféry Rohů", "Sféru Rohů", "Sférou Rohů"},
    {1, "Сфера Рогов", "Сферы Рогов", "Сферу Рогов", "Сферой Рогов"},
    {0/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"},
    {0/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"/*MISSING*/ , "Orb of Horns"},
    }},
  {0x44c19f73, 2, { // ""Crossroads""
    {2, "Skrzyżowanie", "Skrzyżowanie", "Skrzyżowanie", "na Skrzyżowaniu"},
    {2, "Arayollar", "Arayollar", "Arayolları", "Arayollarda"},
    {1, "Křižovatka", "Křižovatky", "Křižovatku", "na Křižovatce"},
    {3, "Перекрёсток", "Перекрёстки", "Перекрёсток", "на Перекрёстке"},
    {1, "Kreuzungen", "Kreuzungen", "auf den Kreuzungen", "auf den Kreuzungen"},
    {1, "Enruzilhada", "Enruzilhadas"/*MISSING*/ , "Crossroads", "Enruzilhada"},
    }},
  {0x45a84b21, 0, { // ""Gadfly""
    {0, "Giez", "Gzy", "Gza", "Gzem"},
    {0/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"},
    {0, "Ovád", "Ovádi", "Ováda", "Ovádem"},
    {0, "Овод", "Оводы", "Овода", "Оводом"},
    {0/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"},
    {0/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"/*MISSING*/ , "Gadfly"},
    }},
  {0x45f826dd, 0, { // ""Orb of Yendor""
    {1, "Sfera Yendoru", "Sfery Yendoru", "Sferę Yendoru", "Sferą Yendoru"},
    {1, "Yendor Küresi", "Yendor Küreleri", "Yendor Küresini", "Yendor Küresiyle"},
    {1, "Yendorská sféra", "Yendorské sféry", "Yendorskou sféru", "Yendorskou sférou"},
    {1, "Сфера Йендора", "Сферы Йендора", "Сферу Йендора", "Сферой Йендора"},
    {0, "Orb des Yendor", "Orbs des Yendor", "Orb des Yendor", "Orb des Yendor"},
    {0, "Orbe de Yendor", "Orbes de Yendor"/*MISSING*/ , "Orb of Yendor", "Orbe de Yendor"},
    }},
  {0x46dfcc8f, 0, { // ""Orb of the Warp""
    {1, "Sfera Zakrzywienia", "Sfery Zakrzywienia", "Sferę Zakrzywienia", "Sferą Zakrzywienia"},
    {0/*MISSING*/ , "Orb of the Warp"/*MISSING*/ , "Orb of the Warp"/*MISSING*/ , "Orb of the Warp"/*MISSING*/ , "Orb of the Warp"},
    {1, "Sféra Pokřivení", "Sféry Pokřivení", "Sféru Pokřivení", "Sférou Pokřivení"},
    {1, "Сфера Искривления", "Сферы Искривления", "Сферу Искривления", "Сферой Искривления"},
    {0, "Orb der Verzerrung", "Orbs der Verzerrung", "Orb der Verzerrung", "Orb der Verzerrung"},
    {0/*MISSING*/ , "Orb of the Warp"/*MISSING*/ , "Orb of the Warp"/*MISSING*/ , "Orb of the Warp"/*MISSING*/ , "Orb of the Warp"},
    }},
  {0x489faba1, 0, { // ""Princess""
    {1, "Księżniczka", "Księżniczki", "Księżniczkę", "Księżniczką"},
    {1, "Prenses", "Prensesi", "Prensesler", "Prensesle"},
    {1, "Princezna", "Princezny", "Princeznu", "Princeznou"},
    {1, "Принцесса", "Принцессы", "Принцессу", "Принцессой"},
    {1, "Prinzessin", "Prinzessinnen", "Prinzessin", "Prinzessin"},
    {0/*MISSING*/ , "Princess"/*MISSING*/ , "Princess"/*MISSING*/ , "Princess"/*MISSING*/ , "Princess"},
    }},
  {0x48f159ba, 0, { // ""Active Ivy""
    {0, "Aktywny Bluszcz", "Aktywne Bluszcze", "Aktywny Bluszcz", "Aktywnym Bluszczem"},
    {0, "Canlı Sarmaşık", "Canlı Sarmaşıklar", "Canlı Sarmaşığı", "Canlı Sarmaşığıyla"},
    {3, "Aktivní břečťan", "Aktivní břečťany", "Aktivní břečťan", "Aktivním břečťanem"},
    {3, "Активный плющ", "Активные плющи", "Активный плющ", "Активным плющом"},
    {0, "Lebender Efeu", "Lebende Efeue", "Lebenden Efeu", "Lebenden Efeu"},
    {1, "Hera Ativa", "Heras Ativas"/*MISSING*/ , "Active Ivy", "Hera Ativa"},
    }},
  {0x49886610, 0, { // ""dead floor""
    {1, "martwa ziemia", "martwe ziemie", "martwą ziemię", "martwą ziemią"},
    {1, "ölü zemin", "ölü zeminler", "ölü zemini", "ölü zeminde"},
    {1, "mrtvá podlaha", "mrtvé podlahy", "mrtvou podlahu", "mrtvou podlahou"},
    {3, "мёртвый пол", "мёртвые полы", "мёртвый пол", "мёртвым полом"},
    {1, "toter Boden", "tote Böden", "toten Boden", "toten Boden"},
    {0/*MISSING*/ , "dead floor"/*MISSING*/ , "dead floor"/*MISSING*/ , "dead floor"/*MISSING*/ , "dead floor"},
    }},
  {0x4ad32242, 0, { // ""Orb of the Dragon""
    {1, "Sfera Smoka", "Sfery Smoka", "Sferę Smoka", "Sferą Smoka"},
    {1, "Ejderha Küresi", "Ejderha Küreleri", "Ejderha Küresini", "Ejderha Küresiyle"},
    {1, "Sféra Draka", "Sféry Draka", "Sféru Draka", "Sférou Draka"},
    {1, "Сфера Дракона", "Сферы Дракона", "Сферу Дракона", "Сферой Дракона"},
    {0, "Orb des Drachen", "Orbs des Drachen", "Orb des Drachen", "Orb des Drachen"},
    {0/*MISSING*/ , "Orb of the Dragon"/*MISSING*/ , "Orb of the Dragon"/*MISSING*/ , "Orb of the Dragon"/*MISSING*/ , "Orb of the Dragon"},
    }},
  {0x4ae2d4e6, 0, { // ""Speed Witch""
    {1, "Czarownica Szybka", "Czarownice Szybkie", "Czarownicę Szybką", "Czarownicą Szybką"},
    {1, "Hız Cadısı", "Hız Cadıları", "Hız Cadısını", "Hız Cadısıyla"},
    {1, "Čarodějnice Rychlosti", "Čarodějnice Rychlosti", "Čarodějnici Rychlosti", "Čarodějnicí Rychlosti"},
    {1, "Ведьма скорости", "Ведьмы скорости", "Ведьму скорости", "Ведьмой скорости"},
    {1, "Geschwindigkeitshexe", "Geschwindigkeitshexen", "Geschwindigkeitshexe", "Geschwindigkeitshexe"},
    {0/*MISSING*/ , "Speed Witch"/*MISSING*/ , "Speed Witch"/*MISSING*/ , "Speed Witch"/*MISSING*/ , "Speed Witch"},
    }},
  {0x4badb8db, 0, { // ""fresh grave""
    {3, "świeży grób", "świeże groby", "świeży grób", "świeżym grobem"},
    {3, "Taze Mezar", "Taze Mezarlar", "Taze Mezarı", "Taze Mezarda"},
    {3, "čerstvý hrob", "čerstvé hroby", "čerstvý hrob", "čerstvým hrobem"},
    {1, "свежая могила", "свежие могилы", "свежую могилу", "свежей могилой"},
    {0, "Frische Grab", "Frische Gräber", "Frisches Grab", "Frisches Grab"},
    {0, "túmulo recente", "túmulos recentes"/*MISSING*/ , "fresh grave", "túmulo recente"},
    }},
  {0x4c77a976, 0, { // ""Orb of Stunning""
    {1, "Sfera Ogłuszania", "Sfery Ogłuszania", "Sferę Ogłuszania", "Sferą Ogłuszania"},
    {1, "Sersemletme Küresi", "Sersemletme Küreleri", "Sersemletme Küresini", "Sersemletme Küresiyle"},
    {1, "Sféra Omráčení", "Sféry Omráčení", "Sféru Omráčení", "Sférou Omráčení"},
    {1, "Сфера Оглушения", "Сферы Оглушения", "Сферу Оглушения", "Сферой Оглушения"},
    {0, "Orb der Betäubung", "Orbs der Betäubung", "Orb der Betäubung", "Orb der Betäubung"},
    {0/*MISSING*/ , "Orb of Stunning"/*MISSING*/ , "Orb of Stunning"/*MISSING*/ , "Orb of Stunning"/*MISSING*/ , "Orb of Stunning"},
    }},
  {0x4c909cae, 0, { // ""Orb of Freedom""
    {1, "Sfera Wolności", "Sfery Wolności", "Sferę Wolności", "Sferą Wolności"},
    {0/*MISSING*/ , "Orb of Freedom"/*MISSING*/ , "Orb of Freedom"/*MISSING*/ , "Orb of Freedom"/*MISSING*/ , "Orb of Freedom"},
    {1, "Sféra Svobody", "Sféry Svobody", "Sféru Svobody", "Sférou Svobody"},
    {1, "Сфера Свободы", "Сферы Свободы", "Сферу Свободы", "Сферой Свободы"},
    {0, "Orb der Freiheit", "Orbs der Freiheit", "Orb der Freiheit", "Orb der Freiheit"},
    {0/*MISSING*/ , "Orb of Freedom"/*MISSING*/ , "Orb of Freedom"/*MISSING*/ , "Orb of Freedom"/*MISSING*/ , "Orb of Freedom"},
    }},
  {0x4df3f5d0, 0, { // ""Cultist""
    {0, "Kultysta", "Kultyści", "Kultystę", "Kultystą"},
    {0, "Mürit", "Müritler", "Müridi", "Müritle"},
    {0, "Kultista", "Kultisté", "Kultistu", "Kultistou"},
    {0, "Культист", "Культисты", "Культиста", "Культистом"},
    {0, "Kultanhänger", "Kultanhänger", "Kultanhänger", "Kultanhänger"},
    {0, "Cultista", "Cultistas"/*MISSING*/ , "Cultist", "Cultista"},
    }},
  {0x4fc909c3, 0, { // ""no monster""
    {3, "brak potwora", "braki potwora", "braku potwora", "brakiem potwora"},
    {0/*MISSING*/ , "no monster"/*MISSING*/ , "no monster"/*MISSING*/ , "no monster"/*MISSING*/ , "no monster"},
    {0, "žádný netvor", "žádní netvoři", "žádného netvora", "žádným netvorem"},
    {3, "нет монстра", "нет монстра", "нет монстра", "нет монстра"},
    {2, "kein Monster", "keine Monster", "kein Monster", "kein Monster"},
    {0/*MISSING*/ , "no monster"/*MISSING*/ , "no monster"/*MISSING*/ , "no monster"/*MISSING*/ , "no monster"},
    }},
  {0x4fd9547e, 0, { // ""big statue of Cthulhu""
    {3, "pomnik Cthulhu", "pomniki Cthulhu", "pomnik Cthulhu", "pomnikiem Cthulhu"},
    {3, "Cthulhu'nun Putu", "Cthulhu'nun Putları", "Cthulhu'nun Putunu", "Cthulhu'nun Putuyla"},
    {1, "Cthulhuova socha", "Cthulhuovy sochy", "Cthulhuovu sochu", "Cthulhuovou sochou"},
    {3, "памятник Ктулху", "памятники Ктулху", "памятник Ктулху", "памятником Ктулху"},
    {1, "große Statue des Cthulhu", "großen Statuen des Cthulhu", "großen Statue des Cthulhu", "großen Statue des Cthulhu"},
    {0/*MISSING*/ , "big statue of Cthulhu"/*MISSING*/ , "big statue of Cthulhu"/*MISSING*/ , "big statue of Cthulhu"/*MISSING*/ , "big statue of Cthulhu"},
    }},
  {0x4fe64013, 0, { // ""Mutant Sapling""
    {1, "Zmutowana Sadzonka", "Zmutowane Sadzonki", "Zmutowaną Sadzonkę", "Zmutowaną Sadzonką"},
    {1, "Mutant Fidan", "Mutant Fidanlar", "Mutant Fidanı", "Mutant Fidanla"},
    {3, "Mutantní semenáček", "Mutantní semenáčky", "Mutantní semenáček", "Mutantním semenáčkem"},
    {3, "Саженец-мутант", "Саженцы-мутанты", "Саженец-мутант", "Саженцем-мутантом"},
    {0, "Mutierter Sätzling", "Mutierte Sätzlinge", "Mutierten Sätzling", "Mutierten Sätzling"},
    {0/*MISSING*/ , "Mutant Sapling"/*MISSING*/ , "Mutant Sapling"/*MISSING*/ , "Mutant Sapling"/*MISSING*/ , "Mutant Sapling"},
    }},
  {0x5212cc51, 0, { // ""sandstone wall""
    {1, "piaskowa ściana", "piaskowe ściany", "piaskową scianę", "piaskową ścianą"},
    {1, "kumtaşı duvarı", "kumtaşı duvarları", "kumtaşı duvarını", "kumtaşı duvarla"},
    {1, "pískovcová zeď", "pískovcové zdi", "pískovcovou zeď", "pískovcovou zdí"},
    {1, "песчаниковая стена", "песчаниковые стены", "песчаниковую стену", "песчаниковой стеной"},
    {1, "Sandsteinwand", "Sandsteinwände", "Sandsteinwand", "Sandsteinwand"},
    {0/*MISSING*/ , "sandstone wall"/*MISSING*/ , "sandstone wall"/*MISSING*/ , "sandstone wall"/*MISSING*/ , "sandstone wall"},
    }},
  {0x521ec5ac, 0, { // ""Necromancer's Totem""
    {3, "Totem Nekromanty", "Totemy Nekromanty", "Totem Nekromanty", "Totemem Nekromanty"},
    {3, "Ölüçağıran Totemi", "Ölüçağıran Totemleri", "Ölüçağıran Totemini", "Ölüçağıran Totemiyle"},
    {3, "Nekromancerův totem", "Nekromancerovy totemy", "Nekromancerův totem", "Nekromancerovým totemem"},
    {3, "Тотем некроманта", "Тотемы некроманта", "Тотем некроманта", "Тотемом некроманта"},
    {2, "Totenbeschwörer-Totem", "Totenbeschwörer-Totems", "Totenbeschwörer-Totem", "Totenbeschwörer-Totem"},
    {0, "Totem do Necromante", "Totens do Necromante"/*MISSING*/ , "Necromancer's Totem", "Totem do Necromante"},
    }},
  {0x52e162a1, 0, { // ""Ratling Avenger""
    {0, "Szczurak-Mściciel", "Szczuraki-Mściciele", "Szczuraka-Mściciela", "Szczurakiem-Mścicielem"},
    {0/*MISSING*/ , "Ratling Avenger"/*MISSING*/ , "Ratling Avenger"/*MISSING*/ , "Ratling Avenger"/*MISSING*/ , "Ratling Avenger"},
    {0, "Krysák mstitel", "Krysáci mstitelé", "Krysáka mstitele", "Krysákem mstitelem"},
    {0, "Крысолюд-мститель", "Крысолюди-мстители", "Крысолюда-мстителя", "Крысолюдом-мстителем"},
    {0, "Rattenmenschen-Rächer", "Rattenmenschen-Rächer", "Rattenmenschen-Rächer", "Rattenmenschen-Rächer"},
    {0/*MISSING*/ , "Ratling Avenger"/*MISSING*/ , "Ratling Avenger"/*MISSING*/ , "Ratling Avenger"/*MISSING*/ , "Ratling Avenger"},
    }},
  {0x532d17d5, 0, { // ""Bomberbird Egg""
    {2, "Jajo Bombardiera", "Jaja Bombardiera", "Jajo Bombardiera", "Jajem Bombardiera"},
    {2, "Bombacıkuş Yumurtası", "Bombacıkuş Yumurtaları", "Bombacıkuş Yumurtasını", "Bombacıkuş Yumurtasıyla"},
    {2, "Vejce Bombarďáka", "Vejce Bombarďáka", "Vejce Bombarďáka", "Vejcem Bombarďáka"},
    {2, "Яйцо бомбардира", "Яйца бомбардира", "Яйцо бомбардира", "Яйцом бомбардира"},
    {2, "Bombenvogelei", "Bombenvogeleier", "Bombenvogelei", "Bombenvogelei"},
    {0/*MISSING*/ , "Bomberbird Egg"/*MISSING*/ , "Bomberbird Egg"/*MISSING*/ , "Bomberbird Egg"/*MISSING*/ , "Bomberbird Egg"},
    }},
  {0x54279e12, 0, { // ""Ivory Tower""
    {1, "Wieża z Kości Słoniowej", "Wieże z Kości Słoniowej", "Wieżą z Kości Słoniowej", "na Wieży z Kości Słoniowej"},
    {1, "Fildişi Kule", "Fildişi Kuleler", "Fildişi Kuleyi", "Fildişi Kulede"},
    {1, "Věž ze slonové kosti", "Věže ze slonové kosti", "Věž ze slonové kosti", "ve Věži ze slonové kosti"},
    {1, "Башня из слоновой кости", "Башни из слоновой кости", "Башню из слоновой кости", "в Башне из слоновой кости"},
    {0, "Elfenbeinturm", "Elfenbeintürme", "Elfenbeinturm", "Elfenbeinturm"},
    {0/*MISSING*/ , "Ivory Tower"/*MISSING*/ , "Ivory Tower"/*MISSING*/ , "Ivory Tower"/*MISSING*/ , "Ivory Tower"},
    }},
  {0x55dea6bd, 0, { // ""Orb of Aether""
    {1, "Sfera Eteru", "Sfery Eteru", "Sferę Eteru", "Sferą Eteru"},
    {1, "Eter Küresi", "Eter Küreleri", "Eter Küresini", "Eter Küresiyle"},
    {1, "Sféra Éteru", "Sféry Éteru", "Sféru Éteru", "Sférou Éteru"},
    {1, "Сфера Эфира", "Сферы Эфира", "Сферу Эфира", "Сферой Эфира"},
    {0, "Orb des Äthers", "Orbs des Äthers", "Orb des Äthers", "Orb des Äthers"},
    {0/*MISSING*/ , "Orb of Aether"/*MISSING*/ , "Orb of Aether"/*MISSING*/ , "Orb of Aether"/*MISSING*/ , "Orb of Aether"},
    }},
  {0x55eafd16, 0, { // ""Statue of Cthulhu""
    {1, "Statuetka", "Statuetki", "Statuetkę", "Statuetką"},
    {1, "Cthulhu Heykeli", "Cthulhu Heykelleri", "Cthulhu Heykelini", "Cthulhu Heykeliyle"},
    {1, "Cthulhuova soška", "Cthulhuovy sošky", "Cthulhuovu sošku", "Cthulhuovou soškou"},
    {1, "Статуя Ктулху", "Статуи Ктулху", "Статую Ктулху", "Статуей Ктулху"},
    {1, "Cthulhu Statue", "Cthulhu Statuen", "Cthulhu Statue", "Cthulhu Statue"},
    {1, "Estátua do Cthulhu", "Estátuas do Cthulhu"/*MISSING*/ , "Statue of Cthulhu", "Estátua do Cthulhu"},
    }},
  {0x56dcb9b1, 0, { // ""Tentacle W""
    {1, "Macka W", "Macki W", "Mackę W", "Macką W"},
    {0, "Dokunaç W", "Dokunaçlar W", "Dokunacı W", "Dokunaçla W"},
    {2, "Chapadlo W", "Chapadla W", "Chapadlo W", "Chapadlem W"},
    {2, "Щупальце W", "Щупальца W", "Щупальце W", "Щупальцем W"},
    {0, "Tentakel W", "Tentakel W", "Tentakel W", "Tentakel W"},
    {0, "Tentáculo W", "Tentáculos W"/*MISSING*/ , "Tentacle W", "Tentáculo W"},
    }},
  {0x576cd2d8, 0, { // ""temporary bridge""
    {3, "tymczasowy most", "tymczasowe mosty", "tymczasowy most", "tymczasowym mostem"},
    {3, "geçici köprü", "geçici köprüler", "geçici köprüyü", "geçici köprüyle"},
    {3, "dočasný most", "dočasné mosty", "dočasný most", "dočasným mostem"},
    {3, "временный мост", "временные мосты", "временный мост", "временным мостом"},
    {1, "temporäre Brücke", "temporäre Brücken", "temporäre Brücke", "temporäre Brücke"},
    {0/*MISSING*/ , "temporary bridge"/*MISSING*/ , "temporary bridge"/*MISSING*/ , "temporary bridge"/*MISSING*/ , "temporary bridge"},
    }},
  {0x591a8afe, 0, { // ""Rock Troll""
    {0, "Skalny Troll", "Skalne Trolle", "Skalnego Trolla", "Skalnym Trollem"},
    {0, "Kaya Trolü", "Kaya Trolleri", "Kaya Trolünü", "Kaya Trolüyle"},
    {0, "Šutroll", "Šutrollové", "Šutrolla", "Šutrollem"},
    {0, "Горный тролль", "Горные тролли", "Горного Тролля", "Горным троллем"},
    {0, "Steintroll", "Steintrolle", "Steintroll", "Steintroll"},
    {0, "Troll de Pedra", "Trolls de Pedra"/*MISSING*/ , "Rock Troll", "Troll de Pedra"},
    }},
  {0x5b31dfeb, 0, { // ""Orb of the Frog""
    {1, "Sfera Żaby", "Sfery Żaby", "Sferę Żaby", "Sferą Żaby"},
    {1, "Kurbağa Küresi", "Kurbağa Küreleri", "Kurbağa Küresini", "Kurbağa Küresiyle"},
    {1, "Sféra Žáby", "Sféry Žáby", "Sféru Žáby", "Sférou Žáby"},
    {1, "Сфера Жабы", "Сферы Жабы", "Сферу Жабы", "Сферой Жабы"},
    {0, "Orb des Frosches", "Orbs des Frosches", "Orb des Frosches", "Orb des Frosches"},
    {0/*MISSING*/ , "Orb of the Frog"/*MISSING*/ , "Orb of the Frog"/*MISSING*/ , "Orb of the Frog"/*MISSING*/ , "Orb of the Frog"},
    }},
  {0x5b8f4cbc, 0, { // ""Emerald""
    {3, "Szmaragd", "Szmaragdy", "Szmaragd", "Szmaragdem"},
    {3, "Zümrüt", "Zümrütler", "Zümrütü", "Zümrütle"},
    {3, "Smaragd", "Smaragdy", "Smaragd", "Smaragdem"},
    {3, "Изумруд", "Изумруды", "Изумруд", "Изумрудом"},
    {0, "Smaragd", "Smaragde", "Smaragd", "Smaragd"},
    {0/*MISSING*/ , "Emerald"/*MISSING*/ , "Emerald"/*MISSING*/ , "Emerald"/*MISSING*/ , "Emerald"},
    }},
  {0x5b929d56, 0, { // ""crystal cabinet""
    {1, "kryształowa gablota", "kryształowe gabloty", "kryształową gablotę", "kryształową gablotą"},
    {1, "Kristal Kabin", "Kristal Kabinler", "Kristal Kabini", "Kristal Kabinle"},
    {1, "křišťálová vitrína", "křišťálové vitríny", "křišťálovou vitrínu", "křišťálovou vitrínou"},
    {3, "кристальный кабинет", "кристальные кабинеты", "кристальный кабинет", "кристальным кабинетом"},
    {1, "Kristallvitrine", "Kristallvitrinen", "Kristallvitrine", "Kristallvitrine"},
    {0/*MISSING*/ , "crystal cabinet"/*MISSING*/ , "crystal cabinet"/*MISSING*/ , "crystal cabinet"/*MISSING*/ , "crystal cabinet"},
    }},
  {0x5c708ba6, 0, { // ""Rose Lord""
    {0, "Różany Pan", "Różani Panowie", "Różanego Pana", "Różanym Panem"},
    {0/*MISSING*/ , "Rose Lord"/*MISSING*/ , "Rose Lord"/*MISSING*/ , "Rose Lord"/*MISSING*/ , "Rose Lord"},
    {1, "Růžový pán", "Rúžoví pánové", "Růžového pána", "Růžovým pánem"},
    {0, "Розовый лорд", "Розовые лорды", "Розового лорда", "Розовым лордом"},
    {0, "Rosenherr", "Rosenherren", "Rosenherr", "Rosenherr"},
    {0/*MISSING*/ , "Rose Lord"/*MISSING*/ , "Rose Lord"/*MISSING*/ , "Rose Lord"/*MISSING*/ , "Rose Lord"},
    }},
  {0x5e748a12, 0, { // ""Gargoyle""
    {0, "Rzygacz", "Rzygacze", "Rzygacza", "Rzygaczem"},
    {0, "Çörten", "Çörtenler", "Çörteni", "Çörtenle"},
    {3, "Chrlič", "Chrliče", "Chrlič", "Chrličem"},
    {1, "Горгулья", "Горгульи", "Горгулью", "Горгульей"},
    {0, "Gargoyle", "Gargoyles", "Gargoyle", "Gargoyle"},
    {0/*MISSING*/ , "Gargoyle"/*MISSING*/ , "Gargoyle"/*MISSING*/ , "Gargoyle"/*MISSING*/ , "Gargoyle"},
    }},
  {0x5ee2290b, 0, { // ""platform""
    {1, "platforma", "platformy", "platformę", "platformą"},
    {0/*MISSING*/ , "platform"/*MISSING*/ , "platform"/*MISSING*/ , "platform"/*MISSING*/ , "platform"},
    {1, "plošinka", "plošinky", "plošinku", "plošinkou"},
    {1, "платформа", "платформы", "платформу", "платформой"},
    {1, "Plattform", "Plattformen", "Plattform", "Plattform"},
    {0/*MISSING*/ , "platform"/*MISSING*/ , "platform"/*MISSING*/ , "platform"/*MISSING*/ , "platform"},
    }},
  {0x5f9e644d, 0, { // ""Jungle""
    {1, "Dżungla", "Dżungla", "Dżunglę", "w Dżungli"},
    {1, "Vahşi Orman", "Vahşi Ormanlar", "Vahşi Ormanı", "Vahşi Ormanda"},
    {1, "Džungle", "Džungle", "Džungli", "v džungli"},
    {2, "Джунгли", "Джунгли", "Джунгли", "в Джунглях"},
    {0, "Dschungel", "Dschungel", "im Dschungel", "im Dschungel"},
    {1, "Floresta", "Florestas"/*MISSING*/ , "Jungle", "Floresta"},
    }},
  {0x60737bbc, 0, { // ""Winter Witch""
    {1, "Czarownica Zimy", "Czarownice Zimy", "Czarownicę Zimy", "Czarownicą Zimy"},
    {1, "Kış Cadısı", "Kış Cadıları", "Kış Cadısını", "Kış Cadısıyla"},
    {1, "Čarodějnice Zimy", "Čarodějnice Zimy", "Čarodějnici Zimy", "Čarodějnicí Zimy"},
    {1, "Ведьма зимы", "Ведьмы зимы", "Ведьму зимы", "Ведьмой зимы"},
    {1, "Frosthexe", "Frosthexen", "Frosthexe", "Frosthexe"},
    {0/*MISSING*/ , "Winter Witch"/*MISSING*/ , "Winter Witch"/*MISSING*/ , "Winter Witch"/*MISSING*/ , "Winter Witch"},
    }},
  {0x60ea65ce, 0, { // ""Reptile bridge""
    {3, "Jaszczurzy most", "Jaszczurze mosty", "Jaszczurzy most", "Jaszczurzym mostem"},
    {0/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"},
    {3, "Ještěrčí most", "Ještěrčí mosty", "Ještěrčí most", "Ještěrčím mostem"},
    {3, "Ящеричный мост", "Ящеричные мосты", "Ящеричный мост", "Ящеричным мостом"},
    {0/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"},
    {0/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"/*MISSING*/ , "Reptile bridge"},
    }},
  {0x63e60e0f, 0, { // ""Emerald Mine""
    {1, "Kopalnia Szmaragdów", "Kopalnia Szmaragdów", "Kopalnię Szmaragdów", "w Kopalni Szmaragdów"},
    {1, "Zümrüt Madeni", "Zümrüt Madenleri", "Zümrüt Madenini", "Zümrüt Madeninde"},
    {3, "Smaragdový důl", "Smaragdové doly", "Smaragdový důl", "ve Smaragdovém dole"},
    {1, "Изумрудная шахта", "Изумрудные шахты", "Изумрудную шахту", "в Изумрудной шахте"},
    {1, "Smaragdmine", "Smaragdminen", "in der Smaragdmine", "in der Smaragdmine"},
    {0/*MISSING*/ , "Emerald Mine"/*MISSING*/ , "Emerald Mine"/*MISSING*/ , "Emerald Mine"/*MISSING*/ , "Emerald Mine"},
    }},
  {0x647840a5, 0, { // ""Fireball""
    {1, "Ognista Kula", "Ogniste Kule", "Ognistą Kulę", "Ognistą Kulą"},
    {1, "Ateş Topu", "Ateş Topları", "Ateş Topunu", "Ateş Topuyla"},
    {1, "Ohnivá koule", "Ohnivé koule", "Ohnivou kouli", "Ohnivou koulí"},
    {3, "Огненный Шар", "Огненные Шары", "Огненный Шар", "Огненным Шаром"},
    {0, "Feuerball", "Feuerbälle", "Feuerball", "Feuerball"},
    {0/*MISSING*/ , "Fireball"/*MISSING*/ , "Fireball"/*MISSING*/ , "Fireball"/*MISSING*/ , "Fireball"},
    }},
  {0x64ac7100, 0, { // ""Earth Shard""
    {3, "Okruch Ziemi", "Okruchy Ziemi", "Okruch Ziemi", "Okruchem Ziemi"},
    {0, "Toprak Parçası", "Toprak Parçaları", "Toprak Parçasını", "Toprak Parçasıyla"},
    {3, "Úlomek Země", "Úlomky Země", "Úlomek Země", "Úlomkem Země"},
    {3, "Осколок Земли", "Осколки Земли", "Осколок Земли", "Осколком Земли"},
    {1, "Erdscherbe", "Erdscherben", "Erdscherbe", "Erdscherbe"},
    {0/*MISSING*/ , "Earth Shard"/*MISSING*/ , "Earth Shard"/*MISSING*/ , "Earth Shard"/*MISSING*/ , "Earth Shard"},
    }},
  {0x652bd1e5, 0, { // ""Caribbean""
    {2, "Karaiby", "Morze Karaibskie", "Karaibów", "na Karaibach"},
    {0, "Karayipler", "Karayip", "Karayipleri", "Karayiplerde"},
    {0, "Karibik", "Karibiky", "Karibuku", "v Karibiku"},
    {0, "Карибы", "Карибы", "Карибы", "на Карибах"},
    {1, "Karibik", "Karibiken", "in der Karibik", "in der Karibik"},
    {0/*MISSING*/ , "Caribbean"/*MISSING*/ , "Caribbean"/*MISSING*/ , "Caribbean"/*MISSING*/ , "Caribbean"},
    }},
  {0x654fc32d, 0, { // ""partial fire""
    {0, "częściowy ogień", "częściowe ognie", "częściowy ogień", "częściowym ogniem"},
    {0, "kısmî ateş", "kısmî ateşler", "kısmî ateşi", "kısmî ateşle"},
    {0, "částečný oheň", "částečné ohně", "částečný oheň", "částečným ohněm"},
    {3, "частичный огонь", "частичные огни", "частичный огонь", "частичным огнём"},
    {2, "partielles Feuer", "partielle Feuer", "partielle Feuer", "partielle Feuer"},
    {0/*MISSING*/ , "partial fire"/*MISSING*/ , "partial fire"/*MISSING*/ , "partial fire"/*MISSING*/ , "partial fire"},
    }},
  {0x6578b7c3, 0, { // ""closing plate""
    {1, "zamykająca płyta", "zamykające płyty", "zamykającą płytę", "zamykającą płytą"},
    {1, "kapanan levha", "kapanan levhalar", "kapanan levhayı", "kapanan levhayla"},
    {1, "zavírací deska", "zavírací desky", "zavírací desku", "zavírací deskou"},
    {1, "закрывающая плита", "закрывающие плиты", "закрывающую плиту", "закрывающей плитой"},
    {0, "Schließmechanismus", "Schließmechanismen", "Schließmechanismus", "Schließmechanismus"},
    {0/*MISSING*/ , "closing plate"/*MISSING*/ , "closing plate"/*MISSING*/ , "closing plate"/*MISSING*/ , "closing plate"},
    }},
  {0x65eeb55f, 0, { // ""Friendly Ivy""
    {3, "Przyjazny Bluszcz", "Przyjazne Bluszcze", "Przyjazny Bluszcz", "Przyjaznym Bluszczem"},
    {0/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"},
    {3, "Přátelský břečťan", "Přátelské břečťany", "Přátelský břečťan", "Přátelským břečťanem"},
    {3, "Ручной плющ", "Ручные плющи", "Ручной плющ", "Ручным плющом"},
    {0/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"},
    {0/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"/*MISSING*/ , "Friendly Ivy"},
    }},
  {0x66adb446, 0, { // ""tree""
    {2, "drzewo", "drzewo", "drzewo", "drzewem"},
    {2, "Ağaç", "Ağaçlar", "Ağacı", "Ağaçla"},
    {3, "strom", "stromy", "strom", "stromem"},
    {2, "дерево", "деревья", "дерево", "деревом"},
    {0, "Baum", "Bäume", "Baum", "Baum"},
    {1, "árvore", "árvores"/*MISSING*/ , "tree", "árvore"},
    }},
#if REPEATED
  {0x66adb447, 0, { // ""tree [repeat]""
    {2, "drzewo", "drzewa", "drzewo", "drzewem"},
    {2, "ağaç", "ağaçlar", "ağacı", "ağaçla"},
    {3, "strom", "stromy", "strom", "stromem"},
    {2, "дерево", "деревья", "дерево", "деревом"},
    {0, "Baum", "Bäume", "Baum", "Baum"},
    {0/*MISSING*/ , "tree [repeat]"/*MISSING*/ , "tree [repeat]"/*MISSING*/ , "tree [repeat]"/*MISSING*/ , "tree [repeat]"},
    }},
#endif
  {0x671492bd, 0, { // ""living floor""
    {1, "żywa ziemia", "żywe ziemie", "żywą ziemię", "żywą ziemią"},
    {1, "Canlı Zemin", "Canlı Zeminler", "Canlı Zemini", "Canlı Zeminde"},
    {1, "živoucí podlaha", "živoucí podlahy", "živoucí podlahu", "živoucí podlahou"},
    {3, "живой пол", "живые полы", "живой пол", "живым полом"},
    {0, "Lebender Boden", "Lebende Böden", "Lebender Boden", "Lebender Boden"},
    {0, "chão vivo", "chãos vivos"/*MISSING*/ , "living floor", "chão vivo"},
    }},
  {0x67a2b7fb, 0, { // ""Striped Dog""
    {0, "Pasiasty Pies", "Pasiate Psy", "Pasiastego Psa", "Pasiastym Psem"},
    {0, "Şeritli Köpek", "Şeritli Köpekler", "Şeritli Köpeği", "Şeritli Köpekle"},
    {0, "Pruhovaný pes", "Pruhovaní psi", "Pruhovaného psa", "Pruhovaným psem"},
    {0, "Полосатый пёс", "Полосатые псы", "Полосатого пса", "Полосатым псом"},
    {0, "Gestreifter Hund", "Gestreifte Hunde", "Gestreiften Hund", "Gestreiften Hund"},
    {0/*MISSING*/ , "Striped Dog"/*MISSING*/ , "Striped Dog"/*MISSING*/ , "Striped Dog"/*MISSING*/ , "Striped Dog"},
    }},
  {0x682e01d4, 2, { // ""Crossroads III""
    {2, "Skrzyżowanie III", "Skrzyżowania III", "Skrzyżowanie III", "na Skrzyżowaniu III"},
    {2, "Arayollar III", "Arayollar III", "Arayolları III", "Arayollarda III"},
    {1, "Křižovatka III", "Křižovatky III", "Křižovatku III", "na Křižovatc III"},
    {3, "Перекрёсток III", "Перекрёстки III", "Перекрёсток III", "на Перекрёстке III"},
    {1, "Kreuzungen III", "Kreuzungen III", "auf den Kreuzungen III", "auf den Kreuzungen III"},
    {0/*MISSING*/ , "Crossroads III"/*MISSING*/ , "Crossroads III"/*MISSING*/ , "Crossroads III"/*MISSING*/ , "Crossroads III"},
    }},
  {0x6867efd1, 0, { // ""Lesser Demon""
    {0, "Mniejszy Demon", "Mniejsze Demony", "Mniejszego Demona", "Mniejszym Demonem"},
    {0, "Küçük Şeytan", "Küçük Şeytanlar", "Küçük Şeytanı", "Küçük Şeytanla"},
    {0, "Čert", "Čerti", "Čerta", "Čertem"},
    {0, "Младший демон", "Младшие демоны", "Младшего демона", "Младшим демоном"},
    {0, "Gewöhnliche Dämon", "Gewöhnlichen Dämonen", "Gewöhnlichen Dämon", "Gewöhnlichen Dämon"},
    {0, "Demônio Menor", "Demônios Menores"/*MISSING*/ , "Lesser Demon", "Demônio Menor"},
    }},
  {0x6a3f72bc, 0, { // ""Round Table""
    {3, "Okrągły Stół", "Okrągłe Stoły", "Okrągły Stół", "Okrągłym Stołem"},
    {3, "Yuvarlak Masa", "yuvarlak Masalar", "Yuvarlak Masayı", "Yuvarlak Masayla"},
    {3, "Kulatý stůl", "Kulaté stoly", "Kulatý stůl", "Kulatým stolem"},
    {3, "Круглый Стол", "Круглые Столы", "Круглый Стол", "Круглому Столу"},
    {1, "Tafelrunde", "Tafelrunde", "Tafelrunde", "Tafelrunde"},
    {0/*MISSING*/ , "Round Table"/*MISSING*/ , "Round Table"/*MISSING*/ , "Round Table"/*MISSING*/ , "Round Table"},
    }},
  {0x6ac36d20, 0, { // ""canopy""
    {2, "listowie", "listowia", "listowie", "listowiem"},
    {0/*MISSING*/ , "canopy"/*MISSING*/ , "canopy"/*MISSING*/ , "canopy"/*MISSING*/ , "canopy"},
    {1, "koruna", "koruny", "korunu", "korunou"},
    {1, "листва", "листва", "листву", "листвой"},
    {2, "Kronendach", "Kronendächer", "Kronendach", "Kronendach"},
    {0/*MISSING*/ , "canopy"/*MISSING*/ , "canopy"/*MISSING*/ , "canopy"/*MISSING*/ , "canopy"},
    }},
  {0x6af295a7, 0, { // ""rosebush""
    {3, "krzak róży", "krzaki róży", "krzak róży", "krzakiem róży"},
    {0/*MISSING*/ , "rosebush"/*MISSING*/ , "rosebush"/*MISSING*/ , "rosebush"/*MISSING*/ , "rosebush"},
    {3, "růžový keřík", "růžové keříky", "růžový keřík", "růžovým keříkem"},
    {3, "куст розы", "кусты розы", "куст розы", "кустом розы"},
    {0, "Rosenbusch", "Rosenbüsche", "Rosenbusch", "Rosenbusch"},
    {0/*MISSING*/ , "rosebush"/*MISSING*/ , "rosebush"/*MISSING*/ , "rosebush"/*MISSING*/ , "rosebush"},
    }},
  {0x6bd7c035, 0, { // ""Sea Border""
    {3, "Granica morska", "Granice morskie", "Granicy morskiej", "na Granicy morskiej"},
    {3, "Deniz Sınırı", "Deniz Sınırları", "Deniz Sınırını", "Deniz Sınırında"},
    {3, "Mořská hranice", "Mořské hranice", "Mořskou hranici", "na Mořské hranici"},
    {3, "Граница моря", "Границы моря", "Границу моря", "на Границе моря"},
    {1, "Seegrenze", "Seegrenzen", "Seegrenze", "Seegrenze"},
    {0/*MISSING*/ , "Sea Border"/*MISSING*/ , "Sea Border"/*MISSING*/ , "Sea Border"/*MISSING*/ , "Sea Border"},
    }},
  {0x6c135277, 0, { // ""lake""
    {2, "jezioro", "jeziora", "jezioro", "jeziorem"},
    {2, "Göl", "Göller", "Gölü", "Gölde"},
    {2, "jezero", "jezera", "jezero", "jezerem"},
    {2, "озеро", "озёра", "озеро", "озером"},
    {0, "See", "Seen", "See", "See"},
    {0, "lago", "lagos"/*MISSING*/ , "lake", "lago"},
    }},
  {0x6c2b5c5d, 0, { // ""Overgrown Woods""
    {3, "Przerośnięty Las", "Przerośnięte Lasy", "Przerośnięty Las", "w Przerośniętym Lesie"},
    {3, "Azman Orman", "Azman Ormanlar", "Azman Ormanı", "Azman Ormanda"},
    {3, "Prales", "Pralesy", "Prales", "v Pralese"},
    {1, "Заросли", "Заросли", "Заросли", "в Зарослях"},
    {0, "Verwilderter Wald", "Verwilderte Wälder", "im Verwilderten Wald", "im Verwilderten Wald"},
    {0/*MISSING*/ , "Overgrown Woods"/*MISSING*/ , "Overgrown Woods"/*MISSING*/ , "Overgrown Woods"/*MISSING*/ , "Overgrown Woods"},
    }},
  {0x6c49496c, 0, { // ""Yendorian Forest""
    {3, "Las Yendoriański", "Lasy Yendoriańskie", "Yendoriańskiego Lasu", "w Yendoriańskim Lesie"},
    {0/*MISSING*/ , "Yendorian Forest"/*MISSING*/ , "Yendorian Forest"/*MISSING*/ , "Yendorian Forest"/*MISSING*/ , "Yendorian Forest"},
    {3, "Yendorský les", "Yendorské lesy", "Yendorský les", "v Yendorském lese"},
    {3, "Йендорский Лес", "Йендорские Леса", "Йендорский Лес", "в Йендорском Лесу"},
    {3, "Yendorianischer Wald", "Yendorianische Wälder", "im Yendorianischen Wald", "im Yendorianischen Wald"},
    {0/*MISSING*/ , "Yendorian Forest"/*MISSING*/ , "Yendorian Forest"/*MISSING*/ , "Yendorian Forest"/*MISSING*/ , "Yendorian Forest"},
    }},
  {0x6c8e75d1, 0, { // ""Burial Grounds""
    {3, "Kurhany", "Kurhany", "Kurhany", "w Kurhanach"},
    {0/*MISSING*/ , "Burial Grounds"/*MISSING*/ , "Burial Grounds"/*MISSING*/ , "Burial Grounds"/*MISSING*/ , "Burial Grounds"},
    {2, "Pohřebiště", "Pohřebiště", "Pohřebiště", "na Pohřebišti"},
    {3, "Курганы", "Курганы", "Курганы", "в Курганах"},
    {1, "Begräbnisstätten", "Begräbnisstätten", "in den Begräbnisstätten", "in den Begräbnisstätten"},
    {0/*MISSING*/ , "Burial Grounds"/*MISSING*/ , "Burial Grounds"/*MISSING*/ , "Burial Grounds"/*MISSING*/ , "Burial Grounds"},
    }},
  {0x6d5cef1f, 0, { // ""Bull Dash""
    {1, "Kraina Byków", "Krainy Byków", "Krainę Byków", "w Krainie Byków"},
    {0/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"},
    {3, "Býčí kraj", "Býčí kraje", "Býčí kraj", "v Býčím kraji"},
    {1, "Земля быков", "Земли быков", "Землю быков", "на Земле быков"},
    {0/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"},
    {0/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"/*MISSING*/ , "Bull Dash"},
    }},
  {0x6d815a1e, 0, { // ""Orb of the Stone""
    {1, "Sfera Kamienia", "Sfery Kamienia", "Sferę Kamienia", "Sferą Kamienia"},
    {0/*MISSING*/ , "Orb of the Stone"/*MISSING*/ , "Orb of the Stone"/*MISSING*/ , "Orb of the Stone"/*MISSING*/ , "Orb of the Stone"},
    {1, "Sféra Kamene", "Sféry Kamene", "Sféru Kamene", "Sférou Kamene"},
    {1, "Сфера Камня", "Сферы Камня", "Сферу Камня", "Сферой Камня"},
    {0, "Orb des Steines", "Orbs des Steines", "Orb des Steines", "Orb des Steines"},
    {0/*MISSING*/ , "Orb of the Stone"/*MISSING*/ , "Orb of the Stone"/*MISSING*/ , "Orb of the Stone"/*MISSING*/ , "Orb of the Stone"},
    }},
  {0x6df7abdd, 0, { // ""Fire Elemental""
    {0, "Ognisty Żywiołak", "Ogniste Żywiołaki", "Ognistego Żywiołaka", "Ognistym Żywiołakiem"},
    {0, "Ateş Unsuru", "Ateş Unsurları", "Ateş Unsurunu", "Ateş Unsuruyla"},
    {0, "Elementál Ohně", "Elementálové Ohně", "Elementála Ohně", "Elementálem Ohně"},
    {0, "Огненный Элементаль", "Огненные Элементали", "Огненного Элементаля", "Огненным Элементалем"},
    {0, "Feuerelementar", "Feuerelementare", "Feuerelementar", "Feuerelementar"},
    {0/*MISSING*/ , "Fire Elemental"/*MISSING*/ , "Fire Elemental"/*MISSING*/ , "Fire Elemental"/*MISSING*/ , "Fire Elemental"},
    }},
  {0x6e596c8e, 0, { // ""Zebra""
    {1, "Zebra", "Zebry", "Zebrą", "w Zebrze"},
    {1, "Zebra", "Zebralar", "Zebrayı", "Zebrada"},
    {1, "Zebra", "Zebry", "Zebru", "v Zebře"},
    {1, "Зебра", "Зебры", "Зебру", "в Зебре"},
    {2, "Zebra", "Zebras", "in Zebra", "in Zebra"},
    {0/*MISSING*/ , "Zebra"/*MISSING*/ , "Zebra"/*MISSING*/ , "Zebra"/*MISSING*/ , "Zebra"},
    }},
  {0x6ef82407, 0, { // ""Ice Diamond""
    {3, "Lodowy Diament", "Lodowe Diamenty", "Lodowy Diament", "Lodowym Diamentem"},
    {3, "Buz Elması", "Buz Elmasları", "Buz Elmasını", "Buz Elmasıyla"},
    {3, "Ledový diamant", "Ledové diamanty", "Ledový diamant", "Ledovým diamantem"},
    {3, "Ледяной алмаз", "Ледяные алмазы", "Ледяной алмаз", "Ледяным алмазом"},
    {0, "Frostdiamant", "Frostdiamanten", "Frostdiamant", "Frostdiamant"},
    {0, "Diamante de Gelo", "Diamantes de Gelo"/*MISSING*/ , "Ice Diamond", "Diamante de Gelo"},
    }},
  {0x6fded839, 0, { // ""Orb of Summoning""
    {1, "Sfera Przywołania", "Sfery Przywołania", "Sferę Przywołania", "Sferą Przywołania"},
    {1, "Çağırma Küresi", "Çağırma Küreleri", "Çağırma Küresini", "Çağırma Küresiyle"},
    {1, "Sféra Vyvolání", "Sféry Vyvolání", "Sféru Vyvolání", "Sférou Vyvolání"},
    {1, "Сфера Призыва", "Сферы Призыва", "Сферу Призыва", "Сферой Призыва"},
    {0, "Orb der Beschwörung", "Orbs der Beschwörung", "Orb der Beschwörung", "Orb der Beschwörung"},
    {0/*MISSING*/ , "Orb of Summoning"/*MISSING*/ , "Orb of Summoning"/*MISSING*/ , "Orb of Summoning"/*MISSING*/ , "Orb of Summoning"},
    }},
  {0x6fe5cc7a, 0, { // ""charged wall""
    {1, "naładowana ściana", "naładoawne ściany", "naładowaną scianę", "naładowaną ścianą"},
    {1, "yüklü duvar", "yüklü duvarlar", "yüklü duvarı", "yüklü duvarla"},
    {1, "nabitá zeď", "nabité zdi", "nabitou zeď", "nabitou zdí"},
    {1, "заряженная стена", "заряженные стены", "заряженную стену", "заряженной стеной"},
    {1, "geladene Wand", "geladene Wände", "geladene Wand", "geladene Wand"},
    {0/*MISSING*/ , "charged wall"/*MISSING*/ , "charged wall"/*MISSING*/ , "charged wall"/*MISSING*/ , "charged wall"},
    }},
  {0x701e2c54, 0, { // ""Grimoire""
    {1, "Księga", "Księgi", "Księgę", "Księgą"},
    {1, "Kara Kitap", "Kara Kitaplar", "Kara Kitabı", "Kara Kitapla"},
    {3, "Grimoár", "Grimoáry", "Grimoár", "Grimoárem"},
    {3, "Гримуар", "Гримуары", "Гримуар", "Гримуаром"},
    {2, "Grimoire", "Grimoires", "Grimoire", "Grimoire"},
    {0/*MISSING*/ , "Grimoire"/*MISSING*/ , "Grimoire"/*MISSING*/ , "Grimoire"/*MISSING*/ , "Grimoire"},
    }},
  {0x703225bc, 0, { // ""Golem""
    {0, "Golem", "Golemy", "Golema", "Golemem"},
    {0, "Golem", "Golemler", "Golemi", "Golemle"},
    {0, "Golem", "Golemové", "Golema", "Golemem"},
    {0, "Голем", "Големы", "Голема", "Големом"},
    {0, "Golem", "Golems", "Golem", "Golem"},
    {0, "Golem", "Golems"/*MISSING*/ , "Golem", "Golem"},
    }},
  {0x70660f35, 0, { // ""Plane of Air""
    {1, "Strefa Powietrza", "Strefy Powietrza", "Strefę Powietrza", "w Strefie Powietrza"},
    {1, "Ateş Düzlemi", "Ateş Düzlemleri", "Hava Düzlemini", "Hava Düzleminde"},
    {1, "Rovina Vzduchu", "Roviny Vzduchu", "Rovinu Vzduchu", "v Rovině Vzduchu"},
    {1, "Плоскость Воздуха", "Плоскости Воздуха", "Плоскость Воздуха", "в Плоскости Воздуха"},
    {1, "Luftebene", "Luftebenen", "Luftebene", "Luftebene"},
    {0/*MISSING*/ , "Plane of Air"/*MISSING*/ , "Plane of Air"/*MISSING*/ , "Plane of Air"/*MISSING*/ , "Plane of Air"},
    }},
  {0x739ca4ff, 0, { // ""Orb of Nature""
    {1, "Sfera Natury", "Sfery Natury", "Sferę Natury", "Sferą Natury"},
    {0/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"},
    {1, "Sféra Přírody", "Sféry Přírody", "Sféru Přírody", "Sférou Přírody"},
    {1, "Сфера Природы", "Сферы Природы", "Сферу Природы", "Сферой Природы"},
    {0/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"},
    {0/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"/*MISSING*/ , "Orb of Nature"},
    }},
  {0x74510512, 0, { // ""blue slime""
    {1, "niebieska maź", "niebieskie mazie", "niebieską maź", "niebieską mazią"},
    {1, "Mavi Balçık", "Mavi Balçıklar", "Mavi Balçığı", "Mavi Balçıkta"},
    {3, "modrý sliz", "modré slizy", "modrý sliz", "modrým slizem"},
    {1, "синяя слизь", "синяя слизь", "синюю слизь", "синей слизью"},
    {0, "Blauer Schleim", "Blauen Schleime", "Blauen Schleim", "Blauen Schleim"},
    {0, "lodo azul", "lodos azuis"/*MISSING*/ , "blue slime", "lodo azul"},
    }},
  {0x74fc70a9, 0, { // ""cell without mine""
    {2, "pole bez miny", "pola bez min", "pole bez min", "polem bez min"},
    {2, "mayınsız hücre", "mayınsız hücreler", "mayınsız hücreyi", "mayınsız hücreyle"},
    {2, "pole bez miny", "pole bez min", "pole bez miny", "polem bez min"},
    {2, "поле без мин", "поля без мин", "поле без мин", "полем без мин"},
    {1, "minenlose Zelle", "minenlose Zellen", "minenlose Zelle", "minenlose Zelle"},
    {0/*MISSING*/ , "cell without mine"/*MISSING*/ , "cell without mine"/*MISSING*/ , "cell without mine"/*MISSING*/ , "cell without mine"},
    }},
  {0x752ff5e6, 0, { // ""Forest Troll""
    {0, "Leśny Troll", "Leśne Trolle", "Leśnego Trolla", "Leśnym Trollem"},
    {0, "Orman Trolü", "Orman Trolleri", "Orman Trolünü", "Orman Trolüyle"},
    {0, "Lestroll", "Lestrollové", "Lestrolla", "Lestrollem"},
    {0, "Лесной тролль", "Лесные тролли", "Лесного тролля", "Лесным троллем"},
    {0, "Waldtroll", "Waldtrolle", "Waldtroll", "Waldtroll"},
    {0/*MISSING*/ , "Forest Troll"/*MISSING*/ , "Forest Troll"/*MISSING*/ , "Forest Troll"/*MISSING*/ , "Forest Troll"},
    }},
  {0x7595ad5e, 0, { // ""rock III""
    {1, "skała III", "skały III", "skałę III", "skałą III"},
    {1, "kaya III", "kayalar III", "kayayı III", "kayayla III"},
    {1, "skála III", "skály III", "skálu III", "skálou III"},
    {1, "скалы III", "скалы III", "скалу III", "скалой III"},
    {0, "Fels III", "Felsen III", "Fels III", "Fels III"},
    {0/*MISSING*/ , "rock III"/*MISSING*/ , "rock III"/*MISSING*/ , "rock III"/*MISSING*/ , "rock III"},
    }},
  {0x75e56b56, 0, { // ""Whirlpool""
    {3, "Wir", "Wiry", "Wir", "w Wirze"},
    {3, "Tayfun", "Tayfunlar", "Tayfunu", "Tayfunda"},
    {3, "Vír", "Víry", "Vír", "ve Víru"},
    {3, "Водоворот", "Водовороты", "Водоворот", "в Водовороте"},
    {0, "Strudel", "Strudel", "im Strudel", "im Strudel"},
    {0/*MISSING*/ , "Whirlpool"/*MISSING*/ , "Whirlpool"/*MISSING*/ , "Whirlpool"/*MISSING*/ , "Whirlpool"},
    }},
  {0x76ef96b5, 0, { // ""great wall""
    {1, "wielka ściana", "wielkie ściany", "wielką ścianę", "wielką ścianą"},
    {1, "Çin Seddi", "Çin Setleri", "Çin Seddini", "Çin Seddinde"},
    {1, "velká zeď", "velké zdi", "velkou zeď", "velkou zdí"},
    {1, "великая стена", "великие стены", "великую стену", "великой стеной"},
    {1, "Große Mauer", "Großen Mauern", "Große Mauer", "Große Mauer"},
    {1, "muralha", "muralhas"/*MISSING*/ , "great wall", "muralha"},
    }},
  {0x7792b48f, 0, { // ""Mutant Fruit""
    {3, "Zmutowany Owoc", "Zmutowane Owoce", "Zmutowany Owoc", "Zmutowanym Owocem"},
    {0/*MISSING*/ , "Mutant Fruit"/*MISSING*/ , "Mutant Fruit"/*MISSING*/ , "Mutant Fruit"/*MISSING*/ , "Mutant Fruit"},
    {2, "Mutantní ovoce", "Mutantní ovoce", "Mutantní ovoce", "Mutantní ovoce"},
    {3, "Фрукт-мутант", "Фрукты-мутанты", "Фрукт-мутант", "Фруктом-мутантом"},
    {1, "Mutierte Frucht", "Mutierte Früchte", "Mutierte Frucht", "Mutierte Frucht"},
    {0/*MISSING*/ , "Mutant Fruit"/*MISSING*/ , "Mutant Fruit"/*MISSING*/ , "Mutant Fruit"/*MISSING*/ , "Mutant Fruit"},
    }},
  {0x78391eaf, 0, { // ""Desert Man""
    {0, "Człowiek Pustyni", "Ludzie Pustyni", "Człowieka Pustyni", "Człowiekiem Pustyni"},
    {0, "Bedevi", "Bedeviler", "Bedeviyi", "Bedeviyle"},
    {0, "Pouštní muž", "Pouštní muži", "Pouštního muže", "Pouštním mužem"},
    {0, "Человек пустыни", "Люди пустыни", "Человека пустыни", "Человеком пустыни"},
    {0, "Wüstenmann", "Wüstenmänner", "Wüstenmann", "Wüstenmann"},
    {0, "Homem do Deserto", "Homens do Deserto"/*MISSING*/ , "Desert Man", "Homem do Deserto"},
    }},
  {0x786f5e72, 0, { // ""Dungeon""
    {3, "Loch", "Lochy", "Loch", "w Lochu"},
    {0/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"},
    {3, "Žalář", "Žaláře", "Žalář", "v Žaláři"},
    {2, "Подземелье", "Подземелья", "Подземелье", "в Подземелье"},
    {0/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"},
    {0/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"/*MISSING*/ , "Dungeon"},
    }},
  {0x796d5023, 0, { // ""rock II""
    {1, "skała II", "skały II", "skałę II", "skałą II"},
    {1, "kaya II", "kayalar  II", "kayayı II", "kayayla II"},
    {1, "skála II", "skály II", "skálu II", "skálou II"},
    {1, "скалы II", "скалы II", "скалу II", "скалой II"},
    {0, "Fels II", "Felsen II", "Fels II", "Fels II"},
    {0/*MISSING*/ , "rock II"/*MISSING*/ , "rock II"/*MISSING*/ , "rock II"/*MISSING*/ , "rock II"},
    }},
  {0x79b02b48, 0, { // ""Fern Flower""
    {3, "Kwiat Paproci", "Kwiaty Paproci", "Kwiat Paproci", "Kwiatem Paproci"},
    {3, "Eğreltiotu Çiçeği", "Eğreltiotu Çiçekleri", "Eğreltiotu Çiçeğini", "Eğreltiotu Çiçeğiyle"},
    {3, "Kapradinový květ", "Kapradinové květy", "Kapradinový květ", "Kapradinové květy"},
    {3, "Цветок папоротника", "Цветы папоротника", "Цветок папоротника", "Цветком папоротника"},
    {0, "Farnblüte", "Farnblüten", "Farnblüte", "Farnblüte"},
    {1, "Flor de Samambaia", "Flores de Samambaia"/*MISSING*/ , "Fern Flower", "Flor de Samambaia"},
    }},
  {0x79cad766, 0, { // ""Rose Garden""
    {3, "Różany Ogród", "Różane Ogrody", "Różany Ogród", "w Różanym Ogrodzie"},
    {0/*MISSING*/ , "Rose Garden"/*MISSING*/ , "Rose Garden"/*MISSING*/ , "Rose Garden"/*MISSING*/ , "Rose Garden"},
    {1, "Růžová zahrada", "Růžové zahrady", "Růžovou zahradu", "v Růžové zahradě"},
    {3, "Розарий", "Розарии", "Розарий", "в Розарии"},
    {0, "Rosengarten", "Rosengärten", "im Rosengarten", "im Rosengarten"},
    {0/*MISSING*/ , "Rose Garden"/*MISSING*/ , "Rose Garden"/*MISSING*/ , "Rose Garden"/*MISSING*/ , "Rose Garden"},
    }},
  {0x7af26a7f, 0, { // ""closed gate""
    {1, "zamknięta brama", "zamknięte bramy", "zamkniętą bramę", "zamkniętą bramą"},
    {1, "kapalı kapı", "kapalı kapılar", "kapalı kapıyı", "kapalı kapıyla"},
    {1, "zavřená brána", "zavřené brány", "zavřenou bránu", "zavřenou bránou"},
    {3, "закрытый проход", "закрытые проходы", "закрытый проход", "закрытым проходом"},
    {2, "Geschlossenes Tor", "Geschlossene Tore", "Geschlossene Tor", "Geschlossene Tor"},
    {0/*MISSING*/ , "closed gate"/*MISSING*/ , "closed gate"/*MISSING*/ , "closed gate"/*MISSING*/ , "closed gate"},
    }},
  {0x7b02aea0, 0, { // ""Demon Shark""
    {0, "Rekin-Demon", "Rekiny-Demony", "Rekina-Demona", "Rekinem-Demonem"},
    {0, "Şeytan Köpekbalığı", "Şeytan Köpekbalıkları", "Şeytan Köpekbalığını", "Şeytan Köpekbalığıyla"},
    {0, "Ďábelský žralok", "Ďábelští žraloci", "Ďábelského žraloka", "Ďábelským žralokem"},
    {1, "Акула-демон", "Акулы-демоны", "Акулу-демона", "Акулой-демоном"},
    {0, "Dämonenhai", "Dämonenhaie", "Dämonenhai", "Dämonenhai"},
    {0, "Tubarão Demônio", "Tubarões Demônio"/*MISSING*/ , "Demon Shark", "Tubarão Demônio"},
    }},
  {0x7b736f7f, 0, { // ""Handsome Gardener""
    {0, "Przystojny Ogrodnik", "Przystojni Ogrodnicy", "Przystojnego Ogrodnika", "Przystojnym Ogrodnikiem"},
    {0/*MISSING*/ , "Handsome Gardener"/*MISSING*/ , "Handsome Gardener"/*MISSING*/ , "Handsome Gardener"/*MISSING*/ , "Handsome Gardener"},
    {1, "Pohledný zahradník", "Pohlední zahradníci", "Pohledného zahradníka", "Pohledným zahradníkem"},
    {0, "Красивый садовод", "Красивые садоводы", "Красивого садовника", "Красивым садовником"},
    {0, "Hübscher Gärtner", "Hübsche Gärtner", "Hübschen Gärtner", "Hübschen Gärtner"},
    {0/*MISSING*/ , "Handsome Gardener"/*MISSING*/ , "Handsome Gardener"/*MISSING*/ , "Handsome Gardener"/*MISSING*/ , "Handsome Gardener"},
    }},
  {0x7ce94873, 0, { // ""Bounty""
    {1, "Nagroda", "Nagrody", "Nagrodę", "Nagrodą"},
    {1, "Ödül", "Ödüller", "Ödülü", "Ödülle"},
    {1, "Odměna", "Odměny", "Odměnu", "Odměnou"},
    {1, "Награда", "Награды", "Награду", "Наградой"},
    {2, "Kopfgeld", "Kopfgelder", "Kopfgeld", "Kopfgeld"},
    {0/*MISSING*/ , "Bounty"/*MISSING*/ , "Bounty"/*MISSING*/ , "Bounty"/*MISSING*/ , "Bounty"},
    }},
  {0x7d9792d4, 0, { // ""Orb of Empathy""
    {1, "Sfera Empatii", "Sfery Empatii", "Sferę Empatii", "Sferą Empatii"},
    {0/*MISSING*/ , "Orb of Empathy"/*MISSING*/ , "Orb of Empathy"/*MISSING*/ , "Orb of Empathy"/*MISSING*/ , "Orb of Empathy"},
    {1, "Sféra Empatie", "Sféry Empatie", "Sféru Empatie", "Sférou Empatie"},
    {1, "Сфера Сочувствия", "Сферы Сочувствия", "Сферу Сочувствия", "Сферой Сочувствия"},
    {0, "Orb der Empathie", "Orbs der Empathie", "Orb der Empathie", "Orb der Empathie"},
    {0/*MISSING*/ , "Orb of Empathy"/*MISSING*/ , "Orb of Empathy"/*MISSING*/ , "Orb of Empathy"/*MISSING*/ , "Orb of Empathy"},
    }},
  {0x7fc6879d, 0, { // ""Bat""
    {3, "Nietoperz", "Nietoperze", "Nietoperza", "Nietoperzem"},
    {0/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"},
    {0, "Netopýr", "Netopýři", "Netopýra", "Netopýrem"},
    {1, "Летучая мышь", "Летучие мыши", "Летучую мышь", "Летучей мышью"},
    {0/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"},
    {0/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"/*MISSING*/ , "Bat"},
    }},
  {0x8056e574, 0, { // ""Prairie""
    {1, "Preria", "Prerie", "Prerię", "na Prerii"},
    {0/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"},
    {1, "Prérie", "Prérii", "Prérii", "na Prérii"},
    {1, "Прерия", "Прерии", "Прерию", "в Прерии"},
    {0/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"},
    {0/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"/*MISSING*/ , "Prairie"},
    }},
  {0x820738c1, 0, { // ""your orbs""
    {1, "Twoje Sfery", "Twoje Sfery", "Twoje Sfery", "Twoje Sfery"},
    {0/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"},
    {1, "Tvé Sféry", "Tvé Sféry", "Tvé Sféry", "Tvé Sféry"},
    {1, "Твои Сферы", "Твои Сферы", "Твои Сферы", "Твоим Сферами"},
    {0/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"},
    {0/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"/*MISSING*/ , "your orbs"},
    }},
  {0x85a9052d, 0, { // ""Ranger""
    {0, "Strażnik", "Strażnicy", "Strażnika", "Strażnikiem"},
    {0, "Korucu", "Korucular", "Korucuyu", "Korucuyla"},
    {0, "Strážce", "Strážci", "Strážce", "Strážcem"},
    {0, "Странник", "Странники", "Странника", "Странником"},
    {0, "Ranger", "Ranger", "Ranger", "Ranger"},
    {0, "Guarda", "Guardas"/*MISSING*/ , "Ranger", "Guarda"},
    }},
  {0x85fea19b, 0, { // ""Reflection""
    {2, "Odbicie", "Odbicia", "Odbicie", "w Odbiciu"},
    {0/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"},
    {3, "Odraz", "Odrazy", "Odraz", "v Odraze"},
    {2, "Отражение", "Отражения", "Отражение", "в Отражении"},
    {0/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"},
    {0/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"/*MISSING*/ , "Reflection"},
    }},
  {0x864ac83d, 2, { // ""Crossroads II""
    {2, "Skrzyżowanie II", "Skrzyżowania II", "Skrzyżowanie II", "na Skrzyżowaniu II"},
    {2, "Arayollar 2", "Arayollar 2'ler", "Arayollar 2'yi", "Arayollar 2'de"},
    {2, "Křižovatka II", "Křižovatky II", "Křižovatku II", "na Křižovatce II"},
    {2, "Перекрёсток II", "Перекрёстки II", "Перекрёсток II", "на Перекрёстке II"},
    {2, "Kreuzungen II", "Kreuzungen II", "auf den Kreuzungen II", "auf den Kreuzungen II"},
    {0/*MISSING*/ , "Crossroads II"/*MISSING*/ , "Crossroads II"/*MISSING*/ , "Crossroads II"/*MISSING*/ , "Crossroads II"},
    }},
  {0x864ac84a, 2, { // ""Crossroads IV""
    {3, "Skrzyżowanie IV", "Skrzyżowania IV", "Skrzyżowanie IV", "na Skrzyżowaniu IV"},
    {0/*MISSING*/ , "Crossroads IV"/*MISSING*/ , "Crossroads IV"/*MISSING*/ , "Crossroads IV"/*MISSING*/ , "Crossroads IV"},
    {3, "Křižovatka IV", "Křižovatky IV", "Křižovatku IV", "na Křižovatce IV"},
    {3, "Перекрёсток IV", "Перекрёстки IV", "Перекрёсток IV", "на Перекрёстке IV"},
    {1, "Kreuzungen IV", "Kreuzungen IV", "auf den Kreuzungen IV", "auf den Kreuzungen IV"},
    {0/*MISSING*/ , "Crossroads IV"/*MISSING*/ , "Crossroads IV"/*MISSING*/ , "Crossroads IV"/*MISSING*/ , "Crossroads IV"},
    }},
  {0x8818aa45, 0, { // ""Water Shard""
    {3, "Okruch Wody", "Okruchy Wody", "Okruch Wody", "Okruchem Wody"},
    {0, "Su Parçası", "Su Parçaları", "Su Parçasını", "Su Parçasıyla"},
    {3, "Úlomek Vody", "Úlomky Vody", "Úlomek Vody", "Úlomkem Vody"},
    {3, "Осколок Воды", "Осколки Воды", "Осколок Воды", "Осколком Воды"},
    {1, "Wasserscherbe", "Wasserscherben", "Wasserscherbe", "Wasserscherbe"},
    {0/*MISSING*/ , "Water Shard"/*MISSING*/ , "Water Shard"/*MISSING*/ , "Water Shard"/*MISSING*/ , "Water Shard"},
    }},
  {0x88408040, 0, { // ""Game Board""
    {1, "Plansza", "Plansza", "Planszę", "na Planszy"},
    {1, "Tahta", "Tahtalar", "Tahtayı", "Tahtada"},
    {1, "Herní deska", "Herní desky", "Herní desku", "na Herní desce"},
    {1, "Игровая Доска", "Игровые Доски", "Игровую Доску", "на Игровой Доске"},
    {2, "Spielbrett", "Spielbretter", "auf dem Spielbrett", "auf dem Spielbrett"},
    {1, "Mesa de Jogo", "Mesas de Jogos"/*MISSING*/ , "Game Board", "Mesa de Jogo"},
    }},
  {0x8867fc7b, 0, { // ""Hypersian Rug""
    {3, "Hiperski Dywan", "Hiperskie Dywany", "Hiperski Dywan", "Hiperskim Dywanem"},
    {3, "Aşkınlı Kilim", "Aşkınlı Kilimler", "Aşkınlı Kilimi", "Aşkınlı Kilimle"},
    {3, "Hyperský koberec", "Hyperské koberce", "Hyperský koberec", "Hyperským kobercem"},
    {3, "Гиперсидский ковёр", "Гиперсидские ковры", "Гиперсидский ковёр", "Гиперсидским ковром"},
    {0, "Hypersischer Teppich", "Hypersische Teppiche", "Hypersischer Teppich", "Hypersischer Teppich"},
    {0/*MISSING*/ , "Hypersian Rug"/*MISSING*/ , "Hypersian Rug"/*MISSING*/ , "Hypersian Rug"/*MISSING*/ , "Hypersian Rug"},
    }},
  {0x88bf8c73, 0, { // ""metal wall""
    {1, "metalowa ściana", "metalowe ściany", "metalową scianę", "metalową ścianą"},
    {1, "metal duvar", "metal duvarlar", "metal duvarı", "metal duvarla"},
    {1, "kovové zeď", "kovové zdi", "kovovou zeď", "kovovou zdí"},
    {1, "металлическая стена", "металлические стены", "металлическую стену", "металлической стеной"},
    {1, "Metallwand", "Metallwände", "Metallwand", "Metallwand"},
    {0/*MISSING*/ , "metal wall"/*MISSING*/ , "metal wall"/*MISSING*/ , "metal wall"/*MISSING*/ , "metal wall"},
    }},
  {0x88f7630f, 0, { // ""Haunted Woods""
    {3, "Nawiedzony Las", "Nawiedzone Lasy", "Nawiedzony Las", "w Nawiedzonym Lesie"},
    {0/*MISSING*/ , "Haunted Woods"/*MISSING*/ , "Haunted Woods"/*MISSING*/ , "Haunted Woods"/*MISSING*/ , "Haunted Woods"},
    {3, "Strašidelný hvozd", "Strašidelné hvozdy", "Strašidelný hvozd", "ve Strašidelném hvozdu"},
    {3, "Лес с привидениями", "Леса с привидениями", "Лес с привидениями", "в Лесу с привидениями"},
    {0, "Verwunschener Wald", "Verwunschene Wälder", "im Verwunschenen Wald", "im Verwunschenen Wald"},
    {0/*MISSING*/ , "Haunted Woods"/*MISSING*/ , "Haunted Woods"/*MISSING*/ , "Haunted Woods"/*MISSING*/ , "Haunted Woods"},
    }},
  {0x892e723f, 0, { // ""Ice Sapphire""
    {3, "Lodowy Szafir", "Lodowe Szafiry", "Lodowy Szafir", "Lodowym Szafirem"},
    {3, "Buz Safiri", "Buz Safirleri", "Buz Safirini", "Buz Safiriyle"},
    {3, "Ledový safír", "Ledové safíry", "Ledový safír", "Ledovým safírem"},
    {3, "Ледяной сапфир", "Ледяные сапфиры", "Ледяной сапфир", "Ледяным сапфиром"},
    {0, "Frostsaphir", "Frostsaphire", "Frostsaphir", "Frostsaphir"},
    {1, "Safira de Gelo", "Safiras de Gelo"/*MISSING*/ , "Ice Sapphire", "Safira de Gelo"},
    }},
  {0x89b196d4, 0, { // ""Dodecahedron""
    {3, "Dwunastościan", "Dwunastościany", "Dwunastościan", "Dwunastościanem"},
    {0/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"},
    {3, "Dvanáctistěn", "Dvanáctistěny", "Dvanáctistěn", "Dvanáctistěnem"},
    {3, "Додекаэдр", "Додекаэдры", "Додекаэдр", "Додекаэдром"},
    {0/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"},
    {0/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"/*MISSING*/ , "Dodecahedron"},
    }},
  {0x8b1cc002, 0, { // ""Slime Mold""
    {3, "Śluzowiec", "Śluzowce", "Śluzowiec", "Śluzowcem"},
    {0/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"},
    {1, "Hlenka", "Hlenky", "Hlenku", "Hlenkou"},
    {0, "Слизень", "Слизни", "Слизня", "Слизнем"},
    {0/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"},
    {0/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"/*MISSING*/ , "Slime Mold"},
    }},
  {0x8d8bfe35, 0, { // ""Clearing""
    {1, "Polana", "Polany", "Polanę", "na Polanie"},
    {0/*MISSING*/ , "Clearing"/*MISSING*/ , "Clearing"/*MISSING*/ , "Clearing"/*MISSING*/ , "Clearing"},
    {1, "Mýtina", "Mýtiny", "Mítinu", "na Mýtině"},
    {1, "Поляна", "Поляны", "Поляну", "на Поляне"},
    {1, "Lichtung", "Lichtungen", "auf der Lichtung", "auf der Lichtung"},
    {0/*MISSING*/ , "Clearing"/*MISSING*/ , "Clearing"/*MISSING*/ , "Clearing"/*MISSING*/ , "Clearing"},
    }},
  {0x8dd246a8, 0, { // ""Ice Shark""
    {0, "Lodowy Rekin", "Lodowe Rekiny", "Lodowego Rekina", "Lodowym Rekinem"},
    {0, "Buz Köpekbalığı", "Buz Köpekbalıkları", "Buz Köpekbalığını", "Buz Köpekbalığıyla"},
    {0, "Ledový žralok", "Ledoví žraloci", "Ledového žraloka", "Ledovým žralokem"},
    {1, "Ледяная акула", "Ледяные акулы", "Ледяную акулу", "Ледяной акулой"},
    {0, "Frosthai", "Frosthaie", "Frosthai", "Frosthai"},
    {0, "Tubarão de Gelo", "Tubarões de Gelo"/*MISSING*/ , "Ice Shark", "Tubarão de Gelo"},
    }},
  {0x8e69ee81, 0, { // ""Orb of Recall""
    {1, "Sfera Powrotu", "Sfery Powrotu", "Sferę Powrotu", "Sferą Powrotu"},
    {0/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"},
    {1, "Sféra Návratu", "Sféry Návratu", "Sféru Návratu", "Sférou Návratu"},
    {1, "Сфера Возвращения", "Сферы Возвращения", "Сферу Возвращения", "Сферой Возвращения"},
    {0/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"},
    {0/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"/*MISSING*/ , "Orb of Recall"},
    }},
  {0x8e8a4b25, 0, { // ""Halloween""
    {3, "Halloween", "Halloween", "Halloween", "w Halloween"},
    {0/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"},
    {3, "Halloween", "Halloweeny", "Halloween", "v Halloweenu"},
    {3, "Хэллоуин", "Хэллоуин", "Хэллоуин", "в Хэллоуине"},
    {0/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"},
    {0/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"/*MISSING*/ , "Halloween"},
    }},
  {0x90992e59, 0, { // ""Air Elemental""
    {0, "Powietrzny Żywiołak", "Powietrzne Żywiołaki", "Powietrznego Żywiołaka", "Powietrznym Żywiołakiem"},
    {0, "Hava Unsuru", "Hava Unsurları", "Hava Unsurunu", "Hava Unsuruyla"},
    {0, "Elementál Vzduchu", "Elementálové Vzduchu", "Elementála Vzduchu", "Elementálem Vzduchu"},
    {0, "Воздушный Элементаль", "Воздушные Элементали", "Воздушного Элементаля", "Воздушным Элементалем"},
    {0, "Luftelementar", "Luftelementare", "Luftelementar", "Luftelementar"},
    {0/*MISSING*/ , "Air Elemental"/*MISSING*/ , "Air Elemental"/*MISSING*/ , "Air Elemental"/*MISSING*/ , "Air Elemental"},
    }},
  {0x916932f5, 0, { // ""Great Wall""
    {2, "Wielka Ściana", "Wielka Ściana", "Wielką Ścianę", "na Wielkiej Ścianie"},
    {2, "Yüksek Duvar", "Yüksek Duvarlar", "Yüksek Duvarı", "Yüksek Duvarda"},
    {1, "Velká zeď", "Velké zdi", "Velkou zeď", "na Velké zdi"},
    {1, "Великая стена", "Великие стены", "Великую стену", "на Великой стене"},
    {1, "Große Mauer", "Großen Mauern", "auf der Großen Mauer", "auf der Großen Mauer"},
    {1, "Muralha", "Muralhas"/*MISSING*/ , "Great Wall", "Muralha"},
    }},
  {0x91835dc5, 0, { // ""Familiar""
    {0, "Chowaniec", "Chowańce", "Chowańca", "Chowańcem"},
    {0/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"},
    {0, "Pomocník", "Pomocníci", "Pomocníka", "Pomocníkem"},
    {0, "Знакомый", "Знакомые", "Знакомого", "Знакомым"},
    {0/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"},
    {0/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"/*MISSING*/ , "Familiar"},
    }},
  {0x921a3033, 0, { // ""Fire Fairy""
    {1, "Ognista Wiła", "Ogniste Wiły", "Ognistą Wiłę", "Ognistą Wiłą"},
    {1, "Ateş Perisi", "Ateş Perileri", "Ateş Perisini", "Ateş Perisiyle"},
    {0, "Ohnivý mužíček", "Ohniví mužíčci", "Ohnivého mužíčka", "Ohnivým mužíčkem"},
    {1, "Огненная фея", "Огненный феи", "Огненную фею", "Огненной феей"},
    {1, "Feuerfee", "Feuerfeen", "Feuerfee", "Feuerfee"},
    {1, "Fada de Fogo", "Fadas de Fogo"/*MISSING*/ , "Fire Fairy", "Fada de Fogo"},
    }},
  {0x9221100c, 0, { // ""Vineyard""
    {1, "Winnica", "Winnica", "Winnicę", "w Winnicy"},
    {1, "Şaraplık", "Şaraplıklar", "Şaraplığı", "Şaraplıkta"},
    {1, "Vinice", "Vinice", "Vinici", "na Vinici"},
    {3, "Виноградник", "Виноградники", "Виноградник", "в Винограднике"},
    {0, "Weingarten", "Weingärten", "im Weingarten", "im Weingarten"},
    {0/*MISSING*/ , "Vineyard"/*MISSING*/ , "Vineyard"/*MISSING*/ , "Vineyard"/*MISSING*/ , "Vineyard"},
    }},
  {0x9285a10b, 0, { // ""Plane of Fire""
    {1, "Strefa Ognia", "Strefy Ognia", "Strefę Ognia", "w Strefie Ognia"},
    {1, "Ateş Düzlemi", "Ateş Düzlemleri", "Ateş Düzlemini", "Ateş Düzleminde"},
    {1, "Rovina Ohně", "Roviny Ohně", "Rovinu Ohně", "v Rovině Ohně"},
    {1, "Плоскость Огня", "Плоскости Огня", "Плоскость Огня", "в Плоскости Огня"},
    {1, "Feuerebene", "Feuerebenen", "Feuerebene ", "Feuerebene "},
    {0/*MISSING*/ , "Plane of Fire"/*MISSING*/ , "Plane of Fire"/*MISSING*/ , "Plane of Fire"/*MISSING*/ , "Plane of Fire"},
    }},
  {0x930c9b5e, 0, { // ""Tentacle+Ghost""
    {1, "Macka+Duch", "Macki+Duchy", "Mackę+Ducha", "Macką+Duchem"},
    {1, "Dokungaç&Hayalet", "Dokungaç&Hayaletler", "Dokungaç&Hayaleti", "Dokungaç&Hayaletle"},
    {0, "Chapadlo+Duch", "Chapadla+Duchové", "Chapadlo+Ducha", "Chapadlem+Duchem"},
    {2, "Щупальце+Дух", "Щупальца+Духи", "Щупальце+Духа", "Щупальцем+Духом"},
    {1, "Tentakel+Geist", "Tentakel+Geister", "Tentakel+Geist", "Tentakel+Geist"},
    {0/*MISSING*/ , "Tentacle+Ghost"/*MISSING*/ , "Tentacle+Ghost"/*MISSING*/ , "Tentacle+Ghost"/*MISSING*/ , "Tentacle+Ghost"},
    }},
  {0x9319ab65, 0, { // ""Holy Grail""
    {0, "Święty Graal", "Święte Graale", "Świętego Graala", "Świętym Graalem"},
    {3, "Kutsal Kâse", "Kutsal Kâseler", "Kutsal Kâseyi", "Kutsal Kâseyle"},
    {3, "Svatý grál", "Svaté grály", "Svatý grál", "Svatým grálem"},
    {3, "Святой Грааль", "Святые Граали", "Святой Грааль", "Святым Граалем"},
    {0, "Heiliger Gral", "Heilige Grale", "Heiligen Gral", "Heiligen Gral"},
    {0/*MISSING*/ , "Holy Grail"/*MISSING*/ , "Holy Grail"/*MISSING*/ , "Holy Grail"/*MISSING*/ , "Holy Grail"},
    }},
  {0x934bf032, 0, { // ""Orb of the Bull""
    {1, "Sfera Byka", "Sfery Byka", "Sferę Byka", "Sferą Byka"},
    {0/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"},
    {1, "Sféra Býka", "Sféry Býka", "Sféru Býka", "Sférou Býka"},
    {1, "Сфера Быка", "Сферы Быка", "Сферу Быка", "Сферой Быка"},
    {0/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"},
    {0/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"/*MISSING*/ , "Orb of the Bull"},
    }},
  {0x93fdcaec, 0, { // ""Tongue""
    {3, "Język", "Języki", "Język", "Językiem"},
    {3, "Dil", "Diller", "Dili", "Dille"},
    {3, "Jazyk", "Jazyky", "Jazyk", "Jazykem"},
    {3, "Язык", "Языки", "Язык", "Языком"},
    {1, "Zunge", "Zungen", "Zunge", "Zunge"},
    {0/*MISSING*/ , "Tongue"/*MISSING*/ , "Tongue"/*MISSING*/ , "Tongue"/*MISSING*/ , "Tongue"},
    }},
  {0x946ae4a5, 0, { // ""Greater Demon""
    {0, "Wielki Demon", "Wielkie Demony", "Wielkiego Demona", "Wielkim Demonem"},
    {0, "Büyük Şeytan", "Büyük Şeytanlar", "Büyük Şeytanı", "Büyük Şeytanla"},
    {0, "Ďábel", "Ďáblové", "Ďábla", "Ďáblem"},
    {0, "Старший демон", "Старшие демоны", "Старшего демона", "Старшим демоном"},
    {0, "Mächtige Dämon", "Mächtigen Dämonen", "Mächtige Dämon", "Mächtige Dämon"},
    {0, "Demônio Maior", "Demônios Maiores"/*MISSING*/ , "Greater Demon", "Demônio Maior"},
    }},
  {0x97257147, 0, { // ""electric discharge""
    {2, "wyładowanie", "wyładowania", "wyładowanie", "wyładowaniem"},
    {2, "elektrik boşalımı", "elektrik boşalımları", "elektrik boşalımını", "elektrik boşalımıyla"},
    {3, "výboj", "výboje", "výboj", "výbojem"},
    {3, "разряд", "разряды", "разряд", "разрядом"},
    {2, "Elektrische Entladung", "Elektrische Entladungen", "Elektrische Entladung", "Elektrische Entladung"},
    {0/*MISSING*/ , "electric discharge"/*MISSING*/ , "electric discharge"/*MISSING*/ , "electric discharge"/*MISSING*/ , "electric discharge"},
    }},
  {0x97d20ec7, 0, { // ""Vizier""
    {0, "Wezyr", "Wezyrowie", "Wezyra", "Wezyrem"},
    {0, "Vezir", "Vezirler", "Veziri", "Vezirler"},
    {0, "Vezír", "Vezírové", "Vezíra", "Vezírem"},
    {0, "Визирь", "Визири", "Визиря", "Визирем"},
    {0, "Wesir", "Wesire", "Wesir", "Wesir"},
    {0/*MISSING*/ , "Vizier"/*MISSING*/ , "Vizier"/*MISSING*/ , "Vizier"/*MISSING*/ , "Vizier"},
    }},
  {0x97d3de2f, 0, { // ""Sand Worm""
    {0, "Pustynny Czerw", "Pustynne Czerwie", "Pustynnego Czerwia", "Pustynnym Czerwiem"},
    {0, "Kumkurdu", "Kumkurtları", "Kumkurdunu", "Kumkurduyla"},
    {0, "Písečný červ", "Píseční červi", "Písečného červa", "Písečným červem"},
    {0, "Пустынный червь", "Пустынные черви", "Пустынного червя", "Пустынным червём"},
    {0, "Sandwurm ", "Sandwürmer", "Sandwurm", "Sandwurm"},
    {1, "Minhoca de Areia", "Minhocas de Areia"/*MISSING*/ , "Sand Worm", "Minhoca de Areia"},
    }},
  {0x983c8c0d, 0, { // ""Hedgehog Warrior""
    {0, "Wojownik-Jeż", "Wojownicy-Jeże", "Wojownika-Jeża", "Wojownikiem-Jeżem"},
    {0, "Kirpi Savaşçı", "Kirpi Savaşçılar", "Kirpi Savaşçıyı", "Kirpi Savaşçıyla"},
    {0, "Ježčí válečník", "Ježčí válečníci", "Ježčího válečníka", "Ježčím válečníkem"},
    {0, "Ёж-воин", "Ежи-воины", "Ежа-воина", "Ежом-воином"},
    {0, "Igelkrieger", "Igelkrieger", "Igelkrieger", "Igelkrieger"},
    {0, "Guerreiro Ouriço", "Guerreiros Ouriço"/*MISSING*/ , "Hedgehog Warrior", "Guerreiro Ouriço"},
    }},
  {0x98748a3a, 0, { // ""Orb of Friendship""
    {1, "Sfera Przyjaźni", "Sfery Przyjaźni", "Sferę Przyjaźni", "Sferą Przyjaźni"},
    {1, "Dostluk Küresi", "Dostluk Küreleri", "Dostluk Küresini", "Dostluk Küresiyle"},
    {1, "Sféra Přátelství", "Sféry Přátelství", "Sféru Přátelství", "Sférou Přátelství"},
    {1, "Сфера Дружбы", "Сферы Дружбы", "Сферу Дружбы", "Сферой Дружбы"},
    {0, "Orb der Freundschaft", "Orbs der Freundschaft", "Orb der Freundschaft", "Orb der Freundschaft"},
    {0/*MISSING*/ , "Orb of Friendship"/*MISSING*/ , "Orb of Friendship"/*MISSING*/ , "Orb of Friendship"/*MISSING*/ , "Orb of Friendship"},
    }},
  {0x9a71dcab, 0, { // ""trapdoor""
    {2, "zapadnia", "zapadnie", "zapadnię", "zapadnią"},
    {2, "tuzak kapısı", "tuzak kapıları", "tuzak kapısını", "tuzak kapısıyla"},
    {2, "propadlo", "propadla", "propadlo", "propadlem"},
    {1, "ловушка", "ловушки", "ловушку", "ловушкой"},
    {1, "Falltür", "Falltüren", "Falltür", "Falltür"},
    {0/*MISSING*/ , "trapdoor"/*MISSING*/ , "trapdoor"/*MISSING*/ , "trapdoor"/*MISSING*/ , "trapdoor"},
    }},
  {0x9b74105e, 0, { // ""Shard""
    {3, "Odłamek", "Odłamki", "Odłamek", "Odłamkiem"},
    {3, "Cam Parçası", "Cam Parçaları", "Cam Parçasını", "Cam Parçasıyla"},
    {1, "Střepina", "Střepiny", "Střepinu", "Střepinou"},
    {3, "Осколок", "Осколки", "Осколок", "Осколком"},
    {1, "Scherbe", "Scherben", "Scherbe", "Scherbe"},
    {0, "Fragmento", "Fragmentos"/*MISSING*/ , "Shard", "Fragmento"},
    }},
  {0x9b741065, 0, { // ""Shark""
    {0, "Rekin", "Rekiny", "Rekina", "Rekinem"},
    {0, "Köpekbalığı", "Köpekbalıkları", "Köpekbalığını", "Köpekbalığıyla"},
    {0, "Žralok", "Žraloci", "Žraloka", "Žralokem"},
    {1, "Акула", "Акулы", "Акулу", "Акулой"},
    {0, "Hai", "Haie", "Hai", "Hai"},
    {0/*MISSING*/ , "Shark"/*MISSING*/ , "Shark"/*MISSING*/ , "Shark"/*MISSING*/ , "Shark"},
    }},
  {0x9c67ce3d, 0, { // ""Icewolf""
    {0, "Lodowy Wilk", "Lodowe Wilki", "Lodowego Wilka", "Lodowym Wilkiem"},
    {0, "Buzkurdu", "Buzkurtları", "Buzkurdunu", "Buzkurduyla"},
    {0, "Ledový vlk", "Ledoví vlci", "Ledového vlka", "Ledovým vlkem"},
    {0, "Ледяной волк", "Ледяные волки", "Ледяного волка", "Ледяным волком"},
    {0, "Eiswolf", "Eiswölfe", "Eiswolf", "Eiswolf"},
    {0, "Lobo de Gelo", "Lobos de Gelo"/*MISSING*/ , "Icewolf", "Lobo de Gelo"},
    }},
  {0x9cca513f, 0, { // ""Pikeman""
    {0, "Pikinier", "Pikinierzy", "Pikiniera", "Pikinierem"},
    {0, "Mızraklı", "Mızraklılar", "Mızraklıyı", "Mızraklıyla"},
    {0, "Kopiník", "Kopiníci", "Kopiníka", "Kopiníkem"},
    {0, "Копейщик", "Копейщики", "Копейщика", "Копейщиком"},
    {0, "Bergmann", "Bergmänner", "Bergmann", "Bergmann"},
    {0/*MISSING*/ , "Pikeman"/*MISSING*/ , "Pikeman"/*MISSING*/ , "Pikeman"/*MISSING*/ , "Pikeman"},
    }},
  {0x9d071523, 0, { // ""Amber""
    {3, "Bursztyn", "Bursztyny", "Bursztyn", "Bursztynem"},
    {3, "Kehribar", "Kehribarlar", "Kehribarı", "Kehribarla"},
    {3, "Jantar", "Jantary", "Jantar", "Jantarem"},
    {3, "Янтарь", "Янтари", "Янтарь", "Янтарём"},
    {0, "Bernstein", "Bernsteine", "Bernstein", "Bernstein"},
    {0/*MISSING*/ , "Amber"/*MISSING*/ , "Amber"/*MISSING*/ , "Amber"/*MISSING*/ , "Amber"},
    }},
  {0x9f911f3d, 0, { // ""Skeleton""
    {0, "Kościotrup", "Kościotrupy", "Kościotrupa", "Kościotrupem"},
    {0, "Kościotrup", "Kościotrupy", "Kościotrupa", "Kościotrupem"},
    {0, "Kostlivec", "Kostlivci", "Kostlivce", "Kostlivcem"},
    {0, "Скелет", "Скелеты", "Скелета", "Скелетом"},
    {0, "Skelett", "Skelette", "Skelett", "Skelett"},
    {0/*MISSING*/ , "Skeleton"/*MISSING*/ , "Skeleton"/*MISSING*/ , "Skeleton"/*MISSING*/ , "Skeleton"},
    }},
  {0xa059a1b8, 0, { // ""Cocytus""
    {0, "Kocyt", "Kocyt", "Kocytu", "w Kocycie"},
    {0, "Kokitos", "Kokitoslar", "Kokitosu", "Kokitosta"},
    {3, "Cocytus", "Cocyty", "Cocytus", "v Cocytu"},
    {3, "Коцит", "Коцит", "Коцит", "в Коците"},
    {0, "Kokytus", "Kokytus", "im Kokytus", "im Kokytus"},
    {0, "Cócito", "Cócito"/*MISSING*/ , "Cocytus", "Cócito"},
    }},
  {0xa0601d59, 0, { // ""Tortoise""
    {0, "Żółw", "Żółwie", "Żółwia", "Żółwiem"},
    {0/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"},
    {1, "Želva", "Želvy", "Želva", "Želvou"},
    {1, "Черепаха", "Черепахи", "Черепаху", "Черепахой"},
    {0/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"},
    {0/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"/*MISSING*/ , "Tortoise"},
    }},
  {0xa0fdad5a, 0, { // ""Witch Apprentice""
    {1, "Uczennica", "Uczennice", "Uczennicę", "Uczennicą"},
    {1, "Cadı Kalfası", "Cadı Kalfaları", "Cadı Kalfasını", "Cadı Kalfasıyla"},
    {1, "Učednice", "Učednice", "Učednici", "Učednicí"},
    {1, "Ведьма-ученица", "Ведьмы-ученицы", "Ведьму-ученицу", "Ведьмой-ученицей"},
    {0, "Hexenlehrling", "Hexenlehrlinge", "Hexenlehrling", "Hexenlehrling"},
    {0/*MISSING*/ , "Witch Apprentice"/*MISSING*/ , "Witch Apprentice"/*MISSING*/ , "Witch Apprentice"/*MISSING*/ , "Witch Apprentice"},
    }},
  {0xa37f5f1f, 0, { // ""Miner""
    {0, "Górnik", "Górnicy", "Górnika", "Górnikiem"},
    {0, "Madenci", "Madenciler", "Madenciyi", "Madenciyle"},
    {0, "Horník", "Horníci", "Horníka", "Horníkem"},
    {0, "Шахтёр", "Шахтёры", "Шахтёра", "Шахтёром"},
    {0, "Minenarbeiter", "Minenarbeiter", "Minenarbeiter", "Minenarbeiter"},
    {0/*MISSING*/ , "Miner"/*MISSING*/ , "Miner"/*MISSING*/ , "Miner"/*MISSING*/ , "Miner"},
    }},
  {0xa400d12d, 0, { // ""Demon Daisy""
    {1, "Czarcie Ziele", "Czarcie Ziela", "Czarcie Ziele", "Czarcim Zielem"},
    {1, "Şeytan Papatyası", "Şeytan Papatyaları", "Şeytan Papatyasını", "Şeytan Papatyasıyla"},
    {2, "Čertovo kvítko", "Čertova kvítka", "Čertovo kvítko", "Čertovým kvítkem"},
    {1, "Адская ромашка", "Адские ромашки", "Адскую ромашку", "Адской ромашкой"},
    {0, "Dämonenblümchen", "Dämonenblümchen", "Dämonenblümchen", "Dämonenblümchen"},
    {1, "Margarida Demônio", "Margaridas Demônio"/*MISSING*/ , "Demon Daisy", "Margarida Demônio"},
    }},
  {0xa550aecd, 0, { // ""Knight""
    {0, "Rycerz", "Rycerze", "Rycerza", "Rycerzem"},
    {0, "Şövalye", "Şövalyeler", "Şövalyeyi", "Şövalyeyle"},
    {0, "Rytíř", "Rytíři", "Rytíře", "Rytířem"},
    {0, "Рыцарь", "Рыцари", "Рыцаря", "Рыцарем"},
    {0, "Ritter", "Ritter", "Ritter", "Ritter"},
    {0/*MISSING*/ , "Knight"/*MISSING*/ , "Knight"/*MISSING*/ , "Knight"/*MISSING*/ , "Knight"},
    }},
  {0xa7441417, 0, { // ""frozen lake""
    {2, "zamarznięte jezioro", "zamarznięte jeziora", "zamarznięte jezioro", "zamarzniętym jeziorem"},
    {2, "Donmuş Göl", "Donmuş Göller", "Donmuş Gölü", "Donmuş Gölde"},
    {2, "zamrzlé jezero", "zamrzlá jezera", "zamrzlé jezero", "zemrzlým jezerem"},
    {2, "замёрзшее озеро", "замёрзшие озёра", "замёрзшее озеро", "замёрзшим озером"},
    {0, "Gefrorene See", "Gefrorene Seen", "Gefrorene See", "Gefrorene See"},
    {0, "lago congelado", "lagos congelados"/*MISSING*/ , "frozen lake", "lago congelado"},
    }},
  {0xa7829bdc, 0, { // ""White Dove Feather""
    {2, "Gołębie Pióro", "Gołębie Pióra", "Gołębie Pióro", "Gołębim Piórem"},
    {0/*MISSING*/ , "White Dove Feather"/*MISSING*/ , "White Dove Feather"/*MISSING*/ , "White Dove Feather"/*MISSING*/ , "White Dove Feather"},
    {2, "Holubičí pero", "Holubičí pera", "Holubičí pero", "Holubičím perem"},
    {2, "Голубиное перо", "Голубиные перья", "Голубиное перо", "Голубиным пером"},
    {1, "Feder der weißen Taube", "Federn der weißen Taube", "Feder der weißen Taube", "Feder der weißen Taube"},
    {0/*MISSING*/ , "White Dove Feather"/*MISSING*/ , "White Dove Feather"/*MISSING*/ , "White Dove Feather"/*MISSING*/ , "White Dove Feather"},
    }},
  {0xa7abedd3, 0, { // ""Flail Guard""
    {0, "Strażnik z cepem", "Strażnicy z cepami", "Strażnika z cepem", "Strażnikiem z cepem"},
    {0, "Topuz Muhafızı", "Topuz Muhafızları", "Topuz Muhafızını", "Topuz Muhafazıyla"},
    {0, "Strážce s řemdihem", "Strážci s řemdihy", "Strážce s řemdihem", "Strážcem s řemdihem"},
    {0, "Воин с цепом", "Воины с цепом", "Воина с цепом", "Воином с цепом"},
    {0, "Morgenstern-Wächter", "Morgenstern-Wächter", "Morgenstern-Wächter", "Morgenstern-Wächter"},
    {0/*MISSING*/ , "Flail Guard"/*MISSING*/ , "Flail Guard"/*MISSING*/ , "Flail Guard"/*MISSING*/ , "Flail Guard"},
    }},
  {0xa92b4827, 0, { // ""Dark Troll""
    {0, "Mroczny Troll", "Mroczne Trolle", "Mrocznego Trolla", "Mrocznym Trollem"},
    {0, "Kara Trol", "Kara Troller", "Kara Trolü", "Kara Trolle"},
    {0, "Temný troll", "Temní trollové", "Temného trolla", "Temným trollem"},
    {0, "Тёмный тролль", "Тёмные тролли", "Тёмного тролля", "Тёмным троллем"},
    {0, "Dunkle Troll", "Dunkle Trolle", "Dunklen Troll", "Dunklen Troll"},
    {0/*MISSING*/ , "Dark Troll"/*MISSING*/ , "Dark Troll"/*MISSING*/ , "Dark Troll"/*MISSING*/ , "Dark Troll"},
    }},
  {0xa94c7a1b, 0, { // ""stranded boat""
    {1, "łódka na mieliźnie", "łódki na mieliźnie", "łódkę na mieliźnie", "w łódce na mieliźnie"},
    {1, "terkedilmiş bot", "terkedilmiş botlar", "terkedilmiş botu", "terkedilmiş botla"},
    {3, "člun na mělčině", "čluny na mělčině", "člun na mělčině", "ve člunu na mělčině"},
    {1, "лодка на мели", "лодки на мели", "лодку на мели", "в лодке на мели"},
    {2, "gestrandetes Boot", "gestrandete Boote", "gestrandete Boot", "gestrandete Boot"},
    {0/*MISSING*/ , "stranded boat"/*MISSING*/ , "stranded boat"/*MISSING*/ , "stranded boat"/*MISSING*/ , "stranded boat"},
    }},
  {0xa9f08e70, 0, { // ""Sleeping Bull""
    {0, "Śpiący Byk", "Śpiące Byki", "Śpiącego Byka", "Śpiącym Bykiem"},
    {0/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"},
    {0, "Spící Býk", "Spící Býci", "Spícího Býka", "Spícím Býkem"},
    {0, "Спящий бык", "Спящие быки", "Спящего быка", "Спящим быком"},
    {0/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"},
    {0/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"/*MISSING*/ , "Sleeping Bull"},
    }},
  {0xa9f73f16, 0, { // ""rock I""
    {1, "skała I", "skały I", "skałę I", "skałą I"},
    {1, "kaya I", "kayalar  I", "kayayı I", "kayayla I"},
    {1, "skála I", "skály I", "skálu I", "skálou I"},
    {1, "скалы I", "скалы I", "скалу I", "скалой I"},
    {0, "Fels I", "Felsen I", "Fels I", "Fels I"},
    {0/*MISSING*/ , "rock I"/*MISSING*/ , "rock I"/*MISSING*/ , "rock I"/*MISSING*/ , "rock I"},
    }},
  {0xaa24f3f1, 0, { // ""Orb of Trickery""
    {1, "Sfera Sztuczek", "Sfery Sztuczek", "Sferę Sztuczek", "Sferą Sztuczek"},
    {1, "Kandırma Küresi", "Kandırma Küreleri", "Kandırma Küresini", "Kandırma Küresiyle"},
    {1, "Sféra Klamu", "Sféry Klamu", "Sféru Klamu", "Sférou Klamu"},
    {1, "Сфера Хитрости", "Сферы Хитрости", "Сферу Хитрости", "Сферой Хитрости"},
    {0, "Orb des Truges", "Orbs des Truges", "Orb des Truges", "Orb des Truges"},
    {0/*MISSING*/ , "Orb of Trickery"/*MISSING*/ , "Orb of Trickery"/*MISSING*/ , "Orb of Trickery"/*MISSING*/ , "Orb of Trickery"},
    }},
  {0xab04e11c, 0, { // ""Icy Land""
    {1, "Lodowa Kraina", "Lodowa Kraina", "Lodową Krainę", "w Lodowej Krainie"},
    {1, "Buz Diyarı", "Buz Diyarları", "Buz Diyarını", "Buz Diyarında"},
    {3, "Ledový kraj", "Ledové kraje", "Ledový kraj", "v Ledovém kraji"},
    {1, "Ледяная земля", "Ледяные земли", "Ледяную землю", "в Ледяной земле"},
    {2, "Eisiges Land", "Eisige Länder", "im Eisigen Land", "im Eisigen Land"},
    {1, "Terra Gélida", "Terras Gélidas"/*MISSING*/ , "Icy Land", "Terra Gélida"},
    }},
  {0xac9d8155, 0, { // ""Garnet""
    {3, "Granat", "Granaty", "Granat", "Granatem"},
    {3, "Nartaşı", "Nartaşları", "Nartaşını", "Nartaşıyla"},
    {3, "Granát", "Granáty", "Granát", "Granátem"},
    {3, "Гранат", "Гранаты", "Гранат", "Гранатом"},
    {0, "Granat", "Granate", "Granat", "Granat"},
    {0/*MISSING*/ , "Garnet"/*MISSING*/ , "Garnet"/*MISSING*/ , "Garnet"/*MISSING*/ , "Garnet"},
    }},
  {0xad105028, 0, { // ""Dead Ivy""
    {0, "Martwy Bluszcz", "Martwe Bluszcze", "Martwy Bluszcz", "Martwym Bluszczem"},
    {0, "Ölü Sarmaşık", "Ölü Sarmaşıklar", "Ölü Sarmaşığı", "Ölü Sarmaşıkla"},
    {0, "Mrtvý břečťan", "Mrtvé břečťany", "Mrtvý břečťan", "Mrtvým břečťanem"},
    {3, "Мёртвый плющ", "Мёртвые плющи", "Мёртвый плющ", "Мёртвым плющом"},
    {0, "Toter Efeu", "Tote Efeue", "Toten Efeu", "Toten Efeu"},
    {0/*MISSING*/ , "Dead Ivy"/*MISSING*/ , "Dead Ivy"/*MISSING*/ , "Dead Ivy"/*MISSING*/ , "Dead Ivy"},
    }},
  {0xad8cbc5b, 0, { // ""Butterfly""
    {0, "Motylek", "Motylki", "Motylka", "Motylkiem"},
    {0/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"},
    {0, "Motýlek", "Motýlci", "Motýlka", "Motýlkem"},
    {1, "Бабочка", "Бабочки", "Бабочку", "Бабочкой"},
    {0/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"},
    {0/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"/*MISSING*/ , "Butterfly"},
    }},
  {0xade3ee7a, 0, { // ""Orb of Love""
    {1, "Sfera Miłości", "Sfery Miłości", "Sferę Miłości", "Sferą Miłości"},
    {1, "Aşk Küresi", "Aşk Küreleri", "Aşk Küresini", "Aşk Küresiyle"},
    {1, "Sféra Lásky", "Sféry Lásky", "Sféru Lásky", "Sférou Lásky"},
    {1, "Сфера Любви", "Сферы Любви", "Сферу Любви", "Сферой Любви"},
    {0, "Orb der Liebe", "Orbs der Liebe", "Orb der Liebe", "Orb der Liebe"},
    {0/*MISSING*/ , "Orb of Love"/*MISSING*/ , "Orb of Love"/*MISSING*/ , "Orb of Love"/*MISSING*/ , "Orb of Love"},
    }},
  {0xaefc7f29, 0, { // ""Eternal Fire""
    {3, "Wieczny Ogień", "Wieczne Ognie", "Wieczny Ogień", "Wiecznym Ogniem"},
    {3, "Sonsuz Ateş", "Sonsuz Ateşler", "Sonsuz Ateşi", "Sonsuz Ateşle"},
    {3, "Věčný Oheň", "Věčné Ohně", "Věčný Oheň", "Věčným Ohněm"},
    {3, "Вечный огонь", "Вечные огни", "Вечный огонь", "Вечным огнём"},
    {2, "Ewiges Feuer", "Ewige Feuer", "Ewige Feuer", "Ewige Feuer"},
    {0/*MISSING*/ , "Eternal Fire"/*MISSING*/ , "Eternal Fire"/*MISSING*/ , "Eternal Fire"/*MISSING*/ , "Eternal Fire"},
    }},
  {0xaf24c8c9, 0, { // ""Orb of Shielding""
    {1, "Sfera Tarczy", "Sfery Tarczy", "Sferę Tarczy", "Sferą Tarczy"},
    {1, "Kalkan Küresi", "Kalkan Küreleri", "Kalkan Küresini", "Kalkan Küresiyle"},
    {1, "Sféra Štítu", "Sféry Štítu", "Sféru Štítu", "Sférou Štítu"},
    {1, "Сфера Щита", "Сферы Щита", "Сферу Щита", "Сферой Щита"},
    {0, "Orb der Abschirmung", "Orbs der Abschirmung", "Orb der Abschirmung", "Orb der Abschirmung"},
    {0, "Orbe de Proteção", "Orbes de Proteção"/*MISSING*/ , "Orb of Shielding", "Orbe de Proteção"},
    }},
  {0xaf37b549, 0, { // ""Green Grass""
    {1, "Zielona Trawa", "Zielone Trawy", "Zieloną Trawę", "Zieloną Trawą"},
    {0/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"},
    {1, "Zelená Tráva", "Zelené Trávy", "Zelenou Trávu", "Zelenou Trávou"},
    {1, "Зелёная трава", "Зелёные травы", "Зелёную траву", "Зелёной травой"},
    {0/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"},
    {0/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"/*MISSING*/ , "Green Grass"},
    }},
  {0xb005fb30, 0, { // ""Orb of Flash""
    {1, "Sfera Błysku", "Sfery Błysku", "Sferę Błysku", "Sferą Błysku"},
    {1, "Şimşek Küresi", "Şimşek Küreleri", "Şimşek Küresini", "Şimşek Küresiyle"},
    {1, "Sféra Záblesku", "Sféry Záblesku", "Sféru Záblesku", "Sférou Záblesku"},
    {1, "Сфера Вспышки", "Сферы Вспышки", "Сферу Вспышки", "Сферой Вспышки"},
    {0, "Orb des Blitzes", "Orbs des Blitzes", "Orb des Blitzes", "Orb des Blitzes"},
    {0, "Orbe de Clarão", "Orbes de Clarão"/*MISSING*/ , "Orb of Flash", "Orbe de Clarão"},
    }},
  {0xb170338a, 0, { // ""Storm Troll""
    {0, "Burzowy Troll", "Burzowe Trolle", "Burzowego Trolla", "Burzowym Trollem"},
    {0, "Fırtına Trolü", "Fırtına Trolleri", "Fırtına Trolünü", "Fırtına Trolüyle"},
    {0, "Bouřkotroll", "Bouřkotrollové", "Bouřkotrolla", "Bouřkotrollem"},
    {0, "Тролль бури", "Тролли бури", "Тролля бури", "Троллем бури"},
    {0, "Sturmtroll", "Sturmtrolle", "Sturmtroll", "Sturmtroll"},
    {0/*MISSING*/ , "Storm Troll"/*MISSING*/ , "Storm Troll"/*MISSING*/ , "Storm Troll"/*MISSING*/ , "Storm Troll"},
    }},
  {0xb21c83e5, 0, { // ""ancient grave""
    {3, "stary grób", "stare groby", "stary grób", "starym grobem"},
    {3, "Antik Mezar", "Antik Mezarlar", "Antik Mezarı", "Antik Mezarda"},
    {3, "starý hrob", "staré hroby", "starý hrob", "starým hrobem"},
    {1, "древняя могила", "древние могилы", "древнюю могилу", "древней могилой"},
    {0, "Antike Grab", "Antike Gräber", "Antike Grab", "Antike Grab"},
    {0, "túmulo ancião", "túmulos anciãos"/*MISSING*/ , "ancient grave", "túmulo ancião"},
    }},
  {0xb25cbbd6, 0, { // ""Metal Beast""
    {3, "Metalowy Stwór", "Metalowe Stwory", "Metalowego Stwora", "Metalowym Stworem"},
    {3, "Metal Canavar", "Metal Canavarları", "Metal Canavarını", "Metal Canavarıyla"},
    {1, "Kovová bestie", "Kovové bestie", "Kovovou bestii", "Kovovou bestií"},
    {0, "Железный зверь", "Железные звери", "Железного зверя", "Железным зверем"},
    {2, "Metallbiest", "Metallbiester", "Metallbiest", "Metallbiest"},
    {0/*MISSING*/ , "Metal Beast"/*MISSING*/ , "Metal Beast"/*MISSING*/ , "Metal Beast"/*MISSING*/ , "Metal Beast"},
    }},
  {0xb29e8496, 0, { // ""Zombie""
    {0, "Zombie", "Zombies", "Zombie", "Zombie"},
    {0, "Hortlak", "Hortlaklar", "Hortlağı", "Hortlakla"},
    {1, "Zombie", "Zombie", "Zombii", "Zombií"},
    {0, "Зомби", "Зомби", "Зомби", "Зомби"},
    {0, "Zombie", "Zombies", "Zombie", "Zombie"},
    {0, "Zumbi", "Zumbis"/*MISSING*/ , "Zombie", "Zumbi"},
    }},
  {0xb2e3b8ed, 0, { // ""Hell""
    {2, "Piekło", "Piekło", "Piekło", "w Piekle"},
    {2, "Cehennem", "Cehennemler", "Cehennemi", "Cehennemde"},
    {2, "Peklo", "Pekla", "Peklo", "v Pekle"},
    {3, "Ад", "Ад", "Ад", "в Аду"},
    {1, "Hölle", "Höllen", "in der Hölle", "in der Hölle"},
    {0, "Inferno", "Infernos"/*MISSING*/ , "Hell", "Inferno"},
    }},
  {0xb4205a72, 0, { // ""saloon wall""
    {1, "ściana salonu", "ściany salonu", "ścianę salonu", "ścianą salonu"},
    {1, "bar duvarı", "bar duvarları", "bar duvarını", "bar duvarıyla"},
    {1, "zeď saloonu", "zdi saloonu", "zeď saloonu", "zdí saloonu"},
    {1, "стена салуна", "стены салунов", "стену салуна", "стеной салуна"},
    {1, "Saloon Wand", "Saloon Wände", "Saloon Wand", "Saloon Wand"},
    {0/*MISSING*/ , "saloon wall"/*MISSING*/ , "saloon wall"/*MISSING*/ , "saloon wall"/*MISSING*/ , "saloon wall"},
    }},
  {0xb4c63743, 0, { // ""Ratling""
    {0, "Szczurak", "Szczuraki", "Szczuraka", "Szczurakiem"},
    {0/*MISSING*/ , "Ratling"/*MISSING*/ , "Ratling"/*MISSING*/ , "Ratling"/*MISSING*/ , "Ratling"},
    {0, "Krysák", "Krysáci", "Krysáka", "Krysákem"},
    {0, "Крысолюд", "Крысолюди", "Крысолюда", "Крысолюдом"},
    {0, "Rattenmensch", "Rattenmenschen", "Rattenmensch", "Rattenmensch"},
    {0/*MISSING*/ , "Ratling"/*MISSING*/ , "Ratling"/*MISSING*/ , "Ratling"/*MISSING*/ , "Ratling"},
    }},
  {0xb4fd21ea, 0, { // ""ladder""
    {1, "drabina", "drabiny", "drabinę", "drabiną"},
    {1, "merdiven", "merdivenler", "merdiveni", "merdivenle"},
    {3, "žebřík", "žebříky", "žebřík", "žebříkem"},
    {1, "лестница", "лестницы", "лестницу", "лестницей"},
    {1, "Leiter", "Leitern", "Leiter", "Leiter"},
    {0/*MISSING*/ , "ladder"/*MISSING*/ , "ladder"/*MISSING*/ , "ladder"/*MISSING*/ , "ladder"},
    }},
  {0xb52f7b64, 0, { // ""Orb of Life""
    {1, "Sfera Życia", "Sfery Życia", "Sferę Życia", "Sferą Życia"},
    {1, "Yaşam Küresi", "Yaşam Küreleri", "Yaşam Küresini", "Yaşam Küresiyle"},
    {1, "Sféra Života", "Sféry Života", "Sféru Života", "Sférou Života"},
    {1, "Сфера Жизни", "Сферы Жизни", "Сферу Жизни", "Сферой Жизни"},
    {0, "Orb des Lebens", "Orbs des Lebens", "Orb des Lebens", "Orb des Lebens"},
    {0, "Orbe de Saúde", "Orbes de Saúde"/*MISSING*/ , "Orb of Life", "Orbe de Saúde"},
    }},
  {0xb5fa80f5, 0, { // ""Rose Beauty""
    {1, "Różana Piękność", "Różane Piękności", "Różaną Piękność", "Różaną Piękność"},
    {0/*MISSING*/ , "Rose Beauty"/*MISSING*/ , "Rose Beauty"/*MISSING*/ , "Rose Beauty"/*MISSING*/ , "Rose Beauty"},
    {1, "Růžová kráska", "Růžové krásky", "Růžovou krásku", "Růžovou krásku"},
    {1, "Розовая красотка", "Розовые красотки", "Розовую красотку", "Розовой красоткой"},
    {1, "Rosenschönheit", "Rosenschönheiten", "Rosenschönheit", "Rosenschönheit"},
    {0/*MISSING*/ , "Rose Beauty"/*MISSING*/ , "Rose Beauty"/*MISSING*/ , "Rose Beauty"/*MISSING*/ , "Rose Beauty"},
    }},
  {0xb675c845, 0, { // ""Plane of Earth""
    {1, "Strefa Ziemi", "Strefy Ziemi", "Strefę Ziemi", "w Strefie Ziemi"},
    {1, "Toprak Düzlemi", "Toprak Düzlemleri", "Toprak Düzlemini", "Toprak Düzleminde"},
    {1, "Rovina Země", "Roviny Země", "Rovinu Země", "v Rovině Země"},
    {1, "Плоскость Земли", "Плоскости Земли", "Плоскость Земли", "в Плоскости Земли"},
    {1, "Erdebene ", "Erdebenen", "Erdebene", "Erdebene"},
    {0/*MISSING*/ , "Plane of Earth"/*MISSING*/ , "Plane of Earth"/*MISSING*/ , "Plane of Earth"/*MISSING*/ , "Plane of Earth"},
    }},
  {0xb6ca63c6, 0, { // ""Crystal Sage""
    {0, "Kryształowy Mędrzec", "Kryształowi Mędrcy", "Kryształowego Mędrca", "Kryształowym Mędrcem"},
    {0, "Buz Piri", "Buz Pirleri", "Buz Pirini", "Buz Piriyle"},
    {0, "Křišťálový mudrc", "Křišťáloví mudrci", "Křišťálového mudrce", "Křišťálovým mudrcem"},
    {0, "Кристальный мудрец", "Кристальные мудрецы", "Кристального мудреца", "Кристальным мудрецом"},
    {0, "Kristallweise", "Kristallweisen", "Kristallweisen", "Kristallweisen"},
    {0, "Sábio de Cristal", "Sábios de Cristal"/*MISSING*/ , "Crystal Sage", "Sábio de Cristal"},
    }},
  {0xb6db0aeb, 0, { // ""mirror wall""
    {1, "lustrzana ściana", "lustrzane ściany", "lustrzaną ścianę", "lustrzaną ścianą"},
    {0/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"},
    {1, "zrcadlová stìna", "zrcadlové stìny", "zrcadlovou stìnu", "zrcadlovou stìnou"},
    {1, "зеркальная стена", "зеркальные стены", "зеркальную стену", "зеркальной стеной"},
    {0/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"},
    {0/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"/*MISSING*/ , "mirror wall"},
    }},
  {0xb7d5f347, 0, { // ""Cult Leader""
    {0, "Przywódca Kultu", "Przywódcy Kultu", "Przywódcę Kultu", "Przywódcą Kultu"},
    {0, "Kült Lideri", "Kült Liderleri", "Kült Liderini", "Kült Lideriyle"},
    {0, "Vůdce kultu", "Vůdcové kultu", "Vůdce kultu", "Vůdcem kultu"},
    {0, "Лидер культа", "Лидеры культа", "Лидера культа", "Лидером культа"},
    {0, "Anführer des Kultes", "Anführer des Kultes", "Anführer des Kultes", "Anführer des Kultes"},
    {0/*MISSING*/ , "Cult Leader"/*MISSING*/ , "Cult Leader"/*MISSING*/ , "Cult Leader"/*MISSING*/ , "Cult Leader"},
    }},
  {0xb80cce44, 0, { // ""trunk""
    {3, "pień", "pnie", "pień", "pniem"},
    {0/*MISSING*/ , "trunk"/*MISSING*/ , "trunk"/*MISSING*/ , "trunk"/*MISSING*/ , "trunk"},
    {3, "kmen", "kmeny", "kmen", "kmenem"},
    {3, "пень", "пни", "пень", "пнём"},
    {0, "Baumstamm", "Baumstämme", "Baumstamm", "Baumstamm"},
    {0/*MISSING*/ , "trunk"/*MISSING*/ , "trunk"/*MISSING*/ , "trunk"/*MISSING*/ , "trunk"},
    }},
  {0xb819b29d, 0, { // ""Fire Witch""
    {1, "Czarownica Ognia", "Czarownice Ognia", "Czarownicę Ognia", "Czarownicą Ognia"},
    {1, "Ateş Cadısı", "Ateş Cadıları", "Ateş Cadısını", "Ateş Cadısıyla"},
    {1, "Čarodějnice Ohně", "Čarodějnice Ohně", "Čarodějnici Ohně", "Čarodějnicí Ohně"},
    {1, "Ведьма огня", "Ведьмы огня", "Ведьму огня", "Ведьмой огня"},
    {1, "Feuerhexe", "Feuerhexen", "Feuerhexe", "Feuerhexe"},
    {0/*MISSING*/ , "Fire Witch"/*MISSING*/ , "Fire Witch"/*MISSING*/ , "Fire Witch"/*MISSING*/ , "Fire Witch"},
    }},
  {0xb89f31c7, 0, { // ""sea""
    {1, "morze", "morza", "morze", "morzem"},
    {1, "deniz", "denizler", "denizi", "denizle"},
    {2, "moře", "moře", "moře", "mořem"},
    {3, "море", "моря", "море", "морем"},
    {0, "See", "Seen", "See", "See"},
    {0/*MISSING*/ , "sea"/*MISSING*/ , "sea"/*MISSING*/ , "sea"/*MISSING*/ , "sea"},
    }},
  {0xb8b94a1c, 0, { // ""Aether Witch""
    {1, "Czarownica Eteru", "Czarownice Eteru", "Czarownicę Eteru", "Czarownicą Eteru"},
    {1, "Eter Cadısı", "Eter Cadıları", "Eter Cadısını", "Eter Cadısıyla"},
    {1, "Čarodějnice Éteru", "Čarodějnice Éteru", "Čarodějnici Éteru", "Čarodějnicí Éteru"},
    {1, "Ведьма эфира", "Ведьмы эфира", "Ведьму эфира", "Ведьмой эфира"},
    {1, "Ätherhexe", "Ätherhexen", "Ätherhexe", "Ätherhexe"},
    {0/*MISSING*/ , "Aether Witch"/*MISSING*/ , "Aether Witch"/*MISSING*/ , "Aether Witch"/*MISSING*/ , "Aether Witch"},
    }},
  {0xba60cabe, 0, { // ""big tree""
    {2, "duże drzewo", "duże drzewa", "duże drzewo", "dużym drzewem"},
    {2, "Büyük Ağaç", "Büyük Ağaçlar", "Büyük Ağacı", "Büyük Ağaçla"},
    {3, "velký strom", "velké stromy", "velký strom", "velkým stromem"},
    {2, "большое дерево", "большие деревья", "большое дерево", "большим деревом"},
    {0, "Großer Baum", "Große Bäume", "Großer Baum", "Großer Baum"},
    {1, "grande árvore", "grandes árvores"/*MISSING*/ , "big tree", "grande árvore"},
    }},
  {0xbbe20fc2, 0, { // ""Mirror Spirit""
    {0, "Duch Lustra", "Duchy Lustra", "Ducha Lustra", "Duchem Lustra"},
    {0/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"},
    {0, "Zrcadlový duch", "Zrcadloví duchové", "Zrcadlového ducha", "Zrcadlovým duchem"},
    {0, "Зеркальный дух", "Зеркальные духи", "Зеркального духа", "Зеркальным духом"},
    {0/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"},
    {0/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"/*MISSING*/ , "Mirror Spirit"},
    }},
  {0xbbf36b59, 0, { // ""Orb of the Shell""
    {1, "Sfera Skorupy", "Sfery Skorupy", "Sferę Skorupy", "Sferą Skorupy"},
    {0/*MISSING*/ , "Orb of the Shell"/*MISSING*/ , "Orb of the Shell"/*MISSING*/ , "Orb of the Shell"/*MISSING*/ , "Orb of the Shell"},
    {1, "Sféra Krunýře", "Sféry Krunýře", "Sféru Krunýře", "Sférou Krunýře"},
    {1, "Сфера Панциря", "Сферы Панциря", "Сферу Панциря", "Сферой Панциря"},
    {0, "Orb des Panzers", "Orbs des Panzers", "Orb des Panzers", "Orb des Panzers"},
    {0/*MISSING*/ , "Orb of the Shell"/*MISSING*/ , "Orb of the Shell"/*MISSING*/ , "Orb of the Shell"/*MISSING*/ , "Orb of the Shell"},
    }},
  {0xbc651aba, 0, { // ""Orb of Storms""
    {1, "Sfera Burzy", "Sfery Burzy", "Sferę Burzy", "Sferą Burzy"},
    {1, "Fırtına Küresi", "Fırtına Küreleri", "Fırtına Küresini", "Fırtına Küresiyle"},
    {1, "Sféra Bouří", "Sféry Bouří", "Sféru Bouří", "Sférou Bouří"},
    {1, "Сфера Бури", "Сферы Бури", "Сферу Бури", "Сферой Бури"},
    {0, "Orb des Sturms", "Orbs des Sturms", "Orb des Sturms", "Orb des Sturms"},
    {0, "Orbe de Tempestades", "Orbes de Tempestades"/*MISSING*/ , "Orb of Storms", "Orbe de Tempestades"},
    }},
  {0xbda6c7e3, 1, { // ""Camelot""
    {3, "Camelot", "Camelot", "Camelot", "w Camelot"},
    {3, "Kamelot", "Kamelotlar", "Kamelotu", "Kamelotta"},
    {3, "Camelot", "Cameloty", "Camelotu", "v Camelotu"},
    {3, "Камелот", "Камелот", "Камелот", "в Камелоте"},
    {2, "Camelot", "Camelot", "in Camelot", "in Camelot"},
    {0/*MISSING*/ , "Camelot"/*MISSING*/ , "Camelot"/*MISSING*/ , "Camelot"/*MISSING*/ , "Camelot"},
    }},
  {0xbdebcdca, 0, { // ""weak branch""
    {1, "słaba gałąź", "słabe gałęzie", "słabą gałąź", "słabą gałęzią"},
    {0/*MISSING*/ , "weak branch"/*MISSING*/ , "weak branch"/*MISSING*/ , "weak branch"/*MISSING*/ , "weak branch"},
    {1, "tenká větev", "tenké větve", "tenkou větev", "tenkou větví"},
    {1, "тонкая ветка", "тонкие ветки", "тонкую ветку", "тонкой веткой"},
    {0, "schwacher Ast", "schwachen Äste", "schwachen Ast", "schwachen Ast"},
    {0/*MISSING*/ , "weak branch"/*MISSING*/ , "weak branch"/*MISSING*/ , "weak branch"/*MISSING*/ , "weak branch"},
    }},
  {0xbe4d2c8c, 0, { // ""stone gargoyle bridge""
    {3, "most z rzygacza", "mosty z rzygacza", "most z rzygacza", "mostem z rzygacza"},
    {3, "taş çörten köprüsü", "taş çörten köprüleri", "taş çörten köprüsünü", "taş çörten köprüsüyle"},
    {3, "most z chrliče", "mosty z chrličů", "most z chrliče", "mostem z chrliče"},
    {3, "мост с горгульей", "мосты с горгульями", "мост с горгульей", "мостом с горгульями"},
    {1, "Steingargoyle-Brücke", "Steingargoyle-Brücken", "Steingargoyle-Brücke", "Steingargoyle-Brücke"},
    {0/*MISSING*/ , "stone gargoyle bridge"/*MISSING*/ , "stone gargoyle bridge"/*MISSING*/ , "stone gargoyle bridge"/*MISSING*/ , "stone gargoyle bridge"},
    }},
  {0xc040393b, 0, { // ""barrow wall""
    {1, "ściana kurhanu", "ściany kurhanu", "ścianę kurhanu", "ścianą kurhanu"},
    {0/*MISSING*/ , "barrow wall"/*MISSING*/ , "barrow wall"/*MISSING*/ , "barrow wall"/*MISSING*/ , "barrow wall"},
    {1, "stěna mohyly", "stěny mohyly", "stěně mohyly", "stěnou mohyly"},
    {1, "стена кургана", "стены кургана", "стену кургана", "стеной кургана"},
    {1, "Hügelwand", "Hügelwände", "Hügelwand", "Hügelwand"},
    {0/*MISSING*/ , "barrow wall"/*MISSING*/ , "barrow wall"/*MISSING*/ , "barrow wall"/*MISSING*/ , "barrow wall"},
    }},
  {0xc0576c8e, 0, { // ""Ivy Root""
    {0, "Korzeń Bluszczu", "Korzenie Bluszczu", "Korzeń Bluszczu", "Korzeniem Bluszczu"},
    {0, "Sarmaşık Kökü", "Sarmaşık Kökleri", "Sarmaşık Kökünü", "Sarmaşık Köküyle"},
    {3, "Kořen břečťanu", "Kořeny břečťanu", "Kořen břečťanu", "Kořenem břečťanu"},
    {3, "Корень плюща", "Корни плюща", "Корень плюща", "Корнем плюща"},
    {1, "Efeu Wurzel", "Efeu Wurzeln", "Efeu Wurzel", "Efeu Wurzel"},
    {1, "Raiz de Hera", "Raízes de Hera"/*MISSING*/ , "Ivy Root", "Raiz de Hera"},
    }},
  {0xc0c33bfe, 0, { // ""opening plate""
    {1, "otwierająca płyta", "otwierające płyty", "otwierającą płytę", "otwierającą płytą"},
    {1, "açılan levha", "açılan levhalar", "açılan levhayı", "açılan levhayla"},
    {1, "otvírací deska", "otvírací desky", "otvírací desku", "otvírací deskou"},
    {1, "открывающая плита", "открывающие плиты", "открывающую плиту", "открывающей плитой"},
    {0, "Öffnungsmechanismus", "Öffnungsmechanismen", "Öffnungsmechanismus", "Öffnungsmechanismus"},
    {0/*MISSING*/ , "opening plate"/*MISSING*/ , "opening plate"/*MISSING*/ , "opening plate"/*MISSING*/ , "opening plate"},
    }},
  {0xc21160da, 0, { // ""Hall of Mirrors""
    {1, "Lustrzana Sala", "Lustrzane Sale", "Lustzaną Salę", "w Lustrzanej Sali"},
    {0/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"},
    {1, "Zrcadlový sál", "Zrcadlové sály", "Zrcadlový sál", "v Zrcadlovém sále"},
    {3, "Зеркальный зал", "Зеркальные залы", "Зеркальный зал", "в Зеркальном зале"},
    {0/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"},
    {0/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"/*MISSING*/ , "Hall of Mirrors"},
    }},
  {0xc3512f3c, 0, { // ""Orb of Beauty""
    {1, "Sfera Piękna", "Sfery Piękna", "Sferę Piękna", "Sferą Piękna"},
    {0/*MISSING*/ , "Orb of Beauty"/*MISSING*/ , "Orb of Beauty"/*MISSING*/ , "Orb of Beauty"/*MISSING*/ , "Orb of Beauty"},
    {1, "Sféra Krásy", "Sféry Krásy", "Sféru Krásy", "Sférou Krásy"},
    {1, "Сфера красоты", "Сферы красоты", "Сферу красоты", "Сферой красоты"},
    {0, "Orb der Schönheit", "Orbs der Schönheit", "Orb der Schönheit", "Orb der Schönheit"},
    {0/*MISSING*/ , "Orb of Beauty"/*MISSING*/ , "Orb of Beauty"/*MISSING*/ , "Orb of Beauty"/*MISSING*/ , "Orb of Beauty"},
    }},
  {0xc66ea1de, 0, { // ""Parrot""
    {1, "Papuga", "Papugi", "Papugę", "Papugą"},
    {1, "Papağan", "Papağanlar", "Papağanı", "Papağanla"},
    {0, "Papoušek", "Papoušci", "Papouška", "Papouškem"},
    {0, "Попугай", "Попугаи", "Попугая", "Попугаем"},
    {0, "Papagei", "Papageien", "Papagei", "Papagei"},
    {0/*MISSING*/ , "Parrot"/*MISSING*/ , "Parrot"/*MISSING*/ , "Parrot"/*MISSING*/ , "Parrot"},
    }},
  {0xc6e66c93, 0, { // ""Vampire Bat""
    {0, "Wampir", "Wampiry", "Wampira", "Wampirem"},
    {0/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"},
    {0, "Vampýr", "Vampýři", "Vampýra", "Vampýrem"},
    {0, "Вампир", "Вампиры", "Вампира", "Вампиром"},
    {0/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"},
    {0/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"/*MISSING*/ , "Vampire Bat"},
    }},
  {0xc701c53b, 0, { // ""Mimic""
    {0, "Mimik", "Mimiki", "Mimika", "Mimikiem"},
    {0/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"},
    {0, "Mimik", "Mimikové", "Mimika", "Mimikem"},
    {0, "Мимик", "Мимики", "Мимика", "Мимиком"},
    {0/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"},
    {0/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"/*MISSING*/ , "Mimic"},
    }},
  {0xc875735e, 2, { // ""Kraken Depths""
    {1, "Głębiny Krakenów", "Głębiny Krakenów", "Głębiny Krakenów", "w Głębinach Krakenów"},
    {0/*MISSING*/ , "Kraken Depths"/*MISSING*/ , "Kraken Depths"/*MISSING*/ , "Kraken Depths"/*MISSING*/ , "Kraken Depths"},
    {1, "Krakení hlubiny", "Krakení hlubiny", "Krakení hlubiny", "v Krakeních hlubinách"},
    {1, "Глубины кракенов", "Глубины кракенов", "Глубины кракенов", "в Глубинах кракенов"},
    {1, "Krakentiefen", "Krakentiefen", "in den Krakentiefen", "in den Krakentiefen"},
    {0/*MISSING*/ , "Kraken Depths"/*MISSING*/ , "Kraken Depths"/*MISSING*/ , "Kraken Depths"/*MISSING*/ , "Kraken Depths"},
    }},
  {0xc8c66d41, 0, { // ""Coral""
    {3, "Koral", "Korale", "Koral", "Koralem"},
    {0/*MISSING*/ , "Coral"/*MISSING*/ , "Coral"/*MISSING*/ , "Coral"/*MISSING*/ , "Coral"},
    {3, "Korál", "Korály", "Korál", "Korálem"},
    {3, "Коралл", "Кораллы", "Коралл", "Кораллом"},
    {3, "Koralle", "Korallen", "Koralle", "Koralle"},
    {0/*MISSING*/ , "Coral"/*MISSING*/ , "Coral"/*MISSING*/ , "Coral"/*MISSING*/ , "Coral"},
    }},
  {0xc93e7fe9, 0, { // ""giant rug""
    {3, "wielki dywan", "wielkie dywany", "wielki dywan", "wielkim dywanem"},
    {3, "devasa kilim", "devasa kilimler", "devasa kilimi", "devasa kilimle"},
    {3, "obří koberec", "obří koberce", "obří koberec", "obřím kobercem"},
    {3, "огромный ковёр", "огромные ковры", "огромный ковёр", "огромным ковром"},
    {3, "Riesiger Teppich", "Riesige Teppiche", "Riesigen Teppich", "Riesigen Teppich"},
    {0/*MISSING*/ , "giant rug"/*MISSING*/ , "giant rug"/*MISSING*/ , "giant rug"/*MISSING*/ , "giant rug"},
    }},
  {0xc956dc2a, 0, { // ""Trollheim""
    {3, "Trollheim", "Trollheim", "Trollheim", "w Trollheim"},
    {0/*MISSING*/ , "Trollheim"/*MISSING*/ , "Trollheim"/*MISSING*/ , "Trollheim"/*MISSING*/ , "Trollheim"},
    {3, "Trollheim", "Trollheimy", "Trollheim", "v Trollheimu"},
    {3, "Тролльхейм", "Тролльхеймы", "Тролльхейм", "в Тролльхейме"},
    {3, "Trollheim", "Trollheim", "Trollheim", "Trollheim"},
    {0/*MISSING*/ , "Trollheim"/*MISSING*/ , "Trollheim"/*MISSING*/ , "Trollheim"/*MISSING*/ , "Trollheim"},
    }},
  {0xc9a76da2, 0, { // ""Phoenix Feather""
    {2, "Pióro Feniksa", "Pióra Feniksa", "Pióro Feniksa", "Piórem Feniksa"},
    {2, "Anka Kuşu Tüyü", "Anka Kuşu Tüyleri", "Anka Kuşu Tüyünü", "Anka Kuşu Tüyüyle"},
    {2, "Fénixovo pírko", "Fénixova pírka", "Fénixovo pírko", "Fénixovým pírkem"},
    {2, "Перо феникса", "Перья феникса", "Перо феникса", "Пером феникса"},
    {1, "Phönixfeder", "Phönixfedern", "Phönixfeder", "Phönixfeder"},
    {1, "Pena de Fênix", "Penas de Fênix"/*MISSING*/ , "Phoenix Feather", "Pena de Fênix"},
    }},
  {0xc9e27f87, 0, { // ""Reptile floor""
    {1, "Jaszczurza podłoga", "Jaszczurze podłogi", "Jaszczurzą podłogę", "Jaszczurzą podłogą"},
    {0/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"},
    {1, "Ještěrčí podlaha", "Ještěrščí podlahy", "Ještěrčí podlahu", "Ještěrčí podlahou"},
    {3, "Ящеричный пол", "Ящеричные полы", "Ящеричный пол", "Ящеричным полом"},
    {0/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"},
    {0/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"/*MISSING*/ , "Reptile floor"},
    }},
  {0xca5028f9, 0, { // ""Hyperstone""
    {3, "Hiperkamień", "Hiperkamienie", "Hiperkamień", "Hiperkamieniem"},
    {3, "Aşkıntaş", "Aşkıntaşlar", "Aşkıntaşı", "Aşkıntaşla"},
    {3, "Hyperkam", "Hyperkamy", "Hyperkam", "Hyperkamem"},
    {3, "Гиперкамень", "Гиперкамни", "Гиперкамень", "Гиперкамнем"},
    {0, "Hyperstein", "Hypersteine", "Hyperstein", "Hyperstein"},
    {1, "Pedra do Ímpeto", "Pedras do Ímpeto"/*MISSING*/ , "Hyperstone", "Pedra do Ímpeto"},
    }},
  {0xcad7fc4d, 1, { // ""R'Lyeh""
    {2, "R'Lyeh", "R'Lyeh", "R'Lyeh", "w R'Lyeh"},
    {2, "R'Lyeh", "R'Lyeh", "R'Lyeh", "R'Lyeh"},
    {2, "R'Lyeh", "R'Lyeh", "R'Lyeh", "v R'Lyeh"},
    {3, "Р'Льех", "Р'Льех", "Р'Льех", "в Р'Льехе"},
    {2, "R'Lyeh", "R'Lyeh", "in R'Lyeh", "in R'Lyeh"},
    {0, "R'Lyeh", "R'Lyeh"/*MISSING*/ , "R'Lyeh", "R'Lyeh"},
    }},
  {0xcb051b8b, 0, { // ""Running Dog""
    {0, "Biegnący Pies", "Biegnące Psy", "Biegnącego Psa", "Biegnącym Psem"},
    {0, "Koşan Köpek", "Koşan Köpekler", "Koşan Köpeği", "Koşan Köpekle"},
    {0, "Běžící pes", "Běžící psi", "Běžícího psa", "Běžícím psem"},
    {1, "Бегущая собака", "Бегущие собаки", "Бегущую собаку", "Бегущей собакой"},
    {0, "Laufhund", "Laufhunde", "Laufhund", "Laufhund"},
    {0, "Cachorro Correndo", "Cachorros Correndo"/*MISSING*/ , "Running Dog", "Cachorro Correndo"},
    }},
  {0xcba185cf, 0, { // ""stone statue""
    {3, "kamienny pomnik", "kamienne pomniki", "kamienny pomnik", "kamiennym pomnikiem"},
    {0/*MISSING*/ , "stone statue"/*MISSING*/ , "stone statue"/*MISSING*/ , "stone statue"/*MISSING*/ , "stone statue"},
    {1, "kamenná socha", "kamenné sochy", "kamennou sochu", "kamennou sochou"},
    {1, "каменная статуя", "каменные статуи", "каменную статую", "каменной статуей"},
    {1, "Steinstatue", "Steinstatuen", "Steinstatue", "Steinstatue"},
    {0/*MISSING*/ , "stone statue"/*MISSING*/ , "stone statue"/*MISSING*/ , "stone statue"/*MISSING*/ , "stone statue"},
    }},
  {0xcbe8c992, 0, { // ""Dry Forest""
    {0, "Puszcza", "Puszcza", "Puszczę", "w Puszczy"},
    {0, "Kara Orman", "Kara Ormanlar", "Kara Ormanı", "Kara Ormanda"},
    {3, "Suchý hvozd", "Suché hvozdy", "Suchý hvozd", "v Suchém hvozdu"},
    {3, "Сухой Лес", "Сухие Леса", "Сухой Лес", "в Сухом Лесу"},
    {0, "Trockenwald", "Trockenwälder", "im Trockenwald", "im Trockenwald"},
    {1, "Floresta Seca", "Florestas Secas"/*MISSING*/ , "Dry Forest", "Floresta Seca"},
    }},
  {0xcc1be799, 0, { // ""open gate""
    {1, "otwarta brama", "otwarte bramy", "otwartą bramę", "otwartą bramą"},
    {1, "açık kapı", "açık kapılar", "açık kapıyı", "açık kapıyla"},
    {1, "otevřená brána", "otevřené brány", "otevřenou bránu", "otevřenou bránou"},
    {3, "открытый проход", "открытые проходы", "открытый проход", "открытым проходом"},
    {2, "Offenes Tor", "Offene Tor", "Offenes Tor", "Offenes Tor"},
    {0/*MISSING*/ , "open gate"/*MISSING*/ , "open gate"/*MISSING*/ , "open gate"/*MISSING*/ , "open gate"},
    }},
  {0xcd5954e2, 0, { // ""Treat""
    {3, "Cukierek", "Cukierki", "Cukierka", "Cukierkiem"},
    {0/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"},
    {1, "Pochoutka", "Pochoutky", "Pochutku", "Pochoutkou"},
    {1, "конфета", "конфеты", "конфету", "конфетой"},
    {0/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"},
    {0/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"/*MISSING*/ , "Treat"},
    }},
  {0xcdca3c6e, 0, { // ""Fat Guard""
    {0, "Gruby Strażnik", "Grubi Strażnicy", "Grubego Strażnika", "Grubym Strażnikiem"},
    {0, "Şişman Muhafız", "Şişman Muhafızlar", "Şişman Muhafızı", "Şişman Muhafızla"},
    {0, "Tlustý strážce", "Tlustí strážci", "Tlustého strážce", "Tlustým strážcem"},
    {0, "Тучный стражник", "Тучные стражники", "Тучного стражника", "Тучным стражником"},
    {1, "Dicke Wache", "Dicken Wachen", "Dicke Wache", "Dicke Wache"},
    {0/*MISSING*/ , "Fat Guard"/*MISSING*/ , "Fat Guard"/*MISSING*/ , "Fat Guard"/*MISSING*/ , "Fat Guard"},
    }},
  {0xceab75b0, 0, { // ""Palace""
    {3, "Pałac", "Pałace", "Pałac", "w Pałacu"},
    {3, "Saray", "Saraylar", "Sarayı", "Sarayda"},
    {3, "Palác", "Paláce", "Palác", "v Paláci"},
    {3, "Дворец", "Дворцы", "Дворец", "в Дворце"},
    {0, "Palast", "Paläste", "im Palast", "im Palast"},
    {0/*MISSING*/ , "Palace"/*MISSING*/ , "Palace"/*MISSING*/ , "Palace"/*MISSING*/ , "Palace"},
    }},
  {0xcf0477ad, 0, { // ""Warped Coast""
    {3, "Zakrzywiony Brzeg", "Zakrzywiony Brzegi", "Zakrzywiony Brzeg", "na Zakrzywionym Brzegu"},
    {0/*MISSING*/ , "Warped Coast"/*MISSING*/ , "Warped Coast"/*MISSING*/ , "Warped Coast"/*MISSING*/ , "Warped Coast"},
    {2, "Pokřivené pobřeží", "Pokřivená pobřeží", "Pokřivené pobřeží", "na Pokřiveném pobřeží"},
    {3, "Искривлённый берег", "Искривлённые берега", "Искривлённый берег", "на Искривлённом берегу"},
    {1, "Verzerrte Küste", "Verzerrte Küsten", "an der Verzerrten Küste", "an der Verzerrten Küste"},
    {0/*MISSING*/ , "Warped Coast"/*MISSING*/ , "Warped Coast"/*MISSING*/ , "Warped Coast"/*MISSING*/ , "Warped Coast"},
    }},
  {0xcf0a2594, 0, { // ""Outlaw""
    {0, "Bandyta", "Bandyci", "Bandytę", "Bandytą"},
    {0, "Haydut", "Haydutlar", "Haydutu", "Haydutla"},
    {0, "Psanec", "Psanci", "Psance", "Psancem"},
    {0, "Бандит", "Бандиты", "Бандита", "Бандитом"},
    {0, "Bandit", "Banditen", "Bandit", "Bandit"},
    {0/*MISSING*/ , "Outlaw"/*MISSING*/ , "Outlaw"/*MISSING*/ , "Outlaw"/*MISSING*/ , "Outlaw"},
    }},
  {0xcf36b6e8, 2, { // ""Elemental Planes""
    {1, "Strefa Żywiołów", "Strefy Żywiołów", "Strefę Żywiołów", "w Strefie Żywiołów"},
    {1, "Unsur Düzlemleri", "Unsur Düzlemleri", "Unsur Düzlemlerini", "Unsur Düzlemlerinde"},
    {1, "Elementální roviny", "Elementální roviny", "Elementální roviny", "v Elementálních rovinách"},
    {1, "Плоскость Стихий", "Плоскости Стихий", "Плоскость Стихий", "в Плоскости Стихий"},
    {1, "Elementare Ebene", "Elementare Ebene", "auf der Elementaren Ebene", "auf der Elementaren Ebene"},
    {0/*MISSING*/ , "Elemental Planes"/*MISSING*/ , "Elemental Planes"/*MISSING*/ , "Elemental Planes"/*MISSING*/ , "Elemental Planes"},
    }},
  {0xd0300026, 0, { // ""Fire""
    {3, "Ogień", "Ognie", "Ogień", "Ogniem"},
    {3, "Ateş", "Ateşler", "Ateşi", "Ateşle"},
    {3, "Oheň", "Ohně", "Oheň", "Ohněm"},
    {3, "Огонь", "Огни", "Огонь", "Огнём"},
    {2, "Feuer", "Feuer", "Feuer", "Feuer"},
    {0/*MISSING*/ , "Fire"/*MISSING*/ , "Fire"/*MISSING*/ , "Fire"/*MISSING*/ , "Fire"},
    }},
  {0xd0fd9d9a, 0, { // ""Rogue""
    {0, "Cwaniak", "Cwaniaki", "Cwaniaka", "Cwaniakiem"},
    {0, "Seyyah", "Seyyahlar", "Seyyahı", "Seyyahla"},
    {0, "Chytrák", "Chytráci", "Chytráka", "Chytrákem"},
    {0, "Разбойник", "Разбойники", "Разбойника", "Разбойником"},
    {0, "Schurke", "Schurken", "Schurken", "Schurken"},
    {0/*MISSING*/ , "Rogue"/*MISSING*/ , "Rogue"/*MISSING*/ , "Rogue"/*MISSING*/ , "Rogue"},
    }},
  {0xd1cca586, 0, { // ""dead wall""
    {1, "martwa ściana", "martwe ściany", "martwą ścianę", "martwą ścianą"},
    {1, "ölü duvar", "ölü duvarlar", "ölü duvarı", "ölü duvarda"},
    {1, "mrtvá zeď", "mrtvé zdi", "mrtvou zeď", "mrtvou zdí"},
    {1, "мёртвая стена", "мёртвые стены", "мёртвую стену", "мёртвой стеной"},
    {1, "tote Wand", "tote Wände", "tote Wand", "tote Wand"},
    {0/*MISSING*/ , "dead wall"/*MISSING*/ , "dead wall"/*MISSING*/ , "dead wall"/*MISSING*/ , "dead wall"},
    }},
  {0xd21228c2, 0, { // ""dead rock troll""
    {0, "martwy skalny troll", "martwe skalne trolle", "martwego skalnego trolla", "martwym skalnym trollem"},
    {0, "ölü mağara trolü", "ölü mağara trolleri", "ölü mağara trolünü", "ölü mağara trolüyle"},
    {0, "mrtvý Šutroll", "mrtví Šutrollové", "mrtvého Šutrolla", "mrtvým Šutrollem"},
    {0, "мёртвый пещерный тролль", "мёртвые пещерные тролли", "мёртвого пещерного тролля", "мёртвым пещерным троллем"},
    {0, "toter Felsentroll", "tote Felsentrolle", "toten Felsentroll", "toten Felsentroll"},
    {0/*MISSING*/ , "dead rock troll"/*MISSING*/ , "dead rock troll"/*MISSING*/ , "dead rock troll"/*MISSING*/ , "dead rock troll"},
    }},
  {0xd234e251, 0, { // ""living wall""
    {1, "żywa ściana", "żywe ściany", "żywą ścianę", "żywą ścianą"},
    {1, "Canlı Duvar", "Canlı Duvarlar", "Canlı Duvarı", "Canlı Duvarda"},
    {1, "živoucí zeď", "živoucí zdi", "živoucí zeď", "živoucí zdí"},
    {1, "живая стена", "живые стены", "живую стену", "живой стеной"},
    {1, "Lebende Wand", "Lebende Wände", "Lebende Wand", "Lebende Wand"},
    {1, "parede viva", "paredes vivas"/*MISSING*/ , "living wall", "parede viva"},
    }},
  {0xd3c9c3a1, 0, { // ""temporary wall""
    {1, "tymczasowa ściana", "tymczasowe ściany", "tymczasową ścianę", "tymczasową ścianą"},
    {1, "geçici duvar", "geçici duvarlar", "geçici duvarı", "geçici duvarla"},
    {1, "dočasná zeď", "dočasné zdi", "dočasnou zeď", "dočasnou zdí"},
    {1, "временная стена", "временные стены", "временную стену", "временной стеной"},
    {1, "temporäre Wand", "temporäre Wände", "temporäre Wand", "temporäre Wand"},
    {0/*MISSING*/ , "temporary wall"/*MISSING*/ , "temporary wall"/*MISSING*/ , "temporary wall"/*MISSING*/ , "temporary wall"},
    }},
  {0xd43d87d8, 0, { // ""Gold""
    {2, "Złoto", "Złota", "Złoto", "Złotem"},
    {2, "Altın", "Altınlar", "Altını", "Altınla"},
    {3, "Zlatý valoun", "Zlaté valouny", "Zlatý valoun", "Zlatým valounem"},
    {2, "Золото", "Золото", "Золото", "Золотом"},
    {2, "Gold", "Gold", "Gold", "Gold"},
    {0, "Ouro", "Ouros"/*MISSING*/ , "Gold", "Ouro"},
    }},
  {0xd47a1579, 0, { // ""no wall""
    {3, "brak ściany", "braki ściany", "braku ściany", "brakiem ściany"},
    {0/*MISSING*/ , "no wall"/*MISSING*/ , "no wall"/*MISSING*/ , "no wall"/*MISSING*/ , "no wall"},
    {1, "žádné stěna", "žádné stěny", "žádnou stěnu", "žádnou stěnou"},
    {3, "нет стены", "нет стены", "нет стены", "нет стены"},
    {1, "keine Wand", "keine Wände", "keine Wand", "keine Wand"},
    {0/*MISSING*/ , "no wall"/*MISSING*/ , "no wall"/*MISSING*/ , "no wall"/*MISSING*/ , "no wall"},
    }},
  {0xd491de5e, 0, { // ""Pearl""
    {1, "Perła", "Perły", "Perłę", "Perłą"},
    {1, "İnci", "İnciler", "İnciyi", "İncide"},
    {1, "Perla", "Perly", "Perlu", "Perlou"},
    {1, "Жемчужина", "Жемчужины", "Жемчужину", "Жемчужиной"},
    {1, "Perle", "Perlen", "Perle", "Perle"},
    {0/*MISSING*/ , "Pearl"/*MISSING*/ , "Pearl"/*MISSING*/ , "Pearl"/*MISSING*/ , "Pearl"},
    }},
  {0xd6128c30, 0, { // ""Windy Plains""
    {1, "Wietrzna Równina", "Wietrzne Równiny", "Wietrzną Równinę", "na Wietrznej Równinie"},
    {0/*MISSING*/ , "Windy Plains"/*MISSING*/ , "Windy Plains"/*MISSING*/ , "Windy Plains"/*MISSING*/ , "Windy Plains"},
    {1, "Větrná pláň", "Větrné pláně", "Větrnou pláň", "na Větrné pláni"},
    {1, "Ветреная равнина", "Ветреные равнины", "Ветреную равнину", "на Ветреной равнине"},
    {1, "Windige Ebenen", "Windige Ebenen", "in den Windigen Ebenen", "in den Windigen Ebenen"},
    {0/*MISSING*/ , "Windy Plains"/*MISSING*/ , "Windy Plains"/*MISSING*/ , "Windy Plains"/*MISSING*/ , "Windy Plains"},
    }},
  {0xd614f371, 0, { // ""Sand Worm Tail""
    {0, "Ogon Czerwia", "Ogony Czerwia", "Ogon Czerwia", "Ogonem Czerwia"},
    {0, "Kumkurdu Kuyruğu", "Kumkurdu Kuyrukları", "Kumkurdu Kuyruğunu", "Kumkurdu Kuyruğuyla"},
    {3, "Červí ocas", "Červí ocasy", "Červí ocas", "Červím ocasem"},
    {0, "Хвост червя", "Хвосты червя", "Хвоста червя", "Хвостом червя"},
    {0, "Sandwurm Schweif", "Sandwurm Schweife", "Sandwurm Schweif", "Sandwurm Schweif"},
    {1, "Cauda de Minhoca de Areia", "Caudas de Minhoca de Areia"/*MISSING*/ , "Sand Worm Tail", "Minhoca de Areia"},
    }},
  {0xd6651cb3, 0, { // ""Dragon""
    {0, "Smok", "Smoki", "Smoka", "Smokiem"},
    {0/*MISSING*/ , "Dragon"/*MISSING*/ , "Dragon"/*MISSING*/ , "Dragon"/*MISSING*/ , "Dragon"},
    {0, "Drak", "Draci", "Draka", "Drakem"},
    {0, "Дракон", "Драконы", "Дракона", "Драконом"},
    {0, "Drachen", "Drachen", "Drachen", "Drachen"},
    {0/*MISSING*/ , "Dragon"/*MISSING*/ , "Dragon"/*MISSING*/ , "Dragon"/*MISSING*/ , "Dragon"},
    }},
  {0xd6acd1c8, 0, { // ""none""
    {3, "nic", "nic", "niczego", "niczym"},
    {3, "hiç", "hiçler", "hiçi", "hiçle"},
    {2, "nic", "nic", "nic", "ničím"},
    {2, "ничто", "ничто", "ничего", "ничем"},
    {3, "keines", "keines", "keines", "keines"},
    {0, "nada", "nada"/*MISSING*/ , "none", "nada"},
    }},
  {0xd6eceeba, 0, { // ""Orb of the Sword II""
    {1, "Sfera Ostrza II", "Sfery Ostrza II", "Sferę Ostrza II", "Sferą Ostrza II"},
    {0/*MISSING*/ , "Orb of the Sword II"/*MISSING*/ , "Orb of the Sword II"/*MISSING*/ , "Orb of the Sword II"/*MISSING*/ , "Orb of the Sword II"},
    {1, "Sféra Meče II", "Sféry Meče II", "Sféru Meče II", "Sférou Meče II"},
    {1, "Сфера Меча II", "Сферы Меча II", "Сферу Меча II", "Сферой Меча II"},
    {0, "Orb des Schwertes II", "Orbs des Schwertes II", "Orb des Schwertes II", "Orb des Schwertes II"},
    {0/*MISSING*/ , "Orb of the Sword II"/*MISSING*/ , "Orb of the Sword II"/*MISSING*/ , "Orb of the Sword II"/*MISSING*/ , "Orb of the Sword II"},
    }},
  {0xd7fe2ac7, 0, { // ""Wind Crow""
    {1, "Wietrzna Wrona", "Wietrzne Wrony", "Wietrzną Wronę", "Wietrzną Wroną"},
    {0/*MISSING*/ , "Wind Crow"/*MISSING*/ , "Wind Crow"/*MISSING*/ , "Wind Crow"/*MISSING*/ , "Wind Crow"},
    {1, "Větrná vrána", "Větrné vrány", "Větrnou vránu", "Větrnou vránou"},
    {1, "Ветряная ворона", "Ветряные вороны", "Ветряную ворону", "Ветряной вороной"},
    {1, "Windkrähe", "Windkrähen", "Windkrähe", "Windkrähe"},
    {0/*MISSING*/ , "Wind Crow"/*MISSING*/ , "Wind Crow"/*MISSING*/ , "Wind Crow"/*MISSING*/ , "Wind Crow"},
    }},
  {0xd7fe825d, 0, { // ""Dormant Ivy""
    {0, "Uśpiony Bluszcz", "Uśpione Bluszcze", "Uśpiony Bluszcz", "Uśpionym Bluszczem"},
    {0, "Uyuyan Sarmaşık", "Uyuyan Sarmaşıklar", "Uyuyan Sarmaşığı", "Uyuyan Sarmaşıkla"},
    {0, "Neaktivní břečťan", "Neaktivní břečťany", "Neaktivní břečťan", "Neaktivním břečťanem"},
    {3, "Спящий плющ", "Спящие плющи", "Спящий плющ", "Спящим плющом"},
    {0, "Ruhender Efeu", "Ruhende Efeue", "Ruhenden Efeu", "Ruhenden Efeu"},
    {1, "Hera Dormente", "Heras Dormentes"/*MISSING*/ , "Dormant Ivy", "Hera Dormente"},
    }},
  {0xd8b8981e, 0, { // ""Elemental Gem""
    {3, "Kamień Żywiołów", "Kamienie Żywiołów", "Kamień Żywiołów", "Kamieniem Żywiołów"},
    {0, "Unsurtaşı", "Unsurtaşları", "Unsurtaşını", "Unsurtaşıyla"},
    {3, "Elementální drahokam", "Elementální drahokamy", "Elementální drahokam", "Elementálním drahokamem"},
    {3, "Камень стихий", "Камни стихий", "Камень стихий", "Камнем стихий"},
    {2, "Elementjuwel", "Elementjuwelen", "Elementjuwel", "Elementjuwel"},
    {0/*MISSING*/ , "Elemental Gem"/*MISSING*/ , "Elemental Gem"/*MISSING*/ , "Elemental Gem"/*MISSING*/ , "Elemental Gem"},
    }},
  {0xda56e219, 0, { // ""Servant""
    {0, "Sługa", "Słudzy", "Sługę", "Sługą"},
    {0, "Hizmetkâr", "Hizmetkârlar", "Hizmetkârı", "Hizmetkârla"},
    {0, "Služebník", "Služebníci", "Služebníka", "Služebníkem"},
    {0, "Слуга", "Слуги", "Слугу", "Слугой"},
    {0, "Diener", "Diener", "Diener", "Diener"},
    {0/*MISSING*/ , "Servant"/*MISSING*/ , "Servant"/*MISSING*/ , "Servant"/*MISSING*/ , "Servant"},
    }},
  {0xdbb078c8, 0, { // ""Powerstone""
    {0, "Kamień Mocy", "Kamienie Mocy", "Kamień Mocy", "Kamieniem Mocy"},
    {0, "Güç Taşı", "Güç Taşları", "Güç Taşını", "Güç Taşıyla"},
    {3, "Mocikam", "Mocikamy", "Mocikam", "Mocikamem"},
    {3, "Камень силы", "Камни силы", "Камень силы", "Камнем силы"},
    {0, "Machtstein", "Machtsteine", "Machtstein", "Machtstein"},
    {0/*MISSING*/ , "Powerstone"/*MISSING*/ , "Powerstone"/*MISSING*/ , "Powerstone"/*MISSING*/ , "Powerstone"},
    }},
  {0xdc1bb91a, 0, { // ""Orb of Earth""
    {1, "Sfera Ziemi", "Sfery Ziemi", "Sferę Ziemi", "Sferą Ziemi"},
    {1, "Toprak Küresi", "Toprak Küreleri", "Toprak Küresini", "Toprak Küresiyle"},
    {1, "Sféra Země", "Sféry Země", "Sféru Země", "Sférou Země"},
    {1, "Сфера Земли", "Сферы Земли", "Сферу Земли", "Сферой Земли"},
    {0, "Orb der Erde", "Orbs der Erde", "Orb der Erde", "Orb der Erde"},
    {0/*MISSING*/ , "Orb of Earth"/*MISSING*/ , "Orb of Earth"/*MISSING*/ , "Orb of Earth"/*MISSING*/ , "Orb of Earth"},
    }},
  {0xdc6e66c3, 0, { // ""Tentacle (withdrawing)""
    {1, "Macka (uciekająca)", "Macki (uciekające)", "Mackę (uciekającą)", "Macką (uciekającą)"},
    {1, "Dokunaç (geri çekiliyor)", "Dokunaçlar (geri çekiliyor)", "Dokunacı (geri çekiliyor)", "Dokunaçla (geri çekiliyor)"},
    {1, "Chapadlo (ustupující)", "Chapadla (ustupující)", "Chapadlo (ustupující)", "Chapadlem (ustupujícím)"},
    {2, "Щупальце (исчезает)", "Щупальца(исчезает)", "Щупальце(исчезает)", "Щупальцем(исчезает)"},
    {0, "Fliehende Tentakel", "Fliehenden Tentakel", "Fliehende Tentakel", "Fliehende Tentakel"},
    {0, "Tentáculo (Retirando-se)", "Tentáculos (Retirando-se)"/*MISSING*/ , "Tentacle (withdrawing)", "Tentáculo (Retirando-se)"},
    }},
  {0xdc707305, 0, { // ""Temple of Cthulhu""
    {1, "Świątynia Cthulhu", "Świątynia Cthulhu", "Świątynię Cthulhu", "w Świątyni Cthulhu"},
    {1, "Cthulhu'nun Tapınağı", "Cthulhu'nun Tapınakları", "Cthulhu'nun Tapınağını", "Cthulhu'nun Tapınağında"},
    {3, "Cthulhuův chrám", "Cthulhuovy chrámy", "Cthulhuův chrám", "v Cthulhuově chrámu"},
    {3, "Храм Ктулху", "Храмы Ктулху", "Храм Ктулху", "в Храме Ктулху"},
    {0, "Tempel des Cthulhu", "Tempel des Cthulhu", "im Tempel des Cthulhu", "im Tempel des Cthulhu"},
    {0/*MISSING*/ , "Temple of Cthulhu"/*MISSING*/ , "Temple of Cthulhu"/*MISSING*/ , "Temple of Cthulhu"/*MISSING*/ , "Temple of Cthulhu"},
    }},
  {0xdf01075f, 0, { // ""Cloud of Mirage""
    {1, "Chmura Mirażowa", "Chmury Mirażowe", "Chmurę Mirażową", "Chmurą Mirażową"},
    {1, "Bulut Yansısı", "Bulut Yansıları", "Bulut Yansısını", "Bulut Yansısıyla"},
    {3, "Oblak přeludů", "Oblaky přeludů", "Oblak přeludů", "Oblakem přeludů"},
    {2, "Облако миражей", "Облака миражей", "Облако миражей", "Облаком миражей"},
    {1, "Illusionswolke", "Illusionswolken", "Illusionswolke", "Illusionswolke"},
    {1, "Nuvem de Miragem", "Nuvens de Miragens"/*MISSING*/ , "Cloud of Mirage", "Nuvem de Miragem"},
    }},
  {0xdfd173e9, 0, { // ""Illusion""
    {1, "Iluzja", "Iluzje", "Iluzję", "Iluzją"},
    {1, "İlüzyon", "İlüzyonlar", "İlüzyonu", "İlüzyonla"},
    {1, "Iluze", "Iluze", "Iluzi", "Iluzí"},
    {1, "Иллюзия", "Иллюзии", "Иллюзию", "Иллюзией"},
    {1, "Illusion", "Illusionen", "Illusion", "Illusion"},
    {0/*MISSING*/ , "Illusion"/*MISSING*/ , "Illusion"/*MISSING*/ , "Illusion"/*MISSING*/ , "Illusion"},
    }},
  {0xe042e7a8, 0, { // ""Bomberbird""
    {3, "Bombardier", "Bombardiery", "Bombardiera", "Bombardierem"},
    {3, "Bombacıkuş", "Bombacıkuşlar", "Bombacıkuşunu", "Bombacıkuşuyla"},
    {0, "Bombarďák", "Bombarďáci", "Bombarďáka", "Bombarďákem"},
    {0, "Бомбардир", "Бомбардиры", "Бомбардира", "Бомбардиром"},
    {0, "Bombenvogel", "Bombenvögel", "Bombenvogel", "Bombenvogel"},
    {0/*MISSING*/ , "Bomberbird"/*MISSING*/ , "Bomberbird"/*MISSING*/ , "Bomberbird"/*MISSING*/ , "Bomberbird"},
    }},
  {0xe13510da, 0, { // ""vine""
    {1, "winorośl", "winorośle", "winorośl", "winoroślą"},
    {1, "şarap", "şaraplar", "şarabı", "şarapla"},
    {1, "réva", "réva", "révu", "révou"},
    {1, "лоза", "лозы", "лозы", "лозой"},
    {1, "Rebe", "Reben", "Rebe", "Rebe"},
    {0/*MISSING*/ , "vine"/*MISSING*/ , "vine"/*MISSING*/ , "vine"/*MISSING*/ , "vine"},
    }},
  {0xe16ebb5e, 0, { // ""Tentacle Tail""
    {1, "Ogon Macki", "Ogony Macki", "Ogon Macki", "Ogonem Macki"},
    {0, "Dokunaç Devamı", "Dokunaç Devamları", "Dokunaç Devamını", "Dokunaç Devamıyla"},
    {2, "Chapadlo", "Chapadla", "Chapadlo", "Chapadlem"},
    {3, "Хвост щупальца", "Хвосты щупальца", "Хвост щупальца", "Хвостом щупальца"},
    {0, "Tentakel Schweif", "Tentakel Schweif", "Tentakel Schweif", "Tentakel Schweif"},
    {1, "Cauda de Tentáculo", "Caudas de Tentáculos"/*MISSING*/ , "Tentacle Tail", "Cauda de Tentáculo"},
    }},
  {0xe29ae5d6, 0, { // ""Ivory Figurine""
    {1, "Figurka", "Figurki", "Figurkę", "Figurką"},
    {1, "Fildişi Biblo", "Fildişi Biblolar", "Fildişi Bibloyu", "Fildişi Bibloyla"},
    {1, "Figurka", "Figurky", "Figurku", "Figurkou"},
    {1, "Фигурка", "Фигурки", "Фигурку", "Фигуркой"},
    {1, "Elfenbeinfigur", "Elfenbeinfiguren", "Elfenbeinfigur", "Elfenbeinfigur"},
    {0/*MISSING*/ , "Ivory Figurine"/*MISSING*/ , "Ivory Figurine"/*MISSING*/ , "Ivory Figurine"/*MISSING*/ , "Ivory Figurine"},
    }},
  {0xe388cddf, 0, { // ""lake of sulphur""
    {2, "jezioro siarki", "jeziora siarki", "jezioro siarki", "jeziorem siarki"},
    {2, "Sülfür Gölü", "Sülfür Gölleri", "Sülfür Gölünü", "Sülfür Gölünde"},
    {2, "sirné jezírko", "sirná jezírka", "sirné jezírko", "sirným jezírkem"},
    {2, "озеро серы", "озёра серы", "озеро серы", "озером серы"},
    {0, "Schwefelsee", "Schwefelseen", "Schwefelsee", "Schwefelsee"},
    {0, "lago de enxofre", "lagos de enxofre"/*MISSING*/ , "lake of sulphur", "lago de enxofre"},
    }},
  {0xe4a5cfa5, 0, { // ""Rock Snake""
    {0, "Skalny Wąż", "Skalne Węże", "Skalnego Węża", "Skalnym Wężem"},
    {0, "Kaya Yılanı", "Kaya Yılanları", "Kaya Yılanını", "Kaya Yılanıyla"},
    {0, "Kamenný had", "Kamenní hadi", "Kamenného hada", "Kamenným hadem"},
    {1, "Каменная змея", "Каменные змеи", "Каменную змею", "Каменной змеёй"},
    {1, "Felsschlange", "Felsschlangen", "Felsschlange", "Felsschlange"},
    {0/*MISSING*/ , "Rock Snake"/*MISSING*/ , "Rock Snake"/*MISSING*/ , "Rock Snake"/*MISSING*/ , "Rock Snake"},
    }},
  {0xe57d27eb, 0, { // ""Rose Lady""
    {1, "Różana Dama", "Różane Damy", "Różaną Damę", "Różaną Damą"},
    {0/*MISSING*/ , "Rose Lady"/*MISSING*/ , "Rose Lady"/*MISSING*/ , "Rose Lady"/*MISSING*/ , "Rose Lady"},
    {1, "Růžová paní", "Růžové paní", "Růžovou dámu", "Růžovou paní"},
    {1, "Розовая дама", "Розовые дамы", "Розовую даму", "Розовой дамой"},
    {1, "Rosendame", "Rosendamen", "Rosendame", "Rosendame"},
    {0/*MISSING*/ , "Rose Lady"/*MISSING*/ , "Rose Lady"/*MISSING*/ , "Rose Lady"/*MISSING*/ , "Rose Lady"},
    }},
  {0xe6a497fc, 0, { // ""Herd Bull""
    {0, "Stadny Byk", "Stadne Byki", "Stadnego Byka", "Stadnym Bykiem"},
    {0/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"},
    {0, "Stádový Býk", "Stádoví Býci", "Stádového Býka", "Stádovým Býkem"},
    {0, "Стадный бык", "Стадные быки", "Стадного быка", "Стадным быком"},
    {0/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"},
    {0/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"/*MISSING*/ , "Herd Bull"},
    }},
  {0xe6de3079, 0, { // ""Vine Beast""
    {0, "Winny Stwór", "Winne Stwory", "Winnego Stwora", "Winnym Stworem"},
    {0, "Şarap Canavarı", "Şarap Canavarları", "Şarap Canavarını", "Şarap Canavarıyla"},
    {0, "Révový netvor", "Révoví netvoři", "Révového netvora", "Révovým netvorem"},
    {0, "Винный зверь", "Винные звери", "Винного зверя", "Винным зверем"},
    {1, "Rebenbestie", "Rebenbestien", "Rebenbestie", "Rebenbestie"},
    {0/*MISSING*/ , "Vine Beast"/*MISSING*/ , "Vine Beast"/*MISSING*/ , "Vine Beast"/*MISSING*/ , "Vine Beast"},
    }},
  {0xe890513d, 0, { // ""Seep""
    {0, "Sączak", "Sączaki", "Sączaka", "Sączakiem"},
    {0, "Sızıntı", "Sızıntılar", "Sızıntıyı", "Sızıntıyla"},
    {0, "Průsak", "Průsakové", "Průsaka", "Průsakem"},
    {0, "Горный призрак", "Горные призраки", "Горного призрака", "Горным призраком"},
    {0, "Tropfer", "Tropfer", "Tropfer", "Tropfer"},
    {0, "Infiltrado", "Infiltrados"/*MISSING*/ , "Seep", "Infiltrado"},
    }},
  {0xe92c565d, 0, { // ""Orb of Winter""
    {1, "Sfera Zimy", "Sfery Zimy", "Sferę Zimy", "Sferą Zimy"},
    {1, "Kış Küresi", "Kış Küreleri", "Kış Küresini", "Kış Küresiyle"},
    {1, "Sféra Zimy", "Sféry Zimy", "Sféru Zimy", "Sférou Zimy"},
    {1, "Сфера Зимы", "Сферы Зимы", "Сферу Зимы", "Сферой Зимы"},
    {0, "Orb des Winters", "Orbs des Winters", "Orb des Winters", "Orb des Winters"},
    {0, "Orbe de Inverno", "Orbes de Inverno"/*MISSING*/ , "Orb of Winter", "Orbe de Inverno"},
    }},
  {0xea5a8f39, 0, { // ""Dead Cave""
    {1, "Martwa Jaskinia", "Martwa Jaskinia", "Martwą Jaskinię", "w Martwej Jaskini"},
    {1, "Ölü Mağara", "Ölü Mağaralar", "Ölü Mağarayı", "Ölü Mağarada"},
    {1, "Mrtvá jeskyně", "Mrtvé jeskyně", "Mrtvou jeskyni", "v Mrtvé jeskyni"},
    {1, "Мёртвая пещера", "Мёртвые пещеры", "Мёртвую пещеру", "в Мёртвой пещере"},
    {1, "Tote Höhle", "Tote Höhlen", "in der Toten Höhle", "in der Toten Höhle"},
    {0/*MISSING*/ , "Dead Cave"/*MISSING*/ , "Dead Cave"/*MISSING*/ , "Dead Cave"/*MISSING*/ , "Dead Cave"},
    }},
  {0xea683a5f, 0, { // ""Orb of Invisibility""
    {1, "Sfera Niewidzialności", "Sfery Niewidzialności", "Sferę Niewidzialności", "Sferą Niewidzialności"},
    {1, "Görünmezlik Küresi", "Görünmezlik Küreleri", "Görünmezlik Küresini", "Görünmezlik Küresiyle"},
    {1, "Sféra Neviditelnosti", "Sféry Neviditelnosti", "Sféru Neviditelnosti", "Sférou Neviditelnosti"},
    {1, "Сфера Невидимости", "Сферы Невидимости", "Сферу Невидимости", "Сферой Невидимости"},
    {0, "Orb der Transparenz", "Orbs der Transparenz", "Orb der Transparenz", "Orb der Transparenz"},
    {0/*MISSING*/ , "Orb of Invisibility"/*MISSING*/ , "Orb of Invisibility"/*MISSING*/ , "Orb of Invisibility"/*MISSING*/ , "Orb of Invisibility"},
    }},
  {0xea69ead7, 0, { // ""Sunken Treasure""
    {3, "Zatopiony Skarb", "Zatopione Skarby", "Zatopiony Skarb", "Zatopionym Skarbem"},
    {0/*MISSING*/ , "Sunken Treasure"/*MISSING*/ , "Sunken Treasure"/*MISSING*/ , "Sunken Treasure"/*MISSING*/ , "Sunken Treasure"},
    {3, "Potopený poklad", "Potopené poklady", "Potopený poklad", "Potopeným pokladem"},
    {3, "Затонувшее сокровище", "Затонувшие сокровища", "Затонувшее сокровище", "Затонувшим сокровищем"},
    {0, "Versunkener Schatz", "Versunkene Schätze", "Versunkenen Schatz", "Versunkenen Schatz"},
    {0/*MISSING*/ , "Sunken Treasure"/*MISSING*/ , "Sunken Treasure"/*MISSING*/ , "Sunken Treasure"/*MISSING*/ , "Sunken Treasure"},
    }},
  {0xebfa53de, 0, { // ""chasm""
    {2, "przepaść", "przepaści", "przepaść", "przepaścią"},
    {2, "Uçurum", "Uçurumlar", "Uçurumu", "Uçurumda"},
    {1, "propast", "propasti", "propast", "propastí"},
    {1, "пропасть", "пропасти", "пропасть", "пропастью"},
    {0, "Abgrund", "Abgründe", "Abgrund", "Abgrund"},
    {0, "abismo", "abismos"/*MISSING*/ , "chasm", "abismo"},
    }},
  {0xec3d06d9, 0, { // ""Orb of Matter""
    {1, "Sfera Materii", "Sfery Materii", "Sferę Materii", "Sferą Materii"},
    {1, "Madde Küresi", "Madde Küreleri", "Madde Küresini", "Madde Küresiyle"},
    {1, "Sféra Hmoty", "Sféry Hmoty", "Sféru Hmoty", "Sférou Hmoty"},
    {1, "Сфера Вещества", "Сферы Вещества", "Сферу Вещества", "Сферой Вещества"},
    {0, "Orb der Materie", "Orbs der Materie", "Orb der Materie", "Orb der Materie"},
    {0/*MISSING*/ , "Orb of Matter"/*MISSING*/ , "Orb of Matter"/*MISSING*/ , "Orb of Matter"/*MISSING*/ , "Orb of Matter"},
    }},
  {0xec821dae, 0, { // ""Fjord Troll""
    {0, "Fjordtroll", "Fjordtrolle", "Fjordtrolla", "Fjordtrollem"},
    {0, "Fiyort Trolü", "Fiyort Trolleri", "Fiyort Trolünü", "Fiyort Trolüyle"},
    {0, "Fjordtroll", "Fjordtrollové", "Fjordtrolla", "Fjordtrollem"},
    {0, "Тролль фьорда", "Тролли фьорда", "Тролля фьорда", "Троллем фьорда"},
    {0, "Fjord Troll", "Fjord Trolle", "Fjord Troll", "Fjord Troll"},
    {0/*MISSING*/ , "Fjord Troll"/*MISSING*/ , "Fjord Troll"/*MISSING*/ , "Fjord Troll"/*MISSING*/ , "Fjord Troll"},
    }},
  {0xeca4511f, 0, { // ""Slime Beast""
    {0, "Mazisty Stwór", "Maziste Stwory", "Mazistego Stwora", "Mazistym Stworem"},
    {1, "Balçık Canavarı", "Balçık Canavarları", "Balçık Canavarını", "Balçık Canavarıyla"},
    {0, "Slizový netvor", "Slizoví netvoři", "Slizového netvora", "Slizovým netvorem"},
    {1, "Живая слизь", "Живая слизь", "Живую слизь", "Живой слизью"},
    {1, "Schleimbestie", "Schleimbestien", "Schleimbestie", "Schleimbestie"},
    {1, "Besta de Lodo", "Bestas de Lodo"/*MISSING*/ , "Slime Beast", "Besta de Lodo"},
    }},
  {0xecbb1f9a, 0, { // ""Dragon Chasms""
    {1, "Smocze Otchłanie", "Smocze Otchłanie", "Smocze Otchłanie", "w Smoczych Otchłaniach"},
    {0/*MISSING*/ , "Dragon Chasms"/*MISSING*/ , "Dragon Chasms"/*MISSING*/ , "Dragon Chasms"/*MISSING*/ , "Dragon Chasms"},
    {1, "Dračí propasti", "Dračí propasti", "Dračí propasti", "v Dračích propastech"},
    {1, "Драконья бездна", "Драконьи бездны", "Драконью бездну", "в Драконьей бездне"},
    {1, "Drachenschluchten", "Drachenschluchten", "in den Drachenschluchten", "in den Drachenschluchten"},
    {0/*MISSING*/ , "Dragon Chasms"/*MISSING*/ , "Dragon Chasms"/*MISSING*/ , "Dragon Chasms"/*MISSING*/ , "Dragon Chasms"},
    }},
  {0xed9fa709, 0, { // ""Yeti""
    {0, "Yeti", "Yeti", "Yeti", "Yeti"},
    {0, "Yeti", "Yeti", "Yeti", "Yeti"},
    {0, "Yeti", "Yetiové", "Yetiho", "Yetim"},
    {0, "Йети", "Йети", "Йети", "Йети"},
    {0, "Yeti", "Yetis", "Yeti", "Yeti"},
    {0, "Yeti", "Yetis"/*MISSING*/ , "Yeti", "Yeti"},
    }},
  {0xede91476, 0, { // ""Black Lotus""
    {3, "Czarny Lotos", "Czarne Lotosy", "Czarny Lotos", "Czarnym Lotosem"},
    {0/*MISSING*/ , "Black Lotus"/*MISSING*/ , "Black Lotus"/*MISSING*/ , "Black Lotus"/*MISSING*/ , "Black Lotus"},
    {3, "Černý lotos", "Černé lotosy", "Černý lotos", "Černým lotosem"},
    {3, "Чёрный лотос", "Чёрные лотосы", "Чёрный лотос", "Чёрным лотосом"},
    {0, "Schwarzer Lotus", "Schwarze Lotus", "Schwarzen Lotus", "Schwarzen Lotus"},
    {0/*MISSING*/ , "Black Lotus"/*MISSING*/ , "Black Lotus"/*MISSING*/ , "Black Lotus"/*MISSING*/ , "Black Lotus"},
    }},
  {0xee28d696, 0, { // ""Spice""
    {1, "Przyprawa", "Przyprawy", "Przyprawę", "Przyprawą"},
    {1, "Baharat", "Baharatlar", "Baharatı", "Baharatla"},
    {2, "Koření", "Koření", "Koření", "Kořením"},
    {1, "Пряность", "Пряности", "Пряность", "Пряностью"},
    {2, "Spice", "Spice", "Spice", "Spice"},
    {1, "Especiaria", "Especiarias"/*MISSING*/ , "Spice", "Especiaria"},
    }},
  {0xee6e1823, 0, { // ""Giant Ape""
    {1, "Wielka Małpa", "Wielkie Małpy", "Wielką Małpę", "Wielką Małpą"},
    {1, "Dev Maymun", "Dev Maymunlar", "Dev Maymunu", "Dev Maymunla"},
    {1, "Obří opice", "Obří opice", "Obří opici", "Obří opicí"},
    {1, "Огромная обезьяна", "Огромные обезьяны", "Огромную обезьяну", "Огромной обезьяной"},
    {0, "Riesenaffe", "Riesenaffen", "Riesenaffen", "Riesenaffen"},
    {0, "Macaco Gigante", "Macacos Gigantes"/*MISSING*/ , "Giant Ape", "Macaco Gigante"},
    }},
  {0xee987540, 0, { // ""Air Shard""
    {3, "Okruch Powietrza", "Okruchy Powietrza", "Okruch Powietrza", "Okruchem Powietrza"},
    {0, "Hava Parçası", "Hava Parçaları", "Hava Parçasını", "Hava Parçasıyla"},
    {3, "Úlomek Vzduchu", "Úlomky Vzduchu", "Úlomek Vzduchu", "Úlomkem Vzduchu"},
    {3, "Осколок Воздуха", "Осколки Воздуха", "Осколок Воздуха", "Осколком Воздуха"},
    {1, "Luftscherbe", "Luftscherben", "Luftscherbe", "Luftscherbe"},
    {0/*MISSING*/ , "Air Shard"/*MISSING*/ , "Air Shard"/*MISSING*/ , "Air Shard"/*MISSING*/ , "Air Shard"},
    }},
  {0xeeca94f9, 0, { // ""Sparrowhawk""
    {0, "Krogulec", "Krogulce", "Krogulca", "Krogulcem"},
    {0/*MISSING*/ , "Sparrowhawk"/*MISSING*/ , "Sparrowhawk"/*MISSING*/ , "Sparrowhawk"/*MISSING*/ , "Sparrowhawk"},
    {0, "Krahujec", "Krahujci", "Krahujce", "Krahujcem"},
    {0, "Ястреб", "Ястребы", "Ястреба", "Ястребом"},
    {0, "Sperber", "Sperber", "Sperber", "Sperber"},
    {0/*MISSING*/ , "Sparrowhawk"/*MISSING*/ , "Sparrowhawk"/*MISSING*/ , "Sparrowhawk"/*MISSING*/ , "Sparrowhawk"},
    }},
  {0xef45398d, 0, { // ""Narcissist""
    {0, "Narcyz", "Narcyzy", "Narcyza", "Narcyzem"},
    {0/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"},
    {0, "Narcis", "Narcisové", "Narcisa", "Narcisem"},
    {0, "Нарциссист", "Нарциссисты", "Нарциссиста", "Нарциссистом"},
    {0/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"},
    {0/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"/*MISSING*/ , "Narcissist"},
    }},
  {0xef8d54c7, 0, { // ""Graveyard""
    {3, "Cmentarz", "Cmentarz", "Cmentarz", "na Cmentarzu"},
    {3, "Mezarlık", "Mezarlıklar", "Mezarlığı", "Mezarlıkta"},
    {3, "Hřbitov", "Hřbitovy", "Hřbitov", "na Hřbitově"},
    {2, "Кладбище", "Кладбища", "Кладбище", "на Кладбище"},
    {0, "Friedhof", "Friedhöfe", "auf dem Friedhof", "auf dem Friedhof"},
    {0, "Cemitério", "Cemitérios"/*MISSING*/ , "Graveyard", "Cemitério"},
    }},
  {0xeff05e04, 0, { // ""Living Cave""
    {1, "Żywa Jaskinia", "Żywa Jaskinia", "Żywą Jaskinię", "w Żywej Jaskini"},
    {1, "Yaşayan Mağara", "Yaşayan Mağaralar", "Yaşayan Mağarayı", "Yaşayan Mağarada"},
    {1, "Živoucí jeskyně", "Živoucí jeskyně", "Živoucí jeskyni", "v Živoucí jeskyni"},
    {1, "Живая Пещера", "Живые Пещеры", "Живую Пещеру", "в Живой Пещере"},
    {1, "Lebendige Höhle", "Lebendige Höhlen", "in der Lebendigen Höhle", "in der Lebendigen Höhle"},
    {1, "Caverna Viva", "Cavernas Vivas"/*MISSING*/ , "Living Cave", "Caverna Viva"},
    }},
  {0xf086bced, 0, { // ""Orb of Time""
    {1, "Sfera Czasu", "Sfery Czasu", "Sferę Czasu", "Sferą Czasu"},
    {1, "Zaman Küresi", "Zaman Küreleri", "Zaman Küresini", "Zaman Küresiyle"},
    {1, "Sféra Času", "Sféry Času", "Sféru Času", "Sférou Času"},
    {1, "Сфера Времени", "Сферы Времени", "Сферу Времени", "Сферой Времени"},
    {0, "Orb der Zeit", "Orbs der Zeit", "Orb der Zeit", "Orb der Zeit"},
    {0/*MISSING*/ , "Orb of Time"/*MISSING*/ , "Orb of Time"/*MISSING*/ , "Orb of Time"/*MISSING*/ , "Orb of Time"},
    }},
  {0xf098af77, 0, { // ""Ghost""
    {0, "Duch", "Duchy", "Ducha", "Duchem"},
    {0, "Hayalet", "Hayaletler", "Hayaleti", "Hayaletle"},
    {0, "Duch", "Duchové", "Ducha", "Duchem"},
    {0, "Дух", "Духи", "Духа", "Духом"},
    {0, "Geist", "Geister", "Geist", "Geist"},
    {0, "Fantasma", "Fantasmas"/*MISSING*/ , "Ghost", "Fantasmas"},
    }},
  {0xf0cdc208, 0, { // ""boat""
    {1, "łódka", "łódki", "łódkę", "łódką"},
    {1, "kayık", "kayıklar", "kayığı", "kayıkla"},
    {3, "člun", "čluny", "člun", "člunem"},
    {1, "лодка", "лодки", "лодку", "лодкой"},
    {2, "Boot", "Boote", "Boot", "Boot"},
    {0/*MISSING*/ , "boat"/*MISSING*/ , "boat"/*MISSING*/ , "boat"/*MISSING*/ , "boat"},
    }},
  {0xf0e0b88f, 0, { // ""Mutant Ivy""
    {3, "Zmutowany Bluszcz", "Zmutowane Bluszcze", "Zmutowany Bluszcz", "Zmutowanym Bluszczem"},
    {3, "Mutant Sarmaşık", "Mutant Sarmaşıklar", "Mutant Sarmaşığı", "Mutant Sarmaşıkla"},
    {3, "Mutantní břečťan", "Mutantní břečťany", "Mutantní břečťan", "Mutantním břečťanem"},
    {3, "Плющ-мутант", "Плющи-мутанты", "Плюш-мутант", "Плющом-мутантом"},
    {0, "Mutierter Efeu", "Mutierte Efeue", "Mutierten Efeu", "Mutierten Efeu"},
    {0/*MISSING*/ , "Mutant Ivy"/*MISSING*/ , "Mutant Ivy"/*MISSING*/ , "Mutant Ivy"/*MISSING*/ , "Mutant Ivy"},
    }},
  {0xf125c9ed, 0, { // ""temporary floor""
    {1, "tymczasowa podłoga", "tymczasowe podłogi", "tymczasową podłogę", "tymczasową podłogą"},
    {1, "geçici zemin", "geçici zeminler", "geçici zemini", "geçici zeminle"},
    {1, "dočasná podlaha", "dočasné podlahy", "dočasnou podlahu", "dočasnou podlahou"},
    {3, "временный пол", "временные полы", "временный пол", "временным полом"},
    {0, "temporärer Boden", "temporäre Böden", "temporären Boden", "temporären Boden"},
    {0/*MISSING*/ , "temporary floor"/*MISSING*/ , "temporary floor"/*MISSING*/ , "temporary floor"/*MISSING*/ , "temporary floor"},
    }},
  {0xf28de481, 0, { // ""Orb of Luck""
    {1, "Sfera Szczęścia", "Sfery Szczęścia", "Sferę Szczęścia", "Sferą Szczęścia"},
    {1, "Şans Küresi", "Şans Küreleri", "Şans Küresini", "Şans Küresiyle"},
    {1, "Sféra Štěstí", "Sféry Štěstí", "Sféru Štěstí", "Sférou Štěstí"},
    {1, "Сфера Удачи", "Сферы Удачи", "Сферу Удачи", "Сферой Удачи"},
    {0, "Orb des Glücks", "Orbs des Glücks", "Orb des Glücks", "Orb des Glücks"},
    {0/*MISSING*/ , "Orb of Luck"/*MISSING*/ , "Orb of Luck"/*MISSING*/ , "Orb of Luck"/*MISSING*/ , "Orb of Luck"},
    }},
  {0xf2f1046e, 0, { // ""Blue Hyperbug""
    {0, "Niebieski Hiperinsekt", "Niebieskie Hiperinsekty", "Niebieskiego Hiperinsekta", "Niebieskim Hiperinsektem"},
    {0, "Mavi Aşkınböcek", "Mavi Aşkınböcekler", "Mavi Aşkınböceği", "Mavi Aşkınböcekle"},
    {0, "Modrý hyperhmyzák", "Modří hyperhmyzáci", "Modrého hyperhmyzáka", "Modrým hyperhmyzákem"},
    {0, "Синий гипержук", "Синие гипержуки", "Синего гипержука", "Синим гипержуком"},
    {0, "Blauer Hyperkäfer", "Blaue Hyperkäfer", "Blauen Hyperkäfer", "Blauen Hyperkäfer"},
    {0/*MISSING*/ , "Blue Hyperbug"/*MISSING*/ , "Blue Hyperbug"/*MISSING*/ , "Blue Hyperbug"/*MISSING*/ , "Blue Hyperbug"},
    }},
  {0xf342c876, 0, { // ""Golden Egg""
    {2, "Złote Jajo", "Złote Jaja", "Złote Jajo", "Złotym Jajem"},
    {0/*MISSING*/ , "Golden Egg"/*MISSING*/ , "Golden Egg"/*MISSING*/ , "Golden Egg"/*MISSING*/ , "Golden Egg"},
    {2, "Zlaté vejce", "Zlatá vejce", "Zlaté vejce", "Zlatým vejcem"},
    {2, "Золотое яйцо", "Золотые яйца", "Золотое яйцо", "Золотым яйцом"},
    {2, "Goldenes Ei", "Goldene Eier", "Goldene Ei", "Goldene Ei"},
    {0/*MISSING*/ , "Golden Egg"/*MISSING*/ , "Golden Egg"/*MISSING*/ , "Golden Egg"/*MISSING*/ , "Golden Egg"},
    }},
  {0xf358ed0c, 0, { // ""Elixir of Life""
    {3, "Eliksir Życia", "Eliksiry Życia", "Eliksir Życia", "Eliksirem Życia"},
    {3, "Yaşam İksiri", "Yaşam İksirleri", "Yaşam İksirini", "Yaşam İksiriyle"},
    {3, "Elixír života", "Elixíry života", "Elixír života", "Elixírem života"},
    {3, "Эликсир жизни", "Эликсиры жизни", "Эликсир жизни", "Эликсиром жизни"},
    {2, "Lebenselixir", "Lebenselixir", "Lebenselixir", "Lebenselixir"},
    {0, "Elixir", "Elixires"/*MISSING*/ , "Elixir of Life", "Elixir"},
    }},
  {0xf3926b40, 0, { // ""Orb of Vaulting""
    {1, "Sfera Przeskoku", "Sfery Przeskoku", "Sferę Przeskoku", "Sferą Przeskoku"},
    {0/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"},
    {1, "Sféra Přeskoku", "Sféry Přeskoku", "Sféru Přeskoku", "Sférou Přeskoku"},
    {1, "Сфера Рывка", "Сферы Рывка", "Сферу Рывка", "Сферой Рывка"},
    {0/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"},
    {0/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"/*MISSING*/ , "Orb of Vaulting"},
    }},
  {0xf3b4b660, 0, { // ""Red Gem""
    {3, "Czerwony Kamień", "Czerwone Kamienie", "Czerwony Kamień", "Czerwonym Kamieniem"},
    {0, "Kızıl Mücevher", "Kızıl Mücevherler", "Kızıl Mücevheri", "Kızıl Mücevherle"},
    {3, "Červený drahokam", "Červené drahokamy", "Červený drahokam", "Červeným drahokamem"},
    {3, "Красный камень", "Красные камни", "Красный камень", "Красным камнем"},
    {0, "Roter Stein", "Rote Steine", "Roten Stein", "Roten Stein"},
    {0/*MISSING*/ , "Red Gem"/*MISSING*/ , "Red Gem"/*MISSING*/ , "Red Gem"/*MISSING*/ , "Red Gem"},
    }},
  {0xf3c25ba9, 0, { // ""strong wind""
    {3, "silny wiatr", "silne wiatry", "silny wiatr", "silnym wiatrem"},
    {0/*MISSING*/ , "strong wind"/*MISSING*/ , "strong wind"/*MISSING*/ , "strong wind"/*MISSING*/ , "strong wind"},
    {3, "silný vítr", "silné větry", "silný vítr", "silným větrem"},
    {3, "сильный ветер", "сильные ветры", "сильный ветер", "сильным ветром"},
    {0, "starker Wind", "starke Winde", "starken Wind", "starken Wind"},
    {0/*MISSING*/ , "strong wind"/*MISSING*/ , "strong wind"/*MISSING*/ , "strong wind"/*MISSING*/ , "strong wind"},
    }},
  {0xf3f23cda, 0, { // ""Orb of Air""
    {1, "Sfera Powietrza", "Sfery Powietrza", "Sferę Powietrza", "Sferą Powietrza"},
    {1, "Hava Küresi", "Hava Küreleri", "Hava Küresini", "Hava Küresiyle"},
    {1, "Sféra Vzduchu", "Sféry Vzduchu", "Sféru Vzduchu", "Sférou Vzduchu"},
    {1, "Сфера Воздуха", "Сферы Воздуха", "Сферу Воздуха", "Сферой Воздуха"},
    {0, "Orb der Luft", "Orbs der Luft", "Orb der Luft", "Orb der Luft"},
    {0/*MISSING*/ , "Orb of Air"/*MISSING*/ , "Orb of Air"/*MISSING*/ , "Orb of Air"/*MISSING*/ , "Orb of Air"},
    }},
  {0xf505646d, 0, { // ""Palace Guard""
    {0, "Strażnik Pałacu", "Strażnicy Pałacu", "Strażnika Pałacu", "Strażnikiem Pałacu"},
    {0, "Saray Muhafızı", "Saray Muhafızları", "Saray Muhafızını", "Saray Muhafızıyla"},
    {0, "Strážce Paláce", "Strážci Paláce", "Strážce Paláce", "Strážcem Paláce"},
    {0, "Дворцовый стражник", "Дворцовые стражники", "Дворцового стражника", "Дворцовым стражником"},
    {1, "Palastwache", "Palastwachen", "Palastwache", "Palastwache"},
    {0/*MISSING*/ , "Palace Guard"/*MISSING*/ , "Palace Guard"/*MISSING*/ , "Palace Guard"/*MISSING*/ , "Palace Guard"},
    }},
  {0xf6325a4d, 0, { // ""Wild West""
    {3, "Dziki Zachód", "Dzikie Zachody", "Dziki Zachód", "na Dzikim Zachodzie"},
    {3, "Vahşi Batı", "Vahşi Batılar", "Vahşi Batıyı", "Vahşi Batıda"},
    {3, "Divoký západ", "Divoké západy", "Divoký západ", "na Divokém západě"},
    {3, "Дикий Запад", "Дикие Запады", "Дикий Запад", "на Диком Западе"},
    {0, "Wilder Westen", "Wilde Westen", "im Wilden Westen", "im Wilden Westen"},
    {0/*MISSING*/ , "Wild West"/*MISSING*/ , "Wild West"/*MISSING*/ , "Wild West"/*MISSING*/ , "Wild West"},
    }},
  {0xf6d7d83e, 0, { // ""Warped Sea""
    {1, "Zakrzywione Morze", "Zakrzywione Morza", "Zakrzywione Morze", "na Zakrzywionym Morzu"},
    {0/*MISSING*/ , "Warped Sea"/*MISSING*/ , "Warped Sea"/*MISSING*/ , "Warped Sea"/*MISSING*/ , "Warped Sea"},
    {1, "Pokřivené moře", "Pokřivená moře", "Pokřivené moře", "na Pokřiveném moři"},
    {2, "Искривлённое море", "Искривлённые моря", "Искривлённое море", "в Искривлённом море"},
    {1, "Verzerrte See", "Verzerrten Seen", "auf der Verzerrten See", "auf der Verzerrten See"},
    {0/*MISSING*/ , "Warped Sea"/*MISSING*/ , "Warped Sea"/*MISSING*/ , "Warped Sea"/*MISSING*/ , "Warped Sea"},
    }},
  {0xf6fecf4f, 0, { // ""Flash Witch""
    {1, "Czarownica Błysku", "Czarownice Błysku", "Czarownicę Błysku", "Czarownicą Błysku"},
    {1, "Işık Cadısı", "Işık Cadıları", "Işık Cadısını", "Işık Cadısıyla"},
    {1, "Čarodějnice Záblesku", "Čarodějnice Záblesku", "Čarodějnicí Záblesku", "Čarodějnicí Záblesku"},
    {1, "Ведьма вспышки", "Ведьмы вспышки", "Ведьму вспышки", "Ведьмой вспышки"},
    {1, "Donnerhexe", "Donnerhexen", "Donnerhexe", "Donnerhexe"},
    {0/*MISSING*/ , "Flash Witch"/*MISSING*/ , "Flash Witch"/*MISSING*/ , "Flash Witch"/*MISSING*/ , "Flash Witch"},
    }},
  {0xf78ea3ca, 0, { // ""Magic Mirror""
    {2, "Magiczne Lustro", "Magiczne Lustra", "Magiczne Lustro", "Magicznym Lustrem"},
    {2, "Sihirli Ayna", "Sihirli Aynalar", "Sihirli Aynayı", "Sihirli Aynayla"},
    {2, "Kouzelné zrcadlo", "Kouzelná zrcadla", "Kouzelné zrcadlo", "Kouzelným zrcadlem"},
    {2, "Волшебное зеркало", "Волшебные зеркала", "Волшебное зеркало", "Волшебным зеркалом"},
    {0, "Zauberspiegel", "Zauberspiegel", "Zauberspiegel", "Zauberspiegel"},
    {0, "Espelho Mágico", "Espelhos Mágicos"/*MISSING*/ , "Magic Mirror", "Espelho Mágico"},
    }},
  {0xf7a1921e, 0, { // ""Orb of the Sword""
    {1, "Sfera Ostrza", "Sfery Ostrza", "Sferę Ostrza", "Sferą Ostrza"},
    {0/*MISSING*/ , "Orb of the Sword"/*MISSING*/ , "Orb of the Sword"/*MISSING*/ , "Orb of the Sword"/*MISSING*/ , "Orb of the Sword"},
    {1, "Sféra Meče", "Sféry Meče", "Sféru Meče", "Sférou Meče"},
    {1, "Сфера Меча", "Сферы Меча", "Сферу Меча", "Сферой Меча"},
    {0, "Orb des Schwertes", "Orbs des Schwertes", "Orb des Schwertes", "Orb des Schwertes"},
    {0/*MISSING*/ , "Orb of the Sword"/*MISSING*/ , "Orb of the Sword"/*MISSING*/ , "Orb of the Sword"/*MISSING*/ , "Orb of the Sword"},
    }},
  {0xf7f22746, 0, { // ""False Princess""
    {1, "Fałszywa Księżniczka", "Fałszywe Księżniczki", "Fałszywą Księżniczkę", "Fałszywą Księżniczką"},
    {0/*MISSING*/ , "False Princess"/*MISSING*/ , "False Princess"/*MISSING*/ , "False Princess"/*MISSING*/ , "False Princess"},
    {1, "Falešná princezna", "Falešné princezny", "Falešnou princeznu", "Falešnou princeznou"},
    {1, "Ложная принцесса", "Ложные принцессы", "Ложную принцессу", "Ложной принцессой"},
    {1, "Falsche Prinzessin", "Falsche Prinzessinen", "Falsche Prinzessin", "Falsche Prinzessin"},
    {0/*MISSING*/ , "False Princess"/*MISSING*/ , "False Princess"/*MISSING*/ , "False Princess"/*MISSING*/ , "False Princess"},
    }},
  {0xf8c01ee7, 0, { // ""Goblin""
    {0, "Goblin", "Gobliny", "Goblina", "Goblinem"},
    {0, "Goblin", "Goblinler", "Goblini", "Goblinle"},
    {0, "Goblin", "Goblini", "Goblina", "Goblinem"},
    {0, "Гоблин", "Гоблины", "Гоблина", "Гоблином"},
    {0, "Goblin", "Goblins", "Goblin", "Goblin"},
    {0, "Goblin", "Goblins"/*MISSING*/ , "Goblin", "Goblin"},
    }},
  {0xf928aba7, 0, { // ""ice wall""
    {1, "lodowa ściana", "lodowe ściany", "lodową ścianę", "lodową ścianą"},
    {1, "Buzdan Duvar", "Buzdan Duvarlar", "Buzdan Duvarı", "Buzdan Duvarda"},
    {1, "ledová zeď", "ledové zdi", "ledovou zeď", "ledovou zdí"},
    {1, "ледяная стена", "ледяные стены", "ледяную стену", "ледяной стеной"},
    {1, "Eiswand", "Eiswände", "Eiswand", "Eiswand"},
    {1, "parede de gelo", "paredes de gelo"/*MISSING*/ , "ice wall", "parede de gelo"},
    }},
  {0xf9638f5a, 0, { // ""Mirror Image""
    {2, "Odbicie Lustrzane", "Odbicia Lustrzane", "Odbicie Lustrzane", "Odbiciem Lustrzanym"},
    {2, "Yansıma", "Yansımalar", "Yansımayı", "Yansımayla"},
    {3, "Zrcadlový obraz", "Zrcadlové obrazy", "Zrcadlový obraz", "Zrcadlovým obrazem"},
    {2, "Зеркальное отражение", "Зеркальные отражения", "Зеркальное отражение", "Зеркальным отражением"},
    {2, "Spiegelbild", "Spiegelbilder", "Spiegelbild", "Spiegelbild"},
    {1, "Imagem de Espelho", "Imagens de Espelho"/*MISSING*/ , "Mirror Image", "Imagem de Espelho"},
    }},
  {0xf9db8922, 0, { // ""Compass""
    {0, "Kompas", "Kompasy", "Kompas", "Kompasem"},
    {0, "Pusula", "Pusulalar", "Pusulayı", "Pusulayla"},
    {0, "Kompas", "Kompasy", "Kompas", "Kompasem"},
    {3, "Компас", "Компасы", "Компас", "Компасом"},
    {0, "Kompass", "Kompasse", "Kompass", "Kompass"},
    {0/*MISSING*/ , "Compass"/*MISSING*/ , "Compass"/*MISSING*/ , "Compass"/*MISSING*/ , "Compass"},
    }},
  {0xf9df8cef, 0, { // ""solid branch""
    {1, "solidna gałąź", "solidne gałęzie", "solidną gałąź", "solidną gałęzią"},
    {0/*MISSING*/ , "solid branch"/*MISSING*/ , "solid branch"/*MISSING*/ , "solid branch"/*MISSING*/ , "solid branch"},
    {1, "pevná větev", "pevné větve", "pevnou větev", "pevnou větví"},
    {1, "толстая ветка", "толстые ветки", "толстую ветку", "толстой веткой"},
    {0, "solider Ast", "solide Äste", "soliden Ast", "soliden Ast"},
    {0/*MISSING*/ , "solid branch"/*MISSING*/ , "solid branch"/*MISSING*/ , "solid branch"/*MISSING*/ , "solid branch"},
    }},
  {0xfa0b397e, 0, { // ""Ocean""
    {3, "Ocean", "Oceany", "Ocean", "na Oceanie"},
    {3, "Okyanus", "Okyanuslar", "Okyanusu", "Okyanusla"},
    {3, "Oceán", "Oceány", "Oceán", "na Oceánu"},
    {3, "Океан", "Океаны", "Океан", "в Океане"},
    {0, "Ozean", "Ozeane", "im Ozean", "im Ozean"},
    {0/*MISSING*/ , "Ocean"/*MISSING*/ , "Ocean"/*MISSING*/ , "Ocean"/*MISSING*/ , "Ocean"},
    }},
  {0xfac23af3, 0, { // ""warp gate""
    {1, "wrota zakrzywienia", "wrota zakrzywienia", "wrota zakrzywienia", "wrotami zakrzywienia"},
    {0/*MISSING*/ , "warp gate"/*MISSING*/ , "warp gate"/*MISSING*/ , "warp gate"/*MISSING*/ , "warp gate"},
    {1, "křivá brána", "křivé brány", "křivou bránu", "křivou branou"},
    {1, "врата искривления", "врата искривления", "врата искривления", "вратами искривления"},
    {2, "verzerrtes Tor", "verzerrte Tore", "verzerrte Tor", "verzerrte Tor"},
    {0/*MISSING*/ , "warp gate"/*MISSING*/ , "warp gate"/*MISSING*/ , "warp gate"/*MISSING*/ , "warp gate"},
    }},
  {0xfcb2047b, 0, { // ""Revolver""
    {3, "Rewolwer", "Rewolwery", "Rewolwer", "Rewolwerem"},
    {3, "Altıpatlar", "Altıpatlarlar", "Altıpatları", "Altıpatlarla"},
    {3, "Revolver", "Revolvery", "Revolver", "Revolverem"},
    {3, "Револьвер", "Револьверы", "Револьвер", "Револьвером"},
    {0, "Revolver", "Revolver", "Revolver", "Revolver"},
    {0/*MISSING*/ , "Revolver"/*MISSING*/ , "Revolver"/*MISSING*/ , "Revolver"/*MISSING*/ , "Revolver"},
    }},
  {0xfdadd7ea, 0, { // ""Rich Metal Beast""
    {3, "Bogaty Stwór", "Bogate Stwory", "Bogatego Stwora", "Bogatym Stworem"},
    {3, "Zengin Metal Canavar", "Zengin Metal Canavarları", "Zengin Metal Canavarını", "Zengin Metal Canavarıyla"},
    {1, "Bohatá bestie", "Bohaté bestie", "Bohatou bestii", "Bohatou bestií"},
    {0, "Богатый железный зверь", "Богатые железные звери", "Богатого железного зверя", "Богатым железным зверем"},
    {2, "Ergiebiges Metallbiest", "Ergiebige Metallbiester", "Ergiebige Metallbiest", "Ergiebige Metallbiest"},
    {0/*MISSING*/ , "Rich Metal Beast"/*MISSING*/ , "Rich Metal Beast"/*MISSING*/ , "Rich Metal Beast"/*MISSING*/ , "Rich Metal Beast"},
    }},
  {0xfec6df36, 0, { // ""Flail""
    {3, "Cep", "Cepy", "Cep", "Cepem"},
    {0, "Topuz", "Topuzlar", "Topuzu", "Topuzla"},
    {3, "Řemdih", "Řemdihy", "Řemdih", "Řemdihem"},
    {3, "Цеп", "Цепы", "Цеп", "Цепом"},
    {0, "Morgenstern", "Morgensterne", "Morgenstern", "Morgenstern"},
    {0/*MISSING*/ , "Flail"/*MISSING*/ , "Flail"/*MISSING*/ , "Flail"/*MISSING*/ , "Flail"},
    }},
  {0xff0c4ea1, 0, { // ""Spinel""
    {3, "Spinel", "Spinele", "Spinel", "Spinelem"},
    {0/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"},
    {3, "Spinel", "Spinely", "Spinel", "Spinelem"},
    {1, "Шпинель", "Шпинели", "Шпинель", "Шпинелью"},
    {0/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"},
    {0/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"/*MISSING*/ , "Spinel"},
    }},
  {0xff1eb57a, 0, { // ""Apple""
    {2, "Jabłko", "Jabłka", "Jabłko", "Jabłkiem"},
    {0/*MISSING*/ , "Apple"/*MISSING*/ , "Apple"/*MISSING*/ , "Apple"/*MISSING*/ , "Apple"},
    {2, "Jablko", "Jablka", "Jablko", "Jablkem"},
    {2, "Яблоко", "Яблоки", "Яблоко", "Яблоком"},
    {0, "Apfel", "Äpfel", "Apfel", "Apfel"},
    {0/*MISSING*/ , "Apple"/*MISSING*/ , "Apple"/*MISSING*/ , "Apple"/*MISSING*/ , "Apple"},
    }},
  {0xff6e2c87, 0, { // ""Albatross""
    {3, "Albatros", "Albatrosy", "Albatrosa", "Albatrosem"},
    {3, "Albatros", "Albatroslar", "Albatrosu", "Albatrosla"},
    {3, "Albatros", "Albatrosové", "Albatrosa", "Albatrosem"},
    {0, "Альбатрос", "Альбатросы", "Альбатроса", "Альбатросом"},
    {0, "Albatros", "Albatrosse", "Albatros", "Albatros"},
    {0/*MISSING*/ , "Albatross"/*MISSING*/ , "Albatross"/*MISSING*/ , "Albatross"/*MISSING*/ , "Albatross"},
    }},
  {0xffc7337b, 0, { // ""fan""
    {3, "Wentylator", "Wentylatory", "Wentylator", "Wentylatorem"},
    {0/*MISSING*/ , "fan"/*MISSING*/ , "fan"/*MISSING*/ , "fan"/*MISSING*/ , "fan"},
    {3, "Větrák", "Větráky", "Větrák", "Větrákem"},
    {3, "Вентилятор", "Вентиляторы", "Вентилятор", "Вентилятором"},
    {0, "Ventilator", "Ventilatoren", "Ventilator", "Ventilator"},
    {0/*MISSING*/ , "fan"/*MISSING*/ , "fan"/*MISSING*/ , "fan"/*MISSING*/ , "fan"},
    }},
  };
